Encoding: UTF-8
Package: piecepackr
Type: Package
Title: Board Game Graphics
Version: 1.16.1
Description: Functions to make board game graphics with the 'ggplot2', 'grid', 'rayrender', 'rayvertex', and 'rgl' packages.  Specializes in game diagrams, animations, and "Print & Play" layouts for the 'piecepack' <https://www.ludism.org/ppwiki> but can make graphics for other board game systems.  Includes configurations for several public domain game systems such as checkers, (double-18) dominoes, go, 'piecepack', playing cards, etc.
Authors@R: c(person("Trevor L.", "Davis", role=c("aut", "cre"),
             email="trevor.l.davis@gmail.com",
             comment = c(ORCID = "0000-0001-6341-4639")),
             person("Linux Foundation", role="dtc",
                    comment = 'Uses some data from the "SPDX License List" <https://github.com/spdx/license-list-XML>'),
             person("Delapouite <https://delapouite.com/>", role="ill",
                     comment = 'Meeple shape extracted from "Meeple icon" <https://game-icons.net/1x1/delapouite/meeple.html> / "CC BY 3.0" <https://creativecommons.org/licenses/by/3.0/>'),
             person("Creative Commons", role="ill",
                    comment = '`save_print_and_play()` uses "license badges" from Creative Commons to describe the generated print-and-play file\'s license'))
License: MIT + file LICENSE
URL: https://trevorldavis.com/piecepackr/ (blog),
        https://trevorldavis.com/R/piecepackr/ (pkgdown),
        https://groups.google.com/forum/#!forum/piecepackr (forum)
BugReports: https://github.com/piecepackr/piecepackr/issues
LazyData: true
LazyLoad: yes
Imports: affiner (>= 0.3.1), grid, gridGeometry, grImport2, grDevices,
        purrr, jpeg, png, R6, rlang, stringr, tibble, tools, utils
Suggests: animation (>= 2.7), ggplot2, gifski, gridpattern, magick,
        pdftools, rayrender (>= 0.34.3), rayvertex (>= 0.10.4), readobj
        (>= 0.4.0), rgl (>= 0.106.8), scales (>= 0.5.0), svglite (>=
        2.2.2), systemfonts, testthat, tweenr, vdiffr, xmpdf (>=
        0.1.1), XML (>= 3.99-0.9)
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-12 04:08:55 UTC; trevorld
Author: Trevor L. Davis [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6341-4639>),
  Linux Foundation [dtc] (Uses some data from the "SPDX License List"
    <https://github.com/spdx/license-list-XML>),
  Delapouite <https://delapouite.com/> [ill] (Meeple shape extracted from
    "Meeple icon" <https://game-icons.net/1x1/delapouite/meeple.html> /
    "CC BY 3.0" <https://creativecommons.org/licenses/by/3.0/>),
  Creative Commons [ill] (`save_print_and_play()` uses "license badges"
    from Creative Commons to describe the generated print-and-play
    file's license)
Maintainer: Trevor L. Davis <trevor.l.davis@gmail.com>
Depends: R (>= 2.10)
Repository: CRAN
Date/Publication: 2026-05-12 05:40:02 UTC
Built: R 4.5.2; ; 2026-05-12 06:29:47 UTC; unix

Encoding: UTF-8
Package: piecepackr
Type: Package
Title: Board Game Graphics
Version: 1.16.1
Description: Functions to make board game graphics with the 'ggplot2', 'grid', 'rayrender', 'rayvertex', and 'rgl' packages.  Specializes in game diagrams, animations, and "Print & Play" layouts for the 'piecepack' <https://www.ludism.org/ppwiki> but can make graphics for other board game systems.  Includes configurations for several public domain game systems such as checkers, (double-18) dominoes, go, 'piecepack', playing cards, etc.
Authors@R: c(person("Trevor L.", "Davis", role=c("aut", "cre"),
             email="trevor.l.davis@gmail.com",
             comment = c(ORCID = "0000-0001-6341-4639")),
             person("Linux Foundation", role="dtc",
                    comment = 'Uses some data from the "SPDX License List" <https://github.com/spdx/license-list-XML>'),
             person("Delapouite <https://delapouite.com/>", role="ill",
                     comment = 'Meeple shape extracted from "Meeple icon" <https://game-icons.net/1x1/delapouite/meeple.html> / "CC BY 3.0" <https://creativecommons.org/licenses/by/3.0/>'),
             person("Creative Commons", role="ill",
                    comment = '`save_print_and_play()` uses "license badges" from Creative Commons to describe the generated print-and-play file\'s license'))
License: MIT + file LICENSE
URL: https://trevorldavis.com/piecepackr/ (blog),
        https://trevorldavis.com/R/piecepackr/ (pkgdown),
        https://groups.google.com/forum/#!forum/piecepackr (forum)
BugReports: https://github.com/piecepackr/piecepackr/issues
LazyData: true
LazyLoad: yes
Imports: affiner (>= 0.3.1), grid, gridGeometry, grImport2, grDevices,
        purrr, jpeg, png, R6, rlang, stringr, tibble, tools, utils
Suggests: animation (>= 2.7), ggplot2, gifski, gridpattern, magick,
        pdftools, rayrender (>= 0.34.3), rayvertex (>= 0.10.4), readobj
        (>= 0.4.0), rgl (>= 0.106.8), scales (>= 0.5.0), svglite (>=
        2.2.2), systemfonts, testthat, tweenr, vdiffr, xmpdf (>=
        0.1.1), XML (>= 3.99-0.9)
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-12 04:08:55 UTC; trevorld
Author: Trevor L. Davis [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6341-4639>),
  Linux Foundation [dtc] (Uses some data from the "SPDX License List"
    <https://github.com/spdx/license-list-XML>),
  Delapouite <https://delapouite.com/> [ill] (Meeple shape extracted from
    "Meeple icon" <https://game-icons.net/1x1/delapouite/meeple.html> /
    "CC BY 3.0" <https://creativecommons.org/licenses/by/3.0/>),
  Creative Commons [ill] (`save_print_and_play()` uses "license badges"
    from Creative Commons to describe the generated print-and-play
    file's license)
Maintainer: Trevor L. Davis <trevor.l.davis@gmail.com>
Depends: R (>= 2.10)
Repository: CRAN
Date/Publication: 2026-05-12 05:40:02 UTC
Built: R 4.6.0; ; 2026-05-12 07:08:59 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.list,pp_cfg)
S3method(grobCoords,basic_piece_side)
S3method(grobCoords,basic_projected_token)
S3method(grobCoords,coords_xyl)
S3method(grobCoords,general_projected_token)
S3method(grobCoords,holed_board)
S3method(grobCoords,pmap_piece)
S3method(grobCoords,pp_grobCoords)
S3method(grobCoords,pp_ps_transformation)
S3method(grobCoords,pyramid_top)
S3method(grobPoints,pmap_piece)
S3method(grobPoints,pp_grobCoords)
S3method(makeContent,alquerque_board)
S3method(makeContent,basic_ellipsoid)
S3method(makeContent,basic_grob_with_bleed)
S3method(makeContent,basic_piece_side)
S3method(makeContent,basic_projected_die)
S3method(makeContent,basic_projected_token)
S3method(makeContent,card_side)
S3method(makeContent,checkered_board)
S3method(makeContent,checkers)
S3method(makeContent,d4_die)
S3method(makeContent,domino_side)
S3method(makeContent,face_card_side)
S3method(makeContent,fudge_die)
S3method(makeContent,general_projected_token)
S3method(makeContent,geom_piece_grob)
S3method(makeContent,gridlines)
S3method(makeContent,hexlines)
S3method(makeContent,holed_board)
S3method(makeContent,joker_card_side)
S3method(makeContent,lined_board)
S3method(makeContent,mat)
S3method(makeContent,morris_board)
S3method(makeContent,piece)
S3method(makeContent,pipped)
S3method(makeContent,pp_cropmark)
S3method(makeContent,pp_crosshair)
S3method(makeContent,pp_picture)
S3method(makeContent,pp_polyclip)
S3method(makeContent,pp_ps_transformation)
S3method(makeContent,projected_ellipsoid)
S3method(makeContent,projected_holed_board)
S3method(makeContent,projected_pyramid_side)
S3method(makeContent,projected_pyramid_top)
S3method(makeContent,projected_rpg_die)
S3method(makeContent,pyramid_top)
S3method(makeContext,basic_piece_side)
S3method(makeContext,checkers)
S3method(makeContext,geom_piece_grob)
S3method(makeContext,gridlines)
S3method(makeContext,hexlines)
S3method(makeContext,mat)
S3method(makeContext,pp_cropmark)
export(AA_to_R)
export(R_to_AA)
export(R_x)
export(R_y)
export(R_z)
export(aabb_piece)
export(aes_piece)
export(animate_piece)
export(as_pp_cfg)
export(basicPieceGrob)
export(breaks_counting)
export(circledSegmentsCrosshairGrob)
export(cleave)
export(cropmarkGrob)
export(crosshairGrob)
export(file2grob)
export(game_systems)
export(geom_piece)
export(get_embedded_font)
export(grid.cropmark)
export(grid.crosshair)
export(grid.piece)
export(has_font)
export(inch)
export(install_ppverse)
export(is_color_invisible)
export(is_pp_cfg)
export(label_counting)
export(label_letter)
export(marbles_transform)
export(op_transform)
export(picturePieceGrobFn)
export(piece)
export(piece3d)
export(pieceGrob)
export(piece_mesh)
export(pkgs_ppverse)
export(pmap_piece)
export(pp_cfg)
export(pp_shape)
export(previewLayoutGrob)
export(pyramidTopGrob)
export(render_piece)
export(save_ellipsoid_obj)
export(save_peg_doll_obj)
export(save_piece_images)
export(save_piece_obj)
export(save_print_and_play)
export(scale_x_piece)
export(scale_y_piece)
export(segmentsCrosshairGrob)
export(squaredSegmentsCrosshairGrob)
export(squaresCrosshairGrob)
export(to_degrees)
export(to_hexpack)
export(to_r)
export(to_radians)
export(to_subpack)
export(to_t)
export(to_x)
export(to_y)
import(grid)
importFrom(R6,R6Class)
importFrom(affiner,as_coord2d)
importFrom(affiner,as_coord3d)
importFrom(affiner,as_ellipse2d)
importFrom(affiner,as_plane3d)
importFrom(affiner,as_polygon2d)
importFrom(affiner,as_segment2d)
importFrom(affiner,convex_hull2d)
importFrom(affiner,degrees)
importFrom(affiner,distance2d)
importFrom(affiner,distance3d)
importFrom(affiner,has_overlap2d)
importFrom(affiner,is_angle)
importFrom(affiner,painter_depth)
importFrom(affiner,painter_order)
importFrom(affiner,regular_ngon_polygon2d)
importFrom(grDevices,col2rgb)
importFrom(grDevices,dev.capabilities)
importFrom(grDevices,dev.list)
importFrom(grDevices,dev.new)
importFrom(grDevices,devAskNewPage)
importFrom(grDevices,rgb)
importFrom(rlang,"%||%")
importFrom(rlang,.data)
importFrom(rlang,abort)
importFrom(rlang,check_dots_empty)
importFrom(rlang,inform)
importFrom(rlang,warn)
importFrom(stringr,str_count)
importFrom(stringr,str_glue)
importFrom(stringr,str_pad)
importFrom(stringr,str_split)
importFrom(tibble,tibble)
importFrom(utils,hasName)
importFrom(utils,head)
importFrom(utils,packageDescription)
importFrom(utils,packageVersion)
importFrom(utils,tail)

piecepackr: Board Game Graphics
===============================

[![CRAN Status Badge](https://www.r-pkg.org/badges/version/piecepackr)](https://cran.r-project.org/package=piecepackr)

[![Coverage Status](https://codecov.io/github/piecepackr/piecepackr/branch/master/graph/badge.svg)](https://app.codecov.io/github/piecepackr/piecepackr?branch=master)

[![R-CMD-check](https://github.com/piecepackr/piecepackr/actions/workflows/R-CMD-check.yaml/badge.svg?branch=master)](https://github.com/piecepackr/piecepackr/actions)

<img src="man/figures/logo.png" align="right" width="200px" alt="piecepackr hex sticker">

`piecepackr` is an [R](https://www.r-project.org/) package designed to
make configurable board game graphics. It can be used with the
[ggplot2](https://ggplot2.tidyverse.org/),
[grid](https://www.rdocumentation.org/packages/grid),
[rayrender](https://www.rayrender.net/),
[rayvertex](https://www.rayvertex.com/), and
[rgl](https://www.rdocumentation.org/packages/rgl) graphics packages to
make board game diagrams, board game animations, and custom [Print &
Play
layouts](https://trevorldavis.com/piecepackr/pages/print-and-play-pdfs.html).
By default it is configured to make [piecepack](#piecepack) game
diagrams, animations, and "Print & Play" layouts but can be configured
to make graphics for other board game systems as well.

Built-in Game Systems
---------------------

The function `game_systems()` returns configurations for multiple public
domain game systems:

### Checkers

`game_systems()` returns a `checkers1` and `checkers2` configuration
which has checkered and lined "boards" with matching checker "bits" in
various sizes and colors.

    library("piecepackr")
    library("ppdf") # remotes::install_github("piecepackr/ppdf")
    stopifnot(packageVersion("ppdf") >= "0.2.0-13")

    df_board <- checker_board()
    df_w <- checker_bits(suit = "white", x = rep(1:8, 2), y = rep(1:2, each=8))
    df_b <- checker_bits(suit = "red", x = rep(1:8, 2), y = rep(7:8, each=8))
    df <- rbind(df_board, df_w, df_b)

    envir <- game_systems()
    pmap_piece(df, envir=envir, default.units="in", trans=op_transform, op_scale=0.5)

![Starting position for Dan Troyka's abstract game
Breakthrough](man/figures/README-breakthrough-1.png)

### Dice

`game_systems()` returns several configurations for dice:

-   The `dice` configuration makes standard 6-sided dice with pips.
-   The `dominoes_chinese` and `dominoes_chinese_black` configurations
    have Asian-style pipped dice.
-   The `dice_d4`, `dice_numeral`, `dice_d8`, `dice_d10`,
    `dice_d10_percentile`, `dice_d12`, and `dice_d20` configurations
    provide the seven [polyhedral
    dice](https://en.wikipedia.org/wiki/Dice#Polyhedral_dice) most
    commonly used by wargames, roleplaying games, and trading card
    games.
-   The `dice_fudge` configuration make the six-sided [Fudge
    dice](https://en.wikipedia.org/wiki/Fudge_(role-playing_game_system)#Fudge_dice)
    with two plus, two minus, and two blank faces most commonly used in
    the
    [Fudge](https://en.wikipedia.org/wiki/Fudge_(role-playing_game_system))
    and [Fate](https://fate-srd.com/) roleplaying games.

<!-- -->

    library("piecepackr")
    envir <- game_systems()
    dice <-  c("d4", "numeral", "d8", "d10_percentile", "d10", "d12", "d20")
    cfg <- paste0("dice_", dice)
    grid.piece("die_face", suit = c(1:6, 1), rank = 1:7,
               cfg = cfg, envir = envir, x = 1:7, y = 1,
               default.units = "in", op_scale = 0.5)

![Polyhedral dice](man/figures/README-polyhedral-1.png)

### Dominoes

`game_systems()` returns seven different configurations for double-18
dominoes:

1.  `dominoes`
2.  `dominoes_black`
3.  `dominoes_blue`
4.  `dominoes_green`
5.  `dominoes_red`
6.  `dominoes_white` (identical to `dominoes`)
7.  `dominoes_yellow`

The `dominoes_chinese` and `dominoes_chinese_black` configurations
support [Chinese
dominoes](https://www.pagat.com/class/domino.html#chinese).

    library("piecepackr")
    library("tibble")

    colors <- rep(c("black", "red", "green", "blue", "yellow", "white"), 2L)
    df <- tibble(piece_side = "tile_face",
                 suit=1:12, rank=7:18+1,
                 cfg = paste0("dominoes_", colors),
                 x=rep(4:1, 3), y=rep(2*3:1, each=4) - 0.5)

    envir <- game_systems(round = TRUE)
    pmap_piece(df, default.units="in", envir=envir, op_scale=0.5, trans=op_transform)

![Double-18 dominoes and standard dice in a variety of
colors](man/figures/README-dominoes-1.png)

### Go

`game_systems()` returns a `go` configuration for
[Go](https://en.wikipedia.org/wiki/Go_(game)) boards and stones in a
variety of colors and sizes. Here are is [an example
diagram](https://trevorldavis.com/piecepackr/go.html) for a game of
[Multi-player
go](https://en.wikipedia.org/wiki/Go_variants#Multi-player_Go) plotted
using [rgl](https://www.rdocumentation.org/packages/rgl):

    library("ppdf")
    library("piecepackr")
    stopifnot(packageVersion("ppdf") >= "0.2.0-13")

    dfb <- go_board()
    dfB <- go_bits(suit = "blue",
                   x = c(2,2,2, 3,3, 4,4, 5,5, 6,6,6, 7,7, 8),
                   y = c(11,13,14, 12,15, 13,16, 12,16, 12,13,16, 15,17, 16))
    dfR <- go_bits(suit = "red",
                   x = c(3,3, 4, 5, 6, 7, 8,8, 9),
                   y = c(2:3, 4, 3, 4, 4, 2:3, 2))
    dfY <- go_bits(suit = "yellow",
                   x = 20 - c(3, 4, 5,5, 7, 8, 9,9,9),
                   y = 20 - c(3, 2, 2:3, 2, 5, 2:4))
    dfG <- go_bits(suit = "green",
                   x = 20 - rep.int(3:6, times = c(5L, 4L, 3L, 2L)),
                   y = c(4,5,7,12,13, 6,7, 10,11, 7,9,10, 8,9))
    dfK <- go_bits(
        suit = "black",
        x = rep.int(2:17,
                    times = c(5L, 5L, 4L, 5L, 2L, 4L, 7L, 5L, 5L, 4L, 5L, 6L, 2L, 6L, 6L, 2L)),
        y = c(2:3,8:9,12, 4:5,10,13:14, 5,10,14:15, 6,8,10:11,14, 5,11,
              5,7,12,14, 4,7,9,11:13,15, 3:4,11,14,16, 8:9,14:16,
              3,6,8,15, 9:11,16,18, 6,8:9,11:12,16, 6:7,
              5:6,8,11:12,16, 3,5,12,14,16:17, 3,16)
    )
    dfW <- go_bits(
        suit = "white",
        x = rep.int(3:14,
                    times = c(1L, 1L, 1L, 3L, 3L, 2L, 5L, 3L, 4L, 4L, 2L, 1L)),
        y = c(9,9,9, 7:8,10, 8,10:11, 8,10, 8:10,12:13, 10:11,13,
              9,11,13:14, 7:8,12:13, 7,10, 11)
    )
    df <- rbind(dfb, dfB, dfR, dfY, dfG, dfK, dfW)

    envir <- game_systems(shading = TRUE, background_color = "burlywood")
    pmap_piece(df, envir = envir, default.units = "in")

![Multi-player Go diagram](man/figures/README-go-1.png)

### Marbles

`game_systems()` returns a `marbles` configuration with spherical
marbles in a variety of colors and sizes along with holed square boards
sized for the 1" diameter marbles.

    library("ppdf")
    library("piecepackr")
    stopifnot(packageVersion("ppdf") >= "0.2.0-13")

    set.seed(42)
    dfb <- marble_board(suit = "green", nrows = 4L, ncol = 4L, x0 = 1, y0 = 1)
    dfm <- marble_bits(
        suit = sample.int(6L, 30L, replace = TRUE),
        x = c(rep(1:4, 4L), rep(0.5 + rep(1:3, 3L)), rep(2:3, 2L), 2.5),
        y = c(rep(1:4, each = 4L), rep(0.5 + rep(1:3, each = 3L)), rep(2:3, each = 2L), 2.5)
    )
    df <- rbind(dfb, dfm)

    envir <- game_systems(round = TRUE, shading = TRUE)
    pmap_piece(df, default.units = "in", envir = envir,
               trans = marbles_transform, op_scale = 0.5)

![4x4 square pyramidal marble
stacking](man/figures/README-marbles-1.png)

### Piecepack

`game_systems()` returns three different [piecepack](#piecepack)
configurations:

1.  `piecepack`
2.  `playing_cards_expansion`
3.  `dual_piecepacks_expansion`

Plus a configuration for a `subpack` aka "mini" piecepack and a
`hexpack` configuration.

The piecepack configurations also contain common piecepack accessories
like piecepack pyramids, piecepack matchsticks, and piecepack saucers.

### Playing Cards

`game_systems()` returns `playing_cards`, `playing_cards_colored`, and
`playing_cards_tarot` (French Tarot) configurations for making diagrams
with various decks of playing cards.

    library("piecepackr")
    library("tibble")

    df <- tibble(piece_side = "card_face",
                 x=1.25 + 2.5 * 0:3, y=2,
                 suit=1:4, rank=c(1,6,9,12),
                 cfg = "playing_cards")

    envir <- game_systems("dejavu", round=TRUE)
    pmap_piece(df, default.units="in", envir=envir)

![Playing Cards](man/figures/README-cards-1.png)

### Other included games and components

-   An `alquerque` configuration that produces "boards"/"bits" for
    [Alquerque](https://en.wikipedia.org/wiki/Alquerque) in a variety of
    colors.
-   `chess1` and `chess2` configurations with checkered "boards" and
    matching chess "bits" (currently "disc" pieces instead of "Staunton"
    pieces).
-   A `meeples` configuration that produces "meeple" bits in a variety
    of colors.
-   A `morris` configuration that can produce
    [Three/Six/Seven/Nine/Twelve men's
    morris](https://en.wikipedia.org/wiki/Nine_men%27s_morris)
    "board"/"bits" in a variety of colors.
-   A `reversi` configuration that can produce "boards"/"bits" for
    [Reversi](http://www.cyningstan.com/game/73/reversi) in a variety of
    colors.

Looney Pyramids
---------------

Configurations for the proprietary Looney Pyramids aka Icehouse Pieces
game system by Andrew Looney can be found in the companion R package
`piecenikr`: <https://github.com/piecepackr/piecenikr>

API Intro
---------

### grid.piece() ({grid})

`grid.piece()` is the core function that can used to draw board game
components (by default [piecepack](#piecepack) game components) using
[grid](https://www.rdocumentation.org/packages/grid):

    library("piecepackr")
    g.p <- function(...) { grid.piece(..., default.units="in") }
    g.p("tile_back", x=0.5+c(3,1,3,1), y=0.5+c(3,3,1,1))
    g.p("tile_back", x=0.5+3, y=0.5+1)
    g.p("tile_back", x=0.5+3, y=0.5+1)
    g.p("die_face", suit=3, rank=5, x=1, y=1)
    g.p("pawn_face", x=1, y=4, angle=90)
    g.p("coin_back", x=3, y=4, angle=180)
    g.p("coin_back", suit=4, x=3, y=4, angle=180)
    g.p("coin_back", suit=2, x=3, y=1, angle=90)

![Piecepack diagram with default
configuration](man/figures/README-intro1-1.png)

### configuration lists

One can use [lists to
configure](https://trevorldavis.com/piecepackr/configuration-lists.html)
to quickly adjust the appearance of the game components drawn by
`grid.piece`:

    library("piecepackr")
    dark_colorscheme <- list(
        suit_color="darkred,black,darkgreen,darkblue,black",
        invert_colors.suited=TRUE, border_color="black", border_lex=2
    )
    piecepack_suits <- list(
        suit_text="\U0001f31e\ufe0e,\U0001f31c\ufe0e,\U0001f451\ufe0e,\u269c,\uaa5c", # 🌞,🌜,👑,⚜,꩜
        suit_fontfamily="Noto Emoji,Noto Sans Symbols2,Noto Emoji,Noto Sans Symbols,Noto Sans Cham",
        suit_cex="0.6,0.7,0.75,0.9,0.9"
    )
    traditional_ranks <- list(use_suit_as_ace=TRUE, rank_text=",a,2,3,4,5")
    cfg <- c(piecepack_suits, dark_colorscheme, traditional_ranks)
    g.p <- function(...) {
        grid.piece(..., default.units="in", cfg=pp_cfg(cfg))
    }
    g.p("tile_back", x=0.5+c(3,1,3,1), y=0.5+c(3,3,1,1))
    g.p("tile_back", x=0.5+3, y=0.5+1)
    g.p("tile_back", x=0.5+3, y=0.5+1)
    g.p("die_face", suit=3, rank=5, x=1, y=1)
    g.p("pawn_face", x=1, y=4, angle=90)
    g.p("coin_back", x=3, y=4, angle=180)
    g.p("coin_back", suit=4, x=3, y=4, angle=180)
    g.p("coin_back", suit=2, x=3, y=1, angle=90)

![Piecepack diagram with custom
configuration](man/figures/README-config-1.png)

### custom grob functions

One can even specify [custom grob
functions](https://trevorldavis.com/piecepackr/custom-grob-functions.html)
to completely customize the appearance of one's game pieces. <span
class="title-ref">piecepackr</span> comes with a variety of convenience
functions such as <span class="title-ref">pp\_shape()</span> to
facilitate creating custom game pieces. Here is an example of creating
"patterned" checkers using `pp_shape()` objects' `pattern()` method
powered by the suggested package
[gridpattern](https://github.com/trevorld/gridpattern):

    library("grid")
    library("gridpattern")
    library("piecepackr")
    library("ppdf") # remotes::install_github("piecepackr/ppdf")
    stopifnot(packageVersion("ppdf") >= "0.2.0-13")

    tilings <- c("hexagonal", "snub_square", "pythagorean",
                 "truncated_square", "triangular", "trihexagonal")
    patternedCheckerGrobFn <- function(piece_side, suit, rank, cfg) {
        opt <- cfg$get_piece_opt(piece_side, suit, rank)
        shape <- pp_shape(opt$shape, opt$shape_t, opt$shape_r, opt$back)
        gp <- gpar(col=opt$suit_color, fill=c(opt$background_color, "white"))
        pattern_grob <- shape$pattern("polygon_tiling", type = tilings[suit],
                                      spacing = 0.3, name = "pattern",
                                      gp = gp, angle = 0)
        gp_border <- gpar(col=opt$border_color, fill=NA, lex=opt$border_lex)
        border_grob <- shape$shape(gp=gp_border, name = "border")
        grobTree(pattern_grob, border_grob)
    }
    envir <- game_systems()
    checkers1 <- as.list(envir$checkers1)
    checkers1$grob_fn.bit <- patternedCheckerGrobFn
    envir$checkers1 <- pp_cfg(checkers1)

    x1 <- c(1:3, 1:2, 1)
    x2 <- c(6:8, 7:8, 8)
    df_board <- checker_board(suit = 6L)
    df_checkers <- checker_bits(suit = rep(c(1L, 3L, 4L, 5L), each = 6L),
                                x = c(x1, rev(x1), x2, rev(x2)),
                                y = rep(c(1,1,1, 2,2, 3, 6, 7,7, 8,8,8), 2))
    df <- rbind(df_board, df_checkers)

    pmap_piece(df, envir=envir, default.units="in")

![Patterned checkers via custom grob
function](man/figures/README-pattern-1.png)

### oblique 3D projection

`grid.piece` even has some support for drawing 3D diagrams with an
[oblique
projection](https://trevorldavis.com/piecepackr/3d-projections.html):

    library("piecepackr")
    cfg3d <- list(width.pawn=0.75, height.pawn=0.75, depth.pawn=1,
                  dm_text.pawn="", shape.pawn="convex6",
                  invert_colors.pawn=TRUE,
                  edge_color.coin="tan", edge_color.tile="tan")
    cfg <- pp_cfg(c(cfg, cfg3d))
    g.p <- function(...) {
      grid.piece(..., op_scale=0.5, op_angle=45, cfg=cfg, default.units="in")
    }
    g.p("tile_back", x=0.5+c(3,1,3,1), y=0.5+c(3,3,1,1))
    g.p("tile_back", x=0.5+3, y=0.5+1, z=1/4+1/8)
    g.p("tile_back", x=0.5+3, y=0.5+1, z=2/4+1/8)
    g.p("die_face", suit=3, rank=5, x=1, y=1, z=1/4+1/4)
    g.p("pawn_face", x=1, y=4, z=1/4+1/2, angle=90)
    g.p("coin_back", x=3, y=4, z=1/4+1/16, angle=180)
    g.p("coin_back", suit=4, x=3, y=4, z=1/4+1/8+1/16, angle=180)
    g.p("coin_back", suit=2, x=3, y=1, z=3/4+1/8, angle=90)

![Piecepack diagram in an oblique
projection](man/figures/README-proj-1.png)

### save\_print\_and\_play() and save\_piece\_images()

`save_print_and_play()` makes a "Print & Play" pdf of a configured
piecepack, `save_piece_images()` makes individual images of each
piecepack component:

    save_print_and_play(cfg, "my_piecepack.pdf", size="letter")
    save_piece_images(cfg)

### pmap\_piece()

If you are comfortable using R data frames there is also `pmap_piece()`
that processes data frame input. It accepts an optional `trans` argument
for a function to pre-process the data frames, in particular if desiring
to draw a 3D [oblique
projection](https://trevorldavis.com/piecepackr/3d-projections.html) one
can use the function `op_transform()` to guess both the pieces'
z-coordinates and an appropriate re-ordering of the data frame given the
desired angle of the oblique projection.

    library("dplyr", warn.conflicts=FALSE)
    library("piecepackr")
    library("tibble")
    df_tiles <- tibble(piece_side="tile_back",
                       x=0.5+c(3,1,3,1,1,1),
                       y=0.5+c(3,3,1,1,1,1))
    df_coins <- tibble(piece_side="coin_back",
                       x=rep(1:4, 4),
                       y=rep(c(4,1), each=8),
                       suit=1:16%%2+rep(c(1,3), each=8),
                       angle=rep(c(180,0), each=8))
    df <- bind_rows(df_tiles, df_coins)
    cfg <- game_systems("dejavu")$piecepack
    pmap_piece(df, cfg=cfg, default.units="in", trans=op_transform,
               op_scale=0.5, op_angle=135)

!['pmap\_piece()' lets you use data frames as
input](man/figures/README-pmap-1.png)

### geom\_piece() ({ggplot2})

`geom_piece()` creates [ggplot2](https://ggplot2.tidyverse.org/) "geom"
objects.

    library("ggplot2")
    library("piecepackr")
    envir <- game_systems("sans")
    df_board <- tibble(piece_side = "board_face", suit = 3, rank = 12,
                       x = 4, y = 4)
    df_b <- tibble(piece_side = "bit_back", suit = 2, rank = 1,
                   x = c(2, 3, 3, 4, 4), y = c(6, 5, 4, 5, 2))
    df_w <- tibble(piece_side = "bit_back", suit = 1, rank = 1,
                   x = c(2, 2, 3, 4, 5, 5), y = c(4, 3, 6, 5, 4, 6))
    df <- rbind(df_board, df_w, df_b)

    ggplot(df, aes_piece(df)) +
        geom_piece(cfg = "morris", envir = envir) +
        coord_fixed() +
        scale_x_piece(limits = c(0.5, 7.5)) +
        scale_y_piece(limits = c(0.5, 7.5)) +
        theme_minimal(32) +
        theme(panel.grid = element_blank())

![Twelve men's morris game diagram](man/figures/README-ggplot2_2d-1.png)

    library("ggplot2")
    library("piecepackr")
    library("ppdf") # remotes::install_github("piecepackr/ppdf")
    library("withr")
    new <- list(piecepackr.cfg = "piecepack",
                piecepackr.envir = game_systems("dejavu", pawn="joystick"),
                piecepackr.op_angle = 90,
                piecepackr.op_scale = 0.80)
    dfc <- ppdf::piecepack_fujisan(seed = 42)
    withr::with_options(new, {
      dft <- op_transform(dfc, as_top = "pawn_face", cfg_class = "character")
      ggplot(dft, aes_piece(dft)) +
          geom_piece() +
          coord_fixed() +
          theme_void()
    })

![Fuji-san starting diagram in an oblique
projection](man/figures/README-ggplot2-1.png)

### piece3d() ({rgl})

`piece3d()` draws pieces using
[rgl](https://www.rdocumentation.org/packages/rgl) graphics.

    library("piecepackr")
    library("piecenikr") # remotes::install_github("piecepackr/piecenikr")
    library("rgl")
    invisible(rgl::open3d())
    rgl::view3d(phi=-45, zoom = 0.9)

    df <- icehouse_martian_chess()
    envir <- c(looney_pyramid_game_system(border = FALSE),
               game_systems(border = FALSE))
    pmap_piece(df, piece3d, envir = envir, trans=op_transform,
               scale = 0.98, res = 150, lit = TRUE)

![3D render with rgl package](man/figures/README-rgl_snapshot.png)

### piece() ({rayrender})

`piece()` creates [rayrender](https://www.rayrender.net/) objects.

    library("piecepackr")
    library("ppdf") # remotes::install_github("piecepackr/ppdf")
    library("rayrender", warn.conflicts = FALSE)
    df <- ppdf::piecepack_xiangqi() |>
        transform(cfg = "dual_piecepacks_expansion")
    envir <- game_systems("dejavu", border = FALSE, round = TRUE, pawn = "peg-doll")
    l <- pmap_piece(df, piece, envir = envir, trans=op_transform,
                    scale = 0.98, res = 150, as_top="pawn_face")
    light <- sphere(x=5,y=-4, z=30, material=light(intensity=420))
    table <- sphere(z=-1e3, radius=1e3, material=diffuse(color="green")) |>
             add_object(light)
    scene <- Reduce(rayrender::add_object, l, init=table)
    rayrender::render_scene(scene, preview = FALSE,
                            lookat = c(5, 5, 0), lookfrom = c(5, -7, 25),
                            sample_method = "sobol", clamp_value = 12,
                            width = 480, height = 480, samples = 200)

![3D render with rayrender package](man/figures/README-rayrender-1.png)

### piece\_mesh() ({rayvertex})

`piece_mesh()` creates [rayvertex](https://www.rayvertex.com/) objects.

    library("piecepackr")
    library("ppdf") # remotes::install_github("piecepackr/ppdf")
    library("rayvertex", warn.conflicts = FALSE)
    df <- ppdf::piecepack_international_chess() |>
        transform(cfg = "dual_piecepacks_expansion")
    envir <- game_systems("dejavu", border = FALSE, round = TRUE, pawn = "joystick")
    l <- pmap_piece(df, piece_mesh, envir = envir, trans=op_transform,
                    scale = 0.98, res = 150, as_top="pawn_face")
    table <- sphere_mesh(c(0, 0, -1e3), radius=1e3,
                         material = material_list(diffuse="grey40"))
    scene <- rayvertex::scene_from_list(l) |> add_shape(table)
    light_info <- directional_light(c(5, -7, 7), intensity = 2.5)
    rayvertex::rasterize_scene(scene,
                               lookat = c(4.5, 4, 0),
                               lookfrom=c(4.5, -16, 20),
                               light_info = light_info)

![3D render with rayvertex package](man/figures/README-rayvertex-1.png)

### animate\_piece()

`animate_piece()` creates animations.

    library("gifski")
    library("piecepackr")
    library("ppn") # remotes::install_github("piecepackr/ppn")
    library("tweenr")

    envir <- game_systems("dejavu")
    cfg <- as.list(envir$piecepack)
    cfg$suit_color <- "black"
    cfg$background_color.r1 <- "#E69F00"
    cfg$background_color.r2 <- "#56B4E9"
    cfg$background_color.r3 <- "#009E73"
    cfg$background_color.r4 <- "#F0E442"
    cfg$background_color.r5 <- "#D55E00"
    cfg$background_color.r6 <- "#F079A7"
    envir$piecepack <- pp_cfg(cfg)

    ppn_file <- system.file("ppn/relativity.ppn", package = "ppn")
    game <- read_ppn(ppn_file)[[1]]
    animate_piece(game$dfs, file = "man/figures/README-relativity.gif",
                  annotate = FALSE,
                  envir = envir, trans = op_transform, op_scale = 0.5,
                  n_transitions = 3, n_pauses = 2, fps = 7)

<figure>
<img src="man/figures/README-relativity.gif" class="align-center" alt="" /><figcaption>Animation of Marty and Ron Hale-Evans' abstract game Relativity</figcaption>
</figure>

### Further documentation

A slightly longer [intro to piecepackr's
API](https://trevorldavis.com/piecepackr/intro-to-piecepackrs-api.html)
plus several other [piecepackr
articles](https://trevorldavis.com/piecepackr/tag/piecepackr-features.html)
are available at piecepackr's [companion
website](https://trevorldavis.com/piecepackr/) as well as some
[demos](https://trevorldavis.com/piecepackr/pages/about.html#demos) and
pre-configured [Print & Play
PDFs](https://trevorldavis.com/piecepackr/pages/print-and-play-pdfs.html).
More API documentation is also available in the package's built-in [man
pages](https://trevorldavis.com/R/piecepackr/reference/index.html).

### Tak Example

Here we'll show an example of configuring piecepackr to draw diagrams
for the abstract board game [Tak](https://crabfragmentlabs.com/tak)
(designed by James Ernest and Patrick Rothfuss).

Since one often plays Tak on differently sized boards one common Tak
board design is to have boards made with colored cells arranged in rings
from the center plus extra symbols in rings placed at the points so it
is easy to see smaller sub-boards. To start we'll write a function to
draw the Tak board.

    library("grid", warn.conflicts=FALSE)
    library("piecepackr")
    grobTakBoard <- function(...) {
        g <- "darkgreen"
        w <- "grey"
        fill <- c(rep(g, 5),
                  rep(c(g, rep(w, 3), g),3),
                  rep(g, 5))
        inner <- rectGrob(x = rep(1:5, 5), y = rep(5:1, each=5),
                     width=1, height=1, default.units="in",
                     gp=gpar(col="gold", fill=fill, lwd=3))
        outer <- rectGrob(gp=gpar(col="black", fill="grey", gp=gpar(lex=2)))
        circles <- circleGrob(x=0.5+rep(1:4, 4),
                              y=0.5+rep(4:1, each=4),
                              r=0.1, default.units="in",
                              gp=gpar(col=NA, fill="gold"))
        rects <- rectGrob(x=0.5+c(0:5, rep(c(0,5), 4), 0:5),
                          y=0.5+c(rep(5,6), rep(c(4:1), each=2), rep(0, 6)),
                          width=0.2, height=0.2,
                          gp=gpar(col=NA, fill="orange"), default.units="in")
        grobTree(outer, inner, circles, rects)
    }

Then we'll configure a Tak set and write some helper functions to draw
Tak pieces with it.

    cfg <- pp_cfg(list(suit_text=",,,", suit_color="white,tan4,", invert_colors=TRUE,
                       ps_text="", dm_text="",
                       width.board=6, height.board=6,
                       depth.board=1/4, grob_fn.board=grobTakBoard,
                       width.r1.bit=0.6, height.r1.bit=0.6,
                       depth.r1.bit=1/4, shape.r1.bit="rect",
                       width.r2.bit=0.6, height.r2.bit=1/4,
                       depth.r2.bit=0.6, shape.r2.bit="rect",
                       width.pawn=0.5, height.pawn=0.5,
                       depth.pawn=0.8, shape.pawn="circle",
                       edge_color="white,tan4", edge_color.board="tan",
                       border_lex=3, border_color="black"))
    g.p <- function(...) {
        grid.piece(..., cfg=cfg, default.units="in",
                   op_scale=0.7, op_angle=45)
    }
    draw_tak_board <- function(x, y) {
        g.p("board_back", x=x+0.5, y=y+0.5)
    }
    draw_flat_stone <- function(x, y, suit=1) {
        z <- 1/4*seq(along=suit)+1/8
        g.p("bit_back", x=x+0.5, y=y+0.5, z=z, suit=suit, rank=1)
    }
    draw_standing_stone <- function(x, y, suit=1, n_beneath=0, angle=0) {
        z <- (n_beneath+1)*1/4+0.3
        g.p("bit_back", suit=suit, rank=2,
            x=x+0.5, y=y+0.5, z=z, angle=angle)
    }
    draw_capstone <- function(x, y, suit=1, n_beneath=0) {
        z <- (n_beneath+1)*1/4+0.4
        g.p("pawn_back", x=x+0.5, y=y+0.5, z=z, suit=suit)
    }

Then we'll draw an example Tak game diagram:

    pushViewport(viewport(width=inch(6), height=inch(6)))
    draw_tak_board(3, 3)
    draw_flat_stone(1, 1, 1)
    draw_flat_stone(1, 2, 2)
    draw_flat_stone(2, 4, 1)
    draw_capstone(2, 4, 2, n_beneath=1)
    draw_flat_stone(2, 5, 2)
    draw_flat_stone(3, 4, 1:2)
    draw_flat_stone(3, 3, c(2,1,1,2))
    draw_flat_stone(3, 2, 1:2)
    draw_flat_stone(3, 1, 2)
    draw_standing_stone(4, 2, 2, angle=90)
    draw_flat_stone(5, 2, 1)
    draw_capstone(5, 3, 1)
    popViewport()

![Tak game diagram](man/figures/README-diagram-1.png)

Installation
------------

To install the last version released on CRAN use the following command
in [R](https://www.r-project.org/):

    install.packages("piecepackr")

To install the development version use the following commands:

    options(repos = c(
        piecepackr = 'https://piecepackr.r-universe.dev',
        CRAN = 'https://cloud.r-project.org'))
    install.packages("piecepackr")

### Suggested R packages

Although the "core" `{piecepackr}` functionality does not need any
additional software installed some non-"core" functionality needs extra
suggested software to be installed. To install **all** of the suggested
R packages use:

    install.packages("piecepackr", dependencies = TRUE)

or (for the development version from the [piecepackr
universe](#piecepackr-universe)):

    options(repos = c(
        piecepackr = 'https://piecepackr.r-universe.dev',
        CRAN = 'https://cloud.r-project.org'))
    install.packages("piecepackr", dependencies = TRUE)

Suggested R packages:

**animation**  
`animate_piece()` uses the `{animation}` package to save "html" and
"video" (e.g. mp4 and avi) animations. Additionally, if the `{gifski}`
package is not installed `animate_piece()` will fall back to using
`{animation}` to make "gif" animations.

**ggplot2**  
Required by the `{ggplot2}` bindings `geom_piece()` and its helper
functions `aes_piece()`, `scale_x_piece()`, and `scale_y_piece()`.

**gifski**  
`animate_piece()` preferably uses the `{gifski}` package to save "gif"
animations. If `{gifski}` is not available then `animate_piece()` can
fall back on `{animation}` to make "gif" animations.

**gridpattern**  
The `pp_shape()` object's `pattern()` method uses `{gridpattern}` to
make patterned shapes. In particular can be used to make patterned board
game pieces.

**magick**  
`file2grob()` uses `magick::image_read()` to import images that are not
"png", "jpg/jpeg", or "svg/svgz".

**pdftools**  
`get_embedded_font()` uses `pdftools::pdf_fonts()`. It also requires R
compiled with Cairo support (i.e. `capabilities("cairo") == TRUE`). If
the suggested R package `{systemfonts}` is not installed then
`has_font()` can also fall back on using `get_embedded_font()`.

**rayrender**  
Required for the `{rayrender}` binding `piece()` and the `pp_cfg()`
object's `rayrender_fn()` method.

**rayvertex**  
Required for the `{rayvertex}` binding `piece_mesh()` and the `pp_cfg()`
object's `rayvertex_fn()` method.

**readobj**  
Allows the `{rgl}` bindings to support more game piece shapes; in
particular the "meeple", "halma", and "roundrect" shaped token game
pieces.

**rgl**  
Required for the `{rgl}` binding `piece3d()` and the `pp_cfg()` object's
`rgl_fn()` method. Also required for the `obj_fn()` method for game
pieces with ellipsoid shapes (in particular this may effect
`save_piece_obj()`, `piece()`, `piece3d()`, and/or `piece_mesh()` when
used with the go stones and joystick pawns provided by
`game_systems()`). You may need to [install extra
software](https://github.com/dmurdoch/rgl#installing-opengl-support) for
`{rgl}` to support OpenGL (in addition to WebGL). Consider also
installing `{readobj}` which allows the `{rgl}` bindings to support more
game piece shapes; in particular the "meeple", "halma", and "roundrect"
shaped token game pieces.

**systemfonts**  
`has_font()` preferably uses `{systemfonts}` to determine if a given
font is available. If `{systemfonts}` is not available then `has_font()`
can fall back on `{pdftools}` if `capabilities("cairo") == TRUE`.

**tweenr**  
`animate_piece()` needs `{tweenr}` to do animation transitions (i.e. its
`n_transitions` argument is greater than the default zero).

**xmpdf**  
`save_print_and_play()` can use `{xmpdf}` to embed bookmarks,
documentation info, and XMP metadata into pdf print-and-play files. You
may also need the system tools <span
class="title-ref">ghostscript</span>, <span
class="title-ref">pdftk</span>, and/or <span
class="title-ref">exiftool</span>.

### piecepackr universe

The [piecepackr universe](#piecepackr-universe) contains other R
packages that may also be of interest to fans of `piecepackr`. To
install them use:

    # install.packages("piecepackr")
    piecepackr::install_ppverse(free_libre_only = TRUE)

**ppcli**  
Functions to visualize board games in plaintext. Provides colorization
support for the terminal and HTML via `cli`.

**ppdf**  
Generate `tibble` data frames indicating how to set up over a hundred
board games playable with public domain game systems. Data format can be
used by `piecepackr` to generate graphics or by `ppcli` to generate
plaintext graphics with `cli`.

**ppgamer**  
Functions that provide players for piecepack games like a solver for
"Fuji-san".

**ppn**  
Parses "Portable Piecepack Notation" files. This allows you to visualize
the moves for over one hundred board games using `piecepackr`.

**pprules**  
Functions to generate piecepack game rulesets and books.

### Other suggested software

The default piecepackr `pp_cfg()` configuration and the default game
systems returned by `game_systems()` should work out on the box on most
modern OSes including Windows without the user needing to mess with
their system fonts. However `game_systems(style = "dejavu")` requires
that the [Dejavu Sans](https://dejavu-fonts.github.io/Download.html)
font is installed.

For more advanced `piecepackr` configurations you'll want to install
additional Unicode fonts and Windows users are highly recommended to use
and install piecepackr on "Ubuntu on Bash on Windows" if planning on
using Unicode symbols from multiple fonts. The following bash commands
will give you a good selection of fonts (Noto, Quivira, and Dejavu) on
Ubuntu:

    sudo apt install fonts-dejavu fonts-noto
    fonts_dir=${XDG_DATA_HOME:="$HOME/.local/share"}/fonts
    curl -O http://www.quivira-font.com/files/Quivira.otf
    mv Quivira.otf $fonts_dir/
    curl -O https://noto-website-2.storage.googleapis.com/pkgs/NotoEmoji-unhinted.zip
    unzip NotoEmoji-unhinted.zip NotoEmoji-Regular.ttf
    mv NotoEmoji-Regular.ttf $fonts_dir/
    rm NotoEmoji-unhinted.zip

Certain `piecepackr` features works best if the version of R installed
was compiled with support for Cairo:

-   A subset of game system configurations use Unicode glyphs. The
    "cairo" graphics devices support Unicode glyphs.
-   3D [grid](https://www.rdocumentation.org/packages/grid) renderings
    for certain pieces like dice and pyramids are enhanced if the
    graphic device supports the "affine transformation" feature. In
    recent versions of R the "cairo" graphics devices support the
    "affine transformation" feature.
-   The function `get_embedded_font()` needs support for the
    `cairo_pdf()` function (which embeds fonts in the pdf) and by
    default `render_piece()` and `save_print_and_play()` may try to use
    "cairo" graphics devices.

Fortunately R is typically compiled with support for Cairo. One can
confirm that R was compiled with support for Cairo via R's
`capabilities()` function:

    > capabilities("cairo")
    cairo
     TRUE

Frequently Asked Questions
--------------------------

### Where should I ask questions?

-   For general questions about piecepackr one may use the project
    mailing list: <https://groups.google.com/forum/#!forum/piecepackr>
-   If you have a bug report or a feature request please use the issue
    tracker: <https://github.com/piecepackr/piecepackr/issues>

### What is the package licence?

The **code** of this software package is licensed under the [MIT
license](https://opensource.org/license/mit).

Graphical assets generated using configurations returned by
`game_systems()` should be usable without attribution:

1.  Uses fonts which should allow you to embed them in images/documents
    without requiring attribution.
2.  Does not embed any outside copyrighted images.[1]
3.  Only contains public domain game systems which should not suffer
    from copyright / trademark issues.

However, third party game configurations [may be encumbered by copyright
/ trademark
issues](https://trevorldavis.com/piecepackr/licenses-faq.html#piecepackr-output).

### Why does the package sometimes use a different font then the one I instructed it to use for a particular symbol?

Some of R's graphic devices (`cairo_pdf()`, `svg()`, and `png()`) use
`Cairo` which uses `fontconfig` to select fonts. `fontconfig` picks what
it thinks is the 'best' font and sometimes it annoyingly decides that
the font to use for a particular symbol is not the one you asked it to
use (although sometimes the symbol it chooses instead still looks nice
in which case maybe you shouldn't sweat it). It is hard but not
impossible to [configure which
fonts](https://eev.ee/blog/2015/05/20/i-stared-into-the-fontconfig-and-the-fontconfig-stared-back-at-me/)
are dispatched by `fontconfig`. A perhaps easier way to guarantee your
symbols will be dispatched would be to either make a new font and
re-assign the symbols to code points in the Unicode "Private Use Area"
that aren't used by any other font on your system or to simply
temporarily move (or permanently delete) from your system font folders
the undesired fonts that `fontconfig` chooses over your requested fonts:

    # temporarily force fontconfig to use Noto Emoji instead of Noto Color Emoji in my piecepacks on Ubuntu 18.04
    $ sudo mv /usr/share/fonts/truetype/noto/NotoColorEmoji.ttf ~/
    ## Make some piecepacks
    $ sudo mv ~/NotoColorEmoji.ttf /usr/share/fonts/truetype/noto/

Also as a sanity check use the command-line tool `fc-match` (or the R
function `systemfonts::match_font()`) to make sure you specified your
font correctly in the first place (i.e. `fc-match "Noto Sans"` on my
system returns "Noto Sans" but `fc-match "Sans Noto"` returns "DejaVu
Sans" and not "Noto Sans" as one may have expected). To help determine
which fonts are actually being embedded you can use the
`get_embedded_font()` helper function:

    library("piecepackr")
    fonts <- c('Noto Sans Symbols2', 'Noto Emoji', 'sans')
    chars <- c('♥', '♠', '♣', '♦', '🌞' ,'🌜' ,'꩜')
    get_embedded_font(fonts, chars)
    #     char      requested_font            embedded_font
    # 1      ♥ Noto Sans Symbols2 NotoSansSymbols2-Regular
    # 2      ♠ Noto Sans Symbols2 NotoSansSymbols2-Regular
    # 3      ♣ Noto Sans Symbols2 NotoSansSymbols2-Regular
    # 4      ♦ Noto Sans Symbols2 NotoSansSymbols2-Regular
    # 5       🌞Noto Sans Symbols2                NotoEmoji
    # 6       🌜Noto Sans Symbols2                NotoEmoji
    # 7      ꩜ Noto Sans Symbols2     NotoSansCham-Regular
    # 8      ♥         Noto Emoji                NotoEmoji
    # 9      ♠         Noto Emoji                NotoEmoji
    # 10     ♣         Noto Emoji                NotoEmoji
    # 11     ♦         Noto Emoji                NotoEmoji
    # 12      🌞        Noto Emoji                NotoEmoji
    # 13      🌜        Noto Emoji                NotoEmoji
    # 14     ꩜         Noto Emoji     NotoSansCham-Regular
    # 15     ♥               sans                    Arimo
    # 16     ♠               sans                    Arimo
    # 17     ♣               sans                    Arimo
    # 18     ♦               sans                    Arimo
    # 19      🌞              sans                NotoEmoji
    # 20      🌜              sans                NotoEmoji
    # 21     ꩜               sans     NotoSansCham-Regular

[1] The outline for the meeple shape used in the "meeples" configuration
(also used in some face cards in the playing cards configurations) was
extracted (converted into a dataset of normalized x, y coordinates) from
[Meeple icon](https://game-icons.net/1x1/delapouite/meeple.html) by
[Delapouite](https://delapouite.com/) / [CC BY
3.0](https://creativecommons.org/licenses/by/3.0/). Since "simple
shapes" nor data can be copyrighted under American law this meeple
outline is not copyrightable in the United States. However, in other
legal jurisdictions with stricter copyright laws you may need to give
the proper CC BY attribution if you use any of the meeples.

The "svgz" files in `extdata/badges` are Creative Commons License "button marks" from https://github.com/creativecommons/cc-assets made available under the following license (https://creativecommons.org/policies):
  
  ```
  Creative Commons License Buttons and Icons: Creative Commons licenses the use of its button marks that describe a particular legal tool and its icon marks that describe a key license element, such as BY, NC, ND, and SA, on the conditions that you use the mark solely to describe the Creative Commons legal tool that applies to a particular work and, in a manner reasonable to the medium and context, include the URI or hyperlink to the relevant Commons deed on the Creative Commons server.
  ```

  They are solely used by `save_print_and_play()` to describe
  the Creative Commons legal tool that applies to the generated print-and-play file.
  Additionally `save_print_and_play()` also includes in the generated print-and-play file
  a URI to the relevant Commons deed on the Creative Commons server.

#' Draw board game pieces with grid
#'
#' `grid.piece()` draws board game pieces onto the graphics device.
#' `pieceGrob()` is its `grid` \dQuote{grob} counterpart.
#'
#' @param piece_side A string with piece and side separated by a underscore e.g. "coin_face"
#' @param cfg Piecepack configuration list or `pp_cfg` object,
#'        a list of `pp_cfg` objects,
#'        or a character vector referring to names in `envir`
#'        or a character vector referring to object names that
#'        can be retrieved by `base::dynGet()`.
#' @param suit Number of suit (starting from 1).
#' @param rank Number of rank (starting from 1)
#' @param x Where to place piece on x axis of viewport
#' @param y Where to place piece on y axis of viewport
#' @param z z-coordinate of the piece.  Has no effect if `op_scale` is `0`.
#' @param angle Angle (on xy plane) to draw piece at
#' @param width Width of piece
#' @param height Height of piece
#' @param depth Depth (thickness) of piece.  Has no effect if `op_scale` is `0`.
#' @param op_scale How much to scale the depth of the piece in the oblique projection
#'              (viewed from the top of the board).
#'              `0` (the default) leads to an \dQuote{orthographic} projection,
#'              `0.5` is the most common scale used in the \dQuote{cabinet} projection,
#'               and `1.0` is the scale used in the \dQuote{cavalier} projection.
#' @param op_angle What is the angle of the oblique projection?  Has no effect if `op_scale` is `0`.
#' @param default.units  A string indicating the default units to use if
#'   'x', 'y', 'width', and/or 'height' are only given as numeric vectors.
#' @param envir Environment (or named list) containing configuration list(s).
#' @param name A character identifier (for grid)
#' @param gp An object of class \dQuote{gpar}.
#' @param draw A logical value indicating whether graphics output should be produced.
#' @param vp A `grid` viewport object (or `NULL`).
#' @param ... Ignored.
#' @param scale Multiplicative scaling factor to apply to width, height, and depth.
#' @param alpha Alpha channel for transparency.
#' @param type Type of grid grob to use.  Either `"normal"` (default), `"picture"`, `"raster"`, or `"transformation"`.
#'             `"picture"` exports to (temporary) svg and re-imports as a [grImport2::pictureGrob()].
#'             `"raster"` exports to (temporary) png and re-imports as a [grid::rasterGrob()].
#'             `"transformation"` uses the affine transformation feature only supported in
#'             R 4.2+ within select graphic devices.
#'             The latter three can be useful if drawing pieces really big or small and don't want
#'             to mess with re-configuring fontsizes and linewidths.
#' @param bleed If `FALSE` do not add a \dQuote{bleed} zone around the piece,
#'              otherwise add a \dQuote{bleed} zone around the piece:\itemize{
#'                 \item{If `bleed` is `TRUE` we will add 1/8 inch bleeds}
#'                 \item{If `bleed` is a [grid::unit()] we will use it as bleed size}
#'                 \item{If `bleed` is numeric we will convert to [grid::unit()] via `grid::unit(bleed, default.units)`}
#'              }
#'              A non-`FALSE` `bleed` is incompatible with `op_scale > 0` (drawing in an \dQuote{oblique projection}).
#' @return A `grid` grob object.  If `draw` is `TRUE` then as a side effect
#'         `grid.piece()` will also draw it to the graphics device.
#' @examples
#'   if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {
#'     opt <- options(piecepackr.at.inform = FALSE)
#'
#'     draw_pp_diagram <- function(cfg=pp_cfg(), op_scale=0) {
#'         g.p <- function(...) {
#'             grid.piece(..., op_scale=op_scale, cfg=cfg, default.units="in")
#'         }
#'         g.p("tile_back", x=0.5+c(3,1,3,1), y=0.5+c(3,3,1,1))
#'         g.p("tile_back", x=0.5+3, y=0.5+1, z=1/4+1/8)
#'         g.p("tile_back", x=0.5+3, y=0.5+1, z=2/4+1/8)
#'         g.p("die_face", suit=3, rank=5, x=1, y=1, z=1/4+1/4)
#'         g.p("pawn_face", x=1, y=4, z=1/4+1/2, angle=90)
#'         g.p("coin_back", x=3, y=4, z=1/4+1/16, angle=180)
#'         g.p("coin_back", suit=4, x=3, y=4, z=1/4+1/8+1/16, angle=180)
#'         g.p("coin_back", suit=2, x=3, y=1, z=3/4+1/8, angle=90)
#'     }
#'
#'     # default piecepack, orthogonal projection
#'     draw_pp_diagram(cfg=pp_cfg())
#'
#'     options(opt) # Reset options
#'   }
#'   if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {
#'     # custom configuration, orthogonal projection
#'     grid::grid.newpage()
#'     dark_colorscheme <- list(suit_color="darkred,black,darkgreen,darkblue,black",
#'                          invert_colors.suited=TRUE, border_color="black", border_lex=2)
#'     traditional_ranks <- list(use_suit_as_ace=TRUE, rank_text=",a,2,3,4,5")
#'     cfg <- c(dark_colorscheme, traditional_ranks)
#'     draw_pp_diagram(cfg=pp_cfg(cfg))
#'   }
#'   if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {
#'     # custom configuration, oblique projection
#'     grid::grid.newpage()
#'     cfg3d <- list(width.pawn=0.75, height.pawn=0.75, depth.pawn=1,
#'                        dm_text.pawn="", shape.pawn="convex6", invert_colors.pawn=TRUE,
#'                        edge_color.coin="tan", edge_color.tile="tan")
#'     cfg <- pp_cfg(c(cfg, cfg3d))
#'     draw_pp_diagram(cfg=pp_cfg(cfg), op_scale=0.5)
#'   }
#' @seealso [pmap_piece()] which applies `pieceGrob()` over rows of a data frame.
#' @name grid.piece
NULL

pieceGrobHelper <- function(
	piece_side = "tile_back",
	suit = NA,
	rank = NA,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = NA,
	angle = 0,
	width = NA,
	height = NA,
	depth = NA,
	op_scale = 0,
	op_angle = 45,
	default.units = "npc",
	scale = 1,
	alpha = 1,
	type = "normal",
	name = "",
	bleed = FALSE
) {
	stopifnot(isFALSE(bleed) || op_scale < 0.0001)
	if (scale == 0 || alpha == 0) {
		return(nullGrob())
	}
	cfg <- as_pp_cfg(cfg)
	rank <- impute_rank(piece_side, rank, cfg)
	suit <- impute_suit(piece_side, suit, cfg)
	if (is.na(angle)) {
		angle <- 0
	}

	has_bleed <- !isFALSE(bleed)

	if (isTRUE(bleed)) {
		bleed <- unit(0.125, "in")
	}
	if (isFALSE(bleed)) {
		bleed <- unit(0, "in")
	}
	if (!is.unit(bleed)) {
		bleed <- unit(bleed, default.units)
	}

	if (is.na(width)) {
		width <- inch(cfg$get_width(piece_side, suit, rank))
	}
	if (is.na(height)) {
		height <- inch(cfg$get_height(piece_side, suit, rank))
	}
	if (is.na(depth)) {
		depth <- inch(cfg$get_depth(piece_side, suit, rank))
	}
	if (is.na(z)) {
		z <- 0.5 * depth
	}
	if (!is.unit(x)) {
		x <- unit(x, default.units)
	}
	if (!is.unit(y)) {
		y <- unit(y, default.units)
	}
	if (!is.unit(z)) {
		z <- unit(z, default.units)
	}
	if (!is.unit(width)) {
		width <- unit(width, default.units)
	}
	if (!is.unit(height)) {
		height <- unit(height, default.units)
	}
	if (!is.unit(depth)) {
		depth <- unit(depth, default.units)
	}
	width <- scale * width + 2 * bleed
	height <- scale * height + 2 * bleed
	depth <- scale * depth
	angle <- angle %% 360
	op_angle <- op_angle %% 360

	if (op_scale < 0.0001) {
		if (has_bleed) {
			grob <- cfg$get_grob_with_bleed(piece_side, suit, rank)
			if (hasName(grob, "bleed")) {
				grob$bleed <- bleed
			}
		} else {
			grob <- cfg$get_grob(piece_side, suit, rank, type)
		}
		cvp <- viewport(x, y, width, height, angle = angle)
		name <- paste0("piece_side", name)
		grob <- grid::editGrob(grob, vp = cvp)
	} else {
		grob <- cfg$get_op_grob(
			piece_side,
			suit,
			rank,
			x,
			y,
			z,
			angle,
			type,
			width,
			height,
			depth,
			op_scale,
			op_angle
		)
		name <- paste0("projected_piece", name)
	}

	# update name, cex, lex, alpha
	grob <- grid::editGrob(grob, name = name)
	if (hasName(grob, "scale")) {
		grob$scale <- scale
	} else {
		grob <- update_gp(grob, gp = gpar(cex = scale, lex = scale))
	}
	if (!nigh(alpha, 1)) {
		grob <- update_gp(grob, gp = gpar(alpha = alpha))
	}
	grob
}

#' @rdname grid.piece
#' @export
pieceGrob <- function(
	piece_side = "tile_back",
	suit = NA,
	rank = NA,
	cfg = getOption("piecepackr.cfg", pp_cfg()),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = NA,
	angle = 0,
	...,
	width = NA,
	height = NA,
	depth = NA,
	op_scale = getOption("piecepackr.op_scale", 0),
	op_angle = getOption("piecepackr.op_angle", 45),
	default.units = getOption("piecepackr.default.units", "npc"),
	envir = getOption("piecepackr.envir"),
	name = NULL,
	gp = NULL,
	vp = NULL,
	scale = 1,
	alpha = 1,
	type = "normal",
	bleed = FALSE
) {
	stopifnot(!(bleed && op_scale > 0))
	if (any(piece_side == "preview_layout")) {
		warn(
			paste0(
				'The "preview_layout" component is deprecated. ',
				'Use `ppdf::piecepack_preview() |> pmap_piece(cfg = cfg, default.units = "in")` instead.'
			),
			class = "deprecatedWarning"
		)
	}
	if (is_angle(angle)) {
		angle <- as.double(angle, "degrees")
	}
	if (is_angle(op_angle)) {
		op_angle <- as.double(op_angle, "degrees")
	}

	gTree(
		piece_side = piece_side,
		suit = suit,
		rank = rank,
		cfg = cfg,
		x = x,
		y = y,
		z = z,
		angle = angle,
		width = width,
		height = height,
		depth = depth,
		op_scale = op_scale,
		op_angle = op_angle,
		default.units = default.units,
		envir = envir,
		scale = scale,
		alpha = alpha,
		type = type,
		bleed = bleed,
		name = name,
		gp = gp,
		vp = vp,
		cl = c("piece", "pp_grobCoords")
	)
}

#' @export
makeContent.piece <- function(x) {
	nn <- max(lengths(list(
		x$piece_side,
		x$suit,
		x$rank,
		x$x,
		x$y,
		x$z,
		x$angle,
		x$width,
		x$height,
		x$depth,
		x$op_scale,
		x$op_angle,
		x$scale,
		x$alpha,
		x$type,
		x$bleed
	)))
	piece_side <- rep(x$piece_side, length.out = nn)
	suit <- rep(x$suit, length.out = nn)
	rank <- rep(x$rank, length.out = nn)
	xc <- rep(x$x, length.out = nn)
	yc <- rep(x$y, length.out = nn)
	zc <- rep(x$z, length.out = nn)
	angle <- rep(x$angle, length.out = nn)

	width <- rep(x$width, length.out = nn)
	height <- rep(x$height, length.out = nn)
	depth <- rep(x$depth, length.out = nn)

	op_scale <- rep(x$op_scale, length.out = nn)
	op_angle <- rep(x$op_angle, length.out = nn)

	scale <- rep(x$scale, length.out = nn)
	alpha <- rep(x$alpha, length.out = nn)
	type <- rep(x$type, length.out = nn)

	bleed <- rep(x$bleed, length.out = nn)

	cfg <- get_cfg(x$cfg, x$envir)
	cfg <- rep(c(cfg), length.out = nn)

	gl <- gList()
	for (i in seq(nn)) {
		name <- paste0(".", i)
		gl[[i]] <- pieceGrobHelper(
			piece_side[i],
			suit[i],
			rank[i],
			cfg[[i]],
			xc[i],
			yc[i],
			zc[i],
			angle[i],
			width[i],
			height[i],
			depth[i],
			op_scale[i],
			op_angle[i],
			x$default.units,
			scale[i],
			alpha[i],
			type[i],
			name,
			bleed[i]
		)
	}
	setChildren(x, gl)
}

get_cfg <- function(cfg = pp_cfg(), envir = NULL) {
	if (is_pp_cfg(cfg)) {
		cfg <- cfg
	} else if (is.character(cfg)) {
		if (!is.null(envir)) {
			envir <- as.environment(envir)
			cfg <- lapply(cfg, function(cc) as_pp_cfg(envir[[cc]]))
		} else {
			cfg <- lapply(cfg, function(cc) as_pp_cfg(dynGet(cc)))
		}
		if (length(cfg) == 1) {
			cfg <- cfg[[1]]
		}
	} else {
		if (is.list(cfg)) {
			if (!all(sapply(cfg, inherits, "pp_cfg"))) {
				cfg <- pp_cfg(cfg)
			}
		} else {
			abort("Don't know how to parse cfg argument") # nocov
		}
	}
	cfg
}

#' @rdname grid.piece
#' @export
grid.piece <- function(
	piece_side = "tile_back",
	suit = NA,
	rank = NA,
	cfg = getOption("piecepackr.cfg", pp_cfg()),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = NA,
	angle = 0,
	...,
	width = NA,
	height = NA,
	depth = NA,
	op_scale = getOption("piecepackr.op_scale", 0),
	op_angle = getOption("piecepackr.op_angle", 45),
	default.units = getOption("piecepackr.default.units", "npc"),
	envir = getOption("piecepackr.envir"),
	name = NULL,
	gp = NULL,
	draw = TRUE,
	vp = NULL,
	scale = 1,
	alpha = 1,
	type = "normal",
	bleed = FALSE
) {
	grob <- pieceGrob(
		piece_side,
		suit,
		rank,
		cfg,
		x,
		y,
		z,
		angle,
		width = width,
		height = height,
		depth = depth,
		op_scale = op_scale,
		op_angle = op_angle,
		default.units = default.units,
		envir = envir,
		name = name,
		gp = gp,
		vp = vp,
		scale = scale,
		alpha = alpha,
		type = type,
		bleed = bleed
	)
	if (draw) {
		grid.draw(grob)
		invisible(grob)
	} else {
		grob
	}
}

#' @import grid
#' @importFrom affiner as_coord2d as_coord3d as_ellipse2d as_plane3d as_polygon2d as_segment2d convex_hull2d distance2d distance3d degrees has_overlap2d is_angle painter_depth painter_order regular_ngon_polygon2d
#' @importFrom grDevices col2rgb dev.capabilities dev.list dev.new devAskNewPage rgb
#' @importFrom R6 R6Class
#' @importFrom rlang .data abort inform warn %||% check_dots_empty
#' @importFrom stringr str_count str_glue str_pad str_split
#' @importFrom tibble tibble
#' @importFrom utils hasName head packageDescription packageVersion tail
NULL

# https://en.wikipedia.org/wiki/Spherical_circle
# BC^2 = AB^2 + AC^2 <=> AB^2 = BC^2 - AC^2
# where BC = 1.0 diameter marble radius = 0.5 and AC = BC - holed board depth = 0.5 - 0.25
RADIUS_BOARD_HOLES <- sqrt(0.5^2 - 0.25^2)

# Timeline of {piecepackr} archival:
# 2025-10-09 {pieceepackr} submitted (for non-reproducible "donttest" issue)
# 2025-10-09 automatic checks not passed due to "donttest" NOTE of
#   Examples with CPU times > 2.5 times elapsed time
# 2025-10-11 emailed CRAN that this rejection is a false positive
# 2025-10-15 {piecepackr} archived (without reply of my email by CRAN)

# Use for examples that
# may take more than 5 seconds
# or with CPU time > 2.5 times elapsed time on CRAN servers
# theoretically you "should" use \donttest{} but
# in practice this led to package archival without response by CRAN
donttest <- function() {
	interactive() ||
		identical(Sys.getenv("NOT_CRAN"), "true") ||
		identical(Sys.getenv("IN_PKGDOWN"), "true")
}

add_cfg <- function(df, cfg = pp_cfg(), envir = NULL) {
	df <- tibble::as_tibble(df)
	if (hasName(df, "cfg")) {
		df$cfg <- lapply(df$cfg, get_cfg, envir)
	} else {
		df$cfg <- rep(c(get_cfg(cfg, envir)), length.out = nrow(df))
	}
	df
}

add_field <- function(df, key, value) {
	if (hasName(df, key)) {
		df[[key]] <- ifelse(is.na(df[[key]]), value, df[[key]])
	} else {
		df[[key]] <- value
	}
	df
}

add_measurements <- function(df) {
	df <- add_field(df, "width", an_pmap(df, gwh))
	df <- add_field(df, "height", an_pmap(df, ghh))
	df <- add_field(df, "depth", an_pmap(df, gdh))
	df <- add_field(df, "angle", 0)
	df
}

an_pmap <- function(...) as.numeric(purrr::pmap(...))
gwh <- function(piece_side, cfg, ..., suit = 1, rank = 1) cfg$get_width(piece_side, suit, rank)
ghh <- function(piece_side, cfg, ..., suit = 1, rank = 1) cfg$get_height(piece_side, suit, rank)
gdh <- function(piece_side, cfg, ..., suit = 1, rank = 1) cfg$get_depth(piece_side, suit, rank)

#' Oblique projection helper function
#'
#' `op_transform()` guesses `z` coordinates and
#' sorting order to more easily
#' make 3D graphics with [pmap_piece()].
#' `marbles_transform()` is a wrapper around `op_transform()` that handles the special case
#' of stacking 1" marbles in a pyramid with a square base on a holed board.
#'
#' The heuristics used to generate guesses
#' for `z` coordinates and sorting order
#' aren't guaranteed to work in every case.
#' In some cases you may get better sorting results
#' by changing the `op_angle` or the dimensions of pieces.
#' @seealso \url{https://trevorldavis.com/piecepackr/3d-projections.html} for more details
#'   and examples of oblique projections in `piecepackr`.
#' @param df A data frame with coordinates and dimensions in inches
#' @param ... Ignored by `op_transform()`.  `marbles_transform()` passes to `op_transform()`.
#' @param cfg Piecepack configuration list or [pp_cfg()] object,
#'        a list of [pp_cfg()] objects,
#'        or a character vector of the names of [pp_cfg()] objects contained in `envir`.
#' @param envir Environment (or named list) containing configuration list(s).
#' @param op_angle Intended oblique projection angle (used for re-sorting)
#' @param pt_thickness Thickness of pyramid tip i.e. value to add to the z-value of a pyramid top
#'                  if it is a (weakly) smaller ranked pyramid (top)
#'                  placed on top of a larger ranked pyramid (top).
#' @param as_top Character vector of components whose \dQuote{side}
#'               should be converted to \dQuote{top} e.g. `"pawn_face"`.
#' @param cfg_class Either `"list"` (default) or `"character"`.
#'                  Desired class of the `cfg` column in the returned tibble.
#'                  `"list"` is more efficient for use with `pmap_piece()` but
#'                  `geom_piece()` needs `"character"`.
#' @return A tibble with extra columns added
#'         and re-sorted rows
#' @examples
#' df <- tibble::tibble(piece_side="tile_back",
#'                      x=c(2,2,2,4,6,6,4,2,5),
#'                      y=c(4,4,4,4,4,2,2,2,3))
#' cfg <- game_systems()$piecepack
#' pmap_piece(df, op_angle=135, trans=op_transform,
#'            op_scale=0.5, default.units="in", cfg=cfg)
#' @export
op_transform <- function(
	df,
	...,
	cfg = getOption("piecepackr.cfg", pp_cfg()),
	envir = getOption("piecepackr.envir"),
	op_angle = getOption("piecepackr.op_angle", 45),
	pt_thickness = 0.01,
	as_top = character(0),
	cfg_class = "list"
) {
	if (is_angle(op_angle)) {
		op_angle <- as.double(op_angle, "degrees")
	}
	for (ps in as_top) {
		indices <- which(df$piece_side == ps)
		piece <- get_piece(ps)
		df$piece_side[indices] <- gsub(ps, paste0(piece, "_top"), df$piece_side[indices])
	}
	df <- add_3d_info(
		df,
		cfg = cfg,
		envir = envir,
		pt_thickness = pt_thickness,
		cfg_class = cfg_class
	)
	df <- op_sort(df, op_angle = op_angle)
	df
}

#' @rdname op_transform
#' @export
marbles_transform <- function(df, ...) {
	# Make sure all pieces are from `marbles` with first piece `board` and the rest 1" marble `bits`
	stopifnot(
		all(df$cfg == "marbles"),
		grepl("^board_", df[1L, ]$piece_side),
		all(grepl("^bit_", df[-1L, ]$piece_side)),
		all(df[-1L, ]$rank == 9L)
	)
	df <- op_transform(df, ...)
	levels <- as.integer(as.factor(df$z)) # note first level is board, rest bits
	nlevels <- length(unique(levels))
	# Board holes sized so first layer of marbles touches ground
	# so subtract board height from all bits' height
	df[-1L, ]$z <- df[-1L, ]$z - df[1L, ]$depth
	df$zt <- df$z + 0.5 * df$depth
	# > Each time you add a layer of balls to the bottom of the pyramid, the height increases by d√1/2
	# https://www.had2know.org/academics/pyramid-of-balls-height-calculator.html
	if (nlevels > 2L) {
		id_prev <- which(levels == 2L)
		for (i in seq.int(3L, nlevels)) {
			id <- which(levels == i)
			df$zt[id] <- max(df$zt[id_prev]) + sqrt(0.5) * df$depth[id]
			df$z[id] <- df$zt[id] - 0.5 * df$depth[id]
			id_prev <- id
		}
	}
	df$zb <- df$z - 0.5 * df$depth
	df
}

add_3d_info <- function(
	df,
	...,
	cfg = pp_cfg(),
	envir = NULL,
	pt_thickness = 0.01,
	cfg_class = "list"
) {
	# Do more stuff if units aren't in inches?
	if (cfg_class == "character") {
		cfg_character <- get_cfg_character(df, cfg)
	}
	df <- add_cfg(df, cfg, envir)
	df <- add_measurements(df)
	df <- add_bounding_box(df)
	df <- add_z(df, pt_thickness = pt_thickness)
	df <- add_field(df, "zt", df$z + 0.5 * df$depth)
	df <- add_field(df, "zb", df$z - 0.5 * df$depth)
	if (cfg_class == "character") {
		df$cfg <- cfg_character
	}
	df
}

get_cfg_character <- function(df, cfg) {
	if (hasName(df, "cfg")) {
		df$cfg
	} else if (is.null(cfg)) {
		rep_len("piecepack", nrow(df))
	} else if (is.character(cfg)) {
		rep_len(cfg, nrow(df))
	} else {
		abort("Can't compute \"character\" `cfg_class`")
	}
}

op_sort <- function(df, ..., op_angle = 45) {
	op_angle <- op_angle %% 360
	if ((0 <= op_angle) && (op_angle < 90)) {
		df <- df[order(df$zt, -df$yb, -df$xl), ]
	} else if ((90 <= op_angle) && (op_angle < 180)) {
		df <- df[order(df$zt, -df$yb, df$xr), ]
	} else if ((180 <= op_angle) && (op_angle < 270)) {
		df <- df[order(df$zt, df$yt, df$xr), ]
	} else {
		df <- df[order(df$zt, df$yt, -df$xl), ]
	}
	df
}

# Axis-Aligned Bounding Box (AABB)
add_bounding_box <- function(df) {
	zeros <- rep_len(0, nrow(df))
	ones <- rep_len(1, nrow(df))
	ll <- npc_to_in(as_coord2d(zeros, zeros), df$x, df$y, df$width, df$height, df$angle)
	ul <- npc_to_in(as_coord2d(zeros, ones), df$x, df$y, df$width, df$height, df$angle)
	ur <- npc_to_in(as_coord2d(ones, ones), df$x, df$y, df$width, df$height, df$angle)
	lr <- npc_to_in(as_coord2d(ones, zeros), df$x, df$y, df$width, df$height, df$angle)
	df$xll <- ll$x
	df$yll <- ll$y
	df$xul <- ul$x
	df$yul <- ul$y
	df$xlr <- lr$x
	df$ylr <- lr$y
	df$xur <- ur$x
	df$yur <- ur$y
	df$xl <- op_round(pmin(df$xll, df$xul, df$xur, df$xlr))
	df$xr <- op_round(pmax(df$xll, df$xul, df$xur, df$xlr))
	df$yb <- op_round(pmin(df$yll, df$yul, df$yur, df$ylr))
	df$yt <- op_round(pmax(df$yll, df$yul, df$yur, df$ylr))
	df
}

op_round <- function(x) round(x, 9)

do_ranges_overlap <- function(l1, r1, l2, r2) {
	(less_than_equal(l1, l2) & less_than(l2, r1)) |
		(less_than_equal(l2, l1) & less_than(l1, r2))
}

which_AABB_overlap <- function(dfi, dfs) {
	rev(which(
		do_ranges_overlap(dfi$xl, dfi$xr, dfs$xl, dfs$xr) &
			do_ranges_overlap(dfi$yb, dfi$yt, dfs$yb, dfs$yt)
	))
}

less_than <- function(x, y) 1e-6 < y - x # in case of trigonometric precision issues
less_than_equal <- function(x, y) 0 < y - x + 1e-6 # in case of trigonometric precision issues

add_z <- function(df, pt_thickness = 0.01) {
	shapes <- get_shapes(df)
	zp <- 0.5 * df$depth
	for (i in seq(length.out = nrow(df))) {
		dfi <- df[i, ]
		dfs <- df[seq_len(i - 1L), ]
		zc <- zp[i]
		for (j in which_AABB_overlap(dfi, dfs)) {
			if (has_overlap2d(shapes[[i]], shapes[[j]])) {
				zc0 <- compute_z(df, zp, i, j, pt_thickness)
				if (zc0 > zc) zc <- zc0
			}
		}
		zp[i] <- zc
	}
	df <- add_field(df, "z", zp)
	df
}
# @param df Data frame
# @param zp Depths
# @param i Index of top piece
# @param j Index of top overlapping piece underneath
# @param pt_thickness Thickness of tip of pyramid
compute_z <- function(df, zp, i, j, pt_thickness = 0.01) {
	if (all(df$piece_side[c(i, j)] == "pyramid_top")) {
		if (df$rank[i] <= df$rank[j]) {
			# top piece is (weakly) smaller
			zp[j] + 0.5 * df$depth[j] + pt_thickness - 0.5 * df$depth[i]
		} else {
			# top piece is (strictly) bigger
			zp[j] - 0.5 * df$depth[j] + 0.5 * df$depth[i]
		}
	} else {
		zp[j] + 0.5 * df$depth[j] + 0.5 * df$depth[i]
	}
}

get_shapes <- function(df) {
	shapes <- vector("list", nrow(df))
	for (ii in seq(length.out = nrow(df))) {
		dfi <- df[ii, ]
		cfg <- df$cfg[[ii]]
		piece_side <- df$piece_side[ii]
		suit <- ifelse(hasName(df, "suit"), df$suit[ii], NA)
		rank <- ifelse(hasName(df, "rank"), df$rank[ii], NA)
		opt <- try(cfg$get_piece_opt(piece_side, suit, rank), silent = TRUE)
		if (
			inherits(opt, "try-error") || opt$shape %in% c("rect", "halma", "meeple", "roundrect")
		) {
			shapes[[ii]] <- as_polygon2d(
				as_coord2d(
					x = c(dfi$xll, dfi$xul, dfi$xur, dfi$xlr),
					y = c(dfi$yll, dfi$yul, dfi$yur, dfi$ylr)
				),
				convex = TRUE
			)
		} else if (opt$shape == "circle") {
			shapes[[ii]] <- as_ellipse2d(
				as_coord2d(dfi$x, dfi$y),
				r = min(dfi$width / 2, dfi$height / 2)
			)
		} else {
			label <- opt$shape
			if (grepl("^concave", label)) {
				label <- gsub("concave", "convex", label)
			}
			shape <- pp_shape(
				label,
				opt$shape_t,
				opt$shape_r,
				opt$back,
				width = opt$shape_w,
				height = opt$shape_h
			)
			xy_c <- npc_to_in(
				as_coord2d(shape$npc_coords),
				dfi$x,
				dfi$y,
				dfi$width,
				dfi$height,
				dfi$angle
			)
			shapes[[ii]] <- as_polygon2d(xy_c, convex = TRUE)
		}
	}
	shapes
}

#' Create graphics using data frame input
#'
#' `pmap_piece()` operates on the rows of a data frame
#'     applying `.f` to each row (usually `grid.piece`).
#'
#' `pmap_piece()` differs from `purrr::pmap()` in a few ways:
#' \enumerate{
#'    \item{If `cfg` and/or `envir` are missing attempts to set reasonable defaults.}
#'    \item{If not `NULL` will first apply function `trans` to `.l`.}
#'    \item{If the output of `.f` is a grid grob object then `pmap_piece`
#'          will return a `gTree` object with
#'          specified `name`, `gp`, and `vp` values and if `draw` is true draw it.}
#'    \item{If `.l` lacks a `name` column or if `name` column is non-unique
#'          attempts to generate a reasonable new default `name` column
#'          and use that to name the return `gTree` children  or `list` values.}
#'  }
#' @inheritParams grid.piece
#' @param .l A list of vectors, such as a data frame. The length of `.l`
#'           determines the number of arguments that `.f`
#'           will be called  with. List names will be used if present.
#' @param .f Function to be applied to `.l` after adjustments to
#'        `cfg` and `envir` and the application of `trans`.
#'        Usually [grid.piece()], [pieceGrob()], [piece3d()], or [piece()].
#' @param trans Function to modify `.l` before drawing.
#'        Default (`NULL`) is to not modify `.l`.  `op_transform`
#'        can help with using an oblique projection (i.e. `op_scale` over 0).
#' @param ... Extra arguments to pass to `.f`.
#' @examples
#'   if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {
#'        dark_colorscheme <- list(suit_color="darkred,black,darkgreen,darkblue,black",
#'                             invert_colors.suited=TRUE, border_color="black", border_lex=2)
#'        traditional_ranks <- list(use_suit_as_ace=TRUE, rank_text=",a,2,3,4,5")
#'        cfg3d <- list(width.pawn=0.75, height.pawn=0.75, depth.pawn=1,
#'                           dm_text.pawn="", shape.pawn="convex6", invert_colors.pawn=TRUE,
#'                           edge_color.coin="tan", edge_color.tile="tan")
#'        cfg <- pp_cfg(c(dark_colorscheme, traditional_ranks, cfg3d))
#'        grid::grid.newpage()
#'        df_tiles <- data.frame(piece_side="tile_back", x=0.5+c(3,1,3,1), y=0.5+c(3,3,1,1),
#'                               suit=NA, angle=NA, z=NA, stringsAsFactors=FALSE)
#'        df_coins <- data.frame(piece_side="coin_back", x=rep(4:1, 4), y=rep(4:1, each=4),
#'                               suit=1:16%%2+rep(c(1,3), each=8),
#'                               angle=rep(c(180,0), each=8), z=1/4+1/16, stringsAsFactors=FALSE)
#'        df <- rbind(df_tiles, df_coins)
#'        pmap_piece(df, cfg=cfg, op_scale=0.5, default.units="in")
#'   }
#' @seealso [render_piece()] is a higher-level function that wraps this function.
#' @export
pmap_piece <- function(
	.l,
	.f = pieceGrob,
	...,
	cfg = getOption("piecepackr.cfg"),
	envir = getOption("piecepackr.envir"),
	trans = getOption("piecepackr.trans"),
	draw = TRUE,
	name = NULL,
	gp = NULL,
	vp = NULL
) {
	if (length(.l) == 0L || (!is.null(nrow(.l)) && nrow(.l) == 0L)) {
		return(list())
	}
	ce <- default_cfg_envir(cfg, envir)
	cfg <- ce$cfg
	envir <- ce$envir

	.l <- update_name(.l)
	if (is.function(trans)) {
		.l <- trans(.l, ..., cfg = cfg, envir = envir)
		.l <- update_name(.l)
	}
	if (hasName(.l, "cfg")) {
		ll <- purrr::pmap(.l, .f, ..., envir = envir, draw = FALSE)
	} else {
		ll <- purrr::pmap(.l, .f, ..., cfg = cfg, envir = envir, draw = FALSE)
	}
	if (length(ll) && all(sapply(ll, is.grob))) {
		grob <- gTree(children = as.gList(ll), name = name, gp = gp, vp = vp, cl = "pmap_piece")
		if (draw) {
			grid.draw(grob)
		}
		invisible(grob)
	} else {
		names(ll) <- .l$name
		invisible(ll)
	}
}

update_name <- function(.l) {
	if (hasName(.l, "name")) {
		if (!is.character(.l$name)) {
			.l$name <- as.character(.l$name)
		}
		if (anyDuplicated(.l$name)) {
			warn("the name column in .l is not unique, generating new name column")
		} else {
			return(.l)
		}
	}
	if (hasName(.l, "id")) {
		.l$name <- paste0("piece.", as.character(.l$id))
		if (anyDuplicated(.l$name)) {
			warn("the id column in .l is not unique, generating new name column")
		} else {
			return(.l)
		}
	}
	.l$name <- paste0("piece.", as.character(seq(length(.l[[1]]))))
	.l
}

#' @export
grobPoints.pmap_piece <- function(x, closed, ...) {
	grobCoords(x, closed)
}

#' @export
grobCoords.pmap_piece <- function(x, closed, ...) {
	gc <- lapply(x$children, grobCoords, closed = closed)
	f <- function(x, y) gridGeometry::polyclip(x, y, "union")
	Reduce(f, gc)
}

as.gList <- function(ll) {
	gl <- gList()
	for (ii in seq(ll)) {
		gl[[ii]] <- ll[[ii]]
	}
	gl
}

default_cfg_envir <- function(cfg = NULL, envir = NULL) {
	if (is.null(cfg) && is.null(envir)) {
		cfg <- pp_cfg()
		envir <- game_systems()
	} else if (is.null(cfg)) {
		# and !is.null(envir)
		if (hasName(envir, "piecepack")) {
			cfg <- envir[["piecepack"]]
		} else {
			cfg <- pp_cfg()
		}
	} else if (is.null(envir)) {
		# and !is.null(cfg)
		cfg <- get_cfg(cfg)
		envir <- game_systems()
		envir[["piecepack"]] <- cfg
	}
	list(cfg = cfg, envir = envir)
}

test_that("crop_marks work as expected", {
	skip_if_not_installed("vdiffr")
	library("vdiffr")

	cfg <- pp_cfg(list(mat_color = "pink", mat_width = 0.05, border_color = NA))
	df <- data.frame(
		piece_side = "tile_face",
		suit = 2,
		rank = 2,
		x = 2,
		y = 2,
		angle = 0,
		stringsAsFactors = FALSE
	)
	expect_doppelganger("crop-mark-outside", {
		pmap_piece(df, grid.cropmark, cfg = cfg, default.units = "in")
		pmap_piece(df, grid.piece, cfg = cfg, default.units = "in", bleed = TRUE)
	})

	df <- data.frame(
		piece_side = "coin_back",
		suit = 2,
		rank = 2,
		x = 2,
		y = 2,
		angle = 0,
		stringsAsFactors = FALSE
	)
	expect_doppelganger("crop-mark-outside-x2", {
		pmap_piece(
			df,
			grid.cropmark,
			cfg = cfg,
			default.units = "in",
			bleed = TRUE,
			scale = 2,
			cm_select = "1357"
		)
		pmap_piece(df, grid.piece, cfg = cfg, default.units = "in", bleed = TRUE, scale = 2)
	})
})

#' @section Package options:
#' The following piecepackr function arguments may be set globally via [base::options()]:
#'  \describe{
#'    \item{piecepackr.at.inform}{If `FALSE` turns off messages when affine transformation support not detected in active graphics device.}
#'    \item{piecepackr.cfg}{Sets a new default for the `cfg` argument}
#'    \item{piecepackr.check.cairo}{If `FALSE` don't check the version of cairo}
#'    \item{piecepackr.default.units}{Sets a new default for the `default.units` argument}
#'    \item{piecepackr.envir}{Sets a new default for the `envir` argument}
#'    \item{piecepackr.fs.inform}{If `FALSE` turns off messages when stroking and filling path support not detected in active graphics device.}
#'    \item{piecepackr.metadata.inform}{If `FALSE` turns off messages when support for embedding metadata not detected.}
#'    \item{piecepackr.op_angle}{Sets a new default for the `op_angle` argument}
#'    \item{piecepackr.op_scale}{Sets a new default for the `op_scale` argument}
#'    \item{piecepackr.rgr.inform}{If `FALSE` turns off messages when radial gradient support not detected in active graphics device.}
#'    \item{piecepackr.trans}{Sets a new default for the `trans` argument}
#'  }
#' @seealso Useful links:
#'
#'   * blog: \url{https://trevorldavis.com/piecepackr/}
#'   * pkgdown: \url{https://trevorldavis.com/R/piecepackr/}
#'   * forum: \url{https://groups.google.com/forum/#!forum/piecepackr}
#'   * Report bugs: \url{https://github.com/piecepackr/piecepackr/issues}
#'
"_PACKAGE"

#' Animate board game pieces
#'
#' `animate_piece()` animates board game pieces.
#' @param dfs A list of data frames of game data to plot.
#' @param file Filename to save animation unless `NULL`
#'             in which case it uses the current graphics device.
#' @param ... Arguments to [pmap_piece()]
#' @param .f Low level graphics function to use e.g. [grid.piece()], [piece3d()], [piece()], or [piece_mesh()].
#' @param cfg A piecepackr configuration list
#' @param envir Environment (or named list) of piecepackr configuration lists
#' @param width Width of animation (in inches).  Inferred by default.
#' @param height Height of animation (in inches).  Inferred by default.
#' @param ppi Resolution of animation in pixels per inch.
#'            By default set so image max 600 pixels wide or tall.
#' @param annotate If `TRUE` or `"algebraic"` annotate the plot
#'                  with \dQuote{algrebraic} coordinates,
#'                 if `FALSE` or `"none"` don't annotate,
#'                 if `"cartesian"` annotate the plot with \dQuote{cartesian} coordinates.
#' @param annotation_scale Multiplicative factor that scales (stretches) any annotation coordinates.
#'                         By default uses `attr(df, "scale_factor") %||% 1`.
#' @param open_device If `TRUE` open a new graphics device otherwise draw in the active graphics.
#' @param n_transitions Integer, if over zero (the default)
#'                how many transition frames to add between moves.
#' @param n_pauses Integer, how many paused frames per completed move.
#' @param fps Double, frames per second.
#' @param xbreaks,ybreaks Subset (of integers) to provide axis labels for if `annotate` is `TRUE`.
#'                        If `NULL` infer a reasonable choice.
#' @return Nothing, as a side effect creates an animation.
#' @examples
#'   # Basic tic-tac-toe animation
#'   dfs <- list()
#'   d.frame <- function(piece_side = "bit_back", ..., rank = 1L) {
#'                  data.frame(piece_side = piece_side, ..., rank = rank,
#'                             cfg = "checkers1", stringsAsFactors = FALSE)
#'   }
#'   df <- d.frame("board_back", suit = 2L, rank = 3L, x = 2, y = 2, id = "1")
#'   dfs[[1L]] <- df
#'   df <- rbind(df, d.frame(suit = 1L, x = 2, y = 2, id = "2"))
#'   dfs[[2L]] <- df
#'   df <- rbind(df, d.frame(suit = 2L, x = 1, y = 2, id = "3"))
#'   dfs[[3L]] <- df
#'   df <- rbind(df, d.frame(suit = 1L, x = 3, y = 1, id = "4"))
#'   dfs[[4L]] <- df
#'   df <- rbind(df, d.frame(suit = 2L, x = 1, y = 3, id = "5"))
#'   dfs[[5L]] <- df
#'   df <- rbind(df, d.frame(suit = 1L, x = 1, y = 1, id = "6"))
#'   dfs[[6L]] <- df
#'   df <- rbind(df, d.frame(suit = 2L, x = 3, y = 3, id = "7"))
#'   dfs[[7L]] <- df
#'   df <- rbind(df, d.frame(suit = 1L, x = 2, y = 1, id = "8"))
#'   dfs[[8L]] <- df
#'
#'   ## Press enter to walk through moves in a "game" in new graphics device
#'   if (interactive()) {
#'       animate_piece(dfs, file = NULL)
#'   }
#'
#'   ## Save GIF of game with animation transitions
#'   \dontrun{# May take more than 5 seconds on CRAN servers
#'   if ((requireNamespace("animation", quietly = TRUE) ||
#'        requireNamespace("gifski", quietly = TRUE)) &&
#'       requireNamespace("tweenr", quietly = TRUE)) {
#'       file <- tempfile("tic-tac-toe", fileext = ".gif")
#'       animate_piece(dfs, file = file,
#'                     n_transitions = 5L, n_pauses = 2L, fps = 9)
#'   }
#'   }
#'
#' @export
animate_piece <- function(
	dfs,
	file = "animation.gif",
	...,
	.f = piecepackr::grid.piece,
	cfg = getOption("piecepackr.cfg", NULL),
	envir = getOption("piecepackr.envir", game_systems("sans")),
	width = NULL,
	height = NULL,
	ppi = NULL,
	annotate = TRUE,
	annotation_scale = NULL,
	open_device = TRUE,
	n_transitions = 0L,
	n_pauses = 1L,
	fps = n_transitions + n_pauses,
	xbreaks = NULL,
	ybreaks = NULL
) {
	if (n_transitions > 0L) {
		assert_suggested("tweenr")
	}

	ce <- default_cfg_envir(cfg, envir)
	cfg <- ce$cfg
	envir <- ce$envir

	ranges <- lapply(dfs, aabb_piece, cfg = cfg, envir = envir, ...)
	if (n_transitions > 0L) {
		dfs <- get_tweened_dfs(dfs, n_transitions, n_pauses, ..., cfg = cfg, envir = envir)
	} else if (n_pauses != 1) {
		dfs <- rep(dfs, each = n_pauses)
	}

	#### Add grid and comment annotations
	xmax_op <- max(sapply(ranges, function(r) r$x_op[2]), na.rm = TRUE)
	ymax_op <- max(sapply(ranges, function(r) r$y_op[2]), na.rm = TRUE)
	xmin_op <- min(sapply(ranges, function(r) r$x_op[1]), na.rm = TRUE)
	ymin_op <- min(sapply(ranges, function(r) r$y_op[1]), na.rm = TRUE)
	xmax <- max(sapply(ranges, function(r) r$x[2]), na.rm = TRUE)
	ymax <- max(sapply(ranges, function(r) r$y[2]), na.rm = TRUE)
	xoffset <- min2offset(xmin_op)
	yoffset <- min2offset(ymin_op)
	if (is.null(width)) {
		width <- xmax_op + xoffset + 0.50
	}
	if (is.null(height)) {
		height <- ymax_op + yoffset + 0.50
	}
	m <- max(width, height)
	if (is.null(ppi)) {
		ppi <- round(600 / m, 0)
	}
	# mp4 needs even height / weight
	height <- ceiling(ppi * height)
	width <- ceiling(ppi * width)
	height <- height + (height %% 2)
	width <- width + (width %% 2)
	plot_fn <- plot_fn_helper(
		.f,
		xmax,
		ymax,
		xoffset,
		yoffset,
		width,
		height,
		m,
		ppi,
		envir,
		annotate,
		annotation_scale,
		xbreaks,
		ybreaks
	)
	animation_fn(file, open_device)(lapply(dfs, plot_fn, ...), file, width, height, 1 / fps, ppi)
	invisible(NULL)
}
#### How to handle empty tibbles??
animation_fn <- function(file, open_device = TRUE) {
	if (is.null(file)) {
		function(expr, file, width, height, delay, res) {
			if (open_device) {
				dev.new(width = width / res, height = height / res, unit = "in", noRstudioGD = TRUE)
			}
			devAskNewPage(TRUE)
			eval(expr)
			devAskNewPage(getOption("device.ask.default"))
		}
	} else if (grepl(".html$", file)) {
		assert_suggested("animation")
		function(expr, file, width, height, delay, res) {
			animation::saveHTML(
				expr,
				htmlfile = file,
				interval = delay,
				img.name = file,
				ani.height = height,
				ani.width = width,
				ani.res = res,
				ani.dev = "png",
				ani.type = "png",
				title = "Animated game",
				verbose = FALSE
			)
		}
	} else if (grepl(".gif$", file)) {
		if (requireNamespace("gifski", quietly = TRUE)) {
			function(expr, file, width, height, delay, res) {
				gifski::save_gif(expr, file, width, height, delay, res = res, progress = FALSE)
			}
		} else if (requireNamespace("animation", quietly = TRUE)) {
			function(expr, file, width, height, delay, res) {
				animation::saveGIF(
					expr,
					movie.name = file,
					interval = delay,
					img.name = file,
					ani.height = height,
					ani.width = width,
					ani.res = res,
					ani.dev = "png",
					ani.type = "png"
				)
			}
		} else {
			abort(
				c(
					"At least one the suggested packages 'gifski' or 'animation' is required to use 'animate_game()'.",
					i = "Use 'install.packages(\"gifski\")' and/or 'install.packages(\"animation\")' to install them."
				),
				class = "piecepackr_suggested_package"
			)
		}
	} else if (grepl(".bmp$|.jpg$|.jpeg$|.png$|.tiff$", file)) {
		stopifnot(has_c_integer_format(file))
		function(expr, file, width, height, delay, res) {
			pp_device(file, width = width / res, height = height / res, res = res)
			eval(expr)
			grDevices::dev.off()
		}
	} else {
		assert_suggested("animation")
		function(expr, file, width, height, delay, res) {
			animation::saveVideo(
				expr,
				video.name = file,
				interval = delay,
				img.name = file,
				ani.height = height,
				ani.width = width,
				ani.res = res,
				ani.dev = "png",
				ani.type = "png",
				title = "Animated game",
				verbose = FALSE
			)
		}
	}
}

get_tweened_dfs <- function(dfs, n_transitions = 0L, n_pauses = 1L, ...) {
	n <- length(dfs)
	if (n < 2) {
		return(rep(dfs, n_pauses))
	}
	new_dfs <- list()
	for (i in seq_len(n - 1)) {
		new_dfs <- append(new_dfs, rep(dfs[i], n_pauses))
		new_dfs <- append(new_dfs, tween_dfs(dfs[[i]], dfs[[i + 1]], n_transitions))
	}
	new_dfs <- append(new_dfs, rep(dfs[n], n_pauses))
	for (i in seq_along(new_dfs)) {
		attr(new_dfs[[i]], "scale_factor") <- attr(dfs[[1]], "scale_factor")
	}
	new_dfs
}

tween_dfs <- function(df1, df2, n_transitions = 0L) {
	stopifnot(names(df1) == names(df2))
	df_id_cfg <- get_id_cfg(df1, df2)
	if (nrow(df1) == 0 && nrow(df2) == 0) {
		return(rep(list(df1), n_transitions))
	}
	dfs <- init_dfs(df1, df2)
	df <- tweenr::tween_state(
		dfs[[1]],
		dfs[[2]],
		ease = "cubic-in-out",
		nframes = n_transitions + 2L,
		id = .data$id
	)
	df <- merge(df, df_id_cfg, by = "id", all.x = TRUE, sort = FALSE)
	id_frames <- as.list(seq.int(max(df$.frame)))
	l <- lapply(id_frames, function(id_frame) df[which(df$.frame == id_frame), , drop = FALSE])
	l <- head(l, -1L)
	l <- tail(l, -1L)
	l
}

init_dfs <- function(df1, df2) {
	if (nrow(df1) == 0L) {
		df1 <- df2 <- get_tweenr_df(df2)
		df1$scale <- 0
		return(df1, df2)
	}
	if (nrow(df2) == 0L) {
		df1 <- df2 <- get_tweenr_df(df1)
		df2$scale <- 0
		return(df1, df2)
	}
	df1i <- get_tweenr_df(df1)
	df2i <- get_tweenr_df(df2)
	# try to insert pieces 'removed' from `df1` back in order in `df2`
	# with a scale of 0
	df2 <- df2i
	df1_anti <- df1i[which(match(df1i$id, df2i$id, 0L) == 0L), , drop = FALSE]
	df1_anti$scale <- rep_len(0, nrow(df1_anti))
	while (nrow(df1_anti)) {
		row <- df1_anti[1L, ]
		df1_anti <- df1_anti[-1L, ]
		prev_id <- find_good_prev_id(row, df1i, df2i)
		df2 <- insert_df(df2, row, prev_id)
	}
	# for now add `df2`'s 'new' pieces to `df1` with a scale of 0
	# and re-sort `df1` to match `df2` (after insertion of the 'removed' pieces)
	df2_anti <- df2i[which(match(df2i$id, df1i$id, 0L) == 0L), , drop = FALSE]
	df2_anti$scale <- rep_len(0, nrow(df2_anti))
	df1 <- rbind(df1i, df2_anti)
	df1 <- df1[match(df2$id, df1$id), ] # re-sort
	list(df1, df2)
}

# If can't find good previous id return `NA_integer_`
find_good_prev_id <- function(row, df1i, df2i) {
	prev_id <- Inf
	index <- which(df1i$id == row$id)
	while (is.infinite(prev_id)) {
		index <- index - 1
		if (index == 0) {
			prev_id <- NA_integer_
		} else {
			prev <- df1i[index, ]
			index2 <- which(df2i$id == prev$id)
			if (length(index2)) {
				prev2 <- df2i[index2, ]
				if (nigh(prev$x, prev2$x) && nigh(prev$y, prev2$y)) {
					prev_id <- df2i$id[index2]
				}
			}
		}
	}
	prev_id
}

# Insert `df2` into `df1` after `index`
# index = 0 means instead at beginning
insert_df <- function(df1, df2, id = tail(df1$id, 1L)) {
	if (is.na(id)) {
		rbind(df2, df1)
	} else if (id == tail(df1$id, 1L)) {
		rbind(df1, df2)
	} else {
		index <- which(df1$id == id)
		rbind(df1[seq.int(index), ], df2, df1[-seq.int(index), ])
	}
}

get_id_cfg <- function(df1, df2) {
	df <- rbind(df1[, c("id", "cfg")], df2[, c("id", "cfg")])
	unique(df)
}

get_tweenr_df <- function(df, ...) {
	stopifnot(all(hasName(df, c("id", "piece_side", "rank", "suit", "x", "y"))))
	if (!hasName(df, "alpha")) {
		df$alpha <- 1
	}
	if (!hasName(df, "angle")) {
		df$angle <- 0
	}
	if (!hasName(df, "scale")) {
		df$scale <- 1
	}

	if (hasName(df, "z")) {
		columns <- c("id", "piece_side", "suit", "rank", "x", "y", "z", "angle", "scale", "alpha")
	} else {
		columns <- c("id", "piece_side", "suit", "rank", "x", "y", "angle", "scale", "alpha")
	}

	as.data.frame(df[, columns])
}

test_that("animate_piece()", {
	# Prevent weird bug if checked with R compiled with cairo/X11 but X server not started
	skip_if_not(all(capabilities(c("cairo", "png", "X11"))))
	skip_if_not_installed("gifski")
	skip_if_not_installed("tweenr")

	dfs <- list()
	d.frame <- function(piece_side = "bit_back", ..., rank = 1L) {
		data.frame(
			piece_side = piece_side,
			...,
			rank = rank,
			cfg = "checkers1",
			stringsAsFactors = FALSE
		)
	}
	df <- d.frame("board_back", suit = 2L, rank = 3L, x = 2, y = 2, id = "1")
	dfs[[1L]] <- df
	df <- rbind(df, d.frame(suit = 1L, x = 2, y = 2, id = "2"))
	dfs[[2L]] <- df
	df <- rbind(df, d.frame(suit = 2L, x = 1, y = 2, id = "3"))
	dfs[[3L]] <- df
	df <- rbind(df, d.frame(suit = 1L, x = 3, y = 1, id = "4"))
	dfs[[4L]] <- df
	df <- rbind(df, d.frame(suit = 2L, x = 1, y = 3, id = "5"))
	dfs[[5L]] <- df
	df <- rbind(df, d.frame(suit = 1L, x = 1, y = 1, id = "6"))
	dfs[[6L]] <- df
	df <- rbind(df, d.frame(suit = 2L, x = 3, y = 3, id = "7"))
	dfs[[7L]] <- df
	df <- rbind(df, d.frame(suit = 1L, x = 2, y = 1, id = "8"))
	dfs[[8L]] <- df

	gif_filename <- tempfile(fileext = ".gif")
	on.exit(unlink(gif_filename))
	animate_piece(dfs, file = gif_filename, n_transitions = 1L, n_pauses = 1L)
	expect_true(file.exists(gif_filename))
})

#' Miscellaneous `piecepackr` utility functions
#'
#' `cleave()` converts a delimiter separated string into a vector.
#' `inch(x)` is equivalent to `grid::unit(x, "in")`.
#' `is_color_invisible()` tells whether the color is transparent (and hence need not be drawn).
#' @examples
#'  cleave("0.5,0.2,0.4,0.5", float=TRUE)
#'  cleave("black,darkred,#050EAA,,", color=TRUE)
#'
#'  is_color_invisible("transparent")
#'  is_color_invisible(NA)
#'  is_color_invisible("blue")
#'  is_color_invisible("#05AE9C")
#'
#'  if (requireNamespace("grid", quietly = TRUE)) {
#'      identical(inch(1), grid::unit(1, "inch"))
#'  }
#'
#' @name pp_utils
NULL

#' @param col Color
#' @rdname pp_utils
#' @export
is_color_invisible <- function(col) {
	as.logical(grDevices::col2rgb(col, alpha = TRUE)[4, ] == 0)
}

#' @rdname pp_utils
#' @param inches Number representing number of inches
#' @export
inch <- function(inches) unit(inches, "in")

#' @rdname pp_utils
#' @param s String to convert
#' @param sep Delimiter (defaults to ",")
#' @param float If `TRUE` cast to numeric
#' @param color if `TRUE` convert empty strings to `"transparent"`
#' @export
cleave <- function(s, sep = ",", float = FALSE, color = FALSE) {
	vec <- str_split(s, sep)
	if (length(vec)) {
		vec <- vec[[1]]
	}
	if (float) {
		as.numeric(vec)
	} else if (color) {
		gsub("^$", "transparent", vec)
	} else {
		vec
	}
}

cleave2 <- function(s, sep = ",", ...) {
	if (length(s) > 1) {
		s <- paste(s, collapse = sep)
	}
	cleave(s, sep, ...)
}
col_cleave <- function(s, sep = ",") cleave2(s, sep, color = TRUE)
numeric_cleave <- function(s, sep = ",") cleave2(s, sep, float = TRUE)

as_picture <- function(grob, width, height) {
	svg_file <- tempfile(fileext = ".svg")
	on.exit(unlink(svg_file), add = TRUE)

	current_dev <- grDevices::dev.cur()
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	grDevices::svg(svg_file, width = width, height = height, bg = "transparent")
	grid.draw(grob)
	invisible(grDevices::dev.off())
	file2grob(svg_file)
}

#' @rdname pp_utils
#' @param file Filename of image
#' @param distort Logical value of whether one should preserve the aspect ratio
#'                or distort to fit the area it is drawn in
#' @export
file2grob <- function(file, distort = TRUE) {
	current_dev <- grDevices::dev.cur()
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	format <- tools::file_ext(file)
	if (format %in% c("svgz", "svg")) {
		picture <- grImport2::readPicture(file, warn = FALSE)
	} else if (format == "png") {
		picture <- grDevices::as.raster(png::readPNG(file))
	} else if (format %in% c("jpg", "jpeg")) {
		picture <- grDevices::as.raster(jpeg::readJPEG(file))
	} else {
		assert_suggested("magick")
		picture <- magick::image_read(file)
	}
	if (inherits(picture, "Picture")) {
		ppPictureGrob(picture, distort)
	} else {
		# grDevices::is.raster(picture) # nolint
		to_rasterGrob(picture, distort)
	}
}

to_rasterGrob <- function(obj, distort = TRUE) {
	if (distort) {
		rasterGrob(
			grDevices::as.raster(obj),
			height = unit(1, "npc"),
			width = unit(1, "npc"),
			name = "raster"
		)
	} else {
		rasterGrob(grDevices::as.raster(obj), name = "raster")
	}
}

# adds support for a 'vp' viewport plus hiding grob details in `grid.ls()` (i.e. ``grid.revert()`` if "forced")
ppPictureGrob <- function(picture, distort = TRUE, ..., name = NULL, gp = gpar(), vp = NULL) {
	gTree(picture = picture, distort = distort, name = name, gp = gp, vp = vp, cl = "pp_picture")
}

#' @export
makeContent.pp_picture <- function(x) {
	grob <- grImport2::pictureGrob(x$picture, expansion = 0, clip = "off", distort = x$distort)
	setChildren(x, gList(grob))
}

assert_suggested <- function(package) {
	calling_fn <- deparse(sys.calls()[[sys.nframe() - 1]])
	if (!requireNamespace(package, quietly = TRUE)) {
		msg <- c(
			sprintf(
				"You need to install the suggested package %s to use %s.",
				sQuote(package),
				sQuote(calling_fn)
			),
			i = sprintf("Use %s.", sQuote(sprintf('install.packages("%s")', package)))
		)
		abort(msg, class = "piecepackr_suggested_package")
	}
}

# base R's Cairo/Quartz devices as well as {ragg} / {svglite} / {vdiffr} devices
# should support Unicode without complaint
# Notably `pdf()` is a device that does not...
# Any other devices to add?
device_supports_unicode <- function() {
	device <- names(grDevices::dev.cur())
	unicode_devices <- c(
		"agg_capture",
		"agg_jpeg",
		"agg_ppm",
		"agg_png",
		"agg_record",
		"agg_tiff", # {ragg}
		"devSVG",
		"devSVG_vdiffr", # {svglite} / {vdiffr}
		"quartz",
		"quartz_off_screen", # Quartz
		"cairo_pdf",
		"cairo_ps",
		"svg",
		"X11cairo"
	) # Cairo
	if (
		any(vapply(
			unicode_devices,
			function(x) grepl(paste0("^", x), device),
			FUN.VALUE = logical(1L)
		))
	) {
		TRUE
	} else if (device %in% c("bmp", "jpeg", "png", "tiff")) {
		# on unix non-"cairo" type have different device names from "cairo" type
		# but on Windows can't distinguish between `type = "windows"` or `type = "cairo"`
		# Windows device doesn't support new patterns feature
		if (getRversion() >= "4.2.0") {
			"LinearGradient" %in% grDevices::dev.capabilities()$patterns
		} else {
			.Platform$OS.type == "unix"
		}
	} else {
		FALSE
	}
}


# From `{gridpattern}`
update_alpha_col <- function(color, alpha = NA_real_) {
	n <- max(lengths(list(color, alpha)))
	color <- rep_len(color, n)
	alpha <- rep_len(alpha, n)
	m <- grDevices::col2rgb(color, alpha = TRUE) / 255
	m[4, ] <- ifelse(is.na(alpha), m[4, ], alpha)
	apply(m, 2, function(x) grDevices::rgb(x[1], x[2], x[3], x[4]))
}

# From `{oblicubes}`
cheap_darken <- function(color, amount) {
	mat <- col2rgb(color, alpha = TRUE)
	mat[1:3, ] <- mat[1:3, ] * (1 - amount)
	rgb(mat[1, ], mat[2, ], mat[3, ], mat[4, ], maxColorValue = 255)
}

as_fill_stroke_grob <- function(grob, fill, col = "black", lwd = 1.5) {
	if (has_fill_strokes()) {
		gp <- gpar(fill = fill, col = col, lwd = lwd)
		fillStrokeGrob(grob, gp = gp)
	} else {
		fs_inform()
		grob
	}
}

has_transformations <- function() {
	getRversion() >= '4.2.0' && isTRUE(dev.capabilities()$transformations)
}

has_fill_strokes <- function() {
	getRversion() >= '4.2.0' && isTRUE(dev.capabilities()$paths)
}

has_radial_gradients <- function() {
	getRversion() >= '4.1.0' &&
		isTRUE("RadialGradient" %in% dev.capabilities("patterns")$patterns)
}

at_inform <- function(fallback = "picture") {
	if (isFALSE(getOption("piecepackr.at.inform"))) {
		return(invisible(NULL))
	}

	msg <- "Affine transformation support not detected in the active graphics device."
	if (fallback == "picture") {
		msg <- paste(
			msg,
			"Falling back to rendering piece side with `grImport2::pictureGrob(..., distort=TRUE)`."
		)
	} else {
		msg <- paste(msg, "Falling back to rendering piece side with a `grid::polygonGrob()`.")
	}
	if (getRversion() < '4.2.0') {
		msg <- c(
			msg,
			i = paste("Current R is version `%s`", getRversion()),
			i = "Affine transformation support requires R version 4.2 or greater."
		)
	} else {
		msg <- c(
			msg,
			i = "`dev.capabilities()$transformations` is not `TRUE`.",
			i = "Perhaps try one of the cairo devices like `png(..., type='cairo')` or `cairo_pdf()`."
		)
	}
	msg <- c(
		msg,
		i = "These messages can be disabled via `options(piecepackr.at.inform = FALSE)`.",
		i = 'These messages can be suppressed via `suppressMessages(expr, classes = "piecepackr_affine_transformation")`.'
	)
	inform(msg, class = "piecepackr_affine_transformation")
}

fs_inform <- function() {
	if (isFALSE(getOption("piecepackr.fs.inform"))) {
		return(invisible(NULL))
	}

	msg <- paste(
		"Stroking and filling path support not detected in the active graphics device.",
		"Falling back to rendering glyph with a `grid::textGrob()`."
	)
	if (getRversion() < '4.2.0') {
		msg <- c(
			msg,
			i = paste("Current R is version `%s`", getRversion()),
			i = "Stroking and filling path support requires R version 4.2 or greater."
		)
	} else {
		msg <- c(
			msg,
			i = "`dev.capabilities()$paths` is not `TRUE`.",
			i = "Perhaps try one of the cairo devices like `png(..., type='cairo')` or `cairo_pdf()`."
		)
	}
	msg <- c(
		msg,
		i = "These messages can be disabled via `options(piecepackr.fs.inform = FALSE)`.",
		i = 'These messages can be suppressed via `suppressMessages(expr, classes = "piecepackr_fill_stroke")`.'
	)
	inform(msg, class = "piecepackr_fill_stroke")
}

rgr_inform <- function() {
	if (isFALSE(getOption("piecepackr.rgr.inform"))) {
		return(invisible(NULL))
	}

	msg <- paste(
		"Radial gradient support not detected in the active graphics device.",
		"Falling back to rendering without a gradient."
	)
	if (getRversion() < '4.1.0') {
		msg <- c(
			msg,
			i = paste("Current R is version `%s`", getRversion()),
			i = "Radial gradient support requires R version 4.1 or greater."
		)
	} else {
		msg <- c(
			msg,
			i = '`"RadialGradient" not in `dev.capabilities()$patterns`.',
			i = "Perhaps try one of the cairo devices like `png(..., type='cairo')` or `cairo_pdf()`."
		)
	}
	msg <- c(
		msg,
		i = "These messages can be disabled via `options(piecepackr.rgr.inform = FALSE)`.",
		i = 'These messages can be suppressed via `suppressMessages(expr, classes = "piecepackr_radial_gradient")`.'
	)
	inform(msg, class = "piecepackr_radial_gradient")
}

pdf_file <- tempfile(fileext = ".pdf")
svg_file <- tempfile(fileext = ".svg")

old_dev <- dev.cur()
pdf(pdf_file)
expect_false(device_supports_unicode())
dev.off()
if (old_dev > 1L) {
	dev.set(old_dev)
}
unlink(pdf_file)

skip_if(!capabilities("cairo"))
old_dev <- dev.cur()
svg(svg_file)
expect_true(device_supports_unicode())
dev.off()
if (old_dev > 1L) {
	dev.set(old_dev)
}
unlink(svg_file)

COMPONENT_AND_SIDES <- c(
	"tile_back",
	"tile_face",
	"coin_back",
	"coin_face",
	"die_face",
	"suitdie_face",
	"pawn_face",
	"pawn_back",
	"belt_face",
	"saucer_face",
	"saucer_back",
	"pyramid_face",
	"pyramid_back",
	"pyramid_left",
	"pyramid_right",
	"pyramid_top",
	"matchstick_face",
	"matchstick_back"
)

piece_device <- function(
	filename,
	piece_side = NULL,
	cfg = list(),
	angle = 0,
	suit = 1,
	rank = 1,
	width = NULL,
	height = NULL,
	res = 72
) {
	cfg <- as_pp_cfg(cfg)
	width <- width %||% cfg$get_width(piece_side, suit, rank)
	height <- height %||% cfg$get_height(piece_side, suit, rank)
	if (angle %in% c(90, 270)) {
		twidth <- height
		height <- width
		width <- twidth
	}
	pp_device(filename, width, height, res, bg = "transparent")
	pushViewport(viewport(angle = angle, name = "main"))
}

pp_device <- function(filename, width, height, res = 72, bg = "transparent") {
	args <- list(
		filename = filename,
		width = width,
		height = height,
		units = "in",
		res = res,
		bg = bg
	)
	dev_fn <- pp_device_fn(filename)
	args <- args[names(args) %in% names(formals(dev_fn))]
	do.call(dev_fn, args)
}

pp_device_fn <- function(filename) {
	if (is.null(filename)) {
		dev_new <- function(width, height) {
			grDevices::dev.new(width = width, height = height, noRStudioGD = TRUE)
		}
		return(dev_new)
	}
	format <- tools::file_ext(tolower(filename))
	switch(
		format,
		bmp = grDevices::bmp,
		jpeg = grDevices::jpeg,
		jpg = grDevices::jpeg,
		pdf = grDevices::cairo_pdf,
		png = grDevices::png,
		ps = grDevices::cairo_ps,
		svg = grDevices::svg,
		svgz = grDevices::svg,
		tiff = grDevices::tiff
	)
}

piece_filename <- function(directory, piece_side, format, angle, suit = NULL, rank = NULL) {
	filename <- paste0(
		piece_side,
		ifelse(is.null(suit), "", paste0("_s", suit)),
		ifelse(is.null(rank), "", paste0("_r", rank)),
		paste0("_t", angle),
		paste0(".", format)
	)
	file.path(directory, filename)
}

#' Save piecepack images
#'
#' Saves images of all individual piecepack pieces.
#'
#' @param cfg Piecepack configuration list
#' @param directory Directory where to place images
#' @param format Character vector of formats to save images in
#' @param angle Numeric vector of angles to rotate images (in degrees)
#' @examples
#'   \donttest{# May take more than 5 seconds on CRAN server
#'   if (all(capabilities(c("cairo", "png")))) {
#'       cfg <- pp_cfg(list(suit_color="darkred,black,darkgreen,darkblue,grey"))
#'       save_piece_images(cfg, directory=tempdir(), format="svg", angle=0)
#'       save_piece_images(cfg, directory=tempdir(), format="png", angle=90)
#'   }
#'   }
#' @export
save_piece_images <- function(
	cfg = getOption("piecepackr.cfg", pp_cfg()),
	directory = tempdir(),
	format = "svg",
	angle = 0
) {
	current_dev <- grDevices::dev.cur()
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	opt <- options(piecepackr.op_scale = 0)
	on.exit(options(opt), add = TRUE)

	stopifnot(dir.exists(directory))

	for (f in format) {
		for (a in angle) {
			make_images_helper(directory, cfg, f, a)
		}
	}
}

pp_dev_off <- function(f, format) {
	v <- grDevices::dev.off()
	if (format == "svgz") {
		rcon <- file(f, "r")
		svg <- readLines(rcon)
		close(rcon)
		wcon <- gzfile(f, "w")
		writeLines(svg, wcon)
		close(wcon)
	}
	invisible(v)
}

#### Check if configuration has that piece?
#### Extend to boards, bits, and cards?
#### What to do if don't want other side of two sided pieces?
#### Allow list of which pieces to save?
make_images_helper <- function(directory, cfg, format, angle) {
	lacks_rank_lacks_suit <- intersect(
		COMPONENT_AND_SIDES,
		intersect(cfg$lacks_rank, cfg$lacks_suit)
	)
	lacks_rank_has_suit <- intersect(setdiff(COMPONENT_AND_SIDES, cfg$lacks_suit), (cfg$lacks_rank))
	has_rank_lacks_suit <- intersect(setdiff(COMPONENT_AND_SIDES, cfg$lacks_rank), (cfg$lacks_suit))
	has_rank_has_suit <- setdiff(COMPONENT_AND_SIDES, union(cfg$lacks_rank, cfg$lacks_suit))

	suppressWarnings({
		for (cs in lacks_rank_lacks_suit) {
			f <- piece_filename(directory, cs, format, angle)
			piece_device(f, cs, cfg, angle)
			grid.piece(cs, NA, NA, cfg)
			pp_dev_off(f, format)
		}
		for (cs in lacks_rank_has_suit) {
			for (suit in 1:cfg$n_suits) {
				f <- piece_filename(directory, cs, format, angle, suit)
				piece_device(f, cs, cfg, angle)
				grid.piece(cs, suit, NA, cfg)
				pp_dev_off(f, format)
			}
		}
		for (cs in has_rank_lacks_suit) {
			for (rank in 1:cfg$n_ranks) {
				f <- piece_filename(directory, cs, format, angle, rank = rank)
				piece_device(f, cs, cfg, angle)
				grid.piece(cs, NA, rank, cfg)
				pp_dev_off(f, format)
			}
		}
		for (cs in has_rank_has_suit) {
			for (suit in 1:cfg$n_suits) {
				for (rank in 1:cfg$n_ranks) {
					f <- piece_filename(directory, cs, format, angle, suit, rank)
					piece_device(f, cs, cfg, angle, rank = rank)
					grid.piece(cs, suit, rank, cfg)
					pp_dev_off(f, format)
				}
			}
		}
	})
}

get_die_face_info <- function(suit, arrangement = "counter_down") {
	angle <- rep(0, length.out = 6)
	suit <- rep(suit, length.out = 6)

	if (grepl(da_token, arrangement)) {
		s <- strsplit(arrangement, ",")[[1]]
		angle <- sub("[1-6]", "", s)
		angle <- sub("\\^", "0", angle)
		angle <- sub("<", "90", angle)
		angle <- sub(">", "-90", angle)
		angle <- as.numeric(sub("v", "180", angle))
		angle <- ifelse(is.na(angle), 0, angle)
		rank <- as.integer(sub("[\\^<>v]", "", s))
	} else if (arrangement == "opposites_sum_to_5") {
		rank <- c(1, 2, 3, 6, 5, 4)
	} else if (arrangement == "counter_up") {
		rank <- 6:1
	} else if (arrangement == "counter_down") {
		rank <- 1:6
	} else {
		msg <- paste(sprintf('`die_arrangement` = "%s" is unrecognized', arrangement))
		abort(msg, class = "piecepackr_cfg")
	}
	suit <- suit[rank]
	list(rank = rank, suit = suit, angle = angle)
}

da_token <- "^([1-6]([\\^<>v])?,){5}[1-6]([\\^<>v])?$"

# figure out rotation to reach each die face
get_die_rotation <- function(suit, rank, cfg) {
	rs <- get_die_face_info(suit, cfg$die_arrangement)
	i <- which(rs$rank == rank)
	stopifnot(i <= 6)
	dR <- switch(
		i,
		diag(3),
		R_y(-90) %*% R_z(90),
		R_x(90) %*% R_z(-90),
		R_x(180),
		R_y(90) %*% R_z(90),
		R_x(-90) %*% R_z(90)
	)
	dR %*% R_z(-rs$angle[i])
}

get_die_faces <- function(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth) {
	xs <- c(0, 0, 1, 1, 0, 0, 1, 1) - 0.5
	ys <- c(1, 0, 0, 1, 1, 0, 0, 1) - 0.5
	zs <- rep(c(1, 0), each = 4) - 0.5

	dfi <- get_die_face_info(suit, cfg$die_arrangement)
	dR <- get_die_rotation(suit, rank, cfg)
	R <- dR %*% AA_to_R(angle, axis_x, axis_y)
	xyz <- as_coord3d(xs, ys, zs)$scale(width, height, depth)$transform(R)$translate(x, y, z)

	# textured face elements
	f <- list()
	f[[1]] <- xyz[die_cycle(1:4, dfi$angle[1])]
	f[[2]] <- xyz[die_cycle(c(8, 4, 3, 7), dfi$angle[2])]
	f[[3]] <- xyz[die_cycle(c(1, 4, 8, 5), dfi$angle[3])]
	f[[4]] <- xyz[die_cycle(c(6, 5, 8, 7), dfi$angle[4])]
	f[[5]] <- xyz[die_cycle(c(1, 5, 6, 2), dfi$angle[5])]
	f[[6]] <- xyz[die_cycle(c(3, 2, 6, 7), dfi$angle[6])]

	list(f_xyz = f, die_face_info = dfi)
}

die_cycle <- function(indices, angle) {
	if (angle == 90) {
		cycle_elements(indices, -1)
	} else if (angle == 180) {
		cycle_elements(indices, -2)
	} else if (angle == -90) {
		cycle_elements(indices, 1)
	} else {
		indices
	}
}

visible_die_faces <- function(die_faces, op_angle = 45) {
	indices <- 1:6

	i_top <- which.max(sapply(1:6, function(i) mean(die_faces$f_xyz[[i]])$z))
	i_bot <- which.min(sapply(1:6, function(i) mean(die_faces$f_xyz[[i]])$z))
	indices <- setdiff(indices, c(i_top, i_bot))

	face_centers <- as_coord3d(
		x = sapply(indices, function(i) mean(die_faces$f_xyz[[i]])$x),
		y = sapply(indices, function(i) mean(die_faces$f_xyz[[i]])$y),
		z = sapply(indices, function(i) mean(die_faces$f_xyz[[i]])$z)
	)
	indices <- indices[painter_order(face_centers, scale = 1, alpha = degrees(op_angle))]
	utils::tail(indices, 2L)
}

# cache which mimics API of cachem: https://cachem.r-lib.org/
# a variation using `fastmap::fastmap()` was not faster in testing
Cache <- R6Class(
	"pp_cache",
	public = list(
		exists = function(key) hasName(private$cache, key),
		get = function(key, default = key_missing()) {
			if (self$exists(key)) private$cache[[key]] else default
		},
		keys = function() names(private$cache),
		remove = function(key) private$cache[[key]] <- NULL,
		reset = function() {
			for (key in self$keys()) {
				self$remove(key)
			}
		},
		set = function(key, value) private$cache[[key]] <- value,
		size = function() length(private$cache)
	),
	private = list(cache = list())
)

key_missing <- function() structure(list(), class = "key_missing")
is.key_missing <- function(x) inherits(x, "key_missing")

test_that("Cache works as expected", {
	cache <- Cache$new()
	expect_false(cache$exists("foo"))
	expect_true(is.key_missing(cache$get("foo")))
	cache$set("foo", 4L)
	cache$set("bar", 2L)
	expect_true(cache$exists("foo"))
	expect_false(is.key_missing(cache$get("foo")))
	expect_equal(cache$get("foo"), 4L)
	expect_equal(cache$size(), 2L)
	expect_equal(cache$keys(), c("foo", "bar"))
	cache$remove("bar")
	expect_equal(cache$size(), 1L)
	cache$reset()
	expect_equal(cache$size(), 0L)
})

#' Render image of game pieces
#'
#' `render_piece()` renders an image of game pieces to a file or graphics device.
#' It is a wrapper around `pmap_piece()` that can auto-size files and graphic devices,
#' apply axes offsets, annotate coordinates, and set up `rayrender` / `rayvertex` scenes.
#' @inheritParams aabb_piece
#' @param file Filename to save image to unless `NULL`
#'             in which case it either uses the current graphics device or opens a new device
#'             (depending on `open_device` argument).
#' @param ... Arguments to [pmap_piece()]
#' @param .f Low level graphics function to use e.g. [grid.piece()], [piece3d()], [piece_mesh()], or [piece()].
#' @param cfg A piecepackr configuration list
#' @param envir Environment (or named list) of piecepackr configuration lists
#' @param width Width of image (in inches).  Inferred by default.
#' @param height Height of image (in inches).  Inferred by default.
#' @param ppi Resolution of image in pixels per inch.
#' @param bg Background color (use `"transparent"` for transparent)
#' @param xoffset Number to add to the `x` column in `df`.  Inferred by default.
#' @param yoffset Number to add to the `y` column in `df`.  Inferred by default.
#' @param annotate If `TRUE` or `"algebraic"` annotate the plot
#'                 with \dQuote{algrebraic} coordinates,
#'                 if `FALSE` or `"none"` don't annotate,
#'                 if `"cartesian"` annotate the plot with \dQuote{cartesian} coordinates.
#' @param annotation_scale Multiplicative factor that scales (stretches) any annotation coordinates.
#'                         By default uses `attr(df, "scale_factor") %||% 1`.
#' @param dev Graphics device function to use if `open_device` is `FALSE`.
#'            If `NULL` infer a reasonable choice from `file`.
#' @param dev.args Additional arguments to pass to `dev` (besides `filename`, `width`, and `height`).
#'                 Will filter out any names that aren't in `formals(dev)`.
#' @param open_device If `TRUE` open a new graphics device otherwise draw in the active graphics.
#' @param close_device If `TRUE` close the graphics device (if `open_device = TRUE` close the newly opened device otherwise close the previously existing graphics device).
#' @param image Class of image object to return in the `"image"` field of returned list.
#'              If `"NULL"` (the default) return `NULL`.
#'              If `"raster"` or `"nativeRaster" try to return a raster object of the image
#'              using [grDevices::dev.capture()] or `as.raster(magick::image_read())`.
#' @param xbreaks,ybreaks Subset (of integers) to provide axis labels for if `annotate` is `TRUE`.
#'                        If `NULL` infer a reasonable choice.
#' @return An invisible list of the dimensions of the image and
#'         possibly an image object specified by `image`.
#'         As a side effect may save a file and/or open/close a graphics device.
#' @seealso This function is a wrapper around [pmap_piece()].
#' @examples
#'  df_board <- data.frame(piece_side = "board_face", suit = 3, rank = 5,
#'                         x = 3.0, y = 3.0, stringsAsFactors = FALSE)
#'  df_w <- data.frame(piece_side = "bit_back", suit = 6, rank = 1,
#'                     x = rep(1:5, 2), y = rep(1:2, each=5),
#'                     stringsAsFactors = FALSE)
#'  df_b <- data.frame(piece_side = "bit_back", suit = 1, rank = 1,
#'                     x = rep(1:5, 2), y = rep(4:5, each=5),
#'                     stringsAsFactors = FALSE)
#'  df <- rbind(df_board, df_w, df_b)
#'  df$cfg <- "checkers1"
#'
#'  if (requireNamespace("grid", quietly = TRUE)) {
#'    render_piece(df, open_device = FALSE)
#'  }
#'  if (requireNamespace("grid", quietly = TRUE)) {
#'    grid::grid.newpage()
#'    render_piece(df, open_device = FALSE,
#'                 op_scale = 0.5, trans = op_transform,
#'                 annotate = "algrebraic")
#'  }
#'  \dontrun{# May take more than 5 seconds on CRAN servers
#'  if (require(rayvertex)) {
#'    envir3d <- game_systems("sans", border = FALSE)
#'    render_piece(df, .f = piece_mesh, envir = envir3d,
#'                 open_device = FALSE,
#'                 op_scale = 0.5, trans = op_transform)
#'  }
#'  }
#' @export
render_piece <- function(
	df,
	file = NULL,
	...,
	.f = piecepackr::grid.piece,
	cfg = getOption("piecepackr.cfg", NULL),
	envir = getOption("piecepackr.envir", game_systems("sans")),
	width = NULL,
	height = NULL,
	ppi = 72,
	bg = "white",
	xoffset = NULL,
	yoffset = NULL,
	annotate = FALSE,
	annotation_scale = NULL,
	dev = NULL,
	dev.args = list(res = ppi, bg = bg, units = "in"),
	open_device = TRUE,
	close_device = open_device && (!is.null(file) || !is.null(dev)),
	image = c("NULL", "raster", "nativeRaster"),
	xbreaks = NULL,
	ybreaks = NULL
) {
	stopifnot(is.null(dev) || is.function(dev))
	image <- match.arg(image)
	if (image == "NULL") {
		image <- NULL
	}

	# Make sure if we close the graphics device we return to any previous device
	if (close_device) {
		if (open_device) {
			current_dev <- grDevices::dev.cur()
			if (current_dev > 1L) {
				on.exit(grDevices::dev.set(current_dev), add = TRUE)
			}
		} else {
			if (length(dev.list()) > 1L) {
				prev_dev <- grDevices::dev.prev()
				on.exit(grDevices::dev.set(prev_dev), add = TRUE)
			}
		}
	}

	ce <- default_cfg_envir(cfg, envir)
	cfg <- ce$cfg
	envir <- ce$envir

	dfr <- aabb_piece(df, cfg = cfg, envir = envir, ...)
	xmax <- dfr$x[2]
	ymax <- dfr$y[2]
	xoffset <- xoffset %||% min2offset(dfr$x_op[1])
	yoffset <- yoffset %||% min2offset(dfr$y_op[1])
	if (is.null(width)) {
		width <- dfr$x_op[2] + xoffset + 0.50
	}
	if (is.null(height)) {
		height <- dfr$y_op[2] + yoffset + 0.50
	}
	if (is.na(width)) {
		width <- 1
	}
	if (is.na(height)) {
		height <- 1
	}
	m <- max(width, height)

	if (open_device) {
		if (is.null(dev)) {
			dev <- pp_device_fn(file)
		}
		args <- c(list(filename = file, width = width, height = height), dev.args)
		args <- args[names(args) %in% names(formals(dev))]
		do.call(dev, args)
	}

	# plot_fn_helper expected width and height in pixels
	width <- ppi * width
	height <- ppi * height
	l <- list(width = width, height = height, image = NULL)

	fn <- plot_fn_helper(
		.f,
		xmax,
		ymax,
		xoffset,
		yoffset,
		width,
		height,
		m,
		ppi,
		envir,
		annotate,
		annotation_scale,
		xbreaks,
		ybreaks
	)
	fn(df, ...)

	if (!is.null(image)) {
		stopifnot(image %in% c("raster", "nativeRaster"))
		native <- ifelse(image == "nativeRaster", TRUE, FALSE)
		l$image <- suppressWarnings(grDevices::dev.capture(native = native))
		if (is.null(l$image) && is.null(file)) {
			abort("`dev.capture()` failed to capture a raster image with this graphics device.")
		}
	}

	if (close_device) {
		grDevices::dev.off()
	}

	# Fallback to {magick} to capture an image if `dev.capture()` failed
	if (!is.null(image) && is.null(l$image)) {
		assert_suggested("magick")
		l$image <- grDevices::as.raster(magick::image_read(file), native = native)
	}
	invisible(l)
}

plot_fn_helper <- function(
	.f = grid.piece,
	xmax,
	ymax,
	xoffset,
	yoffset,
	width,
	height,
	m,
	ppi,
	envir,
	annotate,
	annotation_scale,
	xbreaks,
	ybreaks
) {
	if (identical(.f, grid.piece)) {
		function(df, ..., scale = 1) {
			annotation_scale <- annotation_scale %||% attr(df, "scale_factor") %||% 1
			df$x <- df$x + xoffset
			df$y <- df$y + yoffset
			df$scale <- if (hasName(df, "scale")) scale * df$scale else scale
			grid::grid.newpage()
			pmap_piece(df, default.units = "in", ..., envir = envir)
			annotate_plot(
				annotate,
				xmax,
				ymax,
				xoffset,
				yoffset,
				annotation_scale,
				xbreaks,
				ybreaks
			)
		}
	} else if (identical(.f, piece3d)) {
		assert_suggested("rgl")
		if (Sys.which("wmctrl") != "") {
			cmd <- paste0("wmctrl -r RGL -e 0,-1,-1,", ceiling(width), ",", ceiling(height))
			system(cmd)
		}
		f <- tempfile(fileext = ".png")
		function(df, ..., scale = 1) {
			df$scale <- if (hasName(df, "scale")) scale * df$scale else scale
			rgl::clear3d()
			rgl::points3d(x = rep(c(0, xmax), 2), y = rep(c(0, ymax), each = 2), z = 0, alpha = 0)
			pmap_piece(df, piece3d, ..., envir = envir)
			Sys.sleep(2)
			rgl::snapshot3d(f, top = FALSE, webshot = FALSE)
			grid::grid.newpage()
			grid::grid.raster(png::readPNG(f))
		}
	} else if (identical(.f, piece)) {
		assert_suggested("rayrender")
		function(
			df,
			...,
			scale = 1,
			fov = 20,
			samples = 100,
			lookat = NULL,
			lookfrom = NULL,
			clamp_value = Inf,
			table = NA,
			interactive = TRUE
		) {
			df$scale <- if (hasName(df, "scale")) scale * df$scale else scale
			df$x <- df$x + xoffset
			df$y <- df$y + yoffset
			l <- pmap_piece(df, piece, ..., envir = envir)
			if (all(is.na(table))) {
				table <- rayrender::sphere(
					z = -1e3,
					radius = 1e3,
					material = rayrender::diffuse(color = "green")
				)
				light <- rayrender::sphere(
					x = 0.5 * width / ppi,
					y = -4,
					z = max(2.0 * m + 1, 20),
					material = rayrender::light(intensity = 420)
				)
				table <- rayrender::add_object(table, light)
			}
			scene <- Reduce(rayrender::add_object, l, init = table)
			if (is.null(lookat)) {
				lookat <- c(0.5 * width / ppi, 0.5 * height / ppi, 0)
			}
			if (is.null(lookfrom)) {
				lookfrom <- c(0.5 * width / ppi, -2.0 * m, 2.0 * m)
			}
			rayrender::render_scene(
				scene,
				fov = fov,
				samples = samples,
				lookat = lookat,
				lookfrom = lookfrom,
				clamp_value = clamp_value,
				width = width,
				height = height,
				interactive = interactive
			)
		}
	} else if (identical(.f, piece_mesh)) {
		assert_suggested("rayvertex")
		function(
			df,
			...,
			scale = 1,
			fov = 20,
			lookat = NULL,
			lookfrom = NULL,
			table = NA,
			light_info = NA
		) {
			df$scale <- if (hasName(df, "scale")) scale * df$scale else scale
			df$x <- df$x + xoffset
			df$y <- df$y + yoffset
			l <- pmap_piece(df, piece_mesh, ..., envir = envir)
			if (all(is.na(table))) {
				table <- rayvertex::sphere_mesh(
					c(0, 0, -1e3),
					radius = 1e3,
					material = rayvertex::material_list(diffuse = "green")
				)
			}
			if (all(is.na(light_info))) {
				light_info <- rayvertex::directional_light(
					c(0.5 * width / ppi, -4, max(2.0 * m + 1, 20)),
					intensity = 2.5
				)
			}
			scene <- Reduce(rayvertex::add_shape, l, init = table)
			if (is.null(lookat)) {
				lookat <- c(0.5 * width / ppi, 0.5 * height / ppi, 0)
			}
			if (is.null(lookfrom)) {
				lookfrom <- c(0.5 * width / ppi, -2.0 * m, 2.0 * m)
			}
			rayvertex::rasterize_scene(
				scene,
				width = width,
				height = height,
				fov = fov,
				lookat = lookat,
				lookfrom = lookfrom,
				light_info = light_info
			)
		}
	} else {
		.f
	}
}

min2offset <- function(min, lbound = 0.5) {
	if (is.na(min)) {
		NA_real_
	} else if (min < lbound) {
		lbound - min
	} else {
		0
	}
}

annotate_plot <- function(
	annotate,
	xmax,
	ymax,
	xoffset = 0,
	yoffset = 0,
	annotation_scale = 1,
	xbreaks = NULL,
	ybreaks = NULL
) {
	if (isFALSE(annotate) || annotate == "none" || is.na(xmax) || is.na(ymax)) {
		return(invisible(NULL))
	}
	gp <- gpar(fontsize = 18, fontface = "bold")

	if (is.null(xbreaks)) {
		x_coords <- seq(annotation_scale, floor(xmax), by = annotation_scale)
	} else {
		xbreaks <- as.integer(xbreaks)
		x_coords <- seq(annotation_scale, by = annotation_scale, length.out = max(xbreaks))
	}
	if (annotate == "cartesian") {
		l <- as.character(seq_along(x_coords))
	} else {
		l <- letters[seq_along(x_coords)]
	}
	if (!is.null(xbreaks)) {
		x_coords <- x_coords[xbreaks]
		l <- l[xbreaks]
	}
	l <- str_pad(l, max(str_count(l)))
	grid.text(l, x = x_coords + xoffset, y = 0.25, default.units = "in", gp = gp)

	if (is.null(ybreaks)) {
		y_coords <- seq(annotation_scale, floor(ymax), by = annotation_scale)
	} else {
		ybreaks <- as.integer(ybreaks)
		y_coords <- seq(annotation_scale, by = annotation_scale, length.out = max(ybreaks))
	}
	n <- as.character(seq_along(y_coords))
	n <- str_pad(n, max(str_count(n)))
	if (!is.null(ybreaks)) {
		y_coords <- y_coords[ybreaks]
		l <- l[ybreaks]
	}
	grid.text(n, x = 0.25, y = y_coords + yoffset, default.units = "in", gp = gp)

	invisible(NULL)
}

df_board <- data.frame(
	piece_side = "board_face",
	suit = 3,
	rank = 8,
	x = 4.5,
	y = 4.5,
	stringsAsFactors = FALSE
)
df_w <- data.frame(
	piece_side = "bit_back",
	suit = 6,
	rank = 1,
	x = rep(1:8, 2),
	y = rep(1:2, each = 8),
	stringsAsFactors = FALSE
)
df_b <- data.frame(
	piece_side = "bit_back",
	suit = 1,
	rank = 1,
	x = rep(1:8, 2),
	y = rep(7:8, each = 8),
	stringsAsFactors = FALSE
)
df <- rbind(df_board, df_w, df_b)
df$cfg <- "checkers1"

test_that("render_piece() works", {
	skip_on_ci()
	skip_if_not_installed("vdiffr")
	library("vdiffr")

	envir <- game_systems("sans")
	expect_doppelganger("render-piece", function() {
		render_piece(df, envir = envir, open_device = FALSE)
	})
	expect_doppelganger("render-piece-cartesian", function() {
		render_piece(df, envir = envir, annotate = TRUE, open_device = FALSE)
	})
	expect_doppelganger("render-piece-algebraic", function() {
		render_piece(
			df,
			envir = envir,
			op_scale = 0.5,
			annotate = "algebraic",
			trans = op_transform,
			open_device = FALSE
		)
	})

	skip_if_not_installed("rayvertex")
	library("rayvertex")
	envir3d <- game_systems("sans", border = FALSE)
	f <- tempfile(fileext = ".jpg")
	render_piece(
		df,
		file = f,
		.f = piece_mesh,
		envir = envir3d,
		op_scale = 0.5,
		trans = op_transform
	)
	expect_true(file.exists(f))
	unlink(f)

	skip_if_not_installed("rayrender")
	library("rayrender")
	f <- tempfile(fileext = ".jpg")
	render_piece(
		df,
		file = f,
		.f = piece,
		envir = envir3d,
		op_scale = 0.5,
		trans = op_transform,
		samples = 5,
		clamp_value = 1,
		interactive = FALSE
	)
	expect_true(file.exists(f))
	unlink(f)

	skip_if_not_installed("rgl")
	library("rgl")
	open3d()
	view3d(0, -30)
	f <- tempfile(fileext = ".tiff")
	render_piece(
		df,
		file = f,
		.f = piece3d,
		envir = envir3d,
		op_scale = 0.5,
		trans = op_transform,
		scale = 0.98
	)
	expect_true(file.exists(f))
	unlink(f)
	close3d()
})

#' Standard game systems
#'
#' `game_systems()` returns a list of [pp_cfg()] objects
#' representing several game systems and pieces.
#' `to_subpack()` and `to_hexpack()` will attempt to generate matching (piecepack stackpack)
#'      subpack and (piecepack) hexpack [pp_cfg()] R6 objects respectively given a piecepack configuration.
#'
#' Contains the following game systems:\describe{
#' \item{alquerque}{Boards and pieces in nine color schemes for Alquerque}
#' \item{checkers1, checkers2}{Checkers and checkered boards in nine color schemes.
#'       Checkers are represented by a piecepackr \dQuote{bit}.
#'       The \dQuote{board} \dQuote{face} is a checkered board and the \dQuote{back} is a lined board.
#'       Color is controlled by suit and number of rows/columns by rank.
#'       `checkers1` has one inch squares and `checkers2` has two inch squares.}
#' \item{chess1, chess2}{Chess pieces, boards, and dice in nine color schemes.
#'       Chess pieces are represented by a \dQuote{bit} (face).
#'       The \dQuote{board} \dQuote{face} is a checkered board and the \dQuote{back} is a lined board.
#'       Color is controlled by suit and number of rows/columns by rank.
#'       `chess1` has one inch squares and `chess2` has two inch squares.
#'       Currently uses print-and-play style discs instead of 3D Staunton chess pieces.}
#' \item{cubes}{Cubes in six sizes and nine colors represented by a \dQuote{bit}.
#'              Color is controlled by suit and size of cube by rank.
#'              Sizes go from 8mm (rank 1) to 25mm (rank 6): 8mm, 10mm, 12mm, 15mm, 20mm, 25mm.}
#' \item{dice}{Traditional six-sided pipped dice in nine color schemes (color controlled by their suit).}
#' \item{dice_d4, dice_numeral, dice_d8, dice_d10, dice_d10_percentile, dice_d12, dice_d20}{
#'       Polyhedral dice most commonly used to play wargames, roleplaying games, and trading card games:\describe{
#'   \item{dice_d4}{Four-sided dice in nine color schemes (color controlled by their suit).
#'                  Tetrahedrons with the rank as a numeral at the top point.}
#'   \item{dice_numeral}{Six-sided dice with numerals instead of pips in nine color schemes (color controlled by their suit).}
#'   \item{dice_d8}{Eight-sided dice in nine color schemes (color controlled by their suit).
#'                  Octahedrons with the rank as a numeral at the top face.}
#'   \item{dice_d10}{Ten-sided dice in nine color schemes (color controlled by their suit).
#'                  Pentagonal trapezohedrons with the rank as a numeral at the top face.
#'                  The rank of ten is represented by a zero.}
#'   \item{dice_d10_percentile}{Ten-sided dice in nine color schemes (color controlled by their suit).
#'                  Pentagonal trapezohedrons with the rank as a numeral followed by a zero at the top face.
#'                  The rank of ten is represented by a zero.}
#'   \item{dice_d12}{Twelve-sided dice in nine color schemes (color controlled by their suit).
#'                  Dodecahedrons with the rank as a numeral at the top face.}
#'   \item{dice_d20}{Twenty-sided dice in nine color schemes (color controlled by their suit).
#'                   Icosahedrons with the rank as a numeral at the top face.}}}
#' \item{dice_fudge}{\dQuote{Fudge} dice in nine color schemes (color controlled by their suit).
#'                   \dQuote{Fudge} dice have three ranks "+", " ", and "-" repeated twice.}
#' \item{dominoes, dominoes_black, dominoes_blue, dominoes_green, dominoes_red, dominoes_white, dominoes_yellow}{
#'      Traditional pipped dominoes in six color schemes (`dominoes` and `dominoes_white` are the same).
#'      In each color scheme the number of pips on the \dQuote{top} of the domino is
#'      controlled by their \dQuote{rank} and on the \dQuote{bottom} by their \dQuote{suit}.
#'      Supports up to double-18 sets.}
#' \item{dominoes_chinese, dominoes_chinese_black}{`dominoes_chinese` has Asian-style six-sided pipped dice with
#'       white background and black and red pips.
#'       The \dQuote{tile}'s are Chinese dominoes (1" x 2.5") whose number of pips are controlled
#'       by both their \dQuote{rank} and their \dQuote{suit}. `dominoes_chinese_black` are like `dominoes_chinese` but the
#'       dice and dominoes have a black background and white and red pips.}
#' \item{go}{Go stones and lined boards in nine color schemes.
#'           Go stones are represented by a \dQuote{bit} and the board is a \dQuote{board}.
#'           Color is controlled by suit and number of rows/columns by rank.}
#' \item{marbles}{Marbles in nine sizes and nine colors represented by a \dQuote{bit}.
#'                Color is controlled by suit and size of marble by rank.
#'                Sizes go from 1/2" (rank 1) to 1" (rank 9) in 1/16" increments.
#'                There are also square holed boards spaced for 1" marbles: their color
#'                is controlled by suit and number of holes per row/column by rank.}
#' \item{meeples}{Standard 16mm x 16mm x 10mm \dQuote{meeples} in nine colors represented by a \dQuote{bit}.}
#' \item{morris}{Various morris aka mills aka merels games in nine colors.
#'               Color is controlled by suit and \dQuote{size} of morris board
#'               is controlled by rank e.g. \dQuote{Six men's morris} corresponds to a rank of 6 and
#'               \dQuote{Nine men's morris} corresponds to a rank of 9.
#'               Game pieces are represented by stones.}
#' \item{piecepack, dual_piecepacks_expansion, playing_cards_expansion, hexpack, subpack, piecepack_inverted}{
#'   The piecepack is a public domain game system invented by James "Kyle" Droscha.
#    There are also several public domain accessories and expansions:
#'   See \url{https://www.ludism.org/ppwiki} for more info about the piecepack and its accessories/expansions.
#'   \describe{
#'   \item{piecepack}{A standard piecepack.  The configuration also contains the following piecepack accessories:\describe{
#'     \item{piecepack dice cards}{An accessory proposed by John Braley.
#'                                 See \url{https://www.ludism.org/ppwiki/PiecepackDiceCards}.}
#'     \item{piecepack matchsticks}{A public domain accessory developed by Dan Burkey.
#'                                 See \url{https://www.ludism.org/ppwiki/PiecepackMatchsticks}.}
#'     \item{piecepack pyramids}{A public domain accessory developed by Tim Schutz.
#'                              See \url{https://www.ludism.org/ppwiki/PiecepackPyramids}.}
#'     \item{piecepack saucers}{A public domain accessory developed by Karol M. Boyle at Mesomorph Games.
#'              See \url{https://web.archive.org/web/20190719155827/http://www.piecepack.org/Accessories.html}.}}}
#' \item{piecepack_inverted}{The standard piecepack with its color scheme inverted.
#'           Intended to aid in highlighting special pieces in diagrams.}
#' \item{dual_piecepacks_expansion}{A companion piecepack with a special suit scheme.
#'               See \url{https://trevorldavis.com/piecepackr/dual-piecepacks-pnp.html}.}
#' \item{playing_cards_expansion}{A piecepack with the standard \dQuote{French} playing card suits.
#'                                See \url{https://www.ludism.org/ppwiki/PlayingCardsExpansion}.}
#' \item{hexpack}{A hexagonal extrapolation of the piecepack designed by Nathan Morse and Daniel Wilcox.
#'                See \url{https://boardgamegeek.com/boardgameexpansion/35424/hexpack}.}
#' \item{subpack}{A mini piecepack.  Designed to be used with the `piecepack` to make piecepack
#'               \dQuote{stackpack} diagrams.  See \url{https://www.ludism.org/ppwiki/StackPack}.}
#'}}
#' \item{playing_cards, playing_cards_colored, playing_cards_tarot}{
#'       Poker-sized \dQuote{card} components for various playing card decks:\describe{
#'        \item{playing_cards}{A traditional deck of playing cards with 4 suits
#'            and 13 ranks (A, 2-10, J, Q, K) plus a 14th \dQuote{Joker} rank.}
#'        \item{playing_cards_colored}{Like `playing_cards` but with five colored suits:
#'            red hearts, black spades, green clubs, blue diamonds, and yellow stars.}
#'        \item{playing_cards_tarot}{A (French Bourgeois) deck of tarot playing cards:
#'            first four suits are hearts, spades, clubs, and diamonds with
#'            14 ranks (ace through jack, knight, queen, king) plus a 15th \dQuote{Joker} rank
#'            and a fifth "suit" of 22 trump cards (1-21 plus an \dQuote{excuse}).}}}
#' \item{reversi}{Boards and pieces for Reversi.
#'    "board_face" provides lined boards with colored backgrounds.
#'    "board_back" provides checkered boards.
#'    "bit_face" / "bit_back" provides circular game tokens with differently colored sides:
#'    red paired with green, black paired with white, and blue paired with yellow.
#'}
#' }
#' @param font If `NULL` (the default) uses suit glyphs from the default \dQuote{sans} font.
#'             If `"dejavu"` it will use suit glyphs from the "DejaVu Sans" font
#'             (must be installed on the system).
#' @param round If `TRUE` the \dQuote{shape} of \dQuote{tiles} and \dQuote{cards} (and some boards)
#'              will be \dQuote{roundrect} instead of \dQuote{rect} (the default).
#' @param pawn If `"token"` (default) the piecepack pawn will be a two-sided token in a \dQuote{halma} outline,
#'             if `"peg-doll"` the piecepack pawn will be a \dQuote{peg doll} style pawn, and
#'             if `"joystick"` the piecepack pawn will be a \dQuote{joystick} style pawn.
#'             Note for the latter two pawn styles only `pawn_top` will work with [grid.piece()].
#' @param ... Must be empty
#' @param shading If `TRUE` add a shading effect to marbles and stones when drawn with [grid.piece()].
#' @param border If `TRUE` draw a black border line on piece edges.  Should normally be `TRUE` when
#'               drawing with `{grid}` graphics and `FALSE` when drawing with 3D graphic systems.
#' @param background_color Background color to use for certain pieces like boards and piecepack tiles.
#' @param edge_color Edge color to use for certain pieces like boards and piecepack tiles.
#' @examples
#' cfgs <- game_systems(pawn = "joystick")
#' names(cfgs)
#'
#' \donttest{# May take more than 5 seconds on CRAN servers
#' # standard dice, meeples, and joystick pawns
#' if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {
#'
#'    opt <- options(piecepackr.at.inform = FALSE)
#'    grid::grid.newpage()
#'    dice <-  c("d4", "numeral", "d8", "d10", "d12", "d20")
#'    cfg <- paste0("dice_", dice)
#'    grid.piece("die_face", suit = c(1:6, 1), rank = 1:6,
#'               cfg = cfg, envir = cfgs, x = 1:6, y = 1,
#'               default.units = "in", op_scale = 0.5)
#'    grid.piece("die_face", rank=1:6, suit=1:6,
#'               x=1:6, y=2, default.units="in",
#'               op_scale=0.5, cfg=cfgs$dice)
#'    grid.piece("bit_face", suit=1:6,
#'               x=1:6, y=3, default.units="in",
#'               op_scale=0.5, cfg=cfgs$meeple)
#'    grid.piece("pawn_top", suit=1:6,
#'               x=1:6, y=4, default.units="in",
#'               op_scale=0.5, cfg=cfgs$piecepack)
#'    options(opt)
#' }
#' }
#' # dominoes
#' if (requireNamespace("grid", quietly = TRUE)) {
#'    grid::grid.newpage()
#'    colors <- c("black", "red", "green", "blue", "yellow", "white")
#'    cfg <- paste0("dominoes_", rep(colors, 2))
#'    grid.piece("tile_face",  suit=1:12, rank=1:12+1,
#'               cfg=cfg, envir=cfgs,
#'               x=rep(6:1, 2), y=rep(2*2:1, each=6),
#'               default.units="in", op_scale=0.5)
#' }
#' # piecepack "playing card expansion"
#' if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {
#'    grid::grid.newpage()
#'    df_tiles <- data.frame(piece_side="tile_back",
#'                           x=0.5+c(3,1,3,1), y=0.5+c(3,3,1,1),
#'                           suit=NA, angle=NA, z=1/8,
#'                           stringsAsFactors=FALSE)
#'    df_coins <- data.frame(piece_side="coin_back",
#'                           x=rep(4:1, 4), y=rep(4:1, each=4),
#'                           suit=c(1,4,1,4,4,1,4,1,2,3,2,3,3,2,3,2),
#'                           angle=rep(c(180,0), each=8), z=1/4+1/16,
#'                           stringsAsFactors=FALSE)
#'    df <- rbind(df_tiles, df_coins)
#'    pmap_piece(df, cfg = cfgs$playing_cards_expansion, op_scale=0.5,
#'               default.units="in")
#' }
#' @seealso [pp_cfg()] for information about the [pp_cfg()] objects returned by `game_systems()`.
#' @export
game_systems <- function(
	font = NULL,
	round = FALSE,
	pawn = "token",
	...,
	shading = FALSE,
	border = TRUE,
	background_color = ifelse(border, "white", "burlywood"),
	edge_color = ifelse(border, "white", "black")
) {
	check_dots_empty()
	font <- game_systems_font(font)

	rect_shape <- ifelse(round, "roundrect", "rect")
	color_list <- color_list_fn(border, background_color, edge_color)

	cards <- playing_cards(font, rect_shape)
	packs <- piecepack(font, color_list, rect_shape, pawn, background_color, edge_color)

	list(
		alquerque = alquerque(1, color_list, shading),
		checkers1 = checkers(font, 1, color_list),
		checkers2 = checkers(font, 2, color_list),
		chess1 = chess(font, 1, color_list),
		chess2 = chess(font, 2, color_list),
		cubes = cubes(color_list),
		dice = dice(color_list, rect_shape),
		dice_d4 = dice_d4(font, color_list),
		dice_d8 = dice_d8(font, color_list),
		dice_d10 = dice_d10(font, color_list),
		dice_d10_percentile = dice_d10_percentile(font, color_list),
		dice_d12 = dice_d12(font, color_list),
		dice_d20 = dice_d20(font, color_list),
		dice_fudge = dice_fudge(color_list, rect_shape),
		dice_numeral = dice_numeral(font, color_list, rect_shape),
		dominoes = dominoes(color_list$suit_color[6], "black", color_list$border_color, rect_shape),
		dominoes_black = dominoes(
			color_list$suit_color[2],
			"white",
			color_list$border_color,
			rect_shape
		),
		dominoes_blue = dominoes(
			color_list$suit_color[4],
			"white",
			color_list$border_color,
			rect_shape
		),
		dominoes_green = dominoes(
			color_list$suit_color[3],
			"white",
			color_list$border_color,
			rect_shape
		),
		dominoes_red = dominoes(
			color_list$suit_color[1],
			"white",
			color_list$border_color,
			rect_shape
		),
		dominoes_white = dominoes(
			color_list$suit_color[6],
			"black",
			color_list$border_color,
			rect_shape
		),
		dominoes_yellow = dominoes(
			color_list$suit_color[5],
			"black",
			color_list$border_color,
			rect_shape
		),
		dominoes_chinese = dominoes_chinese(color_list, rect_shape),
		dominoes_chinese_black = dominoes_chinese(color_list, rect_shape, invert = TRUE),
		dual_piecepacks_expansion = packs$dpe,
		go = go(1, color_list, shading),
		hexpack = packs$hexpack,
		marbles = marbles(color_list, rect_shape, shading),
		meeples = meeples(color_list),
		morris = morris(1, color_list, shading),
		piecepack = packs$base,
		piecepack_inverted = packs$inverted,
		playing_cards = cards$base,
		playing_cards_colored = cards$color,
		playing_cards_tarot = cards$tarot,
		playing_cards_expansion = packs$pce,
		reversi = reversi(1, color_list),
		subpack = packs$subpack
	)
}

color_list_fn <- function(border = TRUE, background_color = "white", edge_color = "white") {
	if (border) {
		list(
			suit_color = cb_suit_colors_impure,
			border_color = "black",
			border_lex = 4,
			background_color = background_color,
			edge_color.board = edge_color
		)
	} else {
		list(
			suit_color = cb_suit_colors_pure,
			border_color = "transparent",
			border_lex = 0,
			background_color = background_color,
			edge_color.board = edge_color
		)
	}
}

game_systems_font <- function(font = "sans") {
	fonts <- c("dejavu", "sans")
	if (!is.null(font) && is.na(match(font, fonts))) {
		abort(paste("Don't have a customized configuration for font", dQuote(font)))
	}

	if (is.null(font)) {
		font <- "sans"
	}

	if (grepl("dejavu", font) && !suppressWarnings(has_font("Dejavu Sans"))) {
		inform(sprintf("`font = \"%s\"` but `has_font(\"Dejavu Sans\")` is `FALSE`", font))
	}

	font
}

# Okabe-Ito palette i.e. `palette.colors(n = 8, palette = "Okabe-Ito", names=TRUE)` plus white
#                         vermillion black    bluishgreen    blue     yellow     white      skyblue reddishpurple orange
#                         red        black      green      blue       yellow     white      cyan       magenta    orange
cb_suit_colors_pure <- c(
	"#D55E00",
	"#000000",
	"#009E73",
	"#0072B2",
	"#F0E442",
	"#FFFFFF",
	"#56B4E9",
	"#CC79A7",
	"#E69F00"
)
cb_suit_colors_impure <- cb_suit_colors_pure
cb_suit_colors_impure[2L] <- "grey30" # a dark gray to contrast with black borders

DICE_PIP_COLORS <- "white,white,white,white,black,black,black,black,white"
N_SUITS_COLOR <- 9L

dice <- function(color_list = color_list_fn(), rect_shape = "rect") {
	dice_list <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 6L,
		rank_text = "1,2,3,4,5,6",
		width.die = 16 / 25.4, # 16 mm dice most common
		background_color = DICE_PIP_COLORS,
		invert_colors = TRUE,
		die_arrangement = "opposites_sum_to_5",
		shape.card = rect_shape,
		grob_fn.card_face = cardGrobFn(grob_type = "circle"),
		grob_fn.die = pippedGrobFn(0, "die")
	)
	dice <- pp_cfg(c(dice_list, color_list))
	dice$has_piecepack <- FALSE
	dice$has_dice <- TRUE
	dice
}

dominoes_chinese <- function(color_list = color_list_fn(), rect_shape = "rect", invert = FALSE) {
	dice_list <- list(
		n_suits = 6L,
		n_ranks = 6L,
		rank_text = "1,2,3,4,5,6",
		width.die = 16 / 25.4, # 16 mm dice most common
		background_color = ifelse(invert, color_list$suit_color[2], "white"),
		suit_color = ifelse(invert, "white", "black"),
		die_arrangement = "1,2,3,6>,5,4",
		grob_fn.die = pippedGrobFn(0, "die_chinese"),
		width.tile = 1,
		height.tile = 2.5,
		depth.tile = 0.5,
		gridline_color.tile_back = "transparent",
		gridline_color.tile_face = "transparent",
		gridline_lex.tile_face = 6,
		shape.tile = rect_shape,
		grob_fn.tile_face = dominoGrobFn(0, "domino_chinese"),
		shape.card = rect_shape,
		grob_fn.card_face = cardGrobFn(grob_type = "circle")
	)
	dice <- pp_cfg(c(dice_list, color_list))
	dice$has_piecepack <- FALSE
	dice$has_dice <- TRUE
	dice
}

dice_fudge <- function(color_list = color_list_fn(), rect_shape = "rect") {
	dice_list <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 6,
		rank_text = "-, ,+,+, ,-",
		dm_text.die = "",
		ps_cex.die = 2,
		grob_fn.die = fudgeGrob,
		width.die = 16 / 25.4, # 16 mm dice most common
		background_color = DICE_PIP_COLORS,
		invert_colors = TRUE,
		die_arrangement = "opposites_sum_to_5",
		shape.card = rect_shape
	)
	dice <- pp_cfg(c(dice_list, color_list))
	dice$has_piecepack <- FALSE
	dice$has_dice <- TRUE
	dice
}

dice_d4 <- function(font = "sans", color_list = color_list_fn()) {
	dice_list <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 4,
		rank_text.die = 1:4,
		dm_text.die = "",
		ps_cex.die = 1,
		fontfamily = ifelse(font == "dejavu", "DejaVu Sans", "sans"),
		grob_fn.die = d4Grob,
		op_grob_fn.die = d4TopGrob,
		obj_fn.die = save_d4_obj,
		width.die = (21 / 25.4) / 0.8660254, # if 21 mm vertex to vertex
		height.die = (21 / 25.4) / 0.8660254, # if 21 mm vertex to vertex
		depth.die = (sqrt(6) / 3) * (21 / 25.4),
		background_color = DICE_PIP_COLORS,
		shape.die_face = "convex3",
		shape_t.die = 90,
		invert_colors = TRUE
	)
	dice <- pp_cfg(c(dice_list, color_list))
	dice$has_piecepack <- FALSE
	dice$has_dice <- TRUE
	dice
}

dice_d8 <- function(font = "sans", color_list = color_list_fn()) {
	dice_list <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 8,
		rank_text.die = 1:8,
		ps_cex.die = 1.15,
		ps_r.die = 0.03,
		ps_t.die = 90,
		dm_text.die = "",
		dm_text.r6.die = "\u2012",
		dm_r.die = ifelse(font == "dejavu", 0.14, 0.13),
		dm_t.die = 270,
		dm_cex.die = 1.5,
		fontfamily = ifelse(font == "dejavu", "DejaVu Sans", "sans"),
		op_grob_fn.die = d8TopGrob,
		obj_fn.die = save_d8_obj,
		width.die = (18 / 25.4) / 0.8660254, # if 18 mm vertex to vertex
		height.die = (18 / 25.4) / 0.8660254, # if 18 mm vertex to vertex
		depth.die = (sqrt(6) / 3) * (18 / 25.4), # inradius = sqrt(6) * a / 6
		background_color = DICE_PIP_COLORS,
		shape.die_face = "convex3",
		shape_t.die = 90,
		invert_colors = TRUE
	)
	dice <- pp_cfg(c(dice_list, color_list))
	dice$has_piecepack <- FALSE
	dice$has_dice <- TRUE
	dice
}

# We'll use alpha-90-beta-90 angle kites
# See `utils-d10.R` for more notes on calculations
dice_d10 <- function(font = "sans", color_list = color_list_fn()) {
	dice_list <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 10L,
		rank_text.die = c(1:9, 0),
		ps_cex.die = 1.15,
		ps_r.die = -0.08,
		ps_t.die = 90,
		dm_text.die = "",
		dm_text.r6.die = "\u2012",
		dm_text.r9.die = "\u2012",
		dm_r.die = ifelse(font == "dejavu", 0.29, 0.28),
		dm_t.die = 270,
		dm_cex.die = 1.5,
		fontfamily = ifelse(font == "dejavu", "DejaVu Sans", "sans"),
		op_grob_fn.die = d10TopGrob,
		obj_fn.die = save_d10_obj,
		width.die = 0.4913446110983896164548, # if kite height 5/8"
		height.die = 5 / 8,
		depth.die = 0.5621585749837085810299,
		background_color = DICE_PIP_COLORS,
		shape.die_face = "kite",
		shape_r.die = 0.5 - 0.1909830056250526320039,
		invert_colors = TRUE
	)
	dice <- pp_cfg(c(dice_list, color_list))
	dice$has_piecepack <- FALSE
	dice$has_dice <- TRUE
	dice
}

dice_d10_percentile <- function(font = "sans", color_list = color_list_fn()) {
	dice_list <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 10L,
		rank_text.die = c(1:9, 0),
		dm_text.die = "0",
		dm_cex.die = 1.15,
		dm_r.die = -0.05,
		dm_t.die = 180,
		ps_cex.die = 1.00,
		ps_r.die = 0.20,
		ps_t.die = 180,
		fontfamily = ifelse(font == "dejavu", "DejaVu Sans", "sans"),
		op_grob_fn.die = d10TopGrob,
		obj_fn.die = save_d10_obj,
		width.die = 5 / 8,
		height.die = 0.4913446110983896164548,
		depth.die = 0.5621585749837085810299,
		background_color = DICE_PIP_COLORS,
		shape.die_face = "kite",
		shape_r.die = 0.5 - 0.1909830056250526320039,
		shape_t.die = 0,
		invert_colors = TRUE
	)
	dice <- pp_cfg(c(dice_list, color_list))
	dice$has_piecepack <- FALSE
	dice$has_dice <- TRUE
	dice
}

dice_d12 <- function(font = "sans", color_list = color_list_fn()) {
	dice_list <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 12L,
		rank_text.die = 1:12,
		ps_cex.die = 1.15,
		ps_r.die = 0.01,
		ps_t.die = 90,
		dm_text.die = "",
		dm_text.r6.die = "\u2012",
		dm_text.r9.die = "\u2012",
		dm_r.die = ifelse(font == "dejavu", 0.25, 0.22),
		dm_t.die = 270,
		dm_cex.die = 1.5,
		fontfamily = ifelse(font == "dejavu", "DejaVu Sans", "sans"),
		op_grob_fn.die = d12TopGrob,
		obj_fn.die = save_d12_obj,
		width.die = (5 / 16) / 0.5877853, # if 5/16" vertex to vertex
		height.die = (5 / 16) / 0.5877853, # if 5/16" vertex to vertex
		depth.die = 2 * 1.113516364 * (5 / 16), # inradius = 1.113516364 * a
		background_color = DICE_PIP_COLORS,
		shape.die_face = "convex5",
		shape_t.die = 90,
		invert_colors = TRUE
	)
	for (i in 10:12) {
		dice_list[[str_glue("ps_cex.r{i}.die")]] <- 1.00
		# dice_list[[str_glue("ps_r.r{i}.die")]] <- 0.02
	}
	dice <- pp_cfg(c(dice_list, color_list))
	dice$has_piecepack <- FALSE
	dice$has_dice <- TRUE
	dice
}

dice_d20 <- function(font = "sans", color_list = color_list_fn()) {
	dice_list <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 20L,
		rank_text.die = 1:20,
		ps_cex.die = 0.65,
		ps_r.die = -0.02,
		ps_t.die = 90,
		fontfamily = ifelse(font == "dejavu", "DejaVu Sans", "sans"),
		dm_text.die = "",
		dm_text.r6.die = "\u2012",
		dm_text.r9.die = "\u2012",
		dm_r.die = ifelse(font == "dejavu", 0.12, 0.13),
		dm_t.die = 270,
		dm_cex.die = 1.0,
		op_grob_fn.die = d20TopGrob,
		obj_fn.die = save_d20_obj,
		width.die = 0.5 / 0.8660254, # if 1/2" vertex to vertex
		height.die = 0.5 / 0.8660254, # if 1/2" vertex to vertex
		depth.die = 2 * 0.5 * (3 * sqrt(3) + sqrt(15)) / 12,
		background_color = DICE_PIP_COLORS,
		shape.die_face = "convex3",
		shape_t.die = 90,
		invert_colors = TRUE
	)
	for (i in 1:9) {
		dice_list[[str_glue("ps_cex.r{i}.die")]] <- 0.72
		dice_list[[str_glue("ps_r.r{i}.die")]] <- 0.02
	}
	dice <- pp_cfg(c(dice_list, color_list))
	dice$has_piecepack <- FALSE
	dice$has_dice <- TRUE
	dice
}

dice_numeral <- function(font = "sans", color_list = color_list_fn(), rect_shape = "rect") {
	dice_list <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 6L,
		fontfamily = ifelse(font == "dejavu", "DejaVu Sans", "sans"),
		rank_text = 1:6,
		ps_cex.die = 1.8,
		dm_text.die = "",
		dm_text.r6.die = "\u2012",
		dm_r.die = 0.315,
		dm_t.die = 270,
		dm_cex.die = 2.4,
		width.die = 16 / 25.4, # 16 mm dice most common
		background_color = DICE_PIP_COLORS,
		invert_colors = TRUE,
		die_arrangement = "1<,2>,3>,6v,5,4",
		shape.card = rect_shape
	)
	dice <- pp_cfg(c(dice_list, color_list))
	dice$has_piecepack <- FALSE
	dice$has_dice <- TRUE
	dice
}

cubes <- function(color_list = color_list_fn()) {
	sizes_mm <- c(8, 10, 12, 15, 20, 25)
	cubes_list <- list(
		shape.bit = "rect",
		n_suits = N_SUITS_COLOR,
		n_ranks = 6L,
		ps_text.bit = "",
		dm_text.bit = "",
		invert_colors.bit = TRUE,
		background_color = "white"
	)
	for (i in seq_along(sizes_mm)) {
		size_in <- sizes_mm[i] / 25.4
		cubes_list[[str_glue("width.r{i}.bit")]] <- size_in
		cubes_list[[str_glue("height.r{i}.bit")]] <- size_in
		cubes_list[[str_glue("depth.r{i}.bit")]] <- size_in
	}
	cubes <- pp_cfg(c(cubes_list, color_list))
	cubes$has_piecepack <- FALSE
	cubes$has_bits <- TRUE
	cubes
}

meeples <- function(color_list = color_list_fn()) {
	meeples_list <- list(
		shape.bit_face = "meeple",
		shape.bit_back = "meeple",
		n_suits = N_SUITS_COLOR,
		width.bit = 16 / 25.4,
		height.bit = 16 / 25.4,
		depth.bit = 10 / 25.4,
		ps_text.bit = "",
		dm_text.bit = "",
		invert_colors.bit = TRUE,
		background_color = "white"
	)
	meeples <- pp_cfg(c(meeples_list, color_list))
	meeples$has_piecepack <- FALSE
	meeples$has_bits <- TRUE
	meeples
}

alquerque <- function(cell_width = 1, color_list = color_list_fn(), shading = FALSE) {
	alquerque <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 1L,
		width.board = 5 * cell_width,
		height.board = 5 * cell_width,
		grob_fn.board_face = alquerqueBoardGrobFn(),
		gridline_color.board_face = cb_suit_colors_pure,
		gridline_color.board_back = "transparent",
		gridline_lex.board = 4,
		suit_color = cb_suit_colors_impure,
		background_color = color_list$background_color,
		gridline_color.s6.board_face = "grey80"
	)
	if (shading) {
		alquerque[["edge_color.board"]] <- cheap_darken(color_list$background_color, 0.3)
	}
	alquerque <- pp_cfg(c(alquerque, go_stone(cell_width / 2, shading), color_list))
	alquerque$has_piecepack <- FALSE
	alquerque$has_boards <- TRUE
	alquerque$has_bits <- TRUE
	alquerque
}

marbles <- function(color_list = color_list_fn(), rect_shape = "rect", shading = FALSE) {
	margin <- 0.25
	marbles <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 9L,
		width.board = 4 + 2 * margin,
		invert_colors.bit = TRUE,
		invert_colors.board = TRUE,
		grob_fn.bit = ellipsoidGrobFn(shading),
		op_grob_fn.bit = basicEllipsoidFn(shading),
		obj_fn.bit = function(...) save_ellipsoid_obj(..., subdivide = 4),
		grob_fn.board = holedBoardGrobFn(4L, 4L, margin),
		op_grob_fn.board = basicHoledBoardFn(4L, 4L, margin),
		obj_fn.board = save_holed_board_obj_fn(4L, 4L, margin),
		shape.board = rect_shape,
		ps_text.bit = "",
		dm_text.bit = "",
		suit_color = cb_suit_colors_impure,
		background_color = color_list$background_color
	)
	if (shading) {
		marbles[["edge_color.board"]] <- cheap_darken(cb_suit_colors_impure, 0.3)
	} else {
		marbles[["edge_color.board"]] <- cb_suit_colors_impure
	}
	for (i in seq.int(9L)) {
		marbles[[str_glue("depth.r{i}.bit")]] <- (7 + i) / 16
		marbles[[str_glue("width.r{i}.bit")]] <- (7 + i) / 16
	}
	for (i in seq.int(2L, 12L)) {
		marbles[[str_glue("width.r{i}.board")]] <- i + 2 * margin
		marbles[[str_glue("height.r{i}.board")]] <- i + 2 * margin
		marbles[[str_glue("grob_fn.r{i}.board")]] <- holedBoardGrobFn(i, i, margin)
		marbles[[str_glue("op_grob_fn.r{i}.board")]] <- basicHoledBoardFn(i, i, margin)
		marbles[[str_glue("obj_fn.r{i}.board")]] <- save_holed_board_obj_fn(i, i, margin)
	}
	marbles <- pp_cfg(c(marbles, color_list))
	marbles$has_piecepack <- FALSE
	marbles$has_boards <- TRUE
	marbles$has_bits <- TRUE
	marbles
}

morris <- function(cell_width = 1, color_list = color_list_fn(), shading = FALSE) {
	morris <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 15L,
		width.board = 7 * cell_width,
		height.board = 7 * cell_width,
		grob_fn.r1.board_face = morrisBoardGrobFn(12),
		gridline_color.board_face = cb_suit_colors_pure,
		gridline_color.board_back = "transparent",
		gridline_lex.board = 4,
		suit_color = cb_suit_colors_impure,
		background_color = color_list$background_color,
		gridline_color.s6.board_face = "grey80"
	)
	for (i in seq.int(2L, 15L)) {
		if (i < 5) {
			morris[[str_glue("width.r{i}.board")]] <- 3 * cell_width
			morris[[str_glue("height.r{i}.board")]] <- 3 * cell_width
		} else if (i < 8) {
			morris[[str_glue("width.r{i}.board")]] <- 5 * cell_width
			morris[[str_glue("height.r{i}.board")]] <- 5 * cell_width
		} else {
			morris[[str_glue("width.r{i}.board")]] <- 7 * cell_width
			morris[[str_glue("height.r{i}.board")]] <- 7 * cell_width
		}
		morris[[str_glue("grob_fn.r{i}.board_face")]] <- morrisBoardGrobFn(i)
	}
	if (shading) {
		morris[["edge_color.board"]] <- cheap_darken(color_list$background_color, 0.3)
	}
	morris <- pp_cfg(c(morris, go_stone(cell_width, shading), color_list))
	morris$has_piecepack <- FALSE
	morris$has_boards <- TRUE
	morris$has_bits <- TRUE
	morris
}

go_stone <- function(cell_width, shading = FALSE) {
	list(
		width.bit = 0.8700787 * cell_width,
		depth.bit = 0.3937008 * cell_width,
		invert_colors.bit = TRUE,
		grob_fn.bit = ellipsoidGrobFn(shading),
		op_grob_fn.bit = basicEllipsoidFn(shading),
		obj_fn.bit = function(...) save_ellipsoid_obj(..., subdivide = 4),
		ps_text.bit = "",
		dm_text.bit = ""
	)
}

dominoes <- function(
	background_color = "white",
	suit_color = "black",
	border_color = "black",
	rect_shape = "rect",
	mat_width = 0
) {
	border_lex <- ifelse(border_color == "black", 4, 0)
	dominoes <- pp_cfg(list(
		n_suits = 18 + 1,
		n_ranks = 18 + 1,
		width.tile = 1,
		height.tile = 2,
		depth.tile = 0.25, # 3/8 professional, 1/2 jumbo
		width.die = 16 / 25.4,
		suit_color = suit_color,
		background_color = background_color,
		mat_width = mat_width,
		mat_color = suit_color,
		border_color = border_color,
		border_lex = border_lex,
		die_arrangement = "opposites_sum_to_5",
		grob_fn.die = pippedGrobFn(0, "die"),
		grob_fn.card_face = cardGrobFn(-1, "card", grob_type = "circle"),
		gridline_color.tile_back = "transparent",
		gridline_color.tile_face = suit_color,
		gridline_lex.tile_face = 6,
		shape.tile = rect_shape,
		shape.card = rect_shape,
		grob_fn.tile_face = dominoGrobFn(-1)
	))
	dominoes$has_piecepack <- FALSE
	dominoes$has_dice <- TRUE
	dominoes$has_tiles <- TRUE
	dominoes
}

checker_piece <- function(cell_width, ps_text.bit_face = "", ps_cex.bit_face = 1.3) {
	list(
		dm_text.bit = "",
		invert_colors.bit = TRUE,
		ps_text.bit_back = "",
		ps_text.bit_face = ps_text.bit_face,
		ps_cex.bit_face = cell_width * ps_cex.bit_face,
		width.bit = 0.75 * cell_width,
		background_color.s5.bit_face = "black",
		background_color.s6.bit_face = "black"
	)
}

checkers <- function(font = "sans", cell_width = 1, color_list = color_list_fn()) {
	checkers <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 12L,
		width.board = 8 * cell_width,
		height.board = 8 * cell_width,
		fontfamily = ifelse(font == "dejavu", "DejaVu Sans", "sans"),
		grob_fn.r1.board_face = checkeredBoardGrobFn(8, 8),
		grob_fn.r1.board_back = linedBoardGrobFn(8, 8),
		gridline_color.board_face = cb_suit_colors_impure,
		gridline_color.board_back = cb_suit_colors_pure,
		gridline_lex.board = 4,
		suit_color = cb_suit_colors_impure,
		background_color = "white",
		gridline_color.s6.board_face = "grey80",
		gridline_color.s6.board_back = "grey80"
	)
	for (i in seq.int(2L, 12L)) {
		checkers[[str_glue("width.r{i}.board")]] <- i * cell_width
		checkers[[str_glue("height.r{i}.board")]] <- i * cell_width
		checkers[[str_glue("grob_fn.r{i}.board_face")]] <- checkeredBoardGrobFn(i, i)
		checkers[[str_glue("grob_fn.r{i}.board_back")]] <- linedBoardGrobFn(i, i)
	}
	# \u25cb white circle \u25cf black circle \u265b black queen \u2605 black star
	# \u272a circled white star
	if (font == "sans") {
		bits <- checker_piece(cell_width, "\u25cb", 1.7)
	} else {
		bits <- checker_piece(cell_width, "\u272a", 1.3)
	}
	checkers <- pp_cfg(c(checkers, bits, color_list))
	checkers$has_piecepack <- FALSE
	checkers$has_boards <- TRUE
	checkers$has_bits <- TRUE
	checkers
}

chess <- function(font = "sans", cell_width = 1, color_list = color_list_fn()) {
	if (font == "sans") {
		rank_cex_die <- 1.3
		black_chess_ranks <- c("p", "n", "b", "r", "q", "k")
		white_chess_ranks <- c("P", "N", "B", "R", "Q", "K")
	} else if (font == "dejavu") {
		rank_cex_die <- 1.5
		black_chess_ranks <- c("\u265f", "\u265e", "\u265d", "\u265c", "\u265b", "\u265a")
		white_chess_ranks <- c("\u2659", "\u2658", "\u2657", "\u2656", "\u2655", "\u2654")
	}
	chess <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 12L,
		width.board = 8 * cell_width,
		height.board = 8 * cell_width,
		width.bit = 0.75 * cell_width,
		ps_text.bit_back = "",
		dm_text.bit = "",
		fontfamily = ifelse(font == "dejavu", "DejaVu Sans", "sans"),
		grob_fn.r1.board_face = checkeredBoardGrobFn(8, 8),
		grob_fn.r1.board_back = linedBoardGrobFn(8, 8),
		grob_fn.s5.bit_face = basicPieceGrobFn(fill_stroke = TRUE),
		grob_fn.s5.die_face = basicPieceGrobFn(fill_stroke = TRUE),
		grob_fn.s7.bit_face = basicPieceGrobFn(fill_stroke = TRUE),
		grob_fn.s7.die_face = basicPieceGrobFn(fill_stroke = TRUE),
		grob_fn.s8.bit_face = basicPieceGrobFn(fill_stroke = TRUE),
		grob_fn.s8.die_face = basicPieceGrobFn(fill_stroke = TRUE),
		# grob_fn.s6.bit_face = basicPieceGrob,
		gridline_color.board_face = cb_suit_colors_impure,
		gridline_color.board_back = cb_suit_colors_pure,
		gridline_lex.board = 4,
		suit_text = "",
		rank_cex.bit = 1.4 * cell_width,
		rank_cex.die = rank_cex_die,
		rank_text = black_chess_ranks,
		rank_text.s6 = white_chess_ranks,
		suit_color = cb_suit_colors_pure,
		suit_color.s6 = "black",
		background_color = "white",
		edge_color.bit = color_list$edge_color.board,
		gridline_color.s6.board_face = "grey80",
		gridline_color.s6.board_back = "grey80"
	)
	for (i in seq.int(2L, 12L)) {
		chess[[str_glue("width.r{i}.board")]] <- i * cell_width
		chess[[str_glue("height.r{i}.board")]] <- i * cell_width
		chess[[str_glue("grob_fn.r{i}.board_face")]] <- checkeredBoardGrobFn(i, i)
		chess[[str_glue("grob_fn.r{i}.board_back")]] <- linedBoardGrobFn(i, i)
	}
	chess <- pp_cfg(c(chess, color_list))
	chess$has_piecepack <- FALSE
	chess$has_boards <- TRUE
	chess$has_bits <- TRUE
	chess$has_dice <- TRUE
	chess
}

go <- function(cell_width = 1, color_list = color_list_fn(), shading = FALSE) {
	go <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 19L,
		width.board = 19 * cell_width,
		height.board = 19 * cell_width,
		grob_fn.board_back = basicPieceGrob,
		grob_fn.r1.board_face = linedBoardGrobFn(18L, 18L, 0.5),
		gridline_color.board_face = cb_suit_colors_pure,
		gridline_color.board_back = "transparent",
		gridline_lex.board = 4,
		suit_color = cb_suit_colors_impure,
		background_color = color_list$background_color,
		gridline_color.s6.board_face = "grey80"
	)
	if (shading) {
		go[["edge_color.board"]] <- cheap_darken(color_list$background_color, 0.3)
	}
	for (i in seq.int(2L, 19L)) {
		go[[str_glue("width.r{i}.board")]] <- i * cell_width
		go[[str_glue("height.r{i}.board")]] <- i * cell_width
		go[[str_glue("grob_fn.r{i}.board_face")]] <- linedBoardGrobFn(i - 1L, i - 1L, 0.5)
	}
	go <- pp_cfg(c(go, go_stone(cell_width, shading), color_list))
	go$has_piecepack <- FALSE
	go$has_boards <- TRUE
	go$has_bits <- TRUE
	go
}

piecepack <- function(
	font = "sans",
	color_list = color_list_fn(),
	rect_shape = "rect",
	pawn = "token",
	background_color = "white",
	edge_color = "white"
) {
	if (font == "sans") {
		piecepack_suit_text <- list(suit_text = "\u263c,\u25d8,\u0238,\u03ee,\u2202", n_suits = 4L)
		pce_suit_text <- list(suit_text = "\u2665,\u2660,\u2663,\u2666,\u2202", n_suits = 4L)
		dpe_suit_text <- list(
			suit_text = "\u2665,\u2660,\u2663,\u2666,*,\u2202",
			suit_cex.s5 = 1.3,
			n_suits = 5L
		)
	} else if (font == "dejavu") {
		piecepack_suit_text <- list(
			suit_text = "\u2742,\u25d0,\u265b,\u269c,\u0ed1",
			suit_cex.s2 = 0.9,
			dm_cex.coin = 0.5,
			n_suits = 4L
		)
		pce_suit_text <- list(suit_text = "\u2665,\u2660,\u2663,\u2666,\u0ed1", n_suits = 4L)
		dpe_suit_text <- list(
			suit_text = "\u2665,\u2660,\u2663,\u2666,\u2605,\u0ed1",
			suit_cex.s5 = 0.92,
			n_suits = 5L,
			grob_fn.s5 = basicPieceGrobFn(fill_stroke = TRUE),
			grob_fn.s5.card_face = cardGrobFn(-1, fill_stroke = TRUE)
		)
	}

	piecepack_base <- list(
		depth.coin = 0.25,
		invert_colors.matchstick = TRUE,
		ps_cex.r2.matchstick = 0.7,
		dm_r.r1.matchstick = 0,
		dm_cex.r1.matchstick = 1.5,
		suit_color.s2.matchstick = "grey30",
		suit_color.s2.bit = "grey30",
		mat_color.tile_back = background_color,
		mat_width.tile_back = 0.05,
		fontfamily = ifelse(font == "dejavu", "DejaVu Sans", "sans"),
		background_color.die = background_color,
		background_color.pyramid = background_color,
		background_color.matchstick = background_color,
		edge_color.tile = edge_color,
		edge_color.coin = edge_color,
		suit_color.unsuited = "black",
		invert_colors.bit = TRUE,
		rank_text = ",a,2,3,4,5",
		use_suit_as_ace = TRUE,
		rank_text.pyramid = LETTERS[1:6],
		suit_text.saucer_back = "",
		suit_cex.saucer_face = 0.9,
		suit_cex.pyramid_face = 0.5,
		ps_r.pyramid_face = -0.08,
		dm_cex.pyramid_face = 4.0,
		dm_text.pyramid_face = "|",
		dm_r.pyramid_face = -0.22,
		ps_cex.pyramid_left = 0.7,
		ps_r.pyramid_left = -0.00,
		ps_cex.pyramid_right = 0.7,
		ps_r.pyramid_right = -0.00,
		grob_fn.card_face = cardGrobFn(-1),
		use_suit_as_ace.pyramid = FALSE,
		shape.tile = rect_shape,
		shape.card = rect_shape
	)
	shapes <- shapes_cfg(color_list)
	pawn <- switch(
		pawn,
		"peg-doll" = peg_doll_pawn(shapes),
		"joystick" = joystick_pawn(shapes),
		NULL
	)
	piecepack <- c(pawn, piecepack_suit_text, color_list, piecepack_base)

	playing_cards_expansion <- c(pce_suit_text, piecepack)
	playing_cards_expansion$suit_color <- "#D55E00,#000000,#000000,#D55E00,#F0E442"
	playing_cards_expansion$suit_color.s3.matchstick <- "grey30"
	playing_cards_expansion$suit_color.s3.bit <- "grey30"

	hexpack <- c(
		piecepack,
		list(
			shape.tile_face = "convex6",
			shape.tile_back = "convex6",
			border_lex = 3,
			shape_t.tile = "60",
			dm_t.tile_face = -90,
			width.tile = 4 / sqrt(3),
			height.tile = 4 / sqrt(3),
			shape.coin_face = "convex3",
			shape.coin_back = "convex3"
		)
	)

	dpe_base <- list(
		invert_colors.suited = TRUE,
		border_color.s2.die = "black",
		border_color.s2.pawn = "black",
		suit_color.s2.board_face = "black"
	)

	dual_piecepacks_expansion <- c(dpe_suit_text, dpe_base, piecepack)

	pi_base <- list(
		invert_colors = TRUE,
		invert_colors.matchstick = FALSE,
		suit_color.s2.matchstick = "black",
		suit_color.s2.board_face = "black",
		suit_color.card_back = "grey30",
		suit_color.coin_face = "grey30",
		edge_color.coin = "white",
		edge_color.tile = "white"
	)
	piecepack_inverted <- c(pi_base, piecepack)

	piecepack <- c(list(suit_color = cb_suit_colors_pure), piecepack)

	base <- pp_cfg(piecepack)
	dpe <- pp_cfg(dual_piecepacks_expansion)
	hexpack <- to_hexpack(piecepack)
	inverted <- pp_cfg(piecepack_inverted)
	pce <- pp_cfg(playing_cards_expansion)
	subpack <- to_subpack(piecepack)

	base$has_cards <- TRUE
	dpe$has_cards <- TRUE
	inverted$has_cards <- TRUE
	pce$has_cards <- TRUE

	base$has_matchsticks <- TRUE
	dpe$has_matchsticks <- TRUE
	inverted$has_matchsticks <- TRUE
	pce$has_matchsticks <- TRUE

	base$has_pyramids <- TRUE
	dpe$has_pyramids <- TRUE
	inverted$has_pyramids <- TRUE
	pce$has_pyramids <- TRUE

	base$has_saucers <- TRUE
	dpe$has_saucers <- TRUE
	inverted$has_saucers <- TRUE
	pce$has_saucers <- TRUE

	list(
		base = base,
		dpe = dpe,
		hexpack = hexpack,
		inverted = inverted,
		pce = pce,
		subpack = subpack
	)
}

playing_cards <- function(font = "sans", rect_shape = "rect") {
	if (font == "sans") {
		face_labels <- c("", "\u050a", "\u046a", "\u0238")
		fool_text <- "*"
		pc_suit_text <- list(suit_text = "\u2665,\u2660,\u2663,\u2666,*", suit_cex.s5 = 1.3)
	} else if (font == "dejavu") {
		face_labels <- c("", "\u265e", "\u265b", "\u265a")
		fool_text <- "\u2605"
		pc_suit_text <- list(suit_text = "\u2665,\u2660,\u2663,\u2666,\u2605", suit_cex.s5 = 0.92)
	}

	playing_cards_list <- list(
		n_ranks = 14,
		rank_text = "A,2,3,4,5,6,7,8,9,10,J,Q,K,\n\n\n\nJ\nO\nK\nE\nR",
		grob_fn.card_face = cardGrobFn(),
		grob_fn.r11.card_face = faceCardGrobFn(face_labels[1]),
		grob_fn.r12.card_face = faceCardGrobFn(face_labels[3]),
		grob_fn.r13.card_face = faceCardGrobFn(face_labels[4]),
		grob_fn.r14.card_face = jokerCardGrobFn(TRUE),
		shape.card = rect_shape,
		border_color = "black",
		border_lex = 4
	)
	playing_cards_list$n_suits <- 4
	playing_cards_list$suit_color <- "#D55E00,#000000,#000000,#D55E00"

	playing_cards <- c(playing_cards_list, pc_suit_text)
	playing_cards$grob_fn.s3.r14.card_face <- jokerCardGrobFn(FALSE)
	playing_cards$grob_fn.s4.r14.card_face <- jokerCardGrobFn(FALSE)
	playing_cards$grob_fn.s5.card_face <- cardGrobFn(fill_stroke = TRUE)
	playing_cards$grob_fn.s5.r11.card_face <- faceCardGrobFn(face_labels[1], fill_stroke = TRUE)
	playing_cards$grob_fn.s5.r12.card_face <- faceCardGrobFn(face_labels[3], fill_stroke = TRUE)
	playing_cards$grob_fn.s5.r13.card_face <- faceCardGrobFn(face_labels[4], fill_stroke = TRUE)
	playing_cards$grob_fn.s5.r14.card_face <- jokerCardGrobFn(TRUE, fill_stroke = TRUE)

	playing_cards <- pp_cfg(playing_cards)
	playing_cards$has_piecepack <- FALSE
	playing_cards$has_cards <- TRUE

	playing_cards_colored <- c(playing_cards_list, pc_suit_text)
	playing_cards_colored$n_suits <- 5
	playing_cards_colored$suit_color <- cb_suit_colors_pure
	playing_cards_colored$grob_fn.s5.card_face <- cardGrobFn(fill_stroke = TRUE)
	playing_cards_colored$grob_fn.s5.r11.card_face <- faceCardGrobFn(
		face_labels[1],
		fill_stroke = TRUE
	)
	playing_cards_colored$grob_fn.s5.r12.card_face <- faceCardGrobFn(
		face_labels[3],
		fill_stroke = TRUE
	)
	playing_cards_colored$grob_fn.s5.r13.card_face <- faceCardGrobFn(
		face_labels[4],
		fill_stroke = TRUE
	)
	playing_cards_colored$grob_fn.s5.r14.card_face <- jokerCardGrobFn(TRUE, fill_stroke = TRUE)

	playing_cards_colored <- pp_cfg(playing_cards_colored)
	playing_cards_colored$has_piecepack <- FALSE
	playing_cards_colored$has_cards <- TRUE

	playing_cards_tarot <- playing_cards_list
	playing_cards_tarot$rank_text <- "A,2,3,4,5,6,7,8,9,10,J,C,Q,K,\n\n\n\nJ\nO\nK\nE\nR"
	playing_cards_tarot$grob_fn.r12.card_face <- faceCardGrobFn(face_labels[2], "low")
	playing_cards_tarot$grob_fn.r13.card_face <- faceCardGrobFn(face_labels[3])
	playing_cards_tarot$grob_fn.r14.card_face <- faceCardGrobFn(face_labels[4])
	playing_cards_tarot$grob_fn.r15.card_face <- jokerCardGrobFn(TRUE)
	playing_cards_tarot$grob_fn.s3.r15.card_face <- jokerCardGrobFn(FALSE)
	playing_cards_tarot$grob_fn.s4.r15.card_face <- jokerCardGrobFn(FALSE)
	playing_cards_tarot$n_suits <- 5
	playing_cards_tarot$n_ranks <- 22

	tarot_suit_text <- "\u2665,\u2660,\u2663,\u2666,"
	playing_cards_tarot$suit_text <- tarot_suit_text
	playing_cards_tarot$suit_text.r14 <- tarot_suit_text
	playing_cards_tarot$suit_color <- "#D55E00,#000000,#000000,#D55E00,#000000"
	playing_cards_tarot$rank_text.s5 <- c(1:21, fool_text)
	playing_cards_tarot$grob_fn.s5.card_face <- cardGrobFn(has_pips = FALSE, fill_stroke = FALSE)
	for (i in 11:15) {
		name <- str_glue("grob_fn.s5.r{i}.card_face")
		playing_cards_tarot[[name]] <- cardGrobFn(has_pips = FALSE, fill_stroke = FALSE)
	}

	playing_cards_tarot <- pp_cfg(playing_cards_tarot)
	playing_cards_tarot$has_piecepack <- FALSE
	playing_cards_tarot$has_cards <- TRUE

	list(base = playing_cards, color = playing_cards_colored, tarot = playing_cards_tarot)
}

reversi <- function(cell_width = 1, color_list = color_list_fn()) {
	reversi <- list(
		n_suits = 8L, # N_SUITS_COLOR?
		n_ranks = 12L,
		width.board = 8 * cell_width,
		height.board = 8 * cell_width,
		grob_fn.r1.board_face = linedBoardGrobFn(8, 8),
		grob_fn.r1.board_back = checkeredBoardGrobFn(8, 8),
		background_color.board_face = cb_suit_colors_impure,
		gridline_color.board_face = "black",
		gridline_lex.board = 4,
		suit_color = cb_suit_colors_impure,
		background_color = "white",
		gridline_color.s6.board_back = "grey80"
	)
	for (i in seq.int(2L, 12L)) {
		reversi[[str_glue("width.r{i}.board")]] <- i * cell_width
		reversi[[str_glue("height.r{i}.board")]] <- i * cell_width
		reversi[[str_glue("grob_fn.r{i}.board_face")]] <- linedBoardGrobFn(i, i)
		reversi[[str_glue("grob_fn.r{i}.board_back")]] <- checkeredBoardGrobFn(i, i)
	}
	reversi <- pp_cfg(c(reversi, reversi_piece(cell_width, color_list), color_list))
	reversi$has_piecepack <- FALSE
	reversi$has_boards <- TRUE
	reversi$has_bits <- TRUE
	reversi
}

shapes_cfg <- function(color_list = color_list_fn()) {
	shapes <- list(
		n_suits = N_SUITS_COLOR,
		n_ranks = 4,
		invert_colors = TRUE,
		ps_text = "",
		dm_text = "",
		background_color = "white",
		width = 1,
		height = 1,
		depth = 1,
		shape.r1.bit_face = "circle",
		shape.r1.bit_back = "circle",
		shape.r2.bit = "circle",
		# don't shade joystick sphere at top if rest of joystick isn't shaded?
		grob_fn.r2.bit = ellipsoidGrobFn(FALSE),
		op_grob_fn.r2.bit = basicEllipsoidFn(FALSE),
		obj_fn.r2.bit = function(...) save_ellipsoid_obj(..., subdivide = 4)
	)
	pp_cfg(c(shapes, color_list))
}

reversi_piece <- function(cell_width = 1, color_list = color_list_fn()) {
	shapes_top <- shapes_cfg(color_list)
	color_list$suit_color <- color_list$suit_color[c(7L, 6L, 8L, 5L, 4L, 2L, 1L, 3L)]
	shapes_bot <- shapes_cfg(color_list)
	envir <- list(shapes_top = shapes_top, shapes_bot = shapes_bot)

	df_reversi <- tibble(
		piece_side = "bit_back",
		rank = 1,
		width = 1,
		height = 1,
		depth = c(0.5, 0.5),
		x = 0.5,
		y = 0.5,
		z = c(0.25, 0.75),
		cfg = c("shapes_bot", "shapes_top")
	)

	reversi_piece <- CompositePiece$new(df_reversi, envir = envir, ref_side = "face")

	list(
		width.bit = 0.75 * cell_width,
		height.bit = 0.70 * cell_width,
		depth.bit = 0.25 * cell_width,
		grob_fn.bit = reversi_piece$grob_fn,
		obj_fn.bit = reversi_piece$obj_fn,
		op_grob_fn.bit = reversi_piece$op_grob_fn
	)
}

peg_doll_pawn <- function(shapes) {
	pegdoll_depth <- c(0.55, 1.0 * 0.75 / 1.5)
	df_pegdoll <- tibble(
		piece_side = "bit_back",
		rank = c(1, 2),
		width = 1,
		height = 1,
		depth = pegdoll_depth,
		x = 0.5,
		y = 0.5,
		z = c(0.5 * pegdoll_depth[1], 1 - 0.5 * pegdoll_depth[2]),
		cfg = "shapes"
	)

	pegdoll <- CompositePiece$new(df_pegdoll, envir = list(shapes = shapes))

	list(
		width.pawn = 0.75,
		depth.pawn = 0.75,
		height.pawn = 1.5,
		edge_color.pawn = cb_suit_colors_pure,
		background_color.belt_face = "white",
		mat_color.belt_face = "transparent",
		suit_cex.belt_face = 1.5,
		obj_fn.pawn = save_peg_doll_obj,
		grob_fn.pawn = pegdoll$grob_fn,
		op_grob_fn.pawn = pegdoll$op_grob_fn
	)
}

joystick_pawn <- function(shapes) {
	jd <- c(0.15, 0.6, 0.5 * 5 / 8)
	df_joystick <- tibble(
		piece_side = "bit_back",
		rank = c(1, 1, 2),
		width = c(1, 0.3, 0.5),
		height = c(1, 0.3, 0.5),
		depth = jd,
		x = rep(0.5, 3),
		y = rep(0.5, 3),
		z = c(0.5 * jd[1], jd[1] + 0.5 * jd[2], 1 - 0.5 * jd[3]),
		cfg = "shapes"
	)

	joystick <- CompositePiece$new(df_joystick, envir = list(shapes = shapes))

	list(
		grob_fn.pawn = joystick$grob_fn,
		obj_fn.pawn = joystick$obj_fn,
		op_grob_fn.pawn = joystick$op_grob_fn,
		width.pawn = 5 / 8,
		height.pawn = 1.0,
		depth.pawn = 5 / 8
	)
}

#' @rdname game_systems
#' @inheritParams pp_cfg
#' @export
to_hexpack <- function(cfg = getOption("piecepackr.cfg", pp_cfg())) {
	cfg <- as_pp_cfg(cfg)
	hexpack <- as.list(cfg)
	hexpack$shape.tile_face <- "convex6"
	hexpack$shape.tile_back <- "convex6"
	hexpack$border_lex <- 3
	hexpack$shape_t.tile <- 60
	hexpack$dm_t.tile_face <- -90
	hexpack$width.tile <- 4 / sqrt(3)
	hexpack$height.tile <- 4 / sqrt(3)
	hexpack$shape.coin_face <- "convex3"
	hexpack$shape.coin_back <- "convex3"
	pp_cfg(hexpack)
}

#' @rdname game_systems
#' @inheritParams pp_cfg
#' @export
to_subpack <- function(cfg = getOption("piecepackr.cfg", pp_cfg())) {
	cfg <- as_pp_cfg(cfg)
	subpack <- as.list(cfg)
	subpack$width.tile <- (3 / 8) * cfg$get_width("tile")
	subpack$width.coin <- 0.4 * cfg$get_width("coin")
	subpack$width.die <- 0.6 * cfg$get_width("die")
	subpack$width.pawn <- 0.4 * cfg$get_width("pawn")
	subpack$height.pawn <- 0.4 * cfg$get_height("pawn")
	subpack$width.saucer <- 0.4 * cfg$get_width("saucer")
	subpack$cex <- 0.4
	pp_cfg(subpack)
}

cfg <- piecepackr::game_systems("sans", border = FALSE)$piecepack
test_that("save_piece_obj works", {
	files <- save_piece_obj("tile_face", z = NA, cfg = cfg)
	expect_length(files, 3)
	files <- save_piece_obj("tile_back", cfg = cfg, angle = NA, axis_x = NA, axis_y = NA)
	expect_length(files, 3)
	files <- save_piece_obj("die_face", cfg = cfg)
	expect_length(files, 3)
	files <- save_piece_obj("pyramid_top", cfg = cfg)
	expect_length(files, 3)
	files <- save_piece_obj("pyramid_left", cfg = cfg)
	expect_length(files, 3)
	files <- save_piece_obj("pyramid_right", cfg = cfg)
	expect_length(files, 3)
	files <- save_piece_obj("pyramid_back", cfg = cfg)
	expect_length(files, 3)
	files <- save_piece_obj("pyramid_face", cfg = cfg)
	expect_length(files, 3)
	expect_error(save_piece_obj("belt_face"))
})

test_that("rgl works", {
	skip_on_cran()
	skip_if_not_installed("rgl")
	skip_if_not_installed("systemfonts") # prevent buggy cairo interaction with "dejavu" font
	library("rgl", quietly = TRUE, include.only = c("clear3d", "close3d", "open3d", "view3d"))
	# go stone OBJ generation uses "rgl" in background
	files <- save_piece_obj("bit_face", cfg = game_systems("dejavu", border = FALSE)$go)
	expect_length(files, 3)

	files <- save_piece_obj(
		"pawn_top",
		cfg = game_systems("sans", border = FALSE, pawn = "peg-doll")$piecepack
	)
	expect_length(files, 3)

	open3d(useNULL = TRUE)
	df <- tibble::tibble(piece_side = "tile_face", scale = c(1, 0))
	pmap_piece(df, piece3d, cfg = cfg)

	clear3d()
	cfg <- game_systems("sans", border = FALSE, pawn = "joystick")$piecepack
	l <- piece3d("pawn_top", x = -1:1, suit = 1:3, cfg = cfg, lit = TRUE)
	expect_true(length(l) > 2)

	clear3d()
	cfg <- game_systems()$dice_d4
	l <- piece3d("die_face", x = 1:4, suit = 1:4, rank = 1:4, cfg = cfg)
	expect_true(length(l) == 4)

	clear3d()
	l <- piece3d("die_face", x = 1:8, suit = c(1:6, 1:2), rank = 1:8, cfg = dice_d8())
	expect_true(length(l) == 8)

	clear3d()
	l <- piece3d(
		"die_face",
		x = rep(1:5, 2),
		y = rep(1:2, each = 5),
		rank = 1:10,
		suit = rep(1:5, 2),
		cfg = dice_d10()
	)
	expect_true(length(l) == 10)

	clear3d()
	l <- piece3d(
		"die_face",
		x = rep(1:5, 2),
		y = rep(1:2, each = 5),
		rank = 1:10,
		suit = rep(1:5, 2),
		cfg = dice_d10_percentile()
	)
	expect_true(length(l) == 10)

	clear3d()
	l <- piece3d(
		"die_face",
		x = rep(1:6, 2),
		y = rep(1:2, each = 6),
		rank = 1:12,
		suit = rep(1:6, length.out = 12),
		cfg = dice_d12()
	)
	expect_true(length(l) == 12)

	clear3d()
	l <- piece3d(
		"die_face",
		x = rep(1:5, 4),
		y = rep(1:4, each = 5),
		rank = 1:20,
		suit = rep(1:6, length.out = 20),
		cfg = dice_d20()
	)
	expect_true(length(l) == 20)

	close3d()
})

test_that("rayrender works", {
	skip_on_cran()
	skip_if_not_installed("rayrender")
	library("rayrender", include.only = "render_scene", quietly = TRUE)
	scene <- piece("coin_face", x = -1:1, rank = 1:3, cfg = cfg)
	f <- tempfile(fileext = ".jpeg")
	png(f)
	render_scene(scene, samples = 1, interactive = FALSE)
	dev.off()
	expect_true(file.exists(f))
	unlink(f)

	expect_null(piece("coin_face", x = -1:1, rank = 1:3, cfg = cfg, scale = 0))

	cfg <- game_systems("sans", border = FALSE, pawn = "joystick")$piecepack
	skip_if_not_installed("rgl") # needed to generate joystick pawn obj
	scene <- piece("pawn_top", x = -1:1, suit = 1:3, cfg = cfg)
	expect_equal(nrow(scene), 9)
	# render_scene(scene, samples = 1, lookfrom=c(0, -5, 5)) # nolint
})

test_that("rayvertex works", {
	skip_on_cran()
	skip_if_not_installed("rayvertex")
	library("rayvertex", quietly = TRUE, include.only = c("directional_light", "rasterize_scene"))
	scene <- piece_mesh("coin_face", x = -1:1, rank = 1:3, cfg = cfg)
	f <- tempfile(fileext = ".png")
	png(f)
	rasterize_scene(scene, light_info = directional_light(c(0, 0, 1)), lookat = c(0, 0, 0))
	dev.off()
	expect_true(file.exists(f))
	unlink(f)

	expect_null(piece_mesh("coin_face", x = -1:1, rank = 1:3, cfg = cfg, scale = 0))

	envir <- game_systems("sans", border = FALSE, pawn = "joystick")
	cfg <- envir$piecepack
	skip_if_not_installed("rgl") # needed to generate joystick pawn obj
	scene <- piece_mesh("pawn_top", x = -1:1, suit = c(1, 3, 4), cfg = cfg)
	# rasterize_scene(scene, lookfrom=c(0, -5, 5),
	#                 light_info = directional_light(c(0, -5, 5))) # nolint

	cfg <- envir$marbles
	scene <- piece_mesh("board_face", x = 0, y = 0, cfg = cfg)
	# rasterize_scene(scene, lookfrom=c(0, -5, 15),
	#                 light_info = directional_light(c(0, -5, 5))) # nolint
})

Token2S <- R6Class(
	"token2s",
	public = list(
		initialize = function(
			shape = pp_shape(),
			whd = list(width = 1, height = 1, depth = 1),
			center = as_coord3d(0, 0, 0),
			R = diag(3)
		) {
			# coordinates
			xyz_scaled <- as_coord3d(shape$npc_coords)$translate(as_coord3d(-0.5, -0.5, 0.5))$scale(
				whd$width,
				whd$height,
				whd$depth
			)
			xyz_f <- xyz_scaled$transform(R)$translate(center)

			xyz_scaled <- as_coord3d(shape$npc_coords)$translate(as_coord3d(
				-0.5,
				-0.5,
				-0.5
			))$scale(whd$width, whd$height, whd$depth)
			xyz_b <- xyz_scaled$transform(R)$translate(center)

			self$xyz <- as_coord3d(
				x = c(xyz_f$x, xyz_b$x),
				y = c(xyz_f$y, xyz_b$y),
				z = c(xyz_f$z, xyz_b$z)
			)

			xy_vp <- as_coord2d(rect_xy)
			xyz_scaled <- as_coord3d(xy_vp)$translate(as_coord3d(-0.5, -0.5, 0.5))$scale(
				whd$width,
				whd$height,
				whd$depth
			)
			self$xyz_vp_face <- xyz_scaled$transform(R)$translate(center)

			# Flip so back also in "upper_left", "lower_left", "lower_right", "upper_right" order
			xy_vp <- as_coord2d(x = rect_xy$x[4:1], y = rect_xy$y[4:1])
			xyz_scaled <- as_coord3d(xy_vp)$translate(as_coord3d(-0.5, -0.5, -0.5))$scale(
				whd$width,
				whd$height,
				whd$depth
			)
			self$xyz_vp_back <- xyz_scaled$transform(R)$translate(center)

			# edges
			edge_partition <- partition_edges(shape)
			n_points <- length(xyz_f)
			n_edges <- length(edge_partition$type)
			edges <- vector("list", n_edges)
			for (i in seq(n_edges)) {
				vertices <- self$xyz[edge_indices(edge_partition$indices[[i]], n_points)]
				type <- edge_partition$type[i]
				edges[[i]] <- edge_class(type, vertices)
			}
			self$edges <- edges
		},
		op_edge_order = function(angle) {
			a <- degrees(angle)
			depths <- sapply(self$edges, function(e) {
				painter_depth(mean(e$vertices), scale = 1, alpha = a)
			})
			order(depths)
		},
		op_edges = function(angle) {
			self$edges[self$op_edge_order(angle)]
		},
		op_xy_vp = function(angle, scale, side) {
			switch(
				side,
				face = as_coord2d(self$xyz_vp_face, alpha = degrees(angle), scale = scale),
				back = as_coord2d(self$xyz_vp_back, alpha = degrees(angle), scale = scale),
				abort(paste("Can't handle side", side))
			)
		},
		#### Handle edge case for token (almost) parallel to xy-axis
		visible_side = function(angle) {
			a <- degrees(angle)
			if (
				painter_depth(mean(self$xyz_face), scale = 1, alpha = a) >
					painter_depth(mean(self$xyz_back), scale = 1, alpha = a)
			) {
				"face"
			} else {
				"back"
			}
		},
		xyz_side = function(side) {
			switch(side, face = self$xyz_face, back = self$xyz_back)
		},
		xyz = NULL,
		edges = NULL,
		xyz_vp_face = NULL,
		xyz_vp_back = NULL
	),
	private = list(),
	active = list(
		xyz_face = function() {
			n <- length(self$xyz) / 2
			self$xyz[seq(n)]
		},
		xyz_back = function() {
			n <- length(self$xyz) / 2
			self$xyz[seq(n + 1, 2 * n)]
		}
	)
)

edge_indices <- function(indices, n_points) {
	c(indices, n_points + rev(indices))
}

edge_class <- function(type, vertices) {
	switch(
		type,
		curved = CurvedEdge$new(vertices),
		flat = FlatEdge$new(vertices),
		ring = RingEdge$new(vertices)
	)
}

partition_edges <- function(shape) {
	classes <- shape$npc_coords$c
	n <- length(classes)
	if (all(classes == "C0")) {
		list(
			type = rep("flat", n),
			indices = lapply(seq_along(classes), function(x) c(x, x %% n + 1))
		)
	} else if (all(classes == "C1")) {
		list(type = "ring", indices = list(seq_along(classes)))
	} else {
		if (!(classes[n] == "C0")) {
			abort("Can only handle case when last class is 'C0'")
		}
		type <- vector("character")
		indices <- vector("list")
		index <- 1
		prev <- classes[n]
		curve_start <- NULL
		for (i in seq_along(classes)) {
			class <- classes[i]
			if (all(c(prev, class) == "C0")) {
				type[index] <- "flat"
				i_prev <- ifelse(i == 1, n, i - 1)
				indices[[index]] <- c(i_prev, i)
				index <- index + 1
			} else if (prev == "C1" && class == "C0") {
				# end of curved
				type[index] <- "curved"
				if (curve_start == 0) {
					indices[[index]] <- c(n, seq(i))
				} else {
					indices[[index]] <- seq(curve_start, i)
				}
				index <- index + 1
			} else if (prev == "C0" && class == "C1") {
				# start of curved
				curve_start <- i - 1
			}
			prev <- class
		}
		list(type = type, indices = indices)
	}
}

Edge <- R6Class(
	"edge",
	public = list(
		vertices = NULL,
		initialize = function(vertices = NULL) self$vertices <- vertices,
		visible_side = function(angle) {
			a <- degrees(angle)
			if (
				painter_depth(mean(self$vertices_face), scale = 1, alpha = a) >
					painter_depth(mean(self$vertices_back), scale = 1, alpha = a)
			) {
				"face"
			} else {
				"back"
			}
		},
		vertices_visible_side = function(angle) {
			side <- self$visible_side(angle)
			switch(side, face = self$vertices_face, back = self$vertices_back)
		}
	),
	private = list(),
	active = list(
		# won't work for "FlatEdge" but doesn't need to
		is_perpendicular_xyplane = function() {
			length(unique(round(self$vertices$z, 6))) == 2
		},
		vertices_face = function() {
			n <- length(self$vertices) / 2
			self$vertices[seq(n)]
		},
		vertices_back = function() {
			n <- length(self$vertices) / 2
			self$vertices[seq(n + 1, 2 * n)]
		}
	)
)

FlatEdge <- R6Class(
	"edge_flat",
	inherit = Edge,
	public = list(op_grob = function(angle, scale, name = NULL) {
		xy <- as_coord2d(self$vertices, alpha = degrees(angle), scale = scale)
		polygonGrob(xy$x, xy$y, default.units = "in", name = name)
	}),
	private = list(),
	active = list()
)

## Convex, closed curved edge
RingEdge <- R6Class(
	"edge_ring",
	inherit = Edge,
	public = list(op_grob = function(angle, scale, name = NULL) {
		if (self$is_perpendicular_xyplane) {
			private$op_grob_perpendicular(angle, scale, name = name)
		} else {
			private$op_grob_nonperpendicular(angle, scale, name = name)
		}
	}),
	private = list(
		op_grob_nonperpendicular = function(angle, scale, name = NULL) {
			xyh <- convex_hull2d(as_coord2d(self$vertices, alpha = degrees(angle), scale = scale))

			xy_visible <- as_coord2d(
				self$vertices_visible_side(angle),
				alpha = degrees(angle),
				scale = scale
			)
			xyh_visible <- convex_hull2d(xy_visible)

			A <- list(list(x = xyh$x, y = xyh$y))
			B <- list(list(x = xyh_visible$x, y = xyh_visible$y))
			xy_minus <- gridGeometry::polyclip(A, B, op = "minus")

			#### also include obscured edge (i.e. A intersection B)?
			gridGeometry::xyListPolygon(xy_minus, name = name)
		},
		op_grob_perpendicular = function(angle, scale, name = NULL) {
			n <- length(self$vertices) / 2
			xy_f <- as_coord2d(self$vertices[seq(n)], alpha = degrees(angle), scale = scale)
			projections <- numeric(n)
			proj_vec <- as_coord2d(degrees(angle - 90), radius = 1)
			for (ii in seq(n)) {
				projections[ii] <- proj_vec * xy_f[ii]
			}
			i_min <- which.min(projections)
			i_max <- which.max(projections)
			if (i_min < i_max) {
				indices1 <- seq(i_min, i_max)
				if (i_max < n) {
					indices2 <- c(seq(i_max + 1, n), seq_len(i_min - 1))
				} else {
					indices2 <- seq(1, i_min - 1)
				}
			} else {
				indices1 <- c(seq(i_min, n), seq(1, i_max))
				indices2 <- seq(i_max + 1, i_min - 1)
			}
			# figure out which part farthest
			a <- degrees(angle)
			if (
				painter_depth(mean(self$vertices[indices1]), scale = 1, alpha = a) >
					painter_depth(mean(self$vertices[indices2]), scale = 1, alpha = a)
			) {
				indices_obscured <- indices2
				indices_visible <- indices1
			} else {
				indices_obscured <- indices1
				indices_visible <- indices2
			}
			xy <- as_coord2d(self$vertices, alpha = degrees(angle), scale = scale)
			x_obscured <- xy$x[full_indices(indices_obscured, n)]
			y_obscured <- xy$y[full_indices(indices_obscured, n)]
			x_visible <- xy$x[full_indices(indices_visible, n)]
			y_visible <- xy$y[full_indices(indices_visible, n)]
			polygonGrob(
				x = c(x_obscured, x_visible),
				y = c(y_obscured, y_visible),
				id.lengths = c(length(x_obscured), length(x_visible)),
				default.units = "in",
				name = name
			)
		}
	),
	active = list()
)

full_indices <- function(indices, n) c(indices, rev(2 * n + 1 - indices))

## Convex, open curved edge
CurvedEdge <- R6Class(
	"edge_curved",
	inherit = Edge,
	public = list(op_grob = function(angle, scale, name = NULL) {
		if (self$is_perpendicular_xyplane) {
			private$op_grob_perpendicular(angle, scale, name = name)
		} else {
			private$op_grob_nonperpendicular(angle, scale, name = name)
		}
	}),
	private = list(
		op_grob_nonperpendicular = function(angle, scale, name = NULL) {
			xy <- as_coord2d(self$vertices, alpha = degrees(angle), scale = scale)
			xyh <- convex_hull2d(xy)

			xy_visible <- as_coord2d(
				self$vertices_visible_side(angle),
				alpha = degrees(angle),
				scale = scale
			)
			xyh_visible <- convex_hull2d(xy_visible)

			A <- list(list(x = xyh$x, y = xyh$y))
			B <- list(list(x = xyh_visible$x, y = xyh_visible$y))
			xy_minus <- gridGeometry::polyclip(A, B, op = "minus")

			#### also include obscured edge (i.e. A intersection B)?
			gridGeometry::xyListPolygon(xy_minus, name = name)
		},
		op_grob_perpendicular = function(angle, scale, name = NULL) {
			n <- length(self$vertices) / 2
			xy_f <- as_coord2d(self$vertices[seq(n)], alpha = degrees(angle), scale = scale)
			projections <- numeric(n)
			proj_vec <- as_coord2d(degrees(angle - 90), radius = 1)
			for (ii in seq(n)) {
				projections[ii] <- proj_vec * xy_f[ii]
			}
			i_min <- which.min(projections)
			i_max <- which.max(projections)
			l_indices <- list()
			i_l <- min(i_min, i_max)
			i_u <- max(i_min, i_max)
			if (i_l == 1 && i_u == n) {
				l_indices[[1]] <- seq(n)
			} else if (i_l == 1) {
				l_indices[[1]] <- seq(i_u + 1, n)
				l_indices[[2]] <- seq(1, i_u)
			} else if (i_u == n) {
				l_indices[[1]] <- seq(1, i_l - 1)
				l_indices[[2]] <- seq(i_l, n)
			} else {
				l_indices[[1]] <- seq(1, i_l - 1)
				l_indices[[2]] <- seq(i_u + 1, n)
				l_indices[[3]] <- seq(i_l, i_u)
			}
			xy <- as_coord2d(self$vertices, alpha = degrees(angle), scale = scale)
			x <- numeric(0)
			y <- numeric(0)
			id <- numeric(0)

			# figure out which part farthest
			a <- degrees(angle)
			dists <- sapply(l_indices, function(x) {
				painter_depth(mean(self$vertices[full_indices(x, n)]), scale = 1, alpha = a)
			})
			l_indices <- l_indices[order(dists)]
			for (i in seq_along(l_indices)) {
				indices <- full_indices(l_indices[[i]], n)
				x <- append(x, xy$x[indices])
				y <- append(y, xy$y[indices])
				id <- append(id, rep(i, length(indices)))
			}
			polygonGrob(x = x, y = y, id = id, default.units = "in", name = name)
		}
	),
	active = list()
)

npc_to_in <- function(xy, x = 0.5, y = 0.5, w = 1, h = 1, t = 0) {
	xy$translate(as_coord2d(-0.5, -0.5))$scale(w, h)$rotate(degrees(t))$translate(as_coord2d(x, y))
	invisible(xy)
}


# Name 'nigh' to avoid potential conflict with 'dplyr::near()'
nigh <- function(x, y, tolerance = 1e-6) {
	if (length(y) < length(x)) {
		y <- rep(y, length.out = length(x))
	}
	isTRUE(all.equal(x, y, tolerance = tolerance))
}

# robust version of `acos()` that returns in degrees
arccos <- function(x) {
	affiner::arccosine(x, "degrees")
}

# Axis-angle representation to rotation matrix
# https://en.wikipedia.org/wiki/Axis-angle_representation
# Because we do rotation matrix post-multiplication instead of pre-multiplication we usually need to multiply angles
# in following algorithms by -1

#' @rdname geometry_utils
#' @inheritParams save_piece_obj
#' @param axis_z Third coordinate of the axis unit vector (usually inferred).
#' @param angle Numeric vector in degrees (counter-clockwise) or an [affiner::angle()] vector.
#' @param ... Ignored
#' @export
AA_to_R <- function(angle = 0, axis_x = 0, axis_y = 0, axis_z = NA, ...) {
	if (is.na(axis_z)) {
		inner <- 1 - axis_x^2 - axis_y^2
		if (nigh(inner, 0)) {
			axis_z <- 0
		} else {
			axis_z <- sqrt(inner)
		}
	}
	axis <- as_coord3d(axis_x, axis_y, axis_z)
	affiner::rotate3d(axis, degrees(angle))[1:3, 1:3]
}

# Rotation matrix to Axis-angle representation
# https://en.wikipedia.org/wiki/Axis-angle_representation
#' @param R 3D rotation matrix (post-multiplied)
#' @rdname geometry_utils
#' @export
R_to_AA <- function(R = diag(3)) {
	l <- affiner::rotate3d_to_AA(R, unit = "degrees")
	list(angle = as.numeric(l$theta), axis_x = l$axis$x, axis_y = l$axis$y, axis_z = l$axis$z)
}

# Basic 3D rotation matrices
# https://en.wikipedia.org/wiki/Rotation_matrix#In_three_dimensions

#' Helper functions for making geometric calculations.
#'
#' `to_x()`, `to_y()`, `to_r()`, `to_t()` convert
#' between polar coordinates (in degrees) and Cartesian coordinates.
#' `to_degrees()` and `to_radians()` converts between degrees and radians.
#' `AA_to_R()` and `R_to_AA()` convert back and forth between (post-multiplied) rotation matrix
#' and axis-angle representations of 3D rotations.
#' `R_x()`, `R_y()`, and `R_z()` build (post-multiplied) rotation matrices for simple rotations around
#' the x, y, and z axes.
#'
#' [pp_cfg()] uses polar coordinates to determine where the "primary" and "directional"
#' symbols are located on a game piece.
#' They are also useful for drawing certain shapes and for making game diagrams on hex boards.
#'
#' `piecepackr` and `grid` functions use angles in degrees
#'  but the `base` trigonometry functions usually use radians.
#'
#' `piecepackr`'s 3D graphics functions [save_piece_obj()], [piece()], [piece3d()], and [piece_mesh()]
#' use the axis-angle representation for 3D rotations.
#' The axis-angle representation involves specifying a unit vector
#' indicating the direction of an axis of rotation and an angle describing the (counter-clockwise)
#' rotation around that axis.  Because it is a unit vector one only needs to specify the first two elements,
#' `axis_x` and `axis_y`, and we are able to infer the 3rd element `axis_z`.  The default of
#' `axis = 0`, `axis_y = 0`, and implied `axis_z = 1`
#' corresponds to a rotation around the z-axis which is reverse-compatible
#' with the originally 2D `angle` interpretation in [grid.piece()].  In order to figure out the appropriate
#' axis-angle representation parameters `R_to_AA()`, `R_x()`, `R_y()`, and `R_z()` allow one
#' to first come up with an appropriate (post-multiplied) 3D rotation matrix by chaining simple rotations
#' and then convert them to the corresponding axis-angle representation.
#' Pieces are rotated as if their center was at the origin.
#'
#' @examples
#'  to_x(90, 1)
#'  to_y(180, 0.5)
#'  to_t(0, -1)
#'  to_r(0.5, 0)
#'  all.equal(pi, to_radians(to_degrees(pi)))
#'  # default axis-angle axis is equivalent to a rotation about the z-axis
#'  all.equal(AA_to_R(angle=60), R_z(angle=60))
#'  # axis-angle representation of 90 rotation about the x-axis
#'  R_to_AA(R_x(90))
#'  # find Axis-Angle representation of first rotating about x-axis 180 degrees
#'  # and then rotating about z-axis 45 degrees
#'  R_to_AA(R_x(180) %*% R_z(45))
#'
#' @seealso \url{https://en.wikipedia.org/wiki/Axis-angle_representation} for more details
#'   about the Axis-angle representation of 3D rotations.
#'   See \code{\link[base]{Trig}} for R's built-in trigonometric functions.
#' @name geometry_utils
NULL

#' @rdname geometry_utils
#' @export
R_x <- function(angle = 0) {
	c <- affiner::cosine(degrees(-angle))
	s <- affiner::sine(degrees(-angle))
	matrix(c(1, 0, 0, 0, c, -s, 0, s, c), ncol = 3, byrow = TRUE)
}

#' @rdname geometry_utils
#' @export
R_y <- function(angle = 0) {
	c <- affiner::cosine(degrees(-angle))
	s <- affiner::sine(degrees(-angle))
	matrix(c(c, 0, s, 0, 1, 0, -s, 0, c), ncol = 3, byrow = TRUE)
}

#' @rdname geometry_utils
#' @export
R_z <- function(angle = 0) {
	c <- affiner::cosine(degrees(-angle))
	s <- affiner::sine(degrees(-angle))
	matrix(c(c, -s, 0, s, c, 0, 0, 0, 1), ncol = 3, byrow = TRUE)
}

#' @rdname geometry_utils
#' @export
to_radians <- function(t) as.numeric(affiner::radians(degrees(t)))

#' @rdname geometry_utils
#' @export
to_degrees <- function(t) as.numeric(degrees(affiner::radians(t)))

#' @rdname geometry_utils
#' @param t Numeric vector in degrees (or radians for `to_degrees()`) or an [affiner::angle()] vector.
#' @param r Radial distance
#' @export
to_x <- function(t, r) {
	r * affiner::cosine(degrees(t))
}

#' @rdname geometry_utils
#' @export
to_y <- function(t, r) {
	r * affiner::sine(degrees(t))
}

#' @rdname geometry_utils
#' @param x Cartesian x coordinate
#' @param y Cartesian y coordinate
#' @export
to_r <- function(x, y) {
	sqrt(x^2 + y^2)
}

#' @rdname geometry_utils
#' @export
to_t <- function(x, y) {
	180 * atan2(y, x) / pi
}

library("tibble")
test_that("3d helper functions work", {
	dft <- tibble(piece_side = "tile_back", x = 1.5, y = 1.5, rank = NA, width = NA)
	dfc <- tibble(
		piece_side = "coin_face",
		x = c(1, 2, 2, 1, 1.5),
		y = c(2, 2, 1, 1, 1.5),
		rank = 1:5,
		width = 0.6
	)
	df <- rbind(dft, dfc)
	ans <- c(0.25, rep(0.375, 5))
	# close circles
	expect_equal(op_transform(df)$zt, ans)
	# close regular convex polygons
	expect_equal(op_transform(df, cfg = list(shape.coin = "convex6"))$zt, ans)
	# close concave polygons (stars)
	expect_equal(op_transform(df, cfg = list(shape.coin = "concave5"))$zt, ans)
	# close "kites"
	expect_equal(op_transform(df, cfg = list(shape.coin = "kite"))$zt, ans)
	# close "pyramids"
	expect_equal(op_transform(df, cfg = list(shape.coin = "pyramid"))$zt, ans)
	expect_warning(pp_cfg(list(shadow_fn = basicTokenEdge)))

	skip_on_ci()
	skip_if_not_installed("vdiffr")
	library("vdiffr")
	opf <- function(op_angle = 45, ...) {
		function() {
			pmap_piece(
				df,
				op_angle = op_angle,
				trans = op_transform,
				op_scale = 0.5,
				default.units = "in",
				...
			)
		}
	}

	# orthogonal tiles
	df <- tibble(
		piece_side = "tile_back",
		x = c(2, 2, 2, 4, 6, 6, 4, 2, 5),
		y = c(4, 4, 4, 4, 4, 2, 2, 2, 3)
	)
	expect_doppelganger("op-transform-045", opf(045))
	expect_doppelganger("op-transform-135", opf(135))
	expect_doppelganger("op-transform-225", opf(225))
	expect_doppelganger("op-transform-315", opf(315))

	# angled rectangles
	df1 <- tibble(
		piece_side = "matchstick_face",
		x = 2 + to_x(45, 0.19 * 1:4),
		y = 2 + to_y(45, 0.19 * 1:4),
		angle = 45,
		rank = 4,
		suit = 1
	)
	df2 <- tibble(
		piece_side = "matchstick_face",
		x = 2 + to_x(-45, 0.19 * 1:4),
		y = 2 + to_y(-45, 0.19 * 1:4),
		angle = -45,
		rank = 4,
		suit = 2
	)
	df <- rbind(df1, df2, df1, df2, df1, df2)
	expect_doppelganger("matchsticks-045", opf(045))

	# rounding error #163
	df <- tibble(
		piece_side = "tile_face",
		x = rep(seq(1, 7, 2), 4),
		y = rep(seq(1, 7, 2), each = 4),
		angle = rep(90 * 0:3, 4),
		suit = rep(1:4, each = 4),
		rank = rep_len(1:6, 16)
	)
	expect_doppelganger("rotated-tile-faces", opf(045))

	# stacked pyramids
	suppressMessages(
		{
			df <- tibble(piece_side = "pyramid_top", x = 2, y = 2, rank = 1:6, suit = c(1:4, 1:2))
			expect_doppelganger("pyramid-tops-larger-on-top", opf(045))
		},
		classes = "piecepackr_affine_transformation"
	)

	skip_if_not_installed("svglite", "2.2.2")
	skip_if_not(getRversion() >= "4.2.0")
	write_svg_bleeding_edge <- function(plot, file, title = "") {
		svglite::svglite(file)
		plot()
		invisible(grDevices::dev.off())
	}
	df <- tibble(piece_side = "pyramid_top", x = 2, y = 2, rank = 6:1, suit = c(1:4, 1:2))
	expect_doppelganger("pyramid-tops-smaller-on-top", opf(045), writer = write_svg_bleeding_edge)

	cfg <- list(
		background_color.pyramid_face = "blue",
		background_color.pyramid_back = "red",
		background_color.pyramid_left = "green",
		background_color.pyramid_right = "grey"
	)
	dft <- tibble(
		piece_side = "tile_back",
		x = rep(c(1, 3, 5), 3),
		y = rep(c(1, 3, 5), each = 3),
		rank = NA,
		suit = NA,
		angle = 0
	)
	dfp1 <- tibble(piece_side = "pyramid_top", x = 3, y = 3, rank = 3, suit = 2, angle = 0)
	dfp2 <- tibble(
		piece_side = rep(
			c("pyramid_face", "pyramid_back", "pyramid_left", "pyramid_right"),
			length.out = 9
		),
		x = rep(c(1, 3, 5), 3),
		y = rep(c(5, 3, 1), each = 3),
		angle = c(45, 0, -45, 90, 0, -90, 135, 180, -135),
		rank = rep(1:6, length.out = 9),
		suit = 1
	)[-5, ]
	df <- rbind(dft, dfp1, dfp2)
	expect_doppelganger("oblique-pyramids", opf(045, cfg = cfg), writer = write_svg_bleeding_edge)

	# pawns far from origin with op_angle=135 (#376)
	df <- tibble(
		piece_side = "pawn_top",
		x = c(4, 6, 7, 8),
		y = 2,
		suit = c(2, 1, 3, 4),
		z = c(1.4375, 2.1875, 2.6875, 2.6875)
	)
	opf_pawn <- function(op_angle = 135) {
		function() {
			pmap_piece(
				df,
				op_scale = 1,
				op_angle = op_angle,
				default.units = "in"
			)
		}
	}
	expect_doppelganger(
		"pawns-far-from-origin-135",
		opf_pawn(135),
		writer = write_svg_bleeding_edge
	)
})

test_that("SAT functions work", {
	r1 <- as_polygon2d(as_coord2d(x = c(0, 0, 1, 1), y = c(0, 1, 1, 0)), convex = TRUE)
	r2 <- as_polygon2d(as_coord2d(x = 0.5 + c(0, 0, 1, 1), y = 0.5 + c(0, 1, 1, 0)), convex = TRUE)
	r3 <- as_polygon2d(as_coord2d(x = 1 + c(0, 0, 1, 1), y = 1 + c(0, 1, 1, 0)), convex = TRUE)
	r4 <- as_polygon2d(
		as_coord2d(x = c(0.25, 0.25, 0.75, 0.75), y = c(0.25, 0.75, 0.75, 0.25)),
		convex = TRUE
	)

	c1 <- as_ellipse2d(as_coord2d(0.5, 0.5), r = 0.5)
	c2 <- as_ellipse2d(as_coord2d(1.0, 1.0), r = 0.5)
	c3 <- as_ellipse2d(as_coord2d(2.0, 2.0), r = 0.5)

	expect_true(has_overlap2d(r1, r2))
	expect_true(has_overlap2d(r2, r3))
	expect_false(has_overlap2d(r1, r3))
	expect_true(has_overlap2d(r1, r4))

	expect_true(has_overlap2d(c1, c2))
	expect_false(has_overlap2d(c2, c3))
	expect_false(has_overlap2d(c1, c3))

	expect_true(has_overlap2d(r1, c1))
	expect_false(has_overlap2d(r1, c3))
	expect_true(has_overlap2d(c1, r1))
	expect_false(has_overlap2d(c3, r1))
})

test_that("3D rotation functions work", {
	expect_equal(AA_to_R(angle = 60), R_z(angle = 60))
	R1 <- R_z(60) %*% R_x(50)
	R2 <- do.call(AA_to_R, R_to_AA(R1))
	expect_equal(R1, R2)
	expect_equal(AA_to_R(angle = 60), R_z(angle = 60))
	expect_equal(AA_to_R(angle = 60), R_z(angle = 60))
	expect_equal(AA_to_R(angle = 45, axis_x = 1), R_x(angle = 45))
	expect_equal(AA_to_R(angle = 30, axis_y = 1), R_y(angle = 30))
	expect_equal(R_to_AA(R_x(90)), list(angle = 90, axis_x = 1, axis_y = 0, axis_z = 0))
	expect_equal(R_to_AA(R_y(90)), list(angle = 90, axis_x = 0, axis_y = 1, axis_z = 0))
	expect_equal(
		R_to_AA(AA_to_R(180, 0.5, 0.5)),
		list(angle = 180, axis_x = 0.5, axis_y = 0.5, axis_z = 0.7071068),
		tolerance = 1e-6
	)
	expect_equal(
		R_to_AA(AA_to_R(180, -0.6, 0.6, 0.8)),
		list(angle = 180, axis_x = -0.5144958, axis_y = 0.5144958, axis_z = 0.6859943),
		tolerance = 1e-6
	)
	expect_equal(
		R_to_AA(AA_to_R(180, 0.6, -0.6)),
		list(angle = 180, axis_x = 0.6, axis_y = -0.6, axis_z = 0.5291503),
		tolerance = 1e-6
	)
	expect_equal(
		R_to_AA(AA_to_R(180, -0.6, -0.6)),
		list(angle = 180, axis_x = -0.6, axis_y = -0.6, axis_z = 0.5291503),
		tolerance = 1e-6
	)
	expect_equal(
		R_to_AA(AA_to_R(0, -0.6, 0.6)),
		list(angle = 0, axis_x = 0, axis_y = 0, axis_z = 1),
		tolerance = 1e-6
	)
})

expect_equal(to_x(90, 1), 0)
expect_equal(to_y(90, 1), 1)
expect_equal(to_t(0, 1), 90)
expect_equal(to_r(0, 1), 1)

f <- function(x) assert_suggested("imaginaryPackage")
expect_error(f(), "You need to install the suggested package")

skip_on_cran()
skip_if_not_installed("pdftools")
skip_if(!capabilities("cairo"))
skip_if(is_cairo_maybe_buggy())

df <- get_embedded_font("sans", c("♥", "♠", "♣", "♦", "🌞", "🌜", "꩜"))
expect_false(has_font("thisshouldnotexist"))

print_and_play_paper_compact <- function(cfg, size, pieces, arrangement, quietly, size_bleed) {
	n_suits <- cfg$n_suits
	n_ranks <- cfg$n_ranks

	stopifnot(n_ranks <= 6)

	if (size == "letter") {
		xl <- inch(LETTER_HEIGHT / 2 - A5W / 2 + size_bleed$left)
		xr <- inch(LETTER_HEIGHT / 2 + A5W / 2 + size_bleed$left)
		y <- inch(LETTER_WIDTH / 2 + size_bleed$bottom)
	} else if (size == "A4") {
		xl <- inch(A4_HEIGHT / 2 - A5W / 2 + size_bleed$left)
		xr <- inch(A4_HEIGHT / 2 + A5W / 2 + size_bleed$left)
		y <- inch(A4_WIDTH / 2 + size_bleed$bottom)
	} else {
		# size is "A5"
		xl <- inch(A5W / 2 + size_bleed$left)
		xr <- inch(A5W / 2 + size_bleed$left)
		y <- inch(A5H / 2 + size_bleed$bottom)
	}

	vpl <- viewport(x = xl, y = y, width = inch(A5W), height = inch(A5H))
	vpr <- viewport(x = xr, y = y, width = inch(A5W), height = inch(A5H))
	gl <- list()
	pl <- list()

	if ("piecepack" %in% pieces || "subpack" %in% pieces) {
		psuits <- seq(cfg$n_suits + 2L)
		lpgf <- lapply(
			psuits,
			a5_piecepack_grob,
			cfg = cfg,
			front = TRUE,
			arrangement = arrangement
		)
		lpgb <- lapply(
			psuits,
			a5_piecepack_grob,
			cfg = cfg,
			front = FALSE,
			arrangement = arrangement
		)
	}

	## Front Matter
	gl <- gappend(gl, a5_title_grob(cfg, pieces, quietly))
	gl <- gappend(gl, a5_inst_grob(cfg, pieces))
	if (arrangement == "double-sided" && size != "A5") {
		gl <- gappend(gl, blank_grob)
		gl <- gappend(gl, blank_grob)
		pl[["Front Matter"]] <- 2
	} else {
		pl[["Front Matter"]] <- 1
	}

	if ("piecepack" %in% pieces) {
		n_pages <- n_suits
		if (is_odd(n_suits) && arrangement == "double-sided" && size != "A5") {
			n_pages <- n_pages + 1
		}
		for (suit in seq(n_pages)) {
			gl <- gappend(gl, lpgf[[suit]])
			gl <- gappend(gl, lpgb[[suit]])
		}
		pl$Piecepack <- n_pages
	}

	#### Fine-tune between pyramids, matchsticks, and misc.?
	#### Add misc, cards, dominoes accessories?
	if ("matchsticks" %in% pieces) {
		n_pages <- (n_suits - 1) %/% 4 + 1
		for (ii in seq(n_pages)) {
			ss <- seq(4 * ii - 3, 4 * ii)
			if (arrangement == "double-sided" && size == "A5") {
				gl <- gappend(gl, a5_matchsticks_grob(ss, cfg, TRUE))
				gl <- gappend(gl, a5_matchsticks_grob(ss, cfg, FALSE))
				gl <- gappend(gl, a5_matchsticks_grob2(ss, cfg, TRUE))
				gl <- gappend(gl, a5_matchsticks_grob2(ss, cfg, FALSE))
			} else {
				gl <- gappend(gl, a5_matchsticks_grob(ss, cfg, TRUE))
				gl <- gappend(gl, a5_matchsticks_grob2(ss, cfg, TRUE))
				if (arrangement == "double-sided") {
					gl <- gappend(gl, a5_matchsticks_grob2(ss, cfg, FALSE))
					gl <- gappend(gl, a5_matchsticks_grob(ss, cfg, FALSE))
				}
			}
		}
		pl$Matchsticks <- n_pages
	}
	if ("pyramids" %in% pieces) {
		n_pages <- (n_suits - 1) %/% 4 + 1
		for (ii in seq(n_pages)) {
			ss <- seq(4 * ii - 3, 4 * ii)
			s1 <- ss[1:2]
			s2 <- ss[3:4]
			if (arrangement == "double-sided" && size == "A5") {
				gl <- gappend(gl, a5_pyramids_grob(s1, cfg, TRUE))
				gl <- gappend(gl, blank_grob)
				gl <- gappend(gl, a5_pyramids_grob(s2, cfg, FALSE))
				gl <- gappend(gl, blank_grob)
			} else {
				gl <- gappend(gl, a5_pyramids_grob(s1, cfg, TRUE))
				gl <- gappend(gl, a5_pyramids_grob(s2, cfg, FALSE))
				if (arrangement == "double-sided") {
					gl <- gappend(gl, blank_grob)
					gl <- gappend(gl, blank_grob)
				}
			}
		}
		pl$Pyramids <- n_pages
	}
	if ("subpack" %in% pieces) {
		vpul <- viewport(x = 0.25, y = 0.75, width = inch(A5W / 2), height = inch(A5H / 2))
		vpur <- viewport(x = 0.75, y = 0.75, width = inch(A5W / 2), height = inch(A5H / 2))
		vpll <- viewport(x = 0.25, y = 0.25, width = inch(A5W / 2), height = inch(A5H / 2))
		vplr <- viewport(x = 0.75, y = 0.25, width = inch(A5W / 2), height = inch(A5H / 2))
		lsgf <- lapply(lpgf, as_picture, width = A5W, height = A5H)
		lsgb <- lapply(lpgb, as_picture, width = A5W, height = A5H)
		n_pages <- (n_suits - 1) %/% 4 + 1
		if (arrangement == "double-sided" && is_odd(n_pages) && size != "A5") {
			n_pages <- n_pages + 1
		}
		for (ii in seq(n_pages)) {
			ss <- seq(4 * ii - 3, 4 * ii)
			ss[which(ss > cfg$n_suits + 2L)] <- cfg$n_suits + 2L
			gf <- grobTree(
				grobTree(lsgf[[ss[1]]], vp = vpul),
				grobTree(lsgb[[ss[1]]], vp = vpur),
				grobTree(lsgf[[ss[2]]], vp = vpll),
				grobTree(lsgb[[ss[2]]], vp = vplr)
			)
			gb <- grobTree(
				grobTree(lsgf[[ss[3]]], vp = vpul),
				grobTree(lsgb[[ss[3]]], vp = vpur),
				grobTree(lsgf[[ss[4]]], vp = vpll),
				grobTree(lsgb[[ss[4]]], vp = vplr)
			)
			gl <- gappend(gl, gf)
			gl <- gappend(gl, gb)
		}
		pl$Subpack <- n_pages
	}
	if (size == "A5") {
		pl <- lapply(pl, function(x) 2 * x)
	}
	for (ii in seq(gl)) {
		if (is_odd(ii)) {
			grid.newpage()
			draw_a5_page(gl[[ii]], vpl)
		} else {
			if (size == "A5") {
				grid.newpage()
			}
			draw_a5_page(gl[[ii]], vpr)
		}
	}

	pl
}

a5_inst_grob <- function(cfg = pp_cfg(), pieces) {
	grob_title <- textGrob(
		"Piece Anatomy",
		y = 0.97,
		just = "center",
		gp = gp_title,
		name = "title"
	)
	# Tile
	yt <- A5H - 1.3
	xt <- 1.5
	df <- tibble(
		piece_side = c("tile_face", "tile_back"),
		suit = 1,
		x = xt + c(-0.5, 0.5),
		y = yt,
		width = 1,
		height = 1
	)
	grob_tile <- grobTree(
		htg("Tile", xt, yt + 0.7),
		tg("Face", xt - 0.5, y = yt + 0.6),
		tg("Back", xt + 0.5, y = yt + 0.6),
		pmap_piece(
			df,
			default.units = "in",
			draw = FALSE,
			cfg = cfg,
			gp = gpar(cex = 0.5, lex = 0.5)
		)
	)
	# curveGrob(x1=xt[1]-0.8, y1=yt-1.1, x2=xt[2]+0.8, y2=yt-1.1,
	#           curvature=0.2, square=FALSE, arrow=arrow(),
	#           default.units="in"))
	# Coin
	cwr <- 0.75 / cfg$get_width("coin_face")
	cw <- 0.75
	yc <- yt - 0.5 - cw
	xc <- 0.5 + 2 * cw
	df <- tibble(
		piece_side = rep(c("coin_back", "coin_face"), 2),
		x = xc + c(1.7, 0.7, -0.7, -1.7) * cw,
		y = rep(yc, 4),
		width = 0.75,
		height = 0.75,
		angle = c(0, 180, 0, 0)
	)
	grob_coin <- grobTree(
		htg("Coin", xc, yc + 0.5 * cw + 0.2),
		tg("Back", xc + c(1.7, -0.7) * cw, yc + 0.5 * cw + 0.1),
		tg("Face", xc + c(0.7, -1.7) * cw, yc + 0.5 * cw + 0.1),
		pmap_piece(
			df,
			default.units = "in",
			cfg = cfg,
			draw = FALSE,
			gp = gpar(lex = cwr, cex = cwr)
		),
		textGrob("or", x = xc, y = yc, default.units = "in")
	)
	# Saucer
	swr <- 0.75 / cfg$get_width("saucer_face")
	sw <- 0.75
	ys <- yt - 1.6 - sw
	xs <- 0.5 + 2 * cw
	df <- tibble(
		piece_side = rep(c("saucer_back", "saucer_face"), 2),
		x = xs + c(1.7, 0.7, -0.7, -1.7) * sw,
		y = rep(ys, 4),
		width = 0.75,
		height = 0.75,
		angle = c(0, 180, 0, 0)
	)
	grob_saucer <- grobTree(
		htg("(Pawn) Saucer", xs, ys + 0.5 * sw + 0.25),
		tg("Back", xs + c(1.7, -0.7) * sw, ys + 0.5 * sw + 0.1),
		tg("Face", xs + c(0.7, -1.7) * sw, ys + 0.5 * sw + 0.1),
		pmap_piece(
			df,
			default.units = "in",
			cfg = cfg,
			draw = FALSE,
			gp = gpar(lex = swr, cex = swr)
		),
		textGrob("or", x = xs, y = ys, default.units = "in")
	)

	# Dice
	dwr <- 0.5 / cfg$get_width("die_face")
	dw <- 0.5
	yd <- 1.4
	d1 <- grobTree(
		piecepackDieGrob(1, cfg, arrangement = "counter_down"),
		vp = viewport(
			x = 0.2,
			y = inch(yd),
			width = inch(4 * dw),
			height = inch(3 * dw),
			gp = gpar(lex = dwr, cex = dwr)
		)
	)
	d2 <- grobTree(
		piecepackDieGrob(1, cfg, arrangement = "counter_up"),
		vp = viewport(
			x = 0.5,
			y = inch(yd),
			width = inch(4 * dw),
			height = inch(3 * dw),
			gp = gpar(lex = dwr, cex = dwr)
		)
	)
	d3 <- grobTree(
		piecepackDieGrob(1, cfg, arrangement = "opposites_sum_to_5"),
		vp = viewport(
			x = 0.8,
			y = inch(yd),
			width = inch(4 * dw),
			height = inch(3 * dw),
			gp = gpar(lex = dwr, cex = dwr)
		)
	)
	grob_die <- grobTree(
		htg("Die", A5W / 2, yd + 1.2),
		tg('"counter_down"\narrangement', 0.3 * A5W, yd + 0.9),
		tg('"counter_up"\narrangement', 0.6 * A5W, yd + 0.9),
		tg('"opposites_sum_to_5"\narrangement', 0.85 * A5W, yd + 0.9),
		d1,
		d2,
		d3
	)
	# Pawn
	y_pawn <- 4.0
	grob_pawn <- grobTree(
		htg("Pawn", A5W - 0.5, y_pawn + 1.4),
		tg("Back", A5W - 0.85, y_pawn + 0.5, rot = 90),
		tg("Face", A5W - 0.85, y_pawn - 0.5, rot = 90),
		tg("Stand", A5W - 0.85, y_pawn - 1.1, rot = 90),
		tg("Stand", A5W - 0.85, y_pawn + 1.0, rot = 90),
		pieceGrob("pawn_layout", 1, cfg = cfg, x = inch(A5W - 0.5), y = inch(y_pawn))
	)
	# Belt
	yb <- yt - 2.0 - sw - 0.5
	xb <- 2
	grob_belt <- grobTree(
		htg("(Pawn) Belt (Face)", xb, yb + 0.4),
		pieceGrob("belt_face", x = xb, y = yb, cfg = cfg, default.units = "in")
	)
	# Pyramids
	yy <- yt + 0.10
	xy <- A5W - 1.2
	df <- tibble(
		piece_side = c("pyramid_layout", "pyramid_top"),
		suit = 1,
		rank = 1,
		x = xy + c(-0.5, 0.6),
		y = yy
	)
	grob_pyramid <- grobTree(
		htg("Pyramid", xy, yy + 0.8),
		htg("Top (View)", xy + 0.7, yy + 0.50),
		tg("Face", xy + 0.6, yy + 0.33),
		tg("Back", xy + 0.6, yy - 0.33),
		tg("Left", xy + 0.25, yy, rot = 90),
		tg("Right", xy + 0.95, yy, rot = -90),
		tg("Back", xy - 0.0, yy, rot = -90),
		tg("Face", xy - 1.0, yy + 0.60, rot = 90),
		tg("Face", xy - 1.0, yy - 0.60, rot = 90),
		tg("Right", xy - 0.05, yy + 0.45),
		tg("Left", xy - 0.05, yy - 0.45),
		pmap_piece(df, default.units = "in", cfg = cfg, draw = FALSE)
	)
	# Matchsticks
	df <- tibble(
		piece_side = c("matchstick_face", "matchstick_back"),
		x = 2.5,
		y = c(0.4, 0.2),
		rank = 3,
		angle = 90
	)
	grob_matchsticks <- grobTree(
		htg("Matchstick", 1, 0.3),
		tg("Face", 1.7, 0.4),
		tg("Back", 1.7, 0.2),
		pmap_piece(df, default.units = "in", cfg = cfg, draw = FALSE)
	)

	grobTree(
		grob_title,
		grob_tile,
		grob_coin,
		grob_saucer,
		grob_die,
		grob_pawn,
		grob_belt,
		grob_pyramid,
		grob_matchsticks,
		name = "instructions",
		vp = a5_vp()
	)
}

a5_matchsticks_grob <- function(suits = 1:4, cfg = pp_cfg(), front = TRUE) {
	n_suits <- length(suits)
	y1t <- A5H - 0.5 * MATCHSTICK_HEIGHTS[1]
	y1b <- A5H - 1.5 * MATCHSTICK_HEIGHTS[1]
	x1s <- (0.5 + 0:(3 * n_suits - 1)) * MATCHSTICK_WIDTHS[1]
	xs <- (0.5 + 0:(6 * n_suits - 1)) * MATCHSTICK_WIDTHS[2]
	y2 <- y1b - 0.5 * MATCHSTICK_HEIGHTS[1] - 0.5 * MATCHSTICK_HEIGHTS[2]
	y3 <- y2 - 0.5 * MATCHSTICK_HEIGHTS[2] - 0.5 * MATCHSTICK_HEIGHTS[3]
	y5 <- y3 - 0.5 * MATCHSTICK_HEIGHTS[3] - 0.5 * MATCHSTICK_HEIGHTS[5]
	y6 <- y5 - 0.5 * MATCHSTICK_HEIGHTS[5] - 0.5 * MATCHSTICK_HEIGHTS[6]

	x <- c(rep(x1s, each = 2), rep(xs, 4))
	y <- c(rep(c(y1t, y1b), 3 * n_suits), rep(c(y2, y3, y5, y6), each = 6 * n_suits))
	suit <- rep(rep(suits, each = 6), 5)
	rank <- rep(c(1:3, 5:6), each = 6 * n_suits)
	if (front) {
		piece_side <- "matchstick_face"
	} else {
		x <- A5W - x
		piece_side <- "matchstick_back"
	}
	df <- tibble(piece_side, x, y, suit, rank)

	pmap_piece(df, cfg = cfg, default.units = "inches", draw = FALSE)
}

a5_matchsticks_grob2 <- function(suits = 1:4, cfg = pp_cfg(), front = TRUE) {
	n_suits <- length(suits)
	y4s <- A5H - (0.5 + 0:3) * MATCHSTICK_HEIGHTS[4]
	x4s <- (0.5 + 0:5) * MATCHSTICK_WIDTHS[6]

	x <- rep(x4s, 4)
	y <- rep(c(y4s), each = 6)
	suit <- rep(suits, each = 6)
	rank <- 4
	if (front) {
		piece_side <- "matchstick_face"
	} else {
		x <- A5W - x
		piece_side <- "matchstick_back"
	}
	df <- tibble(piece_side, x, y, suit, rank)

	# add some extra dice if there is some space
	if (cfg$get_width("die_face") <= 0.6) {
		xd1 <- max(x4s) + 0.0 + c(0.6) * cfg$get_width("die_layoutLF")
		dfd <- tibble::tibble(
			piece_side = "die_layoutLF",
			x = xd1,
			y = unique(y),
			suit = suits,
			rank = NA
		)
		df <- rbind(df, dfd)
	}
	if (cfg$get_width("die_face") <= 0.55) {
		xd2 <- xd1 + c(0.7) * cfg$get_width("die_layoutLF")
		dfd <- tibble::tibble(
			piece_side = "die_layoutLF",
			x = xd2,
			y = unique(y),
			suit = suits,
			rank = NA
		)
		df <- rbind(df, dfd)
	}

	pmap_piece(df, cfg = cfg, default.units = "inches", draw = FALSE)
}

a5_piecepack_grob <- function(suit, cfg = pp_cfg(), front = TRUE, arrangement = "single-sided") {
	die_width <- cfg$get_width("die_width")
	pawn_width <- cfg$get_width("pawn_width")
	tile_width <- cfg$get_width("tile_back")
	xtl <- 1.5 * tile_width
	xtr <- 0.5 * tile_width
	ytb <- 0.5 * tile_width
	ytm <- 1.5 * tile_width
	ytt <- 2.5 * tile_width
	xc <- A5W - 0.25 * tile_width
	ycs <- rev((0.50 + seq(0, 5)) * 0.8)
	xdr <- c(0.5, 1.5, 2.5) * die_width
	ydt <- A5H - 0.5 * die_width
	ydb <- A5H - 1.5 * die_width
	xp <- A5W - 0.5 * cfg$get_height("pawn_layout")
	yp <- A5H - 0.5 * pawn_width
	xb <- A5W - 0.5 * cfg$get_width("belt_face")
	yb <- A5H - pawn_width - 0.5 * cfg$get_height("belt_face") - 0.25
	xsr <- A5W - 0.25 * tile_width
	ysb <- 2.75 * tile_width
	dft <- tibble(
		piece_side = "tile_face",
		x = rep(c(xtr, xtl), 3),
		y = rep(c(ytt, ytm, ytb), each = 2),
		suit,
		rank = 1:6,
		angle = 0
	)
	dfc <- tibble(
		piece_side = "coin_back",
		x = rep(xc, 6),
		y = ycs,
		suit,
		rank = 1:6,
		angle = ifelse(front, cfg$coin_arrangement, 0)
	)
	dfd <- tibble(
		piece_side = "die_face",
		x = rep(xdr, 2),
		y = rep(c(ydt, ydb), each = 3),
		suit,
		rank = 1:6,
		angle = 0
	)
	dfp <- tibble(piece_side = "pawn_layout", x = xp, y = yp, suit, rank = NA, angle = 90)
	dfb <- tibble(piece_side = "belt_face", x = xb, y = yb, suit, rank = NA, angle = 0)
	dfs <- tibble(
		piece_side = "saucer_face",
		x = xsr,
		y = ysb,
		suit,
		rank = NA,
		angle = ifelse(front, 0, cfg$coin_arrangement)
	)
	df <- rbind(dfc, dfd, dfp, dfb, dfs, dft)
	if (!front) {
		df$x <- A5W - df$x
		df$piece_side <- c(
			rep("coin_face", 6),
			rep("die_face", 6),
			"pawn_layout",
			"belt_face",
			"saucer_back",
			rep("tile_back", 6)
		)
		# Rotate tile backs to partially hide direction of face if tile back are not fully symmetric
		df$angle <- c(rep(0, 12), 270, 0, 0, 90 * ((df$suit[16:21] + df$rank[16:21]) %% 4))

		if (arrangement == "double-sided") {
			df <- df[df$piece_side %in% c("tile_back", "coin_face", "saucer_back"), ]
		}
	}
	pmap_piece(df, cfg = cfg, default.units = "inches", draw = FALSE)
}

a5_pyramids_grob <- function(suit = 1:2, cfg = pp_cfg(), front = TRUE) {
	grobTree(pyramid_grob_helper(suit[1], cfg, 0), pyramid_grob_helper(suit[2], cfg, A5W / 2))
}

pyramid_grob_helper <- function(suit, cfg = pp_cfg(), xleft = 0) {
	rank <- c(2, 4, 5, 1, 3, 6)
	plh <- PYRAMID_LAYOUT_HEIGHTS[rank]
	y_up <- cumsum(plh) - 0.5 * plh
	x_up <- 0.5 * PYRAMID_LAYOUT_WIDTHS[rank]
	x <- xleft + c(x_up[1:3], A5W / 2 - x_up[4:6])
	y <- c(y_up[1:3], c(plh[5] + plh[6] + 0.5 * plh[4], plh[6] + 0.5 * plh[5], 0.5 * plh[6]))
	piece_side <- "pyramid_layout"
	angle <- c(rep(0, 3), rep(180, 3))
	df <- tibble::tibble(piece_side, x, y, suit, rank, angle)
	# space for a pawn layout on top
	dfp <- tibble::tibble(
		piece_side = "pawn_layout",
		x = xleft + 0.25 * A5W,
		y = A5H - 0.5 * cfg$get_width("pawn_layout"),
		suit,
		rank = 1,
		angle = 90
	)
	df <- rbind(dfp, df)
	pmap_piece(df, cfg = cfg, default.units = "inches", draw = FALSE)
}

#' Render board game pieces with rgl
#'
#' `piece3d()` draws board games pieces using the rgl package.
#' @inheritParams grid.piece
#' @inheritParams save_piece_obj
#' @param lit logical, specifying if rgl lighting calculation should take place.
#' @param shininess Properties for rgl lighting calculation.
#' @param textype Use `"rgba"` when sure texture will have alpha transparency.
#'                Use `"rgb"` when sure texture will not have alpha transparency
#'                (in particular `rgl`'s WebGL export will likely work better).
#'                If `NA` we will read the texture and figure out a reasonable value.
#' @return A numeric vector of rgl object IDs.
#' @examples
#' if (requireNamespace("rgl", quietly = TRUE) && all(capabilities(c("cairo", "png")))) {
#'   rgl::open3d()
#'   cfg <- game_systems("sans", border = FALSE)$piecepack
#'   piece3d("tile_back", suit = 3, rank = 3, cfg = cfg, x = 0, y = 0, z = 0)
#'   piece3d("coin_back", suit = 4, rank = 2, cfg = cfg, x = 0.5, y = 0.5, z = 0.25)
#'   piece3d("pawn_top", suit = 1, cfg = cfg, x = -0.5, y = 0.5, z = 0.6)
#'   piece3d("die_face", suit = 3, cfg = cfg, x = -0.5, y = -0.5, z = 0.375)
#'   piece3d("pyramid_top", suit = 2, rank = 3, cfg = cfg, x = 1.5, y = 0.0, z = 0.31875)
#'   invisible(NULL)
#' }
#' @export
#' @seealso See \code{\link[rgl]{rgl-package}} for more information about the `rgl` package.
#'          See [rgl::material3d()] for more info about setting `rgl` material properties.
#'          See \code{\link{geometry_utils}} for a discussion of the 3D rotation parameterization.
piece3d <- function(
	piece_side = "tile_back",
	suit = NA,
	rank = NA, # nolint
	cfg = getOption("piecepackr.cfg", pp_cfg()),
	x = 0,
	y = 0,
	z = NA,
	angle = 0,
	axis_x = 0,
	axis_y = 0,
	width = NA,
	height = NA,
	depth = NA,
	envir = getOption("piecepackr.envir"),
	...,
	scale = 1,
	res = 72,
	alpha = 1.0,
	lit = FALSE,
	shininess = 50.0,
	textype = NA
) {
	assert_suggested("rgl")
	if (is_angle(angle)) {
		angle <- as.double(angle, "degrees")
	}

	nn <- max(lengths(list(
		piece_side,
		suit,
		rank,
		x,
		y,
		z,
		angle,
		axis_x,
		axis_y,
		width,
		height,
		depth
	)))
	piece_side <- rep(piece_side, length.out = nn)
	suit <- rep(suit, length.out = nn)
	rank <- rep(rank, length.out = nn)
	x <- rep(x, length.out = nn)
	y <- rep(y, length.out = nn)
	z <- rep(z, length.out = nn)
	angle <- rep(angle, length.out = nn)
	axis_x <- rep(axis_x, length.out = nn)
	axis_y <- rep(axis_y, length.out = nn)
	width <- rep(width, length.out = nn)
	height <- rep(height, length.out = nn)
	depth <- rep(depth, length.out = nn)

	scale <- rep(scale, length.out = nn)
	alpha <- rep(alpha, length.out = nn)
	lit <- rep(lit, length.out = nn)
	shininess <- rep(shininess, length.out = nn)
	textype <- rep(textype, length.out = nn)

	cfg <- get_cfg(cfg, envir)
	cfg <- rep(c(cfg), length.out = nn)

	l <- lapply(seq(nn), function(i) {
		cfg[[i]]$rgl(
			piece_side[i],
			suit[i],
			rank[i],
			x[i],
			y[i],
			z[i],
			angle[i],
			axis_x[i],
			axis_y[i],
			width[i],
			height[i],
			depth[i],
			scale = scale[i],
			res = res,
			alpha = alpha[i],
			lit = lit[i],
			shininess = shininess[i],
			textype = textype[i]
		)
	})
	unlist(l)
}

rgl_piece_helper <- function(
	piece_side = "tile_back",
	suit = NA,
	rank = NA,
	cfg = pp_cfg(), # nolint
	x = 0,
	y = 0,
	z = NA,
	angle = 0,
	axis_x = 0,
	axis_y = 0,
	width = NA,
	height = NA,
	depth = NA,
	scale = 1,
	res = 72,
	alpha = 1,
	lit = FALSE,
	shininess = 50.0,
	textype = NA
) {
	if (scale == 0 || alpha == 0) {
		return(invisible(numeric(0)))
	}
	l_obj <- save_piece_obj(
		piece_side,
		suit,
		rank,
		cfg,
		x = x,
		y = y,
		z = z,
		angle = angle,
		axis_x = axis_x,
		axis_y = axis_y,
		width = width,
		height = height,
		depth = depth,
		scale = scale,
		res = res
	)

	l <- purrr::pmap(
		l_obj,
		rgl_piece_helper_obj,
		alpha = alpha,
		lit = lit,
		shininess = shininess,
		textype = textype
	)
	unlist(l)
}

rgl_piece_helper_obj <- function(
	obj,
	mtl,
	png,
	alpha = 1,
	lit = FALSE,
	shininess = 50.0,
	textype = NA
) {
	if (is.na(textype)) {
		r <- png::readPNG(png)
		textype <- ifelse(dim(r)[3] > 3 && mean(r[,, 4]) < 0.999, "rgba", "rgb")
	}
	material <- list(
		color = "white",
		alpha = alpha,
		lit = lit,
		shininess = shininess,
		front = "filled",
		back = "filled",
		texture = png,
		textype = textype
	)
	if (requireNamespace("readobj", quietly = TRUE)) {
		mesh <- readobj::read.obj(obj, convert.rgl = TRUE)
	} else {
		mesh <- suppressWarnings(rgl::readOBJ(obj))
	}
	id <- rgl::shade3d(mesh, material = material)
	invisible(as.numeric(id))
}

# Simplify 'grobCoords' methods

coords_to_xylist <- function(coords) {
	if (inherits(coords, "GridGTreeCoords")) {
		xyl <- lapply(coords, coords_to_xylist)
		f <- function(x, y) gridGeometry::polyclip(x, y, "union")
	} else {
		xyl <- lapply(coords, identity)
		# Use "xor" to preserve holes (e.g. marbles board) #368
		# For non-overlapping shapes xor == union
		f <- function(x, y) gridGeometry::polyclip(x, y, "xor")
	}
	xyl <- Filter(function(x) length(x) > 0, xyl)
	Reduce(f, xyl)
}

xylists_to_grobcoords <- function(xyl, name, closed = TRUE) {
	if (getRversion() >= '4.2.0' && !closed) {
		return(emptyGrobCoords(name))
	}

	if (getRversion() < '4.2.0') {
		if (!is.list(xyl[[1]])) {
			list(xyl)
		} else {
			xyl
		}
	} else {
		if (!is.list(xyl[[1]])) {
			xyl <- list(gridCoords(xyl$x, xyl$y))
		} else {
			xyl <- lapply(xyl, function(x) gridCoords(x$x, x$y))
		}
		gridGrobCoords(xyl, name)
	}
}

#' @export
grobCoords.pp_grobCoords <- function(x, closed, ...) {
	if (getRversion() >= '4.2.0' && !closed) {
		return(emptyGrobCoords(x$name))
	}

	coords <- NextMethod()
	xyl <- coords_to_xylist(coords)
	xylists_to_grobcoords(xyl, x$name)
}

#' @export
grobPoints.pp_grobCoords <- function(x, closed, ...) {
	grobCoords(x, closed, ...)
}

print_and_play_4x6 <- function(cfg, pieces, quietly, bleed, size_bleed) {
	n_suits <- cfg$n_suits
	n_ranks <- cfg$n_ranks

	stopifnot(n_ranks <= 6)
	stopifnot(!bleed)

	if ("matchsticks" %in% pieces) {
		abort('"matchsticks" `pieces` not currently supported for `size = "4x6"`')
	}
	if ("pyramids" %in% pieces) {
		abort('"pyramids" `pieces` not currently supported for `size = "4x6"`')
	}
	if ("subpack" %in% pieces) {
		abort('"subpack" `pieces` not currently supported for `size = "4x6"`')
	}

	pl <- list()

	vp <- viewport(
		x = inch(3 + size_bleed$left),
		y = inch(2 + size_bleed$bottom),
		width = inch(6),
		height = inch(4)
	)

	draw_4x6_title(cfg, pieces, quietly, vp)

	n_pages_tiles <- 2 * n_suits
	for (i_suit in seq.int(n_suits)) {
		draw_4x6_tiles(cfg, i_suit, vp)
	}

	n_pages_coins <- (n_suits - 1) %/% 4 + 1
	for (i_page in seq.int(n_pages_coins)) {
		suits <- seq(4 * i_page - 3, 4 * i_page)
		draw_4x6_coins(cfg, suits, vp)
	}

	n_pages_dice <- (n_suits - 1) %/% 4 + 1
	for (i_page in seq.int(n_pages_dice)) {
		suits <- seq(4 * i_page - 3, 4 * i_page)
		draw_4x6_dice(cfg, suits, vp)
	}

	n_pages_pawns <- (n_suits - 1) %/% 4 + 1
	for (i_page in seq.int(n_pages_pawns)) {
		suits <- seq(4 * i_page - 3, 4 * i_page)
		draw_4x6_pawns(cfg, suits, vp)
	}

	n_pages <- 1 + n_pages_tiles + 2 * n_pages_coins + n_pages_dice + n_pages_pawns

	pl$Piecepack <- n_pages
	pl
}

draw_4x6_title <- function(cfg, pieces, quietly, vp) {
	grob <- a5_title_grob(cfg, pieces, quietly, FALSE)
	grob <- editGrob(grob, vp = vp)
	grid.draw(grob)
}

draw_4x6_tiles <- function(cfg, i_suit, vp) {
	df <- tibble::tibble(
		piece_side = "tile_face",
		x = rep(1:3 / 3 - 1 / 6, each = 2),
		y = rep(2:1 / 2 - 1 / 4, 3),
		suit = i_suit,
		rank = 1:6
	)
	grid.newpage()
	pushViewport(vp)
	pmap_piece(df, cfg = cfg, default.units = "npc")
	popViewport()

	df$piece_side = "tile_back"
	# Rotate tile backs to partially hide direction of face if tile back are not fully symmetric
	df$angle <- 90 * ((df$suit + df$rank) %% 4)
	grid.newpage()
	pushViewport(vp)
	pmap_piece(df, cfg = cfg, default.units = "npc")
	popViewport()
}

draw_4x6_coins <- function(cfg, suits, vp) {
	df <- tibble::tibble(
		piece_side = "coin_back",
		x = rep(1:6 / 6 - 1 / 12, 4),
		y = rep(4:1 / 4 - 1 / 8, each = 6),
		suit = rep(suits, each = 6),
		rank = rep(1:6, 4)
	)

	grid.newpage()
	pushViewport(vp)
	pmap_piece(df, cfg = cfg, default.units = "npc")
	popViewport()

	df$piece_side = "coin_face"
	grid.newpage()
	pushViewport(vp)
	pmap_piece(df, cfg = cfg, default.units = "npc")
	popViewport()
}

draw_4x6_dice <- function(cfg, suits, vp) {
	df <- tibble::tibble(
		piece_side = "die_face",
		x = rep(1:6 / 6 - 1 / 12, 4),
		y = rep(4:1 / 4 - 1 / 8, each = 6),
		suit = rep(suits, each = 6),
		rank = rep(1:6, 4)
	)

	grid.newpage()
	pushViewport(vp)
	pmap_piece(df, cfg = cfg, default.units = "npc")
	popViewport()
}

draw_4x6_pawns <- function(cfg, suits, vp) {
	dfp <- tibble::tibble(
		piece_side = "pawn_layout",
		x = 0.25,
		y = 4:1 / 4 - 1 / 8,
		angle = -90,
		suit = suits
	)
	#### Saucers in middle?
	dfb <- tibble::tibble(
		piece_side = "belt_face",
		x = 0.75,
		y = 4:1 / 4 - 1 / 8,
		angle = 0,
		suit = suits
	)
	df <- rbind(dfp, dfb)

	grid.newpage()
	pushViewport(vp)
	pmap_piece(df, cfg = cfg, default.units = "npc")
	popViewport()
}

# https://mathworld.wolfram.com/RegularOctahedron.html

## d8 Top
d8TopGrob <- function(
	piece_side,
	suit,
	rank,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = unit(0, "npc"),
	angle = 0,
	type = "normal",
	width = NA,
	height = NA,
	depth = NA,
	op_scale = 0,
	op_angle = 45
) {
	cfg <- as_pp_cfg(cfg)

	x <- convertX(x, "in", valueOnly = TRUE)
	y <- convertY(y, "in", valueOnly = TRUE)
	z <- convertX(z, "in", valueOnly = TRUE)
	width <- convertX(width, "in", valueOnly = TRUE)
	height <- convertY(height, "in", valueOnly = TRUE)
	depth <- convertX(depth, "in", valueOnly = TRUE)

	#### allow limited 3D rotation #281
	axis_x <- 0
	axis_y <- 0
	####
	xyz <- d8t_xyz(x, y, z, angle, axis_x, axis_y, width, height, depth)
	p <- regular_ngon_polygon2d(6, x = 0.5, y = 0.5, theta = degrees(0))
	edge_types <- paste0(
		"d8_",
		c("right", "back", "left", "opposite_left", "opposite_back", "opposite_right")
	)
	order <- painter_order(p, alpha = degrees(op_angle))
	df <- tibble(index = 1:6, edge = edge_types, level = c(2, 1, 2, 1, 2, 1))[order, ]
	df <- df[order(df$level), ]
	gl <- gList()
	for (i in 1:nrow(df)) {
		edge <- df$edge[i]
		edge_rank <- d8_edge_rank(edge, rank)
		opt <- cfg$get_piece_opt("die_face", suit, edge_rank)
		gp <- gpar(col = opt$border_color, lex = opt$border_lex, fill = opt$background_color)
		xyz_polygon <- switch(
			edge,
			d8_left = xyz[c(1, 5, 2)],
			d8_back = xyz[c(1, 6, 5)],
			d8_right = xyz[c(1, 3, 6)],
			d8_opposite_left = xyz[c(4, 2, 5)],
			d8_opposite_back = xyz[c(4, 3, 2)],
			d8_opposite_right = xyz[c(4, 6, 3)]
		)
		xy_polygon <- as_coord2d(xyz_polygon, alpha = degrees(op_angle), scale = op_scale)
		xy_vp <- xy_vp_convex(xyz_polygon, op_scale, op_angle)
		gl[[i]] <- at_ps_grob("die_face", suit, edge_rank, cfg, xy_vp, xy_polygon, name = edge)
	}
	# face
	xyz_polygon <- xyz[1:3]
	xy_polygon <- as_coord2d(xyz_polygon, alpha = degrees(op_angle), scale = op_scale)
	xy_vp <- xy_vp_convex(xyz_polygon, op_scale, op_angle)
	gl[[nrow(df) + 1]] <- at_ps_grob(
		"die_face",
		suit,
		rank,
		cfg,
		xy_vp,
		xy_polygon,
		name = "d8_face"
	)

	# pre-compute grobCoords
	coords_xyl <- as.list(convex_hull2d(as_coord2d(
		xyz,
		alpha = degrees(op_angle),
		scale = op_scale
	)))

	gTree(
		scale = 1,
		coords_xyl = coords_xyl,
		children = gl,
		cl = c("projected_rpg_die", "coords_xyl")
	)
}

d8t_xyz <- function(x, y, z, angle, axis_x, axis_y, width, height, depth) {
	xyz_t <- as_coord3d(regular_ngon_polygon2d(3, x = 0, y = 0, theta = degrees(90)), z = 0.5)
	xyz_b <- as_coord3d(
		regular_ngon_polygon2d(3, x = 0, y = 0, theta = degrees(-90))[c(1, 3, 2)],
		z = -0.5
	)
	xs <- c(xyz_t$x, xyz_b$x)
	ys <- c(xyz_t$y, xyz_b$y)
	zs <- c(xyz_t$z, xyz_b$z)
	as_coord3d(xs, ys, zs)$scale(width, height, depth)$transform(AA_to_R(
		angle,
		axis_x,
		axis_y
	))$translate(as_coord3d(x, y, z))
}

save_d8_obj <- function(
	piece_side,
	suit,
	rank,
	cfg,
	x,
	y,
	z,
	angle,
	axis_x,
	axis_y,
	width,
	height,
	depth,
	filename,
	scale,
	res
) {
	cfg <- as_pp_cfg(cfg)
	opt <- cfg$get_piece_opt(piece_side, suit, rank)
	xyz <- d8_xyz(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth)

	p <- regular_ngon_polygon2d(3, x = 0.5, y = 0.5, theta = degrees(90))
	xy_vt <- list(
		x = rep(p$x / 2, 8) + rep(rep(0:1 / 2, each = 3), 4),
		y = rep(p$y / 4, 8) + rep(rep(3:0 / 4, each = 3), each = 2)
	)

	# textured face elements: face, left, right, back, opposite, opposite left, opposite right, opposite back
	f <- list()
	fn_vt <- function(r) switch(r, 1:3, 4:6, 7:9, 10:12, 13:15, 16:18, 19:21, 22:24)
	f[[1]] <- list(v = 1:3, vt = fn_vt(rank))
	f[[2]] <- list(v = 4:6, vt = fn_vt(d8_edge_rank("d8_left", rank)))
	f[[3]] <- list(v = 7:9, vt = fn_vt(d8_edge_rank("d8_right", rank)))
	f[[4]] <- list(v = 10:12, vt = fn_vt(d8_edge_rank("d8_back", rank)))
	f[[5]] <- list(v = 13:15, vt = fn_vt(d8_edge_rank("d8_opposite", rank)))
	f[[6]] <- list(v = 16:18, vt = fn_vt(d8_edge_rank("d8_opposite_left", rank)))
	f[[7]] <- list(v = 19:21, vt = fn_vt(d8_edge_rank("d8_opposite_right", rank)))
	f[[8]] <- list(v = 22:24, vt = fn_vt(d8_edge_rank("d8_opposite_back", rank)))

	ext <- tools::file_ext(filename)
	mtl_filename <- gsub(paste0("\\.", ext, "$"), ".mtl", filename)
	png_filename <- gsub(paste0("\\.", ext, "$"), ".png", filename)

	write_obj(filename, v = xyz, vt = xy_vt, f = f)
	write_d8_texture("die_face", suit, rank, cfg, filename = png_filename, res = res)

	invisible(list(obj = filename, mtl = mtl_filename, png = png_filename))
}

d8_edge_rank <- function(edge, rank) {
	switch(
		edge,
		d8_left = d8_rank(rank + 2),
		d8_back = d8_rank(rank + 4),
		d8_right = d8_rank(rank + 6),
		d8_opposite = 9 - rank,
		d8_opposite_left = d8_rank(9 - rank + 2),
		d8_opposite_back = d8_rank(9 - rank + 4),
		d8_opposite_right = d8_rank(9 - rank + 6)
	)
}

d8_rank <- function(x) {
	(x - 1) %% 8 + 1
}

d8_xyz <- function(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth) {
	xyz_t <- as_coord3d(regular_ngon_polygon2d(3, x = 0, y = 0, theta = degrees(90)), z = 0.5)
	xyz_b <- as_coord3d(
		regular_ngon_polygon2d(3, x = 0, y = 0, theta = degrees(-90))[c(1, 3, 2)],
		z = -0.5
	)

	# face
	xs_f <- xyz_t$x
	ys_f <- xyz_t$y
	zs_f <- xyz_t$z

	# left
	xs_l <- c(xyz_t$x[1], xyz_b$x[2], xyz_t$x[2])
	ys_l <- c(xyz_t$y[1], xyz_b$y[2], xyz_t$y[2])
	zs_l <- c(xyz_t$z[1], xyz_b$z[2], xyz_t$z[2])

	# right
	xs_r <- c(xyz_t$x[1], xyz_t$x[3], xyz_b$x[3])
	ys_r <- c(xyz_t$y[1], xyz_t$y[3], xyz_b$y[3])
	zs_r <- c(xyz_t$z[1], xyz_t$z[3], xyz_b$z[3])

	# back
	xs_b <- c(xyz_t$x[1], xyz_b$x[3], xyz_b$x[2])
	ys_b <- c(xyz_t$y[1], xyz_b$y[3], xyz_b$y[2])
	zs_b <- c(xyz_t$z[1], xyz_b$z[3], xyz_b$z[2])

	# opposite
	xs_o <- xyz_b$x
	ys_o <- xyz_b$y
	zs_o <- xyz_b$z

	# opposite left
	xs_ol <- c(xyz_b$x[1], xyz_t$x[2], xyz_b$x[2])
	ys_ol <- c(xyz_b$y[1], xyz_t$y[2], xyz_b$y[2])
	zs_ol <- c(xyz_b$z[1], xyz_t$z[2], xyz_b$z[2])

	# opposite right
	xs_or <- c(xyz_b$x[1], xyz_b$x[3], xyz_t$x[3])
	ys_or <- c(xyz_b$y[1], xyz_b$y[3], xyz_t$y[3])
	zs_or <- c(xyz_b$z[1], xyz_b$z[3], xyz_t$z[3])

	# opposite back
	xs_ob <- c(xyz_b$x[1], xyz_t$x[3], xyz_t$x[2])
	ys_ob <- c(xyz_b$y[1], xyz_t$y[3], xyz_t$y[2])
	zs_ob <- c(xyz_b$z[1], xyz_t$z[3], xyz_t$z[2])

	xs <- c(xs_f, xs_l, xs_r, xs_b, xs_o, xs_ol, xs_or, xs_ob)
	ys <- c(ys_f, ys_l, ys_r, ys_b, ys_o, ys_ol, ys_or, ys_ob)
	zs <- c(zs_f, zs_l, zs_r, zs_b, zs_o, zs_ol, zs_or, zs_ob)

	as_coord3d(xs, ys, zs)$scale(width, height, depth)$transform(AA_to_R(
		angle,
		axis_x,
		axis_y
	))$translate(as_coord3d(x, y, z))
}

write_d8_texture <- function(
	piece_side = "die_face",
	suit = 1,
	rank = 1,
	cfg = pp_cfg(),
	...,
	filename = tempfile(fileext = ".png"),
	res = 72
) {
	current_dev <- grDevices::dev.cur()
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	width <- cfg$get_width("die_face", suit, rank)

	args <- list(
		filename = filename,
		height = 4 * width,
		width = 2 * width,
		units = "in",
		res = res,
		bg = "transparent"
	)
	if (capabilities("cairo")) {
		args$type <- "cairo"
	}
	do.call(grDevices::png, args)

	xs <- rep(1:2 / 2 - 1 / 4, 4)
	ys <- rep(4:1 / 4 - 1 / 8, each = 2)
	for (i in 1:8) {
		pushViewport(viewport(x = xs[i], width = 1 / 2, y = ys[i], height = 1 / 4))
		draw_piece_and_bleed("die_face", suit, i, cfg)
		popViewport()
	}

	grDevices::dev.off()
	invisible(filename)
}

#' @importFrom tibble tibble
xya_pips_cards <- function(n_pips) {
	if (n_pips == 0) {
		return(tibble(x = numeric(0), y = numeric(0), angle = numeric(0)))
	}
	if (n_pips > 10) {
		abort(paste0("Don't know pip pattern for ", n_pips, " pips"))
	}
	switch(
		n_pips,
		tibble(x = 0.5, y = 0.5, angle = 0), # 1
		tibble(x = 0.5, y = c(0.2, 0.8), angle = c(180, 0)), # 2
		tibble(x = 0.5, y = c(0.2, 0.5, 0.8), angle = c(180, 0, 0)), # 3
		tibble(
			x = rep(c(0.25, 0.75), 2),
			y = rep(c(0.2, 0.8), each = 2), # 4
			angle = rep(c(180, 0), each = 2)
		),
		tibble(
			x = c(rep(c(0.25, 0.75), 2), 0.5), # 5
			y = c(rep(c(0.2, 0.8), each = 2), 0.5),
			angle = c(rep(180, 2), rep(0, 3))
		),
		tibble(
			x = rep(c(0.25, 0.75), 3),
			y = rep(c(0.2, 0.5, 0.8), each = 2), # 6
			angle = c(rep(180, 2), rep(0, 4))
		),
		tibble(
			x = c(rep(c(0.25, 0.75), 3), 0.5), # 7
			y = c(rep(c(0.2, 0.5, 0.75), each = 2), 0.55),
			angle = c(rep(180, 2), rep(0, 5))
		),
		tibble(
			x = rep(c(0.25, 0.75), 4), # 8
			y = rep(c(0.2, 0.4, 0.6, 0.8), each = 2),
			angle = c(rep(180, 4), rep(0, 4))
		),
		tibble(
			x = c(rep(c(0.25, 0.75), 4), 0.5), # 9
			y = c(rep(c(0.2, 0.4, 0.6, 0.8), each = 2), 0.7),
			angle = c(rep(180, 4), rep(0, 5))
		),
		tibble(
			x = c(0.5, rep(c(0.25, 0.75), 4), 0.5), # 10
			y = c(0.3, rep(c(0.2, 0.4, 0.6, 0.8), each = 2), 0.7),
			angle = c(rep(180, 5), rep(0, 5))
		)
	)
}

xya_pips_dominoes <- function(n_pips, die = FALSE, chinese = FALSE, other_pips = NULL) {
	if (n_pips == 0) {
		return(tibble(x = numeric(0), y = numeric(0), angle = numeric(0)))
	}
	if (die && !chinese) {
		high <- 0.75
	} else if (chinese && die) {
		if (n_pips == 4) {
			high <- 0.65
		} else if (n_pips == 6) {
			high <- 0.75
		} else {
			high <- 0.68
		}
	} else {
		high <- 0.78
	}
	low <- 1 - high
	if (chinese && !die) {
		low <- 0.2
		if (n_pips == 3 && other_pips < 5) {
			high <- 1.0
		} else if (n_pips == 4 && other_pips == 4) {
			high <- 0.75
		} else {
			high <- 0.84
		}
	}
	if (chinese && die && n_pips == 6) {
		left <- 0.36
	} else if (chinese && !die) {
		left <- 0.30
	} else {
		left <- low
	}
	right <- 1 - left
	mid <- 0.5 * (high + low)
	if (n_pips > 18L) {
		abort(paste0("Don't know pip pattern for ", n_pips, " pips"))
	}
	switch(
		n_pips,
		if (chinese && !die) {
			tibble(x = 0.5, y = low, angle = 0)
		} else {
			tibble(x = 0.5, y = mid, angle = 0)
		}, # 1
		if (chinese) {
			if (die) {
				tibble(x = 0.5, y = c(low, high), angle = c(0, 180))
			} else {
				tibble(x = c(left, right), y = low, angle = c(0, 180))
			}
		} else {
			tibble(x = c(left, right), y = c(high, low), angle = c(0, 180))
		}, # 2
		tibble(x = c(left, 0.5, right), y = c(high, mid, low), angle = c(0, 0, 180)), # 3
		if (chinese && !die && other_pips != 4) {
			tibble(
				x = rep(c(left, right), 2), # 4
				y = rep(c(low, mid), each = 2),
				angle = rep(c(180, 0), each = 2)
			)
		} else {
			tibble(
				x = rep(c(left, right), 2), # 4
				y = rep(c(low, high), each = 2),
				angle = rep(c(180, 0), each = 2)
			)
		},
		tibble(
			x = c(rep(c(left, right), 2), 0.5), # 5
			y = c(rep(c(low, high), each = 2), mid),
			angle = c(rep(180, 2), rep(0, 3))
		),
		tibble(
			x = rep(c(left, right), 3),
			y = rep(c(low, mid, high), each = 2), # 6
			angle = c(rep(180, 2), rep(0, 4))
		),
		tibble(
			x = c(rep(c(left, right), 3), 0.5), # 7
			y = c(rep(c(low, 0.5, high), each = 2), 0.50),
			angle = c(rep(180, 2), rep(0, 5))
		),
		tibble(
			x = c(0.5, rep(c(left, right), 3), 0.5), # 8
			y = c(low, rep(c(low, 0.5, high), each = 2), high),
			angle = c(rep(180, 3), rep(0, 5))
		),
		tibble(
			x = c(rep(c(left, 0.5, right), 3)), # 9
			y = c(rep(c(low, 0.5, high), each = 3)),
			angle = c(rep(180, 3), rep(0, 6))
		),
		tibble(
			x = c(0.5, rep(c(0.20, 0.80), 4), 0.5), # 10
			y = c(0.2, rep(c(0.2, 0.4, 0.6, 0.8), each = 2), 0.8),
			angle = c(rep(180, 5), rep(0, 5))
		),
		tibble(
			x = c(0.5, rep(c(0.20, 0.80), 4), 0.5, 0.5), # 11
			y = c(0.2, rep(c(0.2, 0.4, 0.6, 0.8), each = 2), 0.5, 0.8),
			angle = c(rep(180, 5), rep(0, 6))
		),
		tibble(
			x = c(rep(c(0.20, 0.5, 0.80), 4)), # 12
			y = c(rep(c(0.2, 0.4, 0.6, 0.8), each = 3)),
			angle = c(rep(180, 6), rep(0, 6))
		),
		tibble(
			x = c(rep(c(0.2, 0.4, 0.6, 0.8), 3), 0.5), # 13
			y = c(rep(c(0.2, 0.4, 0.6), each = 4), 0.8),
			angle = c(rep(180, 8), rep(0, 5))
		),
		tibble(
			x = c(rep(c(0.2, 0.4, 0.6, 0.8), 3), 0.3, 0.7), # 14
			y = c(rep(c(0.2, 0.4, 0.6), each = 4), 0.8, 0.8),
			angle = c(rep(180, 8), rep(0, 6))
		),
		tibble(
			x = c(rep(c(0.2, 0.4, 0.6, 0.8), 3), 0.3, 0.5, 0.7), # 15
			y = c(rep(c(0.2, 0.4, 0.6), each = 4), 0.8, 0.8, 0.8),
			angle = c(rep(180, 8), rep(0, 7))
		),
		tibble(
			x = c(rep(c(0.2, 0.4, 0.6, 0.8), 4)), # 16
			y = c(rep(c(0.2, 0.4, 0.6, 0.8), each = 4)),
			angle = c(rep(180, 8), rep(0, 8))
		),
		tibble(
			x = c(rep(c(0.140, 0.340, 0.660, 0.860), 4), 0.5), # 17
			y = c(rep(c(0.2, 0.4, 0.6, 0.8), each = 4), 0.5),
			angle = c(rep(180, 8), rep(0, 9))
		),
		tibble(
			x = c(rep(c(0.140, 0.340, 0.660, 0.860), 4), 0.5, 0.5), # 18
			y = c(rep(c(0.2, 0.4, 0.6, 0.8), each = 4), 0.3, 0.7),
			angle = c(rep(180, 8), rep(0, 8), 180, 0)
		)
	)
}

cardGrobFn <- function(
	rank_offset = 0,
	type = "card",
	grob_type = "text",
	has_pips = TRUE,
	fill_stroke = FALSE
) {
	force(type)
	force(rank_offset)
	force(grob_type)
	force(has_pips)
	force(fill_stroke)
	function(piece_side, suit, rank, cfg = pp_cfg()) {
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)

		gTree(
			opt = opt,
			piece_side = piece_side,
			suit = suit,
			rank = rank,
			cfg = cfg,
			type = type,
			rank_offset = rank_offset,
			grob_type = grob_type,
			fill_stroke = fill_stroke,
			scale = 1,
			border = TRUE,
			has_pips = has_pips,
			cl = c("card_side", "basic_piece_side")
		)
	}
}

#' @export
makeContent.card_side <- function(x) {
	opt <- x$opt

	type <- x$type
	rank_offset <- x$rank_offset
	grob_type <- x$grob_type

	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	background_grob <- shape$shape(
		gp = gpar(col = NA, fill = opt$background_color),
		name = "background"
	)
	gp_mat <- gpar(col = NA, lwd = 0, fill = opt$mat_color)
	mat_grob <- shape$mat(opt$mat_width, gp = gp_mat, name = "mat")

	gp_rank <- gpar(
		col = opt$ps_color,
		fontsize = opt$ps_fontsize,
		fontfamily = opt$ps_fontfamily,
		fontface = opt$ps_fontface,
		lineheight = 0.8
	)
	rank_grob <- textGrob(
		label = opt$ps_text,
		gp = gp_rank,
		x = c(0.1, 0.9),
		y = c(0.9, 0.1),
		rot = c(0, 180),
		hjust = 0.5,
		vjust = 0.5,
		name = "rank"
	)
	if (isTRUE(x$fill_stroke)) {
		rank_grob <- as_fill_stroke_grob(rank_grob, fill = opt$ps_color)
	}

	gp_suit <- gpar(
		col = opt$dm_color,
		fontsize = opt$dm_fontsize,
		fontfamily = opt$dm_fontfamily,
		fontface = opt$dm_fontface
	)
	suit_grob <- textGrob(
		label = opt$dm_text,
		gp = gp_suit,
		x = c(0.1, 0.9),
		y = c(0.8, 0.2),
		rot = c(0, 180),
		hjust = 0.5,
		vjust = 0.5,
		name = "suit"
	)
	if (isTRUE(x$fill_stroke)) {
		suit_grob <- as_fill_stroke_grob(suit_grob, fill = opt$dm_color)
	}

	# Pips
	if (x$has_pips) {
		tfn <- pippedGrobFn(
			rank_offset,
			type,
			grob_type,
			border = FALSE,
			mat = FALSE,
			fill_stroke = x$fill_stroke
		)
		pip_grob <- tfn(x$piece_side, x$suit, x$rank, x$cfg)
		pip_grob <- grid::editGrob(
			pip_grob,
			vp = viewport(height = 1.0, width = 0.80, gp = gpar(cex = 1.8)),
			name = "pips"
		)
	} else {
		pip_grob <- nullGrob(name = "pips")
	}

	# Border
	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- shape$shape(gp = gp_border, name = "border")
	} else {
		border_grob <- nullGrob(name = "border")
	}

	gl <- gList(background_grob, pip_grob, rank_grob, suit_grob, mat_grob, border_grob)
	setChildren(x, gl)
}

jokerCardGrobFn <- function(star = TRUE, fill_stroke = FALSE) {
	force(star)
	force(fill_stroke)
	function(piece_side, suit, rank, cfg = pp_cfg()) {
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)

		gTree(
			opt = opt,
			star = star,
			fill_stroke = fill_stroke,
			scale = 1,
			border = TRUE,
			cl = c("joker_card_side", "basic_piece_side")
		)
	}
}

#' @export
makeContent.joker_card_side <- function(x) {
	opt <- x$opt

	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	background_grob <- shape$shape(
		gp = gpar(col = NA, fill = opt$background_color),
		name = "background"
	)
	gp_mat <- gpar(col = NA, lwd = 0, fill = opt$mat_color)
	mat_grob <- shape$mat(opt$mat_width, gp = gp_mat, name = "mat")

	gp_rank <- gpar(
		col = opt$ps_color,
		fontsize = opt$ps_fontsize,
		fontfamily = opt$ps_fontfamily,
		fontface = opt$ps_fontface,
		lineheight = 0.8
	)
	rank_grob <- textGrob(
		label = opt$ps_text,
		gp = gp_rank,
		x = c(0.1, 0.9),
		y = c(0.9, 0.1),
		rot = c(0, 180),
		hjust = 0.5,
		vjust = 0.5,
		name = "rank"
	)
	if (isTRUE(x$fill_stroke)) {
		rank_grob <- as_fill_stroke_grob(rank_grob, fill = opt$ps_color)
	}

	# Joker
	joker_grob <- jokerGrob(opt, x$star, x$fill_stroke)

	# Border
	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- shape$shape(gp = gp_border, name = "border")
	} else {
		border_grob <- nullGrob(name = "border")
	}

	gl <- gList(background_grob, joker_grob, rank_grob, mat_grob, border_grob)
	setChildren(x, gl)
}

jokerGrob <- function(opt, star = TRUE, fill_stroke = FALSE) {
	gp_meeple <- gpar(col = opt$ps_color, lwd = 4)
	vp_meeple <- viewport(height = inch(1.3), width = inch(1.2), y = 0.4)
	meeple_grob <- pp_shape("meeple")$shape(gp = gp_meeple, vp = vp_meeple)
	if (fill_stroke) {
		gp_triangle <- gpar(fill = opt$ps_color, col = opt$border_col, lwd = 1.5)
		gp_circle <- gpar(fill = opt$ps_color, col = opt$border_col, lwd = 1.5)
	} else {
		gp_triangle <- gpar(fill = opt$ps_color, col = NA_character_)
		gp_circle <- gpar(col = opt$ps_color, lwd = 4)
	}
	y_triangle <- unit(0.4, "npc") + inch(0.5 * 1.3) + inch(0.5 * 0.8) + inch(-0.10)
	vp_triangle <- viewport(height = inch(0.8), width = inch(0.45), y = y_triangle)
	triangle_grob <- pp_shape("pyramid")$shape(gp = gp_triangle, vp = vp_triangle)
	y_circle <- y_triangle + inch(0.5 * 0.8) + inch(0.5 * 0.2) + inch(-0.02)
	vp_circle <- viewport(height = inch(0.2), width = inch(0.2), y = y_circle)
	circle_grob <- circleGrob(vp = vp_circle, gp = gp_circle)
	if (star) {
		star_grob <- pp_shape("concave5")$shape(gp = gp_triangle, vp = vp_circle)
	} else {
		star_grob <- NULL
	}
	grobTree(
		meeple_grob,
		triangle_grob,
		circle_grob,
		star_grob,
		vp = viewport(width = 0.8, height = 0.9),
		cl = "joker"
	)
}

faceCardGrobFn <- function(label = "", placement = "high", fill_stroke = FALSE) {
	force(label)
	force(placement)
	force(fill_stroke)
	function(piece_side, suit, rank, cfg = pp_cfg()) {
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)

		gTree(
			opt = opt,
			label = label,
			placement = placement,
			scale = scale,
			border = TRUE,
			fill_stroke = fill_stroke,
			cl = c("face_card_side", "basic_piece_side")
		)
	}
}

#' @export
makeContent.face_card_side <- function(x) {
	opt <- x$opt
	label <- x$label
	placement <- x$placement

	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	background_grob <- shape$shape(
		gp = gpar(col = NA, fill = opt$background_color),
		name = "background"
	)
	gp_mat <- gpar(col = NA, lwd = 0, fill = opt$mat_color)
	mat_grob <- shape$mat(opt$mat_width, gp = gp_mat, name = "mat")

	gp_rank <- gpar(
		col = opt$ps_color,
		fontsize = opt$ps_fontsize,
		fontfamily = opt$ps_fontfamily,
		fontface = opt$ps_fontface,
		lineheight = 0.8
	)
	rank_grob <- textGrob(
		label = opt$ps_text,
		gp = gp_rank,
		x = c(0.1, 0.9),
		y = c(0.9, 0.1),
		rot = c(0, 180),
		hjust = 0.5,
		vjust = 0.5,
		name = "rank"
	)
	if (isTRUE(x$fill_stroke)) {
		rank_grob <- as_fill_stroke_grob(rank_grob, fill = opt$ps_color)
	}

	gp_suit <- gpar(
		col = opt$dm_color,
		fontsize = opt$dm_fontsize,
		fontfamily = opt$dm_fontfamily,
		fontface = opt$dm_fontface
	)
	suit_grob <- textGrob(
		label = opt$dm_text,
		gp = gp_suit,
		x = c(0.1, 0.9),
		y = c(0.8, 0.2),
		rot = c(0, 180),
		hjust = 0.5,
		vjust = 0.5,
		name = "suit"
	)
	if (isTRUE(x$fill_stroke)) {
		suit_grob <- as_fill_stroke_grob(suit_grob, fill = opt$dm_color)
	}

	# Face
	top_grob <- faceGrob(opt, label, placement, fill_stroke = x$fill_stroke)
	top_grob <- grid::editGrob(
		top_grob,
		vp = viewport(y = 0.75, height = 0.5, width = 0.80),
		name = "top"
	)
	bot_grob <- faceGrob(opt, label, placement, fill_stroke = x$fill_stroke)
	bot_grob <- grid::editGrob(
		top_grob,
		vp = viewport(y = 0.25, height = 0.5, width = 0.80, angle = 180),
		name = "bottom"
	)

	# Border
	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- shape$shape(gp = gp_border, name = "border")
	} else {
		border_grob <- nullGrob(name = "border")
	}

	gl <- gList(background_grob, top_grob, bot_grob, rank_grob, suit_grob, mat_grob, border_grob)
	setChildren(x, gl)
}

faceGrob <- function(opt, label = "", placement = "high", fill_stroke = FALSE) {
	if (placement == "high") {
		y <- 0.7
		x_meeple <- 0.5
		y_meeple <- 0.3
		in_meeple <- 1
		cex <- 1.5
	} else {
		y <- 0.25
		x_meeple <- 0.57
		y_meeple <- 0.42
		in_meeple <- 0.6
		cex <- 3.0
	}
	rot <- ifelse(label %in% c("\u046a", "\u050a"), 180, 0)
	gp_label <- gpar(
		col = opt$ps_color,
		fontsize = opt$ps_fontsize,
		fontfamily = opt$ps_fontfamily,
		fontface = opt$ps_fontface,
		lineheight = 0.8,
		cex = cex
	)
	label_grob <- textGrob(
		label = label,
		gp = gp_label,
		x = 0.5,
		y = y,
		hjust = 0.5,
		vjust = 0.5,
		rot = rot,
		name = "label"
	)
	if (isTRUE(fill_stroke)) {
		label_grob <- as_fill_stroke_grob(label_grob, fill = opt$ps_color)
	}
	gp_suit <- gpar(
		col = opt$dm_color,
		fontsize = opt$dm_fontsize,
		fontfamily = opt$dm_fontfamily,
		fontface = opt$dm_fontface,
		cex = in_meeple
	)
	suit_grob <- textGrob(
		label = opt$dm_text,
		gp = gp_suit,
		x = x_meeple,
		y = y_meeple,
		hjust = 0.5,
		vjust = 0.5,
		name = "suit"
	)
	if (isTRUE(fill_stroke)) {
		suit_grob <- as_fill_stroke_grob(suit_grob, fill = opt$dm_color)
	}
	gp_meeple <- gpar(col = opt$ps_color, lwd = 4)
	vp_meeple <- viewport(
		height = inch(in_meeple),
		width = inch(in_meeple),
		x = x_meeple,
		y = y_meeple
	)
	meeple_grob <- pp_shape("meeple")$shape(gp = gp_meeple, vp = vp_meeple)
	grobTree(meeple_grob, label_grob, suit_grob, cl = "face")
}

dominoGrobFn <- function(rank_offset = 0, type = "domino", grob_type = "circle") {
	force(rank_offset)
	force(type)
	force(grob_type)
	function(piece_side, suit, rank, cfg = pp_cfg()) {
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)

		gTree(
			opt = opt,
			cfg = cfg,
			suit = suit,
			rank = rank,
			type = type,
			rank_offset = rank_offset,
			grob_type = grob_type,
			scale = 1,
			border = TRUE,
			cl = c("domino_side", "basic_piece_side")
		)
	}
}

#' @export
makeContent.domino_side <- function(x) {
	opt <- x$opt
	suit <- x$suit
	rank <- x$rank
	type <- x$type
	rank_offset <- x$rank_offset
	grob_type <- x$grob_type

	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	background_grob <- shape$shape(
		gp = gpar(col = NA, fill = opt$background_color),
		name = "background"
	)

	gp_mat <- gpar(col = NA, lwd = 0, fill = opt$mat_color)
	mat_grob <- shape$mat(opt$mat_width, gp = gp_mat, name = "mat")

	# Top (Rank)
	tfn <- pippedGrobFn(
		rank_offset,
		type,
		grob_type,
		border = FALSE,
		mat = FALSE,
		background = !(type == "domino_chinese"),
		other_pips = suit
	)
	top_grob <- tfn("tile_face", rank, rank, x$cfg)
	top_grob <- grid::editGrob(
		top_grob,
		vp = viewport(height = 0.5, y = 0.75, angle = 180),
		name = "top_rank"
	)

	# Bottom (Suit)
	bfn <- pippedGrobFn(
		rank_offset,
		type,
		grob_type,
		border = FALSE,
		mat = FALSE,
		background = !(type == "domino_chinese"),
		other_pips = rank
	)
	bot_grob <- bfn("tile_face", suit, suit, x$cfg)
	bot_grob <- grid::editGrob(
		bot_grob,
		vp = viewport(height = 0.5, y = 0.25),
		name = "bottom_suit"
	)

	gp_gl <- gpar(col = opt$gridline_color, lex = opt$gridline_lex)
	gl_grob <- segmentsGrob(x0 = 0.1, x1 = 0.9, y0 = 0.5, y1 = 0.5, gp = gp_gl, name = "gridlines")

	# Border
	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- shape$shape(gp = gp_border, name = "border")
	} else {
		border_grob <- nullGrob(name = "border")
	}

	gl <- gList(background_grob, top_grob, bot_grob, gl_grob, mat_grob, border_grob)
	setChildren(x, gl)
}

pippedGrobFn <- function(
	rank_offset = 0,
	type = "die",
	grob_type = "circle",
	border = TRUE,
	mat = TRUE,
	background = TRUE,
	other_pips = NULL,
	fill_stroke = FALSE
) {
	force(type)
	force(rank_offset)
	force(grob_type)
	force(border)
	force(mat)
	force(background)
	force(other_pips)
	force(fill_stroke)
	function(piece_side, suit, rank, cfg = pp_cfg()) {
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)
		gTree(
			opt = opt,
			n_pips = rank + rank_offset,
			type = type,
			grob_type = grob_type,
			border = border,
			mat = mat,
			background = background,
			other_pips = other_pips + rank_offset,
			fill_stroke = fill_stroke,
			name = NULL,
			gp = gpar(),
			vp = NULL,
			cl = "pipped"
		)
	}
}

#' @export
makeContent.pipped <- function(x) {
	opt <- x$opt
	if (x$type == "card") {
		xya <- xya_pips_cards(x$n_pips)
	} else if (x$type == "die") {
		xya <- xya_pips_dominoes(x$n_pips, die = TRUE)
	} else if (x$type == "die_chinese") {
		xya <- xya_pips_dominoes(x$n_pips, die = TRUE, chinese = TRUE)
	} else if (x$type == "domino_chinese") {
		xya <- xya_pips_dominoes(x$n_pips, die = FALSE, chinese = TRUE, other_pips = x$other_pips)
	} else {
		xya <- xya_pips_dominoes(x$n_pips)
	}
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	# Background
	if (x$background) {
		background_grob <- shape$shape(
			gp = gpar(col = NA, fill = opt$background_color),
			name = "background"
		)
	} else {
		background_grob <- NULL
	}

	# Mat
	if (x$mat) {
		gp_mat <- gpar(col = NA, lwd = 0, fill = opt$mat_color)
		mat_grob <- shape$mat(opt$mat_width, gp = gp_mat, name = "mat")
	} else {
		mat_grob <- NULL
	}

	# Pips
	if (nrow(xya) > 0) {
		if (x$grob_type == "circle") {
			if (x$type == "die_chinese" && (x$n_pips == 1 || x$n_pips == 4)) {
				pip_color <- "#D55E00" # or just use "red"?
			} else if (x$type == "domino_chinese" && (x$n_pips == 1 || x$n_pips == 4)) {
				pip_color <- "#D55E00" # or just use "red"?
			} else if (x$type == "domino_chinese" && x$n_pips == 6 && x$other_pips == 6) {
				pip_color <- rep(c(opt$dm_color, "#D55E00"), 3L)
			} else {
				pip_color <- opt$dm_color
			}
			gp_pip <- gpar(col = pip_color, fill = pip_color)
			if (x$type != "card" && x$n_pips > 9) {
				r <- 0.06
			} else if (x$type == "die_chinese") {
				if (x$n_pips == 1) {
					r <- 0.20
				} else if (x$n_pips == 2) {
					r <- 0.10
				} else {
					r <- 0.10
				}
			} else if (x$type == "domino_chinese") {
				r <- 0.16
			} else {
				r <- 0.08
			}
			pip_grob <- circleGrob(x = xya$x, y = xya$y, r = r, gp = gp_pip, name = "pips")
		} else {
			gp_pip <- gpar(
				col = opt$dm_color,
				fontsize = opt$dm_fontsize,
				fontfamily = opt$dm_fontfamily,
				fontface = opt$dm_fontface
			)
			pip_grob <- textGrob(
				opt$dm_text,
				x = xya$x,
				y = xya$y,
				rot = xya$angle,
				gp = gp_pip,
				hjust = 0.5,
				vjust = 0.5,
				name = "pips"
			)
			if (isTRUE(x$fill_stroke)) {
				pip_grob <- as_fill_stroke_grob(pip_grob, fill = opt$dm_color)
			}
		}
	} else {
		pip_grob <- nullGrob(name = "pips")
	}

	# Border
	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- shape$shape(gp = gp_border, name = "border")
	} else {
		border_grob <- NULL
	}

	gl <- gList(background_grob, mat_grob, pip_grob, border_grob)
	setChildren(x, gl)
}

test_that("no regressions in `game_systems()`", {
	expect_error(game_systems("boobear"), "Don't have a customized configuration for font")
})


test_that("removed `style` argument errors", {
	expect_snapshot(error = TRUE, game_systems(style = "dejavu3d"))
})

test_that("no regressions in `game_systems()` figures", {
	skip_on_ci()
	skip_if_not_installed("vdiffr")
	library("vdiffr")
	# d6 dice
	envir <- game_systems()
	suppressMessages(
		{
			expect_doppelganger("dice-d6", function() {
				grid.piece(
					"die_face",
					x = 1:6,
					y = 1,
					default.units = "in",
					rank = 1:6,
					suit = 1:6,
					op_scale = 0.5,
					cfg = envir$dice
				)
				grid.piece(
					"die_face",
					x = 1:6,
					y = 2,
					default.units = "in",
					rank = 1:6,
					suit = 1:6,
					op_scale = 0.5,
					cfg = envir$dice_numeral
				)
				grid.piece(
					"die_face",
					x = 1:6,
					y = 3,
					default.units = "in",
					rank = 1:6,
					suit = 1:6,
					op_scale = 0.5,
					cfg = envir$dice_fudge
				)
			})
		},
		classes = "piecepackr_affine_transformation"
	)

	# dominoes
	skip_if_not_installed("systemfonts") # prevent buggy cairo interaction with "dejavu" style
	envir <- game_systems("dejavu")
	expect_doppelganger("dominoes", function() {
		grid.piece(
			"tile_face",
			x = rep(6:1, 3),
			y = rep(2 * 3:1, each = 6),
			suit = 1:18,
			rank = 1:18 + 1,
			cfg = paste0(
				"dominoes_",
				rep(rev(c("black", "red", "green", "blue", "yellow", "white")), 2)
			),
			default.units = "in",
			envir = envir,
			op_scale = 0.5
		)
	})
	expect_error(xya_pips_cards(11), "Don't know pip pattern for 11 pips")
	expect_error(xya_pips_dominoes(19), "Don't know pip pattern for 19 pips")
	for (i in 0:10) {
		expect_equal(nrow(xya_pips_cards(i)), i)
	}

	# checkers
	expect_doppelganger("checkers", function() {
		df2 <- tibble(
			piece_side = "board_face",
			x = c(4, 6),
			y = c(4, 6),
			suit = c(2, 4),
			rank = c(4, 2),
			cfg = "checkers2"
		)
		df1f <- tibble(
			piece_side = "board_face",
			x = c(1, 2),
			y = c(1, 6),
			suit = c(1, 5),
			rank = c(2, 4),
			cfg = "checkers1"
		)
		df1b <- tibble(
			piece_side = "board_back",
			x = 5,
			y = 1,
			suit = 1,
			rank = 2,
			cfg = "checkers1"
		)
		df_bit1 <- tibble(
			piece_side = "bit_back",
			x = rep_len(1, 6L),
			y = rep_len(1, 6L),
			suit = 1:6,
			rank = 1:6,
			cfg = "checkers1"
		)
		df_bit2 <- tibble(
			piece_side = "bit_face",
			x = 2:7,
			y = 2:7,
			suit = 6:1,
			rank = 1:6,
			cfg = "checkers1"
		)
		df_bit3 <- tibble(
			piece_side = "bit_back",
			x = 5,
			y = 7,
			suit = 4L,
			rank = 4L,
			cfg = "checkers2"
		)
		df <- rbind(df2, df1f, df1b, df_bit3, df_bit2, df_bit1)
		df_bit4 <- tibble(
			piece_side = "bit_face",
			x = 7,
			y = 5,
			suit = 2L,
			rank = 4L,
			cfg = "checkers2"
		)
		df <- rbind(df2, df1f, df1b, df_bit4, df_bit3, df_bit2, df_bit1)
		pmap_piece(df, default.units = "in", envir = envir, op_scale = 0.5, trans = op_transform)
	})

	# chess
	# df <- tibble(piece_side = "bit_face", x = rep(1:6, 6L), y = rep(6:1, each=6L),
	#              rank = x, suit = y, cfg = "chess1")
	# pmap_piece(df, default.units="in", envir=envir, op_scale=0.5, trans=op_transform)
	# df <- tibble(piece_side = "bit_face", x = rep(seq.int(1, by=2, length.out=6L), 6L), y = rep(6:1, each=6L),
	#              rank = x, suit = y, cfg = "chess2")
	# pmap_piece(df, default.units="in", envir=envir, op_scale=0.5, trans=op_transform)

	# joystick pawns
	cfg <- game_systems(pawn = "joystick")$piecepack
	df <- tibble(piece_side = "pawn_top", x = 1:4, y = 1:4, suit = 1:4)
	expect_doppelganger("joystick", function() {
		pmap_piece(df, default.units = "in", cfg = cfg)
	})
	expect_doppelganger("joystick-op", function() {
		pmap_piece(df, default.units = "in", cfg = cfg, op_scale = 0.5)
	})

	# go
	expect_doppelganger("go", function() {
		dfb <- tibble(piece_side = "board_face", x = 10, y = 10, suit = 2)
		dfB <- tibble(
			piece_side = "bit_face",
			x = c(2, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 6, 7, 7, 8),
			y = c(11, 13, 14, 12, 15, 13, 16, 12, 16, 12, 13, 16, 15, 17, 16),
			suit = 4
		)
		dfR <- tibble(
			piece_side = "bit_face",
			x = c(3, 3, 4, 5, 6, 7, 8, 8, 9),
			y = c(2:3, 4, 3, 4, 4, 2:3, 2),
			suit = 1
		)
		dfY <- tibble(
			piece_side = "bit_face",
			x = 20 - c(3, 4, 5, 5, 7, 8, 9, 9, 9),
			y = 20 - c(3, 2, 2:3, 2, 5, 2:4),
			suit = 5
		)
		dfG <- tibble(
			piece_side = "bit_face",
			x = 20 - c(3, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 6, 6),
			y = c(4, 5, 7, 12, 13, 6, 7, 10, 11, 7, 9, 10, 8, 9),
			suit = 3
		)
		dfK <- tibble(
			piece_side = "bit_face",
			x = c(
				rep(2, 5),
				rep(3, 5),
				rep(4, 4),
				rep(5, 5),
				rep(6, 2),
				rep(7, 4),
				rep(8, 7),
				rep(9, 5),
				rep(10, 5),
				rep(11, 4),
				rep(12, 5),
				rep(13, 6),
				rep(14, 2),
				rep(15, 6),
				rep(16, 6),
				rep(17, 2)
			),
			y = c(
				2:3,
				8:9,
				12,
				4:5,
				10,
				13:14,
				5,
				10,
				14:15,
				6,
				8,
				10:11,
				14,
				5,
				11,
				5,
				7,
				12,
				14,
				4,
				7,
				9,
				11:13,
				15,
				3:4,
				11,
				14,
				16,
				8:9,
				14:16,
				3,
				6,
				8,
				15,
				9:11,
				16,
				18,
				6,
				8:9,
				11:12,
				16,
				6:7,
				5:6,
				8,
				11:12,
				16,
				3,
				5,
				12,
				14,
				16:17,
				3,
				16
			),
			suit = 2
		)
		dfW <- tibble(
			piece_side = "bit_face",
			x = c(
				3:5,
				rep(6, 3),
				rep(7, 3),
				rep(8, 2),
				rep(9, 5),
				rep(10, 3),
				rep(11, 4),
				rep(12, 4),
				rep(13, 2),
				14
			),
			y = c(
				9,
				9,
				9,
				7:8,
				10,
				8,
				10:11,
				8,
				10,
				8:10,
				12:13,
				10:11,
				13,
				9,
				11,
				13:14,
				7:8,
				12:13,
				7,
				10,
				11
			),
			suit = 6
		)
		df <- rbind(dfb, dfB, dfR, dfY, dfG, dfK, dfW)

		cfg <- game_systems("sans")$go
		pmap_piece(
			df,
			grid.piece,
			cfg = cfg,
			trans = op_transform,
			default.units = "in",
			op_scale = 0.5
		)
	})

	# playing cards
	cfg <- game_systems("sans", border = FALSE)$playing_cards_colored
	expect_doppelganger("ten-of-clubs", function() {
		grid.piece("card_face", suit = 3, rank = 10, cfg = cfg, default.units = "npc")
	})
	expect_doppelganger("king-of-stars", function() {
		suppressMessages(
			{
				grid.piece("card_face", suit = 5, rank = 13, cfg = cfg, default.units = "npc")
			},
			classes = "piecepackr_fill_stroke"
		)
	})
	expect_doppelganger("red-joker", function() {
		grid.piece("card_face", suit = 1, rank = 14, cfg = cfg, default.units = "npc")
	})
	cfg <- game_systems("sans", border = FALSE)$playing_cards_tarot
	expect_doppelganger("black-joker-no-star", function() {
		grid.piece("card_face", suit = 3, rank = 15, cfg = cfg, default.units = "npc")
	})
	expect_doppelganger("knight-of-diamonds", function() {
		grid.piece("card_face", suit = 4, rank = 12, cfg = cfg, default.units = "npc")
	})

	# morris
	cfg <- envir$morris
	expect_doppelganger("two-morris", function() {
		grid.piece("board_face", x = 4, y = 4, suit = 3, rank = 2, cfg = cfg, default.units = "in")
		grid.piece("bit_back", x = 0:5, y = 0:5, suit = 1:6, cfg = cfg, default.units = "in")
	})
	expect_doppelganger("three-morris", function() {
		grid.piece("board_face", x = 4, y = 4, suit = 3, rank = 3, cfg = cfg, default.units = "in")
		grid.piece("bit_back", x = 0:5, y = 0:5, suit = 1:6, cfg = cfg, default.units = "in")
	})
	expect_doppelganger("six-morris", function() {
		grid.piece("board_face", x = 4, y = 4, suit = 3, rank = 6, cfg = cfg, default.units = "in")
		grid.piece("bit_back", x = 0:5, y = 0:5, suit = 1:6, cfg = cfg, default.units = "in")
	})
	expect_doppelganger("seven-morris", function() {
		grid.piece("board_face", x = 4, y = 4, suit = 3, rank = 7, cfg = cfg, default.units = "in")
		grid.piece("bit_back", x = 0:5, y = 0:5, suit = 1:6, cfg = cfg, default.units = "in")
	})
	expect_doppelganger("nine-morris", function() {
		grid.piece("board_face", x = 4, y = 4, suit = 3, rank = 9, cfg = cfg, default.units = "in")
		grid.piece("bit_back", x = 0:5, y = 0:5, suit = 1:6, cfg = cfg, default.units = "in")
	})
	expect_doppelganger("twelve-morris", function() {
		grid.piece("board_face", x = 4, y = 4, suit = 3, rank = 12, cfg = cfg, default.units = "in")
		grid.piece("bit_back", x = 0:5, y = 0:5, suit = 1:6, cfg = cfg, default.units = "in")
	})

	# alquerque
	cfg <- envir$alquerque
	expect_doppelganger("alquerque", function() {
		grid.piece("board_face", x = 3, y = 3, suit = 3, cfg = cfg, default.units = "in")
		grid.piece("bit_back", x = 1:5, y = 1:5, suit = 1:5, cfg = cfg, default.units = "in")
	})

	# reversi
	expect_doppelganger("reversi", function() {
		envir <- game_systems()
		df_board <- tibble(
			piece_side = "board_face",
			x = 4.5,
			y = 4.5,
			suit = 4,
			rank = 8,
			cfg = "reversi"
		)
		df_bit_face <- tibble(
			piece_side = "bit_face",
			x = 1:6,
			y = 1:6,
			suit = 1:6,
			rank = NA,
			cfg = "reversi"
		)
		df_bit_back <- tibble(
			piece_side = "bit_back",
			x = 1:6 + 1L,
			y = 1:6,
			suit = 1:6,
			rank = NA,
			cfg = "reversi"
		)
		df <- rbind(df_board, df_bit_face, df_bit_back)
		pmap_piece(df, default.units = "in", envir = envir, op_scale = 0.5, trans = op_transform)
	})

	marbles_test <- function(..., round = FALSE, shading = FALSE, f = grid.piece) {
		cur_seed <- .Random.seed
		envir <- game_systems(round = round, shading = shading)
		on.exit(.Random.seed <<- cur_seed, add = TRUE)
		set.seed(42)
		dfb <- data.frame(
			piece_side = "board_face",
			x = 2.5,
			y = 2.5,
			suit = 4L,
			rank = 4L,
			cfg = "marbles"
		)
		dfm <- data.frame(
			piece_side = "bit_back",
			x = c(rep(1:4, 4L), rep(0.5 + rep(1:3, 3L)), rep(2:3, 2L), 2.5),
			y = c(rep(1:4, each = 4L), rep(0.5 + rep(1:3, each = 3L)), rep(2:3, each = 2L), 2.5),
			suit = sample.int(6L, 30L, replace = TRUE),
			rank = 9L,
			cfg = "marbles"
		)
		df <- rbind(dfb, dfm)
		pmap_piece(df, f, envir = envir, ..., default.units = "in")
	}
	# marbles
	expect_doppelganger("marbles", function() {
		marbles_test()
	})
	expect_doppelganger("marbles-op", function() {
		marbles_test(op_scale = 0.5, trans = marbles_transform)
	})

	skip_if_not_installed("svglite", "2.2.2")
	skip_if_not(getRversion() >= "4.2.0")
	write_svg_bleeding_edge <- function(plot, file, title = "") {
		svglite::svglite(file)
		plot()
		invisible(grDevices::dev.off())
	}

	# standard d4 dice
	cfg <- envir$dice_d4
	expect_doppelganger(
		"dice-d4",
		function() {
			grid.piece(
				"die_face",
				x = 1:4,
				y = 1,
				default.units = "in",
				rank = 1:4,
				suit = 1:4,
				op_scale = 0.0,
				cfg = cfg
			)
			grid.piece(
				"die_face",
				x = 1:4,
				y = 2,
				default.units = "in",
				rank = 1:4,
				suit = 1:4,
				op_scale = 0.5,
				cfg = cfg
			)
		},
		writer = write_svg_bleeding_edge
	)

	# standard d8 dice
	cfg <- envir$dice_d8
	expect_doppelganger(
		"dice-d8",
		function() {
			grid.piece(
				"die_face",
				x = 1:8,
				y = 1,
				default.units = "in",
				rank = 1:8,
				suit = c(1:6, 1:2),
				op_scale = 0.0,
				cfg = cfg
			)
			grid.piece(
				"die_face",
				x = 1:8,
				y = 2,
				default.units = "in",
				rank = 1:8,
				suit = c(1:6, 1:2),
				op_scale = 0.5,
				cfg = cfg
			)
		},
		writer = write_svg_bleeding_edge
	)

	# standard d10 dice
	cfg <- envir$dice_d10
	expect_doppelganger(
		"dice-d10",
		function() {
			grid.piece(
				"die_face",
				x = rep(1:5, 2),
				y = rep(1:2, each = 5),
				default.units = "in",
				rank = 1:10,
				suit = rep(1:5, 2),
				op_scale = 0.5,
				cfg = cfg
			)
		},
		writer = write_svg_bleeding_edge
	)

	# percentile d10 dice
	cfg <- envir$dice_d10_percentile
	expect_doppelganger(
		"dice-d10-percentile",
		function() {
			grid.piece(
				"die_face",
				x = rep(1:5, 2),
				y = rep(1:2, each = 5),
				default.units = "in",
				rank = 1:10,
				suit = rep(1:5, 2),
				op_scale = 0.5,
				cfg = cfg
			)
		},
		writer = write_svg_bleeding_edge
	)

	# standard d12 dice
	cfg <- envir$dice_d12
	expect_doppelganger(
		"dice-d12",
		function() {
			grid.piece(
				"die_face",
				x = rep(1:6, 2),
				y = rep(1:2, each = 6),
				default.units = "in",
				rank = 1:12,
				suit = rep(1:6, 2),
				op_scale = 0.5,
				cfg = cfg
			)
		},
		writer = write_svg_bleeding_edge
	)

	# standard d20 dice
	cfg <- envir$dice_d20
	expect_doppelganger(
		"dice-d20",
		function() {
			grid.piece(
				"die_face",
				x = rep(1:5, 4),
				y = rep(1:4, each = 5),
				default.units = "in",
				rank = 1:20,
				suit = rep(1:6, length.out = 20),
				op_scale = 0.5,
				cfg = cfg
			)
		},
		writer = write_svg_bleeding_edge
	)

	# Chinese dominoes
	expect_doppelganger(
		"dominoes-chinese",
		function() {
			df1 <- tibble(
				piece_side = "tile_face",
				suit = c(rep(1, 6L), 2L, rep(2, 4L), rep(3, 3L), 3, rep(4, 3), 5, 5, 6),
				rank = c(1:6, 2L, 3:6, 3:5, 6, 4:6, 5:6, 6),
				cfg = c(rep(
					c("dominoes_chinese", "dominoes_chinese_black", "dominoes_chinese"),
					each = 7L
				)),
				x = rep(1:7, 3L),
				y = c(rep(c(1.5, 4.0, 6.5), each = 7L))
			)
			df2 <- tibble(
				piece_side = "die_face",
				suit = 1,
				rank = 1:6,
				cfg = rep(
					c("dominoes_chinese", "dominoes_chinese_black", "dominoes_chinese"),
					each = 2L
				),
				x = 8.5,
				y = c(1, 2.25, 3.5, 4.75, 6.0, 7.25)
			)
			df <- rbind(df1, df2)
			pmap_piece(
				df,
				envir = game_systems(),
				default.units = "in",
				trans = op_transform,
				op_scale = 0.5
			)
		},
		writer = write_svg_bleeding_edge
	)

	# king of stars (fill + stroke)
	cfg <- game_systems("sans", border = FALSE)$playing_cards_colored
	expect_doppelganger(
		"king-of-stars-fill-stroke",
		function() {
			grid.piece("card_face", suit = 5, rank = 13, cfg = cfg, default.units = "npc")
		},
		writer = write_svg_bleeding_edge
	)

	expect_doppelganger(
		"marbles-shading",
		function() {
			marbles_test(op_scale = 0.5, shading = TRUE, trans = marbles_transform)
		},
		writer = write_svg_bleeding_edge
	)

	# cubes
	expect_doppelganger("cubes", function() {
		grid.piece(
			"bit_face",
			x = rep(6:1, 6),
			y = rep(6:1, each = 6),
			suit = rep(1:6, each = 6),
			rank = rep(6:1, 6),
			default.units = "in",
			op_scale = 0.5,
			cfg = envir$cubes
		)
	})
})

# Get xyz value for vertices

die_xyz <- function(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth) {
	xs <- c(0, 0, 1, 1, 0, 0, 1, 1) - 0.5
	ys <- c(1, 0, 0, 1, 1, 0, 0, 1) - 0.5
	zs <- rep(c(1, 0), each = 4) - 0.5

	dfi <- get_die_face_info(suit, cfg$die_arrangement)
	dR <- get_die_rotation(suit, rank, cfg)
	R <- dR %*% AA_to_R(angle, axis_x, axis_y)

	as_coord3d(xs, ys, zs)$scale(width, height, depth)$transform(R)$translate(x, y, z)
}

rounded_die_xyz <- function(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth) {
	stopifnot(width == height, width == depth)
	opt <- cfg$get_piece_opt("die_face", suit, rank)
	# Sphere radius
	# 3D Euclidean distance to (0.5 - opt$shape_r, 0.5, 0.5)
	s <- sqrt((0.5 - opt$shape_r)^2 + 0.50)
	xy <- expand.grid(x = seq(0.0, 1.0, 0.05), y = seq(0.0, 1.0, 0.05))
	xy <- xy[which(xy$x^2 + xy$y^2 <= 1), ]
	nz <- sqrt(1 - xy$x^2 - xy$y^2)
	ppp <- data.frame(x = s * xy$x, y = s * xy$y, z = s * nz)
	ppn <- data.frame(x = s * xy$x, y = s * xy$y, z = -s * nz)
	pnp <- data.frame(x = s * xy$x, y = -s * xy$y, z = s * nz)
	pnn <- data.frame(x = s * xy$x, y = -s * xy$y, z = -s * nz)
	nnp <- data.frame(x = -s * xy$x, y = -s * xy$y, z = s * nz)
	npp <- data.frame(x = -s * xy$x, y = s * xy$y, z = s * nz)
	npn <- data.frame(x = -s * xy$x, y = s * xy$y, z = -s * nz)
	nnn <- data.frame(x = -s * xy$x, y = -s * xy$y, z = -s * nz)
	df <- rbind(ppp, ppn, pnp, pnn, nnp, npp, npn, nnn)
	# Flatten: clamp coords beyond each die face to the face boundary
	df$x <- pmin(pmax(df$x, -0.5), 0.5)
	df$y <- pmin(pmax(df$y, -0.5), 0.5)
	df$z <- pmin(pmax(df$z, -0.5), 0.5)
	dR <- get_die_rotation(suit, rank, cfg)
	R <- dR %*% AA_to_R(angle, axis_x, axis_y)
	as_coord3d(df)$scale(width, height, depth)$transform(R)$translate(x, y, z)
}

ellipse_xyz <- function() {
	xy <- expand.grid(x = seq(0.0, 1.0, 0.05), y = seq(0.0, 1.0, 0.05))
	xy <- xy[which(xy$x^2 + xy$y^2 <= 1), ]
	z <- sqrt(1 - xy$x^2 - xy$y^2)
	ppp <- data.frame(x = xy$x, y = xy$y, z = z)
	ppn <- data.frame(x = xy$x, y = xy$y, z = -z)
	pnp <- data.frame(x = xy$x, y = -xy$y, z = z)
	pnn <- data.frame(x = xy$x, y = -xy$y, z = -z)
	nnp <- data.frame(x = -xy$x, y = -xy$y, z = z)
	npp <- data.frame(x = -xy$x, y = xy$y, z = z)
	npn <- data.frame(x = -xy$x, y = xy$y, z = -z)
	nnn <- data.frame(x = -xy$x, y = -xy$y, z = -z)
	df <- 0.5 * rbind(ppp, ppn, pnp, pnn, nnp, npp, npn, nnn)
	as_coord3d(df)
}


# pyramid top
pt_xyz <- function(x, y, z, angle, axis_x, axis_y, width, height, depth) {
	xy <- as_coord2d(rect_xy)$translate(-0.5, -0.5)
	xyz_t <- as_coord3d(x = 0.0, y = 0.0, z = 0.5)
	xyz_b <- as_coord3d(xy, z = -0.5)
	xs <- c(xyz_t$x, xyz_b$x)
	ys <- c(xyz_t$y, xyz_b$y)
	zs <- c(xyz_t$z, xyz_b$z)
	as_coord3d(xs, ys, zs)$scale(width, height, depth)$transform(AA_to_R(
		angle,
		axis_x,
		axis_y
	))$translate(as_coord3d(x, y, z))
}

# pyramid (on its) side
ps_xyz <- function(x, y, z, angle, axis_x, axis_y, width, height, depth) {
	xy <- as_coord2d(pyramid_xy)$translate(-0.5, -0.5)
	xyz_b <- as_coord3d(xy, z = -0.5)
	theta <- 2 * asin(0.5 * width / height)
	xyz_t <- as_coord3d(x = c(-0.5, 0.5), y = 0.5 - cos(theta), z = 0.5)
	xs <- c(xyz_t$x, xyz_b$x)
	ys <- c(xyz_t$y, xyz_b$y)
	zs <- c(xyz_t$z, xyz_b$z)
	as_coord3d(xs, ys, zs)$scale(width, height, depth)$transform(AA_to_R(
		angle,
		axis_x,
		axis_y
	))$translate(x, y, z)
}

#' Piece Grob Functions
#'
#' `basicPieceGrob()` is the most common
#' \dQuote{grob} function that [grid.piece()] uses
#' to create `grid` graphical `grob` objects.
#' `picturePieceGrobFn()` is a function that returns a \dQuote{grob} function
#' that imports graphics from files found in its `directory` argument.
#'
#' @rdname basicPieceGrobs
#' @name basicPieceGrobs
#' @param cfg Piecepack configuration list or [pp_cfg()] object.
#' @inheritParams grid.piece
#' @examples
#'
#'  if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {
#'    cfg <- pp_cfg(list(grob_fn.tile=basicPieceGrob, invert_colors=TRUE))
#'    grid.piece("tile_face", suit=1, rank=3, cfg=cfg)
#'  }
#'
#'  \donttest{# May take more than 5 seconds on CRAN servers
#'  try({
#'    if (requireNamespace("grid", quietly = TRUE) && capabilities(c("cairo"))) {
#'      cfg <- pp_cfg(list(grob_fn.tile=basicPieceGrob, invert_colors=TRUE))
#'      directory <- tempdir()
#'      save_piece_images(cfg, directory=directory, format="svg", angle=0)
#'      cfg2 <- pp_cfg(list(grob_fn=picturePieceGrobFn(directory)))
#'
#'      grid::grid.newpage()
#'      grid.piece("coin_back", suit=3, rank=5, cfg=cfg2)
#'    }
#'  })
#'  }
#' @export
basicPieceGrob <- function(piece_side, suit, rank, cfg = pp_cfg()) {
	cfg <- as_pp_cfg(cfg)
	opt <- cfg$get_piece_opt(piece_side, suit, rank)
	gTree(
		opt = opt,
		border = TRUE,
		flip = FALSE,
		scale = 1,
		fill_stroke = FALSE,
		name = NULL,
		gp = gpar(),
		vp = NULL,
		cl = "basic_piece_side"
	)
}

basicPieceGrobFn <- function(fill_stroke = FALSE) {
	force(fill_stroke)
	function(piece_side, suit, rank, cfg = pp_cfg()) {
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)
		gTree(
			opt = opt,
			border = TRUE,
			flip = FALSE,
			scale = 1,
			fill_stroke = fill_stroke,
			name = NULL,
			gp = gpar(),
			vp = NULL,
			cl = "basic_piece_side"
		)
	}
}

ellipsoidGrobFn <- function(shading = FALSE) {
	force(shading)
	function(piece_side, suit, rank, cfg = pp_cfg()) {
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)

		gTree(
			opt = opt,
			border = TRUE,
			flip = FALSE,
			scale = 1,
			shading = shading,
			name = NULL,
			gp = gpar(),
			vp = NULL,
			cl = c("basic_ellipsoid", "basic_piece_side")
		)
	}
}

#' @export
makeContext.basic_piece_side <- function(x) {
	x <- update_gp(x, gp = gpar(cex = x$scale, lex = x$scale))
	x
}

#' @export
grobCoords.basic_piece_side <- function(x, closed, ...) {
	opt <- x$opt
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)
	grobCoords(shape$shape(vp = x$vp), closed = closed, ...)
}

#' @export
makeContent.basic_piece_side <- function(x) {
	opt <- x$opt
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)
	# Possibly shrink background and gridlines to not overlap mat
	# which sometimes prevents visual glitch if no border line
	# but do not do this if mat color is transparent.
	if (any(as.logical(grDevices::col2rgb(opt$mat_color, alpha = TRUE)[4, ] < 255))) {
		bg_mat_width <- 0
	} else {
		bg_mat_width <- opt$mat_width
	}

	# Background
	background_grob <- shape$shape(
		gp = gpar(col = NA, fill = opt$background_color),
		name = "background",
		mat_width = bg_mat_width
	)
	gl_grob <- shape$gridlines(
		gp = gpar(col = opt$gridline_color, lex = opt$gridline_lex),
		name = "gridlines",
		mat_width = bg_mat_width
	)
	gp_mat <- gpar(col = NA, lwd = 0, fill = opt$mat_color)
	mat_grob <- shape$mat(opt$mat_width, gp = gp_mat, name = "mat")

	# Primary symbol
	gp_ps <- gpar(
		col = opt$ps_color,
		fontsize = opt$ps_fontsize,
		fontfamily = opt$ps_fontfamily,
		fontface = opt$ps_fontface
	)
	ps_grob <- textGrob(
		opt$ps_text,
		x = opt$ps_x,
		y = opt$ps_y,
		hjust = 0.5,
		vjust = 0.5,
		gp = gp_ps,
		name = "primary_symbol"
	)
	if (isTRUE(x$fill_stroke)) {
		ps_grob <- as_fill_stroke_grob(ps_grob, fill = opt$ps_color)
	}

	# Directional mark
	gp_dm <- gpar(
		col = opt$dm_color,
		fontsize = opt$dm_fontsize,
		fontfamily = opt$dm_fontfamily,
		fontface = opt$ps_fontface
	)
	dm_grob <- textGrob(
		opt$dm_text,
		x = opt$dm_x,
		y = opt$dm_y,
		hjust = 0.5,
		vjust = 0.5,
		gp = gp_dm,
		name = "directional_mark"
	)
	if (isTRUE(x$fill_stroke)) {
		dm_grob <- as_fill_stroke_grob(dm_grob, fill = opt$dm_color)
	}

	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- shape$shape(gp = gp_border, name = "border")
	} else {
		border_grob <- nullGrob(name = "border")
	}
	if (x$flip) {
		gl <- gList(dm_grob, ps_grob, mat_grob, gl_grob, background_grob, border_grob)
	} else {
		gl <- gList(background_grob, gl_grob, mat_grob, ps_grob, dm_grob, border_grob)
	}

	setChildren(x, gl)
}

#' @export
makeContent.basic_ellipsoid <- function(x) {
	opt <- x$opt
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	gp_background <- gpar(col = NA, fill = opt$background_color)
	background_grob <- shape$shape(gp = gp_background, name = "background")

	if (x$shading) {
		if (has_radial_gradients()) {
			rgr <- radialGradient(
				c(update_alpha_col(opt$background_color, 0.500), "#00000080"),
				r1 = 0.1
			)
			gp_gr <- gpar(col = opt$border_color, fill = rgr)
			shading_grob <- shape$shape(gp = gp_gr, name = "shading")
		} else {
			rgr_inform()
			shading_grob <- nullGrob(name = "shading")
		}
	} else {
		shading_grob <- nullGrob(name = "shading")
	}

	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- shape$shape(gp = gp_border, name = "border")
	} else {
		border_grob <- nullGrob(name = "border")
	}
	if (x$flip) {
		# Do anything with shading?
		gl <- gList(background_grob, border_grob)
	} else {
		gl <- gList(background_grob, shading_grob, border_grob)
	}

	setChildren(x, gl)
}

#' @rdname basicPieceGrobs
#' @param directory Directory that `picturePieceGrobFn` will look in for piece graphics.
#' @param filename_fn Function that takes arguments `directory`, `piece_side`, `suit`,
#'        `rank`, and optionally `cfg` and returns the (full path) filename of the image that the
#'        function returned by `picturePieceGrobFn` should import.
#' @export
picturePieceGrobFn <- function(directory, filename_fn = find_pp_file) {
	function(piece_side, suit, rank, cfg) {
		if (hasName(formals(filename_fn), "cfg")) {
			f <- filename_fn(directory, piece_side, suit, rank, cfg)
		} else {
			f <- filename_fn(directory, piece_side, suit, rank)
		}
		file2grob(f)
	}
}

find_pp_file <- function(directory, piece_side, suit, rank, cfg) {
	for (format in c("svgz", "svg", "png", "jpg", "jpeg")) {
		f <- piece_filename_helper(directory, piece_side, format, suit, rank, cfg)
		if (file.exists(f)) {
			return(f)
		}
	}
	abort(paste("Couldn't find suitable", piece_side, "image in", directory))
}

piece_filename_helper <- function(directory, piece_side, format, suit, rank, cfg) {
	if (piece_side %in% intersect(cfg$lacks_rank, cfg$lacks_suit)) {
		piece_filename(directory, piece_side, format, 0)
	} else if (piece_side %in% cfg$lacks_rank) {
		piece_filename(directory, piece_side, format, 0, suit = suit)
	} else if (piece_side %in% cfg$lacks_suit) {
		piece_filename(directory, piece_side, format, 0, rank = rank)
	} else {
		piece_filename(directory, piece_side, format, 0, suit = suit, rank = rank)
	}
}

#' @rdname basicPieceGrobs
#' @export
pyramidTopGrob <- function(piece_side, suit, rank, cfg = pp_cfg()) {
	gTree(suit = suit, rank = rank, cfg = cfg, scale = 1, cl = "pyramid_top")
}

#' @export
makeContent.pyramid_top <- function(x) {
	width <- convertWidth(unit(1, "npc"), "in", valueOnly = TRUE)
	height <- convertHeight(unit(1, "npc"), "in", valueOnly = TRUE)

	# face at top
	xy_vp <- list(x = c(width, width, 0, 0), y = c(0.5 * height, height, height, 0.5 * height))
	xy_polygon <- list(x = c(0.5 * height, width, 0), y = c(0.5 * height, height, height))
	g1 <- at_ps_grob("pyramid_face", x$suit, x$rank, x$cfg, xy_vp, xy_polygon, "face")
	g1$scale <- x$scale

	# back at bottom
	xy_vp <- list(x = width * rect_xy$x, y = 0.5 * height * rect_xy$y)
	xy_polygon <- list(x = width * pyramid_xy$x, y = 0.5 * height * pyramid_xy$y)
	g2 <- at_ps_grob("pyramid_back", x$suit, x$rank, x$cfg, xy_vp, xy_polygon, "back")
	g2$scale <- x$scale

	# left at left
	xy_vp <- list(x = c(0.5 * width, 0, 0, 0.5 * width), y = c(height, height, 0, 0))
	xy_polygon <- list(x = c(0.5 * width, 0, 0), y = c(0.5 * height, height, 0))
	g3 <- at_ps_grob("pyramid_left", x$suit, x$rank, x$cfg, xy_vp, xy_polygon, "left")
	g3$scale <- x$scale

	# right at right
	xy_vp <- list(x = c(0.5 * width, width, width, 0.5 * width), y = c(0, 0, height, height))
	xy_polygon <- list(x = c(0.5 * width, width, width), y = c(0.5 * height, 0, height))
	g4 <- at_ps_grob("pyramid_right", x$suit, x$rank, x$cfg, xy_vp, xy_polygon, "right")
	g4$scale <- x$scale

	gl <- gList(g1, g2, g3, g4)

	setChildren(x, gl)
}

#' @export
grobCoords.pyramid_top <- function(x, closed, ...) {
	grobCoords(rectGrob(vp = x$vp), closed = closed, ...)
}

#' @rdname basicPieceGrobs
#' @export
previewLayoutGrob <- function(piece_side, suit, rank, cfg = pp_cfg()) {
	cfg <- as_pp_cfg(cfg)

	t_width <- cfg$get_width("tile_face")
	c_width <- cfg$get_width("coin_face")
	d_width <- cfg$get_width("die_face")
	s_width <- cfg$get_width("saucer_face")
	p_height <- cfg$get_height("pawn_face")
	preview_width <- 3 * t_width

	gl <- gList()
	cs_tiles <- rep("tile_face", 6)
	suit_tiles <- 1:6
	rank_tiles <- rep(2, 6)
	if (cfg$n_suits < 5) {
		cs_tiles[5] <- "tile_back"
		suit_tiles[5] <- NA
		rank_tiles[5] <- NA
	}
	if (cfg$n_suits < 6) {
		cs_tiles[6] <- "tile_back"
		suit_tiles[6] <- NA
		rank_tiles[6] <- NA
	}
	x_tiles <- c(1, 5, 5, 1, 3, 3)
	y_tiles <- c(5, 5, 3, 3, 5, 3)
	gl[["tiles"]] <- pieceGrob(
		cs_tiles,
		suit_tiles,
		rank_tiles,
		cfg,
		x_tiles,
		y_tiles,
		default.units = "in",
		name = "tiles"
	)

	cs_coins <- rep(c("coin_face", "coin_back"), each = 3)
	suit_coins <- c(rep(NA, 3), 4:2)
	rank_coins <- c(1:3, rep(NA, 3))
	x_coins <- rep(1:3, 2) * c_width - 0.5 * c_width
	y_coins <- 0.5 * t_width + rep(c(0.5, -0.5), each = 3) * c_width
	if (cfg$n_suits > 4) {
		suit_coins[5] <- 5
	}
	if (cfg$n_suits > 5) {
		suit_coins[6] <- 6
	}

	gl[["coins"]] <- pieceGrob(
		cs_coins,
		suit_coins,
		rank_coins,
		cfg,
		x_coins,
		y_coins,
		default.units = "in",
		name = "coins"
	)

	gl[["saucers"]] <- pieceGrob(
		c("saucer_face", "saucer_back"),
		c(1, NA),
		NA,
		cfg,
		t_width + 1.5 * d_width,
		0.5 * t_width + c(0.5, -0.5) * s_width,
		default.units = "in",
		name = "saucers"
	)

	gl[["pawns"]] <- pieceGrob(
		c("pawn_face", "pawn_back"),
		2,
		NA,
		cfg,
		t_width + 3 * d_width,
		0.5 * t_width + c(0.5, -0.5) * p_height,
		angle = c(0, 180),
		default.units = "in",
		name = "pawns"
	)

	gl[["suitrankdie"]] <- pieceGrob(
		"suitrankdie_layoutRF",
		NA,
		NA,
		cfg,
		preview_width - 2 * d_width,
		0.5 * t_width,
		default.units = "in",
		name = "suitrankdie"
	)

	gTree(children = gl, cl = "preview_layout")
}

dieLayoutGrobRF <- function(piece_side, suit, rank, cfg) {
	piecepackDieGrob(suit, cfg)
}

dieLayoutGrobLF <- function(piece_side, suit, rank, cfg) {
	piecepackDieGrob(suit, cfg, flip = TRUE)
}

suitdieLayoutGrobRF <- function(piece_side, suit, rank, cfg) {
	suitdieGrob(cfg)
}

suitdieLayouGrobtLF <- function(piece_side, suit, rank, cfg) {
	suitdieGrob(cfg, flip = TRUE)
}

suitrankdieLayoutGrobRF <- function(piece_side, suit, rank, cfg) {
	suit <- get_suit_die_suits(cfg)
	piecepackDieGrob(suit, cfg)
}
suitrankdieLayoutGrobLF <- function(piece_side, suit, rank, cfg) {
	suit <- get_suit_die_suits(cfg)
	piecepackDieGrob(suit, cfg, flip = TRUE)
}

## Layout functions
x_die_layoutRF <- c(1 / 4, 2 / 4, 2 / 4, 3 / 4, 3 / 4, 4 / 4) - 1 / 8
x_die_layoutLF <- c(4 / 4, 3 / 4, 3 / 4, 2 / 4, 2 / 4, 1 / 4) - 1 / 8
y_die_layout <- c(1 / 3, 1 / 3, 2 / 3, 2 / 3, 3 / 3, 3 / 3) - 1 / 6

piecepackDieGrob <- function(suit, cfg, flip = FALSE, arrangement = cfg$die_arrangement) {
	cfg <- as_pp_cfg(cfg)
	rs <- get_die_face_info(suit, arrangement)
	angle <- rep(c(0, -90), 3)
	if (flip) {
		x <- x_die_layoutLF
		angle <- -angle
	} else {
		x <- x_die_layoutRF
	}
	angle <- angle + rs$angle
	gl <- gList()
	for (ii in 1:6) {
		gl[[ii]] <- pieceGrob(
			"die_face",
			rs$suit[ii],
			rs$rank[ii],
			cfg,
			x = x[ii],
			y = y_die_layout[ii],
			default.units = "npc",
			width = 1 / 4,
			height = 1 / 3,
			angle = angle[ii]
		)
	}
	gTree(children = gl, name = "die_layout")
}

suitdieGrob <- function(cfg, flip = FALSE) {
	angle <- rep(c(-90, 0), 3)
	if (flip) {
		x <- x_die_layoutLF
		angle <- -angle
	} else {
		x <- x_die_layoutRF
	}
	suit <- get_suit_die_suits(cfg)
	gl <- gList()
	for (ii in 1:6) {
		vp <- viewport(x = x[ii], y = y_die_layout[ii], angle = angle[ii])
		gl[[ii]] <- pieceGrob("suitdie_face", suit[ii], NA, cfg, default.units = "npc", vp = vp)
	}
	gTree(children = gl, name = "suitdie_layout")
}

get_suit_die_suits <- function(cfg) {
	n_suits <- cfg$n_suits
	if (n_suits < 4) {
		rep(n_suits:1, length.out = 6)
	} else if (n_suits == 4) {
		c(5, 6, 4:1)
	} else if (n_suits > 4) {
		6:1
	}
}

pawnLayoutGrob <- function(piece_side, suit, rank, cfg) {
	gl <- gList()
	suppressWarnings({
		ph <- cfg$get_height("pawn_face")
		pb <- 3 / 7 * ph
		denominator <- 2 * (ph + pb)
		yf <- 0.5 * ph / denominator
		height <- ph / denominator
		gl[[1]] <- pieceGrob(
			"pawn_face",
			suit,
			NA,
			cfg,
			y = 1 / 2 - yf,
			height = height,
			default.units = "npc",
			name = "pawn_face"
		)
		gl[[2]] <- pieceGrob(
			"pawn_back",
			suit,
			NA,
			cfg,
			y = 1 / 2 + yf,
			height = height,
			angle = 180,
			default.units = "npc",
			name = "pawn_back"
		)
	})
	opt <- cfg$get_piece_opt("pawn_face", suit, 0)
	gp <- gpar(col = opt$border_color, fill = NA)
	gl[[3]] <- linesGrob(y = 0.5, gp = gpar(col = opt$border_color, fill = NA, lty = "dashed"))
	gl[[4]] <- rectGrob(gp = gp)
	ll <- 0.07
	gl[[5]] <- segmentsGrob(0.5, 0, 0.5, ll, gp = gp)
	gl[[6]] <- segmentsGrob(0.5, 1, 0.5, 1 - ll, gp = gp)
	gTree(children = gl, name = "pawn_layout")
}

pyramidLayoutGrob <- function(piece_side, suit, rank, cfg) {
	w <- cfg$get_width("pyramid_face", rank = rank)
	h <- cfg$get_height("pyramid_face", rank = rank)
	d <- sqrt(h^2 + (0.5 * w)^2)
	t <- c(72, 36, 0, -36, -72)
	r <- h
	x <- 0.5 * to_x(t, r)
	y <- d + 0.5 * to_y(t, r)
	pieces <- c("pyramid_face", "pyramid_right", "pyramid_back", "pyramid_left", "pyramid_face")
	angles <- c(90 + 72, 90 + 36, 90, 90 - 36, 90 - 72)
	df <- tibble(piece_side = pieces, x = x, y = y, suit = suit, rank = rank, angle = angles)
	g <- pmap_piece(df, cfg = cfg, default.units = "inches", draw = FALSE)
	grobTree(g, name = "pyramid_layout")
}

# pieces that are composites of other pieces
# earlier pieces (in df) are "beneath" later pieces (as viewed from `ref_side`)
CompositePiece <- R6Class(
	"pp_composite",
	public = list(
		initialize = function(df = tibble(), envir = list(), ref_side = "top") {
			private$df <- df
			private$envir <- envir
			private$ref_side <- ref_side
		}
	),
	active = list(
		grob_fn = function() {
			function(piece_side, suit, rank, cfg) {
				df <- private$relative_df(piece_side)
				pmap_piece(df, suit = suit, envir = private$envir, draw = FALSE)
			}
		},
		obj_fn = function() {
			function(
				piece_side,
				suit,
				rank,
				cfg,
				x,
				y,
				z,
				angle,
				axis_x,
				axis_y,
				width,
				height,
				depth,
				filename,
				res
			) {
				suit <- ifelse(is.na(suit), 1, suit)
				rank <- ifelse(is.na(rank), 1, rank)
				if (is.na(angle)) {
					angle <- 0
				}
				if (is.na(axis_x)) {
					axis_x <- 0
				}
				if (is.na(axis_y)) {
					axis_y <- 0
				}
				if (is.na(width)) {
					width <- cfg$get_width(piece_side, suit, rank)
				}
				if (is.na(height)) {
					height <- cfg$get_height(piece_side, suit, rank)
				}
				if (is.na(depth)) {
					depth <- cfg$get_depth(piece_side, suit, rank)
				}
				if (is.na(z)) {
					z <- 0.5 * depth
				}

				df <- private$transform_df(
					piece_side,
					x,
					y,
					z,
					angle,
					width,
					height,
					depth,
					axis_x,
					axis_y
				)
				ext <- tools::file_ext(filename)
				filename <- gsub(paste0("\\.", ext, "$"), paste0("_%03d.", ext), filename)
				filename <- rep(filename, length.out = nrow(df))
				filename <- sprintf(filename, seq_along(filename))
				stopifnot(!anyDuplicated(filename))
				df$filename <- filename
				l <- pmap_piece(
					df,
					suit = suit,
					envir = private$envir,
					.f = save_piece_obj,
					res = res
				)
				do.call(rbind, l)
			}
		},
		op_grob_fn = function() {
			function(
				piece_side,
				suit,
				rank,
				cfg,
				x,
				y,
				z,
				angle,
				type,
				width,
				height,
				depth,
				op_scale,
				op_angle,
				scale = 1
			) {
				x <- convertX(x, "in", valueOnly = TRUE)
				y <- convertY(y, "in", valueOnly = TRUE)
				z <- convertX(z, "in", valueOnly = TRUE)
				width <- convertX(width, "in", valueOnly = TRUE)
				height <- convertY(height, "in", valueOnly = TRUE)
				depth <- convertX(depth, "in", valueOnly = TRUE)
				df <- private$relative_df(piece_side)
				relative_side <- get_relative_side(piece_side, private$ref_side)
				df <- scale_df(df, relative_side, width, height, depth)
				df <- translate_df(df, relative_side, x, y, z, angle, width, height, depth)
				# adjust scale for proper adjustment of `cex` / `lex`
				df <- adjust_scale_df(df, scale = scale)
				#### in future need to re-order elements using `op_sort()`?
				pmap_piece(
					df,
					suit = suit,
					envir = private$envir,
					draw = FALSE,
					default.units = "in",
					op_scale = op_scale,
					op_angle = op_angle,
					scale = scale
				)
			}
		}
	),
	private = list(
		df = NULL,
		envir = NULL,
		ref_side = NULL,
		relative_df = function(piece_side) {
			side = get_relative_side(piece_side, private$ref_side)
			switch(
				side,
				face = private$df,
				back = back_df(private$df),
				abort(paste("`CompositePiece()$relative_df()` can't handle relative side", side))
			)
		},
		transform_df = function(
			piece_side,
			x,
			y,
			z,
			angle,
			width,
			height,
			depth,
			axis_x = 0,
			axis_y = 0,
			scale = 1
		) {
			side <- get_side(piece_side)
			relative_side <- get_relative_side(piece_side, private$ref_side)

			df <- scale_df(private$df, relative_side, width, height, depth, scale)
			whd <- get_scaling_factors(
				relative_side,
				width = scale * width,
				height = scale * height,
				depth = scale * depth
			)
			R <- side_R_rev(private$ref_side) %*% side_R(side) %*% AA_to_R(angle, axis_x, axis_y)
			xyz <- as_coord3d(x = df$x, y = df$y, z = df$z)$translate(-0.5, -0.5, -0.5)$scale(
				whd$width,
				whd$height,
				whd$depth
			)$transform(R)$translate(x, y, z)
			df$x <- xyz$x
			df$y <- xyz$y
			df$z <- xyz$z
			#### Doesn't yet handle cases where components in df have their own angle, axis_x, axis_y values
			df <- cbind(df, R_to_AA(R))
			df
		}
	)
)

# Get "relative" side of `piece_side` given "reference" `ref_side`
get_relative_side <- function(piece_side, ref_side = "top") {
	side <- get_side(piece_side)
	switch(
		ref_side,
		face = switch(
			side,
			face = "face",
			right = "right",
			back = "back",
			left = "left",
			top = "top",
			base = "base"
		),
		right = switch(
			side,
			face = "left",
			right = "face",
			back = "right",
			left = "back",
			top = "top",
			base = "base"
		),
		back = switch(
			side,
			face = "back",
			right = "left",
			back = "face",
			left = "right",
			top = "top",
			base = "base"
		),
		left = switch(
			side,
			face = "right",
			right = "back",
			back = "left",
			left = "face",
			top = "top",
			base = "base"
		),
		top = switch(
			side,
			face = "top",
			right = "right",
			back = "base",
			left = "left",
			top = "face",
			base = "back"
		),
		base = switch(
			side,
			face = "top",
			right = "left",
			back = "base",
			left = "right",
			top = "back",
			base = "face"
		)
	)
}

scale_df <- function(df, relative_side, width, height, depth, scale = 1) {
	whd <- get_scaling_factors(
		relative_side,
		width = scale * width,
		height = scale * height,
		depth = scale * depth
	)
	df$height <- whd$height * df$height
	df$width <- whd$width * df$width
	df$depth <- whd$depth * df$depth
	df
}

# Simple case with no `axis_x` or `axis_y`
translate_df <- function(df, relative_side, x, y, z, angle, width, height, depth, scale = 1) {
	whd <- get_scaling_factors(
		relative_side,
		width = scale * width,
		height = scale * height,
		depth = scale * depth
	)
	R <- R_z(angle)
	xyz <- as_coord3d(x = df$x, y = df$y, z = df$z)$translate(-0.5, -0.5, -0.5)$scale(
		whd$width,
		whd$height,
		whd$depth
	)$transform(R)$translate(x, y, z)
	df$x <- xyz$x
	df$y <- xyz$y
	df$z <- xyz$z
	df$angle <- angle
	df
}

adjust_scale_df <- function(df, scale = 1) {
	df$height <- df$height / scale
	df$width <- df$width / scale
	df$depth <- df$depth / scale
	df
}

back_df = function(df) {
	df$z <- 1 - df$z
	pieces <- sapply(strsplit(df$piece_side, "_"), function(x) x[1])
	sides <- sapply(strsplit(df$piece_side, "_"), function(x) x[2])
	opposites <- sapply(sides, function(side) {
		switch(
			side,
			face = "back",
			back = "face",
			left = "right",
			right = "left",
			top = "base",
			base = "top"
		)
	})
	df$piece_side <- paste(pieces, opposites, sep = "_")
	rev_df(df)
}
rev_df <- function(df) {
	if (nrow(df) > 1L) {
		df[nrow(df):1, ]
	} else {
		df
	}
}

# We'll use the alpha-90-beta-90 angle kites as described in
# "Mathematical Analysis of Trapezohedron with 2n congruent right kite faces" by Harish Chandra Rajpoot
# alpha = to_degrees(2 * atan(sqrt(tanpi(0.2) * tanpi(0.1)))) ~= 51.8272923729877490473
# beta = 180 - alpha ~= 128.1727076270122438473
# a = sqrt(tanpi(0.2) * tanpi(0.1)) * b ~= 0.48586827175664565326 * b := m_b * b
# inner radius = b/2 => die "depth" = b
#
# https://en.wikipedia.org/wiki/Right_kite
# diagonal that is a line of symmetry
# p = sqrt(a^2 + b^2) = sqrt(1 + m_b^2) * b
# p := 5 / 8 = 0.625
# => b = 0.625 / sqrt(1 + m_b^2) ~= 0.5621585749837085810299
# => a = m_b * b ~= 0.2731350152805132047718
# other diagonal
# q = 2ab / sqrt(a^2 + b^2) = (2 * m_b / sqrt(1 + m_b^2)) * b ~= 0.4913446110983896164548
# segments extending from intersection of diagonal to vertices in clockwise order d1, d2, d3, d4 then
# d1 * d3 = d2 * d4 = (p - d3) * d3 = (q / 2) * (q / 2) => d3^2 - p * d3 + q^2 / 4 = 0
# => d3 = (p - sqrt(p^2 - q^2)) / 2   OR  d3 = (p + sqrt(p^2 - q^2)) / 2  (latter seems too big!)
# => d3 ~ 0.1193643785156578984719 => d3 / p = 0.1909830056250526320039
# => kite "radius" = 0.5 - d3 / p ~= 0.3090169943749473402406
#
# Other links
# https://rechneronline.de/pi/right-kite.php
# http://www.dicecollector.com/JM/D10.HTM
# https://rpg.stackexchange.com/questions/114672/is-there-a-standard-d10-dimension-ratio
# https://math.stackexchange.com/questions/2464000/pentagonal-trapezohedron-with-face-perpendicular-to-side
# https://www.youtube.com/watch?v=eYO19tYYqXk
# https://www.athenopolis.net/2018/03/the-dimensions-of-ten-sided-die.html

d10TopGrob <- function(
	piece_side,
	suit,
	rank,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = unit(0, "npc"),
	angle = 0,
	type = "normal",
	width = NA,
	height = NA,
	depth = NA,
	op_scale = 0,
	op_angle = 45
) {
	cfg <- as_pp_cfg(cfg)
	opt <- cfg$get_piece_opt("die_face", suit, rank)

	x <- convertX(x, "in", valueOnly = TRUE)
	y <- convertY(y, "in", valueOnly = TRUE)
	z <- convertX(z, "in", valueOnly = TRUE)
	width <- convertX(width, "in", valueOnly = TRUE)
	height <- convertY(height, "in", valueOnly = TRUE)
	depth <- convertX(depth, "in", valueOnly = TRUE)

	#### allow limited 3D rotation #281
	axis_x <- 0
	axis_y <- 0

	xyz <- d10_xyz(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth)
	l_xyz <- list()
	dfc <- data.frame(x = numeric(0), y = numeric(0), z = numeric(0))
	for (i in 1:10) {
		idx <- seq.int(4 * i - 3, length.out = 4)
		dfc <- rbind(dfc, as.data.frame(mean(xyz[idx])))
	}

	dfc$edge <- paste0(
		"d10_",
		c(
			"face",
			"left1",
			"left2",
			"right1",
			"right2",
			"opposite",
			"opposite_left1",
			"opposite_left2",
			"opposite_right1",
			"opposite_right2"
		)
	)
	dfc$rank <- vapply(dfc$edge, d10_edge_rank, numeric(1), rank = rank, USE.NAMES = FALSE)
	dfc$idx <- 1:10
	dfc <- dfc[
		painter_order(as_coord3d(dfc$x, dfc$y, dfc$z), scale = op_scale, alpha = degrees(op_angle)),
	]

	gl <- gList()
	fn_idx <- function(idx) {
		switch(idx, 1:4, 5:8, 9:12, 13:16, 17:20, 21:24, 25:28, 29:32, 33:36, 37:40)
	}
	if (nigh(opt$shape_t, 90)) {
		idx_vp <- 1:4
	} else if (nigh(opt$shape_t, 0)) {
		idx_vp <- c(2:4, 1)
	} else if (nigh(opt$shape_t, 180)) {
		idx_vp <- c(4, 1:3)
	} else if (nigh(opt$shape_t, 270)) {
		idx_vp <- c(3:4, 1:2)
	}
	for (i in 1:nrow(dfc)) {
		edge_rank <- dfc$rank[i]
		opt <- cfg$get_piece_opt("die_face", suit, edge_rank)
		gp <- gpar(col = opt$border_color, lex = opt$border_lex, fill = opt$background_color)
		xyz_polygon <- xyz[fn_idx(dfc$idx[i])]
		xy_polygon <- as_coord2d(xyz_polygon, alpha = degrees(op_angle), scale = op_scale)
		xy_vp <- xy_vp_kite(xyz_polygon, op_scale, op_angle)[idx_vp]
		gl[[i]] <- at_ps_grob(
			"die_face",
			suit,
			edge_rank,
			cfg,
			xy_vp,
			xy_polygon,
			name = dfc$edge[i]
		)
	}

	# pre-compute grobCoords
	coords_xyl <- as.list(convex_hull2d((xy_polygon)))

	gTree(
		scale = 1,
		coords_xyl = coords_xyl,
		children = gl,
		cl = c("projected_rpg_die", "coords_xyl")
	)
}

# Assumes "acute" angle on top (`theta = 90`)
xy_vp_kite <- function(xyz_polygon, op_scale, op_angle) {
	p_mid <- mean(xyz_polygon[c(2, 4)])
	p_left <- xyz_polygon[2]
	p_right <- xyz_polygon[4]

	up_diff <- xyz_polygon[1] - p_mid
	down_diff <- xyz_polygon[3] - p_mid
	p_ul <- p_left + up_diff
	p_ur <- p_right + up_diff
	p_ll <- p_left + down_diff
	p_lr <- p_right + down_diff

	x <- c(p_ul$x, p_ll$x, p_lr$x, p_ur$x)
	y <- c(p_ul$y, p_ll$y, p_lr$y, p_ur$y)
	z <- c(p_ul$z, p_ll$z, p_lr$z, p_ur$z)

	as_coord2d(as_coord3d(x, y, z), alpha = degrees(op_angle), scale = op_scale)
}

d10_xyz <- function(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth) {
	opt <- cfg$get_piece_opt("die_face", suit, rank)
	xy <- kite_xy(r = 0.5 - 0.1909830056250526320039)
	if (nigh(opt$shape_t, 90) || nigh(opt$shape_t, 270)) {
		xyz_t <- as_coord3d(xy$x, xy$y, 0.0)$translate(as_coord3d(-0.5, -0.5, 0.5))$scale(
			width,
			height,
			depth
		)
		xyz_b <- as_coord3d(xy$x, xy$y, 0.0)$translate(as_coord3d(-0.5, -0.5, -0.5))$rotate(
			"z-axis",
			degrees(180)
		)$scale(width, height, depth)
	} else {
		xyz_t <- as_coord3d(xy$x, xy$y, z = 0.0)$translate(as_coord3d(-0.5, -0.5, 0.5))$scale(
			height,
			width,
			depth
		)
		xyz_b <- as_coord3d(xy$x, xy$y, z = 0.0)$translate(as_coord3d(-0.5, -0.5, -0.5))$rotate(
			"z-axis",
			degrees(180)
		)$scale(height, width, depth)
	}
	xyz_b <- xyz_b[c(1, 4, 3, 2)]

	xyz_l2 <- kite_xyz_missing_right(xyz_t[1], xyz_b[3], xyz_b[2]) # left2 face
	xyz_r2 <- kite_xyz_missing_left(xyz_t[1], xyz_b[4], xyz_b[3]) # right2 face
	xyz_ol2 <- kite_xyz_missing_right(xyz_b[1], xyz_t[3], xyz_t[2]) # opposite left2 face
	xyz_or2 <- kite_xyz_missing_left(xyz_b[1], xyz_t[4], xyz_t[3]) # opposite right2 face

	# top face
	x_face <- xyz_t$x
	y_face <- xyz_t$y
	z_face <- xyz_t$z

	# left1
	x_left1 <- c(xyz_l2$x[c(1, 4)], xyz_ol2$x[4], xyz_t$x[2])
	y_left1 <- c(xyz_l2$y[c(1, 4)], xyz_ol2$y[4], xyz_t$y[2])
	z_left1 <- c(xyz_l2$z[c(1, 4)], xyz_ol2$z[4], xyz_t$z[2])

	# left2
	x_left2 <- xyz_l2$x
	y_left2 <- xyz_l2$y
	z_left2 <- xyz_l2$z

	# right1
	x_right1 <- c(xyz_t$x[c(1, 4)], xyz_or2$x[2], xyz_r2$x[2])
	y_right1 <- c(xyz_t$y[c(1, 4)], xyz_or2$y[2], xyz_r2$y[2])
	z_right1 <- c(xyz_t$z[c(1, 4)], xyz_or2$z[2], xyz_r2$z[2])

	# right2
	x_right2 <- xyz_r2$x
	y_right2 <- xyz_r2$y
	z_right2 <- xyz_r2$z

	# opposite (bottom)
	x_opposite <- xyz_b$x
	y_opposite <- xyz_b$y
	z_opposite <- xyz_b$z

	# opposite left1
	x_opposite_left1 <- c(xyz_ol2$x[c(1, 4)], xyz_l2$x[4], xyz_b$x[2])
	y_opposite_left1 <- c(xyz_ol2$y[c(1, 4)], xyz_l2$y[4], xyz_b$y[2])
	z_opposite_left1 <- c(xyz_ol2$z[c(1, 4)], xyz_l2$z[4], xyz_b$z[2])

	# opposite left2
	x_opposite_left2 <- xyz_ol2$x
	y_opposite_left2 <- xyz_ol2$y
	z_opposite_left2 <- xyz_ol2$z

	# opposite right1
	x_opposite_right1 <- c(xyz_b$x[c(1, 4)], xyz_r2$x[2], xyz_or2$x[2])
	y_opposite_right1 <- c(xyz_b$y[c(1, 4)], xyz_r2$y[2], xyz_or2$y[2])
	z_opposite_right1 <- c(xyz_b$z[c(1, 4)], xyz_r2$z[2], xyz_or2$z[2])

	# opposite right2
	x_opposite_right2 <- xyz_or2$x
	y_opposite_right2 <- xyz_or2$y
	z_opposite_right2 <- xyz_or2$z

	xs <- c(
		x_face,
		x_left1,
		x_left2,
		x_right1,
		x_right2,
		x_opposite,
		x_opposite_left1,
		x_opposite_left2,
		x_opposite_right1,
		x_opposite_right2
	)
	ys <- c(
		y_face,
		y_left1,
		y_left2,
		y_right1,
		y_right2,
		y_opposite,
		y_opposite_left1,
		y_opposite_left2,
		y_opposite_right1,
		y_opposite_right2
	)
	zs <- c(
		z_face,
		z_left1,
		z_left2,
		z_right1,
		z_right2,
		z_opposite,
		z_opposite_left1,
		z_opposite_left2,
		z_opposite_right1,
		z_opposite_right2
	)

	pc <- as_coord3d(x, y, z)
	R <- R_z(opt$shape_t - 90) %*% AA_to_R(angle, axis_x, axis_y)
	as_coord3d(xs, ys, zs)$transform(R)$translate(pc)
}

kite_xyz_missing_right <- function(p_top, p_left, p_bottom) {
	p_mid <- p_bottom + 0.1909830056250526320039 * (p_top - p_bottom)
	p_right <- p_mid + (p_mid - p_left)
	as_coord3d(
		x = c(p_top$x, p_left$x, p_bottom$x, p_right$x),
		y = c(p_top$y, p_left$y, p_bottom$y, p_right$y),
		z = c(p_top$z, p_left$z, p_bottom$z, p_right$z)
	)
}

kite_xyz_missing_left <- function(p_top, p_bottom, p_right) {
	p_mid <- p_bottom + 0.1909830056250526320039 * (p_top - p_bottom)
	p_left <- p_mid + (p_mid - p_right)
	as_coord3d(
		x = c(p_top$x, p_left$x, p_bottom$x, p_right$x),
		y = c(p_top$y, p_left$y, p_bottom$y, p_right$y),
		z = c(p_top$z, p_left$z, p_bottom$z, p_right$z)
	)
}

save_d10_obj <- function(
	piece_side,
	suit,
	rank,
	cfg,
	x,
	y,
	z,
	angle,
	axis_x,
	axis_y,
	width,
	height,
	depth,
	filename,
	scale,
	res
) {
	cfg <- as_pp_cfg(cfg)
	opt <- cfg$get_piece_opt(piece_side, suit, rank)

	xyz <- d10_xyz(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth)

	xy <- kite_xy(r = 0.5 - 0.1909830056250526320039)
	xy_vt <- list(
		x = rep(xy$x / 2, 10) + rep(rep(0:1 / 2, each = 4), 5),
		y = rep(xy$y / 5, 10) + rep(rep(4:0 / 5, each = 4), each = 2)
	)

	# textured face elements: face, left1, left2, right1, right2
	#                      opposite, opposite left, opposite right, opposite back
	f <- list()
	fn_vt <- function(r) switch(r, 1:4, 5:8, 9:12, 13:16, 17:20, 21:24, 25:28, 29:32, 33:36, 37:40)
	f[[1]] <- list(v = 1:4, vt = fn_vt(rank))
	f[[2]] <- list(v = 5:8, vt = fn_vt(d10_edge_rank("d10_left1", rank)))
	f[[3]] <- list(v = 9:12, vt = fn_vt(d10_edge_rank("d10_left2", rank)))
	f[[4]] <- list(v = 13:16, vt = fn_vt(d10_edge_rank("d10_right1", rank)))
	f[[5]] <- list(v = 17:20, vt = fn_vt(d10_edge_rank("d10_right2", rank)))
	f[[6]] <- list(v = 21:24, vt = fn_vt(d10_edge_rank("d10_opposite", rank)))
	f[[7]] <- list(v = 25:28, vt = fn_vt(d10_edge_rank("d10_opposite_left1", rank)))
	f[[8]] <- list(v = 29:32, vt = fn_vt(d10_edge_rank("d10_opposite_left2", rank)))
	f[[9]] <- list(v = 33:36, vt = fn_vt(d10_edge_rank("d10_opposite_right1", rank)))
	f[[10]] <- list(v = 37:40, vt = fn_vt(d10_edge_rank("d10_opposite_right2", rank)))

	ext <- tools::file_ext(filename)
	mtl_filename <- gsub(paste0("\\.", ext, "$"), ".mtl", filename)
	png_filename <- gsub(paste0("\\.", ext, "$"), ".png", filename)

	write_obj(filename, v = xyz, vt = xy_vt, f = f)
	write_d10_texture("die_face", suit, rank, cfg, filename = png_filename, res = res)

	invisible(list(obj = filename, mtl = mtl_filename, png = png_filename))
}

d10_edge_rank <- function(edge, rank) {
	switch(
		edge,
		d10_face = rank,
		d10_left1 = d10_rank(rank + 2),
		d10_left2 = d10_rank(rank + 4),
		d10_right1 = d10_rank(rank + 8),
		d10_right2 = d10_rank(rank + 6),
		d10_opposite = 11 - rank,
		d10_opposite_left1 = d10_rank(11 - rank + 2),
		d10_opposite_left2 = d10_rank(11 - rank + 4),
		d10_opposite_right1 = d10_rank(11 - rank + 8),
		d10_opposite_right2 = d10_rank(11 - rank + 6)
	)
}

d10_rank <- function(x) {
	(x - 1) %% 10 + 1
}

write_d10_texture <- function(
	piece_side = "die_face",
	suit = 1,
	rank = 1,
	cfg = pp_cfg(),
	...,
	filename = tempfile(fileext = ".png"),
	res = 72
) {
	current_dev <- grDevices::dev.cur()
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	opt <- cfg$get_piece_opt(piece_side, suit, rank)
	height <- cfg$get_height("die_face", suit, rank)
	width <- cfg$get_width("die_face", suit, rank)

	if (nigh(opt$shape_t, 90) || nigh(opt$shape_t, 270)) {
		args <- list(
			filename = filename,
			height = 5 * height,
			width = 2 * width,
			units = "in",
			res = res,
			bg = "transparent"
		)
	} else {
		args <- list(
			filename = filename,
			height = 5 * width,
			width = 2 * height,
			units = "in",
			res = res,
			bg = "transparent"
		)
	}
	angle <- 90 - opt$shape_t
	if (capabilities("cairo")) {
		args$type <- "cairo"
	}
	do.call(grDevices::png, args)

	xs <- rep(1:2 / 2 - 1 / 4, 5)
	ys <- rep(5:1 / 5 - 1 / 10, each = 2)
	for (i in 1:10) {
		pushViewport(viewport(
			x = xs[i],
			width = inch(width),
			y = ys[i],
			height = inch(height),
			angle = angle
		))
		draw_piece_and_bleed("die_face", suit, i, cfg)
		popViewport()
	}

	grDevices::dev.off()
	invisible(filename)
}

#' Shape object for generating various grobs
#'
#' `pp_shape()` creates an R6 object with methods for creating various shape based grobs.
#'
#' `pp_shape` objects serve the following purposes:\enumerate{
#'  \item{Make it easier for developers to customize game piece appearances
#'        either through a "grob_fn" or "op_grob_fn" styles in `pp_cfg()`
#'        or manipulate a piece post drawing via functions like `grid::grid.edit()`.}
#'  \item{Used internally to generate `piecepackr`'s built-in game piece grobs.}
#'   }
#'
#' @section `pp_shape` R6 Class Method Arguments:\describe{
#' \item{`mat_width`}{Numeric vector of mat widths.}
#' \item{`clip`}{\dQuote{clip grob} to perform polyclip operation with.
#'                    See [gridGeometry::grid.polyclip()] for more info.}
#' \item{`op`}{Polyclip operation to perform.
#'                  See [gridGeometry::grid.polyclip()] for more info.}
#' \item{`pattern`}{Pattern to fill in shape with.
#'                  See [gridpattern::patternGrob()] for more info.}
#' \item{`...`}{Passed to `gridpattern::patternGrob()`.}
#' \item{`name`}{Grid grob `name` value.}
#' \item{`gp`}{Grid `gpar` list.  See [grid::gpar()] for more info.}
#' \item{`vp`}{Grid viewport or `NULL`.}
#' }
#' @section `pp_shape` R6 Class Methods:\describe{
#' \item{`checkers(name = NULL, gp = gpar(), vp = NULL)`}{Returns a grob of checkers for that shape.}
#' \item{`gridlines(name = NULL, gp = gpar(), vp = NULL)`}{Returns a grob of gridlines for that shape.}
#' \item{`hexlines(name = NULL, gp = gpar(), vp = NULL)`}{Returns a grob of hexlines for that shape.}
#' \item{`mat(mat_width = 0, name = NULL, gp = gpar(), vp = NULL)`}{
#'    Returns a grob for a matting \dQuote{mat} for that shape.}
#' \item{`pattern(pattern = "stripe", ..., name = NULL, gp = gpar(), vp = NULL)`}{
#'       Fills in the shape's `npc_coords` with a pattern.
#'       See [gridpattern::patternGrob()] for more information.
#' }
#' \item{`polyclip(clip, op = "intersection", name = NULL, gp = gpar(), vp = NULL)`}{
#'    Returns a grob that is an \dQuote{intersection}, \dQuote{minus}, \dQuote{union}, or \dQuote{xor} of another grob.
#'    Note unlike `gridGeometry::polyclipGrob` it can directly work with a `pieceGrob` "clip grob" argument.}
#' \item{`shape(name = NULL, gp = gpar(), vp = NULL)`}{Returns a grob of the shape.}
#' }
#' @section `pp_shape` R6 Class Active Bindings:\describe{
#' \item{`label`}{The shape's label.}
#' \item{`theta`}{The shape's theta.}
#' \item{`radius`}{The shape's radius.}
#' \item{`back`}{A boolean of whether this is the shape's \dQuote{back} side.}
#' \item{`npc_coords`}{A named list of \dQuote{npc} coordinates along the perimeter of the shape.}
#' }
#' @param label Label of the shape.  One of \describe{
#'      \item{\dQuote{circle}}{Circle.}
#'      \item{\dQuote{convexN}}{An `N`-sided convex polygon.
#'                              `theta` controls which direction the first vertex is drawn.}
#'      \item{\dQuote{concaveN}}{A \dQuote{star} (concave) polygon with `N` \dQuote{points}.
#'                              `theta` controls which direction the first point is drawn.
#'                              `radius` controls the distance of the \dQuote{inner} vertices from the center.}
#'      \item{\dQuote{halma}}{A 2D outline of a \dQuote{Halma pawn}.}
#'      \item{\dQuote{kite}}{\dQuote{Kite} quadrilateral shape.}
#'      \item{\dQuote{meeple}}{A 2D outline of a \dQuote{meeple}.}
#'      \item{\dQuote{oval}}{Oval.}
#'      \item{\dQuote{pyramid}}{An \dQuote{Isosceles} triangle whose base is the bottom of the viewport.
#'                              Typically used to help draw the face of the \dQuote{pyramid} piece.}
#'      \item{\dQuote{rect}}{Rectangle.}
#'      \item{\dQuote{roundrect}}{\dQuote{Rounded} rectangle.  `radius` controls curvature of corners.}
#' }
#' @param theta `convex` and `concave` polygon shapes
#'                    use this to determine where the first point is drawn.
#' @param radius `concave` polygon and `roundrect` use this
#'                     to control appearance of the shape.
#' @param back Whether the shape should be reflected across a vertical line in the middle of the viewport.
#' @param ... Should be empty
#' @param width,height Width and height of the piece.
#' @examples
#'  if (require("grid", quietly = TRUE)) {
#'      gp <- gpar(col="black", fill="yellow")
#'      rect <- pp_shape(label="rect")
#'      convex6 <- pp_shape(label="convex6")
#'      circle <- pp_shape(label="circle")
#'
#'      pushViewport(viewport(x=0.25, y=0.75, width=1/2, height=1/2))
#'      grid.draw(rect$shape(gp=gp))
#'      grid.draw(rect$gridlines(gp=gpar(col="blue", lex=4)))
#'      grid.draw(rect$hexlines(gp=gpar(col="green")))
#'      popViewport()
#'
#'      pushViewport(viewport(x=0.75, y=0.75, width=1/2, height=1/2))
#'      grid.draw(convex6$shape(gp=gp))
#'      grid.draw(convex6$checkers(gp=gpar(fill="blue")))
#'      popViewport()
#'
#'      pushViewport(viewport(x=0.25, y=0.25, width=1/2, height=1/2))
#'      grid.draw(circle$shape(gp=gp))
#'      grid.draw(circle$mat(mat_width=0.2, gp=gpar(fill="blue")))
#'      popViewport()
#'
#'      pushViewport(viewport(x=0.75, y=0.25, width=1/2, height=1/2))
#'      grid.draw(rect$shape(gp=gp))
#'      grid.draw(rect$mat(mat_width=c(0.2, 0.1, 0.3, 0.4), gp=gpar(fill="blue")))
#'      popViewport()
#'  }
#'  if (require("grid", quietly = TRUE)) {
#'      grid.newpage()
#'      gp <- gpar(col="black", fill="yellow")
#'
#'      vp <- viewport(x=1/4, y=1/4, width=1/2, height=1/2)
#'      grid.draw(pp_shape("halma")$shape(gp=gp, vp=vp))
#'      vp <- viewport(x=3/4, y=1/4, width=1/2, height=1/2)
#'      grid.draw(pp_shape("pyramid")$shape(gp=gp, vp=vp))
#'      vp <- viewport(x=3/4, y=3/4, width=1/2, height=1/2)
#'      grid.draw(pp_shape("kite")$shape(gp=gp, vp=vp))
#'      vp <- viewport(x=1/4, y=3/4, width=1/2, height=1/2)
#'      grid.draw(pp_shape("meeple")$shape(gp=gp, vp=vp))
#'  }
#'  if (require("grid", quietly = TRUE)) {
#'      grid.newpage()
#'      vp <- viewport(x=1/4, y=1/4, width=1/2, height=1/2)
#'      grid.draw(pp_shape("convex3", 0)$shape(gp=gp, vp=vp))
#'      vp <- viewport(x=3/4, y=1/4, width=1/2, height=1/2)
#'      grid.draw(pp_shape("convex4", 90)$shape(gp=gp, vp=vp))
#'      vp <- viewport(x=3/4, y=3/4, width=1/2, height=1/2)
#'      grid.draw(pp_shape("convex5", 180)$shape(gp=gp, vp=vp))
#'      vp <- viewport(x=1/4, y=3/4, width=1/2, height=1/2)
#'      grid.draw(pp_shape("convex6", 270)$shape(gp=gp, vp=vp))
#'  }
#'  if (require("grid", quietly = TRUE)) {
#'      grid.newpage()
#'      vp <- viewport(x=1/4, y=1/4, width=1/2, height=1/2)
#'      grid.draw(pp_shape("concave3", 0, 0.1)$shape(gp=gp, vp=vp))
#'      vp <- viewport(x=3/4, y=1/4, width=1/2, height=1/2)
#'      grid.draw(pp_shape("concave4", 90, 0.2)$shape(gp=gp, vp=vp))
#'      vp <- viewport(x=3/4, y=3/4, width=1/2, height=1/2)
#'      grid.draw(pp_shape("concave5", 180, 0.3)$shape(gp=gp, vp=vp))
#'      vp <- viewport(x=1/4, y=3/4, width=1/2, height=1/2)
#'      grid.draw(pp_shape("concave6", 270)$shape(gp=gp, vp=vp))
#'  }
#'  if (require("grid", quietly = TRUE) &&
#'      requireNamespace("gridpattern", quietly = TRUE)) {
#'      grid.newpage()
#'      hex <- pp_shape("convex6")
#'      gp <- gpar(fill = c("blue", "yellow", "red"), col = "black")
#'      grid.draw(hex$pattern("polygon_tiling", gp = gp, spacing = 0.1,
#'                            type = "truncated_trihexagonal"))
#'      gp <- gpar(fill = "black", col = NA)
#'      grid.draw(hex$mat(mat_width = 0.025, gp = gp))
#'  }
#' @export
pp_shape <- function(
	label = "rect",
	theta = 90,
	radius = 0.2,
	back = FALSE,
	...,
	width = 1,
	height = 1
) {
	check_dots_empty()
	Shape$new(label, theta, radius, back, width, height)
}

Shape <- R6Class(
	"pp_shape",
	public = list(
		initialize = function(
			label = "rect",
			theta = 90,
			radius = 0.2,
			back = FALSE,
			width = 1,
			height = 1
		) {
			if (!is_known_shape_label(label)) {
				abort(paste("Don't recognize shape label", label))
			}
			if (is_angle(theta)) {
				theta <- as.double(theta, "degrees")
			}
			private$shape_label <- label
			private$shape_theta <- theta
			private$shape_radius <- radius
			private$shape_back <- back
			private$shape_width <- width
			private$shape_height <- height
		},
		shape = function(name = NULL, gp = gpar(), vp = NULL, mat_width = 0) {
			private$shape_grob_fn(mat_width = mat_width)(name = name, gp = gp, vp = vp)
		},
		gridlines = function(name = NULL, gp = gpar(), vp = NULL, mat_width = 0) {
			gTree(
				shape = self,
				mat_width = mat_width,
				name = name,
				gp = gp,
				vp = vp,
				cl = "gridlines"
			)
		},
		mat = function(mat_width = 0, name = NULL, gp = gpar(), vp = NULL) {
			gTree(mat_width = mat_width, shape = self, name = name, gp = gp, vp = vp, cl = "mat")
		},
		checkers = function(name = NULL, gp = gpar(), vp = NULL) {
			gTree(shape = self, name = name, gp = gp, vp = vp, cl = "checkers")
		},
		hexlines = function(name = NULL, gp = gpar(), vp = NULL) {
			gTree(shape = self, name = name, gp = gp, vp = vp, cl = "hexlines")
		},
		pattern = function(pattern = "stripe", ..., name = NULL, gp = gpar(), vp = NULL) {
			assert_suggested("gridpattern")
			xy <- self$npc_coords
			gridpattern::patternGrob(pattern, xy$x, xy$y, ..., name = name, gp = gp, vp = vp)
		},
		polyclip = function(clip, op = "intersection", name = NULL, gp = gpar(), vp = NULL) {
			gTree(
				shape = self,
				clip = clip,
				op = op,
				name = name,
				gp = gp,
				vp = vp,
				cl = "pp_polyclip"
			)
		}
	),
	active = list(
		back = function() private$shape_back,
		convex = function() {
			if (
				grepl(self$label, "^concave") ||
					self$label %in% c("halma", "meeple")
			) {
				FALSE
			} else {
				TRUE
			}
		},
		label = function() private$shape_label,
		npc_coords = function() {
			label <- self$label
			theta <- self$theta
			radius <- self$radius
			back <- self$back
			width <- self$width
			height <- self$height
			if (label == "rect") {
				rect_xy
			} else if (grepl(label, "circle|oval")) {
				xy <- convex_xy(72, 90) #### increase number of vertices?
				xy$c <- rep("C1", length(xy$x))
				xy
			} else if (label == "kite") {
				if (back) {
					theta <- 180 - theta
				}
				kite_xy(theta, radius)
			} else if (label == "halma") {
				halma_xy()
			} else if (label == "meeple") {
				meeple_xy
			} else if (label == "pyramid") {
				pyramid_xy
			} else if (label == "roundrect") {
				roundrect_xy(radius, width, height)
			} else if (grepl("^concave", label)) {
				if (back) {
					theta <- 180 - theta
				}
				concave_xy(get_n_vertices(label), theta, radius)
			} else if (grepl("^convex", label)) {
				if (back) {
					theta <- 180 - theta
				}
				convex_xy(get_n_vertices(label), theta)
			}
		},
		radius = function() private$shape_radius,
		width = function() private$shape_width,
		height = function() private$shape_height,
		theta = function() private$shape_theta
	),
	private = list(
		shape_label = NULL,
		shape_theta = NULL,
		shape_radius = NULL,
		shape_back = NULL,
		shape_width = NULL,
		shape_height = NULL,
		shape_grob_fn = function(mat_width = 0) {
			label <- self$label
			theta <- self$theta
			back <- self$back
			if (label == "rect") {
				rectGrobFn(mat_width)
			} else if (label == "circle") {
				circleGrobFn(r = 0.5 - mat_width[1])
			} else if (label == "roundrect") {
				roundrectGrobFn(self$radius)
				# coords <- roundrect_xy(self$radius)
				# polygonGrobFn(coords$x, coords$y)
			} else if (grepl("^convex", label)) {
				if (back) {
					theta <- 180 - theta
				}
				coords <- convex_xy(get_n_vertices(label), theta, 0.5 - mat_width[1])
				polygonGrobFn(coords$x, coords$y)
			} else if (label == "oval") {
				coords <- convex_xy(72, 90, 0.5 - mat_width[1])
				polygonGrobFn(coords$x, coords$y)
			} else {
				#### for halma and roundrect no mat shrinkage (yet)
				#### For other shapes no mat support at all
				coords <- self$npc_coords
				polygonGrobFn(coords$x, coords$y)
			}
		}
	)
)

rectGrobFn <- function(mat_width = 0) {
	mat_width <- rep(mat_width, length.out = 4L)
	t <- 1 - mat_width[1]
	r <- 1 - mat_width[2]
	b <- mat_width[3]
	l <- mat_width[4]
	function(name = NULL, gp = gpar(), vp = NULL) {
		polygonGrob(x = c(l, l, r, r), y = c(t, b, b, t), name = name, gp = gp, vp = vp)
	}
}

circleGrobFn <- function(r = 0.5) {
	function(name = NULL, gp = gpar(), vp = NULL) {
		circleGrob(r = unit(r, "snpc"), name = name, gp = gp, vp = vp)
	}
}

polygonGrobFn <- function(x, y) {
	function(name = NULL, gp = gpar(), vp = NULL) polygonGrob(x, y, name = name, gp = gp, vp = vp)
}

roundrectGrobFn <- function(r = 0.05) {
	function(name = NULL, gp = gpar(), vp = NULL) {
		roundrectGrob(r = unit(r, "snpc"), name = name, gp = gp, vp = vp)
	}
}

is_known_shape_label <- function(label) {
	if (label %in% c("circle", "halma", "kite", "meeple", "oval", "pyramid", "rect", "roundrect")) {
		return(TRUE)
	}
	if (grepl("^concave[0-9]+$", label)) {
		return(TRUE)
	}
	if (grepl("^convex[0-9]+$", label)) {
		return(TRUE)
	}
	FALSE
}

#' @export
makeContext.gridlines <- function(x) {
	x$gp$col <- x$gp$col %||% "black"
	if (x$shape$label %in% c("rect", "roundrect")) {
		x$gp$lwd <- 8
		x$gp$lineend <- "butt"
	} else {
		x$gp$lwd <- 4
	}
	x
}

#' @export
makeContent.gridlines <- function(x) {
	label <- x$shape$label
	theta <- x$shape$theta
	radius <- x$shape$radius
	back <- x$shape$back
	if (is_color_invisible(x$gp$col)) {
		return(setChildren(x, gList(nullGrob(name = "invisible"))))
	}
	if (label %in% c("circle", "kite", "halma", "oval", "meeple")) {
		abort(paste("Don't know how to add grid lines to shape", label))
	}
	gl <- gList()
	if (label %in% c("rect", "roundrect")) {
		width <- rep(x$mat_width, 4)
		gl[[1]] <- segmentsGrob(0.5, width[3], 0.5, 1 - width[1], name = "line1")
		gl[[2]] <- segmentsGrob(width[4], 0.5, 1 - width[2], 0.5, name = "line2")
	} else {
		theta <- if (back) 180 - theta else theta
		n_vertices <- get_n_vertices(label)
		t <- seq(0, 360, length.out = n_vertices + 1) + theta
		t <- t[1:(length(t) - 1)]
		nt <- length(t)
		n <- floor(nt / 2)
		r <- (0.5 - x$mat_width[1]) - 0.01
		xc <- 0.5 + to_x(t, r)
		yc <- 0.5 + to_y(t, r)
		for (ii in 1:nt) {
			i_next <- ii + n
			if (i_next > nt) {
				i_next <- i_next %% nt
			}
			gl[[ii]] <- segmentsGrob(
				xc[ii],
				yc[ii],
				xc[i_next],
				yc[i_next],
				name = paste0("line", ii)
			)
		}
	}
	setChildren(x, gl)
}

#' @export
makeContext.mat <- function(x) {
	x$gp$col <- x$gp$col %||% NA_character_
	x$gp$fill <- x$gp$fill %||% NA_character_
	x
}

#' @export
makeContent.mat <- function(x) {
	mat_width <- x$mat_width
	label <- x$shape$label
	theta <- x$shape$theta
	radius <- x$shape$radius
	back <- x$shape$back

	if (is_color_invisible(x$gp$fill) || nigh(mat_width, 0)) {
		gl <- gList(nullGrob())
	} else if (label == "rect") {
		gl <- gList(rectMatGrobFn(mat_width)())
	} else if (label == "oval") {
		gl <- gList(convexMatGrobFn(72, 90, mat_width[1])())
	} else if (grepl("^convex", label)) {
		if (back) {
			theta <- 180 - theta
		}
		gl <- gList(convexMatGrobFn(get_n_vertices(label), theta, mat_width[1])())
	} else if (label == "halma") {
		gl <- gList(halmaMatGrobFn(mat_width)())
	} else if (label %in% c("circle", "roundrect")) {
		gl <- gList(diffMatGrobFn(mat_width, x$shape)())
	} else {
		abort(paste("Don't know how to add mat to shape", label))
	}
	setChildren(x, gl)
}

#' @export
makeContext.checkers <- function(x) {
	x$gp$col <- x$gp$col %||% NA_character_
	x
}

#' @export
makeContent.checkers <- function(x) {
	label <- x$shape$label
	theta <- x$shape$theta
	back <- x$shape$back
	if (is_color_invisible(x$gp$fill)) {
		gl <- gList(nullGrob())
	} else if (label == "rect") {
		gl <- gList(
			rectGrob(x = 0.25, y = 0.25, width = 0.5, height = 0.5),
			rectGrob(x = 0.75, y = 0.75, width = 0.5, height = 0.5)
		)
	} else if (grepl("^concave", label) || grepl("^convex", label)) {
		if (back) {
			theta <- 180 - theta
		}
		n_vertices <- get_n_vertices(label)
		t <- seq(0, 360, length.out = n_vertices + 1) + theta
		nt <- length(t) - 1
		r <- 0.5
		xc <- 0.5 + to_x(t, r)
		yc <- 0.5 + to_y(t, r)
		gl <- gList()
		for (ii in 1:nt) {
			if (ii %% 2) {
				xs <- c(0.5, xc[ii], xc[ii + 1])
				ys <- c(0.5, yc[ii], yc[ii + 1])
				gl[[ii]] <- polygonGrob(xs, ys)
			}
		}
	} else {
		squares <- gList(
			rectGrob(x = 0.25, y = 0.25, width = 0.5, height = 0.5),
			rectGrob(x = 0.75, y = 0.75, width = 0.5, height = 0.5)
		)
		gl <- gList(x$shape$polyclip(squares, "intersection"))
	}
	setChildren(x, gl)
}

#' @export
makeContext.hexlines <- function(x) {
	x$gp$lwd <- x$gp$lwd %||% 4
	x
}

#' @export
makeContent.hexlines <- function(x) {
	label <- x$shape$label
	if (is_color_invisible(x$gp$col)) {
		gl <- gList(nullGrob())
	} else if (label %in% c("rect", "roundrect")) {
		ho <- 0.25
		gl <- gList(
			segmentsGrob(0, 1 - ho, ho, 1),
			segmentsGrob(0, ho, ho, 0),
			segmentsGrob(1, ho, 1 - ho, 0),
			segmentsGrob(1, 1 - ho, 1 - ho, 1)
		)
	} else {
		abort(paste("Don't know how to add hexlines to shape", label))
	}
	setChildren(x, gl)
}

#' @export
makeContent.pp_polyclip <- function(x) {
	subject <- x$shape$shape()
	clip <- x$clip
	if (inherits(clip, "pp_grobCoords") || inherits(clip, "pmap_piece")) {
		xylist <- grid::grobCoords(clip, closed = TRUE)
		clip <- gridGeometry::xyListPath(xylist)
	}
	gl <- gList(gridGeometry::polyclipGrob(subject, clip, x$op))
	setChildren(x, gl)
}

# xy functions and/or constants
roundrect_xy <- function(shape_r, width = 1, height = 1) {
	# Always querying in square `pdf(NULL)` device fixes #254
	current_dev <- grDevices::dev.cur()
	grDevices::pdf(NULL, width = width, height = height)
	on.exit(grDevices::dev.off(), add = TRUE)
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	coords <- grid::grobCoords(
		grid::roundrectGrob(r = grid::unit(shape_r, "snpc")),
		closed = TRUE
	)[[1]]
	x <- convertX(grid::unit(coords$x, "in"), "npc", valueOnly = TRUE)
	y <- convertY(grid::unit(coords$y, "in"), "npc", valueOnly = TRUE)
	list(x = x, y = y, c = rep("C1", length(x)))
}

kite_xy <- function(t = 90, r = 0.25) {
	t <- t %% 360
	stopifnot(nigh(t, 0) || nigh(t, 90) || nigh(t, 180) || nigh(t, 270))
	if (nigh(t, 0)) {
		list(x = c(1, 0.5 - r, 0, 0.5 - r), y = c(0.5, 1, 0.5, 0), c = rep("C0", 4))
	} else if (nigh(t, 90)) {
		list(x = c(0.5, 0, 0.5, 1), y = c(1, 0.5 - r, 0, 0.5 - r), c = rep("C0", 4))
	} else if (nigh(t, 180)) {
		list(x = c(0, 0.5 + r, 1, 0.5 + r), y = c(0.5, 0, 0.5, 1), c = rep("C0", 4))
	} else {
		list(x = c(0.5, 1, 0.5, 0), y = c(0, 0.5 + r, 1, 0.5 + r), c = rep("C0", 4))
	}
}
pyramid_xy <- list(x = c(0.5, 0, 1), y = c(1, 0, 0), c = rep("C0", 3))
rect_xy <- list(x = c(0, 0, 1, 1), y = c(1, 0, 0, 1), c = rep("C0", 4))

convex_xy <- function(n_vertices, t, r = 0.5) {
	t <- seq(0, 360, length.out = n_vertices + 1) + t
	x <- to_x(t, r) + 0.5
	y <- to_y(t, r) + 0.5
	list(x = utils::head(x, -1), y = utils::head(y, -1), c = rep("C0", n_vertices))
}

halma_xy <- function() {
	y_cutoff <- 0.55
	y_frac <- 0.5
	t <- seq(0, 360, length.out = 36 + 1) - 90
	r <- 0.5
	x <- 0.5 + to_x(t, r)
	y <- 1 + y_frac * (to_y(t, r) - r)
	indices <- which(y >= y_cutoff)
	list(
		x = c(1, 1, x[indices], 0, 0),
		y = c(0, 0.3, y[indices], 0.3, 0),
		c = c("C0", "C0", "C0", rep("C1", length(indices) - 2), "C0", "C0", "C0")
	)
}

concave_xy <- function(n_vertices, t, r = 0.2) {
	t_outer <- seq(0, 360, length.out = n_vertices + 1) + t
	n_degrees <- 360 / n_vertices / 2
	t_inner <- seq(n_degrees, 360 - n_degrees, length.out = n_vertices) + t
	x_outer <- to_x(t_outer, 0.5) + 0.5
	x_inner <- to_x(t_inner, r) + 0.5
	y_outer <- to_y(t_outer, 0.5) + 0.5
	y_inner <- to_y(t_inner, r) + 0.5
	x <- splice(x_outer, x_inner)
	y <- splice(y_outer, y_inner)
	list(x = x, y = y, c = rep("C0", length(x)))
}

splice <- function(x0, x1) {
	vec <- as.numeric()
	for (ii in seq_along(x1)) {
		vec <- append(vec, x0[ii])
		vec <- append(vec, x1[ii])
	}
	append(vec, x0[ii + 1])
}

library("grid", quietly = TRUE)
test_that("pp_shape() works as expected", {
	circle <- pp_shape("circle")
	current_dev <- grDevices::dev.cur()
	pdf_file <- tempfile(fileext = ".pdf")
	expect_error({
		pdf(pdf_file)
		grid.draw(circle$hexlines(gp = gpar(col = "yellow")))
	})
	unlink(pdf_file)
	dev.off()
	if (current_dev > 1) {
		grDevices::dev.set(current_dev)
	}
	skip_on_ci()
	skip_if_not_installed("vdiffr")
	library("vdiffr")
	expect_doppelganger("add-checkers", function() {
		rect <- pp_shape("rect")
		pushViewport(viewport(width = 0.9, height = 0.9))
		grid.draw(rect$shape(gp = gpar(fill = "grey", col = "NA")))
		grid.draw(rect$checkers(gp = gpar(fill = "purple")))
		grid.draw(rect$gridlines(gp = gpar(col = "black")))
		grid.draw(rect$mat(mat_width = 0.05, gp = gpar(fill = "green")))
		grid.draw(rect$hexlines(gp = gpar(col = "yellow")))
		grid.draw(rect$shape(gp = gpar(col = "black", fill = "NA", lex = 4)))
	})
	expect_doppelganger("add-checkers-transparent", function() {
		rect <- pp_shape("rect")
		grid.draw(rect$checkers(gp = gpar(fill = "transparent")))
		grid.draw(rect$hexlines(gp = gpar(col = "transparent")))
	})
	expect_doppelganger("add-checkers-convex8", function() {
		convex8 <- pp_shape("convex8", back = TRUE)
		pushViewport(viewport(width = 0.9, height = 0.9))
		grid.draw(convex8$shape(gp = gpar(fill = "grey", col = "NA")))
		grid.draw(convex8$checkers(gp = gpar(fill = "purple")))
		grid.draw(convex8$gridlines(gp = gpar(col = "black")))
		grid.draw(convex8$mat(mat_width = 0.05, gp = gpar(fill = "green")))
		grid.draw(convex8$shape(gp = gpar(col = "black", fill = "NA", lex = 4)))
	})
	expect_doppelganger("roundrect", function() {
		roundrect <- pp_shape("roundrect")
		pushViewport(viewport(width = 0.9, height = 0.9))
		grid.draw(roundrect$shape(gp = gpar(fill = "grey", col = "NA")))
		grid.draw(roundrect$checkers(gp = gpar(fill = "purple")))
		grid.draw(roundrect$gridlines(gp = gpar(col = "black")))
		grid.draw(roundrect$mat(mat_width = 0.05, gp = gpar(fill = "green")))
		grid.draw(roundrect$hexlines(gp = gpar(col = "yellow")))
		grid.draw(roundrect$shape(gp = gpar(col = "black", fill = "NA", lex = 4)))
	})
	expect_doppelganger("circle", function() {
		circle <- pp_shape("circle")
		pushViewport(viewport(width = 0.9, height = 0.9))
		grid.draw(circle$shape(gp = gpar(fill = "grey", col = "NA")))
		grid.draw(circle$checkers(gp = gpar(fill = "purple")))
		grid.draw(circle$mat(mat_width = 0.05, gp = gpar(fill = "green")))
		# grid.draw(circle$hexlines(gp=gpar(col="yellow"))) # nolint
		# grid.draw(circle$gridlines(gp=gpar(col="black"))) # nolint
		grid.draw(circle$shape(gp = gpar(col = "black", fill = "NA", lex = 4)))
	})
	expect_doppelganger("polyclip-circle", function() {
		circle <- pp_shape("circle")
		g <- pieceGrob("coin_face", op_scale = 1, default.units = "npc")
		grid.draw(circle$polyclip(g, "minus", gp = gpar(fill = "blue")))
	})
	expect_doppelganger("polyclip-holed-board", function() {
		rect <- pp_shape("rect")
		g <- pieceGrob("board_face", cfg = game_systems()$marbles, default.units = "npc")
		grid.draw(rect$polyclip(g, "minus", gp = gpar(col = NA, fill = "red")))
	})
	expect_doppelganger("polyclip-holed-board-op", function() {
		rect <- pp_shape("rect")
		g <- pieceGrob(
			"board_face",
			cfg = game_systems()$marbles,
			op_scale = 0.5,
			default.units = "npc"
		)
		grid.draw(g)
		grid.draw(rect$polyclip(g, "minus", gp = gpar(col = NA, fill = "red")))
	})
	skip_if_not_installed("gridpattern")
	expect_doppelganger("hex-pattern", function() {
		hex <- pp_shape("convex6")
		gp <- gpar(fill = c("blue", "yellow", "red"), col = "black")
		grid.draw(hex$pattern(
			"polygon_tiling",
			gp = gp,
			spacing = 0.1,
			type = "truncated_trihexagonal"
		))
		gp <- gpar(fill = "black", col = NA)
		grid.draw(hex$mat(mat_width = 0.025, gp = gp))
	})
})

test_that("npc_coords() works", {
	expect_equal(pp_shape("kite")$npc_coords, kite_xy(r = 0.2))
	expect_equal(pp_shape("kite", radius = 0.25)$npc_coords, kite_xy(r = 0.25))
	expect_equal(pp_shape("halma")$npc_coords, halma_xy())
	expect_equal(pp_shape("pyramid")$npc_coords, pyramid_xy)
	expect_equal(pp_shape("concave4")$npc_coords, concave_xy(4, 90, 0.2))
	expect_error(pp_shape("boobah")$npc_coords, "Don't recognize shape label boobah")
})

test_that("partition_edges works", {
	s <- pp_shape("kite")
	expect_equal(partition_edges(s)$type, rep("flat", 4))
	expect_equal(partition_edges(s)$indices, list(1:2, 2:3, 3:4, c(4, 1)))

	s <- pp_shape("oval")
	expect_equal(partition_edges(s)$type, "ring")
	expect_equal(partition_edges(s)$indices, list(1:72))

	s <- pp_shape("halma")
	expect_equal(partition_edges(s)$type, c("flat", "flat", "flat", "curved", "flat", "flat"))
	expect_equal(partition_edges(s)$indices, list(c(33, 1), 1:2, 2:3, 3:31, 31:32, 32:33))
})

test_that("Token2S works as expected", {
	shape <- pp_shape("kite")
	token <- Token2S$new(shape)
	expect_length(token$edges, 4)
})

at_vp_define <- function(piece_side, suit, rank, cfg) {
	viewport(
		width = inch(cfg$get_width(piece_side, suit, rank)),
		height = inch(cfg$get_height(piece_side, suit, rank))
	)
}

at_ps_grob <- function(piece_side, suit, rank, cfg, xy_vp, xy_polygon, name = "piece_side") {
	at_settings <- affiner::affine_settings(as.data.frame(xy_vp))
	vp_define <- at_vp_define(piece_side, suit, rank, cfg)

	if (nigh(at_settings$width, 0) || nigh(at_settings$height, 0)) {
		ps_grob <- nullGrob()
	} else if (
		nigh(at_settings$width, vp_define$width) &&
			nigh(at_settings$height, vp_define$height) &&
			nigh(at_settings$sx, 0) &&
			!at_settings$flipX
	) {
		ps_grob <- cfg$get_grob(piece_side, suit, rank)
		ps_grob$vp <- at_settings$vp
	} else if (has_transformations()) {
		#### && !at_settings$flipX ?
		grob <- cfg$get_grob(piece_side, suit, rank)
		has_border <- hasName(grob, "border")
		if (has_border) {
			grob$border <- FALSE
		}
		if (at_settings$flipX && hasName(grob, "flip")) {
			grob$flip <- TRUE
		}
		ps_grob <- affiner::affineGrob(
			grob,
			vp_define = vp_define,
			vp_use = at_settings$vp,
			transform = at_settings$transform
		)
		if (has_border) {
			opt <- cfg$get_piece_opt(piece_side, suit, rank)
			gp <- gpar(col = opt$border_color, fill = "transparent", lex = opt$border_lex)
			border_grob <- polygonGrob(
				x = xy_polygon$x,
				y = xy_polygon$y,
				default.units = "in",
				gp = gp
			)
			ps_grob <- gList(ps_grob, border_grob)
		}
	} else if (nigh(at_settings$sx, 0) && !at_settings$flipX) {
		at_inform(fallback = "picture")
		ps_grob <- cfg$get_grob(piece_side, suit, rank, "picture")
		ps_grob$vp <- at_settings$vp
	} else {
		at_inform(fallback = "polygon")
		opt <- cfg$get_piece_opt(piece_side, suit, rank)
		gp <- gpar(col = opt$border_color, fill = opt$background_color, lex = opt$border_lex)
		ps_grob <- polygonGrob(x = xy_polygon$x, y = xy_polygon$y, default.units = "in", gp = gp)
	}
	gTree(
		scale = 1,
		xy_polygon = xy_polygon,
		name = name,
		children = gList(ps_grob),
		cl = "pp_ps_transformation"
	)
}

#' @export
grobCoords.pp_ps_transformation <- function(x, closed, ...) {
	if (
		getRversion() >= '4.2.0' &&
			(!closed || inherits(x$children[[1]], c("null", "grob")))
	) {
		return(emptyGrobCoords(x$name))
	}

	grobCoords(
		polygonGrob(
			x = x$xy_polygon$x,
			y = x$xy_polygon$y,
			default.units = "in",
			vp = x$vp,
			name = x$name
		),
		closed = closed,
		...
	)
}

#' @export
makeContent.pp_ps_transformation <- function(x) {
	if (length(x$children) == 1) {
		grob <- x$children[[1]]
		if (inherits(grob, c("polygon", "grob"))) {
			grob <- update_gp(grob, gp = gpar(cex = x$scale, lex = x$scale))
		} else if (hasName(grob, "scale")) {
			grob$scale <- x$scale
		}
		x$children[[1]] <- grob
	} else {
		# transformation grob plus manual border
		x$children[[2]] <- update_gp(x$children[[2]], gp = gpar(cex = x$scale, lex = x$scale))
	}
	x
}

print_and_play_paper_bleed <- function(cfg, size, pieces, arrangement, quietly, size_bleed) {
	n_suits <- cfg$n_suits
	n_ranks <- cfg$n_ranks

	stopifnot(n_ranks <= 6)
	if ("matchsticks" %in% pieces) {
		abort('"matchsticks" `pieces` not currently supported for `bleed = TRUE`')
	}
	if ("pyramids" %in% pieces) {
		abort('"pyramids" `pieces` not currently supported for `bleed = TRUE`')
	}
	if ("subpack" %in% pieces) {
		abort('"subpack" `pieces` not currently supported for `bleed = TRUE`')
	}

	if (size == "letter") {
		xl <- inch(LETTER_HEIGHT / 2 - A5W / 2 + size_bleed$left)
		xr <- inch(LETTER_HEIGHT / 2 + A5W / 2 + size_bleed$left)
		y <- inch(LETTER_WIDTH / 2 + size_bleed$bottom)
	} else if (size == "A4") {
		xl <- inch(A4_HEIGHT / 2 - A5W / 2 + size_bleed$left)
		xr <- inch(A4_HEIGHT / 2 + A5W / 2 + size_bleed$left)
		y <- inch(A4_WIDTH / 2 + size_bleed$bottom)
	} else {
		# size is "A5"
		xl <- inch(A5W / 2 + size_bleed$left)
		xr <- inch(A5W / 2 + size_bleed$left)
		y <- inch(A5H / 2 + size_bleed$bottom)
	}

	vpl <- viewport(x = xl, y = y, width = inch(A5W), height = inch(A5H))
	vpr <- viewport(x = xr, y = y, width = inch(A5W), height = inch(A5H))
	gl <- list()
	pl <- list()

	## Front Matter
	gl <- gappend(gl, a5_title_grob(cfg, pieces, quietly, bleed = TRUE))
	gl <- gappend(gl, a5_inst_grob_bleed(cfg, pieces, arrangement, size))
	if (arrangement == "double-sided" && size != "A5") {
		gl <- gappend(gl, blank_grob)
		gl <- gappend(gl, blank_grob)
		pl[["Front Matter"]] <- 2
	} else {
		pl[["Front Matter"]] <- 1
	}

	## Piecepack
	if ("piecepack" %in% pieces) {
		n_pages_tiles <- n_suits
		if (is_odd(n_pages_tiles) && arrangement == "double-sided" && size != "A5") {
			n_pages_tiles <- n_pages_tiles + 1
		}
		for (suit in seq(n_pages_tiles)) {
			gl <- gappend(gl, a5_tile_grob(suit, cfg, TRUE, arrangement, size))
			gl <- gappend(gl, a5_tile_grob(suit, cfg, FALSE, arrangement, size))
		}
		pl$Tiles <- n_pages_tiles

		n_pages_coins <- (n_suits - 1) %/% 4 + 1
		for (ii in seq(n_pages_coins)) {
			ss <- seq(4 * ii - 3, 4 * ii)
			gl <- gappend(gl, a5_coin_grob(ss, cfg, TRUE, arrangement, size))
			gl <- gappend(gl, a5_coin_grob(ss, cfg, FALSE, arrangement, size))
		}
		if (is_odd(n_pages_coins) && arrangement == "double-sided" && size != "A5") {
			n_pages_coins <- n_pages_coins + 1
			ss <- 1:4
			gl <- gappend(gl, a5_coin_grob(ss, cfg, TRUE, arrangement, size))
			gl <- gappend(gl, a5_coin_grob(ss, cfg, FALSE, arrangement, size))
		}
		pl$Coins <- n_pages_coins

		n_pages_dice <- (n_suits - 1) %/% 4 + 1
		for (ii in seq(n_pages_dice)) {
			ss <- seq(4 * ii - 3, 4 * ii)
			if (arrangement == "double-sided" && size == "A5") {
				gl <- gappend(gl, a5_die_grob(ss, cfg, TRUE, arrangement))
				gl <- gappend(gl, blank_grob)
				gl <- gappend(gl, a5_pawn_grob(ss, cfg, TRUE, arrangement))
				gl <- gappend(gl, blank_grob)
			} else {
				gl <- gappend(gl, a5_die_grob(ss, cfg, TRUE, arrangement))
				gl <- gappend(gl, a5_pawn_grob(ss, cfg, FALSE, arrangement))
				if (arrangement == "double-sided") {
					gl <- gappend(gl, blank_grob)
					gl <- gappend(gl, blank_grob)
				}
			}
		}
		if (arrangement == "double-sided") {
			n_pages_dice <- 2 * n_pages_dice
		}
		pl$`Dice and Pawns` <- n_pages_dice
	}
	if (size == "A5") {
		pl <- lapply(pl, function(x) 2 * x)
	}
	for (ii in seq(gl)) {
		if (is_odd(ii)) {
			grid.newpage()
			draw_a5_page(gl[[ii]], vpl)
		} else {
			if (size == "A5") {
				grid.newpage()
			}
			draw_a5_page(gl[[ii]], vpr)
		}
	}

	pl
}

a5_inst_grob_bleed <- function(cfg, pieces, arrangement, size) {
	y_inst <- unit(1, "npc") - unit(0.2, "in")
	inst <- c("\u25cf See https://www.ludism.org/ppwiki/MakingPiecepacks for general advice")

	# Settings
	components <- paste(paste0('"', pieces, '"'), collapse = ", ")
	if (TRUE) {
		arrangement <- sprintf('\t\u25cb "%s" arrangement with "bleed" zones', arrangement)
	}
	inst <- c(
		inst,
		"\u25cf This print-and-play layout was generated for:",
		sprintf('\t\u25cb %s components', components),
		arrangement,
		sprintf('\t\u25cb "%s" paper size', size)
	)

	# Tiles
	tile_crop <- c('\t\u25cb Use "crop" marks as guide to where to cut them out')
	if (size == "a5") {
		inst <- c(
			inst,
			'\u25cf For each suit we have a page of tile "faces" then a page of tile "backs"',
			tile_crop,
			'\t\u25cb May place whole sheet on one side and next sheet on the opposite side',
			'\t\t of target material before cutting out'
		)
	} else {
		inst <- c(
			inst,
			'\u25cf We have a page of tiles for each suit: "faces" on left and "backs" on right',
			tile_crop,
			'\t\u25cb Central "gutter" line may help place sheet on both sides of target material',
			'\t\t  Line up "gutter" line on the center of edge of target material, fold,',
			'\t\t  and cut tiles out.',
			'\t\u25cb Alternatively place whole sheet on one side and another on the opposite side'
		)
	}

	# Coins
	coin_crop <- c(
		'\t\u25cb For those lacking circular cutting tools one may use "crop" marks',
		'\t\t as guide to cut out (inferior) square coins',
		'\t\u25cb Otherwise use "crop" marks to help center placement of circular cutting tool'
	)
	if (size == "a5") {
		inst <- c(
			inst,
			"\u25cf We have a page of coin faces then a page of coin backs (per 4 suits)",
			coin_crop
		)
	} else {
		inst <- c(
			inst,
			'\u25cf We have a page of coins (per 4 suits): "backs" on left and "faces" on right',
			coin_crop,
			'\t\u25cb Central "gutter" line may help place sheet on both sides of target material',
			'\t\t  Line up "gutter" line on the center of edge of target material, fold,',
			'\t\t  and cut coins out.',
			'\t\u25cb Alternatively place whole sheet on one side and a copy on the opposite side'
		)
	}

	# Dice and Pawns
	dice_crop <- c(
		'\t\u25cb Depending on size of target dice use inner or outer "crop" marks',
		'\t\t as guide to cut out dice',
		'\t\u25cb Alternatively use "crop" marks to help center placement of circular cutting tool',
		'\t\u25cb Use "crop" marks to cut out pawn "belts"',
		'\t\u25cb Use "crop" marks to cut out pawn "tokens" (depending on shape of token)',
		'\t\t "gutter" line can be used to help place pawn "tokens" on both sides',
		'\t\t  of target material ("gutter" line on center of edge of target material)'
	)
	if (size == "a5") {
		inst <- c(
			inst,
			"\u25cf We have a page of dice faces then a page of pawn belts and pawn tokens (per 4 suits)",
			dice_crop
		)
	} else {
		inst <- c(
			inst,
			'\u25cf We have a page of dice faces, pawn belts, and pawn tokens (per 4 suits):',
			'\t\u25cb Dice faces on left, pawn "belts" on top right, pawn "tokens" on bottom right',
			dice_crop
		)
	}

	inst <- paste(inst, collapse = "\n")

	gTree(
		name = "instructions",
		children = gList(
			# rectGrob(gp=gpar(fill="grey90")),
			textGrob(
				"Instructions",
				x = unit(0.5, "cm"),
				y = y_inst,
				just = "left",
				gp = gp_header
			),
			textGrob(
				inst,
				x = unit(0.5, "cm"),
				y = y_inst - unit(0.2, "in"),
				just = c(0, 1),
				gp = gp_text
			)
		)
	)
}

a5_tile_grob <- function(i_suit, cfg, front, arrangement, size) {
	tile_width <- cfg$get_width("tile_back")
	xtr <- 0.5 * tile_width + 1 / 8 + 1 / 8
	xtl <- xtr + tile_width + 1 / 4

	content_height <- 3 * tile_width + 3 / 4
	ytb <- 0.5 * tile_width + 1 / 8 + (A5H - content_height) / 2
	ytm <- ytb + tile_width + 1 / 4
	ytt <- ytm + tile_width + 1 / 4

	df <- tibble::tibble(
		piece_side = "tile_face",
		x = rep(c(xtr, xtl), 3),
		y = rep(c(ytt, ytm, ytb), each = 2),
		suit = i_suit,
		rank = 1:6
	)

	if (!front) {
		df$piece_side = "tile_back"
		df$x <- A5W - df$x
		# Rotate tile backs to partially hide direction of face if tile back are not fully symmetric
		df$angle <- 90 * ((df$suit + df$rank) %% 4)
	}
	if (!front && size != "A5") {
		vline <- linesGrob(x = 0, gp = gpar(lty = "dashed"))
	} else {
		vline <- nullGrob()
	}

	cm_grob <- pmap_piece(
		df,
		cropmarkGrob,
		cfg = cfg,
		default.units = "in",
		bleed = TRUE,
		draw = FALSE
	)
	ps_grob <- pmap_piece(
		df,
		pieceGrob,
		cfg = cfg,
		default.units = "in",
		bleed = TRUE,
		draw = FALSE
	)

	ch_size <- 1 / 4
	side_x <- if (front) xtr - tile_width / 2 else A5W - xtr + tile_width / 2
	top_y <- ytt + tile_width / 2 + 1 / 4
	bottom_y <- ytb - tile_width / 2 - 1 / 4
	reg_grob <- gList(
		circledSegmentsCrosshairGrob(
			x = inch(side_x),
			y = inch(top_y),
			width = inch(ch_size),
			height = inch(ch_size)
		),
		squaredSegmentsCrosshairGrob(
			x = inch(side_x),
			y = inch(bottom_y),
			width = inch(ch_size),
			height = inch(ch_size)
		)
	)

	gList(cm_grob, ps_grob, vline, reg_grob)
}

a5_coin_grob <- function(suit, cfg, front, arrangement, size) {
	coin_width <- cfg$get_width("coin_back")
	xt1 <- 0.5 * coin_width + 1 / 8 + 1 / 8
	xt2 <- xt1 + coin_width + 1 / 4 + 1 / 8
	xt3 <- xt2 + coin_width + 1 / 4 + 1 / 8
	xt4 <- xt3 + coin_width + 1 / 4 + 1 / 8

	content_height <- 5 * (coin_width + 1 / 4 + 1 / 8) + coin_width + 1 / 4
	yt1 <- 0.5 * coin_width + 1 / 8 + (A5H - content_height) / 2
	yt2 <- yt1 + coin_width + 1 / 4 + 1 / 8
	yt3 <- yt2 + coin_width + 1 / 4 + 1 / 8
	yt4 <- yt3 + coin_width + 1 / 4 + 1 / 8
	yt5 <- yt4 + coin_width + 1 / 4 + 1 / 8
	yt6 <- yt5 + coin_width + 1 / 4 + 1 / 8

	df <- tibble::tibble(
		piece_side = "coin_back",
		x = rep(c(xt1, xt2, xt3, xt4), each = 6),
		y = rep(c(yt1, yt2, yt3, yt4, yt5, yt6), 4),
		suit = rep(suit, each = 6),
		rank = rep(1:6, 4)
	)

	if (!front) {
		df$piece_side = "coin_face"
		df$x <- A5W - df$x
	} else {
		# Rotate coin faces to desired direction
		df$angle <- cfg$coin_arrangement
	}
	if (!front && size != "A5") {
		vline <- linesGrob(x = 0, gp = gpar(lty = "dashed"))
	} else {
		vline <- nullGrob()
	}

	cm_grob <- pmap_piece(
		df,
		cropmarkGrob,
		cfg = cfg,
		default.units = "in",
		bleed = TRUE,
		draw = FALSE
	)
	ps_grob <- pmap_piece(
		df,
		pieceGrob,
		cfg = cfg,
		default.units = "in",
		bleed = TRUE,
		draw = FALSE
	)

	ch_size <- 1 / 4
	side_x <- if (front) xt1 - coin_width / 2 - 1 / 8 else A5W - xt1 + coin_width / 2 + 1 / 8
	top_y <- yt6 + coin_width / 2 + 1 / 4
	bottom_y <- yt1 - coin_width / 2 - 1 / 4
	reg_grob <- gList(
		circledSegmentsCrosshairGrob(
			x = inch(side_x),
			y = inch(top_y),
			width = inch(ch_size),
			height = inch(ch_size)
		),
		squaredSegmentsCrosshairGrob(
			x = inch(side_x),
			y = inch(bottom_y),
			width = inch(ch_size),
			height = inch(ch_size)
		)
	)

	gList(cm_grob, ps_grob, vline, reg_grob)
}

a5_die_grob <- function(suit, cfg, front, arrangement) {
	die_width <- cfg$get_width("die_face")
	xt1 <- 0.5 * 3 / 4 + 1 / 8 + 1 / 8
	xt2 <- xt1 + 3 / 4 + 1 / 4 + 1 / 8
	xt3 <- xt2 + 3 / 4 + 1 / 4 + 1 / 8
	xt4 <- xt3 + 3 / 4 + 1 / 4 + 1 / 8

	content_height <- 5 * (3 / 4 + 1 / 4 + 1 / 8) + 3 / 4 + 1 / 4
	yt1 <- 0.5 * 3 / 4 + 1 / 8 + (A5H - content_height) / 2
	yt2 <- yt1 + 3 / 4 + 1 / 4 + 1 / 8
	yt3 <- yt2 + 3 / 4 + 1 / 4 + 1 / 8
	yt4 <- yt3 + 3 / 4 + 1 / 4 + 1 / 8
	yt5 <- yt4 + 3 / 4 + 1 / 4 + 1 / 8
	yt6 <- yt5 + 3 / 4 + 1 / 4 + 1 / 8

	df <- tibble::tibble(
		piece_side = "die_face",
		x = rep(c(xt1, xt2, xt3, xt4), each = 6),
		y = rep(c(yt1, yt2, yt3, yt4, yt5, yt6), 4),
		suit = rep(suit, each = 6),
		rank = rep(1:6, 4)
	)

	if (!front) {
		df$x <- A5W - df$x
	}

	cm_grob1 <- pmap_piece(
		df,
		cropmarkGrob,
		cfg = cfg,
		default.units = "in",
		width = die_width,
		height = die_width,
		bleed = (3 / 4 - die_width) / 2 + 1 / 8,
		draw = FALSE
	)
	cm_grob2 <- pmap_piece(
		df,
		cropmarkGrob,
		cfg = cfg,
		default.units = "in",
		width = 3 / 4,
		height = 3 / 4,
		bleed = 1 / 8,
		draw = FALSE
	)
	ps_grob <- pmap_piece(
		df,
		pieceGrob,
		cfg = cfg,
		default.units = "in",
		bleed = (3 / 4 - die_width) / 2 + 1 / 8,
		draw = FALSE
	)
	gList(cm_grob1, cm_grob2, ps_grob)
}

a5_pawn_grob <- function(suit, cfg, front, arrangement) {
	belt_width <- cfg$get_width("belt_face")
	pawn_width <- cfg$get_width("pawn_face")
	pawn_height <- cfg$get_height("pawn_face")

	x <- A5W * (1:4 / 4 - 1 / 8)
	yb <- A5H - (0.5 * belt_width + 1 / 8 + 1 / 8)

	ys <- 0.5 * A5H

	df_b <- tibble::tibble(piece_side = "belt_face", x = x, suit = suit, y = yb, angle = 90)

	# Center token group in space below belt, leaving equal margin for registration marks
	avail <- yb - belt_width / 2 - 1 / 8
	ypb <- avail / 2 - pawn_height / 2 - 3 / 8
	ypt <- (ypb + 0.5 * pawn_height + 1 / 8) + 1 / 2 + (1 / 8 + 0.5 * pawn_height) # 1/2" gutter = jumbo domino

	if (!front) {
		df_pb <- tibble::tibble(piece_side = "pawn_back", x = x, suit = suit, y = ypb, angle = 180)
		df_pt <- tibble::tibble(piece_side = "pawn_face", x = x, suit = suit, y = ypt, angle = 0)

		# df_s <- tibble::tibble(piece_side = "saucer_face",
		#                        x = x, suit = suit, y = ys, angle = 0)
		df <- rbind(df_b, df_pb, df_pt)

		df$x <- A5W - df$x
	} else {
		df_pb <- tibble::tibble(piece_side = "pawn_face", x = x, suit = suit, y = ypb, angle = 180)
		df_pt <- tibble::tibble(piece_side = "pawn_back", x = x, suit = suit, y = ypt, angle = 0)

		# df_s <- tibble::tibble(piece_side = "saucer_back",
		#                        x = x, suit = suit, y = ys, angle = 0)
		df <- rbind(df_b, df_pb, df_pt)
	}

	hline <- linesGrob(y = inch(0.5 * (ypb + ypt)), gp = gpar(lty = "dashed"))

	cm_grob <- pmap_piece(
		df,
		cropmarkGrob,
		cfg = cfg,
		default.units = "in",
		bleed = TRUE,
		draw = FALSE
	)
	ps_grob <- pmap_piece(
		df,
		pieceGrob,
		cfg = cfg,
		default.units = "in",
		bleed = TRUE,
		draw = FALSE
	)

	ch_size <- 1 / 4
	left_x <- x[1] - pawn_width / 2
	right_x <- x[length(x)] + pawn_width / 2
	top_y <- ypt + pawn_height / 2 + 1 / 4
	bottom_y <- ypb - pawn_height / 2 - 1 / 4
	reg_grob <- gList(
		circledSegmentsCrosshairGrob(
			x = inch(left_x),
			y = inch(top_y),
			width = inch(ch_size),
			height = inch(ch_size)
		),
		squaredSegmentsCrosshairGrob(
			x = inch(right_x),
			y = inch(top_y),
			width = inch(ch_size),
			height = inch(ch_size)
		),
		circledSegmentsCrosshairGrob(
			x = inch(left_x),
			y = inch(bottom_y),
			width = inch(ch_size),
			height = inch(ch_size)
		),
		squaredSegmentsCrosshairGrob(
			x = inch(right_x),
			y = inch(bottom_y),
			width = inch(ch_size),
			height = inch(ch_size)
		)
	)

	gList(cm_grob, ps_grob, hline, reg_grob)
}

# unlike `grid::editGrob()` *updates* previous cex, lex, alpha values
# by multiplying new values with previous values.
update_gp <- function(grob, gp = gpar()) {
	stopifnot(all(names(gp) %in% c("alpha", "cex", "lex")))
	if (is.null(grob$gp)) {
		grob$gp <- gp
	} else {
		for (name in c("alpha", "cex", "lex")) {
			if (hasName(grob$gp, name) && hasName(gp, name)) {
				grob$gp[[name]] <- grob$gp[[name]] * gp[[name]]
			} else if (hasName(gp, name)) {
				grob$gp[[name]] <- gp[[name]]
			}
		}
	}
	grob
}

#' Draw crosshairs with grid
#'
#' `grid.crosshair()` draws crosshairs at the corners of a rectangular area.
#' `crosshairGrob()` is its grid grob counterpart.
#' They are intended for use in adding crosshairs at the corners of
#' game pieces in print-and-play layouts.
#' One can also use the lower level
#' `circledSegmentsCrosshairGrob()`, `segmentsCrosshairGrob()`, `squaredSegmentsCrosshairGrob()`, and `squaresCrosshairGrob()`
#' (which can be drawn with [grid::grid.draw()])
#' to add individual crosshairs to specified (x,y) locations (e.g. for use as registration marks).
#' @inheritParams pieceGrob
#' @param ch_width Width/height of `ch_grob`'s viewport.
#' @param ch_grob Crosshair grob.  Will be drawn in each corner of the piece in a viewport with `ch_width` width and `ch_width` height.
#'                `segmentsCrosshairGrob()` simply wraps [grid::segmentsGrob()].
#'                `circledSegmentsCrosshairGrob()` wraps [grid::segmentsGrob()] and [grid::circleGrob()].
#'                `squaredSegmentsCrosshairGrob()` wraps [grid::segmentsGrob()] and [grid::rectGrob()].
#'                `squaresCrosshairGrob()` wraps [grid::rectGrob()] and alternates
#'                 black/white squares for visibility on both light and dark backgrounds.
#' @return A grid grob.
#' @examples
#' if (requireNamespace("grid", quietly = TRUE) &&
#'     piecepackr:::device_supports_unicode()) {
#'   cfg <- pp_cfg(list(border_color = NA))
#'   grid::grid.newpage()
#'   df <- data.frame(piece_side = "tile_face", suit = 2, rank = 2,
#'                    x = 2, y = 2, angle = 0,
#'                    stringsAsFactors = FALSE)
#'   pmap_piece(df, grid.piece, cfg = cfg, default.units = "in")
#'   pmap_piece(df, grid.crosshair, cfg = cfg, default.units = "in")
#' }
#' if (requireNamespace("grid", quietly = TRUE) &&
#'     piecepackr:::device_supports_unicode()) {
#'   grid::grid.newpage()
#'   pmap_piece(df, grid.piece, cfg = cfg, default.units = "in")
#'   pmap_piece(df, grid.crosshair, cfg = cfg, default.units = "in",
#'              ch_grob = segmentsCrosshairGrob())
#' }
#' if (requireNamespace("grid", quietly = TRUE)) {
#'   grid::grid.newpage()
#'   fns <- list(squaresCrosshairGrob, segmentsCrosshairGrob,
#'               circledSegmentsCrosshairGrob, squaredSegmentsCrosshairGrob)
#'   labels <- c("squaresCrosshairGrob()", "segmentsCrosshairGrob()",
#'               "circledSegmentsCrosshairGrob()", "squaredSegmentsCrosshairGrob()")
#'   xs <- c(0.25, 0.75, 0.25, 0.75)
#'   ys <- c(0.70, 0.70, 0.30, 0.30)
#'   label_ys <- c(0.55, 0.55, 0.15, 0.15)
#'   for (i in seq_along(fns)) {
#'     grid::grid.draw(fns[[i]](x = grid::unit(xs[i], "npc"),
#'                              y = grid::unit(ys[i], "npc"),
#'                              width = grid::unit(0.15, "npc"),
#'                              height = grid::unit(0.15, "npc")))
#'     grid::grid.text(labels[i], x = xs[i], y = label_ys[i], gp = grid::gpar(cex = 0.7))
#'   }
#' }
#' @name grid.crosshair
NULL

#' @rdname grid.crosshair
#' @export
crosshairGrob <- function(
	...,
	piece_side = "tile_back",
	suit = NA,
	rank = NA,
	cfg = getOption("piecepackr.cfg", pp_cfg()),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	angle = 0,
	width = NA,
	height = NA,
	scale = 1,
	default.units = "npc",
	envir = getOption("piecepackr.envir"),
	name = NULL,
	gp = gpar(),
	vp = NULL,
	ch_width = unit(1 / 6, "in"),
	ch_grob = squaresCrosshairGrob()
) {
	if (is.na(width) || is.na(height)) {
		cfg <- get_cfg(cfg, envir)
		rank <- impute_rank(piece_side, rank, cfg)
		suit <- impute_suit(piece_side, suit, cfg)
		if (is.na(width)) {
			width <- inch(cfg$get_width(piece_side, suit, rank))
		}
		if (is.na(height)) height <- inch(cfg$get_height(piece_side, suit, rank))
	}

	if (is_angle(angle)) {
		angle <- as.double(angle, "degrees")
	}

	if (!is.unit(x)) {
		x <- unit(x, default.units)
	}
	if (!is.unit(y)) {
		y <- unit(y, default.units)
	}
	if (!is.unit(width)) {
		width <- unit(width, default.units)
	}
	if (!is.unit(height)) {
		height <- unit(height, default.units)
	}
	if (!is.unit(ch_width)) {
		ch_width <- unit(ch_width, default.units)
	}

	width <- scale * width
	height <- scale * height

	gTree(
		piece_side = piece_side,
		suit = suit,
		rank = rank,
		cfg = cfg,
		x = x,
		y = y,
		angle = angle,
		width = width,
		height = height,
		default.units = default.units,
		envir = envir,
		scale = scale,
		ch_width = ch_width,
		ch_grob = ch_grob,
		name = name,
		gp = gp,
		vp = vp,
		cl = "pp_crosshair"
	)
}

#' @export
makeContent.pp_crosshair <- function(x) {
	nn <- max(lengths(list(
		x$piece_side,
		x$suit,
		x$rank,
		x$x,
		x$y,
		x$angle,
		x$width,
		x$height,
		x$scale,
		x$bleed,
		x$ch_width
	)))
	piece_side <- rep(x$piece_side, length.out = nn)
	suit <- rep(x$suit, length.out = nn)
	rank <- rep(x$rank, length.out = nn)
	xc <- rep(x$x, length.out = nn)
	yc <- rep(x$y, length.out = nn)
	angle <- rep(x$angle, length.out = nn)

	width <- rep(x$width, length.out = nn)
	height <- rep(x$height, length.out = nn)

	scale <- rep(x$scale, length.out = nn)

	cfg <- get_cfg(x$cfg, x$envir)
	cfg <- rep(c(cfg), length.out = nn)

	ch_width <- rep(x$ch_width, length.out = nn)

	gl <- gList()
	for (i in seq(nn)) {
		name <- paste0(".", i)
		gl[[i]] <- crosshairGrobHelper(
			piece_side[i],
			suit[i],
			rank[i],
			cfg[[i]],
			xc[i],
			yc[i],
			angle[i],
			width[i],
			height[i],
			x$default.units,
			scale[i],
			name,
			ch_width[i],
			x$ch_grob
		)
	}
	setChildren(x, gl)
}

crosshairGrobHelper <- function(
	piece_side = "tile_back",
	suit = NA,
	rank = NA,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	angle = 0,
	width = NA,
	height = NA,
	default.units = "npc",
	scale = 1,
	name = "",
	ch_width = unit(0.25, "mm"),
	ch_grob = unit(0.125, "in")
) {
	if (is.na(width) || is.na(height)) {
		cfg <- as_pp_cfg(cfg)
		rank <- impute_rank(piece_side, rank, cfg)
		suit <- impute_suit(piece_side, suit, cfg)
		if (is.na(width)) {
			width <- inch(cfg$get_width(piece_side, suit, rank))
		}
		if (is.na(height)) height <- inch(cfg$get_height(piece_side, suit, rank))
	}

	if (is.na(angle)) {
		angle <- 0
	} else {
		angle <- angle %% 360
	}

	if (!is.unit(x)) {
		x <- unit(x, default.units)
	}
	if (!is.unit(y)) {
		y <- unit(y, default.units)
	}
	if (!is.unit(width)) {
		width <- unit(width, default.units)
	}
	if (!is.unit(height)) {
		height <- unit(height, default.units)
	}
	if (!is.unit(ch_width)) {
		ch_width <- unit(ch_width, default.units)
	}

	width <- scale * width
	height <- scale * height

	vp_piece <- viewport(x = x, y = y, angle = angle, width = width, height = height)

	vp_ll <- viewport(x = 0, y = 0, width = ch_width, height = ch_width)
	vp_ul <- viewport(x = 0, y = 1, width = ch_width, height = ch_width)
	vp_lr <- viewport(x = 1, y = 0, width = ch_width, height = ch_width)
	vp_ur <- viewport(x = 1, y = 1, width = ch_width, height = ch_width)

	ch_ll <- grobTree(ch_grob, name = "crosshair_ll", vp = vp_ll)
	ch_ul <- grobTree(ch_grob, name = "crosshair_ul", vp = vp_ul)
	ch_lr <- grobTree(ch_grob, name = "crosshair_lr", vp = vp_lr)
	ch_ur <- grobTree(ch_grob, name = "crosshair_ur", vp = vp_ur)

	name <- paste0("crosshair", name)

	grobTree(ch_ll, ch_ul, ch_lr, ch_ur, vp = vp_piece, name = name)
}

#' @rdname grid.crosshair
#' @param ... `crosshairGrob()` ignores; `grid.crosshair()` passes to `crosshairGrob()`.
#' @export
grid.crosshair <- function(..., draw = TRUE) {
	grob <- crosshairGrob(...)
	if (draw) {
		grid.draw(grob)
		invisible(grob)
	} else {
		grob
	}
}

#' @rdname grid.crosshair
#' @export
segmentsCrosshairGrob <- function(
	...,
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	width = unit(1, "snpc"),
	height = unit(1, "snpc"),
	default.units = "npc",
	name = NULL,
	gp = gpar(),
	vp = NULL
) {
	col <- gp$col %||% gp$fill %||% "black"
	lwd <- gp$lwd %||% 1
	lty <- gp$lty %||% "solid"

	gp <- gpar(col = col, lwd = lwd, lty = lty)

	nn <- max(lengths(list(x, y, width, height)))
	x <- rep(x, length.out = nn)
	y <- rep(y, length.out = nn)
	width <- rep(width, length.out = nn)
	height <- rep(height, length.out = nn)
	if (!is.unit(x)) {
		x <- unit(x, default.units)
	}
	if (!is.unit(y)) {
		y <- unit(y, default.units)
	}
	if (!is.unit(width)) {
		width <- unit(width, default.units)
	}
	if (!is.unit(height)) {
		height <- unit(height, default.units)
	}

	gl <- gList()
	for (i in seq(nn)) {
		vpi <- viewport(x = x[i], y = y[i], width = width[i], height = height[i])
		gl[[i]] <- grobTree(
			segmentsGrob(x0 = 0.5, x1 = 0.5, y0 = 0, y1 = 1, default.units = "snpc"),
			segmentsGrob(x0 = 0, x1 = 1, y0 = 0.5, y1 = 0.5, default.units = "snpc"),
			name = paste0("crosshair.", i),
			vp = vpi
		)
	}
	gTree(name = name, gp = gp, vp = vp, children = gl, cl = "pp_segments_crosshair")
}

#' @rdname grid.crosshair
#' @export
circledSegmentsCrosshairGrob <- function(
	...,
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	width = unit(1, "snpc"),
	height = unit(1, "snpc"),
	default.units = "npc",
	name = NULL,
	gp = gpar(),
	vp = NULL
) {
	col <- gp$col %||% gp$fill %||% "black"
	lwd <- gp$lwd %||% 1
	lty <- gp$lty %||% "solid"

	gp <- gpar(col = col, lwd = lwd, lty = lty)

	nn <- max(lengths(list(x, y, width, height)))
	x <- rep(x, length.out = nn)
	y <- rep(y, length.out = nn)
	width <- rep(width, length.out = nn)
	height <- rep(height, length.out = nn)
	if (!is.unit(x)) {
		x <- unit(x, default.units)
	}
	if (!is.unit(y)) {
		y <- unit(y, default.units)
	}
	if (!is.unit(width)) {
		width <- unit(width, default.units)
	}
	if (!is.unit(height)) {
		height <- unit(height, default.units)
	}

	gl <- gList()
	for (i in seq(nn)) {
		vpi <- viewport(x = x[i], y = y[i], width = width[i], height = height[i])
		gl[[i]] <- grobTree(
			segmentsGrob(x0 = 0.5, x1 = 0.5, y0 = 0, y1 = 1, default.units = "snpc"),
			segmentsGrob(x0 = 0, x1 = 1, y0 = 0.5, y1 = 0.5, default.units = "snpc"),
			circleGrob(r = unit(0.25, "snpc"), gp = gpar(fill = NA)),
			name = paste0("crosshair.", i),
			vp = vpi
		)
	}
	gTree(name = name, gp = gp, vp = vp, children = gl, cl = "pp_circled_segments_crosshair")
}

#' @rdname grid.crosshair
#' @export
squaredSegmentsCrosshairGrob <- function(
	...,
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	width = unit(1, "snpc"),
	height = unit(1, "snpc"),
	default.units = "npc",
	name = NULL,
	gp = gpar(),
	vp = NULL
) {
	col <- gp$col %||% gp$fill %||% "black"
	lwd <- gp$lwd %||% 1
	lty <- gp$lty %||% "solid"

	gp <- gpar(col = col, lwd = lwd, lty = lty)

	nn <- max(lengths(list(x, y, width, height)))
	x <- rep(x, length.out = nn)
	y <- rep(y, length.out = nn)
	width <- rep(width, length.out = nn)
	height <- rep(height, length.out = nn)
	if (!is.unit(x)) {
		x <- unit(x, default.units)
	}
	if (!is.unit(y)) {
		y <- unit(y, default.units)
	}
	if (!is.unit(width)) {
		width <- unit(width, default.units)
	}
	if (!is.unit(height)) {
		height <- unit(height, default.units)
	}

	gl <- gList()
	for (i in seq(nn)) {
		vpi <- viewport(x = x[i], y = y[i], width = width[i], height = height[i])
		gl[[i]] <- grobTree(
			segmentsGrob(x0 = 0.5, x1 = 0.5, y0 = 0, y1 = 1, default.units = "snpc"),
			segmentsGrob(x0 = 0, x1 = 1, y0 = 0.5, y1 = 0.5, default.units = "snpc"),
			rectGrob(width = unit(0.5, "snpc"), height = unit(0.5, "snpc"), gp = gpar(fill = NA)),
			name = paste0("crosshair.", i),
			vp = vpi
		)
	}
	gTree(name = name, gp = gp, vp = vp, children = gl, cl = "pp_squared_segments_crosshair")
}

#' @rdname grid.crosshair
#' @export
squaresCrosshairGrob <- function(
	...,
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	width = unit(1, "snpc"),
	height = unit(1, "snpc"),
	default.units = "npc",
	name = NULL,
	gp = gpar(),
	vp = NULL
) {
	fill <- gp$fill %||% gp$col %||% c("black", "white")
	if (length(fill) == 1L) {
		fill <- c(fill, opposite_col(fill))
	}

	gp1 <- gpar(col = NA, fill = fill[1L])
	gp2 <- gpar(col = NA, fill = fill[2L])

	nn <- max(lengths(list(x, y, width, height)))
	x <- rep(x, length.out = nn)
	y <- rep(y, length.out = nn)
	width <- rep(width, length.out = nn)
	height <- rep(height, length.out = nn)
	if (!is.unit(x)) {
		x <- unit(x, default.units)
	}
	if (!is.unit(y)) {
		y <- unit(y, default.units)
	}
	if (!is.unit(width)) {
		width <- unit(width, default.units)
	}
	if (!is.unit(height)) {
		height <- unit(height, default.units)
	}

	gl <- gList()
	for (i in seq(nn)) {
		vpi <- viewport(x = x[i], y = y[i], width = width[i], height = height[i])
		gl[[i]] <- grobTree(
			rectGrob(
				x = c(0.1, 0.3, 0.5, 0.5, 0.5, 0.5, 0.5, 0.7, 0.9),
				y = c(0.5, 0.5, 0.1, 0.3, 0.5, 0.7, 0.9, 0.5, 0.5),
				width = 0.1,
				height = 0.1,
				gp = gp1,
				default.units = "snpc"
			),
			rectGrob(
				x = c(0.2, 0.4, 0.5, 0.5, 0.5, 0.5, 0.6, 0.8),
				y = c(0.5, 0.5, 0.2, 0.4, 0.6, 0.8, 0.5, 0.5),
				width = 0.1,
				height = 0.1,
				gp = gp2,
				default.units = "snpc"
			),
			name = paste0("crosshair.", i),
			vp = vpi
		)
	}
	gTree(name = name, gp = gp, vp = vp, children = gl, cl = "pp_squares_crosshair")
}

# Opposite color in RGB space
opposite_col <- function(col) {
	stopifnot(length(col) == 1L)
	x <- grDevices::col2rgb(col)
	grDevices::rgb(255 - x[[1, 1]], 255 - x[[2, 1]], 255 - x[[3, 1]], maxColorValue = 255)
}

test_that("crosshairs work as expected", {
	skip_if_not_installed("vdiffr")
	library("vdiffr")

	cfg <- pp_cfg(list(border_color = NA))
	df <- data.frame(
		piece_side = "tile_face",
		suit = 2,
		rank = 2,
		x = 2,
		y = 2,
		angle = 0,
		stringsAsFactors = FALSE
	)

	expect_doppelganger("crosshairs-circled-segments", {
		pmap_piece(df, grid.piece, cfg = cfg, default.units = "in")
		pmap_piece(
			df,
			grid.crosshair,
			cfg = cfg,
			default.units = "in",
			ch_grob = circledSegmentsCrosshairGrob()
		)
	})

	expect_doppelganger("crosshairs-segments", {
		pmap_piece(df, grid.piece, cfg = cfg, default.units = "in")
		pmap_piece(
			df,
			grid.crosshair,
			cfg = cfg,
			default.units = "in",
			ch_grob = segmentsCrosshairGrob()
		)
	})

	expect_doppelganger("crosshairs-squared-segments", {
		pmap_piece(df, grid.piece, cfg = cfg, default.units = "in")
		pmap_piece(
			df,
			grid.crosshair,
			cfg = cfg,
			default.units = "in",
			ch_grob = squaredSegmentsCrosshairGrob()
		)
	})

	expect_doppelganger("crosshairs-squares", {
		pmap_piece(df, grid.piece, cfg = cfg, default.units = "in")
		pmap_piece(df, grid.crosshair, cfg = cfg, default.units = "in")
	})
})

#' Create rayvertex board game piece objects
#'
#' `piece_mesh()` creates 3d board game piece objects for use with the rayvertex package.
#' @inheritParams piece3d
#' @return A rayvertex object.
#' @examples
#' cfg <- game_systems("sans", border = FALSE)$piecepack
#' rasterize_mesh <- function(mesh) {
#'   opt <- options(cores = getOption("Ncpus"))
#'   light <- rayvertex::directional_light(c(0, 0, 1))
#'   lookat <- c(0, 0, 0)
#'   rayvertex::rasterize_scene(mesh, light_info = light, lookat = lookat)
#'   invisible(options(opt))
#' }
#' should_run_piece_mesh_example <- piecepackr:::donttest() &&
#'   requireNamespace("rayvertex", quietly = TRUE) &&
#'   all(capabilities(c("cairo", "png")))
#' if (should_run_piece_mesh_example) {
#'   mesh <- piece_mesh("tile_face", suit = 3, rank = 3, cfg = cfg)
#'   rasterize_mesh(mesh)
#' }
#' if (should_run_piece_mesh_example) {
#'   mesh <- piece_mesh("coin_back", suit = 4, rank = 2, cfg = cfg)
#'   rasterize_mesh(mesh)
#' }
#' if (should_run_piece_mesh_example) {
#'   mesh <- piece_mesh("pawn_face", suit = 1, cfg = cfg)
#'   rasterize_mesh(mesh)
#' }
#' @export
#' @seealso See \url{https://www.rayvertex.com} for more information about the `rayvertex` package.
#'          See \code{\link{geometry_utils}} for a discussion of the 3D rotation parameterization.
piece_mesh <- function(
	piece_side = "tile_back",
	suit = NA,
	rank = NA,
	cfg = pp_cfg(), # nolint
	x = 0,
	y = 0,
	z = NA,
	angle = 0,
	axis_x = 0,
	axis_y = 0,
	width = NA,
	height = NA,
	depth = NA,
	envir = NULL,
	...,
	scale = 1,
	res = 72
) {
	assert_suggested("rayvertex")
	if (is_angle(angle)) {
		angle <- as.double(angle, "degrees")
	}

	nn <- max(lengths(list(
		piece_side,
		suit,
		rank,
		x,
		y,
		z,
		angle,
		axis_x,
		axis_y,
		width,
		height,
		depth
	)))
	piece_side <- rep(piece_side, length.out = nn)
	suit <- rep(suit, length.out = nn)
	rank <- rep(rank, length.out = nn)
	x <- rep(x, length.out = nn)
	y <- rep(y, length.out = nn)
	z <- rep(z, length.out = nn)
	angle <- rep(angle, length.out = nn)
	axis_x <- rep(axis_x, length.out = nn)
	axis_y <- rep(axis_y, length.out = nn)
	width <- rep(width, length.out = nn)
	height <- rep(height, length.out = nn)
	depth <- rep(depth, length.out = nn)
	scale <- rep(scale, length.out = nn)

	cfg <- get_cfg(cfg, envir)
	cfg <- rep(c(cfg), length.out = nn)
	l <- lapply(seq(nn), function(i) {
		cfg[[i]]$rayvertex(
			piece_side[i],
			suit[i],
			rank[i],
			x[i],
			y[i],
			z[i],
			angle[i],
			axis_x[i],
			axis_y[i],
			width[i],
			height[i],
			depth[i],
			scale = scale[i],
			res = res
		)
	})
	l <- Filter(Negate(is.null), l)
	rv_from_list(l)
}

rv_piece_helper <- function(
	piece_side = "tile_back",
	suit = NA,
	rank = NA,
	cfg = pp_cfg(), # nolint
	x = 0,
	y = 0,
	z = NA,
	angle = 0,
	axis_x = 0,
	axis_y = 0,
	width = NA,
	height = NA,
	depth = NA,
	scale = 1,
	res = 72
) {
	if (scale == 0) {
		return(NULL)
	}
	l_obj <- save_piece_obj(
		piece_side,
		suit,
		rank,
		cfg,
		x = x,
		y = y,
		z = z,
		angle = angle,
		axis_x = axis_x,
		axis_y = axis_y,
		width = width,
		height = height,
		depth = depth,
		scale = scale,
		res = res
	)
	l <- lapply(l_obj$obj, rayvertex::obj_mesh)
	rv_from_list(l)
}

rv_from_list <- function(l) {
	if (length(l) == 1L) {
		l[[1L]]
	} else if (length(l) > 1L) {
		rayvertex::scene_from_list(l)
	} else {
		NULL
	}
}

#' Font utility functions
#'
#' `get_embedded_font()` returns which font is actually embedded
#' by `cairo_pdf()` for a given character.
#' `has_font()` tries to determine if a given font is available on the OS.
#' @name font_utils
#' @rdname font_utils
#' @param font A character vector of font(s).
#' @param char A character vector of character(s) to be embedded by `grid::grid.text()`
#' @return `get_embedded_font()` returns character vector of fonts that were actually embedded by `cairo_pdf()`.
#'         `NA`'s means no embedded font detected: this either means that no font
#'          was found or that a color emoji font was found and instead of a font an image was embedded.
#' @details `get_embedded_font()` depends on the suggested `pdftools` package being installed
#'          and R being compiled with Cairo support.
#'          `has_font()` depends on either the suggested `systemfonts` (preferred) or `pdftools`
#'          packages being installed.
#' @examples
#'  if (requireNamespace("pdftools", quietly = TRUE) &&
#'      capabilities("cairo") &&
#'      !piecepackr:::is_cairo_maybe_buggy()) {
#'    chars <- c("a", "\u2666")
#'    fonts <- c("sans", "Sans Noto", "Noto Sans", "Noto Sans Symbols2")
#'    try(get_embedded_font(fonts, chars))
#'  }
#'
#'  if (requireNamespace("systemfonts", quietly = TRUE) ||
#'      (requireNamespace("pdftools", quietly = TRUE) &&
#'       capabilities("cairo")) && !piecepackr:::is_cairo_maybe_buggy()) {
#'    try(has_font("Dejavu Sans"))
#'  }
#' @export
get_embedded_font <- function(font, char) {
	if (!isTRUE(capabilities("cairo"))) {
		abort("`get_embedded_font()` requires that R has been compiled with `cairo` support.")
	}
	if (!requireNamespace("pdftools", quietly = TRUE)) {
		assert_suggested("pdftools")
	}
	if (!isFALSE(getOption("piecepackr.check.cairo")) && is_cairo_maybe_buggy()) {
		warn(
			c(
				sprintf(
					"Your cairographics (version %s) may embed malformed font names.",
					grDevices::grSoftVersion()[["cairo"]]
				),
				i = "cairographics 1.17.8 is known to embed malformed font names.",
				i = "cairographics's issue tracker suggests 1.17.4 and 1.17.6 may also do this.",
				i = "See https://github.com/piecepackr/piecepackr/issues/334 for more info.",
				i = "These warnings can be disabled via `options(piecepackr.check.cairo = FALSE)`."
			),
			class = "piecepackr_buggy_cairo"
		)
	}
	df <- expand.grid(char, font, stringsAsFactors = FALSE)
	names(df) <- c("char", "requested_font")
	df$embedded_font <- NA
	for (ii in seq_len(nrow(df))) {
		df[ii, 3] <- get_embedded_font_helper(df[ii, 2], df[ii, 1])
	}
	df
}

get_embedded_font_helper <- function(font, char) {
	stopifnot(requireNamespace("pdftools", quietly = TRUE))

	file <- tempfile(fileext = ".pdf")
	on.exit(unlink(file), add = TRUE)
	grDevices::cairo_pdf(file)
	grid::grid.text(char, gp = grid::gpar(fontsize = 72, fontfamily = font))
	invisible(grDevices::dev.off())

	df <- pdftools::pdf_fonts(file)
	if (nrow(df) == 0L) {
		# probably some color emoji font used
		embedded_font <- NA
	} else {
		embedded_font <- gsub("^[^+]*\\+(.*)", "\\1", df$name)
	}
	embedded_font
}

#' @rdname font_utils
#' @export
has_font <- function(font) {
	stopifnot(length(font) == 1)
	if (requireNamespace("systemfonts", quietly = TRUE)) {
		if (packageVersion("systemfonts") >= '1.1.0') {
			font_file <- basename(systemfonts::match_fonts(family = font)$path)
		} else {
			font_file <- basename(systemfonts::match_font(family = font)$path)
		}
		grepl(simplify_font(font), simplify_font(font_file))
	} else if (
		requireNamespace("pdftools", quietly = TRUE) &&
			isTRUE(capabilities("cairo"))
	) {
		embedded_font <- get_embedded_font(font, "A")$embedded_font
		grepl(simplify_font(font), simplify_font(embedded_font))
	} else {
		warn(paste(
			"`has_font()` needs either the suggested 'systemfonts' package installed",
			"or R compiled with 'cairo' support plus the suggested 'pdftools' package installed.",
			"Conservatively returning `FALSE`."
		))
		FALSE
	}
}

simplify_font <- function(font) {
	tolower(gsub(" ", "", font))
}


# `cairo_pdf()` embedded fontnames with are definitely buggy in cairo 1.17.8
# they may also be buggy with cairo 1.17.4 and 1.17.6:
# https://gitlab.freedesktop.org/cairo/cairo/-/issues/449
is_cairo_maybe_buggy <- function() {
	cairo_version <- grDevices::grSoftVersion()[["cairo"]]
	stopifnot("cairographics is not available" = cairo_version != "")
	cairo_version <- numeric_version(cairo_version)
	if (
		cairo_version <= numeric_version("1.17.8") &&
			cairo_version >= numeric_version("1.17.4")
	) {
		TRUE
	} else {
		FALSE
	}
}

#' Calculate axis-aligned bounding box for set of game pieces
#'
#' Calculate axis-aligned bounding box (AABB) for set of game pieces
#' with and without an \dQuote{oblique projection}.
#'
#' The \dQuote{oblique projection} of a set of \eqn{(x,y,z)} points onto the xy-plane
#' is \eqn{(x + \lambda * z * cos(\alpha), y + \lambda * z * sin(\alpha))}
#' where \eqn{\lambda} is the scale factor and \eqn{\alpha} is the angle.
#'
#' @param df A data frame of game piece information with (at least) the
#'        named columns \dQuote{piece_side}, \dQuote{x}, and \dQuote{y}.
#' @inheritParams grid.piece
#' @param ... Ignored
#' @return A named list of ranges with five named elements `x`, `y`, and `z` for
#'         the axis-aligned bounding cube
#'         in xyz-space plus `x_op` and `y_op` for the axis-aligned bounding box
#'         of the \dQuote{oblique projection} onto the xy plane.
#' @examples
#'  df_tiles <- data.frame(piece_side="tile_back", x=0.5+c(3,1,3,1), y=0.5+c(3,3,1,1),
#'                         suit=NA, angle=NA, z=NA, stringsAsFactors=FALSE)
#'  df_coins <- data.frame(piece_side="coin_back", x=rep(4:1, 4), y=rep(4:1, each=4),
#'                         suit=1:16%%2+rep(c(1,3), each=8),
#'                         angle=rep(c(180,0), each=8), z=1/4+1/16, stringsAsFactors=FALSE)
#'  df <- rbind(df_tiles, df_coins)
#'
#'  aabb_piece(df, op_scale = 0)
#'  aabb_piece(df, op_scale = 1, op_angle = 45)
#'  aabb_piece(df, op_scale = 1, op_angle = -90)
#' @export
aabb_piece <- function(
	df,
	cfg = getOption("piecepackr.cfg", pp_cfg()),
	envir = getOption("piecepackr.envir"),
	op_scale = getOption("piecepackr.op_scale", 0),
	op_angle = getOption("piecepackr.op_angle", 45),
	...
) {
	rlang::local_options(affiner_angular_unit = "degrees")
	if (nrow(df) == 0) {
		return(list(
			x = c(NA_real_, NA_real_),
			y = c(NA_real_, NA_real_),
			z = c(NA_real_, NA_real_),
			x_op = c(NA_real_, NA_real_),
			y_op = c(NA_real_, NA_real_)
		))
	}

	df <- add_3d_info(df, cfg = cfg, envir = envir)
	x <- c(df$xl, df$xr)
	y <- c(df$yb, df$yt)
	z <- c(df$zb, df$zt)

	llb <- as_coord2d(as_coord3d(df$xll, df$yll, df$zb), alpha = op_angle, scale = op_scale)
	llt <- as_coord2d(as_coord3d(df$xll, df$yll, df$zt), alpha = op_angle, scale = op_scale)
	ulb <- as_coord2d(as_coord3d(df$xul, df$yul, df$zb), alpha = op_angle, scale = op_scale)
	ult <- as_coord2d(as_coord3d(df$xul, df$yul, df$zt), alpha = op_angle, scale = op_scale)
	lrb <- as_coord2d(as_coord3d(df$xlr, df$ylr, df$zb), alpha = op_angle, scale = op_scale)
	lrt <- as_coord2d(as_coord3d(df$xlr, df$ylr, df$zt), alpha = op_angle, scale = op_scale)
	urb <- as_coord2d(as_coord3d(df$xur, df$yur, df$zb), alpha = op_angle, scale = op_scale)
	urt <- as_coord2d(as_coord3d(df$xur, df$yur, df$zt), alpha = op_angle, scale = op_scale)
	x_op <- c(llb$x, llt$x, ulb$x, ult$x, lrb$x, lrt$x, urb$x, urt$x)
	y_op <- c(llb$y, llt$y, ulb$y, ult$y, lrb$y, lrt$y, urb$y, urt$y)

	list(x = range(x), y = range(y), z = range(z), x_op = range(x_op), y_op = range(y_op))
}

# #' Deprecated functions
# #'
# #' These functions are Deprecated in this release of piecepackr,
# #' they will be marked as Defunct and removed in a future version.
# #'
# #' @name piecepackr-deprecated
# NULL

#' SPDX License List data
#'
#' `spdx_license_list` is a data frame of SPDX License List data.
#'
#' @seealso See \url{https://spdx.org/licenses/} for more information.
#' @format a data frame with eight variables:
#'   \describe{
#'     \item{id}{SPDX Identifier.}
#'      \item{name}{Full name of license.
#'           For Creative Commons licenses these have been tweaked from the SPDX version
#'           to more closely match the full name used by Creative Commons Foundation. }
#'      \item{url}{URL for copy of license located at `spdx.org`}
#'      \item{fsf}{Is this license considered Free/Libre by the FSF?}
#'      \item{osi}{Is this license OSI approved?}
#'      \item{deprecated}{Has this SPDFX Identifier been deprecated by SPDX?}
#'      \item{badge}{Filename of appropriate \dQuote{button mark} badge (if any)
#'                   located in `system.file("extdata/badges", package = "piecepackr")`.}
#'      \item{url_alt}{Alternative URL for license.
#'                    Manually created for a subset of Creative Commons licenses.
#'                    Others taken from \url{https://github.com/sindresorhus/spdx-license-list}.
#'                    }
#'   }
"spdx_license_list"

#' Draw crop marks with grid
#'
#' `grid.cropmark()` draws \dQuote{crop marks} to the active graphics device.
#' `cropmarkGrob()` is its grid grob counterpart.
#' Intended for use in adding crop marks around
#' game pieces in print-and-play layouts.
#'
#' @inheritParams pieceGrob
#' @param bleed Bleed zone size to assume:
#'              \itemize{
#'                 \item{If `bleed` is a [grid::unit()] simply use it}
#'                 \item{If `bleed` is numeric then convert via `grid::unit(bleed, default.units)`}
#'                 \item{If `bleed` is `TRUE` assume 1/8 inch bleed zone size}
#'                 \item{If `bleed` is `FALSE` assume 0 inch bleed zone size}
#'              }
#' @param cm_select A string of integers from "1" to "8" indicating which
#'                  crop marks to draw.  "1" represents the top right crop mark
#'                  then we proceeding clockwise to "8" which represents the
#'                  top left crop mark.
#'                  Default "12345678" draws all eight crop marks.
#' @param cm_width Width of crop mark.
#' @param cm_length Length of crop mark.
#' @return A grid grob.
#' @examples
#' if (requireNamespace("grid", quietly = TRUE) &&
#'     piecepackr:::device_supports_unicode()) {
#'   cfg <- pp_cfg(list(mat_color = "pink", mat_width=0.05, border_color=NA))
#'   grid::grid.newpage()
#'   df <- data.frame(piece_side = "tile_face", suit = 2, rank = 2,
#'                    x = 2, y = 2, angle = 0,
#'                    stringsAsFactors = FALSE)
#'   pmap_piece(df, grid.cropmark, cfg = cfg, default.units = "in")
#'   pmap_piece(df, grid.piece, cfg = cfg, default.units = "in", bleed=TRUE)
#' }
#' if (requireNamespace("grid", quietly = TRUE) &&
#'     piecepackr:::device_supports_unicode()) {
#'   grid::grid.newpage()
#'   df <- data.frame(piece_side = "coin_back", suit = 2, rank = 2,
#'                    x = 2, y = 2, angle = 0,
#'                    stringsAsFactors = FALSE)
#'   pmap_piece(df, grid.cropmark, cfg = cfg, default.units = "in", bleed=TRUE)
#'   pmap_piece(df, grid.piece, cfg = cfg, default.units = "in", bleed=TRUE)
#' }
#' @name grid.cropmark
NULL

#' @rdname grid.cropmark
#' @export
cropmarkGrob <- function(
	...,
	piece_side = "tile_back",
	suit = NA,
	rank = NA,
	cfg = getOption("piecepackr.cfg", pp_cfg()),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	angle = 0,
	width = NA,
	height = NA,
	scale = 1,
	default.units = "npc",
	envir = getOption("piecepackr.envir"),
	name = NULL,
	gp = gpar(),
	vp = NULL,
	bleed = unit(0.125, "in"),
	cm_select = "12345678",
	cm_width = unit(0.25, "mm"),
	cm_length = unit(0.125, "in")
) {
	if (is_angle(angle)) {
		angle <- as.double(angle, "degrees")
	}

	gTree(
		piece_side = piece_side,
		suit = suit,
		rank = rank,
		cfg = cfg,
		x = x,
		y = y,
		angle = angle,
		width = width,
		height = height,
		default.units = default.units,
		envir = envir,
		scale = scale,
		bleed = bleed,
		cm_select = cm_select,
		cm_width = cm_width,
		cm_length = cm_length,
		name = name,
		gp = gp,
		vp = vp,
		cl = "pp_cropmark"
	)
}

#' @rdname grid.cropmark
#' @param ... `cropmarkGrob()` ignores; `grid.cropmark()` passes to `cropmarkGrob()`.
#' @export
grid.cropmark <- function(..., draw = TRUE) {
	grob <- cropmarkGrob(...)
	if (draw) {
		grid.draw(grob)
		invisible(grob)
	} else {
		grob
	}
}

#' @export
makeContext.pp_cropmark <- function(x) {
	x$gp$fill <- x$gp$fill %||% x$gp$col %||% get.gpar("col")$col
	x$gp$col <- NA
	x
}

#' @export
makeContent.pp_cropmark <- function(x) {
	nn <- max(lengths(list(
		x$piece_side,
		x$suit,
		x$rank,
		x$x,
		x$y,
		x$angle,
		x$width,
		x$height,
		x$scale,
		x$bleed,
		x$cm_select,
		x$cm_width,
		x$cm_length
	)))
	piece_side <- rep(x$piece_side, length.out = nn)
	suit <- rep(x$suit, length.out = nn)
	rank <- rep(x$rank, length.out = nn)
	xc <- rep(x$x, length.out = nn)
	yc <- rep(x$y, length.out = nn)
	angle <- rep(x$angle, length.out = nn)

	width <- rep(x$width, length.out = nn)
	height <- rep(x$height, length.out = nn)

	scale <- rep(x$scale, length.out = nn)
	bleed <- rep(x$bleed, length.out = nn)

	cfg <- get_cfg(x$cfg, x$envir)
	cfg <- rep(c(cfg), length.out = nn)

	cm_select <- rep(x$cm_select, length.out = nn)
	cm_width <- rep(x$cm_width, length.out = nn)
	cm_length <- rep(x$cm_length, length.out = nn)

	gl <- gList()
	for (i in seq(nn)) {
		name <- paste0(".", i)
		gl[[i]] <- cropmarkGrobHelper(
			piece_side[i],
			suit[i],
			rank[i],
			cfg[[i]],
			xc[i],
			yc[i],
			angle[i],
			width[i],
			height[i],
			x$default.units,
			scale[i],
			name,
			bleed[i],
			cm_select[i],
			cm_width[i],
			cm_length[i]
		)
	}
	setChildren(x, gl)
}

cropmarkGrobHelper <- function(
	piece_side = "tile_back",
	suit = NA,
	rank = NA,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	angle = 0,
	width = NA,
	height = NA,
	default.units = "npc",
	scale = 1,
	name = "",
	bleed = FALSE,
	cm_select = "12345678",
	cm_width = unit(0.25, "mm"),
	cm_length = unit(0.125, "in")
) {
	if (is.na(width) || is.na(height)) {
		cfg <- as_pp_cfg(cfg)
		rank <- impute_rank(piece_side, rank, cfg)
		suit <- impute_suit(piece_side, suit, cfg)
		if (is.na(width)) {
			width <- inch(cfg$get_width(piece_side, suit, rank))
		}
		if (is.na(height)) height <- inch(cfg$get_height(piece_side, suit, rank))
	}

	if (is.na(angle)) {
		angle <- 0
	} else {
		angle <- angle %% 360
	}

	if (isTRUE(bleed)) {
		bleed <- unit(0.125, "in")
	}
	if (isFALSE(bleed)) {
		bleed <- unit(0, "in")
	}
	if (!is.unit(bleed)) {
		bleed <- unit(bleed, default.units)
	}

	if (!is.unit(x)) {
		x <- unit(x, default.units)
	}
	if (!is.unit(y)) {
		y <- unit(y, default.units)
	}
	if (!is.unit(width)) {
		width <- unit(width, default.units)
	}
	if (!is.unit(height)) {
		height <- unit(height, default.units)
	}
	if (!is.unit(cm_width)) {
		cm_width <- unit(cm_width, default.units)
	}
	if (!is.unit(cm_length)) {
		cm_length <- unit(cm_length, default.units)
	}

	width <- scale * width
	height <- scale * height
	vp <- viewport(
		x = x,
		y = y,
		angle = angle,
		width = width + 2 * bleed + 2 * cm_length,
		height = height + 2 * bleed + 2 * cm_length
	)

	xc <- unit(0.5, "npc")
	yc <- unit(0.5, "npc")

	if (grepl("1", cm_select)) {
		cm1 <- rectGrob(
			x = xc + 0.5 * width,
			y = yc + 0.5 * height + bleed + 0.5 * cm_length,
			width = cm_width,
			height = cm_length,
			name = "crop_mark_1"
		)
	} else {
		cm1 <- nullGrob(name = "crop_mark_1")
	}

	if (grepl("2", cm_select)) {
		cm2 <- rectGrob(
			x = xc + 0.5 * width + bleed + 0.5 * cm_length,
			y = yc + 0.5 * height,
			width = cm_length,
			height = cm_width,
			name = "crop_mark_2"
		)
	} else {
		cm2 <- nullGrob(name = "crop_mark_2")
	}

	if (grepl("3", cm_select)) {
		cm3 <- rectGrob(
			x = xc + 0.5 * width + bleed + 0.5 * cm_length,
			y = yc - 0.5 * height,
			width = cm_length,
			height = cm_width,
			name = "crop_mark_3"
		)
	} else {
		cm3 <- nullGrob(name = "crop_mark_3")
	}

	if (grepl("4", cm_select)) {
		cm4 <- rectGrob(
			x = xc + 0.5 * width,
			y = yc - 0.5 * height - bleed - 0.5 * cm_length,
			width = cm_width,
			height = cm_length,
			name = "crop_mark_4"
		)
	} else {
		cm4 <- nullGrob(name = "crop_mark_4")
	}

	if (grepl("5", cm_select)) {
		cm5 <- rectGrob(
			x = xc - 0.5 * width,
			y = yc - 0.5 * height - bleed - 0.5 * cm_length,
			width = cm_width,
			height = cm_length,
			name = "crop_mark_5"
		)
	} else {
		cm5 <- nullGrob(name = "crop_mark_5")
	}

	if (grepl("6", cm_select)) {
		cm6 <- rectGrob(
			x = xc - 0.5 * width - bleed - 0.5 * cm_length,
			y = yc - 0.5 * height,
			width = cm_length,
			height = cm_width,
			name = "crop_mark_6"
		)
	} else {
		cm6 <- nullGrob(name = "crop_mark_6")
	}

	if (grepl("7", cm_select)) {
		cm7 <- rectGrob(
			x = xc - 0.5 * width - bleed - 0.5 * cm_length,
			y = yc + 0.5 * height,
			width = cm_length,
			height = cm_width,
			name = "crop_mark_7"
		)
	} else {
		cm7 <- nullGrob(name = "crop_mark_7")
	}

	if (grepl("8", cm_select)) {
		cm8 <- rectGrob(
			x = xc - 0.5 * width,
			y = yc + 0.5 * height + bleed + 0.5 * cm_length,
			width = cm_width,
			height = cm_length,
			name = "crop_mark_8"
		)
	} else {
		cm8 <- nullGrob(name = "crop_mark_8")
	}
	name <- paste0("crop_mark", name)

	grobTree(cm1, cm2, cm3, cm4, cm5, cm6, cm7, cm8, name = name, vp = vp)
}

# "mat" grob functions
halmaMatGrobFn <- function(width = 0.2) {
	width <- rep(width, length.out = 2)
	xy_out <- halma_xy()
	xy_in <- npc_to_in(as_coord2d(xy_out), h = 1 - width[1], w = 1 - 2.5 * width[2])
	x <- c(xy_in$x, xy_out$x)
	y <- c(xy_in$y, xy_out$y)
	id <- rep(1:2, each = length(xy_out$x))
	function(name = NULL, gp = gpar(), vp = NULL) {
		pathGrob(x, y, id = id, rule = "evenodd", name = name, gp = gp, vp = vp)
	}
}

convexMatGrobFn <- function(n_vertices, t = 90, width = 0.2) {
	t <- seq(0, 360, length.out = n_vertices + 1) + t
	r <- 0.5 - width[1]
	x_in <- 0.5 + to_x(t, r)
	y_in <- 0.5 + to_y(t, r)
	x_out <- 0.5 + to_x(t, 0.5)
	y_out <- 0.5 + to_y(t, 0.5)
	x <- c(x_in, x_out)
	y <- c(y_in, y_out)
	id <- rep(1:2, each = n_vertices + 1)
	function(name = NULL, gp = gpar(), vp = NULL) {
		pathGrob(x, y, id = id, rule = "evenodd", name = name, gp = gp, vp = vp)
	}
}

diffMatGrobFn <- function(width = 0.2, shape = pp_shape()) {
	width <- rep(width, length.out = 4)
	x_in <- c(width[4], 1 - width[2], 1 - width[2], width[4])
	y_in <- c(1 - width[1], 1 - width[1], width[3], width[3])
	vpi_x <- mean(x_in[1:2])
	vpi_y <- mean(y_in[2:3])
	vpi_w <- x_in[2] - x_in[1]
	vpi_h <- y_in[2] - y_in[3]
	function(name = NULL, gp = gpar(), vp = NULL) {
		vpi <- viewport(x = vpi_x, y = vpi_y, width = vpi_w, height = vpi_h)
		inner <- shape$shape(vp = vpi)
		shape$polyclip(inner, op = "minus", name = name, gp = gp, vp = vp)
	}
}

rectMatGrobFn <- function(width = 0.2) {
	width <- rep(width, length.out = 4)
	is_zero <- vapply(width, nigh, logical(1), y = 0, USE.NAMES = FALSE)
	if (all(is_zero)) {
		function(name = NULL, gp = gpar(), vp = NULL) nullGrob(name = name, vp = vp)
	} else if (
		sum(is_zero) == 3 ||
			all(is_zero == c(TRUE, FALSE, TRUE, FALSE)) ||
			all(is_zero == c(FALSE, TRUE, FALSE, TRUE))
	) {
		rectMatRectGrobFn(width)
	} else if (sum(is_zero) == 0) {
		rectMatPathGrobFn(width)
	} else {
		rectMatPolygonGrob(width)
	}
}

rectMatRectGrobFn <- function(width) {
	if (nigh(width[1], 0)) {
		rt <- nullGrob()
	} else {
		y <- 1 - width[1]
		rt <- polygonGrob(x = c(0, 0, 1, 1), y = c(1, y, y, 1))
	}
	if (nigh(width[2], 0)) {
		rr <- nullGrob()
	} else {
		x <- 1 - width[2]
		rr <- polygonGrob(x = c(x, x, 1, 1), y = c(1, 0, 0, 1))
	}
	if (nigh(width[3], 0)) {
		rb <- nullGrob()
	} else {
		y <- width[3]
		rb <- polygonGrob(x = c(0, 0, 1, 1), y = c(y, 0, 0, y))
	}
	if (nigh(width[4], 0)) {
		rl <- nullGrob()
	} else {
		x <- width[4]
		rl <- rectGrob(x = c(0, 0, x, x), y = c(1, 0, 0, 1))
	}
	function(name = NULL, gp = gpar(), vp = NULL) {
		grobTree(rt, rr, rb, rl, name = name, gp = gp, vp = vp)
	}
}

#### Could figure out actual polygons instead of path hack
rectMatPolygonGrob <- function(width) {
	x_out <- c(0, 1, 1, 0)
	y_out <- c(1, 1, 0, 0)
	x_in <- c(width[4], 1 - width[2], 1 - width[2], width[4])
	y_in <- c(1 - width[1], 1 - width[1], width[3], width[3])
	x <- c(x_in, x_out)
	y <- c(y_in, y_out)
	id <- rep(1:2, each = 4)
	function(name = NULL, gp = gpar(), vp = NULL) {
		pathGrob(x, y, id = id, rule = "evenodd", name = name, gp = gp, vp = vp)
	}
}

rectMatPathGrobFn <- function(width) {
	x_out <- c(0, 1, 1, 0)
	y_out <- c(1, 1, 0, 0)
	x_in <- c(width[4], 1 - width[2], 1 - width[2], width[4])
	y_in <- c(1 - width[1], 1 - width[1], width[3], width[3])
	x <- c(x_in, x_out)
	y <- c(y_in, y_out)
	id <- rep(1:2, each = 4)
	function(name = NULL, gp = gpar(), vp = NULL) {
		pathGrob(x, y, id = id, rule = "evenodd", name = name, gp = gp, vp = vp)
	}
}

## Shared "paper" utilities

A5W <- 5 # 5.83"
A5H <- 7.5 # 8.27"
a5_vp <- function() viewport(width = unit(A5W, "in"), height = unit(A5H, "in"))

draw_a5_page <- function(grob, vp) {
	pushViewport(vp)
	grid.draw(grob)
	upViewport()
}

blank_grob <- textGrob("Intentionally left blank")

is_odd <- function(x) as.logical(x %% 2)

gappend <- function(ll, g) {
	ll[[length(ll) + 1]] <- g
	ll
}

gp_title <- gpar(fontsize = 15, fontfamily = "sans", fontface = "bold")
gp_header <- gpar(fontsize = 12, fontfamily = "sans", fontface = "bold")
gp_text <- gpar(fontsize = 9, fontfamily = "sans")

htg <- function(label, x, y, just = "center", ...) {
	textGrob(label, x = inch(x), y = inch(y), just = just, gp = gp_header, ...)
}
tg <- function(label, x, y, just = "center", ...) {
	textGrob(label, x = inch(x), y = inch(y), just = just, gp = gp_text, ...)
}

a5_title_grob <- function(cfg, pieces, quietly, extra_credit = TRUE, bleed = FALSE) {
	# Title
	y_title <- unit(1, "npc") - unit(0.2, "in")
	if (is.null(cfg$title)) {
		if (!quietly) {
			inform("`cfg$title` is `NULL`, omitting title", class = "piecepackr_missing_metadata")
		}
		grob_title <- nullGrob()
	} else {
		grob_title <- textGrob(
			cfg$title,
			y = y_title,
			just = "center",
			gp = gp_title,
			name = "title"
		)
	}

	# Description
	y_description <- y_title - grobHeight(grob_title) - unit(0.2, "in")
	if (is.null(cfg$description)) {
		if (!quietly) {
			inform(
				"`cfg$description` is `NULL`, omitting description",
				class = "piecepackr_missing_metadata"
			)
		}
		grob_description <- nullGrob()
	} else {
		dtext <- paste(strwrap(cfg$description, 72), collapse = "\n")
		grob_description <- textGrob(
			dtext,
			x = 0.1,
			y = y_description,
			just = c(0, 1),
			gp = gp_text,
			name = "description"
		)
	}

	# License
	y_license <- y_description - grobHeight(grob_description) - unit(0.2, "in")
	if (is.null(cfg$spdx_id)) {
		if (!quietly) {
			inform(
				"`cfg$spdx_id` is `NULL`, omitting license",
				class = "piecepackr_missing_metadata"
			)
		}
		grob_license <- grob_lh <- grob_l <- nullGrob()
	} else {
		stopifnot(cfg$spdx_id %in% piecepackr::spdx_license_list$id)
		url <- piecepackr::spdx_license_list[cfg$spdx_id, "url_alt"]
		if (is.na(url)) {
			url <- piecepackr::spdx_license_list[cfg$spdx_id, "url"]
		}
		full_name <- piecepackr::spdx_license_list[cfg$spdx_id, "name"]
		license <- paste(c(paste("\u25cf", full_name), paste("\t", url)), collapse = "\n")
		grob_lh <- textGrob("License", x = 0.1, y = y_license, just = "left", gp = gp_header)
		badge <- piecepackr::spdx_license_list[cfg$spdx_id, "badge"]
		if (is.na(badge)) {
			grob_cc <- nullGrob()
		} else {
			cc_file <- system.file(paste0("extdata/badges/", badge), package = "piecepackr")
			current_dev <- grDevices::dev.cur() # Workaround for {grImport2} v0.2-0 bug
			cc_picture <- grImport2::readPicture(cc_file)
			if (current_dev > 1) {
				grDevices::dev.set(current_dev)
			}
			grob_cc <- grImport2::symbolsGrob(cc_picture, x = 0.50, y = 0.05, size = inch(0.9))
		}
		grob_l <- textGrob(
			license,
			x = 0.1,
			y = y_license - unit(0.2, "in"),
			just = c(0, 1),
			gp = gp_text
		)

		grob_license <- grobTree(grob_lh, grob_l, grob_cc, name = "license")
	}

	# Copyright
	y_copyright <- y_license - grobHeight(grob_lh) - grobHeight(grob_l) - unit(0.3, "in")
	if (is.null(cfg$copyright) && !quietly) {
		if (!quietly) {
			inform(
				"`cfg$copyright` is `NULL`, omitting copyright",
				class = "piecepackr_missing_metadata"
			)
		}
	}
	if (is.null(cfg$copyright) || cfg$copyright == "") {
		grob_copyright <- grob_ch <- grob_c <- nullGrob()
	} else {
		copyright <- paste(cfg$copyright, collapse = "\n")
		grob_ch <- textGrob("Copyright", x = 0.1, y = y_copyright, just = "left", gp = gp_header)
		grob_c <- textGrob(
			copyright,
			x = 0.1,
			y = y_copyright - unit(0.2, "in"),
			just = c(0, 1),
			gp = gp_text
		)
		grob_copyright <- grobTree(grob_ch, grob_c, name = "copyright")
	}

	# Credits
	y_credits <- y_copyright - grobHeight(grob_ch) - grobHeight(grob_c) - unit(0.3, "in")
	if (is.null(cfg$credit) && !quietly) {
		inform(
			"`cfg$credit` is `NULL`, omitting custom credits",
			class = "piecepackr_missing_metadata"
		)
	}
	credits <- c(
		"\u25cf This print-and-play layout was generated by piecepackr.",
		"\thttps://github.com/piecepackr/piecepackr"
	)
	if (extra_credit) {
		if ("piecepack" %in% pieces) {
			credits <- c(
				credits,
				'\u25cf The piecepack was invented by James "Kyle" Droscha. Public Domain.',
				"\thttps://ludism.org/ppwiki/AnatomyOfAPiecepack"
			)
		}
		if (!bleed && "piecepack" %in% pieces) {
			credits <- c(
				credits,
				"\u25cf Pawn saucers were invented by Karol M. Boyle. Public Domain.",
				"\thttps://web.archive.org/web/2018/http://www.piecepack.org/Accessories.html"
			)
		}
		if ("pyramids" %in% pieces) {
			credits <- c(
				credits,
				"\u25cf Piecepack pyramids were invented by Tim Schutz. Public Domain.",
				"\thttps://www.ludism.org/ppwiki/PiecepackPyramids"
			)
		}
		if ("matchsticks" %in% pieces) {
			credits <- c(
				credits,
				"\u25cf Piecepack matchsticks were invented by Dan Burkey. Public Domain.",
				"\thttps://www.ludism.org/ppwiki/PiecepackMatchsticks"
			)
		}
	} else {
		credits <- c(
			credits,
			'\u25cf The piecepack was invented by James "Kyle" Droscha. Public Domain.',
			"\thttps://ludism.org/ppwiki/AnatomyOfAPiecepack"
		)
	}
	credits <- paste(c(credits, cfg$credit), collapse = "\n")
	grob_credits <- gTree(
		name = "credits",
		children = gList(
			textGrob("Credits", x = 0.1, y = y_credits, just = "left", gp = gp_header),
			textGrob(
				credits,
				x = 0.1,
				y = y_credits - unit(0.2, "in"),
				just = c(0, 1),
				gp = gp_text
			)
		)
	)

	grobTree(
		grob_title,
		grob_description,
		grob_license,
		grob_copyright,
		grob_credits,
		name = "title_page",
		vp = a5_vp()
	)
}

cellsGrob <- function(
	nrows = 8,
	ncols = 8, # nolint
	name = NULL,
	gp = gpar(),
	vp = NULL
) {
	x <- seq(0.5 / ncols, 1 - 0.5 / ncols, length.out = ncols)
	y <- seq(0.5 / nrows, 1 - 0.5 / nrows, length.out = nrows)
	gl <- gList()
	fill <- c(gp$fill %||% NA_character_, NA_character_)
	for (i in seq(nrows)) {
		lwd <- 1
		gp_cell <- gpar(fill = cycle_elements(fill, i - 1))
		gl[[i]] <- rectGrob(
			x,
			y[i],
			width = 1 / ncols,
			height = 1 / nrows,
			default.units = "npc",
			name = paste("cell.", i),
			gp = gp_cell
		)
	}
	gp$col <- gp$col %||% NA_character_
	gp$fill <- NULL
	gp$lineend <- gp$lineend %||% "butt"
	gTree(children = gl, name = name, gp = gp, vp = vp, cl = "cells")
}

checkeredBoardGrobFn <- function(nrows = 8, ncols = 8, margin = 0) {
	# nolint
	force(nrows)
	force(ncols)
	force(margin)
	function(piece_side, suit, rank, cfg = pp_cfg()) {
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)
		gTree(
			opt = opt,
			border = TRUE,
			scale = 1,
			nrows = nrows,
			ncols = ncols,
			margin = margin,
			name = piece_side,
			cl = c("checkered_board", "basic_piece_side")
		)
	}
}

#' @export
makeContent.checkered_board <- function(x) {
	opt <- x$opt
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	background_grob <- shape$shape(
		gp = gpar(col = NA, fill = opt$background_color),
		name = "background"
	)

	vp_cell <- viewport(
		width = x$ncols / (x$ncols + 2 * x$margin),
		height = x$nrows / (x$nrows + 2 * x$margin)
	)
	cell_grob <- cellsGrob(
		x$nrows,
		x$ncols,
		gp = gpar(fill = opt$gridline_color),
		name = "cells",
		vp = vp_cell
	)

	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- shape$shape(gp = gp_border, name = "border")
	} else {
		border_grob <- nullGrob(name = "border")
	}

	gl <- gList(background_grob, cell_grob, border_grob)
	setChildren(x, gl)
}

linedBoardGrobFn <- function(nrows = 8L, ncols = 8L, margin = 0) {
	# nolint
	force(nrows)
	force(ncols)
	force(margin)
	function(piece_side, suit, rank, cfg = pp_cfg()) {
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)
		gTree(
			opt = opt,
			border = TRUE,
			scale = 1,
			nrows = nrows,
			ncols = ncols,
			margin = margin,
			name = piece_side,
			cl = c("lined_board", "basic_piece_side")
		)
	}
}

#' @export
makeContent.lined_board <- function(x) {
	opt <- x$opt
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	background_grob <- shape$shape(
		gp = gpar(col = NA, fill = opt$background_color),
		name = "background"
	)

	vp_cell <- viewport(
		width = x$ncols / (x$ncols + 2 * x$margin),
		height = x$nrows / (x$nrows + 2 * x$margin)
	)
	gp_cell <- gpar(col = opt$gridline_color, lex = opt$gridline_lex)
	cell_grob <- cellsGrob(x$nrows, x$ncols, gp = gp_cell, name = "cells", vp = vp_cell)

	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- shape$shape(
			gp = gpar(col = opt$border_color, fill = NA, lex = opt$border_lex),
			name = "border"
		)
	} else {
		border_grob <- nullGrob(name = "border")
	}

	gl <- gList(background_grob, cell_grob, border_grob)
	setChildren(x, gl)
}

holedBoardGrobFn <- function(nrows = 4L, ncols = 4L, margin = 0) {
	# nolint
	force(nrows)
	force(ncols)
	force(margin)
	function(piece_side, suit, rank, cfg = pp_cfg()) {
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)
		gTree(
			opt = opt,
			border = TRUE,
			scale = 1,
			nrows = nrows,
			ncols = ncols,
			margin = margin,
			name = piece_side,
			cl = c("holed_board", "basic_piece_side")
		)
	}
}

npc2snpc <- function(x) {
	vapply(
		x,
		function(x) {
			max(
				convertX(unit(x, "npc"), "npc", valueOnly = TRUE),
				convertY(unit(x, "npc"), "npc", valueOnly = TRUE)
			)
		},
		FUN.VALUE = numeric(1L)
	)
}

#' @export
grobCoords.holed_board <- function(x, closed, ...) {
	pushViewport(x$vp)
	xyi <- xyi_holed_board(x$opt, x$nrows, x$ncols, x$margin)
	popViewport()
	grobCoords(
		pathGrob(
			x = xyi$x,
			y = xyi$y,
			id = xyi$id,
			rule = "winding",
			default.units = "snpc",
			vp = x$vp
		),
		closed = closed,
		...
	)
}

xyi_holed_board <- function(opt, nrows, ncols, margin) {
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)
	xys <- shape$npc_coords
	xs <- npc2snpc(xys$x)
	ys <- npc2snpc(xys$y)
	dfs <- data.frame(x = xs, y = ys, id = 0L)

	xc <- rep(seq.int(ncols) - 0.5 + margin, times = nrows) / (ncols + 2 * margin)
	yc <- rep(seq.int(nrows) - 0.5 + margin, each = ncols) / (nrows + 2 * margin)
	xc <- npc2snpc(xc)
	yc <- npc2snpc(yc)
	r <- RADIUS_BOARD_HOLES / (min(nrows, ncols) + 2 * margin)

	p <- regular_ngon_polygon2d(36L, x = 0, y = 0, radius = r, theta = degrees(0))
	x_offsets <- rev_shift(p$x)
	y_offsets <- rev_shift(p$y)
	l <- purrr::pmap(list(xc = xc, yc = yc, r = r, id = seq_along(xc)), function(xc, yc, r, id) {
		data.frame(x = x_offsets + xc, y = y_offsets + yc, id = id)
	})
	dfc <- do.call(rbind, l)
	rbind(dfs, dfc)
}

#' @export
makeContent.holed_board <- function(x) {
	opt <- x$opt
	xyi <- xyi_holed_board(x$opt, x$nrows, x$ncols, x$margin)

	background_grob <- pathGrob(
		x = xyi$x,
		y = xyi$y,
		id = xyi$id,
		default.units = "snpc",
		gp = gpar(col = NA, fill = opt$background_color),
		rule = "winding",
		name = "background"
	)

	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- pathGrob(
			x = xyi$x,
			y = xyi$y,
			id = xyi$id,
			default.units = "snpc",
			gp = gpar(col = opt$border_color, fill = NA, lex = opt$border_lex),
			rule = "winding",
			name = "border"
		)
	} else {
		border_grob <- nullGrob(name = "border")
	}

	gl <- gList(background_grob, border_grob)
	setChildren(x, gl)
}


cycle_elements <- function(x, n = 1) {
	l <- length(x)
	if (l < 2 || abs(n) == l || n == 0) {
		x
	} else if (n < -l) {
		cycle_elements(cycle_elements(x, -l), n + l)
	} else if (n < 0) {
		n <- -n
		c(x[(l - n + 1):l], x[1:(l - n)])
	} else if (n < l) {
		c(x[(n + 1):l], x[1:n])
	} else {
		cycle_elements(cycle_elements(x, l), n - l)
	}
}

alquerqueBoardGrobFn <- function() {
	function(piece_side, suit, rank, cfg = pp_cfg()) {
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)

		gTree(
			opt = opt,
			scale = 1,
			border = TRUE,
			name = piece_side,
			cl = c("alquerque_board", "basic_piece_side")
		)
	}
}

#' @export
makeContent.alquerque_board <- function(x) {
	opt <- x$opt
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	background_grob <- shape$shape(
		gp = gpar(col = NA, fill = opt$background_color),
		name = "background"
	)

	gp_line <- gpar(col = NA, fill = opt$gridline_color)
	inner_grob <- alquerqueInnerGrob(gp = gp_line, name = "lines_and_dots")
	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- shape$shape(
			gp = gpar(col = opt$border_color, fill = NA, lex = opt$border_lex),
			name = "border"
		)
	} else {
		border_grob <- nullGrob(name = "border")
	}

	gl <- gList(background_grob, inner_grob, border_grob)
	setChildren(x, gl)
}

morrisBoardGrobFn <- function(n_pieces = 12) {
	force(n_pieces)
	function(piece_side, suit, rank, cfg = pp_cfg()) {
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)

		gTree(
			opt = opt,
			n_pieces = n_pieces,
			scale = 1,
			border = TRUE,
			name = piece_side,
			cl = c("morris_board", "basic_piece_side")
		)
	}
}

#' @export
makeContent.morris_board <- function(x) {
	n_pieces <- x$n_pieces
	opt <- x$opt
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	background_grob <- shape$shape(
		gp = gpar(col = NA, fill = opt$background_color),
		name = "background"
	)

	gp_line <- gpar(col = NA, fill = opt$gridline_color)
	if (n_pieces < 3) {
		inner_grob <- morrisBoard3Grob(gp = gp_line, name = "lines_and_dots", diag = FALSE)
	} else if (n_pieces < 5) {
		inner_grob <- morrisBoard3Grob(gp = gp_line, name = "lines_and_dots")
	} else if (n_pieces < 7) {
		inner_grob <- morrisBoard6Grob(gp = gp_line, name = "lines_and_dots")
	} else if (n_pieces == 7) {
		inner_grob <- morrisBoard6Grob(gp = gp_line, name = "lines_and_dots", cross = TRUE)
	} else if (n_pieces < 11) {
		inner_grob <- morrisBoard9Grob(gp = gp_line, name = "lines_and_dots")
	} else {
		inner_grob <- morrisBoard9Grob(gp = gp_line, name = "lines_and_dots", diag = TRUE)
	}

	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- shape$shape(
			gp = gpar(col = opt$border_color, fill = NA, lex = opt$border_lex),
			name = "border"
		)
	} else {
		border_grob <- nullGrob(name = "border")
	}

	gl <- gList(background_grob, inner_grob, border_grob)
	setChildren(x, gl)
}

alquerqueInnerGrob <- function(gp = gpar(), name = NULL, diag = TRUE) {
	w <- 1 / 64
	wd <- w / sqrt(2) / 2
	vlines <- rectGrob(x = c(0, 1, 2, 3, 4) / 4, y = 0.5, width = w, height = 1 + w)
	hlines <- rectGrob(x = 0.5, y = c(0, 1, 2, 3, 4) / 4, height = w, width = 1 + w)
	tl_br1 <- polygonGrob(
		x = c(1 - wd, 1 + wd, 1 + wd, 0 + wd, 0 - wd, 0 - wd, 1 - wd),
		y = c(0 - wd, 0 - wd, 0 + wd, 1 + wd, 1 + wd, 1 - wd, 0 - wd)
	)
	tl_br2 <- polygonGrob(
		x = c(1 / 2 - wd, 1 / 2 + wd, 1 / 2 + wd, 0 + wd, 0 - wd, 0 - wd, 1 / 2 - wd),
		y = c(0 - wd, 0 - wd, 0 + wd, 1 / 2 + wd, 1 / 2 + wd, 1 / 2 - wd, 0 - wd)
	)
	tl_br3 <- polygonGrob(
		x = c(1 - wd, 1 + wd, 1 + wd, 1 / 2 + wd, 1 / 2 - wd, 1 / 2 - wd, 1 - wd),
		y = c(1 / 2 - wd, 1 / 2 - wd, 1 / 2 + wd, 1 + wd, 1 + wd, 1 - wd, 1 / 2 - wd)
	)
	tr_bl1 <- polygonGrob(
		x = c(0 + wd, 0 - wd, 0 - wd, 1 - wd, 1 + wd, 1 + wd, 0 + wd),
		y = c(0 - wd, 0 - wd, 0 + wd, 1 + wd, 1 + wd, 1 - wd, 0 - wd)
	)
	tr_bl2 <- polygonGrob(
		x = c(0 + wd, 0 - wd, 0 - wd, 1 / 2 - wd, 1 / 2 + wd, 1 / 2 + wd, 0 + wd),
		y = c(1 / 2 - wd, 1 / 2 - wd, 1 / 2 + wd, 1 + wd, 1 + wd, 1 - wd, 1 / 2 - wd)
	)
	tr_bl3 <- polygonGrob(
		x = c(1 / 2 + wd, 1 / 2 - wd, 1 / 2 - wd, 1 - wd, 1 + wd, 1 + wd, 1 / 2 + wd),
		y = c(0 - wd, 0 - wd, 0 + wd, 1 / 2 + wd, 1 / 2 + wd, 1 / 2 - wd, 0 - wd)
	)
	dots <- circleGrob(x = rep(0:4, 5) / 4, y = rep(0:4, each = 5) / 4, r = 1.5 * w)

	grobTree(
		vlines,
		hlines,
		tl_br1,
		tl_br2,
		tl_br3,
		tr_bl1,
		tr_bl2,
		tr_bl3,
		dots,
		name = name,
		cl = "alquerque_board_inner",
		gp = gp,
		vp = viewport(width = 4 / 5, height = 4 / 5)
	)
}

morrisBoard3Grob <- function(gp = gpar(), name = NULL, diag = TRUE) {
	w <- 1 / 64
	wd <- w / sqrt(2) / 2
	vlines <- rectGrob(x = c(0, 1, 2) / 2, y = 0.5, width = w, height = 1 + w)
	hlines <- rectGrob(x = 0.5, y = c(0, 1, 2) / 2, height = w, width = 1 + w)
	if (diag) {
		tl_br <- polygonGrob(
			x = c(1 - wd, 1 + wd, 1 + wd, 0 + wd, 0 - wd, 0 - wd, 1 - wd),
			y = c(0 - wd, 0 - wd, 0 + wd, 1 + wd, 1 + wd, 1 - wd, 0 - wd)
		)
		tr_bl <- polygonGrob(
			x = c(0 + wd, 0 - wd, 0 - wd, 1 - wd, 1 + wd, 1 + wd, 0 + wd),
			y = c(0 - wd, 0 - wd, 0 + wd, 1 + wd, 1 + wd, 1 - wd, 0 - wd)
		)
	} else {
		tl_br <- NULL
		tr_bl <- NULL
	}
	dots <- circleGrob(x = rep(c(0, 1 / 2, 1), 3), y = rep(c(0, 1 / 2, 1), each = 3), r = 1.5 * w)

	grobTree(
		vlines,
		hlines,
		tl_br,
		tr_bl,
		dots,
		name = name,
		cl = "morris_board_inner",
		gp = gp,
		vp = viewport(width = 2 / 3, height = 2 / 3)
	)
}
morrisBoard6Grob <- function(gp = gpar(), name = NULL, cross = FALSE) {
	w <- 1 / 64
	wd <- w / sqrt(2) / 2
	if (cross) {
		dots <- circleGrob(
			x = rep.int(c(0, 1, 2, 3, 4), c(3, 3, 5, 3, 3)) / 4,
			y = c(0, 2, 4, 1:3, 0:4, 1:3, 0, 2, 4) / 4,
			r = 1.5 * w
		)
		vlines <- rectGrob(
			x = c(0, 1, 2, 3, 4) / 4,
			y = c(2, 2, 2, 2, 2) / 4,
			width = w,
			height = c(4, 2, 4, 2, 4) / 4 + w
		)
		hlines <- rectGrob(
			x = c(2, 2, 2, 2, 2) / 4,
			y = c(0, 1, 2, 3, 4) / 4,
			height = w,
			width = c(4, 2, 4, 2, 4) / 4 + w
		)
	} else {
		dots <- circleGrob(
			x = rep.int(c(0, 1, 2, 3, 4), c(3, 3, 4, 3, 3)) / 4,
			y = c(0, 2, 4, 1, 2, 3, 0, 1, 3, 4, 1, 2, 3, 0, 2, 4) / 4,
			r = 1.5 * w
		)
		vlines <- rectGrob(
			x = c(0, 1, 2, 2, 3, 4) / 4,
			y = c(2, 2, 0.5, 3.5, 2, 2) / 4,
			width = w,
			height = c(4, 2, 1, 1, 2, 4) / 4 + w
		)
		hlines <- rectGrob(
			x = c(2, 2, 0.5, 3.5, 2, 2) / 4,
			y = c(0, 1, 2, 2, 3, 4) / 4,
			height = w,
			width = c(4, 2, 1, 1, 2, 4) / 4 + w
		)
	}

	grobTree(
		vlines,
		hlines,
		dots,
		name = name,
		cl = "morris_board_inner",
		gp = gp,
		vp = viewport(width = 4 / 5, height = 4 / 5)
	)
}
morrisBoard9Grob <- function(gp = gpar(), name = NULL, diag = FALSE) {
	w <- 1 / 64
	wd <- w / sqrt(2) / 2
	vlines <- rectGrob(
		x = c(0, 1, 2, 3, 3, 4, 5, 6) / 6,
		y = c(3, 3, 3, 1, 5, 3, 3, 3) / 6,
		width = w,
		height = c(6, 4, 2, 2, 2, 2, 4, 6) / 6 + w
	)
	hlines <- rectGrob(
		x = c(3, 3, 3, 1, 5, 3, 3, 3) / 6,
		y = c(0, 1, 2, 3, 3, 4, 5, 6) / 6,
		height = w,
		width = c(6, 4, 2, 2, 2, 2, 4, 6) / 6 + w
	)
	dots <- circleGrob(
		x = rep.int(c(0, 1, 2, 3, 4, 5, 6), c(3, 3, 3, 6, 3, 3, 3)) / 6,
		y = c(0, 3, 6, 1, 3, 5, 2, 3, 4, 0, 1, 2, 4, 5, 6, 2, 3, 4, 1, 3, 5, 0, 3, 6) / 6,
		r = 1.5 * w
	)
	if (diag) {
		ul <- polygonGrob(
			x = c(2 / 6 - wd, 2 / 6 + wd, 2 / 6 + wd, 0 + wd, 0 - wd, 0 - wd, 2 / 6 - wd),
			y = c(0 - wd, 0 - wd, 0 + wd, 2 / 6 + wd, 2 / 6 + wd, 2 / 6 - wd, 0 - wd) + 4 / 6
		)
		br <- polygonGrob(
			x = c(2 / 6 - wd, 2 / 6 + wd, 2 / 6 + wd, 0 + wd, 0 - wd, 0 - wd, 2 / 6 - wd) + 4 / 6,
			y = c(0 - wd, 0 - wd, 0 + wd, 2 / 6 + wd, 2 / 6 + wd, 2 / 6 - wd, 0 - wd)
		)
		ur <- polygonGrob(
			x = c(0 + wd, 0 - wd, 0 - wd, 2 / 6 - wd, 2 / 6 + wd, 2 / 6 + wd, 0 + wd) + 4 / 6,
			y = c(0 - wd, 0 - wd, 0 + wd, 2 / 6 + wd, 2 / 6 + wd, 2 / 6 - wd, 0 - wd) + 4 / 6
		)
		bl <- polygonGrob(
			x = c(0 + wd, 0 - wd, 0 - wd, 2 / 6 - wd, 2 / 6 + wd, 2 / 6 + wd, 0 + wd),
			y = c(0 - wd, 0 - wd, 0 + wd, 2 / 6 + wd, 2 / 6 + wd, 2 / 6 - wd, 0 - wd)
		)
	} else {
		ul <- NULL
		ur <- NULL
		br <- NULL
		bl <- NULL
	}

	grobTree(
		vlines,
		hlines,
		dots,
		ul,
		ur,
		br,
		bl,
		name = name,
		cl = "morris_board_inner",
		gp = gp,
		vp = viewport(width = 6 / 7, height = 6 / 7)
	)
}

#' Save Wavefront OBJ files of board game pieces
#'
#' `save_piece_obj()` saves Wavefront OBJ files (including associated MTL and texture image) of board game pieces.
#' @inheritParams pieceGrob
#' @param axis_x First coordinate of the axis unit vector.
#' @param axis_y Second coordinate of the axis unit vector.
#' @param filename Name of Wavefront OBJ object.  By default `tempfile(fileext = ".obj")`.
#' @param res Resolution of the faces.
#' @return A data frame with named elements "obj", "mtl", "png" with the created filenames.
#' @examples
#'     if (all(capabilities(c("cairo", "png")))) {
#'       cfg <- game_systems("sans", border = FALSE)$dominoes
#'       files <- save_piece_obj("tile_face", suit = 3+1, rank=6+1, cfg = cfg)
#'       print(files)
#'     }
#' @seealso See \code{\link{geometry_utils}} for a discussion of the 3D rotation parameterization.
#' @export
save_piece_obj <- function(
	piece_side = "tile_face",
	suit = 1,
	rank = 1,
	cfg = getOption("piecepackr.cfg", pp_cfg()),
	...,
	x = 0,
	y = 0,
	z = 0,
	angle = 0,
	axis_x = 0,
	axis_y = 0,
	width = NA,
	height = NA,
	depth = NA,
	envir = getOption("piecepackr.envir"),
	filename = tempfile(fileext = "%03d.obj"),
	scale = 1,
	res = 72
) {
	opt <- options(piecepackr.op_scale = 0)
	on.exit(options(opt), add = TRUE)
	if (is_angle(angle)) {
		angle <- as.double(angle, "degrees")
	}

	nn <- max(lengths(list(
		piece_side,
		suit,
		rank,
		x,
		y,
		z,
		angle,
		axis_x,
		axis_y,
		width,
		height,
		depth
	)))
	piece_side <- rep(piece_side, length.out = nn)
	suit <- rep(suit, length.out = nn)
	rank <- rep(rank, length.out = nn)
	x <- rep(x, length.out = nn)
	y <- rep(y, length.out = nn)
	z <- rep(z, length.out = nn)
	angle <- rep(angle, length.out = nn)
	axis_x <- rep(axis_x, length.out = nn)
	axis_y <- rep(axis_y, length.out = nn)
	width <- rep(width, length.out = nn)
	height <- rep(height, length.out = nn)
	depth <- rep(depth, length.out = nn)
	scale <- rep(scale, length.out = nn)

	cfg <- get_cfg(cfg, envir)
	cfg <- rep(c(cfg), length.out = nn)

	filename <- rep(filename, length.out = nn)
	if (all(grepl("%[#0+=-]*[0-9.]*[diouxX]", filename))) {
		filename <- sprintf(filename, seq_along(filename))
	}
	filename <- normalizePath(filename, mustWork = FALSE)
	stopifnot(!anyDuplicated(filename))

	l <- lapply(seq.int(nn), function(i) {
		df <- save_piece_obj_helper(
			piece_side[i],
			suit[i],
			rank[i],
			cfg[[i]],
			x[i],
			y[i],
			z[i],
			angle[i],
			axis_x[i],
			axis_y[i],
			width[i],
			height[i],
			depth[i],
			filename[i],
			scale[i],
			res
		)
		df
	})
	do.call(rbind, l)
}

save_piece_obj_helper <- function(
	piece_side = "tile_face",
	suit = 1,
	rank = 1,
	cfg = getOption("piecepackr.cfg", pp_cfg()),
	x = 0,
	y = 0,
	z = 0,
	angle = 0,
	axis_x = 0,
	axis_y = 0,
	width = NA,
	height = NA,
	depth = NA,
	filename = tempfile(fileext = ".obj"),
	scale = 1,
	res = 72
) {
	cfg <- as_pp_cfg(cfg)
	suit <- ifelse(is.na(suit), 1, suit)
	rank <- ifelse(is.na(rank), 1, rank)
	if (is.na(angle)) {
		angle <- 0
	}
	if (is.na(axis_x)) {
		axis_x <- 0
	}
	if (is.na(axis_y)) {
		axis_y <- 0
	}
	if (is.na(width)) {
		width <- cfg$get_width(piece_side, suit, rank)
	}
	if (is.na(height)) {
		height <- cfg$get_height(piece_side, suit, rank)
	}
	if (is.na(depth)) {
		depth <- cfg$get_depth(piece_side, suit, rank)
	}
	if (is.na(z)) {
		z <- 0.5 * depth
	}
	width <- scale * width
	height <- scale * height
	depth <- scale * depth

	l <- cfg$save_obj(
		piece_side,
		suit,
		rank,
		x,
		y,
		z,
		angle,
		axis_x,
		axis_y,
		width,
		height,
		depth,
		filename,
		res
	)
	as.data.frame(l)
}

write_mtl <- function(mtl_filename, png_filename) {
	writeLines(c("newmtl material_0", paste("map_Kd", png_filename)), mtl_filename)
}

# 1,2,3,4 -> 1,4,3,2
rev_shift <- function(x) c(x[1], rev(x[-1]))

# (2-sided token) rotation matrix to move from "face" to `side`
side_R <- function(side) {
	switch(
		side,
		back = R_y(180),
		base = R_x(-90),
		left = R_y(-90) %*% R_z(-90),
		right = R_y(90) %*% R_z(90),
		top = R_x(90) %*% R_z(180),
		diag(3)
	)
}
# (2-sided token) rotation matrix to return to "face" from `side`
side_R_rev <- function(side) {
	switch(
		side,
		back = R_y(-180),
		base = R_x(90),
		left = R_z(90) %*% R_y(90),
		right = R_z(-90) %*% R_y(-90),
		top = R_z(-180) %*% R_x(-90),
		diag(3)
	)
}

get_scaling_factors <- function(side, width, height, depth) {
	w <- switch(side, left = depth, right = depth, width)
	h <- switch(side, back = height, face = height, left = width, right = width, depth)
	d <- switch(side, base = height, top = height, left = height, right = height, depth)
	list(width = w, height = h, depth = d)
}

# two-sided objects
save_2s_obj <- function(
	piece_side = "tile_face",
	suit = 1,
	rank = 1,
	cfg = pp_cfg(),
	...,
	x = 0,
	y = 0,
	z = 0,
	angle = 0,
	axis_x = 0,
	axis_y = 0,
	width = NA,
	height = NA,
	depth = NA,
	filename = tempfile(fileext = ".obj"),
	res = 72
) {
	cfg <- as_pp_cfg(cfg)
	piece <- get_piece(piece_side)
	side <- get_side(piece_side)
	opt <- cfg$get_piece_opt(paste0(piece, "_face"), suit, rank)
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	# geometric vertices
	whd <- get_scaling_factors(side, width, height, depth)
	pc <- as_coord3d(x, y, z)
	R <- side_R(side) %*% AA_to_R(angle, axis_x, axis_y)
	xyz <- Token2S$new(shape, whd, pc, R)$xyz

	# texture coordinates
	back <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		!opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)
	xy_npc <- as_coord2d(shape$npc_coords)
	xy_npc_back <- as_coord2d(back$npc_coords)
	xy_vt_f <- xy_npc$scale(0.4, 1)$translate(as_coord2d(0.025, 0))
	xy_vt_b <- xy_npc_back$scale(0.4, 1)$translate(as_coord2d(0.575, 0))
	xy_vt_e <- list(x = c(0.52, 0.48, 0.48, 0.52), y = c(0, 0, 1, 1))
	vt <- list(x = c(xy_vt_f$x, xy_vt_b$x, xy_vt_e$x), y = c(xy_vt_f$y, xy_vt_b$y, xy_vt_e$y))

	# textured face elements
	nv <- length(xyz) / 2
	f <- list()
	f[[1]] <- list(v = seq(nv), vt = seq(nv)) # top
	#### #217 "coin_arrangement"
	f[[2]] <- list(v = rev_shift(nv + seq(nv)), vt = rev_shift(nv + seq(nv))) # bottom
	# sides
	for (i in seq(nv)) {
		ir <- i %% nv + 1
		il <- ir %% nv + 1
		f[[2 + i]] <- list(v = c(ir + nv, il + nv, il, ir), vt = 2 * nv + 1:4)
	}

	ext <- tools::file_ext(filename)
	mtl_filename <- gsub(paste0("\\.", ext, "$"), ".mtl", filename)
	png_filename <- gsub(paste0("\\.", ext, "$"), ".png", filename)

	write_obj(filename, v = xyz, vt = vt, f = f)
	write_2s_texture(piece_side, suit, rank, cfg, filename = png_filename, res = res)

	invisible(list(obj = filename, mtl = mtl_filename, png = png_filename))
}

write_2s_texture <- function(
	piece_side = "tile_face",
	suit = 1,
	rank = 1,
	cfg = pp_cfg(),
	...,
	filename = tempfile(fileext = ".png"),
	res = 72
) {
	current_dev <- grDevices::dev.cur()
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	piece <- get_piece(piece_side)
	piece_face <- paste0(piece, "_face")
	piece_back <- paste0(piece, "_back")
	height <- cfg$get_height(piece_face, suit, rank)
	width <- cfg$get_width(piece_face, suit, rank)
	edge_color <- cfg$get_piece_opt(piece_face, suit, rank)$edge_color

	args <- list(
		filename = filename,
		height = height,
		width = 2.5 * width,
		units = "in",
		res = res,
		bg = "transparent"
	)
	if (capabilities("cairo")) {
		args$type <- "cairo"
	}
	do.call(grDevices::png, args)

	# front
	pushViewport(viewport(x = 0.225, width = 0.45))
	draw_piece_and_bleed(piece_face, suit, rank, cfg)
	popViewport()

	# edge
	pushViewport(viewport(x = 0.5, width = 0.1))
	grid.rect(gp = gpar(col = "transparent", fill = edge_color))
	popViewport()

	# back
	pushViewport(viewport(x = 0.775, width = 0.45))
	draw_piece_and_bleed(piece_back, suit, rank, cfg)
	popViewport()
	grDevices::dev.off()

	invisible(filename)
}

#' Alternative Wavefront OBJ file generators
#'
#' These are alternative Wavefront OBJ generators intended to be used as a `obj_fn` attribute
#' in a [pp_cfg()] \dQuote{configuration list}.
#' `save_ellipsoid_obj()` saves an ellipsoid with a color equal to that piece's `background_color`.
#' `save_peg_doll_obj()` saves a \dQuote{peg doll} style doll with a color equal to that piece's `edge_color`
#' with a \dQuote{pawn belt} around it's waste from that suit's and rank's `belt_face`.
#' @seealso See [pp_cfg()] for a discussion of \dQuote{configuration lists}.
#'          Wavefront OBJ file generators are used by [save_piece_obj()] and (by default)
#'          [piece3d()] (`rgl` wrapper), [piece()] (`rayrender` wrapper),
#'          and [piece_mesh()] (`rayvertex` wrapper).
#' @inheritParams save_piece_obj
#' @param subdivide Increasing this value makes for a smoother ellipsoid (and larger OBJ file and slower render).
#'                  See [rgl::ellipse3d()].
#' @rdname obj_fns
#' @export
save_ellipsoid_obj <- function(
	piece_side = "bit_face",
	suit = 1,
	rank = 1,
	cfg = getOption("piecepackr.cfg", pp_cfg()),
	...,
	x = 0,
	y = 0,
	z = 0,
	angle = 0,
	axis_x = 0,
	axis_y = 0,
	width = NA,
	height = NA,
	depth = NA,
	filename = tempfile(fileext = ".obj"),
	subdivide = 3
) {
	assert_suggested("rgl")
	if (is_angle(angle)) {
		angle <- as.double(angle, "degrees")
	}

	cfg <- as_pp_cfg(cfg)
	piece <- get_piece(piece_side)
	side <- get_side(piece_side)

	# geometric vertices
	R <- side_R(side) %*% AA_to_R(angle, axis_x, axis_y)
	whd <- get_scaling_factors(side, width, height, depth)
	e <- rgl::ellipse3d(diag(3), t = 0.5, subdivide = subdivide, smooth = FALSE)
	xyz <- as.data.frame(t(e$vb))
	names(xyz) <- c("x", "y", "z", "u")
	xyz <- as_coord3d(xyz)$scale(whd$width, whd$height, whd$depth)$translate(as_coord3d(x, y, z))

	# texture coordinates
	xy_vt <- list(x = c(0, 0, 1, 1), y = c(1, 0, 0, 1))

	# face elements
	f <- lapply(seq.int(ncol(e$ib)), function(x) list(v = e$ib[, x], vt = 1:4))

	ext <- tools::file_ext(filename)
	mtl_filename <- gsub(paste0("\\.", ext, "$"), ".mtl", filename)
	png_filename <- gsub(paste0("\\.", ext, "$"), ".png", filename)

	write_obj(filename, v = xyz, vt = xy_vt, f = f)

	current_dev <- grDevices::dev.cur()
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	opt <- cfg$get_piece_opt(paste0(piece, "_face"), suit, rank)
	grDevices::png(png_filename, bg = opt$background_color)
	grid.newpage()
	grDevices::dev.off()

	invisible(list(obj = filename, mtl = mtl_filename, png = png_filename))
}

#' @rdname obj_fns
#' @export
save_peg_doll_obj <- function(
	piece_side = "pawn_top",
	suit = 1,
	rank = 1,
	cfg = getOption("piecepackr.cfg", pp_cfg()),
	...,
	x = 0,
	y = 0,
	z = 0,
	angle = 0,
	axis_x = 0,
	axis_y = 0,
	width = NA,
	height = NA,
	depth = NA,
	filename = tempfile(fileext = ".obj"),
	res = 72
) {
	assert_suggested("rgl")
	if (is_angle(angle)) {
		angle <- as.double(angle, "degrees")
	}

	cfg <- as_pp_cfg(cfg)
	piece <- get_piece(piece_side)
	side <- get_side(piece_side)

	R <- side_R(side) %*% AA_to_R(angle, axis_x, axis_y)
	whd <- get_scaling_factors(side, width, height, depth)
	pc <- as_coord3d(x, y, z)
	# normalized peg doll is laying down face up
	width <- whd$width
	height <- whd$height
	depth <- whd$depth
	if (!nigh(width, depth)) {
		warn("Base of peg doll is not circular")
	}
	# base
	circle <- as_coord2d(pp_shape("convex72", back = TRUE)$npc_coords)$translate(as_coord2d(
		-0.5,
		-0.5
	))
	xyz_base <- data.frame(x = circle$x, y = -0.5, z = circle$y)
	vt_base <- as.data.frame(circle$clone()$scale(0, 0.2)$translate(as_coord2d(0.5, 0.1)))[, c(
		"x",
		"y"
	)]
	n_base <- nrow(xyz_base)
	f_base <- list(list(v = seq(n_base), vt = seq(n_base)))

	# body below
	belt <- cfg$get_height("belt_face", suit, rank) / height
	head <- width / height
	neck <- 0.1 * head
	y_below <- (1 - head - neck - belt) / 2
	xyz_below <- data.frame(x = circle$x, y = y_below - 0.5, z = circle$y)
	vt_body <- data.frame(x = c(0, 0, 1, 1), y = c(0.1, 0, 0, 0.1))

	f_below <- list()
	for (i in seq(n_base)) {
		ir <- i %% n_base + 1
		il <- ir %% n_base + 1
		f_below[[i]] <- list(v = c(n_base + ir, n_base + il, il, ir), vt = n_base + 1:4)
	}
	# belt
	xyz_belt <- data.frame(x = circle$x, y = y_below - 0.5 + belt, z = circle$y)
	vt_belt <- data.frame(
		x = rep(seq(0, 1, length.out = nrow(xyz_belt) + 1L), 2),
		y = rep(c(0.75, 0.25), each = nrow(xyz_belt) + 1L)
	)
	f_belt <- list()
	for (i in seq(n_base)) {
		if (i <= n_base / 2) {
			ir <- i
			il <- ir + 1
			v <- c(n_base + ir, n_base + il, il, ir)
			vt <- c(
				0.5 * n_base + 2 - ir,
				0.5 * n_base + 2 - il,
				1.5 * n_base + 3 - il,
				1.5 * n_base + 3 - ir
			)
			f_belt[[i]] <- list(v = n_base + v, vt = n_base + 4 + vt)
		} else {
			ir <- i
			il <- ir + 1
			ilv <- ifelse(il > n_base, 1, il)
			v <- c(n_base + ir, n_base + ilv, ilv, ir)
			vt <- c(
				1.5 * n_base + 2 - ir,
				1.5 * n_base + 2 - il,
				2.5 * n_base + 3 - il,
				2.5 * n_base + 3 - ir
			)
			f_belt[[i]] <- list(v = n_base + v, vt = n_base + 4 + vt)
		}
	}

	# body above
	xyz_above <- data.frame(x = circle$x, y = 2 * y_below + belt - 0.5, z = circle$y)

	f_above <- list()
	for (i in seq(n_base)) {
		ir <- i %% n_base + 1
		il <- ir %% n_base + 1
		f_above[[i]] <- list(
			v = 2 * n_base + c(n_base + ir, n_base + il, il, ir),
			vt = n_base + 1:4
		)
	}
	# neck
	xyz_neck1 <- data.frame(x = 0.95 * circle$x, y = 0.5 - head - 0.95 * neck, z = 0.95 * circle$y)
	xyz_neck2 <- data.frame(x = 0.85 * circle$x, y = 0.5 - head - 0.7 * neck, z = 0.85 * circle$y)
	xyz_neck3 <- data.frame(x = 0.70 * circle$x, y = 0.5 - head - 0.4 * neck, z = 0.70 * circle$y)
	xyz_neck4 <- data.frame(x = 0.50 * circle$x, y = 0.5 - head, z = 0.50 * circle$y)
	xyz_neck5 <- data.frame(x = 0.50 * circle$x, y = 0.5 - 0.9 * head, z = 0.50 * circle$y)
	xyz_neck <- rbind(xyz_neck1, xyz_neck2, xyz_neck3, xyz_neck4, xyz_neck5)

	f_neck <- vector("list", 5 * n_base)
	for (i in seq(n_base)) {
		ir <- i %% n_base + 1
		il <- ir %% n_base + 1
		f_neck[[i]] <- list(
			v = 3 * n_base + c(n_base + ir, n_base + il, il, ir),
			vt = 3 * n_base + 4 + 2 + 1:4
		)
		f_neck[[n_base + i]] <- list(
			v = 4 * n_base + c(n_base + ir, n_base + il, il, ir),
			vt = 3 * n_base + 4 + 2 + 1:4
		)
		f_neck[[2 * n_base + i]] <- list(
			v = 5 * n_base + c(n_base + ir, n_base + il, il, ir),
			vt = 3 * n_base + 4 + 2 + 1:4
		)
		f_neck[[3 * n_base + i]] <- list(
			v = 6 * n_base + c(n_base + ir, n_base + il, il, ir),
			vt = 3 * n_base + 4 + 2 + 1:4
		)
		f_neck[[4 * n_base + i]] <- list(
			v = 7 * n_base + c(n_base + ir, n_base + il, il, ir),
			vt = 3 * n_base + 4 + 2 + 1:4
		)
	}

	# head
	e <- rgl::ellipse3d(diag(3), t = 0.5, subdivide = 4, smooth = FALSE)
	xyz <- as.data.frame(t(e$vb))
	names(xyz) <- c("x", "y", "z", "w")
	xyz_head <- as_coord3d(xyz)$scale(y = width / height)$translate(as_coord3d(
		x = 0,
		y = 0.5 - 0.5 * width / height
	))
	xyz_head <- as.data.frame(xyz_head)[, c("x", "y", "z")]

	vt_head <- data.frame(x = c(0, 0, 1, 1), y = c(1, 0.90, 0.90, 1))

	f_head <- lapply(seq.int(ncol(e$ib)), function(x) {
		list(v = 9 * n_base + e$ib[, x], vt = 3 * n_base + 4 + 2 + 1:4)
	})

	# write obj
	ext <- tools::file_ext(filename)
	mtl_filename <- gsub(paste0("\\.", ext, "$"), ".mtl", filename)
	png_filename <- gsub(paste0("\\.", ext, "$"), ".png", filename)

	xyz <- rbind(xyz_base, xyz_below, xyz_belt, xyz_above, xyz_neck, xyz_head)
	xyz <- as_coord3d(xyz)$scale(whd$width, whd$height, whd$depth)$transform(R)$translate(pc)

	xy_vt <- rbind(vt_base, vt_body, vt_belt, vt_head)

	faces <- c(f_base, f_below, f_belt, f_above, f_neck, f_head)

	write_obj(filename, v = xyz, vt = xy_vt, f = faces)
	write_peg_doll_texture(piece_side, suit, rank, cfg, filename = png_filename, res = res)

	invisible(list(obj = filename, mtl = mtl_filename, png = png_filename))
}

write_peg_doll_texture <- function(
	piece_side = "pawn_face",
	suit = 1,
	rank = 1,
	cfg = pp_cfg(),
	...,
	filename = tempfile(fileext = ".png"),
	res = 72
) {
	current_dev <- grDevices::dev.cur()
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	height <- cfg$get_height("belt_face", suit, rank)
	width <- cfg$get_width("belt_face", suit, rank)
	piece <- get_piece(piece_side)
	opt <- cfg$get_piece_opt(paste0(piece, "_face"), suit, rank)
	args <- list(
		filename = filename,
		height = 2 * height,
		width = width,
		units = "in",
		res = res,
		bg = "transparent"
	)
	if (capabilities("cairo")) {
		args$type <- "cairo"
	}
	do.call(grDevices::png, args)
	pushViewport(viewport(y = 0.875, height = 0.25))
	grid.rect(gp = gpar(col = NA_character_, fill = opt$edge_color))
	popViewport()
	pushViewport(viewport(y = 0.500, height = 0.50))
	grid.piece("belt_face", suit = suit, rank = rank, cfg = cfg, op_scale = 0)
	popViewport()
	pushViewport(viewport(y = 0.125, height = 0.25))
	grid.rect(gp = gpar(col = NA_character_, fill = opt$edge_color))
	popViewport()

	grDevices::dev.off()
	invisible(filename)
}

save_die_obj <- function(
	piece_side = "die_face",
	suit = 1,
	rank = 1,
	cfg = pp_cfg(),
	...,
	x = 0,
	y = 0,
	z = 0,
	angle = 0,
	axis_x = 0,
	axis_y = 0,
	width = NA,
	height = NA,
	depth = NA,
	filename = tempfile(fileext = ".obj"),
	res = 72
) {
	cfg <- as_pp_cfg(cfg)
	opt <- cfg$get_piece_opt(piece_side, suit, rank)
	xyz <- die_xyz(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth)

	xy_vt <- list(x = rep(c(0, 0.5, 1), 4), y = rep(c(1, 2 / 3, 1 / 3, 0), each = 3))

	# textured face elements
	f <- list()
	f[[1]] <- list(v = 1:4, vt = c(1, 4, 5, 2))
	f[[2]] <- list(v = c(8, 4, 3, 7), vt = c(2, 5, 6, 3))
	f[[3]] <- list(v = c(1, 4, 8, 5), vt = c(4, 7, 8, 5))
	f[[4]] <- list(v = c(6, 5, 8, 7), vt = c(5, 8, 9, 6))
	f[[5]] <- list(v = c(1, 5, 6, 2), vt = c(7, 10, 11, 8))
	f[[6]] <- list(v = c(3, 2, 6, 7), vt = c(8, 11, 12, 9))

	ext <- tools::file_ext(filename)
	mtl_filename <- gsub(paste0("\\.", ext, "$"), ".mtl", filename)
	png_filename <- gsub(paste0("\\.", ext, "$"), ".png", filename)

	write_obj(filename, v = xyz, vt = xy_vt, f = f)
	write_die_texture(piece_side, suit, rank, cfg, filename = png_filename, res = res)

	invisible(list(obj = filename, mtl = mtl_filename, png = png_filename))
}

write_pyramid_texture <- function(
	piece_side = "pyramid_face",
	suit = 1,
	rank = 1,
	cfg = pp_cfg(),
	...,
	filename = tempfile(fileext = ".png"),
	res = 72
) {
	current_dev <- grDevices::dev.cur()
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	height <- cfg$get_height("pyramid_face", suit, rank)
	width <- cfg$get_width("pyramid_face", suit, rank)

	args <- list(
		filename = filename,
		height = height,
		width = 4 * width,
		units = "in",
		res = res,
		bg = "transparent"
	)
	if (capabilities("cairo")) {
		args$type <- "cairo"
	}
	do.call(grDevices::png, args)

	pushViewport(viewport(x = 0.125, width = 0.25))
	draw_piece_and_bleed("pyramid_face", suit, rank, cfg)
	popViewport()

	pushViewport(viewport(x = 0.375, width = 0.25))
	draw_piece_and_bleed("pyramid_left", suit, rank, cfg)
	popViewport()

	pushViewport(viewport(x = 0.625, width = 0.25))
	draw_piece_and_bleed("pyramid_back", suit, rank, cfg)
	popViewport()

	pushViewport(viewport(x = 0.875, width = 0.25))
	draw_piece_and_bleed("pyramid_right", suit, rank, cfg)
	popViewport()

	grDevices::dev.off()
	invisible(filename)
}

write_die_texture <- function(
	piece_side = "die_face",
	suit = 1,
	rank = 1,
	cfg = pp_cfg(),
	...,
	filename = tempfile(fileext = ".png"),
	res = 72
) {
	current_dev <- grDevices::dev.cur()
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	width <- cfg$get_width("die_face", suit, rank)

	args <- list(
		filename = filename,
		height = 3 * width,
		width = 2 * width,
		units = "in",
		res = res,
		bg = "transparent"
	)
	if (capabilities("cairo")) {
		args$type <- "cairo"
	}
	do.call(grDevices::png, args)

	rs <- get_die_face_info(suit, cfg$die_arrangement)
	pushViewport(viewport(x = 0.25, width = 0.5, y = 5 / 6, height = 1 / 3))
	grid.piece(
		"die_face",
		suit = rs$suit[1],
		rank = rs$rank[1],
		cfg = cfg,
		angle = rs$angle[1],
		op_scale = 0
	)
	popViewport()
	pushViewport(viewport(x = 0.75, width = 0.5, y = 5 / 6, height = 1 / 3))
	grid.piece(
		"die_face",
		suit = rs$suit[2],
		rank = rs$rank[2],
		cfg = cfg,
		angle = rs$angle[2],
		op_scale = 0
	)
	popViewport()
	pushViewport(viewport(x = 0.25, width = 0.5, y = 3 / 6, height = 1 / 3))
	grid.piece(
		"die_face",
		suit = rs$suit[3],
		rank = rs$rank[3],
		cfg = cfg,
		angle = rs$angle[3],
		op_scale = 0
	)
	popViewport()
	pushViewport(viewport(x = 0.75, width = 0.5, y = 3 / 6, height = 1 / 3))
	grid.piece(
		"die_face",
		suit = rs$suit[4],
		rank = rs$rank[4],
		cfg = cfg,
		angle = rs$angle[4],
		op_scale = 0
	)
	popViewport()
	pushViewport(viewport(x = 0.25, width = 0.5, y = 1 / 6, height = 1 / 3))
	grid.piece(
		"die_face",
		suit = rs$suit[5],
		rank = rs$rank[5],
		cfg = cfg,
		angle = rs$angle[5],
		op_scale = 0
	)
	popViewport()
	pushViewport(viewport(x = 0.75, width = 0.5, y = 1 / 6, height = 1 / 3))
	grid.piece(
		"die_face",
		suit = rs$suit[6],
		rank = rs$rank[6],
		cfg = cfg,
		angle = rs$angle[6],
		op_scale = 0
	)
	popViewport()

	grDevices::dev.off()
	invisible(filename)
}

draw_piece_and_bleed <- function(piece_side, suit, rank, cfg) {
	g <- pieceGrob(piece_side, suit, rank, cfg, default.units = "npc", op_scale = 0)
	bleed_color <- cfg$get_piece_opt(piece_side, suit, rank)$bleed_color
	b <- pp_shape()$polyclip(g, "minus", gp = gpar(col = NA, fill = bleed_color))
	grid.draw(b)
	grid.draw(g)
}

# pyramid top up
save_pt_obj <- function(
	piece_side = "pyramid_top",
	suit = 1,
	rank = 1,
	cfg = pp_cfg(),
	...,
	x = 0,
	y = 0,
	z = 0,
	angle = 0,
	axis_x = 0,
	axis_y = 0,
	width = NA,
	height = NA,
	depth = NA,
	filename = tempfile(fileext = ".obj"),
	res = 72
) {
	cfg <- as_pp_cfg(cfg)

	# vertices
	xyz <- pt_xyz(x, y, z, angle, axis_x, axis_y, width, height, depth)

	# texture coordinates
	xy_vt <- list(x = seq(0, 1, 0.125), y = rep(c(0, 1), length.out = 9))

	# textured face elements
	f <- list()
	f[[1]] <- list(v = c(2, 3, 1), vt = c(3, 5, 4)) # left
	f[[2]] <- list(v = c(3, 4, 1), vt = c(5, 7, 6)) # back
	f[[3]] <- list(v = c(4, 5, 1), vt = c(7, 9, 8)) # right
	f[[4]] <- list(v = c(5, 2, 1), vt = c(1, 3, 2)) # front

	ext <- tools::file_ext(filename)
	mtl_filename <- gsub(paste0("\\.", ext, "$"), ".mtl", filename)
	png_filename <- gsub(paste0("\\.", ext, "$"), ".png", filename)

	write_obj(filename, v = xyz, vt = xy_vt, f = f)
	write_pyramid_texture(piece_side, suit, rank, cfg, filename = png_filename, res = res)

	invisible(list(obj = filename, mtl = mtl_filename, png = png_filename))
}

# pyramid side down
save_ps_obj <- function(
	piece_side = "pyramid_face",
	suit = 1,
	rank = 1,
	cfg = pp_cfg(),
	...,
	x = 0,
	y = 0,
	z = 0,
	angle = 0,
	axis_x = 0,
	axis_y = 0,
	width = NA,
	height = NA,
	depth = NA,
	filename = tempfile(fileext = ".obj"),
	res = 72
) {
	cfg <- as_pp_cfg(cfg)

	# geometric vertices
	xyz <- ps_xyz(x, y, z, angle, axis_x, axis_y, width, height, depth)

	# texture coordinates
	xy_vt <- list(x = seq(0, 1, 0.125), y = rep(c(0, 1), length.out = 9))

	# textured face elements
	f <- list()
	if (piece_side == "pyramid_face") {
		f[[1]] <- list(v = c(2, 5, 3), vt = c(3, 5, 4)) # left
		f[[2]] <- list(v = c(5, 4, 3), vt = c(5, 7, 6)) # back
		f[[3]] <- list(v = c(4, 1, 3), vt = c(7, 9, 8)) # right
		f[[4]] <- list(v = c(1, 2, 3), vt = c(1, 3, 2)) # front
	} else if (piece_side == "pyramid_left") {
		f[[1]] <- list(v = c(1, 2, 3), vt = c(3, 5, 4)) # left
		f[[2]] <- list(v = c(2, 5, 3), vt = c(5, 7, 6)) # back
		f[[3]] <- list(v = c(5, 4, 3), vt = c(7, 9, 8)) # right
		f[[4]] <- list(v = c(4, 1, 3), vt = c(1, 3, 2)) # front
	} else if (piece_side == "pyramid_back") {
		f[[1]] <- list(v = c(4, 1, 3), vt = c(3, 5, 4)) # left
		f[[2]] <- list(v = c(1, 2, 3), vt = c(5, 7, 6)) # back
		f[[3]] <- list(v = c(2, 5, 3), vt = c(7, 9, 8)) # right
		f[[4]] <- list(v = c(5, 4, 3), vt = c(1, 3, 2)) # front
	} else {
		# pyramid right
		f[[1]] <- list(v = c(5, 4, 3), vt = c(3, 5, 4)) # left
		f[[2]] <- list(v = c(4, 1, 3), vt = c(5, 7, 6)) # back
		f[[3]] <- list(v = c(1, 2, 3), vt = c(7, 9, 8)) # right
		f[[4]] <- list(v = c(2, 5, 3), vt = c(1, 3, 2)) # front
	}

	ext <- tools::file_ext(filename)
	mtl_filename <- gsub(paste0("\\.", ext, "$"), ".mtl", filename)
	png_filename <- gsub(paste0("\\.", ext, "$"), ".png", filename)

	write_obj(filename, v = xyz, vt = xy_vt, f = f)
	write_pyramid_texture(piece_side, suit, rank, cfg, filename = png_filename, res = res)

	invisible(list(obj = filename, mtl = mtl_filename, png = png_filename))
}

# v: v$x v$y v$z
# vt: vt$x vt$y
# f: list of lists of faces: f[[1]]$v, f[[1]]$vt
write_obj <- function(file, v, vt = NULL, f = NULL) {
	ext <- tools::file_ext(file)
	mtl_filename <- gsub(paste0("\\.", ext, "$"), ".mtl", file)
	png_filename <- gsub(paste0("\\.", ext, "$"), ".png", file)
	write_mtl(mtl_filename, basename(png_filename))

	cat(
		paste0("# Written by piecepackr v", packageDescription("piecepackr")$Version),
		"# https://github.com/piecepackr/piecepackr",
		paste("mtllib", basename(mtl_filename)),
		"# geometric vertices",
		paste("v", v$x, v$y, v$z),
		sep = "\n",
		file = file
	)
	if (!is.null(vt)) {
		cat(
			"# texture coordinates",
			paste("vt", vt$x, vt$y),
			sep = "\n",
			file = file,
			append = TRUE
		)
	}
	if (!is.null(f)) {
		cat(
			"# Textured polygonal face element",
			"usemtl material_0",
			sapply(f, function(x) paste("f", paste0(x$v, "/", x$vt, collapse = " "))),
			sep = "\n",
			file = file,
			append = TRUE
		)
	}
	invisible(NULL)
}

# two-sided objects
save_holed_board_obj_fn <- function(nrows = 4L, ncols = 4L, margin = 0) {
	# nolint
	force(nrows)
	force(ncols)
	force(margin)
	stopifnot(nrows == ncols)
	function(
		piece_side = "tile_face",
		suit = 1,
		rank = 1,
		cfg = pp_cfg(),
		...,
		x = 0,
		y = 0,
		z = 0,
		angle = 0,
		axis_x = 0,
		axis_y = 0,
		width = NA,
		height = NA,
		depth = NA,
		filename = tempfile(fileext = ".obj"),
		res = 72
	) {
		current_dev <- grDevices::dev.cur()
		if (current_dev > 1) {
			on.exit(grDevices::dev.set(current_dev), add = TRUE)
		}

		cfg <- as_pp_cfg(cfg)
		piece <- get_piece(piece_side)
		side <- get_side(piece_side)
		opt <- cfg$get_piece_opt(paste0(piece, "_face"), suit, rank)
		shape <- pp_shape(
			opt$shape,
			opt$shape_t,
			opt$shape_r,
			opt$back,
			width = opt$shape_w,
			height = opt$shape_h
		)

		grDevices::pdf(NULL, width = width, height = height)
		# snpc coords are npc if square
		xyi_npc <- xyi_holed_board(opt, nrows, ncols, margin)
		grDevices::dev.off()

		# Interior holes as integers
		xc <- rep(seq.int(ncols), times = nrows)
		yc <- rep(seq.int(nrows), each = ncols)

		# geometric vertices
		whd <- get_scaling_factors(side, width, height, depth)
		pc <- as_coord3d(x, y, z)
		R <- side_R(side) %*% AA_to_R(angle, axis_x, axis_y)

		xyz_scaled <- as_coord3d(xyi_npc)$translate(as_coord3d(-0.5, -0.5, 0.5))$scale(
			whd$width,
			whd$height,
			whd$depth
		)
		xyz_f <- xyz_scaled$transform(R)$translate(pc)

		xyz_scaled <- as_coord3d(xyi_npc)$translate(as_coord3d(-0.5, -0.5, -0.5))$scale(
			whd$width,
			whd$height,
			whd$depth
		)
		xyz_b <- xyz_scaled$transform(R)$translate(pc)

		xyz <- as_coord3d(x = c(xyz_f$x, xyz_b$x), y = c(xyz_f$y, xyz_b$y), z = c(xyz_f$z, xyz_b$z))

		# texture coordinates
		xy_npc_top <- as_coord2d(xyi_npc)
		xy_npc_bot <- as_coord2d(xyi_npc)
		xy_vt_f <- xy_npc_top$scale(0.4, 1)$translate(as_coord2d(0.025, 0))
		xy_vt_b <- xy_npc_bot$scale(0.4, 1)$translate(as_coord2d(0.575, 0))
		xy_vt_e <- list(x = c(0.52, 0.48, 0.48, 0.52), y = c(0, 0, 1, 1))
		vt <- list(x = c(xy_vt_f$x, xy_vt_b$x, xy_vt_e$x), y = c(xy_vt_f$y, xy_vt_b$y, xy_vt_e$y))

		# textured face elements
		nv <- length(xyz) / 2
		f <- list()

		# Outside shape
		xyio <- xyi_npc[which(xyi_npc$id == 0), ]
		nvo <- nrow(xyio)

		# Within a circle which vertices are left, top, right, or bottom?
		nvc <- nrow(xyi_npc[which(xyi_npc$id == 1), ])
		ivc_left <- seq.int(0.25 * nvc + 1, length.out = 0.5 * nvc + 1)
		ivc_top <- c(seq.int(0.50 * nvc + 1, length.out = 0.5 * nvc), 1)
		ivc_right <- c(
			seq.int(0.75 * nvc + 1, length.out = 0.25 * nvc),
			seq.int(1, length.out = 0.25 * nvc + 1)
		)
		ivc_bot <- seq.int(1, length.out = 0.5 * nvc + 1)

		# Left-most
		idc_left <- order(xc, yc)[seq.int(nrows)] # left-most circles
		id_left <- which(xyio$x < 0.5)
		# cycle elements so first one is "top" using `cycle_elements(x, n)`
		id_leftm <- which.max(xyio[id_left, ]$y)
		id_left <- cycle_elements(id_left, id_leftm - 1)
		for (idc in idc_left) {
			id_left <- c(id_left, nvo + nvc * (idc - 1) + ivc_left)
		}
		f[[1L]] <- list(v = id_left, vt = id_left) # top
		f[[2L]] <- list(v = rev_shift(nv + id_left), vt = rev_shift(nv + id_left)) # bottom

		# Right-most
		idc_right <- order(-xc, -yc)[seq.int(nrows)]
		id_right <- which(xyio$x > 0.5)
		id_rightm <- which.min(xyio[id_right, ]$y)
		id_right <- cycle_elements(id_right, id_rightm - 1)
		for (idc in idc_right) {
			id_right <- c(id_right, nvo + nvc * (idc - 1) + ivc_right)
		}
		f[[3L]] <- list(v = id_right, vt = id_right) # top
		f[[4L]] <- list(v = rev_shift(nv + id_right), vt = rev_shift(nv + id_right)) # bottom

		# Top-most
		idc_top <- order(-yc, xc)[seq.int(ncols)]
		id_top <- which(xyio$y > 0.5)
		id_topm <- which.max(xyio[id_top, ]$x)
		id_top <- cycle_elements(id_top, id_topm - 1)
		for (idc in idc_top) {
			id_top <- c(id_top, nvo + nvc * (idc - 1) + ivc_top)
		}
		f[[5L]] <- list(v = id_top, vt = id_top) # top
		f[[6L]] <- list(v = rev_shift(nv + id_top), vt = rev_shift(nv + id_top)) # bottom

		# Bottom-most
		idc_bot <- order(yc, -xc)[seq.int(ncols)]
		id_bot <- which(xyio$y < 0.5)
		id_botm <- which.min(xyio[id_bot, ]$x)
		id_bot <- cycle_elements(id_bot, id_botm - 1)
		for (idc in idc_bot) {
			id_bot <- c(id_bot, nvo + nvc * (idc - 1) + ivc_bot)
		}
		f[[7L]] <- list(v = id_bot, vt = id_bot) # top
		f[[8L]] <- list(v = rev_shift(nv + id_bot), vt = rev_shift(nv + id_bot)) # bottom

		# Interior
		ivc_lr <- seq.int(1, length.out = 0.25 * nvc + 1)
		ivc_ll <- seq.int(0.25 * nvc + 1, length.out = 0.25 * nvc + 1)
		ivc_ul <- seq.int(0.50 * nvc + 1, length.out = 0.25 * nvc + 1)
		ivc_ur <- c(seq.int(0.75 * nvc + 1, length.out = 0.25 * nvc), 1)
		if (nrows > 1L && ncols > 1L) {
			for (ir in seq.int(nrows - 1L)) {
				idc_ul <- which(yc == ir + 1L & xc == 1L)
				id_r <- nvo + nvc * (idc_ul - 1) + ivc_lr
				idc_ll <- which(yc == ir & xc == 1L)
				id_r <- c(id_r, nvo + nvc * (idc_ll - 1) + ivc_ur)
				if (ncols > 2L) {
					idc_b <- order(yc, xc)[ncols * (ir - 1) + seq.int(2, ncols - 1)]
					for (idc in idc_b) {
						id_r <- c(id_r, nvo + nvc * (idc - 1) + ivc_top)
					}
				}
				idc_lr <- which(yc == ir & xc == ncols)
				id_r <- c(id_r, nvo + nvc * (idc_lr - 1) + ivc_ul)
				idc_ur <- which(yc == ir + 1L & xc == ncols)
				id_r <- c(id_r, nvo + nvc * (idc_ur - 1) + ivc_ll)
				if (ncols > 2L) {
					idc_t <- order(yc, -xc)[ncols * ir + seq.int(2, ncols - 1)]
					for (idc in idc_t) {
						id_r <- c(id_r, nvo + nvc * (idc - 1) + ivc_bot)
					}
				}
				f[[length(f) + 1L]] <- list(v = id_r, vt = id_r) # top
				f[[length(f) + 1L]] <- list(v = rev_shift(nv + id_r), vt = rev_shift(nv + id_r)) # bottom
			}
		}

		# Board edges
		np <- 0
		for (id in unique(xyi_npc$id)) {
			no <- length(which(xyi_npc$id == id))
			for (i in seq.int(no)) {
				ir <- i %% no + 1
				il <- ir %% no + 1
				f[[length(f) + 1L]] <- list(
					v = c(ir + nv + np, il + nv + np, il + np, ir + np),
					vt = 2 * nv + 1:4
				)
			}
			np <- np + no
		}

		ext <- tools::file_ext(filename)
		mtl_filename <- gsub(paste0("\\.", ext, "$"), ".mtl", filename)
		png_filename <- gsub(paste0("\\.", ext, "$"), ".png", filename)

		write_obj(filename, v = xyz, vt = vt, f = f)

		grDevices::png(png_filename, bg = opt$background_color)
		grid.newpage()
		grDevices::dev.off()

		invisible(list(obj = filename, mtl = mtl_filename, png = png_filename))
	}
}

# https://mathworld.wolfram.com/RegularDodecahedron.html
# https://en.wikipedia.org/wiki/Regular_dodecahedron
# http://www.dicecollector.com/JM/D12.HTM
# If a = edge length
# inradius ~ 1.113516364 * a
# dihedral angle = arccos(-sqrt(5) / 5)

d12TopGrob <- function(
	piece_side,
	suit,
	rank,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = unit(0, "npc"),
	angle = 0,
	type = "normal",
	width = NA,
	height = NA,
	depth = NA,
	op_scale = 0,
	op_angle = 45
) {
	cfg <- as_pp_cfg(cfg)

	x <- convertX(x, "in", valueOnly = TRUE)
	y <- convertY(y, "in", valueOnly = TRUE)
	z <- convertX(z, "in", valueOnly = TRUE)
	width <- convertX(width, "in", valueOnly = TRUE)
	height <- convertY(height, "in", valueOnly = TRUE)
	depth <- convertX(depth, "in", valueOnly = TRUE)

	#### allow limited 3D rotation #281
	axis_x <- 0
	axis_y <- 0

	xyz <- d12_xyz(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth)
	l_xyz <- list()
	dfc <- data.frame(x = numeric(0), y = numeric(0), z = numeric(0))
	for (i in 1:12) {
		idx <- seq.int(5 * i - 4, length.out = 5)
		dfc <- rbind(dfc, as.data.frame(mean(xyz[idx])))
	}

	dfc$rank <- 1:12
	dfc <- dfc[
		painter_order(as_coord3d(dfc$x, dfc$y, dfc$z), scale = op_scale, alpha = degrees(op_angle)),
	]

	gl <- gList()
	for (i in 1:nrow(dfc)) {
		edge_rank <- dfc$rank[i]
		edge <- paste("d12_", edge_rank)
		opt <- cfg$get_piece_opt("die_face", suit, edge_rank)
		gp <- gpar(col = opt$border_color, lex = opt$border_lex, fill = opt$background_color)
		idx <- seq.int(5 * edge_rank - 4, length.out = 5)
		xyz_polygon <- xyz[idx]
		xy_polygon <- as_coord2d(xyz_polygon, alpha = degrees(op_angle), scale = op_scale)
		xy_vp <- xy_vp_convex(xyz_polygon, op_scale, op_angle)
		gl[[i]] <- at_ps_grob("die_face", suit, edge_rank, cfg, xy_vp, xy_polygon, name = edge)
	}

	# pre-compute grobCoords
	coords_xyl <- as.list(convex_hull2d(as_coord2d(
		xyz,
		alpha = degrees(op_angle),
		scale = op_scale
	)))

	gTree(
		scale = 1,
		coords_xyl = coords_xyl,
		children = gl,
		cl = c("projected_rpg_die", "coords_xyl")
	)
}

# if `rank` on top figure out rotation so one on top instead
d12_rot_from_top <- function(rank) {
	a <- 180 - arccos(-sqrt(5) / 5) # 180 - dihedral angle
	stopifnot(rank <= 12)
	switch(
		rank,
		diag(3), # 01
		d12_rot_from_top(3) %*% R_x(180), # 02
		R_z(72) %*% R_x(-a) %*% R_z(-36), # 03
		d12_rot_from_top(5) %*% R_x(180), # 04
		R_z(72) %*% R_x(-a) %*% R_z(-108), # 05
		d12_rot_from_top(7) %*% R_x(180), # 06
		R_z(72) %*% R_x(-a) %*% R_z(180), # 07
		d12_rot_from_top(9) %*% R_x(180), # 08
		R_z(72) %*% R_x(-a) %*% R_z(108), # 09
		d12_rot_from_top(11) %*% R_x(180), # 10
		R_z(72) %*% R_x(-a) %*% R_z(36), # 11
		R_x(180)
	) # 12
}

# if one on top figure out rotation so `rank` on top instead
d12_rot_to_top <- function(rank) {
	a <- 180 - arccos(-sqrt(5) / 5) # 180 - dihedral angle
	stopifnot(rank <= 12)
	switch(
		rank,
		diag(3), # 01
		R_x(180) %*% d12_rot_to_top(3), # 02
		R_z(36) %*% R_x(a) %*% R_z(-72), # 03
		R_x(180) %*% d12_rot_to_top(5), # 04
		R_z(108) %*% R_x(a) %*% R_z(-72), # 05
		R_x(180) %*% d12_rot_to_top(7), # 06
		R_z(180) %*% R_x(a) %*% R_z(-72), # 07
		R_x(180) %*% d12_rot_to_top(9), # 08
		R_z(-108) %*% R_x(a) %*% R_z(-72), # 09
		R_x(180) %*% d12_rot_to_top(11), # 10
		R_z(-36) %*% R_x(a) %*% R_z(-72), # 11
		R_x(180)
	) # 12
}

d12_xyz <- function(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth) {
	R_rank <- d12_rot_to_top(rank)
	xs <- numeric(0)
	ys <- numeric(0)
	zs <- numeric(0)
	xyz_t <- as_coord3d(
		regular_ngon_polygon2d(5, x = 0, y = 0, theta = degrees(90)),
		z = 0.5
	)$scale(
		width,
		height,
		depth
	)
	for (i in 1:12) {
		R_i <- d12_rot_from_top(i)
		xyz_i <- xyz_t$clone()$transform(R_i %*% R_rank)
		xs <- append(xs, xyz_i$x)
		ys <- append(ys, xyz_i$y)
		zs <- append(zs, xyz_i$z)
	}

	R <- AA_to_R(angle, axis_x, axis_y)
	as_coord3d(xs, ys, zs)$transform(R)$translate(x, y, z)
}

save_d12_obj <- function(
	piece_side,
	suit,
	rank,
	cfg,
	x,
	y,
	z,
	angle,
	axis_x,
	axis_y,
	width,
	height,
	depth,
	filename,
	scale,
	res
) {
	cfg <- as_pp_cfg(cfg)
	opt <- cfg$get_piece_opt(piece_side, suit, rank)

	xyz <- d12_xyz(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth)

	p <- regular_ngon_polygon2d(5, x = 0.5, y = 0.5, theta = degrees(90))
	xy_vt <- list(
		x = rep(p$x / 4, 12) + rep(rep(0:3 / 4, each = 5), 3),
		y = rep(p$y / 3, 12) + rep(rep(2:0 / 3, each = 5), each = 4)
	)

	# textured face elements
	f <- list()
	for (i in 1:12) {
		idx <- seq.int(5 * i - 4, length.out = 5)
		f[[i]] <- list(v = idx, vt = idx)
	}

	ext <- tools::file_ext(filename)
	mtl_filename <- gsub(paste0("\\.", ext, "$"), ".mtl", filename)
	png_filename <- gsub(paste0("\\.", ext, "$"), ".png", filename)

	write_obj(filename, v = xyz, vt = xy_vt, f = f)
	write_d12_texture("die_face", suit, rank, cfg, filename = png_filename, res = res)

	invisible(list(obj = filename, mtl = mtl_filename, png = png_filename))
}

write_d12_texture <- function(
	piece_side = "die_face",
	suit = 1,
	rank = 1,
	cfg = pp_cfg(),
	...,
	filename = tempfile(fileext = ".png"),
	res = 72
) {
	current_dev <- grDevices::dev.cur()
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	width <- cfg$get_width("die_face", suit, rank)

	args <- list(
		filename = filename,
		height = 3 * width,
		width = 4 * width,
		units = "in",
		res = res,
		bg = "transparent"
	)
	if (capabilities("cairo")) {
		args$type <- "cairo"
	}
	do.call(grDevices::png, args)

	xs <- rep(1:4 / 4 - 1 / 8, 3)
	ys <- rep(3:1 / 3 - 1 / 6, each = 4)
	for (i in 1:12) {
		pushViewport(viewport(x = xs[i], width = 1 / 4, y = ys[i], height = 1 / 3))
		draw_piece_and_bleed("die_face", suit, i, cfg)
		popViewport()
	}

	grDevices::dev.off()
	invisible(filename)
}

#' Configuration list R6 object
#'
#' `pp_cfg()` and `as_pp_cfg()` create piecepack configuration list R6 objects.
#' `is_pp_cfg()` returns `TRUE` if object is a piecepack configuration list R6 object.
#' `as.list()` will convert it into a list.
#'
#' `pp_cfg` R6 class objects serve the following purposes:
#'
#' * Customize the appearance of pieces drawn by `grid.piece()`.
#' * Speed up the drawing of graphics through use of caching.
#' * Allow the setting and querying of information about the board game components
#'   that maybe of use to developers:
#'
#'   + Number of suits
#'   + Number of ranks
#'   + Suit colors
#'   + Which types of components are included and/or properly supported
#'   + What would be a good color to use when adding annotations on top of these components.
#'   + Title, Description, Copyright, License, and Credit metadata
#'
#' @section `pp_cfg` R6 Class Method Arguments:\describe{
#'   \item{`piece_side`}{A string with piece and side separated by a underscore e.g. "coin_face".}
#'   \item{`suit`}{Number of suit (starting from 1).}
#'   \item{`rank`}{Number of rank (starting from 1).}
#'   \item{`type`}{Which type of grob to return, either `"normal"`, `"picture"`, `"raster"`, or `"transformation"`.}
#'   \item{`scale`}{"scale" factor}
#'   \item{`alpha`}{"alpha" value}
#' }
#'
#' @section `pp_cfg` R6 Class Methods:\describe{
#'   \item{`get_grob()`}{Returns a `grid` \dQuote{grob} for drawing the piece.}
#'   \item{`get_piece_opt()`}{Returns a list with info useful for drawing the piece.}
#'   \item{`get_suit_color()`}{Returns the suit colors.}
#'   \item{`get_width()`, `get_height()`, `get_depth()`}{
#'         Dimensions (of the bounding cube) of the piece in inches}
#' }
#'
#' @section `pp_cfg` R6 Class Fields and Active Bindings:\describe{
#'   \item{`annotation_color`}{Suggestion of a good color to annotate with}
#'   \item{`cache`}{Cache object which stores intermediate graphical calculations.
#'                  Default is a memory-cache that does not prune.
#'                  This can be replaced by another cache that
#'                  implements the cache API used by the `cachem` package}
#'   \item{`cache_grob`}{Whether we should cache (2D) grobs}
#'   \item{`cache_grob_with_bleed_fn`}{Whether we should cache the grob with bleed functions}
#'   \item{`cache_piece_opt`}{Whether we should cache piece opt information}
#'   \item{`cache_op_fn`}{Whether we should cache the oblique projection functions}
#'   \item{`cache_obj_fn`}{Whether we should cache any 3D rendering functions}
#'   \item{`copyright`}{Design copyright information}
#'   \item{`credit`}{Design credits}
#'   \item{`description`}{Design description}
#'   \item{`fontfamily`}{Main font family}
#'   \item{`has_bits`}{Whether we should assume this supports "bit" pieces}
#'   \item{`has_boards`}{Whether we should assume this supports "board" pieces}
#'   \item{`has_cards`}{Whether we should assume this supports "card" pieces}
#'   \item{`has_coins`}{Whether we should assume this supports "coin" pieces}
#'   \item{`has_dice`}{Whether we should assume this supports "die" pieces}
#'   \item{`has_matchsticks`}{Whether we should assume this supports "matchstick" pieces}
#'   \item{`has_pawns`}{Whether we should assume this supports "pawn" pieces}
#'   \item{`has_piecepack`}{Binding which simultaneously checks/sets
#'                         `has_coins`, `has_tiles`, `has_pawns`, `has_dice`}
#'   \item{`has_pyramids`}{Whether we should assume this supports "pyramid" pieces}
#'   \item{`has_saucers`}{Whether we should assume this supports "saucer" pieces}
#'   \item{`has_tiles`}{Whether we should assume this supports "tile" pieces}
#'   \item{`spdx_id`}{SPDX Identifier for graphical design license.
#'                    See \url{https://spdx.org/licenses/} for full list.}
#'   \item{`title`}{Design title}
#' }
#'
#' @section Defunct `pp_cfg` R6 Class attributes which have been removed:\describe{
#'    \item{`cache_shadow`}{Use `cache_op_fn` instead}
#'    \item{`i_unsuit`}{Instead add `1L` to `n_suits`}
#'    \item{`get_pictureGrob()`}{Use `get_grob(..., type = "picture")` instead}
#'    \item{`get_shadow_fn`}{`get_op_grob()` returns complete oblique projection grob}
#' }
#'
#' @seealso [game_systems()] for functions that return configuration list
#'  objects for several game systems.
#'  <https://trevorldavis.com/piecepackr/configuration-lists.html> for more details
#'  about `piecepackr` configuration lists.
#' @param cfg List of configuration options
#' @examples
#'  cfg <- pp_cfg(list(invert_colors=TRUE))
#'  as.list(cfg)
#'  is_pp_cfg(cfg)
#'  as_pp_cfg(list(suit_color="darkred,black,darkgreen,darkblue,grey"))
#'  cfg$get_suit_color(suit=3)
#'  cfg$annotation_color
#'  cfg$has_matchsticks
#'  cfg$has_matchsticks <- TRUE
#'  cfg$has_matchsticks
#'  cfg$get_width("tile_back")
#'  cfg$get_height("die_face")
#'  cfg$get_depth("coin_face")
#'  \donttest{# May take more than 5 seconds on CRAN servers
#'  # `pp_cfg()` objects use a cache to speed up repeated drawing
#'  pdf(tempfile(fileext = ".pdf"))
#'  cfg <- list()
#'  system.time(replicate(100, grid.piece("tile_back", 4, 4, cfg)))
#'  cfg <- pp_cfg(list())
#'  system.time(replicate(100, grid.piece("tile_back", 4, 4, cfg)))
#'  invisible(dev.off())
#'  }
#'
#' @export
pp_cfg <- function(cfg = list()) {
	if (is_pp_cfg(cfg)) {
		cfg
	} else {
		Config$new(cfg)
	}
}

#' @rdname pp_cfg
#' @export
is_pp_cfg <- function(cfg) inherits(cfg, "pp_cfg")

#' @rdname pp_cfg
#' @export
as_pp_cfg <- pp_cfg

#' @export
as.list.pp_cfg <- function(x, ...) {
	x$as_list()
}

default_lacks_suit <- c("tile_back", "saucer_back", "coin_face", "card_back")
default_lacks_rank <- c(
	"tile_back",
	"coin_back",
	"card_back",
	"pawn_face",
	"pawn_back",
	"belt_face",
	"saucer_face",
	"saucer_back",
	"suitdie_face"
)
impute_rank <- function(piece_side, rank, cfg) {
	ifelse(
		piece_side %in% cfg$lacks_rank,
		cfg$n_ranks + 1L,
		ifelse(is.na(rank), 1L, min(rank, cfg$n_ranks + 2L))
	)
}
impute_suit <- function(piece_side, suit, cfg) {
	ifelse(
		piece_side %in% cfg$lacks_suit,
		cfg$n_suits + 1L,
		ifelse(is.na(suit), 1L, min(suit, cfg$n_suits + 2L))
	)
}

Config <- R6Class(
	"pp_cfg",
	public = list(
		cache = NULL,
		has_bits = FALSE,
		has_boards = FALSE,
		has_cards = FALSE,
		has_coins = TRUE,
		has_dice = TRUE,
		has_matchsticks = FALSE,
		has_pawns = TRUE,
		has_pyramids = FALSE,
		has_saucers = FALSE,
		has_tiles = TRUE,
		initialize = function(cfg = list()) {
			warn_cfg(cfg)
			self$cache <- Cache$new()
			private$cfg <- cfg
			private$cfg$n_suits <- cfg$n_suits %||%
				(length(get_suit_symbols(cfg = cfg, expand = FALSE)) - 1L)
			private$cfg$n_ranks <- cfg$n_ranks %||%
				length(get_rank_symbols(cfg = cfg, expand = FALSE))
			private$cfg$lacks_rank <- cleave2(cfg$lacks_rank %||% default_lacks_rank)
			private$cfg$lacks_suit <- cleave2(cfg$lacks_suit %||% default_lacks_suit)
			# so different cfg objects can share the same cache (with high probability)
			private$prefix <- as.hexmode(sample.int(2147483647L, 1L, useHash = TRUE))
		},
		get_grob = function(
			piece_side,
			suit,
			rank,
			type = c("normal", "picture", "raster", "transformation"),
			...
		) {
			type <- match.arg(type)
			grob <- switch(
				type,
				normal = private$get_grob_normal(piece_side, suit, rank),
				picture = private$get_grob_picture(piece_side, suit, rank),
				raster = to_rasterGrob(self$get_raster(piece_side, suit, rank, ...)),
				transformation = private$get_grob_transformation(piece_side, suit, rank)
			)
			grob
		},
		get_grob_with_bleed = function(piece_side, suit, rank) {
			rank <- impute_rank(piece_side, rank, self)
			suit <- impute_suit(piece_side, suit, self)
			key <- private$opt_cache_key(piece_side, suit, rank, "grob_with_bleed_fn")
			grob_fn <- self$cache$get(key, key_missing())
			if (is.key_missing(grob_fn)) {
				default_grob_fn <- basicGrobWithBleed
				grob_fn <- get_style_element(
					"grob_with_bleed_fn",
					piece_side,
					private$cfg,
					default_grob_fn,
					suit,
					rank
				)
				if (is.character(grob_fn)) {
					grob_fn <- match.fun(grob_fn)
				}
				if (self$cache_grob_with_bleed_fn) self$cache$set(key, grob_fn)
			}
			args <- list(piece_side = piece_side, suit = suit, rank = rank, cfg = self)
			do.call(grob_fn, args)
		},
		get_piece_opt = function(piece_side, suit = NA, rank = NA) {
			if (is.na(rank)) {
				rank <- 1L
			}
			if (is.na(suit)) {
				suit <- self$n_suits + 1L
			}
			key <- private$opt_cache_key(piece_side, suit, rank, "piece_opt")
			opt <- self$cache$get(key, key_missing())
			if (is.key_missing(opt)) {
				opt <- get_piece_opt_helper(piece_side, suit, rank, private$cfg)
				opt$shape_w <- self$get_width(piece_side, suit, rank)
				opt$shape_h <- self$get_height(piece_side, suit, rank)
				if (self$cache_piece_opt) self$cache$set(key, opt)
			}
			opt
		},
		get_suit_color = function(suit = NULL) {
			if (is.null(suit)) {
				suit <- seq(self$n_suits)
			}
			colors <- character(length(suit))
			for (ii in seq(along.with = suit)) {
				colors[ii] <- get_suit_color_helper("pawn_face", suit[ii], rank = 1, private$cfg)
			}
			colors
		},
		get_width = function(piece_side, suit = 1, rank = 1) {
			key <- private$opt_cache_key(piece_side, suit, rank, "width")
			width <- self$cache$get(key, key_missing())
			if (!is.key_missing(width)) {
				return(width)
			}
			if (grepl("die_layout", piece_side)) {
				die_width <- self$get_width("die_face")
				return(4 * die_width)
			}
			if (grepl("pyramid_layout", piece_side)) {
				pyramid_height <- self$get_height("pyramid_face", rank = rank)
				return(pyramid_height)
			}
			if (piece_side == "preview_layout") {
				tile_width <- self$get_width("tile_face")
				return(3 * tile_width)
			}
			piece <- get_piece(piece_side)
			if (grepl("_left$|_right$", piece_side) && piece != "pyramid") {
				ps <- paste0(piece, "_face")
				return(self$get_height(ps, suit = suit, rank = rank))
			}
			default <- switch(
				piece,
				belt = 0.75 * pi, # so can wrap around 3/4" diameter pawns
				bit = 0.75,
				board = 8,
				card = 2.5,
				coin = 0.75,
				die = 0.5,
				matchstick = MATCHSTICK_WIDTHS[rank],
				pawn = 0.5,
				pyramid = PYRAMID_WIDTHS[rank],
				saucer = 0.75, # better than 7/8" for diagrams of hex games played with coin+pawn
				suitdie = 0.5,
				tile = 2,
				abort(paste("Don't know width of piece", piece))
			)
			width <- get_style_element("width", piece_side, private$cfg, default, suit, rank)
			self$cache$set(key, width)
			width
		},
		get_height = function(piece_side, suit = 1, rank = 1) {
			key <- private$opt_cache_key(piece_side, suit, rank, "height")
			height <- self$cache$get(key, key_missing())
			if (!is.key_missing(height)) {
				return(height)
			}
			if (grepl("die_layout", piece_side)) {
				die_height <- self$get_height("die_face")
				return(3 * die_height)
			}
			if (grepl("pyramid_layout", piece_side)) {
				pyramid_height <- self$get_height("pyramid_face", rank = rank)
				pyramid_width <- self$get_width("pyramid_face", rank = rank)
				pyramid_diagonal <- sqrt(pyramid_height^2 + (0.5 * pyramid_width)^2)
				return(2 * pyramid_diagonal)
			}
			if (grepl("pyramid_top", piece_side)) {
				pyramid_width <- self$get_width("pyramid_face", rank = rank)
				return(pyramid_width)
			}
			if (grepl("pawn_layout", piece_side)) {
				pawn_height <- self$get_height("pawn_face")
				return((2.5 / (7 / 8)) * pawn_height)
			}
			if (piece_side == "preview_layout") {
				tile_height <- self$get_height("tile_face")
				return(3 * tile_height)
			}
			piece <- get_piece(piece_side)
			if (grepl("_top$|_base$|_left$|_right$", piece_side) && piece != "pyramid") {
				ps <- paste0(piece, "_face")
				return(self$get_depth(ps, suit = suit, rank = rank))
			}
			width <- self$get_width(piece_side, suit, rank)
			if (piece == "matchstick") {
				W <- ifelse(rank == 1, 0.5 * width, width)
				S <- 0.5 * self$get_width("tile_face")
				ms_default <- (c(
					2 * W,
					S - W,
					sqrt(2) * S - W,
					2 * S - W,
					sqrt(5 * S^2) - W,
					2 * sqrt(2) * S - W
				)[rank])
			}
			default <- switch(
				piece,
				belt = 0.5,
				bit = width,
				board = width,
				card = 1.4 * width,
				coin = width,
				die = width,
				matchstick = ms_default,
				pawn = 0.875,
				pyramid = 1.538842 * width,
				saucer = width,
				suitdie = width,
				tile = width,
				abort(paste("Don't know height of piece", piece))
			) #nocov
			height <- get_style_element("height", piece_side, private$cfg, default, suit, rank)
			self$cache$set(key, height)
			height
		},
		get_depth = function(piece_side, suit = 1, rank = 1) {
			key <- private$opt_cache_key(piece_side, suit, rank, "depth")
			depth <- self$cache$get(key, key_missing())
			if (!is.key_missing(depth)) {
				return(depth)
			}
			width <- self$get_width(piece_side, suit, rank)
			if (grepl("pyramid", piece_side)) {
				height <- self$get_height("pyramid_face", rank = rank)
				if (piece_side == "pyramid_top") {
					return(sqrt(height^2 - (0.5 * width)^2))
				} else {
					theta <- 2 * asin(0.5 * width / height)
					return(sin(theta) * height)
				}
			}
			piece <- get_piece(piece_side)
			if (grepl("_top$|_base$", piece_side)) {
				ps <- paste0(piece, "_face")
				return(self$get_height(ps, suit = suit, rank = rank))
			} else if (grepl("_left$|_right$", piece_side)) {
				ps <- paste0(piece, "_face")
				return(self$get_width(ps, suit = suit, rank = rank))
			}
			default <- switch(
				piece,
				bit = 0.25,
				board = 0.25,
				card = 0.01181102, # 3 mm
				coin = 0.125,
				die = width,
				matchstick = width,
				pawn = 0.25,
				saucer = 0.125,
				suitdie = width,
				tile = 0.25,
				0
			)
			depth <- get_style_element("depth", piece_side, private$cfg, default, suit, rank)
			self$cache$set(key, depth)
			depth
		},
		as_list = function() private$cfg,
		print = function() {
			for (name in sort(names(private$cfg))) {
				if (is.function(private$cfg[[name]])) {
					cat(paste0("$", name, " : ", "a function", "\n"))
				} else {
					cat(paste0("$", name, " : ", paste(private$cfg[[name]], collapse = ",")), "\n")
				}
			}
		},
		get_op_grob = function(
			piece_side,
			suit,
			rank,
			x,
			y,
			z,
			angle,
			type,
			width,
			height,
			depth,
			op_scale,
			op_angle
		) {
			key <- private$opt_cache_key(piece_side, suit, rank, "op_grob")
			grob_fn <- self$cache$get(key, key_missing())
			if (is.key_missing(grob_fn)) {
				default_fn <- switch(
					piece_side,
					die_face = basicDieGrob,
					pyramid_top = basicPyramidTop,
					pyramid_face = basicPyramidSide,
					pyramid_back = basicPyramidSide,
					pyramid_left = basicPyramidSide,
					pyramid_right = basicPyramidSide,
					basicTokenGrob
				)
				grob_fn <- get_style_element(
					"op_grob_fn",
					piece_side,
					private$cfg,
					default_fn,
					suit,
					rank
				)
				if (self$cache_op_fn) self$cache$set(key, grob_fn)
			}
			args <- list(
				piece_side = piece_side,
				suit = suit,
				rank = rank,
				cfg = self,
				x = x,
				y = y,
				z = z,
				angle = angle,
				type = type,
				width = width,
				height = height,
				depth = depth,
				op_scale = op_scale,
				op_angle = op_angle
			)
			grob <- do.call(grob_fn, args)
		},
		get_raster = function(piece_side, suit, rank, res = 72) {
			grob <- private$get_grob_normal(piece_side, suit, rank)
			width <- self$get_width(piece_side, suit, rank)
			height <- self$get_height(piece_side, suit, rank)
			png_file <- tempfile(fileext = ".png")
			on.exit(unlink(png_file), add = TRUE)
			current_dev <- grDevices::dev.cur()
			if (current_dev > 1) {
				on.exit(grDevices::dev.set(current_dev), add = TRUE)
			}
			args <- list(
				filename = png_file,
				width = width,
				height = height,
				units = "in",
				res = res,
				bg = "transparent"
			)
			if (capabilities("cairo")) {
				args$type <- "cairo"
			}
			do.call(png, args)
			grid.draw(grob)
			invisible(grDevices::dev.off())
			as.raster(png::readPNG(png_file))
		},
		rayrender = function(
			piece_side,
			suit,
			rank,
			x,
			y,
			z,
			angle,
			axis_x,
			axis_y,
			width,
			height,
			depth,
			scale = 1,
			res = 72
		) {
			key <- private$opt_cache_key(piece_side, suit, rank, "rayrender_fn")
			rayrender_fn <- self$cache$get(key, key_missing())
			if (is.key_missing(rayrender_fn)) {
				rayrender_fn <- get_style_element(
					"rayrender_fn",
					piece_side,
					private$cfg,
					rr_piece_helper,
					suit,
					rank
				)
				if (self$cache_obj_fn) self$cache$set(key, rayrender_fn)
			}
			args <- list(
				piece_side = piece_side,
				suit = suit,
				rank = rank,
				cfg = self,
				x = x,
				y = y,
				z = z,
				angle = angle,
				axis_x = axis_x,
				axis_y = axis_y,
				width = width,
				height = height,
				depth = depth,
				scale = scale,
				res = res
			)
			do.call(rayrender_fn, args)
		},
		rayvertex = function(
			piece_side,
			suit,
			rank,
			x,
			y,
			z,
			angle,
			axis_x,
			axis_y,
			width,
			height,
			depth,
			scale = 1,
			res = 72
		) {
			key <- private$opt_cache_key(piece_side, suit, rank, "rayvertex_fn")
			rayvertex_fn <- self$cache$get(key, key_missing())
			if (is.key_missing(rayvertex_fn)) {
				rayvertex_fn <- get_style_element(
					"rayvertex_fn",
					piece_side,
					private$cfg,
					rv_piece_helper,
					suit,
					rank
				)
				if (self$cache_obj_fn) self$cache$set(key, rayvertex_fn)
			}
			args <- list(
				piece_side = piece_side,
				suit = suit,
				rank = rank,
				cfg = self,
				x = x,
				y = y,
				z = z,
				angle = angle,
				axis_x = axis_x,
				axis_y = axis_y,
				width = width,
				height = height,
				depth = depth,
				scale = scale,
				res = res
			)
			do.call(rayvertex_fn, args)
		},
		rgl = function(
			piece_side,
			suit,
			rank,
			x,
			y,
			z,
			angle,
			axis_x,
			axis_y,
			width,
			height,
			depth,
			scale = 1,
			res = 72,
			alpha = 1.0,
			lit = FALSE,
			shininess = 50.0,
			textype = "rgba"
		) {
			key <- private$opt_cache_key(piece_side, suit, rank, "rgl_fn")
			rgl_fn <- self$cache$get(key, key_missing())
			if (is.key_missing(rgl_fn)) {
				rgl_fn <- get_style_element(
					"rgl_fn",
					piece_side,
					private$cfg,
					rgl_piece_helper,
					suit,
					rank
				)
				if (self$cache_obj_fn) self$cache$set(key, rgl_fn)
			}
			args <- list(
				piece_side = piece_side,
				suit = suit,
				rank = rank,
				cfg = self,
				x = x,
				y = y,
				z = z,
				angle = angle,
				axis_x = axis_x,
				axis_y = axis_y,
				width = width,
				height = height,
				depth = depth,
				scale = scale,
				res = res,
				alpha = alpha,
				lit = lit,
				shininess = shininess,
				textype = textype
			)
			do.call(rgl_fn, args)
		},
		save_obj = function(
			piece_side,
			suit,
			rank,
			x,
			y,
			z,
			angle,
			axis_x,
			axis_y,
			width,
			height,
			depth,
			filename = tempfile(fileext = ".obj"),
			res = 72
		) {
			key <- private$opt_cache_key(piece_side, suit, rank, "obj_fn")
			obj_fn <- self$cache$get(key, key_missing())
			if (is.key_missing(obj_fn)) {
				if (grepl("tile|coin|pawn|matchstick|bit|board|card|saucer", piece_side)) {
					default_fn <- save_2s_obj
				} else if (grepl("die", piece_side)) {
					default_fn <- save_die_obj
				} else if (piece_side == "pyramid_top") {
					default_fn <- save_pt_obj
				} else if (grepl("pyramid", piece_side)) {
					default_fn <- save_ps_obj
				} else {
					default_fn <- NULL
				}
				obj_fn <- get_style_element(
					"obj_fn",
					piece_side,
					private$cfg,
					default_fn,
					suit,
					rank
				)
				if (is.null(obj_fn)) {
					abort(paste(
						"Don't know how to export ",
						piece_side,
						" to Wavefront OBJ format."
					))
				}
				if (self$cache_obj_fn) self$cache$set(key, obj_fn)
			}
			args <- list(
				piece_side = piece_side,
				suit = suit,
				rank = rank,
				cfg = self,
				x = x,
				y = y,
				z = z,
				angle = angle,
				axis_x = axis_x,
				axis_y = axis_y,
				width = width,
				height = height,
				depth = depth,
				filename = filename,
				res = res
			)
			do.call(obj_fn, args)
		}
	),
	active = list(
		annotation_color = function(value) {
			if (missing(value)) {
				private$cfg$annotation_color %||% "black"
			} else {
				stopifnot(is.character(value))
				private$cfg$annotation_color <- value
			}
		},
		cache_grob = function(value) {
			if (missing(value)) {
				private$cache_grob_bool
			} else {
				private$cache_type(value, "cache_grob_bool", "grob$")
			}
		},
		cache_grob_with_bleed_fn = function(value) {
			if (missing(value)) {
				private$cache_grob_with_bleed_fn_bool
			} else {
				private$cache_type(value, "cache_grob_with_bleed_fn_bool", "grob_with_bleed_fn$")
			}
		},
		cache_obj_fn = function(value) {
			if (missing(value)) {
				private$cache_obj_fn_bool
			} else {
				private$cache_type(value, "cache_obj_fn_bool", "obj_fn$|rgl_fn$|rayrender_fn$")
			}
		},
		cache_piece_opt = function(value) {
			if (missing(value)) {
				private$cache_piece_opt_bool
			} else {
				private$cache_type(value, "cache_piece_opt_bool", "piece_opt$")
			}
		},
		cache_op_fn = function(value) {
			if (missing(value)) {
				private$cache_op_fn_bool
			} else {
				private$cache_type(value, "cache_op_fn_bool", "shadow$|op_grob_fn$")
			}
		},
		coin_arrangement = function(value) {
			if (missing(value)) {
				as.numeric(private$cfg$coin_arrangement %||% 180)
			} else {
				private$cfg$coin_arrangement <- value
			}
		},
		copyright = function(value) {
			if (missing(value)) {
				private$cfg$copyright
			} else {
				stopifnot(is.character(value))
				private$cfg$copyright <- value
			}
		},
		credit = function(value) {
			if (missing(value)) {
				private$cfg$credit
			} else {
				stopifnot(is.character(value))
				private$cfg$credit <- value
			}
		},
		description = function(value) {
			if (missing(value)) {
				private$cfg$description
			} else {
				stopifnot(is.character(value))
				private$cfg$description <- value
			}
		},
		die_arrangement = function(value) {
			if (missing(value)) {
				private$cfg$die_arrangement %||% "counter_down"
			} else {
				stopifnot(is.character(value))
				private$cfg$die_arrangement <- value
			}
		},
		fontfamily = function(value) {
			if (missing(value)) {
				private$cfg$fontfamily %||% "sans"
			} else {
				warn("Must set 'fontfamily' at initialization")
			}
		},
		has_piecepack = function(value) {
			if (missing(value)) {
				self$has_coins && self$has_tiles && self$has_pawns && self$has_dice
			} else {
				stopifnot(is.logical(value))
				self$has_coins <- value
				self$has_tiles <- value
				self$has_pawns <- value
				self$has_dice <- value
			}
		},
		lacks_rank = function(value) {
			if (missing(value)) {
				private$cfg$lacks_rank
			} else {
				warn("Must set 'lacks_rank' at initialization")
			}
		},
		lacks_suit = function(value) {
			if (missing(value)) {
				private$cfg$lacks_suit
			} else {
				warn("Must set 'lacks_suit' at initialization")
			}
		},
		n_ranks = function(value) {
			if (missing(value)) {
				private$cfg$n_ranks
			} else {
				warn("Must set 'n_ranks' at initialization")
			}
		},
		n_suits = function(value) {
			if (missing(value)) {
				private$cfg$n_suits
			} else {
				warn("Must set 'n_suits' at initialization")
			}
		},
		spdx_id = function(value) {
			if (missing(value)) {
				private$cfg$spdx_id
			} else {
				stopifnot(is.character(value))
				check_spdx_id(value)
				private$cfg$spdx_id <- value
			}
		},
		title = function(value) {
			if (missing(value)) {
				private$cfg$title
			} else {
				stopifnot(is.character(value))
				private$cfg$title <- value
			}
		}
	),
	private = list(
		cfg = NULL,
		prefix = NULL,
		cache_grob_bool = TRUE,
		cache_grob_with_bleed_fn_bool = TRUE,
		cache_obj_fn_bool = TRUE,
		cache_piece_opt_bool = TRUE,
		cache_op_fn_bool = TRUE,
		n_suits_val = NULL,
		n_ranks_val = NULL,
		get_grob_normal = function(piece_side, suit, rank) {
			rank <- impute_rank(piece_side, rank, self)
			suit <- impute_suit(piece_side, suit, self)
			key <- private$opt_cache_key(piece_side, suit, rank, "grob")
			grob <- self$cache$get(key, key_missing())
			if (is.key_missing(grob)) {
				default_grob_fn <- switch(
					piece_side,
					board_face = checkeredBoardGrobFn(8, 8),
					board_back = linedBoardGrobFn(8, 8),
					card_face = cardGrobFn(-1),
					basicPieceGrob
				)
				default_fn <- get_style_element(
					"grob_fn",
					piece_side,
					private$cfg,
					default_grob_fn,
					suit,
					rank
				)
				grob_fn <- switch(
					piece_side,
					die_layoutLF = dieLayoutGrobLF,
					die_layoutRF = dieLayoutGrobRF,
					suitdie_layoutLF = suitdieLayouGrobtLF,
					suitdie_layoutRF = suitdieLayoutGrobRF,
					suitrankdie_layoutLF = suitrankdieLayoutGrobLF,
					suitrankdie_layoutRF = suitrankdieLayoutGrobRF,
					pawn_layout = pawnLayoutGrob,
					preview_layout = previewLayoutGrob,
					pyramid_layout = pyramidLayoutGrob,
					pyramid_top = pyramidTopGrob,
					default_fn
				)
				if (is.character(grob_fn)) {
					grob_fn <- match.fun(grob_fn)
				}

				args <- list(piece_side = piece_side, suit = suit, rank = rank, cfg = self)
				grob <- do.call(grob_fn, args)
				if (!inherits(grob, "pp_grobCoords")) {
					class(grob) <- c("pp_grobCoords", class(grob))
				}
				if (self$cache_grob) self$cache$set(key, grob)
			}
			grob
		},
		get_grob_picture = function(piece_side, suit, rank) {
			grob <- private$get_grob_normal(piece_side, suit, rank)
			width <- self$get_width(piece_side, suit, rank)
			height <- self$get_height(piece_side, suit, rank)
			as_picture(grob, width, height)
		},
		get_grob_transformation = function(piece_side, suit, rank) {
			grob <- private$get_grob_normal(piece_side, suit, rank)
			width <- self$get_width(piece_side, suit, rank)
			height <- self$get_height(piece_side, suit, rank)
			vp <- viewport(width = inch(width), height = inch(height))
			affiner::affineGrob(grob, vp_define = vp)
		},
		opt_cache_key = function(piece_side, suit, rank, type) {
			paste(private$prefix, piece_side, suit, rank, type, sep = "-")
		},
		cache_type = function(value, private_key, cache_key) {
			if (isTRUE(value)) {
				private[[private_key]] <- TRUE
			} else if (isFALSE(value)) {
				private[[private_key]] <- FALSE
				keys <- grep(cache_key, self$cache$keys(), value = TRUE)
				for (key in keys) {
					self$cache$remove(key)
				}
			} else {
				abort("value was not a boolean")
			}
		}
	)
)

cfg_default <- pp_cfg(list(title = "default cfg"))
test_that("`pp_cfg()` works as expected", {
	expect_true(inherits(cfg_default, "pp_cfg"))
	expect_equal(class(as.list(cfg_default)), "list")
	expect_output(print(cfg_default), "default cfg")
})

test_that("`update_names()` works as expected", {
	df <- tibble(x = 1:4, name = 1:4)
	expect_equal(update_name(df)$name, as.character(1:4))
	df <- tibble(x = 1:4, id = 1:4)
	expect_equal(update_name(df)$name, paste0("piece.", 1:4))
	df <- tibble(x = 1:4)
	expect_equal(update_name(df)$name, paste0("piece.", 1:4))
	df <- tibble(x = 1:4, name = 1)
	expect_warning(
		update_name(df)$name,
		"the name column in .l is not unique, generating new name column"
	)
	df <- tibble(x = 1:4, id = 1)
	expect_warning(
		update_name(df)$name,
		"the id column in .l is not unique, generating new name column"
	)
})

test_that("`save_print_and_play()` works as expected", {
	skip_on_cran()
	pdf_deck_dir <- tempfile()
	dir.create(pdf_deck_dir)
	on.exit(unlink(pdf_deck_dir, recursive = TRUE))
	pdf_deck_filename <- file.path(pdf_deck_dir, "piecepack_deck.pdf")
	on.exit(unlink(pdf_deck_filename), add = TRUE)
	pdf_deck_filename_a4 <- file.path(pdf_deck_dir, "piecepack_deck_a4.pdf")
	on.exit(unlink(pdf_deck_filename_a4), add = TRUE)
	pdf_deck_filename_a5 <- file.path(pdf_deck_dir, "piecepack_deck_a5.pdf")
	on.exit(unlink(pdf_deck_filename_a5), add = TRUE)
	pdf_deck_filename_5s <- file.path(pdf_deck_dir, "piecepack_deck_5s.pdf")
	on.exit(unlink(pdf_deck_filename_5s), add = TRUE)
	pdf_deck_filename_4x6 <- file.path(pdf_deck_dir, "piecepack_deck_4x6.pdf")
	on.exit(unlink(pdf_deck_filename_4x6), add = TRUE)
	pdf_deck_filename_bleed <- file.path(pdf_deck_dir, "piecepack_deck_bleed.pdf")
	cfg_5s <- list(
		suit_text = "♥,★,♣,♦,♠,꩜",
		suit_color = "darkred,gold,darkgreen,darkblue,black,grey"
	)

	skip_if_not(capabilities("cairo"))
	save_print_and_play(cfg_5s, pdf_deck_filename_5s, "A5", "all", "double-sided", quietly = TRUE)

	save_print_and_play(cfg_default, pdf_deck_filename, "letter", quietly = TRUE)

	cfg_large_dice <- pp_cfg(list(title = "default cfg", width.die = 0.7))
	save_print_and_play(
		cfg_large_dice,
		pdf_deck_filename_a4,
		"A4",
		arrangement = "double-sided",
		quietly = TRUE
	)
	save_print_and_play(cfg_default, pdf_deck_filename_a5, "A5", quietly = TRUE)

	suppressWarnings(
		save_print_and_play(
			cfg_default,
			pdf_deck_filename_4x6,
			"4x6",
			pieces = "piecepack",
			quietly = TRUE
		),
		classes = "deprecatedWarning"
	)

	save_print_and_play(cfg_default, pdf_deck_filename_bleed, bleed = TRUE, quietly = TRUE)

	expect_true(file.exists(pdf_deck_filename))

	skip_if_not_installed("qpdf") # avoid potential "n_pages_gs()" bug on Windows
	skip_if_not_installed("xmpdf")
	expect_equal(xmpdf::n_pages(pdf_deck_filename), 7, ignore_attr = "names")
	expect_equal(xmpdf::n_pages(pdf_deck_filename_a4), 10, ignore_attr = "names")
	expect_equal(xmpdf::n_pages(pdf_deck_filename_a5), 14, ignore_attr = "names")
	expect_equal(xmpdf::n_pages(pdf_deck_filename_4x6), 13, ignore_attr = "names")
	expect_equal(xmpdf::n_pages(pdf_deck_filename_bleed), 7, ignore_attr = "names")
})

test_that('`save_print_and_play(size = "4x6")` is deprecated', {
	skip_if_not(capabilities("cairo"))
	rlang::local_options(piecepackr.metadata.inform = FALSE)
	f <- tempfile(fileext = ".pdf")
	on.exit(unlink(f))
	expect_snapshot(
		save_print_and_play(pp_cfg(), f, size = "4x6", pieces = "piecepack", quietly = TRUE)
	)
})

test_that("`save_piece_images()` works as expected", {
	skip_if_not(capabilities("cairo"))
	directory <- tempfile()
	on.exit(unlink(directory))
	cfg <- pp_cfg(list(grob_fn = picturePieceGrobFn(directory)))
	g.p <- function(...) {
		grid.piece(..., op_scale = 0.5, default.units = "in")
	}

	current_dev <- grDevices::dev.cur()
	expect_error(save_piece_images(cfg_default, directory), "dir.exists\\(directory\\) is not TRUE")
	expect_error(grid.piece("tile_back", cfg = cfg), "Couldn't find suitable")
	grDevices::dev.off()
	if (current_dev > 1) {
		grDevices::dev.set(current_dev)
	}

	dir.create(directory)
	save_piece_images(cfg_default, directory, format = "svgz", angle = c(0, 90))
	expect_equal(length(list.files(directory)), 496)

	announce_snapshot_file(name = "diagram-op-ppgf.svg")
	skip_if_not(Sys.info()[["nodename"]] == "pesd-tld-laptop")
	library("vdiffr")
	suppressMessages(
		{
			expect_doppelganger("diagram-op-ppgf", function() {
				g.p("tile_back", x = 0.5 + c(3, 1, 3, 1), y = 0.5 + c(3, 3, 1, 1), cfg = cfg)
				g.p("tile_back", x = 0.5 + 3, y = 0.5 + 1, z = 1 / 4 + 1 / 8, cfg = cfg)
				g.p("tile_back", x = 0.5 + 3, y = 0.5 + 1, z = 2 / 4 + 1 / 8, cfg = cfg)
				g.p("die_face", x = 1, y = 1, z = 1 / 4 + 1 / 4, cfg = cfg)
				g.p("pawn_face", x = 1, y = 4, z = 1 / 4 + 1 / 8, angle = 90, cfg = cfg)
				g.p("coin_face", x = 3, y = 4, z = 1 / 4 + 1 / 16, angle = 180, cfg = cfg)
				g.p("coin_back", x = 3, y = 4, z = 1 / 4 + 1 / 8 + 1 / 16, angle = 180, cfg = cfg)
				g.p("coin_back", x = 3, y = 1, z = 3 / 4 + 1 / 16, angle = 90, cfg = cfg)
			})
		},
		classes = "piecepackr_affine_transformation"
	)
})

test_that("no regressions in figures", {
	skip_on_ci()
	skip_if_not(capabilities("cairo"))
	skip_if_not_installed("vdiffr")
	library("vdiffr")
	dc <- function(..., cfg = cfg_default) {
		grid.piece(..., cfg = cfg, default.units = "npc")
	}
	dce <- function(...) {
		tmpfile <- tempfile(fileext = ".svg")
		on.exit(unlink(tmpfile))
		svg(tmpfile)
		on.exit(dev.off(), add = TRUE)
		grid.piece(..., default.units = "npc")
	}
	# tile back
	expect_doppelganger("tile-back", function() dc("tile_back", bleed = TRUE))
	expect_doppelganger("tile-back-thickgrid", function() {
		dc("tile_back", cfg = list(gridline_lex.tile_back = 5))
	})
	expect_doppelganger("tile-back-thickborder", function() {
		dc("tile_back", cfg = list(border_lex.tile_back = 5))
	})
	expect_doppelganger("tile-back-svg", function() dc("tile_back", type = "picture"))
	expect_doppelganger("tile-back-hex", function() {
		dc("tile_back", cfg = list(shape.tile_back = "convex6"))
	})
	# tile face
	expect_doppelganger("tile-face-s1-r1", function() dc("tile_face", suit = 1, rank = 1))
	expect_doppelganger("tile-face-s3-r2", function() dc("tile_face", suit = 3, rank = 2))
	expect_doppelganger("tile-face-s3-r2-suit-as-ace", function() {
		dc("tile_face", cfg = list(use_suit_as_ace = TRUE), suit = 3, rank = 2)
	})
	expect_doppelganger("tile-face-s2-r3", function() dc("tile_face", suit = 2, rank = 3))
	df_tile <- tibble::tribble(~x, ~y, ~suit, ~rank,
                               1, 1, NA, NA,
                               3, 1, 3, NA,
                               1, 3, NA, 3,
                               3, 3, 3, 3)
	df_tile$piece_side <- "tile_back"
	df_tile$please_ignore <- "what"
	df_tile$id <- 1:4
	expect_doppelganger("tiles-faces", function() {
		pmap_piece(df_tile, width = 2, height = 2, default.units = "inches", cfg = "cfg_default")
	})
	# 106
	expect_doppelganger("different-sizes", function() {
		pushViewport(viewport(width = inch(6), height = inch(6)))
		dc("tile_face", suit = 2, rank = 3, x = inch(1), y = inch(5))
		dc("tile_face", suit = 2, rank = 3, x = inch(4), y = inch(4.5), cfg = list(width.tile = 3))
		dc(
			"tile_face",
			suit = 2,
			rank = 3,
			x = inch(4),
			y = inch(1.5),
			cfg = list(width.tile = 4, height.tile = 3)
		)
		dc("coin_back", suit = 3, x = inch(1), y = inch(1), cfg = list(mat_width.coin = 0.1))
		dc(
			"coin_back",
			suit = 3,
			x = inch(2),
			y = inch(1),
			cfg = list(width.coin = 1, mat_width.coin = 0.1)
		)
	})

	#### coins appear wrong #99
	# coin back
	expect_doppelganger("coin-back-s4", function() dc("coin_back", suit = 4))
	expect_doppelganger("coin-back-s4-inverted", function() {
		dc("coin_back", cfg = list(invert_colors = TRUE), suit = 4)
	})
	# coin face
	expect_doppelganger("coin-face-r1", function() dc("coin_face", rank = 1))
	expect_doppelganger("coin-face-r2", function() dc("coin_face", rank = 2))
	expect_doppelganger("coin-face-r4", function() dc("coin_face", rank = 4))
	expect_doppelganger("coin-face-r4italic", function() {
		dc("coin_face", cfg = list(ps_fontface.coin_face = "italic"), rank = 4)
	})
	expect_doppelganger("coin-face-r4pyramid", function() {
		dc("coin_face", rank = 4, cfg = list(shape.coin = "pyramid"))
	})
	expect_doppelganger("coin-face-r4convex8", function() {
		dc("coin_face", rank = 4, cfg = list(shape.coin = "convex8"))
	})
	expect_doppelganger("coin-face-r4concave8", function() {
		dc("coin_face", rank = 4, cfg = list(shape.coin = "concave8", shape_r.coin = 0.4))
	})
	expect_doppelganger("coin-face-r4oval", function() {
		dc(
			"coin_face",
			rank = 4,
			op_scale = 0.5,
			cfg = list(shape.coin = "oval", height.coin = 1.3)
		)
	})
	# die faces
	expect_doppelganger("die-face-s4-r1", function() dc("die_face", suit = 4, rank = 1))
	expect_doppelganger("die-face-s3-r2", function() dc("die_face", suit = 3, rank = 2))
	expect_doppelganger("die-face-s2-r5", function() dc("die_face", suit = 2, rank = 5))
	expect_doppelganger("die-face-s2-r5-kite", function() {
		dc(
			"die_face",
			suit = 2,
			rank = 5,
			cfg = list(shape.die_face = "kite", dm_t = 90, suit_cex = "1,1,1,1")
		)
	})
	expect_doppelganger("die-face-s2-r5-convex5mat", function() {
		dc(
			"die_face",
			suit = 2,
			rank = 5,
			cfg = list(shape.die_face = "convex5", mat_width.die_face = 0.1, mat_color = "pink")
		)
	})
	expect_doppelganger("suitdie-face-s1", function() dc("suitdie_face", suit = 1))
	expect_doppelganger("suitdie-face-s5", function() dc("suitdie_face", suit = 5))
	expect_doppelganger("suitdie-face-s6", function() dc("suitdie_face", suit = 6))
	# pawns
	expect_doppelganger("pawn-face-s3", function() dc("pawn_face", suit = 3))
	expect_doppelganger("pawn-back-s2", function() dc("pawn_back", suit = 2))
	expect_doppelganger("belt-face-s1", function() dc("belt_face", suit = 1))
	expect_doppelganger("saucer-face-s4", function() dc("saucer_face", suit = 4))
	expect_doppelganger("saucer-back", function() dc("saucer_back"))
	cfg <- list(shape.pawn = "convex6", height.pawn = 1, width.pawn = 0.5)
	expect_doppelganger("pawn-face-irregular-convex", function() {
		dc("pawn_face", cfg = cfg, op_scale = 0.5)
	})

	# matchsticks
	expect_doppelganger("matchstick-face-s1-r1", function() {
		dc("matchstick_face", suit = 1, rank = 1)
	})
	expect_doppelganger("matchstick-face-s2-r2", function() {
		dc("matchstick_face", suit = 2, rank = 2)
	})
	expect_doppelganger("matchstick-face-s3-r3", function() {
		dc("matchstick_face", suit = 3, rank = 3)
	})
	expect_doppelganger("matchstick-face-s4-r4", function() {
		dc("matchstick_face", suit = 4, rank = 4)
	})
	expect_doppelganger("matchstick-face-s1-r5", function() {
		dc("matchstick_face", suit = 1, rank = 5)
	})
	expect_doppelganger("matchstick-face-s2-r6", function() {
		dc("matchstick_face", suit = 2, rank = 6)
	})
	cfg <- list(invert_colors.suited = TRUE, grob_fn = "basicPieceGrob")
	expect_doppelganger("matchstick-back-s3-r6", function() {
		dc("matchstick_back", cfg = cfg, suit = 3, rank = 6)
	})

	expect_doppelganger("preview", function() suppressWarnings(dc("preview_layout")))
	expect_doppelganger("preview-5s", function() {
		suppressWarnings(dc(
			"preview_layout",
			cfg = list(suit_text = "A,B,C,D,E,", die_arrangement = "counter_up")
		))
	})
	expect_doppelganger("preview-6s", function() {
		suppressWarnings(dc("preview_layout", cfg = list(suit_text = "I,II,III,IV,V,VI,")))
	})

	# dice
	cfg <- list(suit_text = "-,0,+,")
	expect_doppelganger("suitdie-layoutlf", function() dc("suitdie_layoutLF", cfg = cfg))
	expect_doppelganger("suitdie-layoutrf", function() dc("suitdie_layoutRF"))
	expect_doppelganger("suitrankdie-layoutlf", function() dc("suitrankdie_layoutLF"))

	cfg <- list(die_arrangement = "opposites_sum_to_5")
	expect_doppelganger("die-layoutrf-opposites-sum-to-5", function() {
		dc("die_layoutRF", suit = 3, cfg = cfg)
	})

	# 102
	expect_doppelganger("rankdie-layoutrf-suitasace", function() {
		dc("die_layoutRF", suit = 6, cfg = list(use_suit_as_ace = TRUE))
	})

	# edges
	df <- tibble(
		piece_side = c("pawn_top", "coin_left", "tile_base"),
		x = 1:3,
		y = 1:3,
		suit = 1:3,
		rank = 1:3
	)
	cfg <- list(
		border_color = "black",
		edge_color.pawn = "red",
		edge_color.coin = "grey",
		edge_color.tile = "grey"
	)
	expect_doppelganger("piece-edges", function() {
		pushViewport(viewport(width = inch(4), height = inch(4)))
		pmap_piece(df, cfg = cfg, default.units = "in")
	})

	cfg <- list(suit_text = "A,B,C,D,E,F", suit_color = "red,black,green,blue,orange,grey")
	expect_doppelganger("suitdie-layoutrf-5suits", function() dc("suitdie_layoutRF", cfg = cfg))
	expect_doppelganger("suitrankdie-layoutrf-5suits", function() {
		dc("suitrankdie_layoutRF", cfg = cfg)
	})
	cfg <- list(suit_text = "A,B,C,D,E,F,G", suit_color = "red,black,green,blue,orange,purple,grey")
	expect_doppelganger("suitdie-layoutrf-6suits", function() dc("suitdie_layoutRF", cfg = cfg))
	expect_doppelganger("suitrankdie-layoutrf-6suits", function() {
		dc("suitrankdie_layoutRF", cfg = cfg)
	})

	# draw_pieces
	df <- tibble::tribble(~piece_side, ~x, ~y, ~suit, ~rank,
                          "tile_face", 0.25, 0.25, 1, 1,
                          "tile_face", 0.75, 0.25, 2, 2,
                          "tile_face", 0.25, 0.75, 3, 5,
                          "tile_face", 0.75, 0.75, 4, 6)
	expect_doppelganger("pmap-piece", function() {
		pushViewport(viewport(width = inch(4), height = inch(4)))
		pmap_piece(df, cfg = "default", envir = list(default = cfg_default))
	})
	expect_doppelganger("pmap-piece-default", function() {
		pushViewport(viewport(width = inch(4), height = inch(4)))
		pmap_piece(df, envir = list())
	})

	# errors
	expect_error(
		dce("coin_face", rank = 3, cfg = list(gridline_color = "grey")),
		"Don't know how to add grid lines to shape circle"
	)
	expect_error(
		dce("coin_face", rank = 3, cfg = list(shape = "megahex")),
		"Don't recognize shape label megahex"
	)
	expect_error(
		dce(
			"coin_face",
			rank = 3,
			cfg = list(mat_width = 0.2, mat_color = "green", shape = "kite")
		),
		"Don't know how to add mat to shape kite"
	)
	expect_error(cfg_default$get_width("boo_back"), "Don't know width of piece boo")

	# piecepack pyramids
	expect_doppelganger("pyramid-face-s1-r6", function() dc("pyramid_face", suit = 1, rank = 6))
	expect_doppelganger("pyramid-left-s2-r5", function() dc("pyramid_left", suit = 2, rank = 5))
	expect_doppelganger("pyramid-right-s3-r2", function() dc("pyramid_right", suit = 3, rank = 2))
	expect_doppelganger("pyramid-back-s4-r1", function() dc("pyramid_back", suit = 4, rank = 1))
	expect_doppelganger("pyramid-face-s2-r1", function() dc("pyramid_face", suit = 2, rank = 1))
	cfg <- list(invert_colors.suited = TRUE, grob_fn = "basicPieceGrob")
	expect_doppelganger("pyramid-layout-s3-r4", function() {
		dc("pyramid_layout", cfg = cfg, suit = 3, rank = 4)
	})

	announce_snapshot_file(name = "pyramid-face-op.svg")
	announce_snapshot_file(name = "pyramid-top-op.svg")
	announce_snapshot_file(name = "pyramid-top-s4-r3.svg")
	skip_if_not(Sys.info()[["nodename"]] == "pesd-tld-laptop")
	suppressMessages(
		{
			expect_doppelganger("pyramid-top-s4-r3", function() {
				dc("pyramid_top", suit = 4, rank = 3)
			})
			expect_doppelganger("pyramid-top-op", function() {
				dc(
					"pyramid_top",
					rank = 6,
					op_angle = 90,
					op_scale = 0.5,
					cfg = list(invert_colors = TRUE)
				)
			})
			expect_doppelganger("pyramid-face-op", function() {
				dc(
					"pyramid_face",
					rank = 6,
					op_angle = 90,
					op_scale = 0.5,
					cfg = list(invert_colors = TRUE)
				)
			})
		},
		classes = "piecepackr_affine_transformation"
	)
})

test_that("deprecated 'preview_layout' component warning", {
	expect_snapshot(invisible(pieceGrob(
		"preview_layout",
		cfg = cfg_default,
		default.units = "npc"
	)))
})

test_that("oblique projection works", {
	skip_on_ci()
	skip_on_cran()
	skip_if_not(capabilities("cairo"))
	skip_if_not_installed("vdiffr")
	library("vdiffr")
	dc <- function(..., cfg = cfg_default) {
		grid.piece(..., cfg = cfg, op_scale = 0.5, default.units = "npc")
	}
	expect_doppelganger("tile-face-op", function() dc("tile_face"))
	expect_doppelganger("tile-face-op-roundrect", function() {
		dc("tile_face", cfg = list(shape.tile = "roundrect"))
	})
	expect_doppelganger("coin-face-op", function() dc("coin_face"))
	expect_doppelganger("pawn-face-op", function() dc("pawn_face"))
	expect_doppelganger("matchstick-face-op", function() dc("matchstick_face"))
	suppressMessages(
		{
			expect_doppelganger("die-face-op", function() dc("die_face"))
			expect_doppelganger("die-face-op-roundrect", function() {
				dc("die_face", cfg = list(shape.die_face = "roundrect", shape_r.die_face = 0.3))
			})
			g.p <- function(...) {
				grid.piece(..., op_scale = 0.5, default.units = "in")
			}
			cfg <- pp_cfg(list(
				depth.pawn = 1,
				width.pawn = 0.75,
				height.pawn = 0.75,
				dm_text.pawn = "",
				shape.pawn = "convex6",
				invert_colors.pawn = TRUE
			))
			expect_doppelganger("diagram-op", function() {
				g.p("tile_back", x = 0.5 + c(3, 1, 3, 1), y = 0.5 + c(3, 3, 1, 1), cfg = cfg)
				g.p("tile_back", x = 0.5 + 3, y = 0.5 + 1, z = 1 / 4 + 1 / 8, cfg = cfg)
				g.p("tile_back", x = 0.5 + 3, y = 0.5 + 1, z = 2 / 4 + 1 / 8, cfg = cfg)
				g.p("die_face", x = 1, y = 1, z = 1 / 4 + 1 / 4, cfg = cfg)
				g.p("pawn_face", x = 1, y = 4, z = 1 / 4 + 1 / 2, angle = 90, cfg = cfg)
				g.p("coin_back", x = 3, y = 4, z = 1 / 4 + 1 / 16, angle = 180, cfg = cfg)
				g.p("coin_back", x = 3, y = 4, z = 1 / 4 + 1 / 8 + 1 / 16, angle = 180, cfg = cfg)
				g.p("coin_back", x = 3, y = 1, z = 3 / 4 + 1 / 16, angle = 90, cfg = cfg)
			})
		},
		classes = "piecepackr_affine_transformation"
	)

	cfg <- pp_cfg(list(
		depth.pawn = 10 / 25.4,
		width.pawn = 16 / 25.4,
		height.pawn = 16 / 25.4,
		dm_cex.pawn = 0.5,
		shape.pawn = "meeple",
		invert_colors.pawn = TRUE
	))
	expect_doppelganger("meeple", function() {
		g.p("pawn_face", x = 2, y = 3, cfg = cfg, op_angle = 90)
		g.p("pawn_face", x = 3, y = 3, cfg = cfg, op_angle = 45)
		g.p("pawn_face", x = 3, y = 2, cfg = cfg, op_angle = 0)
		g.p("pawn_face", x = 3, y = 1, cfg = cfg, op_angle = -45)
		g.p("pawn_face", x = 2, y = 1, cfg = cfg, op_angle = -90)
		g.p("pawn_face", x = 1, y = 1, cfg = cfg, op_angle = -135)
		g.p("pawn_face", x = 1, y = 2, cfg = cfg, op_angle = 180)
		g.p("pawn_face", x = 1, y = 3, cfg = cfg, op_angle = 135)
		g.p("coin_face", x = 2, y = 2, cfg = cfg, op_angle = 0, angle = 190)
	})
	expect_doppelganger("pawn-top-meeple-op", function() {
		suppressMessages(g.p("pawn_top", cfg = cfg), classes = "piecepackr_affine_transformation")
	})
})

test_that("alpha and scale works", {
	skip_on_ci()
	skip_if_not(capabilities("cairo"))
	skip_if_not_installed("vdiffr")
	library("vdiffr")
	expect_doppelganger("alpha", function() {
		cfg <- pp_cfg(list(shape.coin = "convex6"))
		df <- tibble(piece_side = "coin_back", x = 1:6, y = 1, alpha = seq(0, 1, length.out = 6))
		pmap_piece(df, default.units = "in", cfg = cfg)
	})
	expect_doppelganger("scale", function() {
		cfg <- pp_cfg(list(shape.coin = "convex6"))
		df <- tibble(piece_side = "coin_back", x = 1:6, y = 1, scale = seq(0, 1, length.out = 6))
		pmap_piece(df, default.units = "in", cfg = cfg)
	})
	announce_snapshot_file(name = "alpha-and-scale-op.svg")
	skip_if_not(Sys.info()[["nodename"]] == "pesd-tld-laptop")
	expect_doppelganger("alpha-and-scale-op", function() {
		cfg <- pp_cfg(list(shape.coin = "convex6"))
		df <- tibble(
			piece_side = "coin_back",
			x = 1:6,
			y = 1,
			alpha = seq(0, 1, length.out = 6),
			scale = seq(0, 1, length.out = 6)
		)
		g <- pmap_piece(df, default.units = "in", cfg = cfg, op_scale = 0.5)
		grid.draw(pp_shape()$polyclip(g, "minus", gp = gpar(fill = "yellow")))
	})
})

#' Defunct functions
#'
#' These functions are Defunct and have been removed from piecepackr.
#'
#' \enumerate{
#' \item{For `get_shape_grob_fn` use `pp_shape()$shape` instead.}
#' \item{For `gridlinesGrob()` use `pp_shape()$gridlines()` instead.}
#' \item{For `matGrob()` use `pp_shape()$mat()` instead.}
#' \item{For `checkersGrob()`() use `pp_shape()$checkers()` instead.}
#' \item{For `hexlinesGrob()` use `pp_shape()$hexlines()` instead.}
#' \item{For `halmaGrob()` use `pp_shape("halma")$shape()` instead.}
#' \item{For `kiteGrob()` use `pp_shape("kite")$shape()` instead.}
#' \item{For `pyramidGrob()` use `pp_shape("pyramid")$shape()` instead.}
#' \item{For `convexGrobFn(n, t)` use `pp_shape(paste0("convex", n), t)$shape` instead.}
#' \item{For `concaveGrobFn(n, t, r)` use `pp_shape(paste0("concave", n), t, r)$shape` instead.}
#' }
#' @param ... Ignored
#' @name piecepackr-defunct
NULL

#' @rdname piecepackr-defunct
halmaGrob <- function(...) {
	.Defunct('pp_shape("halma")$shape()')
}

#' @rdname piecepackr-defunct
kiteGrob <- function(...) {
	.Defunct('pp_shape("kite")$shape()')
}

#' @rdname piecepackr-defunct
pyramidGrob <- function(...) {
	.Defunct('pp_shape("pyramid")$shape()')
}

#' @rdname piecepackr-defunct
convexGrobFn <- function(...) {
	.Defunct('pp_shape("convexN", t)$shape')
}

#' @rdname piecepackr-defunct
concaveGrobFn <- function(...) {
	.Defunct('pp_shape("concaveN", t, r)$shape')
}

#' @rdname piecepackr-defunct
gridlinesGrob <- function(...) {
	.Defunct("pp_shape()$gridlines()")
}

#' @rdname piecepackr-defunct
matGrob <- function(...) {
	.Defunct("pp_shape()$mat()")
}

#' @rdname piecepackr-defunct
checkersGrob <- function(...) {
	.Defunct("pp_shape()$checkers()")
}

#' @rdname piecepackr-defunct
hexlinesGrob <- function(...) {
	.Defunct("pp_shape()$hexlines()")
}

#' @rdname piecepackr-defunct
get_shape_grob_fn <- function(...) {
	.Defunct("pp_shape()$shape")
}

test_that("Defunct functions throw errors", {
	skip_on_ci()

	expect_error(checkersGrob(shape = "convex8", col = "purple"), class = "defunctError")
	expect_error(concaveGrobFn(3, 90), class = "defunctError")
	expect_error(convexGrobFn(3, 90), class = "defunctError")
	expect_error(get_shape_grob_fn("rect"), class = "defunctError")
	expect_error(gridlinesGrob(col = "black", shape = "rect"), class = "defunctError")
	expect_error(halmaGrob(), class = "defunctError")
	expect_error(hexlinesGrob(shape = "rect", col = "transparent"), class = "defunctError")
	expect_error(kiteGrob(), class = "defunctError")
	expect_error(matGrob(col = "black", shape = "convex3"), class = "defunctError")
	expect_error(pyramidGrob(), class = "defunctError")
})

#' Install additional packages from the piecepackr universe
#'
#' Uses [utils::install.packages()] to install additional R packages
#' from the piecepackr universe <https://piecepackr.r-universe.dev/builds>.
#' `pkgs_ppverse()` returns a character vector of R packages
#' in the piecepackr universe.
#' @param pkgs Packages to install. Passed to [utils::install.packages()].
#' @param ... Passed to [utils::install.packages()].
#' @param dependencies Logical indicating whether to install dependencies of `pkgs`. Passed to [utils::install.packages()].
#' @examples
#' pkgs_ppverse()
#' \dontrun{# Installs non-CRAN packages from the piecepackr universe
#'   install_ppverse()
#' }
#' @export
install_ppverse <- function(
	pkgs = pkgs_ppverse(free_libre_only),
	...,
	dependencies = TRUE,
	free_libre_only = TRUE
) {
	repos <- c(piecepackr = "https://piecepackr.r-universe.dev/", getOption("repos"))
	utils::install.packages(pkgs, ..., repos = repos, dependencies = dependencies)
}

#' @rdname install_ppverse
#' @param free_libre_only Flag to only include packages that are Free/Libre Open Source.
#' @export
pkgs_ppverse <- function(free_libre_only = TRUE) {
	pkgs <- c("piecepackr", "pprules", "ppcli", "ppgamer", "ppn", "ppdf")
	if (isFALSE(free_libre_only)) {
		pkgs <- append(pkgs, c("piecenikr"))
	}
	pkgs <- sort(pkgs)
	pkgs
}

# Grob for Fudge dice

fudgeGrob <- function(piece_side, suit, rank, cfg = pp_cfg()) {
	cfg <- as_pp_cfg(cfg)
	opt <- cfg$get_piece_opt(piece_side, suit, rank)
	gTree(
		opt = opt,
		border = TRUE,
		flip = FALSE,
		scale = 1,
		name = NULL,
		gp = gpar(),
		vp = NULL,
		cl = c("fudge_die", "basic_piece_side")
	)
}

#' @export
makeContent.fudge_die <- function(x) {
	opt <- x$opt
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	# Background
	background_grob <- shape$shape(
		gp = gpar(col = NA, fill = opt$background_color),
		name = "background"
	)

	# Primary symbol
	gp_ps <- gpar(col = NA, fill = opt$ps_color)
	if (opt$ps_text == "+") {
		ps_grob <- rectGrob(
			width = c(0.12, 0.7),
			height = c(0.7, 0.12),
			default.units = "npc",
			gp = gp_ps
		)
	} else if (opt$ps_text == "-") {
		ps_grob <- rectGrob(width = 0.7, height = 0.12, default.units = "npc", gp = gp_ps)
	} else {
		ps_grob <- nullGrob()
	}
	ps_grob <- editGrob(ps_grob, name = "primary_symbol")

	# Border
	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- shape$shape(gp = gp_border, name = "border")
	} else {
		border_grob <- nullGrob(name = "border")
	}
	if (x$flip) {
		gl <- gList(ps_grob, background_grob, border_grob)
	} else {
		gl <- gList(background_grob, ps_grob, border_grob)
	}

	setChildren(x, gl)
}

get_piece <- function(piece_side) {
	strsplit(piece_side, "_")[[1]][1]
}
get_side <- function(piece_side) {
	strsplit(piece_side, "_")[[1]][2]
}

has_rank <- function(cs) {
	!(cs %in%
		c(
			"tile_back",
			"coin_back",
			"suitdie_face",
			"pawn_face",
			"pawn_back",
			"belt_face",
			"saucer_face",
			"saucer_back"
		))
}

PYRAMID_WIDTHS <- seq(from = 1 / 2, to = 27 / 32, length.out = 6)
PYRAMID_HEIGHTS <- 1.538842 * PYRAMID_WIDTHS
PYRAMID_DIAGONALS <- sqrt(PYRAMID_HEIGHTS^2 + (0.5 * PYRAMID_WIDTHS)^2)
PYRAMID_LAYOUT_WIDTHS <- PYRAMID_HEIGHTS
PYRAMID_LAYOUT_HEIGHTS <- 2 * PYRAMID_DIAGONALS
W <- 3 / 16
S <- 1
MATCHSTICK_WIDTHS <- c(2 * W, rep(W, 5))
MATCHSTICK_HEIGHTS <- c(
	2 * W,
	S - W,
	sqrt(2) * S - W,
	2 * S - W,
	sqrt(5 * S^2) - W,
	2 * sqrt(2) * S - W
)

# nolint start
# piece_side="pawn_face"
# suit=0
# rank=0
# style="font"
#
# if(is.na(piece_side)) {
#     piece <- NULL
# } else {
#     piece <- get_piece(piece_side)
# }
# suited <- is_suited(piece_side, suit, cfg) %||% FALSE
# piece_affixes <- c(paste0(".", piece_side), paste0(".", piece), "")
# piece_score <- c(2, 1, 0)
# rank_affixes <- c(paste0(".r", rank), "")
# rank_score <- c(1, 0)
# suit_affixes <- c(paste0(".s", suit), ifelse(suited, ".suited", ".unsuited"), "")
# suit_score <- c(2, 1, 0)
# dfs <- expand.grid(piece_affixes, rank_affixes, suit_affixes, stringsAsFactors=FALSE)
# names(dfs) <- c("piece", "rank", "suit")
# dfs$order <- 1:nrow(dfs)
# dfs <- left_join(dfs, data.frame(piece=piece_affixes, piece_score, stringsAsFactors=FALSE), by="piece")
# dfs <- left_join(dfs, data.frame(rank=rank_affixes, rank_score, stringsAsFactors=FALSE), by="rank")
# dfs <- left_join(dfs, data.frame(suit=suit_affixes, suit_score, stringsAsFactors=FALSE), by="suit")
# dfs <- mutate(dfs, score = 1e3 * piece_score + 1e2 * rank_score + 1e1 * suit_score,
#               affix = paste0(suit, rank, piece))
# dfs <- arrange(dfs, desc(score))
# dfs$order
# nolint end

style_ordering <- c(1, 7, 13, 4, 10, 16, 2, 8, 14, 5, 11, 17, 3, 9, 15, 6, 12, 18)

get_style_element <- function(
	style,
	piece_side = NA,
	cfg = list(),
	default = NULL,
	suit = 0,
	rank = 0
) {
	if (is.na(piece_side)) {
		piece <- NULL
	} else {
		piece <- get_piece(piece_side)
	}
	suited <- is_suited(piece_side, suit, rank, cfg) %||% FALSE

	piece_affixes <- c(paste0(".", piece_side), paste0(".", piece), "")
	rank_affixes <- c(paste0(".r", rank), "")
	suit_affixes <- c(paste0(".s", suit), ifelse(suited, ".suited", ".unsuited"), "")
	dfs <- expand.grid(piece_affixes, rank_affixes, suit_affixes, stringsAsFactors = FALSE)
	names(dfs) <- c("piece", "rank", "suit")
	affixes <- paste0(dfs$suit, dfs$rank, dfs$piece)[style_ordering]

	for (affix in affixes) {
		string <- paste0(style, affix)
		value <- cfg[[string]]
		if (!is.null(value)) return(value)
	}
	return(default)
}

is_legit_cfg_style <- function(cfg_style) {
	# style
	# style.suit
	# style.rank
	# style.piece
	# style.suit.rank
	# style.suit.piece
	# style.rank.piece
	# style.suit.rank.piece
	ss <- cleave2(cfg_style, sep = "\\.")
	if (!is_legit_style(ss[1])) {
		return(FALSE)
	}
	if (length(ss) == 2) {
		return(is_legit_suit(ss[2]) || is_legit_rank(ss[2]) || is_legit_piece(ss[2]))
	}
	if (length(ss) == 3) {
		return(
			(is_legit_suit(ss[2]) && (is_legit_rank(ss[3]) || is_legit_piece(ss[3]))) ||
				(is_legit_rank(ss[2]) && is_legit_piece(ss[3]))
		)
	}
	if (length(ss) == 4) {
		return(is_legit_suit(ss[2]) && is_legit_rank(ss[3]) && is_legit_piece(ss[4]))
	}
	if (length(ss) > 4) {
		return(FALSE)
	}
	TRUE
}

styles <- c(
	paste(
		c("ps", "dm"),
		rep(c("text", "fontface", "fontfamily", "fontsize", "cex", "t", "r", "color"), each = 2),
		sep = "_"
	),
	"shape",
	"shape_t",
	"shape_r",
	"background_color",
	"invert_colors",
	"border_color",
	"border_lex",
	"gridline_color",
	"gridline_lex",
	"edge_color",
	"annotation_color",
	"mat_color",
	"mat_width",
	"suit_color",
	paste(c("rank", "suit"), rep(c("text", "fontface", "fontfamily", "cex"), each = 2), sep = "_"),
	"use_suit_as_ace",
	"fontfamily",
	"fontface",
	"cex",
	"n_ranks",
	"n_suits",
	"coin_arrangement",
	"die_arrangement",
	"width",
	"height",
	"depth",
	"grob_fn",
	"op_grob_fn",
	"grob_with_bleed_fn",
	"obj_fn",
	"rayrender_fn",
	"rayvertex_fn",
	"rgl_fn",
	"lacks_rank",
	"lacks_suit",
	"title",
	"description",
	"credit",
	"copyright",
	"spdx_id"
)
is_legit_style <- function(style) {
	style %in% styles
}
is_legit_rank <- function(rank) {
	grepl("r[[:digit:]]+", rank)
}
is_legit_suit <- function(suit) {
	(suit %in% c("suited", "unsuited")) || grepl("s[[:digit:]]", suit)
}
pieces <- c(
	paste0(
		rep(
			c("tile", "coin", "pawn", "saucer", "matchstick", "pyramid", "bit", "board", "card"),
			3
		),
		rep(c("", "_face", "_back"), each = 9)
	),
	"die",
	"belt",
	"die_face",
	"belt_face",
	"pyramid_left",
	"pyramid_right"
)
is_legit_piece <- function(piece) {
	piece %in% pieces
}
warn_cfg <- function(cfg) {
	for (nn in names(cfg)) {
		if (!is_legit_cfg_style(nn)) {
			warn(paste(nn, "is not a recognized configuration"))
		}
	}
	if (!is.null(cfg$spdx_id)) {
		check_spdx_id(cfg$spdx_id)
	}
}

check_spdx_id <- function(spdx_id) {
	if (!(spdx_id %in% piecepackr::spdx_license_list$id)) {
		warn(
			c(
				sprintf("'%s' not recognized as an appropriate SPDX Identifier.", spdx_id),
				i = "See https://spdx.org/licenses/ for list of SPDX Identifiers."
			),
			class = "piecepackr_unrecognized_spdx_id"
		)
	} else if (piecepackr::spdx_license_list[spdx_id, "deprecated"]) {
		inform(
			c(
				sprintf("'%s' is a deprecated SPDX Identifier.", spdx_id),
				i = "See https://spdx.org/licenses/ for list of SPDX Identifiers."
			),
			class = "piecepackr_deprecated_spdx_id"
		)
	}
}

make_get_style_fn <- function(style, default) {
	function(cfg = list()) get_style_element(style, cfg = cfg, default = default)
}

is_edge <- function(piece_side) {
	(get_side(piece_side) %in% c("top", "left", "right", "base")) &&
		(get_piece(piece_side) != "pyramid")
}

get_background_color_helper <- function(piece_side, suit, rank, cfg) {
	default <- ifelse(
		is_edge(piece_side),
		get_edge_color(paste0(get_piece(piece_side), "_face"), suit, rank, cfg),
		"white"
	)
	colors <- col_cleave(get_style_element(
		"background_color",
		piece_side,
		cfg,
		default,
		suit,
		rank
	))
	expand_suit_elements(colors, "background_colors", piece_side, cfg)[suit]
}

get_border_colors <- function(piece_side = NA, suit = 0, rank = 0, cfg = list(), expand = TRUE) {
	border_colors <- col_cleave(get_style_element(
		"border_color",
		piece_side,
		cfg,
		"grey",
		suit,
		rank
	))
	if (expand) {
		border_colors <- expand_suit_elements(border_colors, "border_colors", piece_side, cfg)
	}
	border_colors
}

get_border_color <- function(piece_side, suit, rank, cfg) {
	get_border_colors(piece_side, suit, rank, cfg)[suit]
}

get_border_lex <- function(piece_side, suit, rank, cfg) {
	lex <- get_style_element("border_lex", piece_side, cfg, 1, suit, rank)
	lex
}

get_shape_t <- function(piece_side, suit, rank, cfg) {
	t <- numeric_cleave(get_style_element("shape_t", piece_side, cfg, 90, suit, rank))
	t <- expand_suit_elements(t, "shape_t", piece_side, cfg)
	t[suit]
}

get_shape_r <- function(piece_side, suit, rank, cfg) {
	shape <- get_shape(piece_side, suit, rank, cfg)
	piece <- get_piece(piece_side)
	default <- switch(shape, kite = 0.25, roundrect = switch(piece, tile = 0.12, 0.06), 0.2)
	r <- numeric_cleave(get_style_element("shape_r", piece_side, cfg, default, suit, rank))
	r <- expand_suit_elements(r, "shape_r", piece_side, cfg)
	r[suit]
}

get_shape <- function(piece_side, suit, rank, cfg) {
	default <- switch(
		piece_side,
		coin_back = "circle",
		coin_face = "circle",
		bit_face = "circle",
		bit_back = "circle",
		pawn_face = "halma",
		pawn_back = "halma",
		saucer_face = "circle",
		saucer_back = "circle",
		pyramid_face = "pyramid",
		pyramid_left = "pyramid",
		pyramid_right = "pyramid",
		pyramid_back = "pyramid",
		"rect"
	)
	get_style_element("shape", piece_side, cfg, default, suit, rank)
}

get_n_vertices <- function(shape) {
	as.numeric(gsub("convex|concave", "", shape))
}

get_suit_color_helper <- function(piece_side, suit, rank, cfg = list()) {
	default <- ifelse(
		is_edge(piece_side),
		get_edge_color(paste0(get_piece(piece_side), "_face"), suit, rank, cfg),
		"#D55E00,#000000,#009E73,#56B4E9,#E69F00"
	)
	suit_colors <- col_cleave(get_style_element("suit_color", piece_side, cfg, default, suit, rank))
	suit_colors <- expand_suit_elements(suit_colors, "suit_colors", piece_side, cfg)
	n_suits <- cfg$n_suits
	ifelse(suit <= n_suits, suit_colors[suit], suit_colors[n_suits + 1L])
}

should_invert <- function(piece_side, suit, rank, cfg) {
	should_inverts <- get_style_element("invert_colors", piece_side, cfg, FALSE, suit, rank)
	expand_suit_elements(should_inverts, "should_inverts", piece_side, cfg)[suit]
}

is_suited <- function(piece_side, suit, rank, cfg) {
	if (is.na(piece_side)) {
		FALSE
	} else {
		!(piece_side %in% cfg$lacks_suit) && (suit <= cfg$n_suits + 1L)
	}
}

get_dm_t <- function(piece_side, suit, rank, cfg) {
	default <- ifelse(piece_side %in% c("tile_face", "die_face", "suitdie_face"), 135, 90)
	default <- ifelse(piece_side == "matchstick_face" && rank == 1, 135, default)
	t <- numeric_cleave(get_style_element("dm_t", piece_side, cfg, default, suit, rank))
	expand_suit_elements(t, "dm_t", piece_side, cfg)[suit]
}
get_dm_r <- function(piece_side, suit, rank, cfg) {
	shape <- get_shape(piece_side, suit, rank, cfg)
	r_corner <- sqrt(0.25^2 + 0.25^2)
	default <- switch(
		shape,
		rect = switch(piece_side, matchstick_face = ifelse(rank > 1, 0.4, r_corner), r_corner),
		circle = switch(piece_side, saucer_face = 0.3, saucer_back = 0.3, r_corner),
		halma = 0.25,
		pyramid = 0.1,
		0.3
	)
	r <- numeric_cleave(get_style_element("dm_r", piece_side, cfg, default, suit, rank))
	expand_suit_elements(r, "dm_r", piece_side, cfg)[suit]
}

get_ps_t <- function(piece_side, suit, rank, cfg) {
	shape <- get_shape(piece_side, suit, rank, cfg)
	default <- switch(shape, halma = -90, pyramid = -90, 0)
	t <- numeric_cleave(get_style_element("ps_t", piece_side, cfg, default, suit, rank))
	expand_suit_elements(t, "ps_t", piece_side, cfg)[suit]
}
get_ps_r <- function(piece_side, suit, rank, cfg) {
	shape <- get_shape(piece_side, suit, rank, cfg)
	default <- switch(shape, halma = 0.25, pyramid = 0.25, 0.0)
	r <- numeric_cleave(get_style_element("ps_r", piece_side, cfg, default, suit, rank))
	expand_suit_elements(r, "ps_r", piece_side, cfg)[suit]
}

get_dm_symbols <- function(piece_side, suit = 0, rank = 0, cfg = list()) {
	default <- {
		if (is_edge(piece_side)) {
			dm_symbols <- ""
		} else if (piece_side %in% c("coin_back", "coin_face")) {
			dm_symbols <- "\u25cf"
		} else if (piece_side %in% c("saucer_back", "saucer_face")) {
			dm_symbols <- "\u25b2"
		} else if (piece_side %in% c("pawn_face", "pyramid_face")) {
			dm_symbols <- "\u0298\u0298"
		} else if (
			piece_side %in%
				c(
					"suitdie_face",
					"pawn_back",
					"belt_face",
					"tile_back",
					"matchstick_back",
					"card_back",
					"pyramid_left",
					"pyramid_right",
					"pyramid_back"
				)
		) {
			dm_symbols <- ""
		} else {
			dm_symbols <- get_suit_symbols(piece_side, suit, rank, cfg)
		}
	}
	default <- paste(default, collapse = ",")
	dm_symbols <- cleave2(get_style_element("dm_text", piece_side, cfg, default, suit, rank))
	dm_symbols <- expand_suit_elements(dm_symbols, "suit_symbols", piece_side, cfg)
	dm_symbols
}

get_dm_text <- function(piece_side, suit, rank, cfg) {
	get_dm_symbols(piece_side, suit, rank, cfg)[suit]
}

get_dm_color <- function(piece_side, suit, rank, cfg) {
	default <- get_suit_color(piece_side, suit, rank, cfg)
	colors <- get_style_element("dm_color", piece_side, cfg, default, suit, rank)
	colors <- col_cleave(colors)
	colors <- expand_suit_elements(colors, "suit_colors", piece_side, cfg)
	colors[suit]
}

get_background_color <- function(piece_side, suit, rank, cfg) {
	bcol <- get_background_color_helper(piece_side, suit, rank, cfg)
	scol <- get_suit_color_helper(piece_side, suit, rank, cfg)
	if (should_invert(piece_side, suit, rank, cfg)) {
		scol
	} else {
		bcol
	}
}

get_suit_color <- function(piece_side, suit, rank, cfg) {
	bcol <- get_background_color_helper(piece_side, suit, rank, cfg)
	scol <- get_suit_color_helper(piece_side, suit, rank, cfg)
	if (should_invert(piece_side, suit, rank, cfg)) {
		bcol
	} else {
		scol
	}
}
get_gridline_color <- function(piece_side, suit, rank, cfg) {
	default <- switch(
		piece_side,
		tile_back = get_suit_color(piece_side, cfg$n_suits + 1L, rank, cfg),
		board_face = get_suit_color(piece_side, suit, rank, cfg),
		board_back = get_suit_color(piece_side, suit, rank, cfg),
		"transparent"
	)
	colors <- get_style_element("gridline_color", piece_side, cfg, default, suit, rank)
	colors <- col_cleave(colors)
	colors <- expand_suit_elements(colors, "gridline_colors", piece_side, cfg)
	colors[suit]
}

get_gridline_lex <- function(piece_side, suit, rank, cfg) {
	lex <- get_style_element("gridline_lex", piece_side, cfg, 1, suit, rank)
	lex
}

get_mat_color <- function(piece_side, suit, rank, cfg) {
	default <- {
		if (piece_side %in% c("belt_face", "saucer_face", "saucer_back")) {
			get_suit_color(piece_side, suit, rank, cfg)
		} else {
			"transparent"
		}
	}
	colors <- get_style_element("mat_color", piece_side, cfg, default, suit, rank)
	colors <- col_cleave(colors)
	colors <- expand_suit_elements(colors, "suit_colors", piece_side, cfg)
	colors[suit]
}
get_mat_width <- function(piece_side, suit, rank, cfg) {
	default <- {
		if (piece_side == "belt_face") {
			"0.2,0"
		} else if (piece_side %in% c("saucer_face", "saucer_back")) {
			"0.1428571" # 1/7 which mimics 5/8" pawn in 7/8" saucer
		} else {
			"0"
		}
	}
	widths <- get_style_element("mat_width", piece_side, cfg, default, suit, rank)
	widths <- numeric_cleave(widths)
}

get_edge_color <- function(piece_side, suit, rank, cfg) {
	background_color <- get_background_color(piece_side, suit = suit, rank = rank, cfg)
	neutral_col <- get_background_color("tile_back", suit = cfg$n_suits + 1L, rank = 0L, cfg)

	piece <- get_piece(piece_side)
	default <- switch(
		piece,
		bit = background_color,
		pawn = background_color,
		die = background_color,
		matchstick = get_background_color("matchstick_back", suit = suit, rank = rank, cfg),
		pyramid = background_color,
		neutral_col
	)
	colors <- get_style_element("edge_color", piece_side, cfg, default, suit, rank)
	colors <- col_cleave(colors)
	colors <- expand_suit_elements(colors, "suit_colors", piece_side, cfg)
	colors[suit]
}

get_suit_symbol <- function(piece_side, suit, rank, cfg) {
	get_suit_symbols(piece_side, suit, rank, cfg)[suit]
}

get_rank_symbols <- function(piece_side = NA, suit = 0, rank = 0, cfg = list(), expand = TRUE) {
	default <- "n,a,2,3,4,5"
	rank_symbols <- cleave2(get_style_element("rank_text", piece_side, cfg, default, suit, rank))
	if (expand) {
		rank_symbols <- expand_rank_elements(rank_symbols, "rank_symbols", piece_side, cfg)
	}
	rank_symbols
}
get_suit_symbols <- function(piece_side = NA, suit = 0, rank = 0, cfg = list(), expand = TRUE) {
	default <- "\u2665,\u2660,\u2663,\u2666,\u263c"
	suit_symbols <- cleave2(get_style_element("suit_text", piece_side, cfg, default, suit, rank))
	if (expand) {
		suit_symbols <- expand_suit_elements(suit_symbols, "suit_symbols", piece_side, cfg)
	}
	suit_symbols
}

expand_suit_elements <- function(elements, style, piece_side, cfg) {
	n_suits <- cfg$n_suits
	if (length(elements) < n_suits) {
		elements <- rep(elements, length.out = n_suits + 1)
	} else if (length(elements) == n_suits) {
		elements <- c(elements, switch(style, cex = 1.0, suit_colors = "grey", ""))
	}
	if (length(elements) == n_suits + 1L) {
		elements <- c(
			elements,
			switch(
				style,
				suit_symbols = switch(
					ifelse(is.na(piece_side), "NA", piece_side),
					suitdie_face = "",
					die_face = "",
					elements[n_suits + 1L]
				),
				gridline_colors = NA,
				elements[n_suits + 1L]
			)
		)
	}
	elements
}

expand_rank_elements <- function(elements, style, piece_side, cfg) {
	n_ranks <- cfg$n_ranks
	if (length(elements) < n_ranks) {
		elements <- rep(elements, length.out = n_ranks)
	}
	if (length(elements) == n_ranks) {
		elements <- c(elements, switch(style, rank_symbols = "", elements[n_ranks]))
	}
	elements
}

use_suit_as_ace <- function(piece_side = NA, suit = 0, rank = 0, cfg = list()) {
	use_suit_as_ace <- get_style_element("use_suit_as_ace", piece_side, cfg, FALSE, suit, rank)
	ifelse(use_suit_as_ace && rank == 2, TRUE, FALSE)
}

get_rank_symbol <- function(piece_side, suit, rank, cfg) {
	if (use_suit_as_ace(piece_side, suit, rank, cfg)) {
		suit_symbols <- get_suit_symbols(piece_side, suit, rank, cfg, expand = FALSE)
		expand_suit_elements(suit_symbols, "rank_suit_symbols", piece_side, cfg)[suit]
	} else {
		get_rank_symbols(piece_side, suit, rank, cfg)[rank]
	}
}

get_rank_cex <- function(piece_side, suit, rank, cfg) {
	if (use_suit_as_ace(piece_side, suit, rank, cfg)) {
		get_suit_cex(piece_side, suit, rank, cfg)
	} else {
		cex <- numeric_cleave(get_style_element("rank_cex", piece_side, cfg, 1.0, suit, rank))
		expand_rank_elements(cex, "cex", piece_side, cfg)[rank]
	}
}
get_rank_fontfamily <- function(piece_side, suit, rank, cfg) {
	if (use_suit_as_ace(piece_side, suit, rank, cfg)) {
		get_suit_fontfamily(piece_side, suit, rank, cfg)
	} else {
		fontfamilies <- cleave2(get_style_element(
			"rank_fontfamily",
			piece_side,
			cfg,
			cfg$fontfamily %||% "sans",
			suit,
			rank
		))
		expand_rank_elements(fontfamilies, "font", piece_side, cfg)[rank]
	}
}
get_rank_fontface <- function(piece_side, suit, rank, cfg) {
	if (use_suit_as_ace(piece_side, suit, rank, cfg)) {
		get_suit_fontface(piece_side, suit, rank, cfg)
	} else {
		fontfaces <- cleave2(get_style_element(
			"rank_fontface",
			piece_side,
			cfg,
			cfg$fontface %||% "plain",
			suit,
			rank
		))
		expand_rank_elements(fontfaces, "fontface", piece_side, cfg)[rank]
	}
}
get_suit_fonts <- function(piece_side, suit, rank, cfg) {
	fonts <- cleave2(get_style_element(
		"suit_fontfamily",
		piece_side,
		cfg,
		cfg$fontfamily %||% "sans",
		suit,
		rank
	))
	expand_suit_elements(fonts, "font", piece_side, cfg)
}
get_suit_fontfamily <- function(piece_side, suit, rank, cfg) {
	get_suit_fonts(piece_side, suit, rank, cfg)[suit]
}
get_suit_fontface <- function(piece_side, suit, rank, cfg) {
	fontfaces <- cleave2(get_style_element(
		"suit_fontface",
		piece_side,
		cfg,
		cfg$fontface %||% "plain",
		suit,
		rank
	))
	expand_suit_elements(fontfaces, "fontface", piece_side, cfg)[suit]
}
get_suit_cex <- function(piece_side, suit, rank, cfg) {
	cex <- numeric_cleave(get_style_element("suit_cex", piece_side, cfg, 1.0, suit, rank))
	expand_suit_elements(cex, "cex", piece_side, cfg)[suit]
}
are_suits_dm <- function(piece_side, suit, rank, cfg) {
	identical(
		get_dm_symbols(piece_side, suit, rank, cfg),
		get_suit_symbols(piece_side, suit, rank, cfg)
	)
}
get_dm_fontface <- function(piece_side, suit, rank, cfg) {
	if (are_suits_dm(piece_side, suit, rank, cfg)) {
		default <- get_suit_fontface(piece_side, suit, rank, cfg)
	} else {
		default <- cfg$fontface %||% "plain"
	}
	fonts <- cleave2(get_style_element("dm_fontface", piece_side, cfg, default))
	expand_suit_elements(fonts, "fontface", piece_side, cfg)[suit]
}
get_dm_fontfamily <- function(piece_side, suit, rank, cfg) {
	if (are_suits_dm(piece_side, suit, rank, cfg)) {
		default <- get_suit_fontfamily(piece_side, suit, rank, cfg)
	} else {
		default <- cfg$fontfamily %||% "sans"
	}
	fonts <- cleave2(get_style_element("dm_fontfamily", piece_side, cfg, default))
	expand_suit_elements(fonts, "font", piece_side, cfg)[suit]
}
get_dm_cex <- function(piece_side, suit, rank, cfg) {
	if (are_suits_dm(piece_side, suit, rank, cfg)) {
		default <- get_suit_cex(piece_side, suit, rank, cfg)
	} else {
		default <- 1.0
	}
	cex <- numeric_cleave(get_style_element("dm_cex", piece_side, cfg, default, suit, rank))
	expand_suit_elements(cex, "cex", piece_side, cfg)[suit]
}

get_suit_fontsize <- function(piece_side, suit, rank, cfg) {
	default <- switch(
		piece_side,
		"belt_face" = 22,
		"coin_back" = 34,
		"pawn_face" = 28,
		"pawn_back" = 28,
		"bit_back" = 34,
		"saucer_back" = 28,
		"saucer_face" = 28,
		"suitdie_face" = 32,
		"pyramid_face" = 48 * PYRAMID_WIDTHS[rank],
		"pyramid_back" = 48 * PYRAMID_WIDTHS[rank],
		24
	)
	get_style_element("suit_fontsize", piece_side, cfg, default, suit, rank)
}

get_dm_fontsize <- function(piece_side, suit, rank, cfg) {
	default <- switch(
		piece_side,
		"tile_face" = 40,
		"pawn_face" = 12,
		"pawn_back" = 12,
		"pyramid_face" = 12 * PYRAMID_WIDTHS[rank],
		"card_face" = 32,
		12
	)
	get_style_element("dm_fontsize", piece_side, cfg, default, suit, rank)
}

get_rank_fontsize <- function(piece_side, suit, rank, cfg) {
	default <- switch(
		piece_side,
		"die_face" = 20,
		"coin_face" = 28,
		"tile_face" = 72,
		"bit_face" = 28,
		"pyramid_left" = 48 * PYRAMID_WIDTHS[rank],
		"pyramid_right" = 48 * PYRAMID_WIDTHS[rank],
		"card_face" = 28,
		20
	)
	get_style_element("rank_fontsize", piece_side, cfg, default, suit, rank)
}


get_ps_element <- function(piece_side, suit_element, rank_element, neither_element = NA) {
	if (
		piece_side %in%
			c(
				"coin_face",
				"die_face",
				"tile_face",
				"pyramid_left",
				"pyramid_right",
				"matchstick_face",
				"bit_face",
				"card_face"
			)
	) {
		rank_element
	} else if (piece_side %in% c("tile_back", "matchstick_back", "card_back")) {
		neither_element
	} else if (is_edge(piece_side)) {
		neither_element
	} else {
		suit_element
	}
}
get_ps_fontface <- function(piece_side, suit, rank, cfg) {
	rank_fontface <- get_rank_fontface(piece_side, suit, rank, cfg)
	suit_fontface <- get_suit_fontface(piece_side, suit, rank, cfg)
	default <- get_ps_element(piece_side, suit_fontface, rank_fontface, "plain")
	get_style_element("ps_fontface", piece_side, cfg, default, suit, rank)
}
get_ps_fontfamily <- function(piece_side, suit, rank, cfg) {
	rank_font <- get_rank_fontfamily(piece_side, suit, rank, cfg)
	suit_font <- get_suit_fontfamily(piece_side, suit, rank, cfg)
	default <- get_ps_element(piece_side, suit_font, rank_font)
	get_style_element("ps_fontfamily", piece_side, cfg, default, suit, rank)
}
get_ps_fontsize <- function(piece_side, suit, rank, cfg) {
	rank_fontsize <- get_rank_fontsize(piece_side, suit, rank, cfg)
	suit_fontsize <- get_suit_fontsize(piece_side, suit, rank, cfg)
	default <- get_ps_element(piece_side, suit_fontsize, rank_fontsize)
	get_style_element("ps_fontsize", piece_side, cfg, default, suit, rank)
}
get_ps_cex <- function(piece_side, suit, rank, cfg) {
	rank_cex <- get_rank_cex(piece_side, suit, rank, cfg)
	suit_cex <- get_suit_cex(piece_side, suit, rank, cfg)
	default <- get_ps_element(piece_side, suit_cex, rank_cex)
	get_style_element("ps_cex", piece_side, cfg, default, suit, rank)
}
get_ps_text <- function(piece_side, suit, rank, cfg) {
	rank_symbol <- get_rank_symbol(piece_side, suit, rank, cfg)
	suit_symbol <- get_suit_symbol(piece_side, suit, rank, cfg)
	default <- get_ps_element(piece_side, suit_symbol, rank_symbol, "")
	get_style_element("ps_text", piece_side, cfg, default, suit, rank)
}
get_ps_color <- function(piece_side, suit, rank, cfg) {
	default <- get_suit_color(piece_side, suit, rank, cfg)
	get_style_element("ps_color", piece_side, cfg, default, suit, rank)
}
get_piece_opt_helper <- function(piece_side, suit, rank, cfg) {
	# Shape
	shape <- get_shape(piece_side, suit, rank, cfg)
	shape_r <- get_shape_r(piece_side, suit, rank, cfg)
	shape_t <- get_shape_t(piece_side, suit, rank, cfg)
	back <- grepl("back", piece_side)

	# Additional colors
	background_color <- get_background_color(piece_side, suit, rank, cfg)
	border_color <- get_border_color(piece_side, suit, rank, cfg)
	border_lex <- get_border_lex(piece_side, suit, rank, cfg)
	gridline_color <- get_gridline_color(piece_side, suit, rank, cfg)
	gridline_lex <- get_gridline_lex(piece_side, suit, rank, cfg)
	mat_color <- get_mat_color(piece_side, suit, rank, cfg)
	mat_width <- get_mat_width(piece_side, suit, rank, cfg)
	edge_color <- get_edge_color(piece_side, suit, rank, cfg)

	# Overall scaling factor
	cex <- cfg[["cex"]] %||% 1.0

	# Directional mark symbol
	dm_color <- get_dm_color(piece_side, suit, rank, cfg)
	dm_cex <- get_dm_cex(piece_side, suit, rank, cfg)
	dm_fontfamily <- get_dm_fontfamily(piece_side, suit, rank, cfg)
	dm_fontface <- get_dm_fontface(piece_side, suit, rank, cfg)
	dm_fontsize <- cex * dm_cex * get_dm_fontsize(piece_side, suit, rank, cfg)
	dm_text <- get_dm_text(piece_side, suit, rank, cfg)
	dm_t <- get_dm_t(piece_side, suit, rank, cfg)
	dm_r <- get_dm_r(piece_side, suit, rank, cfg)
	dm_x <- to_x(dm_t, dm_r) + 0.5
	dm_y <- to_y(dm_t, dm_r) + 0.5

	# Primary symbol
	ps_color <- get_ps_color(piece_side, suit, rank, cfg)
	ps_cex <- get_ps_cex(piece_side, suit, rank, cfg)
	ps_fontfamily <- get_ps_fontfamily(piece_side, suit, rank, cfg)
	ps_fontface <- get_ps_fontface(piece_side, suit, rank, cfg)
	ps_fontsize <- cex * ps_cex * get_ps_fontsize(piece_side, suit, rank, cfg)
	ps_text <- get_ps_text(piece_side, suit, rank, cfg)
	ps_t <- get_ps_t(piece_side, suit, rank, cfg)
	ps_r <- get_ps_r(piece_side, suit, rank, cfg)
	ps_x <- to_x(ps_t, ps_r) + 0.5
	ps_y <- to_y(ps_t, ps_r) + 0.5

	# Extra info
	rank_text <- get_rank_symbol(piece_side, suit, rank, cfg)
	suit_text <- get_suit_symbol(piece_side, suit, rank, cfg)

	opt <- list(
		shape = shape,
		shape_r = shape_r,
		shape_t = shape_t,
		back = back,
		background_color = background_color,
		edge_color = edge_color,
		border_color = border_color,
		border_lex = border_lex,
		gridline_color = gridline_color,
		gridline_lex = gridline_lex,
		mat_color = mat_color,
		mat_width = mat_width,
		dm_color = dm_color,
		dm_text = dm_text,
		dm_fontsize = dm_fontsize,
		dm_fontfamily = dm_fontfamily,
		dm_fontface = dm_fontface,
		dm_x = dm_x,
		dm_y = dm_y,
		ps_color = ps_color,
		ps_text = ps_text,
		ps_fontsize = ps_fontsize,
		ps_fontfamily = ps_fontfamily,
		ps_fontface = ps_fontface,
		ps_x = ps_x,
		ps_y = ps_y,
		rank_text = rank_text,
		suit_text = suit_text
	)
	opt$bleed_color <- get_bleed_color(opt)
	opt
}

get_bleed_color <- function(opt) {
	if (!is_color_invisible(opt$border_color) && !nigh(opt$border_lex, 0)) {
		return(opt$border_color)
	}
	if (!is_color_invisible(opt$mat_color) && !nigh(opt$mat_width, 0)) {
		return(opt$mat_color)
	}
	if (!is_color_invisible(opt$background_color)) {
		return(opt$background_color)
	}
	return("transparent")
}

test_that("options work as expected", {
	cfg <- list(
		invert_colors.suited = TRUE,
		background_color = "white",
		suit_color = "darkred,black,darkgreen,darkblue,grey",
		lacks_suit = "coin_face",
		n_suits = 4
	)
	expect_equal(get_suit_color("coin_back", 1, 1, cfg), "white")
	expect_equal(get_suit_color("coin_face", 5, 1, cfg), "grey")
	expect_equal(pp_cfg(cfg)$get_suit_color(3:2), c("darkgreen", "black"))
	colors <- c("darkred", "black", "darkgreen", "darkblue", "grey")
	cfg <- list(
		invert_colors.suited = TRUE,
		background_color = "white",
		suit_color = colors,
		lacks_suit = "coin_face",
		n_suits = 4
	)
	expect_equal(get_suit_color("coin_back", 1, 1, cfg), "white")
	expect_equal(get_suit_color("coin_face", 5, 1, cfg), "grey")
	cfg <- pp_cfg(cfg)
	expect_equal(cfg$get_suit_color(3:2), colors[3:2])
	expect_equal(cfg$get_suit_color(), colors[1:4])

	cfg <- list(n_suits = 4)
	expect_equal(should_invert("coin_back", 1, 1, cfg), FALSE)
	expect_equal(should_invert("coin_face", 5, 1, cfg), FALSE)
	cfg <- list(invert_colors = TRUE, n_suits = 4)
	expect_equal(should_invert("coin_back", 1, 1, cfg), TRUE)
	expect_equal(should_invert("coin_face", 5, 1, cfg), TRUE)
	cfg <- list(invert_colors = FALSE, n_suits = 4)
	expect_equal(should_invert("coin_back", 1, 1, cfg), FALSE)
	expect_equal(should_invert("coin_face", 5, 1, cfg), FALSE)
	cfg <- list(invert_colors.suited = TRUE, lacks_suit = "coin_face", n_suits = 4)
	expect_equal(should_invert("coin_back", 1, 1, cfg), TRUE)
	expect_equal(should_invert("coin_face", 5, 1, cfg), FALSE)
	cfg <- list(invert_colors.unsuited = TRUE, lacks_suit = "coin_face", n_suits = 4)
	expect_equal(should_invert("coin_back", 1, 1, cfg), FALSE)
	expect_equal(should_invert("coin_face", 5, 1, cfg), TRUE)
	cfg <- list(invert_colors.coin_back = TRUE, n_suits = 4)
	expect_equal(should_invert("coin_back", 1, 1, cfg), TRUE)
	expect_equal(should_invert("coin_face", 5, 1, cfg), FALSE)
	cfg <- list(invert_colors.coin_face = TRUE, n_suits = 4)
	expect_equal(should_invert("coin_back", 1, 1, cfg), FALSE)
	expect_equal(should_invert("coin_face", 5, 1, cfg), TRUE)

	cfg <- pp_cfg(list(n_ranks = 7, n_suits = 3))
	expect_equal(cfg$n_ranks, 7)
	expect_equal(cfg$n_suits, 3)

	cfg <- list(rank_symbols = "A,B,C,D,E,F", use_suit_as_ace = TRUE, n_ranks = 6, n_suits = 4)
	expect_equal(get_rank_symbol("die_face", 6, 2, cfg), "\u263c")

	cfg <- list(background_color.unsuited = "orange", lacks_suit = "tile_back", n_suits = 4)
	expect_equal(get_background_color("tile_face", 1, 1, cfg), "white")
	expect_equal(get_background_color("tile_back", 1, 1, cfg), "orange")
	expect_equal(is_suited("die_face", 5, 1, list(n_suits = 4)), TRUE)
	expect_equal(is_suited("die_face", 6, 1, list(n_suits = 4)), FALSE)
	expect_equal(get_background_color("die_face", 5, 1, cfg), "white")
	expect_equal(get_background_color("die_face", 6, 1, cfg), "orange")
	cfg1 <- list(dm_text = "", n_suits = 4)
	cfg2 <- list(dm_text = ",,,,", n_suits = 4)
	expect_equal(get_dm_text("suitdie_face", 3, 1, cfg1), "")
	expect_equal(get_dm_text("suitdie_face", 3, 1, cfg2), "")
	expect_equal(get_dm_text("suitdie_face", 6, 1, cfg1), "")
	expect_equal(get_dm_text("suitdie_face", 6, 1, cfg2), "")
	expect_equal(get_dm_text("die_face", 6, 1, cfg2), "")
	cfg <- list(dm_text.saucer_face = "b", n_suits = 4)
	expect_equal(get_dm_text("saucer_face", 3, 1, cfg), "b")
	cfg <- list(dm_text.saucer_back = "b", n_suits = 4)
	expect_equal(get_dm_text("saucer_back", 5, 1, cfg), "b")

	cfg <- list(
		invert_colors.suited = TRUE,
		invert_colors.r1 = FALSE,
		suit_color = "darkred,black,darkgreen,darkblue,grey",
		n_suits = 4
	)
	expect_equal(get_background_color("tile_face", 1, 2, cfg), "darkred")
	expect_equal(get_background_color("tile_face", 1, 1, cfg), "white")
})

test_that("get_piece_opt works as expected", {
	cfg <- pp_cfg(list(suit_color = "darkred,black,darkgreen,darkblue,grey"))
	opt <- cfg$get_piece_opt("coin_face")
	expect_equal(opt$background_color, "white")
	expect_equal(opt$border_color, "grey")
	expect_equal(opt$bleed_color, "grey")
	expect_equal(opt$gridline_color, "transparent")
	expect_equal(opt$ps_color, "grey")
	expect_equal(opt$ps_text, "n")
	expect_equal(opt$ps_fontsize, 28)
	expect_equal(opt$ps_fontfamily, "sans")
	expect_equal(opt$ps_fontface, "plain")
	expect_equal(opt$dm_color, "grey")
	expect_equal(opt$dm_text, "\u25cf")
	expect_equal(opt$shape, "circle")
	expect_equal(opt$shape_r, 0.2)
	expect_equal(opt$shape_t, 90)
	expect_equal(opt$dm_fontsize, 12)
	expect_equal(opt$dm_fontfamily, "sans")
	expect_equal(opt$dm_fontface, "plain")
	expect_equal(opt$dm_x, 0.5)
	expect_equal(opt$dm_y, to_y(90, sqrt(0.25^2 + 0.25^2)) + 0.5)
	expect_equal(opt$ps_x, 0.5)
	expect_equal(opt$ps_y, 0.5)
	expect_equal(opt$dm_text, "\u25cf")

	cfg <- pp_cfg()
	opt <- cfg$get_piece_opt("saucer_back")
	expect_equal(opt$dm_text, "\u25b2")
	opt <- cfg$get_piece_opt("pawn_face")
	expect_equal(opt$dm_text, "\u0298\u0298")
	opt <- cfg$get_piece_opt("pawn_back")
	expect_equal(opt$dm_text, "")

	cfg <- pp_cfg(list(suit_color = "white"))
	opt <- cfg$get_piece_opt("tile_face", suit = 2, rank = 2)
	expect_equal(opt$ps_col, "white")
	cfg <- pp_cfg(list(suit_color.suited = "white"))
	opt <- cfg$get_piece_opt("tile_face", suit = 2, rank = 2)
	expect_equal(opt$ps_col, "white")
	cfg <- pp_cfg(list(suit_color.s2 = "white"))
	opt <- cfg$get_piece_opt("tile_face", suit = 2, rank = 2)
	expect_equal(opt$ps_col, "white")

	cfg <- pp_cfg(list(border_lex = 0))
	expect_equal(cfg$get_piece_opt("tile_face")$bleed_color, "white")
	cfg <- pp_cfg(list(border_color = "transparent"))
	expect_equal(cfg$get_piece_opt("tile_face")$bleed_color, "white")
	cfg <- pp_cfg(list(border_color = "transparent", mat_color = "blue"))
	expect_equal(cfg$get_piece_opt("tile_face")$bleed_color, "white")
	cfg <- pp_cfg(list(border_color = "transparent", mat_color = "blue", mat_width = 0.1))
	expect_equal(cfg$get_piece_opt("tile_face")$bleed_color, "blue")
	cfg <- pp_cfg(list(
		border_color = "transparent",
		mat_color = "blue",
		background_color = "#00000000"
	))
	expect_equal(cfg$get_piece_opt("tile_face")$bleed_color, "transparent")

	# edges
	cfg <- pp_cfg(list(
		suit_color = "darkred,black,darkgreen,darkblue,grey",
		edge_color = "darkred,black,darkgreen,darkblue,grey"
	))
	opt <- cfg$get_piece_opt("pawn_top", 1, 1)
	expect_equal(opt$shape, "rect")
	expect_equal(opt$dm_text, "")
	expect_equal(opt$ps_text, "")
	expect_equal(opt$background_color, "darkred")
})

test_that("pp_cfg querying variables work as expected", {
	cfg <- pp_cfg()
	expect_true(cfg$has_piecepack)

	expect_true(cfg$has_pawns)
	expect_true(cfg$has_coins)
	expect_true(cfg$has_tiles)
	expect_true(cfg$has_dice)

	cfg$has_dice <- FALSE
	cfg$has_tiles <- FALSE
	expect_false(cfg$has_tiles)
	expect_false(cfg$has_dice)
	expect_false(cfg$has_piecepack)
	cfg$has_piecepack <- TRUE
	expect_true(cfg$has_tiles)
	expect_true(cfg$has_dice)
	expect_true(cfg$has_piecepack)

	expect_false(cfg$has_saucers)
	expect_false(cfg$has_pyramids)
	expect_false(cfg$has_matchsticks)
	expect_error(cfg$has_piecepack <- 3, "is.logical\\(value)\\ is not TRUE")

	expect_true(cfg$cache_grob)
	cfg$cache_grob <- FALSE
	expect_false(cfg$cache_grob)
	cfg$cache_piece_opt <- FALSE
	expect_false(cfg$cache_piece_opt)
	cfg$cache_piece_opt <- TRUE
	expect_true(cfg$cache_piece_opt)
	cfg$cache_obj_fn <- FALSE
	expect_false(cfg$cache_obj_fn)
	cfg$cache_op_fn <- FALSE
	expect_false(cfg$cache_op_fn)
})

op_xy <- function(x, y, z, op_angle = 45, op_scale = 0) {
	x <- x + op_scale * z * cos(to_radians(op_angle))
	y <- y + op_scale * z * sin(to_radians(op_angle))
	list(x = x, y = y)
}

## Two-sided token
basicTokenGrob <- function(
	piece_side,
	suit,
	rank,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = unit(0, "npc"),
	angle = 0,
	type = "normal",
	width = NA,
	height = NA,
	depth = NA,
	op_scale = 0,
	op_angle = 45
) {
	side <- get_side(piece_side)

	if (side %in% c("face", "back")) {
		grob <- cfg$get_grob(piece_side, suit, rank, type)
		xy_p <- op_xy(x, y, z + 0.5 * depth, op_angle, op_scale)
		cvp <- viewport(xy_p$x, xy_p$y, width, height, angle = angle)
		grob <- grid::editGrob(grob, name = "piece_side", vp = cvp)

		edge <- basicTokenEdge(
			piece_side,
			suit,
			rank,
			cfg,
			x,
			y,
			z,
			angle,
			width,
			height,
			depth,
			op_scale,
			op_angle
		)
		edge <- grid::editGrob(edge, name = "other_faces")

		gl <- gList(edge, grob)

		gTree(scale = 1, type = type, children = gl, cl = "basic_projected_token")
	} else {
		generalTokenGrob(
			piece_side,
			suit,
			rank,
			cfg,
			x,
			y,
			z,
			angle,
			type,
			width,
			height,
			depth,
			op_scale,
			op_angle
		)
	}
}

#' @export
makeContent.basic_projected_token <- function(x) {
	gp <- gpar(cex = x$scale, lex = x$scale)
	x$children$other_faces <- update_gp(x$children$other_faces, gp)

	if (hasName(x$children$piece_side, "scale")) {
		x$children$piece_side$scale <- x$scale
	} else if (x$type == "normal") {
		x$children$piece_side <- update_gp(x$children$piece_side, gp)
	}

	x
}

#' @export
grobCoords.basic_projected_token <- function(x, closed, ...) {
	if (is.null(x$children$other_faces$coords_xyl)) {
		NextMethod()
	} else {
		xylists_to_grobcoords(x$children$other_faces$coords_xyl, x$name, closed)
	}
}

## Token edge-side grobs
basicTokenEdge <- function(
	piece_side,
	suit,
	rank,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = unit(0, "npc"),
	angle = 0,
	width = NA,
	height = NA,
	depth = NA,
	op_scale = 0,
	op_angle = 45
) {
	cfg <- as_pp_cfg(cfg)
	opt <- cfg$get_piece_opt(piece_side, suit, rank)
	piece <- get_piece(piece_side)
	side <- ifelse(opt$back, "back", "face") #### allow limited 3D rotation #281

	x <- convertX(x, "in", valueOnly = TRUE)
	y <- convertY(y, "in", valueOnly = TRUE)
	z <- convertX(z, "in", valueOnly = TRUE)
	width <- convertX(width, "in", valueOnly = TRUE)
	height <- convertY(height, "in", valueOnly = TRUE)
	depth <- convertX(depth, "in", valueOnly = TRUE)

	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)
	whd <- get_scaling_factors(side, width, height, depth)
	pc <- as_coord3d(x, y, z)
	R <- side_R(side) %*% AA_to_R(angle, axis_x = 0, axis_y = 0)
	token <- Token2S$new(shape, whd, pc, R)

	gl <- gList()
	# opposite side
	#### Could use transformation grob (flipped)
	opp_side <- ifelse(opt$back, "face", "back")
	opp_piece_side <- paste0(piece, "_", opp_side)
	opp_opt <- cfg$get_piece_opt(opp_piece_side, suit, rank)
	gp_opp <- gpar(
		col = opp_opt$border_color,
		fill = opp_opt$background_color,
		lex = opp_opt$border_lex
	)
	xyz_opp <- if (opt$back) token$xyz_face else token$xyz_back
	xy_opp <- as_coord2d(xyz_opp, alpha = degrees(op_angle), scale = op_scale)

	grob_opposite <- polygonGrob(
		x = xy_opp$x,
		y = xy_opp$y,
		default.units = "in",
		gp = gp_opp,
		name = "opposite_piece_side"
	)

	# edges
	edges <- token$op_edges(op_angle)
	for (i in seq_along(edges)) {
		name <- paste0("edge", i)
		gl[[i]] <- edges[[i]]$op_grob(op_angle, op_scale, name = name)
	}
	gp_edge <- gpar(col = opt$border_color, fill = opt$edge_color, lex = opt$border_lex)
	grob_edge <- gTree(children = gl, gp = gp_edge, name = "token_edges")

	# pre-compute grobCoords
	if (shape$convex) {
		coords_xyl <- as.list(convex_hull2d(as_coord2d(
			token$xyz,
			alpha = degrees(op_angle),
			scale = op_scale
		)))
	} else {
		coords_xyl <- NULL
	}

	gTree(coords_xyl = coords_xyl, children = gList(grob_opposite, grob_edge))
}

generalTokenGrob <- function(
	piece_side,
	suit,
	rank,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = unit(0, "npc"),
	angle = 0,
	type = "normal",
	width = NA,
	height = NA,
	depth = NA,
	op_scale = 0,
	op_angle = 45
) {
	cfg <- as_pp_cfg(cfg)
	piece <- get_piece(piece_side)
	side <- get_side(piece_side)
	opt <- cfg$get_piece_opt(paste0(piece, "_face"), suit, rank)
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	x <- convertX(x, "in", valueOnly = TRUE)
	y <- convertY(y, "in", valueOnly = TRUE)
	z <- convertX(z, "in", valueOnly = TRUE)
	width <- convertX(width, "in", valueOnly = TRUE)
	height <- convertY(height, "in", valueOnly = TRUE)
	depth <- convertX(depth, "in", valueOnly = TRUE)

	#### Generalize axis_x, axis_y #281
	axis_x <- 0
	axis_y <- 0

	# geometric vertices
	whd <- get_scaling_factors(side, width, height, depth)
	pc <- as_coord3d(x, y, z)
	R <- side_R(side) %*% AA_to_R(angle, axis_x, axis_y)
	token <- Token2S$new(shape, whd, pc, R)

	gl <- gList()
	edges <- token$op_edges(op_angle)
	for (i in seq_along(edges)) {
		name <- paste0("edge", i)
		gl[[i]] <- edges[[i]]$op_grob(op_angle, op_scale, name = name)
	}
	gp_edge <- gpar(col = opt$border_color, fill = opt$edge_color, lex = opt$border_lex)
	grob_edge <- gTree(children = gl, gp = gp_edge, name = "token_edges")

	side_visible <- token$visible_side(op_angle)
	piece_side <- paste0(piece, "_", side_visible)
	xy_vp <- token$op_xy_vp(op_angle, op_scale, side_visible)
	xy_polygon <- as_coord2d(
		token$xyz_side(side_visible),
		alpha = degrees(op_angle),
		scale = op_scale
	)
	ps_grob <- at_ps_grob(piece_side, suit, rank, cfg, xy_vp, xy_polygon)

	gl <- gList(grob_edge, ps_grob)

	# pre-compute grobCoords
	if (shape$convex) {
		coords_xyl <- as.list(convex_hull2d(as_coord2d(
			token$xyz,
			alpha = degrees(op_angle),
			scale = op_scale
		)))
	} else {
		coords_xyl <- NULL
	}

	gTree(
		scale = 1,
		type = type,
		coords_xyl = coords_xyl,
		children = gl,
		cl = "general_projected_token"
	)
}

#' @export
makeContent.general_projected_token <- function(x) {
	gp <- gpar(cex = x$scale, lex = x$scale)
	x$children$token_edges <- update_gp(x$children$token_edges, gp)
	x$children$piece_side$scale <- x$scale
	x
}

#' @export
grobCoords.general_projected_token <- function(x, closed, ...) {
	if (is.null(x$coords_xyl)) {
		NextMethod()
	} else {
		xylists_to_grobcoords(x$coords_xyl, x$name, closed)
	}
}

## Die
basicDieGrob <- function(
	piece_side,
	suit,
	rank,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = unit(0, "npc"),
	angle = 0,
	type = "normal",
	width = NA,
	height = NA,
	depth = NA,
	op_scale = 0,
	op_angle = 45
) {
	grob <- cfg$get_grob(piece_side, suit, rank, type)
	xy_p <- op_xy(x, y, z + 0.5 * depth, op_angle, op_scale)
	cvp <- viewport(xy_p$x, xy_p$y, width, height, angle = angle)
	grob <- grid::editGrob(grob, name = "top_face", vp = cvp)

	x <- convertX(x, "in", valueOnly = TRUE)
	y <- convertY(y, "in", valueOnly = TRUE)
	z <- convertX(z, "in", valueOnly = TRUE)
	width <- convertX(width, "in", valueOnly = TRUE)
	height <- convertY(height, "in", valueOnly = TRUE)
	depth <- convertX(depth, "in", valueOnly = TRUE)

	#### allow limited 3D rotation #281
	axis_x <- 0
	axis_y <- 0

	edge <- basicDieEdge(
		piece_side,
		suit,
		rank,
		cfg,
		x,
		y,
		z,
		angle,
		axis_x,
		axis_y,
		width,
		height,
		depth,
		op_scale,
		op_angle
	)
	edge <- grid::editGrob(edge, name = "other_faces")

	gl <- gList(edge, grob)

	# pre-compute grobCoords
	coords_xyl <- die_grobcoords_xyl(
		suit,
		rank,
		cfg,
		x,
		y,
		z,
		angle,
		axis_x,
		axis_y,
		width,
		height,
		depth,
		op_scale,
		op_angle
	)

	gTree(
		scale = 1,
		type = type,
		coords_xyl = coords_xyl,
		children = gl,
		cl = c("basic_projected_die", "coords_xyl")
	)
}

#' @export
grobCoords.coords_xyl <- function(x, closed, ...) {
	xylists_to_grobcoords(x$coords_xyl, x$name, closed)
}

#' @export
makeContent.basic_projected_die <- function(x) {
	gp <- gpar(cex = x$scale, lex = x$scale)
	for (i in seq_along(x$children$other_faces$children)) {
		if (hasName(x$children$other_faces$children[[i]], "scale")) {
			x$children$other_faces$children[[i]]$scale <- x$scale
		} else {
			x$children$other_faces$children[[i]] <- update_gp(
				x$children$other_faces$children[[i]],
				gp
			)
		}
	}
	if (hasName(x$children$top_face, "scale")) {
		x$children$top_face$scale <- x$scale
	} else if (x$type == "normal") {
		x$children$top_face <- update_gp(x$children$top_face, gp)
	}
	x
}

die_grobcoords_xyl <- function(
	suit,
	rank,
	cfg,
	x,
	y,
	z,
	angle,
	axis_x,
	axis_y,
	width,
	height,
	depth,
	op_scale,
	op_angle
) {
	opt <- cfg$get_piece_opt("die_face", suit, rank)
	if (opt$shape == "roundrect") {
		xyz <- rounded_die_xyz(
			suit,
			rank,
			cfg,
			x,
			y,
			z,
			angle,
			axis_x,
			axis_y,
			width,
			height,
			depth
		)
	} else {
		xyz <- die_xyz(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth)
	}
	as.list(convex_hull2d(as_coord2d(xyz, alpha = degrees(op_angle), scale = op_scale)))
}

die_face_polygon_3d <- function(f_xyz, shape_r, width, height) {
	npc <- pp_shape("roundrect", radius = shape_r, width = width, height = height)$npc_coords
	# affine_settings expects corners as: upper-left, lower-left, lower-right, upper-right
	ul <- f_xyz[1L]
	ll <- f_xyz[2L]
	lr <- f_xyz[3L]
	as_coord3d(
		ll$x + npc$x * (lr$x - ll$x) + npc$y * (ul$x - ll$x),
		ll$y + npc$x * (lr$y - ll$y) + npc$y * (ul$y - ll$y),
		ll$z + npc$x * (lr$z - ll$z) + npc$y * (ul$z - ll$z)
	)
}

die_roundrect_hull_grob <- function(
	suit,
	rank,
	cfg,
	opt,
	x,
	y,
	z,
	angle,
	axis_x,
	axis_y,
	width,
	height,
	depth,
	op_angle,
	op_scale
) {
	xyz <- rounded_die_xyz(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth)
	hull <- convex_hull2d(as_coord2d(xyz, alpha = degrees(op_angle), scale = op_scale))
	gp <- gpar(col = opt$border_color, fill = opt$edge_color, lex = opt$border_lex)
	polygonGrob(x = hull$x, y = hull$y, default.units = "in", gp = gp, name = "convex_hull")
}

basicDieEdge <- function(
	piece_side,
	suit,
	rank,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = unit(0, "npc"),
	angle = 0,
	axis_x = 0,
	axis_y = 0,
	width = NA,
	height = NA,
	depth = NA,
	op_scale = 0,
	op_angle = 45
) {
	cfg <- as_pp_cfg(cfg)
	opt <- cfg$get_piece_opt(piece_side, suit, rank)

	lf <- get_die_faces(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth)

	indices_visible <- visible_die_faces(lf, op_angle)

	if (opt$shape == "roundrect") {
		hull_grob <- die_roundrect_hull_grob(
			suit,
			rank,
			cfg,
			opt,
			x,
			y,
			z,
			angle,
			axis_x,
			axis_y,
			width,
			height,
			depth,
			op_angle,
			op_scale
		)
		gl <- gList(hull_grob)
	} else {
		gl <- gList()
	}

	for (i in indices_visible) {
		xy <- as_coord2d(lf$f_xyz[[i]], alpha = degrees(op_angle), scale = op_scale)
		rank <- lf$die_face_info$rank[i]
		name <- paste0("die_side", i)
		if (opt$shape == "roundrect") {
			pts3d <- die_face_polygon_3d(lf$f_xyz[[i]], opt$shape_r, width, height)
			xy_polygon <- as_coord2d(pts3d, alpha = degrees(op_angle), scale = op_scale)
		} else {
			xy_polygon <- xy
		}
		gl[[length(gl) + 1L]] <- at_ps_grob(
			piece_side,
			suit,
			rank,
			cfg,
			xy,
			xy_polygon,
			name = name
		)
	}

	gTree(children = gl, name = "die_sides", cl = "basic_projected_die_edge")
}

basicEllipsoidFn <- function(shading = FALSE) {
	force(shading)
	function(
		piece_side,
		suit,
		rank,
		cfg = pp_cfg(),
		x = unit(0.5, "npc"),
		y = unit(0.5, "npc"),
		z = unit(0, "npc"),
		angle = 0,
		type = "normal",
		width = NA,
		height = NA,
		depth = NA,
		op_scale = 0,
		op_angle = 45
	) {
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)

		x <- convertX(x, "in", valueOnly = TRUE)
		y <- convertY(y, "in", valueOnly = TRUE)
		z <- convertX(z, "in", valueOnly = TRUE)
		width <- convertX(width, "in", valueOnly = TRUE)
		height <- convertY(height, "in", valueOnly = TRUE)
		depth <- convertX(depth, "in", valueOnly = TRUE)

		xyz <- ellipse_xyz()$scale(width, height, depth)$translate(x, y, z)
		xy <- as_coord2d(xyz, alpha = degrees(op_angle), scale = op_scale)
		xyh <- convex_hull2d(xy)

		gp_gf <- gpar(col = NA, fill = opt$background_color, lwd = 0)
		gf <- polygonGrob(
			x = xyh$x,
			y = xyh$y,
			default.units = "in",
			gp = gp_gf,
			name = "background"
		)

		gp_gb <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		gb <- polygonGrob(x = xyh$x, y = xyh$y, default.units = "in", gp = gp_gb, name = "border")

		if (shading && getRversion() >= "4.1") {
			# Get top of ellipsoid in npc units for gradient fill
			idm <- which.max(xyz$z)
			cx <- (xy$x[idm] - range(xyh)$x[1L]) / diff(range(xyh)$x)
			cy <- (xy$y[idm] - range(xyh)$y[1L]) / diff(range(xyh)$y)
			rgr <- radialGradient(
				c(update_alpha_col(opt$background_color, 0.500), "#00000080"),
				r1 = 0.1,
				cx1 = cx,
				cx2 = cx,
				cy1 = cy,
				cy2 = cy
			)
			gp_gr <- gpar(col = opt$border_color, fill = rgr)
			gr <- polygonGrob(
				x = xyh$x,
				y = xyh$y,
				default.units = "in",
				gp = gp_gr,
				name = "shading"
			)
		} else {
			gr <- nullGrob(name = "shading")
		}

		gl <- gList(gf, gr, gb)

		coords_xyl <- as.list(xyh)

		gTree(
			scale = 1,
			type = type,
			coords_xyl = coords_xyl,
			shading = shading,
			children = gl,
			cl = c("projected_ellipsoid", "coords_xyl")
		)
	}
}

#' @export
makeContent.projected_ellipsoid <- function(x) {
	if (x$shading && !has_radial_gradients()) {
		rgr_inform()
	}
	x
}

## Pyramid Top
basicPyramidTop <- function(
	piece_side,
	suit,
	rank,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = unit(0, "npc"),
	angle = 0,
	type = "normal",
	width = NA,
	height = NA,
	depth = NA,
	op_scale = 0,
	op_angle = 45
) {
	cfg <- as_pp_cfg(cfg)

	x <- convertX(x, "in", valueOnly = TRUE)
	y <- convertY(y, "in", valueOnly = TRUE)
	z <- convertX(z, "in", valueOnly = TRUE)
	width <- convertX(width, "in", valueOnly = TRUE)
	height <- convertY(height, "in", valueOnly = TRUE)
	depth <- convertX(depth, "in", valueOnly = TRUE)
	xy_b <- npc_to_in(as_coord2d(rect_xy), x, y, width, height, angle)
	p <- as_polygon2d(xy_b)
	edge_types <- paste0("pyramid_", c("left", "back", "right", "face"))
	order <- painter_order(p, alpha = degrees(op_angle))
	df <- tibble(index = 1:4, edge = edge_types)[order, ]
	gl <- gList()
	for (i in 1:4) {
		opt <- cfg$get_piece_opt(df$edge[i], suit, rank)
		gp <- gpar(col = opt$border_color, lex = opt$border_lex, fill = opt$background_color)
		edge <- p$edges[df$index[i]]

		ex <- c(x, edge$p1$x, edge$p2$x)
		ey <- c(y, edge$p1$y, edge$p2$y)
		ez <- c(z + 0.5 * depth, z - 0.5 * depth, z - 0.5 * depth)
		xyz_polygon <- as_coord3d(x = ex, y = ey, z = ez)
		xy_polygon <- as_coord2d(xyz_polygon, alpha = degrees(op_angle), scale = op_scale)
		xy_vp <- xy_vp_ps(xyz_polygon, op_scale, op_angle)

		piece_side <- df$edge[i]
		gl[[i]] <- at_ps_grob(piece_side, suit, rank, cfg, xy_vp, xy_polygon, name = piece_side)
	}

	#### allow limited 3D rotation #281
	axis_x <- 0
	axis_y <- 0
	# pre-compute grobCoords
	coords_xyl <- pt_grobcoords_xyl(
		x,
		y,
		z,
		angle,
		axis_x,
		axis_y,
		width,
		height,
		depth,
		op_scale,
		op_angle
	)

	gTree(
		scale = 1,
		type = type,
		coords_xyl = coords_xyl,
		children = gl,
		cl = c("projected_pyramid_top", "coords_xyl")
	)
}

pt_grobcoords_xyl <- function(
	x,
	y,
	z,
	angle,
	axis_x,
	axis_y,
	width,
	height,
	depth,
	op_scale,
	op_angle
) {
	xyz <- pt_xyz(x, y, z, angle, axis_x, axis_y, width, height, depth)
	as.list(convex_hull2d(as_coord2d(xyz, alpha = degrees(op_angle), scale = op_scale)))
}

#' @export
makeContent.projected_pyramid_top <- function(x) {
	gp <- gpar(cex = x$scale, lex = x$scale)
	for (i in 1:4) {
		if (hasName(x$children[[i]], "scale")) {
			x$children[[i]]$scale <- x$scale
		} else if (x$type == "normal") {
			x$children[[i]] <- update_gp(x$children[[i]], gp)
		}
	}
	x
}

## Pyramid side
basicPyramidSide <- function(
	piece_side,
	suit,
	rank,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = unit(0, "npc"),
	angle = 0,
	type = "normal",
	width = NA,
	height = NA,
	depth = NA,
	op_scale = 0,
	op_angle = 45
) {
	cfg <- as_pp_cfg(cfg)

	x <- convertX(x, "in", valueOnly = TRUE)
	y <- convertY(y, "in", valueOnly = TRUE)
	z <- convertX(z, "in", valueOnly = TRUE)
	width <- convertX(width, "in", valueOnly = TRUE)
	height <- convertY(height, "in", valueOnly = TRUE)
	depth <- convertX(depth, "in", valueOnly = TRUE)

	xy_b <- npc_to_in(as_coord2d(pyramid_xy), x, y, width, height, angle)
	p <- as_polygon2d(xy_b)
	xy_tip <- xy_b[1]

	theta <- 2 * asin(0.5 * width / height)
	yt <- 1 - cos(theta)
	xy_t <- npc_to_in(as_coord2d(x = 0:1, y = yt), x, y, width, height, angle)

	gl <- gList()

	## opposite edge
	opposite_edge <- switch(
		piece_side,
		"pyramid_face" = "pyramid_back",
		"pyramid_back" = "pyramid_face",
		"pyramid_left" = "pyramid_right",
		"pyramid_right" = "pyramid_left"
	)
	xyz_polygon <- as_coord3d(x = xy_b$x[c(1, 3:2)], y = xy_b$y[c(1, 3:2)], z = z - 0.5 * depth)
	xy_polygon <- as_coord2d(xyz_polygon, alpha = degrees(op_angle), scale = op_scale)
	xy_vp <- xy_vp_ps(xyz_polygon, op_scale, op_angle)
	gl[[1]] <- at_ps_grob(opposite_edge, suit, rank, cfg, xy_vp, xy_polygon, name = opposite_edge)

	## side edges
	edge_types <- paste0(
		"pyramid_",
		switch(
			piece_side,
			"pyramid_face" = c("right", "bottom", "left"),
			"pyramid_left" = c("face", "bottom", "back"),
			"pyramid_back" = c("left", "bottom", "right"),
			"pyramid_right" = c("back", "bottom", "face")
		)
	)

	order <- painter_order(p, alpha = degrees(op_angle))
	df <- tibble(index = 1:3, edge = edge_types)[order, ]
	gli <- 2

	for (i in 1:3) {
		edge_ps <- df$edge[i]
		index <- df$index[i]
		if (edge_ps == "pyramid_bottom") {
			next
		}
		opt <- cfg$get_piece_opt(edge_ps, suit, rank)
		gp <- gpar(col = opt$border_color, lex = opt$border_lex, fill = opt$background_color)
		edge <- p$edges[index]
		if (index == 1) {
			# right side viewed top (left side viewed on side)
			ex <- c(edge$p1$x, edge$p2$x, xy_t$x[1])
			ey <- c(edge$p1$y, edge$p2$y, xy_t$y[1])
			ez <- c(z - 0.5 * depth, z - 0.5 * depth, z + 0.5 * depth)
		} else {
			# left side viewed top (right side viewed on side)
			ex <- c(edge$p2$x, xy_t$x[2], edge$p1$x)
			ey <- c(edge$p2$y, xy_t$y[2], edge$p1$y)
			ez <- c(z - 0.5 * depth, z + 0.5 * depth, z - 0.5 * depth)
		}
		xyz_polygon <- as_coord3d(x = ex, y = ey, z = ez)
		xy_polygon <- as_coord2d(xyz_polygon, alpha = degrees(op_angle), scale = op_scale)
		xy_vp <- xy_vp_ps(xyz_polygon, op_scale, op_angle)
		gl[[gli]] <- at_ps_grob(edge_ps, suit, rank, cfg, xy_vp, xy_polygon, name = edge_ps)
		gli <- gli + 1
	}

	## edge facing up
	x_f <- c(xy_tip$x, xy_t$x)
	y_f <- c(xy_tip$y, xy_t$y)
	z_f <- c(z - 0.5 * depth, z + 0.5 * depth, z + 0.5 * depth)
	opt <- cfg$get_piece_opt(piece_side, suit, rank)
	gp <- gpar(col = opt$border_color, lex = opt$border_lex, fill = opt$background_color)
	xyz_polygon <- as_coord3d(x = x_f, y = y_f, z = z_f)
	xy_polygon <- as_coord2d(xyz_polygon, alpha = degrees(op_angle), scale = op_scale)
	xy_vp <- xy_vp_ps(xyz_polygon, op_scale, op_angle)
	gl[[4]] <- at_ps_grob(piece_side, suit, rank, cfg, xy_vp, xy_polygon, name = piece_side)

	#### allow limited 3D rotation #281
	axis_x <- 0
	axis_y <- 0
	# pre-compute grobCoords
	coords_xyl <- ps_grobcoords_xyl(
		x,
		y,
		z,
		angle,
		axis_x,
		axis_y,
		width,
		height,
		depth,
		op_scale,
		op_angle
	)

	gTree(
		scale = 1,
		type = type,
		coords_xyl = coords_xyl,
		children = gl,
		cl = c("projected_pyramid_side", "coords_xyl")
	)
}

ps_grobcoords_xyl <- function(
	x,
	y,
	z,
	angle,
	axis_x,
	axis_y,
	width,
	height,
	depth,
	op_scale,
	op_angle
) {
	xyz <- ps_xyz(x, y, z, angle, axis_x, axis_y, width, height, depth)
	as.list(convex_hull2d(as_coord2d(xyz, alpha = degrees(op_angle), scale = op_scale)))
}

#' @export
makeContent.projected_pyramid_side <- function(x) {
	gp <- gpar(cex = x$scale, lex = x$scale)
	for (i in 1:4) {
		if (hasName(x$children[[i]], "scale")) {
			x$children[[i]]$scale <- x$scale
		} else if (x$type == "normal") {
			x$children[[i]] <- update_gp(x$children[[i]], gp)
		}
	}
	x
}

# compute `xy_vp` for pyramid sides
xy_vp_ps <- function(xyz_polygon, op_scale, op_angle) {
	p_midbottom <- mean(xyz_polygon[2:3])
	p_diff <- xyz_polygon[1] - p_midbottom
	p_ul <- xyz_polygon[2] + p_diff
	p_ur <- xyz_polygon[3] + p_diff

	x <- c(p_ul$x, xyz_polygon$x[2:3], p_ur$x)
	y <- c(p_ul$y, xyz_polygon$y[2:3], p_ur$y)
	z <- c(p_ul$z, xyz_polygon$z[2:3], p_ur$z)

	as_coord2d(as_coord3d(x, y, z), alpha = degrees(op_angle), scale = op_scale)
}

#### `grobCoords()`
basicHoledBoardFn <- function(nrows = 4L, ncols = 4L, margin = 0) {
	# nolint
	force(nrows)
	force(ncols)
	force(margin)
	function(
		piece_side,
		suit,
		rank,
		cfg = pp_cfg(),
		x = unit(0.5, "npc"),
		y = unit(0.5, "npc"),
		z = unit(0, "npc"),
		angle = 0,
		type = "normal",
		width = NA,
		height = NA,
		depth = NA,
		op_scale = 0,
		op_angle = 45
	) {
		side <- get_side(piece_side)
		stopifnot(side %in% c("face", "back"))

		# Top side
		grob <- cfg$get_grob(piece_side, suit, rank, type)
		xy_p <- op_xy(x, y, z + 0.5 * depth, op_angle, op_scale)
		cvp <- viewport(xy_p$x, xy_p$y, width, height, angle = angle)
		grob <- grid::editGrob(grob, name = "piece_side", vp = cvp)

		# Opposite side
		xy_op <- op_xy(x, y, z - 0.5 * depth, op_angle, op_scale)
		cv_op <- viewport(xy_op$x, xy_op$y, width, height, angle = angle)
		grob_op <- grid::editGrob(grob, name = "opposite_piece_side", vp = cv_op)

		# Edges
		cfg <- as_pp_cfg(cfg)
		opt <- cfg$get_piece_opt(piece_side, suit, rank)
		piece <- get_piece(piece_side)
		side <- ifelse(opt$back, "back", "face") #### allow limited 3D rotation #281

		x <- convertX(x, "in", valueOnly = TRUE)
		y <- convertY(y, "in", valueOnly = TRUE)
		z <- convertX(z, "in", valueOnly = TRUE)
		width <- convertX(width, "in", valueOnly = TRUE)
		height <- convertY(height, "in", valueOnly = TRUE)
		depth <- convertX(depth, "in", valueOnly = TRUE)

		# Exterior edges
		shape <- pp_shape(
			opt$shape,
			opt$shape_t,
			opt$shape_r,
			opt$back,
			width = opt$shape_w,
			height = opt$shape_h
		)
		whd <- get_scaling_factors(side, width, height, depth)
		pc <- as_coord3d(x, y, z)
		R <- side_R(side) %*% AA_to_R(angle, axis_x = 0, axis_y = 0)
		token <- Token2S$new(shape, whd, pc, R)

		gl <- gList()
		edges <- token$op_edges(op_angle)
		for (i in seq_along(edges)) {
			name <- paste0("outside_edge", i)
			gl[[i]] <- edges[[i]]$op_grob(op_angle, op_scale, name = name)
		}

		# Interior hole edges
		xc <- rep(seq.int(ncols) - 0.5 + margin, times = nrows) / (ncols + 2 * margin) # npc
		yc <- rep(seq.int(nrows) - 0.5 + margin, each = ncols) / (nrows + 2 * margin) # npc
		r <- RADIUS_BOARD_HOLES / (min(nrows, ncols) + 2 * margin) # snpc
		# Convert npc units to inches, rotate by `angle`, and translate to `x` and `y`
		stopifnot(width == height) # snpc `r` may not work for non-square holed boards
		xc <- width * (xc - 0.5)
		yc <- height * (yc - 0.5)
		wc <- 2 * width * r
		tc <- to_t(xc, yc)
		rc <- to_r(xc, yc)
		xc <- to_x(tc + angle, rc) + x
		yc <- to_y(tc + angle, rc) + y

		shape_circle <- pp_shape("circle")
		lc <- purrr::pmap(list(xc = xc, yc = yc), function(xc, yc) {
			whd <- get_scaling_factors(side, wc, wc, depth)
			pc <- as_coord3d(xc, yc, z)
			token <- Token2S$new(shape_circle, whd, pc)
		})

		for (i in seq_along(lc)) {
			edges <- lc[[i]]$op_edges(op_angle)
			for (j in seq_along(edges)) {
				name <- str_glue("hole{i}_edge{j}")
				gl[[length(gl) + 1L]] <- edges[[j]]$op_grob(op_angle, op_scale, name = name)
			}
		}

		gp_edge <- gpar(col = opt$border_color, fill = opt$edge_color, lex = opt$border_lex)
		grob_edge <- gTree(children = gl, gp = gp_edge, name = "token_edges")

		gl <- gList(grob_op, grob_edge, grob)
		gTree(scale = 1, type = type, children = gl, cl = "projected_holed_board")
	}
}

#' @export
makeContent.projected_holed_board <- makeContent.basic_projected_token

d4Grob <- function(piece_side, suit, rank, cfg = pp_cfg()) {
	cfg <- as_pp_cfg(cfg)
	opt <- cfg$get_piece_opt(piece_side, suit, rank)
	gTree(
		opt = opt,
		border = TRUE,
		flip = FALSE,
		scale = 1,
		name = NULL,
		gp = gpar(),
		vp = NULL,
		cl = c("d4_die", "basic_piece_side")
	)
}

#' @export
makeContent.d4_die <- function(x) {
	opt <- x$opt
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	# Background
	background_grob <- shape$shape(
		gp = gpar(col = NA, fill = opt$background_color),
		name = "background"
	)

	# Top (rank)
	gp_text <- gpar(
		col = opt$ps_color,
		fill = opt$ps_color,
		fontsize = opt$ps_fontsize,
		fontfamily = opt$ps_fontfamily,
		fontface = opt$ps_fontface
	)
	r <- 0.2
	top_grob <- textGrob(
		label = opt$ps_text,
		gp = gp_text,
		x = 0.5,
		y = 0.5 + r,
		hjust = 0.5,
		vjust = 0.5,
		name = "top_symbol"
	)

	left <- switch(opt$ps_text, `1` = 2, `2` = 4, `3` = 4, `4` = 2)
	right <- switch(opt$ps_text, `1` = 3, `2` = 3, `3` = 1, `4` = 1)
	# Left/right
	left_grob <- textGrob(
		label = left,
		gp = gp_text,
		x = 0.5,
		y = 0.5 + r,
		hjust = 0.5,
		vjust = 0.5,
		vp = viewport(angle = 120),
		,
		name = "left_symbol"
	)
	right_grob <- textGrob(
		label = right,
		gp = gp_text,
		x = 0.5,
		y = 0.5 + r,
		hjust = 0.5,
		vjust = 0.5,
		vp = viewport(angle = 240),
		name = "right_symbol"
	)

	# Border
	if (x$border) {
		gp_border <- gpar(col = opt$border_color, fill = NA, lex = opt$border_lex)
		border_grob <- shape$shape(gp = gp_border, name = "border")
	} else {
		border_grob <- nullGrob(name = "border")
	}
	if (x$flip) {
		gl <- gList(top_grob, left_grob, right_grob, background_grob, border_grob)
	} else {
		gl <- gList(background_grob, top_grob, left_grob, right_grob, border_grob)
	}

	setChildren(x, gl)
}

# https://mathworld.wolfram.com/RegularTetrahedron.html

## d4 Top
d4TopGrob <- function(
	piece_side,
	suit,
	rank,
	cfg = pp_cfg(),
	x = unit(0.5, "npc"),
	y = unit(0.5, "npc"),
	z = unit(0, "npc"),
	angle = 0,
	type = "normal",
	width = NA,
	height = NA,
	depth = NA,
	op_scale = 0,
	op_angle = 45
) {
	cfg <- as_pp_cfg(cfg)

	x <- convertX(x, "in", valueOnly = TRUE)
	y <- convertY(y, "in", valueOnly = TRUE)
	z <- convertX(z, "in", valueOnly = TRUE)
	width <- convertX(width, "in", valueOnly = TRUE)
	height <- convertY(height, "in", valueOnly = TRUE)
	depth <- convertX(depth, "in", valueOnly = TRUE)
	xy_b <- npc_to_in(
		regular_ngon_polygon2d(3, x = 0.5, y = 0.5, theta = degrees(90)),
		x,
		y,
		width,
		height,
		angle
	)
	p <- as_polygon2d(xy_b)
	edge_types <- paste0("d4_", c("left", "face", "right"))
	order <- painter_order(p, alpha = degrees(op_angle))
	df <- tibble(index = 1:3, edge = edge_types)[order, ]
	gl <- gList()
	for (i in 1:3) {
		l_edge <- d4_edge_info(df$edge[i], rank, angle)
		opt <- cfg$get_piece_opt("die_face", suit, l_edge$rank)
		gp <- gpar(col = opt$border_color, lex = opt$border_lex, fill = opt$background_color)
		edge <- p$edges[df$index[i]]

		ex <- c(x, edge$p1$x, edge$p2$x)
		ey <- c(y, edge$p1$y, edge$p2$y)
		ez <- c(z + 0.5 * depth, z - 0.5 * depth, z - 0.5 * depth)
		xyz_polygon <- as_coord3d(x = ex, y = ey, z = ez)[cycle_d4(1:3, l_edge$vp_rot)]
		xy_polygon <- as_coord2d(xyz_polygon, alpha = degrees(op_angle), scale = op_scale)
		xy_vp <- xy_vp_convex(xyz_polygon, op_scale, op_angle)

		gl[[i]] <- at_ps_grob(
			"die_face",
			suit,
			l_edge$rank,
			cfg,
			xy_vp,
			xy_polygon,
			name = df$edge[i]
		)
	}

	#### allow limited 3D rotation #281
	axis_x <- 0
	axis_y <- 0
	####
	# pre-compute grobCoords
	coords_xyl <- d4t_grobcoords_xyl(
		x,
		y,
		z,
		angle,
		axis_x,
		axis_y,
		width,
		height,
		depth,
		op_scale,
		op_angle
	)

	gTree(
		scale = 1,
		coords_xyl = coords_xyl,
		children = gl,
		cl = c("projected_rpg_die", "coords_xyl")
	)
}

#' @export
makeContent.projected_rpg_die <- function(x) {
	for (i in seq_along(x$children)) {
		x$children[[i]]$scale <- x$scale
	}
	x
}

# We need to widen/lengthen viewport for "convex" shapes
# Assuming top vertex in viewport at (0.5, 1)
xy_vp_convex <- function(xyz_polygon, op_scale, op_angle) {
	n <- length(xyz_polygon)

	xy <- regular_ngon_polygon2d(n, x = 0.5, y = 0.5, theta = degrees(90))

	# widen viewport since "convex" shape doesn't reach edges of viewport
	i_left <- which.min(xy$x)
	i_right <- n + 2 - i_left

	p_mid_widest <- mean(xyz_polygon[c(i_left, i_right)])
	m_width <- 1 / (xy$x[i_right] - xy$x[i_left])
	p_left <- p_mid_widest + m_width * (xyz_polygon[i_left] - p_mid_widest)
	p_right <- p_mid_widest + m_width * (xyz_polygon[i_right] - p_mid_widest)

	up_diff <- xyz_polygon[1] - p_mid_widest
	p_ul <- p_left + up_diff
	p_ur <- p_right + up_diff

	# lengthen viewport since (odd) "convex" shapes don't reach bottom of viewport
	i_bottom <- which.min(xy$y)
	i_top <- which.max(xy$y)
	m_height <- 1 / (1 - xy$y[i_left])
	p_ll <- p_ul - m_height * up_diff
	p_lr <- p_ur - m_height * up_diff

	x <- c(p_ul$x, p_ll$x, p_lr$x, p_ur$x)
	y <- c(p_ul$y, p_ll$y, p_lr$y, p_ur$y)
	z <- c(p_ul$z, p_ll$z, p_lr$z, p_ur$z)

	as_coord2d(as_coord3d(x, y, z), alpha = degrees(op_angle), scale = op_scale)
}

d4t_grobcoords_xyl <- function(
	x,
	y,
	z,
	angle,
	axis_x,
	axis_y,
	width,
	height,
	depth,
	op_scale,
	op_angle
) {
	xyz <- d4t_xyz(x, y, z, angle, axis_x, axis_y, width, height, depth)
	as.list(convex_hull2d(as_coord2d(xyz, alpha = degrees(op_angle), scale = op_scale)))
}

d4t_xyz <- function(x, y, z, angle, axis_x, axis_y, width, height, depth) {
	xy <- regular_ngon_polygon2d(3, x = 0, y = 0, theta = degrees(90))
	xyz_t <- as_coord3d(x = 0.0, y = 0.0, z = 0.5)
	xyz_b <- as_coord3d(xy, z = -0.5)
	xs <- c(xyz_t$x, xyz_b$x)
	ys <- c(xyz_t$y, xyz_b$y)
	zs <- c(xyz_t$z, xyz_b$z)
	as_coord3d(xs, ys, zs)$scale(width, height, depth)$transform(AA_to_R(
		angle,
		axis_x,
		axis_y
	))$translate(x, y, z)
}

save_d4_obj <- function(
	piece_side,
	suit,
	rank,
	cfg,
	x,
	y,
	z,
	angle,
	axis_x,
	axis_y,
	width,
	height,
	depth,
	filename,
	scale,
	res
) {
	cfg <- as_pp_cfg(cfg)
	opt <- cfg$get_piece_opt(piece_side, suit, rank)
	xyz <- d4_xyz(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth)

	p <- regular_ngon_polygon2d(3, x = 0.5, y = 0.5, theta = degrees(90))
	xy_vt <- list(
		x = rep(c(0.5 * p$x, 0.5 * p$x + 0.5), 2),
		y = c(rep(0.5 * p$y + 0.5, 2), rep(0.5 * p$y, 2))
	)

	# textured face elements: face, left, right, bottom
	f <- list()
	fn_vt <- function(r) switch(r, 1:3, 4:6, 7:9, 10:12)
	f[[1]] <- list(v = 1:3, vt = fn_vt(rank))
	f[[2]] <- list(v = 4:6, vt = fn_vt(d4_edge_info("d4_left", rank)$rank))
	f[[3]] <- list(v = 7:9, vt = fn_vt(d4_edge_info("d4_right", rank)$rank))
	f[[4]] <- list(v = 10:12, vt = fn_vt(d4_edge_info("d4_bottom", rank)$rank))

	ext <- tools::file_ext(filename)
	mtl_filename <- gsub(paste0("\\.", ext, "$"), ".mtl", filename)
	png_filename <- gsub(paste0("\\.", ext, "$"), ".png", filename)

	write_obj(filename, v = xyz, vt = xy_vt, f = f)
	write_d4_texture("die_face", suit, rank, cfg, filename = png_filename, res = res)

	invisible(list(obj = filename, mtl = mtl_filename, png = png_filename))
}

d4_edge_info <- function(edge, rank, angle = 0) {
	if (edge == "d4_left") {
		edge_angle <- (angle - 120) %% 360
		edge_rank <- switch(rank, 4, 4, 2, 2)
		vp_rot <- switch(rank, 120, -120, 120, -120)
	} else if (edge == "d4_right") {
		edge_angle <- (angle + 120) %% 360
		edge_rank <- switch(rank, 3, 1, 1, 3)
		vp_rot <- switch(rank, 120, -120, 120, -120)
	} else if (edge == "d4_bottom") {
		edge_angle <- angle
		edge_rank <- switch(rank, 2, 3, 4, 1)
		vp_rot <- switch(rank, 120, 0, 120, 0)
	} else {
		# d4_face
		edge_angle <- angle
		edge_rank <- rank
		vp_rot <- 0
	}
	list(angle = edge_angle, rank = edge_rank, vp_rot = vp_rot)
}

cycle_d4 <- function(x, vp_rot) {
	if (vp_rot == 120) {
		cycle_elements(x, 1)
	} else if (vp_rot == -120) {
		cycle_elements(x, -1)
	} else {
		x
	}
}

d4_xyz <- function(suit, rank, cfg, x, y, z, angle, axis_x, axis_y, width, height, depth) {
	xyz_b <- as_coord3d(regular_ngon_polygon2d(3, x = 0, y = 0, theta = degrees(90)), z = -0.5)
	top <- as_coord3d(0, 0, 0.5)

	edge_types <- paste0("d4_", c("left", "face", "right"))

	# face
	xs_face <- c(top$x, xyz_b[2:3]$x)
	ys_face <- c(top$y, xyz_b[2:3]$y)
	zs_face <- c(top$z, xyz_b[2:3]$z)

	# left
	lr <- d4_edge_info("d4_left", rank)$vp_rot
	xs_left <- cycle_d4(c(top$x, xyz_b[1:2]$x), lr)
	ys_left <- cycle_d4(c(top$y, xyz_b[1:2]$y), lr)
	zs_left <- cycle_d4(c(top$z, xyz_b[1:2]$z), lr)

	# right
	rr <- d4_edge_info("d4_right", rank)$vp_rot
	xs_right <- cycle_d4(c(top$x, xyz_b[c(3, 1)]$x), rr)
	ys_right <- cycle_d4(c(top$y, xyz_b[c(3, 1)]$y), rr)
	zs_right <- cycle_d4(c(top$z, xyz_b[c(3, 1)]$z), rr)

	# bottom
	br <- d4_edge_info("d4_bottom", rank)$vp_rot
	xs_bottom <- cycle_d4(rev(xyz_b$x), br)
	ys_bottom <- cycle_d4(rev(xyz_b$y), br)
	zs_bottom <- cycle_d4(rev(xyz_b$z), br)

	xs <- c(xs_face, xs_left, xs_right, xs_bottom)
	ys <- c(ys_face, ys_left, ys_right, ys_bottom)
	zs <- c(zs_face, zs_left, zs_right, zs_bottom)

	as_coord3d(xs, ys, zs)$scale(width, height, depth)$transform(AA_to_R(
		angle,
		axis_x,
		axis_y
	))$translate(x, y, z)
}

write_d4_texture <- function(
	piece_side = "die_face",
	suit = 1,
	rank = 1,
	cfg = pp_cfg(),
	...,
	filename = tempfile(fileext = ".png"),
	res = 72
) {
	current_dev <- grDevices::dev.cur()
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	width <- cfg$get_width("die_face", suit, rank)

	args <- list(
		filename = filename,
		height = 2 * width,
		width = 2 * width,
		units = "in",
		res = res,
		bg = "transparent"
	)
	if (capabilities("cairo")) {
		args$type <- "cairo"
	}
	do.call(grDevices::png, args)

	pushViewport(viewport(x = 0.25, width = 0.5, y = 0.75, height = 0.5))
	draw_piece_and_bleed("die_face", suit, 1, cfg)
	popViewport()
	pushViewport(viewport(x = 0.75, width = 0.5, y = 0.75, height = 0.5))
	draw_piece_and_bleed("die_face", suit, 2, cfg)
	popViewport()
	pushViewport(viewport(x = 0.25, width = 0.5, y = 0.25, height = 0.5))
	draw_piece_and_bleed("die_face", suit, 3, cfg)
	popViewport()
	pushViewport(viewport(x = 0.75, width = 0.5, y = 0.25, height = 0.5))
	draw_piece_and_bleed("die_face", suit, 4, cfg)
	popViewport()

	grDevices::dev.off()
	invisible(filename)
}

#' Draw board game pieces with ggplot2
#'
#' `geom_piece()` creates a `ggplot2` geom.
#' `aes_piece()` takes a data frame and generates
#' an appropriate `ggplot2::aes()` mapping.
#'
#' `geom_piece()` requires a fixed scale coordinate system with an aspect
#' ratio of 1 as provided by `ggplot2::coord_fixed()`.
#' `geom_piece()` also requires that `cfg` is a character vector (and not a `pp_cfg()` object).
#' In particular if using `op_transform()` one should set its argument `cfg_class = "character"`
#' if intending for use with `geom_piece()`.
#'
#' @section Aesthetics:
#' `geom_piece()` understands the following aesthetics (required aesthetics are in bold).
#' See [pieceGrob()] for more details.
#'
#' * **`x`**
#' * **`y`**
#' * `z`
#' * `piece_side`
#' * `rank`
#' * `suit`
#' * `cfg`
#' * `width`
#' * `height`
#' * `depth`
#' * `angle`
#' * `scale`
#' * `type`
#'
#' @inheritParams ggplot2::geom_rect
#' @inheritParams pieceGrob
#' @inheritParams render_piece
#' @param  ... Aesthetics, used to set an aesthetic to a fixed value.
#' @seealso `geom_piece()` is a wrapper around [pieceGrob()].
#'         [scale_x_piece()] and [scale_y_piece()] are wrappers
#'         around [ggplot2::scale_x_continuous()] and [ggplot2::scale_y_continuous()]
#'         with better defaults for board game diagrams.
#' @examples
#' if (require("ggplot2", quietly = TRUE) && require("tibble", quietly = TRUE)) {
#'   envir <- game_systems("sans")
#'   df_board <- tibble(piece_side = "board_face", suit = 3, rank = 8,
#'                  x = 4.5, y = 4.5)
#'   df_w <- tibble(piece_side = "bit_back", suit = 6, rank = 1,
#'                  x = rep(1:8, 2), y = rep(1:2, each=8))
#'   df_b <- tibble(piece_side = "bit_back", suit = 1, rank = 1,
#'                  x = rep(1:8, 2), y = rep(7:8, each=8))
#'   df <- rbind(df_board, df_w, df_b)
#'   # 2D example
#'   # `cfg` must be a character vector for `geom_piece()`
#'   ggplot(df, aes_piece(df)) +
#'       geom_piece(cfg = "checkers1", envir = envir) +
#'       coord_fixed() +
#'       scale_x_piece() +
#'       scale_y_piece() +
#'       theme_minimal(28) +
#'       theme(panel.grid = element_blank())
#' }
#' if (require("ggplot2", quietly = TRUE) && require("tibble", quietly = TRUE)) {
#'   # 3D "oblique" projection example
#'   # `cfg_class` must be "character" when using with `geom_piece()`
#'   df3d <- op_transform(df, cfg = "checkers1", envir = envir,
#'                        op_angle = 45, cfg_class = "character")
#'   ggplot(df3d, aes_piece(df3d)) +
#'       geom_piece(cfg = "checkers1", envir = envir,
#'                  op_angle = 45, op_scale = 0.5) +
#'       coord_fixed() +
#'       theme_void()
#' }
#' @export
geom_piece <- function(
	mapping = NULL,
	data = NULL,
	stat = "identity",
	position = "identity",
	...,
	envir = getOption("piecepackr.envir", piecepackr::game_systems()),
	op_scale = getOption("piecepackr.op_scale", 0),
	op_angle = getOption("piecepackr.op_angle", 45),
	inherit.aes = TRUE
) {
	assert_suggested("ggplot2")

	# assert cfg is character
	ggplot2::layer(
		data = data,
		mapping = mapping,
		stat = stat,
		geom = GeomPiece,
		position = position,
		show.legend = FALSE,
		inherit.aes = inherit.aes,
		params = list(
			envir = envir,
			op_scale = op_scale,
			op_angle = op_angle,
			...
		)
	)
}

#' ggplot2 game diagram scales
#'
#' `scale_x_piece()` and `scale_y_piece()` are wrappers
#' around [ggplot2::scale_x_continuous()] and
#' [ggplot2::scale_y_continuous()] with "better"
#' defaults for board game diagrams.
#' `label_letter()` labels breaks with letters
#' and `label_counting()` labels breaks with positive integers
#' to more easily generate (i.e. chess) algebraic notation coordinates.
#' `breaks_counting()` generates breaks of just the positive integers within the limits.
#' @param ... Passed to [ggplot2::scale_x_continuous()] or [ggplot2::scale_y_continuous()].
#' @inheritParams ggplot2::scale_x_continuous
#' @examples
#' if (require("ggplot2", quietly = TRUE) && require("tibble", quietly = TRUE)) {
#'   envir <- game_systems("sans")
#'   df_board <- tibble(piece_side = "board_face", suit = 3, rank = 8,
#'                  x = 4.5, y = 4.5)
#'   df_w <- tibble(piece_side = "bit_back", suit = 6, rank = 1,
#'                  x = rep(1:8, 2), y = rep(1:2, each=8))
#'   df_b <- tibble(piece_side = "bit_back", suit = 1, rank = 1,
#'                  x = rep(1:8, 2), y = rep(7:8, each=8))
#'   df <- rbind(df_board, df_w, df_b)
#'
#'   # `cfg` must be a character vector for `geom_piece()`
#'   ggplot(df, aes_piece(df)) +
#'       geom_piece(cfg = "checkers1", envir = envir) +
#'       coord_fixed() +
#'       scale_x_piece() +
#'       scale_y_piece() +
#'       theme_minimal(28) +
#'       theme(panel.grid = element_blank())
#' }
#' @return `scale_x_piece()` and `scale_y_piece()` return ggplot2 scale objects.
#'         `label_letter()` and `label_counting()` return functions suitable for use with the `labels` scale argument.
#'         `breaks_counting()` returns a function suitable for use with the `breaks` scale argument.
#' @rdname scale_piece
#' @export
scale_x_piece <- function(
	...,
	name = NULL,
	breaks = breaks_counting(),
	minor_breaks = NULL,
	labels = label_letter()
) {
	assert_suggested("ggplot2")
	ggplot2::scale_x_continuous(
		...,
		name = name,
		breaks = breaks,
		minor_breaks = minor_breaks,
		labels = labels
	)
}

#' @rdname scale_piece
#' @export
scale_y_piece <- function(
	...,
	name = NULL,
	breaks = breaks_counting(),
	minor_breaks = NULL,
	labels = label_counting()
) {
	assert_suggested("ggplot2")
	ggplot2::scale_y_continuous(
		...,
		name = name,
		breaks = breaks,
		minor_breaks = minor_breaks,
		labels = labels
	)
}

#' @rdname scale_piece
#' @export
label_letter <- function() {
	function(x) letters[seq_along(x)]
}

#' @rdname scale_piece
#' @export
label_counting <- function() {
	function(x) as.character(seq_along(x))
}

#' @rdname scale_piece
#' @export
breaks_counting <- function() {
	function(x) {
		seq.int(from = max(1L, ceiling(x[1])), to = floor(x[2]), by = 1L)
	}
}

# GeomPiece is defined in `.onLoad()` in `hooks.R` so {ggplot2} can be Suggests instead of Imports
# Because it is defined in `.onLoad()` {covr} can't see it even though it is implicitly tested
# in `geom_piece()` tests
create_GeomPiece <- function() {
	# nocov start
	ggplot2::ggproto(
		"GeomPiece",
		ggplot2::Geom,
		required_aes = c("x", "y"),
		default_aes = ggplot2::aes(
			piece_side = "tile_back",
			rank = 1L,
			suit = 1L,
			cfg = "piecepack",
			z = NA,
			width = NA,
			height = NA,
			depth = NA,
			angle = 0,
			scale = 1,
			type = "normal"
		),
		draw_key = function(self, data, params, size) {
			# data Single row data frame containing scaled aesthetics
			# params A list of additional parameters supplied to the geom
			# size Width and height of key in mm
			ggplot2::zeroGrob()
		},
		draw_panel = function(self, data, panel_params, coord, envir, op_scale, op_angle) {
			if (coord$is_free()) {
				abort("'geom_piece()' will not work correctly if not using a fixed scale.")
			}
			if (hasName(coord, "ratio") && coord$ratio != 1) {
				abort("'geom_piece()' will not work correctly if not using an aspect ratio of 1.")
			}
			data <- gg_impute_missing_data(data, envir, panel_params)
			coord <- coord$transform(data, panel_params)
			grob_piece(coord, panel_params, envir, op_scale, op_angle)
		},
		setup_data = function(data, params) {
			l <- aabb_piece(
				data,
				envir = params$envir,
				op_scale = params$op_scale,
				op_angle = params$op_angle
			)
			data$xmin <- l$x_op[1]
			data$xmax <- l$x_op[2]
			data$ymin <- l$y_op[1]
			data$ymax <- l$y_op[2]
			data
		}
	)
} # nocov end

#' @rdname geom_piece
#' @export
aes_piece <- function(df) {
	assert_suggested("ggplot2")
	stopifnot(hasName(df, "x"), hasName(df, "y"))
	mapping <- ggplot2::aes()
	aesthetics <- c(
		"x",
		"y",
		"z",
		"piece_side",
		"rank",
		"suit",
		"cfg",
		"width",
		"height",
		"depth",
		"angle",
		"scale",
		"type"
	)
	for (a in aesthetics) {
		if (hasName(df, a)) mapping[[a]] <- as.name(a)
	}
	mapping
}

grob_piece <- function(coord, panel_params, envir, op_scale, op_angle) {
	grid::gTree(
		coord = coord,
		panel_params = panel_params,
		envir = envir,
		op_scale = op_scale,
		op_angle = op_angle,
		cl = "geom_piece_grob"
	)
}

#' @import grid
#' @export
makeContent.geom_piece_grob <- function(x) {
	grob <- piecepackr::pmap_piece(
		x$df,
		piecepackr::pieceGrob,
		default.units = "inches",
		draw = FALSE,
		op_scale = x$op_scale,
		op_angle = x$op_angle,
		envir = x$envir,
		name = "geom_piece"
	)
	setChildren(x, gList(grob))
}

#' @export
makeContext.geom_piece_grob <- function(x) {
	x_width_native <- diff(x$panel_params$x$continuous_range)
	x_width_in <- grid::convertWidth(grid::unit(1, "npc"), "in", valueOnly = TRUE)
	gp_scale <- x_width_in / x_width_native
	x$gp <- grid::gpar(cex = gp_scale, lex = gp_scale)
	x$df <- gg_to_in(x$coord, gp_scale)
	x
}

# nolint start
# CoordPiece <- ggplot2::ggproto(
#   "CoordPiece",
#   ggplot2::CoordFixed
# )
# coord_piece <- function(xlim = NULL, ylim = NULL, expand = TRUE, clip = "on") {
#   ggproto(NULL, CoordPiece,
#     limits = list(x = xlim, y = ylim),
#     ratio = 1,
#     expand = expand,
#     clip = clip
#   )
# }
# coord_piece() algebraic?
# nolint end

gg_to_in <- function(df, scale) {
	df$x <- convertX(unit(df$x, "npc"), "in", valueOnly = TRUE)
	df$y <- convertY(unit(df$y, "npc"), "in", valueOnly = TRUE)
	df$z <- scale * df$z
	df$width <- scale * df$width
	df$height <- scale * df$height
	df$depth <- scale * df$depth
	df
}

gg_impute_missing_data <- function(data, envir, panel_params) {
	# impute missing width, height, depth, and z
	i_w <- which(is.na(data$width))
	data$width[i_w] <- purrr::pmap_dbl(data[i_w, ], gg_get_width, envir = envir)
	i_h <- which(is.na(data$height))
	data$height[i_h] <- purrr::pmap_dbl(data[i_h, ], gg_get_height, envir = envir)
	i_d <- which(is.na(data$depth))
	data$depth[i_d] <- purrr::pmap_dbl(data[i_d, ], gg_get_depth, envir = envir)
	i_z <- which(is.na(data$z))
	data$z[i_z] <- 0.5 * data$depth[i_d]
	data
}
gg_get_width <- function(piece_side, suit, rank, cfg, ..., envir) {
	envir[[cfg]]$get_width(piece_side, suit, rank)
}
gg_get_height <- function(piece_side, suit, rank, cfg, ..., envir) {
	envir[[cfg]]$get_height(piece_side, suit, rank)
}
gg_get_depth <- function(piece_side, suit, rank, cfg, ..., envir) {
	envir[[cfg]]$get_depth(piece_side, suit, rank)
}

test_that("geom_piece() works as expected", {
	skip_if_not_installed("vdiffr")
	library("vdiffr")

	skip_if_not_installed("ggplot2")
	library("ggplot2")
	library("tibble")

	envir <- piecepackr::game_systems("sans")

	df_board <- tibble(piece_side = "board_face", suit = 3, rank = 8, x = 4.5, y = 4.5)
	df_w <- tibble(
		piece_side = "bit_back",
		suit = 6,
		rank = 1,
		x = rep(1:8, 2),
		y = rep(1:2, each = 8)
	)
	df_b <- tibble(
		piece_side = "bit_back",
		suit = 1,
		rank = 1,
		x = rep(1:8, 2),
		y = rep(7:8, each = 8)
	)
	df <- rbind(df_board, df_w, df_b)

	gg <- ggplot(df, aes_piece(df)) +
		geom_piece(cfg = "checkers1", envir = envir) +
		coord_fixed() +
		scale_x_piece() +
		scale_y_piece() +
		theme_minimal(32) +
		theme(panel.grid = element_blank())

	expect_doppelganger("geom_piece", gg)
})

LETTER_WIDTH <- 8.5
LETTER_HEIGHT <- 11
A4_WIDTH <- 8.27
A4_HEIGHT <- 11.69

#' Save piecepack print-and-play (PnP) file
#'
#' Save piecepack print-and-play (PnP) file
#'
#' @param cfg Piecepack configuration list or `pp_cfg` object
#' @param output_filename Filename for print-and-play file
#' @param size PnP output size (currently supports either "letter", "A4", or "A5").
#'             This is the targeted \dQuote{trim} size of the print-and-play file
#'             (`size_bleed` can be used to make the print-and-play file larger than this).
#'             "A5" is in \dQuote{portrait} mode whereas the other sizes are in \dQuote{landscape} mode.
#' @param pieces Character vector of desired PnP pieces.
#'        Supports "piecepack", "matchsticks", "pyramids", "subpack", or "all".
#'        If `NULL` and combination of `size` / `bleed` values supports "matchsticks" and "pyramids"
#'        then defaults to `c("piecepack", "pyramids", "matchsticks")` else just "piecepack".
#' @param arrangement Either "single-sided" or "double-sided".
#' @param quietly Whether to hide messages about missing metadata
#'                in the provided configuration.
#' @param ... Currently ignored.
#' @param bleed If `TRUE` produce a variant print-and-play file with "bleed" zones
#'              and "crop marks" around game pieces.
#'              Currently only supports `pieces = "piecepack"`.
#' @param size_bleed A list with names "top", "right", "bottom", "left"
#'                   containing numeric values indicating the inches "bleed" to add to
#'                   the `size` of the print-and-play layout.
#'                   The default `NULL` means no such bleed added to "letter", "A4", and "A5"
#'                   layouts.
#'                   NB. multiply millimeters by `0.0393700787` to convert to inches.
#'                   We currently don't support an asymmetric left/right bleed combined with
#'                   `arrangement = "double-sided"`.
#' @inheritParams render_piece
#' @examples
#' \donttest{# May take more than 5 seconds on CRAN servers
#' if (capabilities("cairo")) {
#'   cfg <- pp_cfg(list(invert_colors.suited=TRUE))
#'   cfg$description <- 'Piecepack with an "inverted" color scheme.'
#'   cfg$title <- '"Inverted" piecepack'
#'   cfg$copyright <- "\u00a9 2022 Trevor L Davis.  Some Right Reserved."
#'   cfg$spdx_id <- "CC-BY-4.0"
#'   cfg$credit <- ""
#'
#'   file <- tempfile("my_pnp_file", fileext = ".pdf")
#'   file_ds <- tempfile("my_pnp_file_ds", fileext = ".pdf")
#'   file_a4 <- tempfile("my_pnp_file_a4", fileext = ".pdf")
#'   file_a5 <- tempfile("my_pnp_file_a5", fileext = ".pdf")
#'
#'   save_print_and_play(cfg, file)
#'   save_print_and_play(cfg, file_ds, arrangement="double-sided")
#'   save_print_and_play(cfg, file_a4, size="A4", pieces="all")
#'   save_print_and_play(cfg, file_a5, size="A5")
#' }
#' }
#' @export
save_print_and_play <- function(
	cfg = getOption("piecepackr.cfg", pp_cfg()),
	output_filename = "piecepack.pdf",
	size = c("letter", "A4", "A5", "4x6"),
	pieces = NULL,
	arrangement = c("single-sided", "double-sided"),
	dev = NULL,
	dev.args = list(family = cfg$fontfamily, onefile = TRUE, units = "in", bg = "white", res = 300),
	quietly = FALSE,
	...,
	bleed = FALSE,
	size_bleed = NULL
) {
	opt <- options(piecepackr.op_scale = 0)
	on.exit(options(opt), add = TRUE)

	stopifnot(is.null(dev) || is.function(dev))
	size <- match.arg(size)
	if (size == "4x6") {
		warn(
			'`size = "4x6"` is deprecated.',
			class = "deprecatedWarning"
		)
	}
	arrangement <- match.arg(arrangement)
	if (is.null(pieces)) {
		if (size == "4x6" || bleed) {
			pieces <- "piecepack"
		} else {
			pieces <- c("piecepack", "pyramids", "matchsticks")
		}
	}
	if ("all" %in% pieces) {
		pieces <- c("piecepack", "pyramids", "matchsticks", "subpack")
	}
	if (is.null(size_bleed)) {
		if (size == "4x6") {
			size_bleed <- list(top = 1 / 16, right = 3 / 32, bottom = 1 / 16, left = 3 / 32)
		} else {
			size_bleed <- list(top = 0, right = 0, bottom = 0, left = 0)
		}
	}
	stopifnot(
		all(c("top", "right", "bottom", "left") %in% names(size_bleed)),
		all(sapply(size_bleed, is.numeric)),
		arrangement == "single-sided" || size_bleed$left == size_bleed$right
	)

	cfg <- as_pp_cfg(cfg)
	current_dev <- grDevices::dev.cur()
	if (current_dev > 1) {
		on.exit(grDevices::dev.set(current_dev), add = TRUE)
	}
	height <- switch(size, letter = LETTER_WIDTH, A4 = A4_WIDTH, A5 = A4_WIDTH, `4x6` = 4) +
		size_bleed$top +
		size_bleed$bottom
	width <- switch(size, letter = LETTER_HEIGHT, A4 = A4_HEIGHT, A5 = A4_HEIGHT / 2, `4x6` = 6) +
		size_bleed$right +
		size_bleed$left

	if (is.null(dev)) {
		dev <- pp_device_fn(output_filename)
	}

	args <- list(filename = output_filename, width = width, height = height)
	args <- c(args, dev.args)
	args <- args[names(args) %in% names(formals(dev))]
	onefile <- has_onefile(dev, args)
	if (!onefile && !has_c_integer_format(output_filename)) {
		abort(paste(
			"`save_print_and_play()` generates multiple pages.",
			'Either `dev` needs to support `onefile` and `isTRUE(dev.args[["onefile"]])`',
			"or else `output_filename` needs a C integer format in the string."
		))
	}
	do.call(dev, args)

	pl <- switch(
		size,
		`4x6` = print_and_play_4x6(cfg, pieces, quietly, bleed, size_bleed),
		print_and_play_paper(cfg, size, pieces, arrangement, quietly, bleed, size_bleed)
	)

	invisible(grDevices::dev.off())

	if (onefile && tools::file_ext(output_filename) == "pdf") {
		if (requireNamespace("xmpdf", quietly = TRUE)) {
			add_pdf_metadata(output_filename, cfg, pl)
		} else if (!isFALSE(getOption("piecepackr.metadata.inform"))) {
			msg <- c(
				x = "Need the {xmpdf} package to embed pdf metadata",
				i = '`install.packages("xmpdf")`',
				i = "These messages can be disabled via `options(piecepackr.metadata.inform = FALSE)`."
			)
			inform(msg, class = "piecepackr_embed_metadata")
		}
	}
	invisible(NULL)
}

has_onefile <- function(dev, args) {
	hasName(formals(dev), "onefile") && isTRUE(args[["onefile"]])
}
has_c_integer_format <- function(filename) {
	grepl("%[[:digit:]]+d", filename)
}

add_pdf_metadata <- function(output_filename, cfg = pp_cfg(), pl = list()) {
	assert_suggested("xmpdf")

	if (xmpdf::supports_set_docinfo()) {
		docinfo <- xmpdf::docinfo(
			title = cfg$title,
			creator = paste0("piecepackr v", packageDescription("piecepackr")$Version),
			subject = paste(cfg$description, collapse = "\n"),
			keywords = "piecepack"
		)
		xmpdf::set_docinfo(docinfo, output_filename)
	} else if (!isFALSE(getOption("piecepackr.metadata.inform"))) {
		msg <- c(
			x = "Unable to embed pdf documentation info metadata",
			xmpdf::enable_feature_message("set_docinfo"),
			i = "These messages can be disabled via `options(piecepackr.metadata.inform = FALSE)`."
		)
		inform(msg, class = "piecepackr_embed_metadata")
	}

	if (xmpdf::supports_set_bookmarks()) {
		starting_pages <- head(c(1, 1 + cumsum(sapply(pl, identity))), -1)
		bookmarks <- data.frame(title = names(pl), page = starting_pages, stringsAsFactors = FALSE)
		xmpdf::set_bookmarks(bookmarks, output_filename)
	} else if (!isFALSE(getOption("piecepackr.metadata.inform"))) {
		msg <- c(
			x = "Unable to embed pdf bookmarks metadata",
			xmpdf::enable_feature_message("set_bookmarks"),
			i = "These messages can be disabled via `options(piecepackr.metadata.inform = FALSE)`."
		)
		inform(msg, class = "piecepackr_embed_metadata")
	}

	if (xmpdf::supports_set_xmp()) {
		x <- xmpdf::xmp(
			title = cfg$title,
			creator_tool = paste0("piecepackr v", packageDescription("piecepackr")$Version),
			description = paste(cfg$description, collapse = "\n"),
			keywords = "piecepack",
			subject = "piecepack",
			rights = cfg$copyright,
			spdx_id = cfg$spdx_id
		)
		xmpdf::set_xmp(x, output_filename)
	} else if (!isFALSE(getOption("piecepackr.metadata.inform"))) {
		msg <- c(
			x = "Unable to embed pdf XMP metadata",
			xmpdf::enable_feature_message("set_xmp"),
			i = "These messages can be disabled via `options(piecepackr.metadata.inform = FALSE)`."
		)
		inform(msg, class = "piecepackr_embed_metadata")
	}
}

print_and_play_paper <- function(cfg, size, pieces, arrangement, quietly, bleed, size_bleed) {
	if (bleed) {
		print_and_play_paper_bleed(cfg, size, pieces, arrangement, quietly, size_bleed)
	} else {
		print_and_play_paper_compact(cfg, size, pieces, arrangement, quietly, size_bleed)
	}
}

basicGrobWithBleed <- function(piece_side, suit, rank, cfg = pp_cfg()) {
	cfg <- as_pp_cfg(cfg)
	opt <- cfg$get_piece_opt(piece_side, suit, rank)
	grob <- cfg$get_grob(piece_side, suit, rank)
	gTree(
		opt = opt,
		grob = grob,
		bleed = inch(0.125),
		scale = 1,
		name = NULL,
		gp = gpar(),
		vp = NULL,
		cl = "basic_grob_with_bleed"
	)
}

#' @export
makeContent.basic_grob_with_bleed <- function(x) {
	stopifnot(all(hasName(x$grob, c("name", "vp"))))
	opt <- x$opt
	shape <- pp_shape(
		opt$shape,
		opt$shape_t,
		opt$shape_r,
		opt$back,
		width = opt$shape_w,
		height = opt$shape_h
	)

	width <- convertWidth(unit(1, "npc"), "in") - 2 * x$bleed
	height <- convertHeight(unit(1, "npc"), "in") - 2 * x$bleed

	cvp <- viewport(width = width, height = height)
	grob <- grid::editGrob(x$grob, vp = cvp, name = "piece")
	if (hasName(grob, "scale")) {
		grob$scale <- x$scale
	}

	width <- convertWidth(width, "in", valueOnly = TRUE)
	height <- convertHeight(height, "in", valueOnly = TRUE)
	bleed <- convertWidth(x$bleed, "in", valueOnly = TRUE)

	# if easy extend mats and/or gridlines into bleed zone...
	if (
		inherits(grob, "basic_piece_side") &&
			(is_color_invisible(opt$border_color) || nigh(opt$border_lex, 0)) &&
			(!is_color_invisible(opt$mat_color) && !nigh(opt$mat_width, 0)) &&
			opt$shape == "rect"
	) {
		bleed <- rect_mat_bleed(opt, width, height, bleed)
	} else if (
		inherits(grob, "basic_piece_side") &&
			(is_color_invisible(opt$border_color) || nigh(opt$border_lex, 0)) &&
			(is_color_invisible(opt$mat_color) || nigh(opt$mat_width, 0)) &&
			(!is_color_invisible(opt$gridline_color) && !nigh(opt$gridline_lex, 0)) &&
			(opt$shape %in% c("rect", "roundrect"))
	) {
		bleed <- rect_gl_bleed(opt)
	} else {
		gp_bleed <- gpar(col = NA, lwd = 0, fill = opt$bleed_color)
		bleed <- rectGrob(gp = gp_bleed, name = "bleed")
	}

	gl <- gList(bleed, grob)

	setChildren(x, gl)
}

rect_mat_bleed <- function(opt, width, height, bleed) {
	gp_bg <- gpar(col = NA, lwd = 0, fill = opt$background_color)
	gp_mat <- gpar(col = NA, lwd = 0, fill = opt$mat_color)

	mat_width <- rep(opt$mat_width, length.out = 4)
	if (!nigh(mat_width[1], 0)) {
		mat_width[1] <- (bleed + mat_width[1] * height) / (height + 2 * bleed)
	}
	if (!nigh(mat_width[2], 0)) {
		mat_width[2] <- (bleed + mat_width[2] * width) / (width + 2 * bleed)
	}
	if (!nigh(mat_width[3], 0)) {
		mat_width[3] <- (bleed + mat_width[3] * height) / (height + 2 * bleed)
	}
	if (!nigh(mat_width[4], 0)) {
		mat_width[4] <- (bleed + mat_width[4] * width) / (width + 2 * bleed)
	}
	mat <- rectMatGrobFn(mat_width)(gp = gp_mat)
	t <- 1 - mat_width[1]
	r <- 1 - mat_width[2]
	b <- mat_width[3]
	l <- mat_width[4]
	bg <- polygonGrob(x = c(l, l, r, r), y = c(t, b, b, t), name = "background", gp = gp_bg)

	gList(bg, mat)
}

rect_gl_bleed <- function(opt) {
	gp_bg <- gpar(col = NA, lwd = 0, fill = opt$background_color)
	gp_gl <- gpar(col = opt$gridline_color, lwd = 8 * opt$gridline_lex, lineend = "butt")
	bg <- rectGrob(gp = gp_bg, name = "background")
	gl_v <- linesGrob(x = 0.5, name = "line1", gp = gp_gl)
	gl_h <- linesGrob(y = 0.5, name = "line2", gp = gp_gl)

	gList(bg, gl_v, gl_h)
}

# We use `<<-` below to modify the package's namespace.
# We don't modify the global environment.
# This is so {ggplot2} can be Suggests instead of Imports
# Define object at build time
GeomPiece <- NULL

.onLoad <- function(libname, pkgname) {
	if (requireNamespace("ggplot2", quietly = TRUE)) {
		# Modify object at load time
		GeomPiece <<- create_GeomPiece()
	}
}

# `save_print_and_play(size = "4x6")` is deprecated

    Code
      save_print_and_play(pp_cfg(), f, size = "4x6", pieces = "piecepack", quietly = TRUE)
    Condition
      Warning:
      `size = "4x6"` is deprecated.

# deprecated 'preview_layout' component warning

    Code
      invisible(pieceGrob("preview_layout", cfg = cfg_default, default.units = "npc"))
    Condition
      Warning:
      The "preview_layout" component is deprecated. Use `ppdf::piecepack_preview() |> pmap_piece(cfg = cfg, default.units = "in")` instead.

# removed `style` argument errors

    Code
      game_systems(style = "dejavu3d")
    Condition
      Error in `game_systems()`:
      ! `...` must be empty.
      x Problematic argument:
      * style = "dejavu3d"

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/animate_piece.R
\name{animate_piece}
\alias{animate_piece}
\title{Animate board game pieces}
\usage{
animate_piece(
  dfs,
  file = "animation.gif",
  ...,
  .f = piecepackr::grid.piece,
  cfg = getOption("piecepackr.cfg", NULL),
  envir = getOption("piecepackr.envir", game_systems("sans")),
  width = NULL,
  height = NULL,
  ppi = NULL,
  annotate = TRUE,
  annotation_scale = NULL,
  open_device = TRUE,
  n_transitions = 0L,
  n_pauses = 1L,
  fps = n_transitions + n_pauses,
  xbreaks = NULL,
  ybreaks = NULL
)
}
\arguments{
\item{dfs}{A list of data frames of game data to plot.}

\item{file}{Filename to save animation unless \code{NULL}
in which case it uses the current graphics device.}

\item{...}{Arguments to \code{\link[=pmap_piece]{pmap_piece()}}}

\item{.f}{Low level graphics function to use e.g. \code{\link[=grid.piece]{grid.piece()}}, \code{\link[=piece3d]{piece3d()}}, \code{\link[=piece]{piece()}}, or \code{\link[=piece_mesh]{piece_mesh()}}.}

\item{cfg}{A piecepackr configuration list}

\item{envir}{Environment (or named list) of piecepackr configuration lists}

\item{width}{Width of animation (in inches).  Inferred by default.}

\item{height}{Height of animation (in inches).  Inferred by default.}

\item{ppi}{Resolution of animation in pixels per inch.
By default set so image max 600 pixels wide or tall.}

\item{annotate}{If \code{TRUE} or \code{"algebraic"} annotate the plot
with \dQuote{algrebraic} coordinates,
if \code{FALSE} or \code{"none"} don't annotate,
if \code{"cartesian"} annotate the plot with \dQuote{cartesian} coordinates.}

\item{annotation_scale}{Multiplicative factor that scales (stretches) any annotation coordinates.
By default uses \code{attr(df, "scale_factor") \%||\% 1}.}

\item{open_device}{If \code{TRUE} open a new graphics device otherwise draw in the active graphics.}

\item{n_transitions}{Integer, if over zero (the default)
how many transition frames to add between moves.}

\item{n_pauses}{Integer, how many paused frames per completed move.}

\item{fps}{Double, frames per second.}

\item{xbreaks, ybreaks}{Subset (of integers) to provide axis labels for if \code{annotate} is \code{TRUE}.
If \code{NULL} infer a reasonable choice.}
}
\value{
Nothing, as a side effect creates an animation.
}
\description{
\code{animate_piece()} animates board game pieces.
}
\examples{
  # Basic tic-tac-toe animation
  dfs <- list()
  d.frame <- function(piece_side = "bit_back", ..., rank = 1L) {
                 data.frame(piece_side = piece_side, ..., rank = rank,
                            cfg = "checkers1", stringsAsFactors = FALSE)
  }
  df <- d.frame("board_back", suit = 2L, rank = 3L, x = 2, y = 2, id = "1")
  dfs[[1L]] <- df
  df <- rbind(df, d.frame(suit = 1L, x = 2, y = 2, id = "2"))
  dfs[[2L]] <- df
  df <- rbind(df, d.frame(suit = 2L, x = 1, y = 2, id = "3"))
  dfs[[3L]] <- df
  df <- rbind(df, d.frame(suit = 1L, x = 3, y = 1, id = "4"))
  dfs[[4L]] <- df
  df <- rbind(df, d.frame(suit = 2L, x = 1, y = 3, id = "5"))
  dfs[[5L]] <- df
  df <- rbind(df, d.frame(suit = 1L, x = 1, y = 1, id = "6"))
  dfs[[6L]] <- df
  df <- rbind(df, d.frame(suit = 2L, x = 3, y = 3, id = "7"))
  dfs[[7L]] <- df
  df <- rbind(df, d.frame(suit = 1L, x = 2, y = 1, id = "8"))
  dfs[[8L]] <- df

  ## Press enter to walk through moves in a "game" in new graphics device
  if (interactive()) {
      animate_piece(dfs, file = NULL)
  }

  ## Save GIF of game with animation transitions
  \dontrun{# May take more than 5 seconds on CRAN servers
  if ((requireNamespace("animation", quietly = TRUE) ||
       requireNamespace("gifski", quietly = TRUE)) &&
      requireNamespace("tweenr", quietly = TRUE)) {
      file <- tempfile("tic-tac-toe", fileext = ".gif")
      animate_piece(dfs, file = file,
                    n_transitions = 5L, n_pauses = 2L, fps = 9)
  }
  }

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_piece.R
\name{render_piece}
\alias{render_piece}
\title{Render image of game pieces}
\usage{
render_piece(
  df,
  file = NULL,
  ...,
  .f = piecepackr::grid.piece,
  cfg = getOption("piecepackr.cfg", NULL),
  envir = getOption("piecepackr.envir", game_systems("sans")),
  width = NULL,
  height = NULL,
  ppi = 72,
  bg = "white",
  xoffset = NULL,
  yoffset = NULL,
  annotate = FALSE,
  annotation_scale = NULL,
  dev = NULL,
  dev.args = list(res = ppi, bg = bg, units = "in"),
  open_device = TRUE,
  close_device = open_device && (!is.null(file) || !is.null(dev)),
  image = c("NULL", "raster", "nativeRaster"),
  xbreaks = NULL,
  ybreaks = NULL
)
}
\arguments{
\item{df}{A data frame of game piece information with (at least) the
named columns \dQuote{piece_side}, \dQuote{x}, and \dQuote{y}.}

\item{file}{Filename to save image to unless \code{NULL}
in which case it either uses the current graphics device or opens a new device
(depending on \code{open_device} argument).}

\item{...}{Arguments to \code{\link[=pmap_piece]{pmap_piece()}}}

\item{.f}{Low level graphics function to use e.g. \code{\link[=grid.piece]{grid.piece()}}, \code{\link[=piece3d]{piece3d()}}, \code{\link[=piece_mesh]{piece_mesh()}}, or \code{\link[=piece]{piece()}}.}

\item{cfg}{A piecepackr configuration list}

\item{envir}{Environment (or named list) of piecepackr configuration lists}

\item{width}{Width of image (in inches).  Inferred by default.}

\item{height}{Height of image (in inches).  Inferred by default.}

\item{ppi}{Resolution of image in pixels per inch.}

\item{bg}{Background color (use \code{"transparent"} for transparent)}

\item{xoffset}{Number to add to the \code{x} column in \code{df}.  Inferred by default.}

\item{yoffset}{Number to add to the \code{y} column in \code{df}.  Inferred by default.}

\item{annotate}{If \code{TRUE} or \code{"algebraic"} annotate the plot
with \dQuote{algrebraic} coordinates,
if \code{FALSE} or \code{"none"} don't annotate,
if \code{"cartesian"} annotate the plot with \dQuote{cartesian} coordinates.}

\item{annotation_scale}{Multiplicative factor that scales (stretches) any annotation coordinates.
By default uses \code{attr(df, "scale_factor") \%||\% 1}.}

\item{dev}{Graphics device function to use if \code{open_device} is \code{FALSE}.
If \code{NULL} infer a reasonable choice from \code{file}.}

\item{dev.args}{Additional arguments to pass to \code{dev} (besides \code{filename}, \code{width}, and \code{height}).
Will filter out any names that aren't in \code{formals(dev)}.}

\item{open_device}{If \code{TRUE} open a new graphics device otherwise draw in the active graphics.}

\item{close_device}{If \code{TRUE} close the graphics device (if \code{open_device = TRUE} close the newly opened device otherwise close the previously existing graphics device).}

\item{image}{Class of image object to return in the \code{"image"} field of returned list.
If \code{"NULL"} (the default) return \code{NULL}.
If \code{"raster"} or \verb{"nativeRaster" try to return a raster object of the image using [grDevices::dev.capture()] or }as.raster(magick::image_read())`.}

\item{xbreaks, ybreaks}{Subset (of integers) to provide axis labels for if \code{annotate} is \code{TRUE}.
If \code{NULL} infer a reasonable choice.}
}
\value{
An invisible list of the dimensions of the image and
possibly an image object specified by \code{image}.
As a side effect may save a file and/or open/close a graphics device.
}
\description{
\code{render_piece()} renders an image of game pieces to a file or graphics device.
It is a wrapper around \code{pmap_piece()} that can auto-size files and graphic devices,
apply axes offsets, annotate coordinates, and set up \code{rayrender} / \code{rayvertex} scenes.
}
\examples{
 df_board <- data.frame(piece_side = "board_face", suit = 3, rank = 5,
                        x = 3.0, y = 3.0, stringsAsFactors = FALSE)
 df_w <- data.frame(piece_side = "bit_back", suit = 6, rank = 1,
                    x = rep(1:5, 2), y = rep(1:2, each=5),
                    stringsAsFactors = FALSE)
 df_b <- data.frame(piece_side = "bit_back", suit = 1, rank = 1,
                    x = rep(1:5, 2), y = rep(4:5, each=5),
                    stringsAsFactors = FALSE)
 df <- rbind(df_board, df_w, df_b)
 df$cfg <- "checkers1"

 if (requireNamespace("grid", quietly = TRUE)) {
   render_piece(df, open_device = FALSE)
 }
 if (requireNamespace("grid", quietly = TRUE)) {
   grid::grid.newpage()
   render_piece(df, open_device = FALSE,
                op_scale = 0.5, trans = op_transform,
                annotate = "algrebraic")
 }
 \dontrun{# May take more than 5 seconds on CRAN servers
 if (require(rayvertex)) {
   envir3d <- game_systems("sans", border = FALSE)
   render_piece(df, .f = piece_mesh, envir = envir3d,
                open_device = FALSE,
                op_scale = 0.5, trans = op_transform)
 }
 }
}
\seealso{
This function is a wrapper around \code{\link[=pmap_piece]{pmap_piece()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/game_systems.R
\name{game_systems}
\alias{game_systems}
\alias{to_hexpack}
\alias{to_subpack}
\title{Standard game systems}
\usage{
game_systems(
  font = NULL,
  round = FALSE,
  pawn = "token",
  ...,
  shading = FALSE,
  border = TRUE,
  background_color = ifelse(border, "white", "burlywood"),
  edge_color = ifelse(border, "white", "black")
)

to_hexpack(cfg = getOption("piecepackr.cfg", pp_cfg()))

to_subpack(cfg = getOption("piecepackr.cfg", pp_cfg()))
}
\arguments{
\item{font}{If \code{NULL} (the default) uses suit glyphs from the default \dQuote{sans} font.
If \code{"dejavu"} it will use suit glyphs from the "DejaVu Sans" font
(must be installed on the system).}

\item{round}{If \code{TRUE} the \dQuote{shape} of \dQuote{tiles} and \dQuote{cards} (and some boards)
will be \dQuote{roundrect} instead of \dQuote{rect} (the default).}

\item{pawn}{If \code{"token"} (default) the piecepack pawn will be a two-sided token in a \dQuote{halma} outline,
if \code{"peg-doll"} the piecepack pawn will be a \dQuote{peg doll} style pawn, and
if \code{"joystick"} the piecepack pawn will be a \dQuote{joystick} style pawn.
Note for the latter two pawn styles only \code{pawn_top} will work with \code{\link[=grid.piece]{grid.piece()}}.}

\item{...}{Must be empty}

\item{shading}{If \code{TRUE} add a shading effect to marbles and stones when drawn with \code{\link[=grid.piece]{grid.piece()}}.}

\item{border}{If \code{TRUE} draw a black border line on piece edges.  Should normally be \code{TRUE} when
drawing with \code{{grid}} graphics and \code{FALSE} when drawing with 3D graphic systems.}

\item{background_color}{Background color to use for certain pieces like boards and piecepack tiles.}

\item{edge_color}{Edge color to use for certain pieces like boards and piecepack tiles.}

\item{cfg}{List of configuration options}
}
\description{
\code{game_systems()} returns a list of \code{\link[=pp_cfg]{pp_cfg()}} objects
representing several game systems and pieces.
\code{to_subpack()} and \code{to_hexpack()} will attempt to generate matching (piecepack stackpack)
subpack and (piecepack) hexpack \code{\link[=pp_cfg]{pp_cfg()}} R6 objects respectively given a piecepack configuration.
}
\details{
Contains the following game systems:\describe{
\item{alquerque}{Boards and pieces in nine color schemes for Alquerque}
\item{checkers1, checkers2}{Checkers and checkered boards in nine color schemes.
Checkers are represented by a piecepackr \dQuote{bit}.
The \dQuote{board} \dQuote{face} is a checkered board and the \dQuote{back} is a lined board.
Color is controlled by suit and number of rows/columns by rank.
\code{checkers1} has one inch squares and \code{checkers2} has two inch squares.}
\item{chess1, chess2}{Chess pieces, boards, and dice in nine color schemes.
Chess pieces are represented by a \dQuote{bit} (face).
The \dQuote{board} \dQuote{face} is a checkered board and the \dQuote{back} is a lined board.
Color is controlled by suit and number of rows/columns by rank.
\code{chess1} has one inch squares and \code{chess2} has two inch squares.
Currently uses print-and-play style discs instead of 3D Staunton chess pieces.}
\item{cubes}{Cubes in six sizes and nine colors represented by a \dQuote{bit}.
Color is controlled by suit and size of cube by rank.
Sizes go from 8mm (rank 1) to 25mm (rank 6): 8mm, 10mm, 12mm, 15mm, 20mm, 25mm.}
\item{dice}{Traditional six-sided pipped dice in nine color schemes (color controlled by their suit).}
\item{dice_d4, dice_numeral, dice_d8, dice_d10, dice_d10_percentile, dice_d12, dice_d20}{
Polyhedral dice most commonly used to play wargames, roleplaying games, and trading card games:\describe{
\item{dice_d4}{Four-sided dice in nine color schemes (color controlled by their suit).
Tetrahedrons with the rank as a numeral at the top point.}
\item{dice_numeral}{Six-sided dice with numerals instead of pips in nine color schemes (color controlled by their suit).}
\item{dice_d8}{Eight-sided dice in nine color schemes (color controlled by their suit).
Octahedrons with the rank as a numeral at the top face.}
\item{dice_d10}{Ten-sided dice in nine color schemes (color controlled by their suit).
Pentagonal trapezohedrons with the rank as a numeral at the top face.
The rank of ten is represented by a zero.}
\item{dice_d10_percentile}{Ten-sided dice in nine color schemes (color controlled by their suit).
Pentagonal trapezohedrons with the rank as a numeral followed by a zero at the top face.
The rank of ten is represented by a zero.}
\item{dice_d12}{Twelve-sided dice in nine color schemes (color controlled by their suit).
Dodecahedrons with the rank as a numeral at the top face.}
\item{dice_d20}{Twenty-sided dice in nine color schemes (color controlled by their suit).
Icosahedrons with the rank as a numeral at the top face.}}}
\item{dice_fudge}{\dQuote{Fudge} dice in nine color schemes (color controlled by their suit).
\dQuote{Fudge} dice have three ranks "+", " ", and "-" repeated twice.}
\item{dominoes, dominoes_black, dominoes_blue, dominoes_green, dominoes_red, dominoes_white, dominoes_yellow}{
Traditional pipped dominoes in six color schemes (\code{dominoes} and \code{dominoes_white} are the same).
In each color scheme the number of pips on the \dQuote{top} of the domino is
controlled by their \dQuote{rank} and on the \dQuote{bottom} by their \dQuote{suit}.
Supports up to double-18 sets.}
\item{dominoes_chinese, dominoes_chinese_black}{\code{dominoes_chinese} has Asian-style six-sided pipped dice with
white background and black and red pips.
The \dQuote{tile}'s are Chinese dominoes (1" x 2.5") whose number of pips are controlled
by both their \dQuote{rank} and their \dQuote{suit}. \code{dominoes_chinese_black} are like \code{dominoes_chinese} but the
dice and dominoes have a black background and white and red pips.}
\item{go}{Go stones and lined boards in nine color schemes.
Go stones are represented by a \dQuote{bit} and the board is a \dQuote{board}.
Color is controlled by suit and number of rows/columns by rank.}
\item{marbles}{Marbles in nine sizes and nine colors represented by a \dQuote{bit}.
Color is controlled by suit and size of marble by rank.
Sizes go from 1/2" (rank 1) to 1" (rank 9) in 1/16" increments.
There are also square holed boards spaced for 1" marbles: their color
is controlled by suit and number of holes per row/column by rank.}
\item{meeples}{Standard 16mm x 16mm x 10mm \dQuote{meeples} in nine colors represented by a \dQuote{bit}.}
\item{morris}{Various morris aka mills aka merels games in nine colors.
Color is controlled by suit and \dQuote{size} of morris board
is controlled by rank e.g. \dQuote{Six men's morris} corresponds to a rank of 6 and
\dQuote{Nine men's morris} corresponds to a rank of 9.
Game pieces are represented by stones.}
\item{piecepack, dual_piecepacks_expansion, playing_cards_expansion, hexpack, subpack, piecepack_inverted}{
The piecepack is a public domain game system invented by James "Kyle" Droscha.
See \url{https://www.ludism.org/ppwiki} for more info about the piecepack and its accessories/expansions.
\describe{
\item{piecepack}{A standard piecepack.  The configuration also contains the following piecepack accessories:\describe{
\item{piecepack dice cards}{An accessory proposed by John Braley.
See \url{https://www.ludism.org/ppwiki/PiecepackDiceCards}.}
\item{piecepack matchsticks}{A public domain accessory developed by Dan Burkey.
See \url{https://www.ludism.org/ppwiki/PiecepackMatchsticks}.}
\item{piecepack pyramids}{A public domain accessory developed by Tim Schutz.
See \url{https://www.ludism.org/ppwiki/PiecepackPyramids}.}
\item{piecepack saucers}{A public domain accessory developed by Karol M. Boyle at Mesomorph Games.
See \url{https://web.archive.org/web/20190719155827/http://www.piecepack.org/Accessories.html}.}}}
\item{piecepack_inverted}{The standard piecepack with its color scheme inverted.
Intended to aid in highlighting special pieces in diagrams.}
\item{dual_piecepacks_expansion}{A companion piecepack with a special suit scheme.
See \url{https://trevorldavis.com/piecepackr/dual-piecepacks-pnp.html}.}
\item{playing_cards_expansion}{A piecepack with the standard \dQuote{French} playing card suits.
See \url{https://www.ludism.org/ppwiki/PlayingCardsExpansion}.}
\item{hexpack}{A hexagonal extrapolation of the piecepack designed by Nathan Morse and Daniel Wilcox.
See \url{https://boardgamegeek.com/boardgameexpansion/35424/hexpack}.}
\item{subpack}{A mini piecepack.  Designed to be used with the \code{piecepack} to make piecepack
\dQuote{stackpack} diagrams.  See \url{https://www.ludism.org/ppwiki/StackPack}.}
}}
\item{playing_cards, playing_cards_colored, playing_cards_tarot}{
Poker-sized \dQuote{card} components for various playing card decks:\describe{
\item{playing_cards}{A traditional deck of playing cards with 4 suits
and 13 ranks (A, 2-10, J, Q, K) plus a 14th \dQuote{Joker} rank.}
\item{playing_cards_colored}{Like \code{playing_cards} but with five colored suits:
red hearts, black spades, green clubs, blue diamonds, and yellow stars.}
\item{playing_cards_tarot}{A (French Bourgeois) deck of tarot playing cards:
first four suits are hearts, spades, clubs, and diamonds with
14 ranks (ace through jack, knight, queen, king) plus a 15th \dQuote{Joker} rank
and a fifth "suit" of 22 trump cards (1-21 plus an \dQuote{excuse}).}}}
\item{reversi}{Boards and pieces for Reversi.
"board_face" provides lined boards with colored backgrounds.
"board_back" provides checkered boards.
"bit_face" / "bit_back" provides circular game tokens with differently colored sides:
red paired with green, black paired with white, and blue paired with yellow.
}
}
}
\examples{
cfgs <- game_systems(pawn = "joystick")
names(cfgs)

\donttest{# May take more than 5 seconds on CRAN servers
# standard dice, meeples, and joystick pawns
if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {

   opt <- options(piecepackr.at.inform = FALSE)
   grid::grid.newpage()
   dice <-  c("d4", "numeral", "d8", "d10", "d12", "d20")
   cfg <- paste0("dice_", dice)
   grid.piece("die_face", suit = c(1:6, 1), rank = 1:6,
              cfg = cfg, envir = cfgs, x = 1:6, y = 1,
              default.units = "in", op_scale = 0.5)
   grid.piece("die_face", rank=1:6, suit=1:6,
              x=1:6, y=2, default.units="in",
              op_scale=0.5, cfg=cfgs$dice)
   grid.piece("bit_face", suit=1:6,
              x=1:6, y=3, default.units="in",
              op_scale=0.5, cfg=cfgs$meeple)
   grid.piece("pawn_top", suit=1:6,
              x=1:6, y=4, default.units="in",
              op_scale=0.5, cfg=cfgs$piecepack)
   options(opt)
}
}
# dominoes
if (requireNamespace("grid", quietly = TRUE)) {
   grid::grid.newpage()
   colors <- c("black", "red", "green", "blue", "yellow", "white")
   cfg <- paste0("dominoes_", rep(colors, 2))
   grid.piece("tile_face",  suit=1:12, rank=1:12+1,
              cfg=cfg, envir=cfgs,
              x=rep(6:1, 2), y=rep(2*2:1, each=6),
              default.units="in", op_scale=0.5)
}
# piecepack "playing card expansion"
if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {
   grid::grid.newpage()
   df_tiles <- data.frame(piece_side="tile_back",
                          x=0.5+c(3,1,3,1), y=0.5+c(3,3,1,1),
                          suit=NA, angle=NA, z=1/8,
                          stringsAsFactors=FALSE)
   df_coins <- data.frame(piece_side="coin_back",
                          x=rep(4:1, 4), y=rep(4:1, each=4),
                          suit=c(1,4,1,4,4,1,4,1,2,3,2,3,3,2,3,2),
                          angle=rep(c(180,0), each=8), z=1/4+1/16,
                          stringsAsFactors=FALSE)
   df <- rbind(df_tiles, df_coins)
   pmap_piece(df, cfg = cfgs$playing_cards_expansion, op_scale=0.5,
              default.units="in")
}
}
\seealso{
\code{\link[=pp_cfg]{pp_cfg()}} for information about the \code{\link[=pp_cfg]{pp_cfg()}} objects returned by \code{game_systems()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/save_piece_obj.R
\name{save_piece_obj}
\alias{save_piece_obj}
\title{Save Wavefront OBJ files of board game pieces}
\usage{
save_piece_obj(
  piece_side = "tile_face",
  suit = 1,
  rank = 1,
  cfg = getOption("piecepackr.cfg", pp_cfg()),
  ...,
  x = 0,
  y = 0,
  z = 0,
  angle = 0,
  axis_x = 0,
  axis_y = 0,
  width = NA,
  height = NA,
  depth = NA,
  envir = getOption("piecepackr.envir"),
  filename = tempfile(fileext = "\%03d.obj"),
  scale = 1,
  res = 72
)
}
\arguments{
\item{piece_side}{A string with piece and side separated by a underscore e.g. "coin_face"}

\item{suit}{Number of suit (starting from 1).}

\item{rank}{Number of rank (starting from 1)}

\item{cfg}{Piecepack configuration list or \code{pp_cfg} object,
a list of \code{pp_cfg} objects,
or a character vector referring to names in \code{envir}
or a character vector referring to object names that
can be retrieved by \code{base::dynGet()}.}

\item{...}{Ignored.}

\item{x}{Where to place piece on x axis of viewport}

\item{y}{Where to place piece on y axis of viewport}

\item{z}{z-coordinate of the piece.  Has no effect if \code{op_scale} is \code{0}.}

\item{angle}{Angle (on xy plane) to draw piece at}

\item{axis_x}{First coordinate of the axis unit vector.}

\item{axis_y}{Second coordinate of the axis unit vector.}

\item{width}{Width of piece}

\item{height}{Height of piece}

\item{depth}{Depth (thickness) of piece.  Has no effect if \code{op_scale} is \code{0}.}

\item{envir}{Environment (or named list) containing configuration list(s).}

\item{filename}{Name of Wavefront OBJ object.  By default \code{tempfile(fileext = ".obj")}.}

\item{scale}{Multiplicative scaling factor to apply to width, height, and depth.}

\item{res}{Resolution of the faces.}
}
\value{
A data frame with named elements "obj", "mtl", "png" with the created filenames.
}
\description{
\code{save_piece_obj()} saves Wavefront OBJ files (including associated MTL and texture image) of board game pieces.
}
\examples{
    if (all(capabilities(c("cairo", "png")))) {
      cfg <- game_systems("sans", border = FALSE)$dominoes
      files <- save_piece_obj("tile_face", suit = 3+1, rank=6+1, cfg = cfg)
      print(files)
    }
}
\seealso{
See \code{\link{geometry_utils}} for a discussion of the 3D rotation parameterization.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install_ppverse.R
\name{install_ppverse}
\alias{install_ppverse}
\alias{pkgs_ppverse}
\title{Install additional packages from the piecepackr universe}
\usage{
install_ppverse(
  pkgs = pkgs_ppverse(free_libre_only),
  ...,
  dependencies = TRUE,
  free_libre_only = TRUE
)

pkgs_ppverse(free_libre_only = TRUE)
}
\arguments{
\item{pkgs}{Packages to install. Passed to \code{\link[utils:install.packages]{utils::install.packages()}}.}

\item{...}{Passed to \code{\link[utils:install.packages]{utils::install.packages()}}.}

\item{dependencies}{Logical indicating whether to install dependencies of \code{pkgs}. Passed to \code{\link[utils:install.packages]{utils::install.packages()}}.}

\item{free_libre_only}{Flag to only include packages that are Free/Libre Open Source.}
}
\description{
Uses \code{\link[utils:install.packages]{utils::install.packages()}} to install additional R packages
from the piecepackr universe \url{https://piecepackr.r-universe.dev/builds}.
\code{pkgs_ppverse()} returns a character vector of R packages
in the piecepackr universe.
}
\examples{
pkgs_ppverse()
\dontrun{# Installs non-CRAN packages from the piecepackr universe
  install_ppverse()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/save_piece_images.R
\name{save_piece_images}
\alias{save_piece_images}
\title{Save piecepack images}
\usage{
save_piece_images(
  cfg = getOption("piecepackr.cfg", pp_cfg()),
  directory = tempdir(),
  format = "svg",
  angle = 0
)
}
\arguments{
\item{cfg}{Piecepack configuration list}

\item{directory}{Directory where to place images}

\item{format}{Character vector of formats to save images in}

\item{angle}{Numeric vector of angles to rotate images (in degrees)}
}
\description{
Saves images of all individual piecepack pieces.
}
\examples{
  \donttest{# May take more than 5 seconds on CRAN server
  if (all(capabilities(c("cairo", "png")))) {
      cfg <- pp_cfg(list(suit_color="darkred,black,darkgreen,darkblue,grey"))
      save_piece_images(cfg, directory=tempdir(), format="svg", angle=0)
      save_piece_images(cfg, directory=tempdir(), format="png", angle=90)
  }
  }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-geometry.R
\name{AA_to_R}
\alias{AA_to_R}
\alias{R_to_AA}
\alias{geometry_utils}
\alias{R_x}
\alias{R_y}
\alias{R_z}
\alias{to_radians}
\alias{to_degrees}
\alias{to_x}
\alias{to_y}
\alias{to_r}
\alias{to_t}
\title{Helper functions for making geometric calculations.}
\usage{
AA_to_R(angle = 0, axis_x = 0, axis_y = 0, axis_z = NA, ...)

R_to_AA(R = diag(3))

R_x(angle = 0)

R_y(angle = 0)

R_z(angle = 0)

to_radians(t)

to_degrees(t)

to_x(t, r)

to_y(t, r)

to_r(x, y)

to_t(x, y)
}
\arguments{
\item{angle}{Numeric vector in degrees (counter-clockwise) or an \code{\link[affiner:angle]{affiner::angle()}} vector.}

\item{axis_x}{First coordinate of the axis unit vector.}

\item{axis_y}{Second coordinate of the axis unit vector.}

\item{axis_z}{Third coordinate of the axis unit vector (usually inferred).}

\item{...}{Ignored}

\item{R}{3D rotation matrix (post-multiplied)}

\item{t}{Numeric vector in degrees (or radians for \code{to_degrees()}) or an \code{\link[affiner:angle]{affiner::angle()}} vector.}

\item{r}{Radial distance}

\item{x}{Cartesian x coordinate}

\item{y}{Cartesian y coordinate}
}
\description{
\code{to_x()}, \code{to_y()}, \code{to_r()}, \code{to_t()} convert
between polar coordinates (in degrees) and Cartesian coordinates.
\code{to_degrees()} and \code{to_radians()} converts between degrees and radians.
\code{AA_to_R()} and \code{R_to_AA()} convert back and forth between (post-multiplied) rotation matrix
and axis-angle representations of 3D rotations.
\code{R_x()}, \code{R_y()}, and \code{R_z()} build (post-multiplied) rotation matrices for simple rotations around
the x, y, and z axes.
}
\details{
\code{\link[=pp_cfg]{pp_cfg()}} uses polar coordinates to determine where the "primary" and "directional"
symbols are located on a game piece.
They are also useful for drawing certain shapes and for making game diagrams on hex boards.

\code{piecepackr} and \code{grid} functions use angles in degrees
but the \code{base} trigonometry functions usually use radians.

\code{piecepackr}'s 3D graphics functions \code{\link[=save_piece_obj]{save_piece_obj()}}, \code{\link[=piece]{piece()}}, \code{\link[=piece3d]{piece3d()}}, and \code{\link[=piece_mesh]{piece_mesh()}}
use the axis-angle representation for 3D rotations.
The axis-angle representation involves specifying a unit vector
indicating the direction of an axis of rotation and an angle describing the (counter-clockwise)
rotation around that axis.  Because it is a unit vector one only needs to specify the first two elements,
\code{axis_x} and \code{axis_y}, and we are able to infer the 3rd element \code{axis_z}.  The default of
\code{axis = 0}, \code{axis_y = 0}, and implied \code{axis_z = 1}
corresponds to a rotation around the z-axis which is reverse-compatible
with the originally 2D \code{angle} interpretation in \code{\link[=grid.piece]{grid.piece()}}.  In order to figure out the appropriate
axis-angle representation parameters \code{R_to_AA()}, \code{R_x()}, \code{R_y()}, and \code{R_z()} allow one
to first come up with an appropriate (post-multiplied) 3D rotation matrix by chaining simple rotations
and then convert them to the corresponding axis-angle representation.
Pieces are rotated as if their center was at the origin.
}
\examples{
 to_x(90, 1)
 to_y(180, 0.5)
 to_t(0, -1)
 to_r(0.5, 0)
 all.equal(pi, to_radians(to_degrees(pi)))
 # default axis-angle axis is equivalent to a rotation about the z-axis
 all.equal(AA_to_R(angle=60), R_z(angle=60))
 # axis-angle representation of 90 rotation about the x-axis
 R_to_AA(R_x(90))
 # find Axis-Angle representation of first rotating about x-axis 180 degrees
 # and then rotating about z-axis 45 degrees
 R_to_AA(R_x(180) \%*\% R_z(45))

}
\seealso{
\url{https://en.wikipedia.org/wiki/Axis-angle_representation} for more details
about the Axis-angle representation of 3D rotations.
See \code{\link[base]{Trig}} for R's built-in trigonometric functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom_piece.R
\name{geom_piece}
\alias{geom_piece}
\alias{aes_piece}
\title{Draw board game pieces with ggplot2}
\usage{
geom_piece(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  envir = getOption("piecepackr.envir", piecepackr::game_systems()),
  op_scale = getOption("piecepackr.op_scale", 0),
  op_angle = getOption("piecepackr.op_angle", 45),
  inherit.aes = TRUE
)

aes_piece(df)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[ggplot2:aes]{aes()}}. If specified and
\code{inherit.aes = TRUE} (the default), it is combined with the default mapping
at the top level of the plot. You must supply \code{mapping} if there is no plot
mapping.}

\item{data}{The data to be displayed in this layer. There are three
options:

If \code{NULL}, the default, the data is inherited from the plot
data as specified in the call to \code{\link[ggplot2:ggplot]{ggplot()}}.

A \code{data.frame}, or other object, will override the plot
data. All objects will be fortified to produce a data frame. See
\code{\link[ggplot2:fortify]{fortify()}} for which variables will be created.

A \code{function} will be called with a single argument,
the plot data. The return value must be a \code{data.frame}, and
will be used as the layer data. A \code{function} can be created
from a \code{formula} (e.g. \code{~ head(.x, 10)}).}

\item{stat}{The statistical transformation to use on the data for this layer.
When using a \verb{geom_*()} function to construct a layer, the \code{stat}
argument can be used to override the default coupling between geoms and
stats. The \code{stat} argument accepts the following:
\itemize{
\item A \code{Stat} ggproto subclass, for example \code{StatCount}.
\item A string naming the stat. To give the stat as a string, strip the
function name of the \code{stat_} prefix. For example, to use \code{stat_count()},
give the stat as \code{"count"}.
\item For more information and other ways to specify the stat, see the
\link[ggplot2:layer_stats]{layer stat} documentation.
}}

\item{position}{A position adjustment to use on the data for this layer. This
can be used in various ways, including to prevent overplotting and
improving the display. The \code{position} argument accepts the following:
\itemize{
\item The result of calling a position function, such as \code{position_jitter()}.
This method allows for passing extra arguments to the position.
\item A string naming the position adjustment. To give the position as a
string, strip the function name of the \code{position_} prefix. For example,
to use \code{position_jitter()}, give the position as \code{"jitter"}.
\item For more information and other ways to specify the position, see the
\link[ggplot2:layer_positions]{layer position} documentation.
}}

\item{...}{Aesthetics, used to set an aesthetic to a fixed value.}

\item{envir}{Environment (or named list) containing configuration list(s).}

\item{op_scale}{How much to scale the depth of the piece in the oblique projection
(viewed from the top of the board).
\code{0} (the default) leads to an \dQuote{orthographic} projection,
\code{0.5} is the most common scale used in the \dQuote{cabinet} projection,
and \code{1.0} is the scale used in the \dQuote{cavalier} projection.}

\item{op_angle}{What is the angle of the oblique projection?  Has no effect if \code{op_scale} is \code{0}.}

\item{inherit.aes}{If \code{FALSE}, overrides the default aesthetics,
rather than combining with them. This is most useful for helper functions
that define both data and aesthetics and shouldn't inherit behaviour from
the default plot specification, e.g. \code{\link[ggplot2:annotation_borders]{annotation_borders()}}.}

\item{df}{A data frame of game piece information with (at least) the
named columns \dQuote{piece_side}, \dQuote{x}, and \dQuote{y}.}
}
\description{
\code{geom_piece()} creates a \code{ggplot2} geom.
\code{aes_piece()} takes a data frame and generates
an appropriate \code{ggplot2::aes()} mapping.
}
\details{
\code{geom_piece()} requires a fixed scale coordinate system with an aspect
ratio of 1 as provided by \code{ggplot2::coord_fixed()}.
\code{geom_piece()} also requires that \code{cfg} is a character vector (and not a \code{pp_cfg()} object).
In particular if using \code{op_transform()} one should set its argument \code{cfg_class = "character"}
if intending for use with \code{geom_piece()}.
}
\section{Aesthetics}{

\code{geom_piece()} understands the following aesthetics (required aesthetics are in bold).
See \code{\link[=pieceGrob]{pieceGrob()}} for more details.
\itemize{
\item \strong{\code{x}}
\item \strong{\code{y}}
\item \code{z}
\item \code{piece_side}
\item \code{rank}
\item \code{suit}
\item \code{cfg}
\item \code{width}
\item \code{height}
\item \code{depth}
\item \code{angle}
\item \code{scale}
\item \code{type}
}
}

\examples{
if (require("ggplot2", quietly = TRUE) && require("tibble", quietly = TRUE)) {
  envir <- game_systems("sans")
  df_board <- tibble(piece_side = "board_face", suit = 3, rank = 8,
                 x = 4.5, y = 4.5)
  df_w <- tibble(piece_side = "bit_back", suit = 6, rank = 1,
                 x = rep(1:8, 2), y = rep(1:2, each=8))
  df_b <- tibble(piece_side = "bit_back", suit = 1, rank = 1,
                 x = rep(1:8, 2), y = rep(7:8, each=8))
  df <- rbind(df_board, df_w, df_b)
  # 2D example
  # `cfg` must be a character vector for `geom_piece()`
  ggplot(df, aes_piece(df)) +
      geom_piece(cfg = "checkers1", envir = envir) +
      coord_fixed() +
      scale_x_piece() +
      scale_y_piece() +
      theme_minimal(28) +
      theme(panel.grid = element_blank())
}
if (require("ggplot2", quietly = TRUE) && require("tibble", quietly = TRUE)) {
  # 3D "oblique" projection example
  # `cfg_class` must be "character" when using with `geom_piece()`
  df3d <- op_transform(df, cfg = "checkers1", envir = envir,
                       op_angle = 45, cfg_class = "character")
  ggplot(df3d, aes_piece(df3d)) +
      geom_piece(cfg = "checkers1", envir = envir,
                 op_angle = 45, op_scale = 0.5) +
      coord_fixed() +
      theme_void()
}
}
\seealso{
\code{geom_piece()} is a wrapper around \code{\link[=pieceGrob]{pieceGrob()}}.
\code{\link[=scale_x_piece]{scale_x_piece()}} and \code{\link[=scale_y_piece]{scale_y_piece()}} are wrappers
around \code{\link[ggplot2:scale_x_continuous]{ggplot2::scale_x_continuous()}} and \code{\link[ggplot2:scale_y_continuous]{ggplot2::scale_y_continuous()}}
with better defaults for board game diagrams.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-font.R
\name{font_utils}
\alias{font_utils}
\alias{get_embedded_font}
\alias{has_font}
\title{Font utility functions}
\usage{
get_embedded_font(font, char)

has_font(font)
}
\arguments{
\item{font}{A character vector of font(s).}

\item{char}{A character vector of character(s) to be embedded by \code{grid::grid.text()}}
}
\value{
\code{get_embedded_font()} returns character vector of fonts that were actually embedded by \code{cairo_pdf()}.
\code{NA}'s means no embedded font detected: this either means that no font
was found or that a color emoji font was found and instead of a font an image was embedded.
}
\description{
\code{get_embedded_font()} returns which font is actually embedded
by \code{cairo_pdf()} for a given character.
\code{has_font()} tries to determine if a given font is available on the OS.
}
\details{
\code{get_embedded_font()} depends on the suggested \code{pdftools} package being installed
and R being compiled with Cairo support.
\code{has_font()} depends on either the suggested \code{systemfonts} (preferred) or \code{pdftools}
packages being installed.
}
\examples{
 if (requireNamespace("pdftools", quietly = TRUE) &&
     capabilities("cairo") &&
     !piecepackr:::is_cairo_maybe_buggy()) {
   chars <- c("a", "\u2666")
   fonts <- c("sans", "Sans Noto", "Noto Sans", "Noto Sans Symbols2")
   try(get_embedded_font(fonts, chars))
 }

 if (requireNamespace("systemfonts", quietly = TRUE) ||
     (requireNamespace("pdftools", quietly = TRUE) &&
      capabilities("cairo")) && !piecepackr:::is_cairo_maybe_buggy()) {
   try(has_font("Dejavu Sans"))
 }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-misc.R
\name{pp_utils}
\alias{pp_utils}
\alias{is_color_invisible}
\alias{inch}
\alias{cleave}
\alias{file2grob}
\title{Miscellaneous \code{piecepackr} utility functions}
\usage{
is_color_invisible(col)

inch(inches)

cleave(s, sep = ",", float = FALSE, color = FALSE)

file2grob(file, distort = TRUE)
}
\arguments{
\item{col}{Color}

\item{inches}{Number representing number of inches}

\item{s}{String to convert}

\item{sep}{Delimiter (defaults to ",")}

\item{float}{If \code{TRUE} cast to numeric}

\item{color}{if \code{TRUE} convert empty strings to \code{"transparent"}}

\item{file}{Filename of image}

\item{distort}{Logical value of whether one should preserve the aspect ratio
or distort to fit the area it is drawn in}
}
\description{
\code{cleave()} converts a delimiter separated string into a vector.
\code{inch(x)} is equivalent to \code{grid::unit(x, "in")}.
\code{is_color_invisible()} tells whether the color is transparent (and hence need not be drawn).
}
\examples{
 cleave("0.5,0.2,0.4,0.5", float=TRUE)
 cleave("black,darkred,#050EAA,,", color=TRUE)

 is_color_invisible("transparent")
 is_color_invisible(NA)
 is_color_invisible("blue")
 is_color_invisible("#05AE9C")

 if (requireNamespace("grid", quietly = TRUE)) {
     identical(inch(1), grid::unit(1, "inch"))
 }

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom_piece.R
\name{scale_x_piece}
\alias{scale_x_piece}
\alias{scale_y_piece}
\alias{label_letter}
\alias{label_counting}
\alias{breaks_counting}
\title{ggplot2 game diagram scales}
\usage{
scale_x_piece(
  ...,
  name = NULL,
  breaks = breaks_counting(),
  minor_breaks = NULL,
  labels = label_letter()
)

scale_y_piece(
  ...,
  name = NULL,
  breaks = breaks_counting(),
  minor_breaks = NULL,
  labels = label_counting()
)

label_letter()

label_counting()

breaks_counting()
}
\arguments{
\item{...}{Passed to \code{\link[ggplot2:scale_x_continuous]{ggplot2::scale_x_continuous()}} or \code{\link[ggplot2:scale_y_continuous]{ggplot2::scale_y_continuous()}}.}

\item{name}{The name of the scale. Used as the axis or legend title. If
\code{waiver()}, the default, the name of the scale is taken from the first
mapping used for that aesthetic. If \code{NULL}, the legend title will be
omitted.}

\item{breaks}{One of:
\itemize{
\item \code{NULL} for no breaks
\item \code{waiver()} for the default breaks computed by the
\link[scales:new_transform]{transformation object}
\item A numeric vector of positions
\item A function that takes the limits as input and returns breaks
as output (e.g., a function returned by \code{\link[scales:breaks_extended]{scales::extended_breaks()}}).
Note that for position scales, limits are provided after scale expansion.
Also accepts rlang \link[rlang:as_function]{lambda} function notation.
}}

\item{minor_breaks}{One of:
\itemize{
\item \code{NULL} for no minor breaks
\item \code{waiver()} for the default breaks (none for discrete, one minor break
between each major break for continuous)
\item A numeric vector of positions
\item A function that given the limits returns a vector of minor breaks. Also
accepts rlang \link[rlang:as_function]{lambda} function notation. When
the function has two arguments, it will be given the limits and major
break positions.
}}

\item{labels}{One of the options below. Please note that when \code{labels} is a
vector, it is highly recommended to also set the \code{breaks} argument as a
vector to protect against unintended mismatches.
\itemize{
\item \code{NULL} for no labels
\item \code{waiver()} for the default labels computed by the
transformation object
\item A character vector giving labels (must be same length as \code{breaks})
\item An expression vector (must be the same length as breaks). See ?plotmath for details.
\item A function that takes the breaks as input and returns labels
as output. Also accepts rlang \link[rlang:as_function]{lambda} function
notation.
}}
}
\value{
\code{scale_x_piece()} and \code{scale_y_piece()} return ggplot2 scale objects.
\code{label_letter()} and \code{label_counting()} return functions suitable for use with the \code{labels} scale argument.
\code{breaks_counting()} returns a function suitable for use with the \code{breaks} scale argument.
}
\description{
\code{scale_x_piece()} and \code{scale_y_piece()} are wrappers
around \code{\link[ggplot2:scale_x_continuous]{ggplot2::scale_x_continuous()}} and
\code{\link[ggplot2:scale_y_continuous]{ggplot2::scale_y_continuous()}} with "better"
defaults for board game diagrams.
\code{label_letter()} labels breaks with letters
and \code{label_counting()} labels breaks with positive integers
to more easily generate (i.e. chess) algebraic notation coordinates.
\code{breaks_counting()} generates breaks of just the positive integers within the limits.
}
\examples{
if (require("ggplot2", quietly = TRUE) && require("tibble", quietly = TRUE)) {
  envir <- game_systems("sans")
  df_board <- tibble(piece_side = "board_face", suit = 3, rank = 8,
                 x = 4.5, y = 4.5)
  df_w <- tibble(piece_side = "bit_back", suit = 6, rank = 1,
                 x = rep(1:8, 2), y = rep(1:2, each=8))
  df_b <- tibble(piece_side = "bit_back", suit = 1, rank = 1,
                 x = rep(1:8, 2), y = rep(7:8, each=8))
  df <- rbind(df_board, df_w, df_b)

  # `cfg` must be a character vector for `geom_piece()`
  ggplot(df, aes_piece(df)) +
      geom_piece(cfg = "checkers1", envir = envir) +
      coord_fixed() +
      scale_x_piece() +
      scale_y_piece() +
      theme_minimal(28) +
      theme(panel.grid = element_blank())
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pieceGrob-grid.R
\name{grid.piece}
\alias{grid.piece}
\alias{pieceGrob}
\title{Draw board game pieces with grid}
\usage{
pieceGrob(
  piece_side = "tile_back",
  suit = NA,
  rank = NA,
  cfg = getOption("piecepackr.cfg", pp_cfg()),
  x = unit(0.5, "npc"),
  y = unit(0.5, "npc"),
  z = NA,
  angle = 0,
  ...,
  width = NA,
  height = NA,
  depth = NA,
  op_scale = getOption("piecepackr.op_scale", 0),
  op_angle = getOption("piecepackr.op_angle", 45),
  default.units = getOption("piecepackr.default.units", "npc"),
  envir = getOption("piecepackr.envir"),
  name = NULL,
  gp = NULL,
  vp = NULL,
  scale = 1,
  alpha = 1,
  type = "normal",
  bleed = FALSE
)

grid.piece(
  piece_side = "tile_back",
  suit = NA,
  rank = NA,
  cfg = getOption("piecepackr.cfg", pp_cfg()),
  x = unit(0.5, "npc"),
  y = unit(0.5, "npc"),
  z = NA,
  angle = 0,
  ...,
  width = NA,
  height = NA,
  depth = NA,
  op_scale = getOption("piecepackr.op_scale", 0),
  op_angle = getOption("piecepackr.op_angle", 45),
  default.units = getOption("piecepackr.default.units", "npc"),
  envir = getOption("piecepackr.envir"),
  name = NULL,
  gp = NULL,
  draw = TRUE,
  vp = NULL,
  scale = 1,
  alpha = 1,
  type = "normal",
  bleed = FALSE
)
}
\arguments{
\item{piece_side}{A string with piece and side separated by a underscore e.g. "coin_face"}

\item{suit}{Number of suit (starting from 1).}

\item{rank}{Number of rank (starting from 1)}

\item{cfg}{Piecepack configuration list or \code{pp_cfg} object,
a list of \code{pp_cfg} objects,
or a character vector referring to names in \code{envir}
or a character vector referring to object names that
can be retrieved by \code{base::dynGet()}.}

\item{x}{Where to place piece on x axis of viewport}

\item{y}{Where to place piece on y axis of viewport}

\item{z}{z-coordinate of the piece.  Has no effect if \code{op_scale} is \code{0}.}

\item{angle}{Angle (on xy plane) to draw piece at}

\item{...}{Ignored.}

\item{width}{Width of piece}

\item{height}{Height of piece}

\item{depth}{Depth (thickness) of piece.  Has no effect if \code{op_scale} is \code{0}.}

\item{op_scale}{How much to scale the depth of the piece in the oblique projection
(viewed from the top of the board).
\code{0} (the default) leads to an \dQuote{orthographic} projection,
\code{0.5} is the most common scale used in the \dQuote{cabinet} projection,
and \code{1.0} is the scale used in the \dQuote{cavalier} projection.}

\item{op_angle}{What is the angle of the oblique projection?  Has no effect if \code{op_scale} is \code{0}.}

\item{default.units}{A string indicating the default units to use if
'x', 'y', 'width', and/or 'height' are only given as numeric vectors.}

\item{envir}{Environment (or named list) containing configuration list(s).}

\item{name}{A character identifier (for grid)}

\item{gp}{An object of class \dQuote{gpar}.}

\item{vp}{A \code{grid} viewport object (or \code{NULL}).}

\item{scale}{Multiplicative scaling factor to apply to width, height, and depth.}

\item{alpha}{Alpha channel for transparency.}

\item{type}{Type of grid grob to use.  Either \code{"normal"} (default), \code{"picture"}, \code{"raster"}, or \code{"transformation"}.
\code{"picture"} exports to (temporary) svg and re-imports as a \code{\link[grImport2:pictureGrob]{grImport2::pictureGrob()}}.
\code{"raster"} exports to (temporary) png and re-imports as a \code{\link[grid:rasterGrob]{grid::rasterGrob()}}.
\code{"transformation"} uses the affine transformation feature only supported in
R 4.2+ within select graphic devices.
The latter three can be useful if drawing pieces really big or small and don't want
to mess with re-configuring fontsizes and linewidths.}

\item{bleed}{If \code{FALSE} do not add a \dQuote{bleed} zone around the piece,
otherwise add a \dQuote{bleed} zone around the piece:\itemize{
\item{If \code{bleed} is \code{TRUE} we will add 1/8 inch bleeds}
\item{If \code{bleed} is a \code{\link[grid:unit]{grid::unit()}} we will use it as bleed size}
\item{If \code{bleed} is numeric we will convert to \code{\link[grid:unit]{grid::unit()}} via \code{grid::unit(bleed, default.units)}}
}
A non-\code{FALSE} \code{bleed} is incompatible with \code{op_scale > 0} (drawing in an \dQuote{oblique projection}).}

\item{draw}{A logical value indicating whether graphics output should be produced.}
}
\value{
A \code{grid} grob object.  If \code{draw} is \code{TRUE} then as a side effect
\code{grid.piece()} will also draw it to the graphics device.
}
\description{
\code{grid.piece()} draws board game pieces onto the graphics device.
\code{pieceGrob()} is its \code{grid} \dQuote{grob} counterpart.
}
\examples{
  if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {
    opt <- options(piecepackr.at.inform = FALSE)

    draw_pp_diagram <- function(cfg=pp_cfg(), op_scale=0) {
        g.p <- function(...) {
            grid.piece(..., op_scale=op_scale, cfg=cfg, default.units="in")
        }
        g.p("tile_back", x=0.5+c(3,1,3,1), y=0.5+c(3,3,1,1))
        g.p("tile_back", x=0.5+3, y=0.5+1, z=1/4+1/8)
        g.p("tile_back", x=0.5+3, y=0.5+1, z=2/4+1/8)
        g.p("die_face", suit=3, rank=5, x=1, y=1, z=1/4+1/4)
        g.p("pawn_face", x=1, y=4, z=1/4+1/2, angle=90)
        g.p("coin_back", x=3, y=4, z=1/4+1/16, angle=180)
        g.p("coin_back", suit=4, x=3, y=4, z=1/4+1/8+1/16, angle=180)
        g.p("coin_back", suit=2, x=3, y=1, z=3/4+1/8, angle=90)
    }

    # default piecepack, orthogonal projection
    draw_pp_diagram(cfg=pp_cfg())

    options(opt) # Reset options
  }
  if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {
    # custom configuration, orthogonal projection
    grid::grid.newpage()
    dark_colorscheme <- list(suit_color="darkred,black,darkgreen,darkblue,black",
                         invert_colors.suited=TRUE, border_color="black", border_lex=2)
    traditional_ranks <- list(use_suit_as_ace=TRUE, rank_text=",a,2,3,4,5")
    cfg <- c(dark_colorscheme, traditional_ranks)
    draw_pp_diagram(cfg=pp_cfg(cfg))
  }
  if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {
    # custom configuration, oblique projection
    grid::grid.newpage()
    cfg3d <- list(width.pawn=0.75, height.pawn=0.75, depth.pawn=1,
                       dm_text.pawn="", shape.pawn="convex6", invert_colors.pawn=TRUE,
                       edge_color.coin="tan", edge_color.tile="tan")
    cfg <- pp_cfg(c(cfg, cfg3d))
    draw_pp_diagram(cfg=pp_cfg(cfg), op_scale=0.5)
  }
}
\seealso{
\code{\link[=pmap_piece]{pmap_piece()}} which applies \code{pieceGrob()} over rows of a data frame.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crosshairGrob.R
\name{grid.crosshair}
\alias{grid.crosshair}
\alias{crosshairGrob}
\alias{segmentsCrosshairGrob}
\alias{circledSegmentsCrosshairGrob}
\alias{squaredSegmentsCrosshairGrob}
\alias{squaresCrosshairGrob}
\title{Draw crosshairs with grid}
\usage{
crosshairGrob(
  ...,
  piece_side = "tile_back",
  suit = NA,
  rank = NA,
  cfg = getOption("piecepackr.cfg", pp_cfg()),
  x = unit(0.5, "npc"),
  y = unit(0.5, "npc"),
  angle = 0,
  width = NA,
  height = NA,
  scale = 1,
  default.units = "npc",
  envir = getOption("piecepackr.envir"),
  name = NULL,
  gp = gpar(),
  vp = NULL,
  ch_width = unit(1/6, "in"),
  ch_grob = squaresCrosshairGrob()
)

grid.crosshair(..., draw = TRUE)

segmentsCrosshairGrob(
  ...,
  x = unit(0.5, "npc"),
  y = unit(0.5, "npc"),
  width = unit(1, "snpc"),
  height = unit(1, "snpc"),
  default.units = "npc",
  name = NULL,
  gp = gpar(),
  vp = NULL
)

circledSegmentsCrosshairGrob(
  ...,
  x = unit(0.5, "npc"),
  y = unit(0.5, "npc"),
  width = unit(1, "snpc"),
  height = unit(1, "snpc"),
  default.units = "npc",
  name = NULL,
  gp = gpar(),
  vp = NULL
)

squaredSegmentsCrosshairGrob(
  ...,
  x = unit(0.5, "npc"),
  y = unit(0.5, "npc"),
  width = unit(1, "snpc"),
  height = unit(1, "snpc"),
  default.units = "npc",
  name = NULL,
  gp = gpar(),
  vp = NULL
)

squaresCrosshairGrob(
  ...,
  x = unit(0.5, "npc"),
  y = unit(0.5, "npc"),
  width = unit(1, "snpc"),
  height = unit(1, "snpc"),
  default.units = "npc",
  name = NULL,
  gp = gpar(),
  vp = NULL
)
}
\arguments{
\item{...}{\code{crosshairGrob()} ignores; \code{grid.crosshair()} passes to \code{crosshairGrob()}.}

\item{piece_side}{A string with piece and side separated by a underscore e.g. "coin_face"}

\item{suit}{Number of suit (starting from 1).}

\item{rank}{Number of rank (starting from 1)}

\item{cfg}{Piecepack configuration list or \code{pp_cfg} object,
a list of \code{pp_cfg} objects,
or a character vector referring to names in \code{envir}
or a character vector referring to object names that
can be retrieved by \code{base::dynGet()}.}

\item{x}{Where to place piece on x axis of viewport}

\item{y}{Where to place piece on y axis of viewport}

\item{angle}{Angle (on xy plane) to draw piece at}

\item{width}{Width of piece}

\item{height}{Height of piece}

\item{scale}{Multiplicative scaling factor to apply to width, height, and depth.}

\item{default.units}{A string indicating the default units to use if
'x', 'y', 'width', and/or 'height' are only given as numeric vectors.}

\item{envir}{Environment (or named list) containing configuration list(s).}

\item{name}{A character identifier (for grid)}

\item{gp}{An object of class \dQuote{gpar}.}

\item{vp}{A \code{grid} viewport object (or \code{NULL}).}

\item{ch_width}{Width/height of \code{ch_grob}'s viewport.}

\item{ch_grob}{Crosshair grob.  Will be drawn in each corner of the piece in a viewport with \code{ch_width} width and \code{ch_width} height.
\code{segmentsCrosshairGrob()} simply wraps \code{\link[grid:segmentsGrob]{grid::segmentsGrob()}}.
\code{circledSegmentsCrosshairGrob()} wraps \code{\link[grid:segmentsGrob]{grid::segmentsGrob()}} and \code{\link[grid:circleGrob]{grid::circleGrob()}}.
\code{squaredSegmentsCrosshairGrob()} wraps \code{\link[grid:segmentsGrob]{grid::segmentsGrob()}} and \code{\link[grid:rectGrob]{grid::rectGrob()}}.
\code{squaresCrosshairGrob()} wraps \code{\link[grid:rectGrob]{grid::rectGrob()}} and alternates
black/white squares for visibility on both light and dark backgrounds.}

\item{draw}{A logical value indicating whether graphics output should be produced.}
}
\value{
A grid grob.
}
\description{
\code{grid.crosshair()} draws crosshairs at the corners of a rectangular area.
\code{crosshairGrob()} is its grid grob counterpart.
They are intended for use in adding crosshairs at the corners of
game pieces in print-and-play layouts.
One can also use the lower level
\code{circledSegmentsCrosshairGrob()}, \code{segmentsCrosshairGrob()}, \code{squaredSegmentsCrosshairGrob()}, and \code{squaresCrosshairGrob()}
(which can be drawn with \code{\link[grid:grid.draw]{grid::grid.draw()}})
to add individual crosshairs to specified (x,y) locations (e.g. for use as registration marks).
}
\examples{
if (requireNamespace("grid", quietly = TRUE) &&
    piecepackr:::device_supports_unicode()) {
  cfg <- pp_cfg(list(border_color = NA))
  grid::grid.newpage()
  df <- data.frame(piece_side = "tile_face", suit = 2, rank = 2,
                   x = 2, y = 2, angle = 0,
                   stringsAsFactors = FALSE)
  pmap_piece(df, grid.piece, cfg = cfg, default.units = "in")
  pmap_piece(df, grid.crosshair, cfg = cfg, default.units = "in")
}
if (requireNamespace("grid", quietly = TRUE) &&
    piecepackr:::device_supports_unicode()) {
  grid::grid.newpage()
  pmap_piece(df, grid.piece, cfg = cfg, default.units = "in")
  pmap_piece(df, grid.crosshair, cfg = cfg, default.units = "in",
             ch_grob = segmentsCrosshairGrob())
}
if (requireNamespace("grid", quietly = TRUE)) {
  grid::grid.newpage()
  fns <- list(squaresCrosshairGrob, segmentsCrosshairGrob,
              circledSegmentsCrosshairGrob, squaredSegmentsCrosshairGrob)
  labels <- c("squaresCrosshairGrob()", "segmentsCrosshairGrob()",
              "circledSegmentsCrosshairGrob()", "squaredSegmentsCrosshairGrob()")
  xs <- c(0.25, 0.75, 0.25, 0.75)
  ys <- c(0.70, 0.70, 0.30, 0.30)
  label_ys <- c(0.55, 0.55, 0.15, 0.15)
  for (i in seq_along(fns)) {
    grid::grid.draw(fns[[i]](x = grid::unit(xs[i], "npc"),
                             y = grid::unit(ys[i], "npc"),
                             width = grid::unit(0.15, "npc"),
                             height = grid::unit(0.15, "npc")))
    grid::grid.text(labels[i], x = xs[i], y = label_ys[i], gp = grid::gpar(cex = 0.7))
  }
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/save_piece_obj.R
\name{save_ellipsoid_obj}
\alias{save_ellipsoid_obj}
\alias{save_peg_doll_obj}
\title{Alternative Wavefront OBJ file generators}
\usage{
save_ellipsoid_obj(
  piece_side = "bit_face",
  suit = 1,
  rank = 1,
  cfg = getOption("piecepackr.cfg", pp_cfg()),
  ...,
  x = 0,
  y = 0,
  z = 0,
  angle = 0,
  axis_x = 0,
  axis_y = 0,
  width = NA,
  height = NA,
  depth = NA,
  filename = tempfile(fileext = ".obj"),
  subdivide = 3
)

save_peg_doll_obj(
  piece_side = "pawn_top",
  suit = 1,
  rank = 1,
  cfg = getOption("piecepackr.cfg", pp_cfg()),
  ...,
  x = 0,
  y = 0,
  z = 0,
  angle = 0,
  axis_x = 0,
  axis_y = 0,
  width = NA,
  height = NA,
  depth = NA,
  filename = tempfile(fileext = ".obj"),
  res = 72
)
}
\arguments{
\item{piece_side}{A string with piece and side separated by a underscore e.g. "coin_face"}

\item{suit}{Number of suit (starting from 1).}

\item{rank}{Number of rank (starting from 1)}

\item{cfg}{Piecepack configuration list or \code{pp_cfg} object,
a list of \code{pp_cfg} objects,
or a character vector referring to names in \code{envir}
or a character vector referring to object names that
can be retrieved by \code{base::dynGet()}.}

\item{...}{Ignored.}

\item{x}{Where to place piece on x axis of viewport}

\item{y}{Where to place piece on y axis of viewport}

\item{z}{z-coordinate of the piece.  Has no effect if \code{op_scale} is \code{0}.}

\item{angle}{Angle (on xy plane) to draw piece at}

\item{axis_x}{First coordinate of the axis unit vector.}

\item{axis_y}{Second coordinate of the axis unit vector.}

\item{width}{Width of piece}

\item{height}{Height of piece}

\item{depth}{Depth (thickness) of piece.  Has no effect if \code{op_scale} is \code{0}.}

\item{filename}{Name of Wavefront OBJ object.  By default \code{tempfile(fileext = ".obj")}.}

\item{subdivide}{Increasing this value makes for a smoother ellipsoid (and larger OBJ file and slower render).
See \code{\link[rgl:ellipse3d]{rgl::ellipse3d()}}.}

\item{res}{Resolution of the faces.}
}
\description{
These are alternative Wavefront OBJ generators intended to be used as a \code{obj_fn} attribute
in a \code{\link[=pp_cfg]{pp_cfg()}} \dQuote{configuration list}.
\code{save_ellipsoid_obj()} saves an ellipsoid with a color equal to that piece's \code{background_color}.
\code{save_peg_doll_obj()} saves a \dQuote{peg doll} style doll with a color equal to that piece's \code{edge_color}
with a \dQuote{pawn belt} around it's waste from that suit's and rank's \code{belt_face}.
}
\seealso{
See \code{\link[=pp_cfg]{pp_cfg()}} for a discussion of \dQuote{configuration lists}.
Wavefront OBJ file generators are used by \code{\link[=save_piece_obj]{save_piece_obj()}} and (by default)
\code{\link[=piece3d]{piece3d()}} (\code{rgl} wrapper), \code{\link[=piece]{piece()}} (\code{rayrender} wrapper),
and \code{\link[=piece_mesh]{piece_mesh()}} (\code{rayvertex} wrapper).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{spdx_license_list}
\alias{spdx_license_list}
\title{SPDX License List data}
\format{
a data frame with eight variables:
\describe{
\item{id}{SPDX Identifier.}
\item{name}{Full name of license.
For Creative Commons licenses these have been tweaked from the SPDX version
to more closely match the full name used by Creative Commons Foundation. }
\item{url}{URL for copy of license located at \code{spdx.org}}
\item{fsf}{Is this license considered Free/Libre by the FSF?}
\item{osi}{Is this license OSI approved?}
\item{deprecated}{Has this SPDFX Identifier been deprecated by SPDX?}
\item{badge}{Filename of appropriate \dQuote{button mark} badge (if any)
located in \code{system.file("extdata/badges", package = "piecepackr")}.}
\item{url_alt}{Alternative URL for license.
Manually created for a subset of Creative Commons licenses.
Others taken from \url{https://github.com/sindresorhus/spdx-license-list}.
}
}
}
\usage{
spdx_license_list
}
\description{
\code{spdx_license_list} is a data frame of SPDX License List data.
}
\seealso{
See \url{https://spdx.org/licenses/} for more information.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/save_print_and_play.R
\name{save_print_and_play}
\alias{save_print_and_play}
\title{Save piecepack print-and-play (PnP) file}
\usage{
save_print_and_play(
  cfg = getOption("piecepackr.cfg", pp_cfg()),
  output_filename = "piecepack.pdf",
  size = c("letter", "A4", "A5", "4x6"),
  pieces = NULL,
  arrangement = c("single-sided", "double-sided"),
  dev = NULL,
  dev.args = list(family = cfg$fontfamily, onefile = TRUE, units = "in", bg = "white",
    res = 300),
  quietly = FALSE,
  ...,
  bleed = FALSE,
  size_bleed = NULL
)
}
\arguments{
\item{cfg}{Piecepack configuration list or \code{pp_cfg} object}

\item{output_filename}{Filename for print-and-play file}

\item{size}{PnP output size (currently supports either "letter", "A4", or "A5").
This is the targeted \dQuote{trim} size of the print-and-play file
(\code{size_bleed} can be used to make the print-and-play file larger than this).
"A5" is in \dQuote{portrait} mode whereas the other sizes are in \dQuote{landscape} mode.}

\item{pieces}{Character vector of desired PnP pieces.
Supports "piecepack", "matchsticks", "pyramids", "subpack", or "all".
If \code{NULL} and combination of \code{size} / \code{bleed} values supports "matchsticks" and "pyramids"
then defaults to \code{c("piecepack", "pyramids", "matchsticks")} else just "piecepack".}

\item{arrangement}{Either "single-sided" or "double-sided".}

\item{dev}{Graphics device function to use if \code{open_device} is \code{FALSE}.
If \code{NULL} infer a reasonable choice from \code{file}.}

\item{dev.args}{Additional arguments to pass to \code{dev} (besides \code{filename}, \code{width}, and \code{height}).
Will filter out any names that aren't in \code{formals(dev)}.}

\item{quietly}{Whether to hide messages about missing metadata
in the provided configuration.}

\item{...}{Currently ignored.}

\item{bleed}{If \code{TRUE} produce a variant print-and-play file with "bleed" zones
and "crop marks" around game pieces.
Currently only supports \code{pieces = "piecepack"}.}

\item{size_bleed}{A list with names "top", "right", "bottom", "left"
containing numeric values indicating the inches "bleed" to add to
the \code{size} of the print-and-play layout.
The default \code{NULL} means no such bleed added to "letter", "A4", and "A5"
layouts.
NB. multiply millimeters by \code{0.0393700787} to convert to inches.
We currently don't support an asymmetric left/right bleed combined with
\code{arrangement = "double-sided"}.}
}
\description{
Save piecepack print-and-play (PnP) file
}
\examples{
\donttest{# May take more than 5 seconds on CRAN servers
if (capabilities("cairo")) {
  cfg <- pp_cfg(list(invert_colors.suited=TRUE))
  cfg$description <- 'Piecepack with an "inverted" color scheme.'
  cfg$title <- '"Inverted" piecepack'
  cfg$copyright <- "\u00a9 2022 Trevor L Davis.  Some Right Reserved."
  cfg$spdx_id <- "CC-BY-4.0"
  cfg$credit <- ""

  file <- tempfile("my_pnp_file", fileext = ".pdf")
  file_ds <- tempfile("my_pnp_file_ds", fileext = ".pdf")
  file_a4 <- tempfile("my_pnp_file_a4", fileext = ".pdf")
  file_a5 <- tempfile("my_pnp_file_a5", fileext = ".pdf")

  save_print_and_play(cfg, file)
  save_print_and_play(cfg, file_ds, arrangement="double-sided")
  save_print_and_play(cfg, file_a4, size="A4", pieces="all")
  save_print_and_play(cfg, file_a5, size="A5")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piece3d-rgl.R
\name{piece3d}
\alias{piece3d}
\title{Render board game pieces with rgl}
\usage{
piece3d(
  piece_side = "tile_back",
  suit = NA,
  rank = NA,
  cfg = getOption("piecepackr.cfg", pp_cfg()),
  x = 0,
  y = 0,
  z = NA,
  angle = 0,
  axis_x = 0,
  axis_y = 0,
  width = NA,
  height = NA,
  depth = NA,
  envir = getOption("piecepackr.envir"),
  ...,
  scale = 1,
  res = 72,
  alpha = 1,
  lit = FALSE,
  shininess = 50,
  textype = NA
)
}
\arguments{
\item{piece_side}{A string with piece and side separated by a underscore e.g. "coin_face"}

\item{suit}{Number of suit (starting from 1).}

\item{rank}{Number of rank (starting from 1)}

\item{cfg}{Piecepack configuration list or \code{pp_cfg} object,
a list of \code{pp_cfg} objects,
or a character vector referring to names in \code{envir}
or a character vector referring to object names that
can be retrieved by \code{base::dynGet()}.}

\item{x}{Where to place piece on x axis of viewport}

\item{y}{Where to place piece on y axis of viewport}

\item{z}{z-coordinate of the piece.  Has no effect if \code{op_scale} is \code{0}.}

\item{angle}{Angle (on xy plane) to draw piece at}

\item{axis_x}{First coordinate of the axis unit vector.}

\item{axis_y}{Second coordinate of the axis unit vector.}

\item{width}{Width of piece}

\item{height}{Height of piece}

\item{depth}{Depth (thickness) of piece.  Has no effect if \code{op_scale} is \code{0}.}

\item{envir}{Environment (or named list) containing configuration list(s).}

\item{...}{Ignored.}

\item{scale}{Multiplicative scaling factor to apply to width, height, and depth.}

\item{res}{Resolution of the faces.}

\item{alpha}{Alpha channel for transparency.}

\item{lit}{logical, specifying if rgl lighting calculation should take place.}

\item{shininess}{Properties for rgl lighting calculation.}

\item{textype}{Use \code{"rgba"} when sure texture will have alpha transparency.
Use \code{"rgb"} when sure texture will not have alpha transparency
(in particular \code{rgl}'s WebGL export will likely work better).
If \code{NA} we will read the texture and figure out a reasonable value.}
}
\value{
A numeric vector of rgl object IDs.
}
\description{
\code{piece3d()} draws board games pieces using the rgl package.
}
\examples{
if (requireNamespace("rgl", quietly = TRUE) && all(capabilities(c("cairo", "png")))) {
  rgl::open3d()
  cfg <- game_systems("sans", border = FALSE)$piecepack
  piece3d("tile_back", suit = 3, rank = 3, cfg = cfg, x = 0, y = 0, z = 0)
  piece3d("coin_back", suit = 4, rank = 2, cfg = cfg, x = 0.5, y = 0.5, z = 0.25)
  piece3d("pawn_top", suit = 1, cfg = cfg, x = -0.5, y = 0.5, z = 0.6)
  piece3d("die_face", suit = 3, cfg = cfg, x = -0.5, y = -0.5, z = 0.375)
  piece3d("pyramid_top", suit = 2, rank = 3, cfg = cfg, x = 1.5, y = 0.0, z = 0.31875)
  invisible(NULL)
}
}
\seealso{
See \code{\link[rgl]{rgl-package}} for more information about the \code{rgl} package.
See \code{\link[rgl:material3d]{rgl::material3d()}} for more info about setting \code{rgl} material properties.
See \code{\link{geometry_utils}} for a discussion of the 3D rotation parameterization.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piecepackr-package.R
\docType{package}
\name{piecepackr-package}
\alias{piecepackr}
\alias{piecepackr-package}
\title{piecepackr: Board Game Graphics}
\description{
\if{html}{\figure{logo.svg}{options: style='float: right' alt='logo' width='120'}}

Functions to make board game graphics with the 'ggplot2', 'grid', 'rayrender', 'rayvertex', and 'rgl' packages. Specializes in game diagrams, animations, and "Print & Play" layouts for the 'piecepack' \url{https://www.ludism.org/ppwiki} but can make graphics for other board game systems. Includes configurations for several public domain game systems such as checkers, (double-18) dominoes, go, 'piecepack', playing cards, etc.
}
\section{Package options}{

The following piecepackr function arguments may be set globally via \code{\link[base:options]{base::options()}}:
\describe{
\item{piecepackr.at.inform}{If \code{FALSE} turns off messages when affine transformation support not detected in active graphics device.}
\item{piecepackr.cfg}{Sets a new default for the \code{cfg} argument}
\item{piecepackr.check.cairo}{If \code{FALSE} don't check the version of cairo}
\item{piecepackr.default.units}{Sets a new default for the \code{default.units} argument}
\item{piecepackr.envir}{Sets a new default for the \code{envir} argument}
\item{piecepackr.fs.inform}{If \code{FALSE} turns off messages when stroking and filling path support not detected in active graphics device.}
\item{piecepackr.metadata.inform}{If \code{FALSE} turns off messages when support for embedding metadata not detected.}
\item{piecepackr.op_angle}{Sets a new default for the \code{op_angle} argument}
\item{piecepackr.op_scale}{Sets a new default for the \code{op_scale} argument}
\item{piecepackr.rgr.inform}{If \code{FALSE} turns off messages when radial gradient support not detected in active graphics device.}
\item{piecepackr.trans}{Sets a new default for the \code{trans} argument}
}
}

\seealso{
Useful links:
\itemize{
\item blog: \url{https://trevorldavis.com/piecepackr/}
\item pkgdown: \url{https://trevorldavis.com/R/piecepackr/}
\item forum: \url{https://groups.google.com/forum/#!forum/piecepackr}
\item Report bugs: \url{https://github.com/piecepackr/piecepackr/issues}
}
}
\author{
\strong{Maintainer}: Trevor L. Davis \email{trevor.l.davis@gmail.com} (\href{https://orcid.org/0000-0001-6341-4639}{ORCID})

Authors:
\itemize{
  \item Trevor L. Davis \email{trevor.l.davis@gmail.com} (\href{https://orcid.org/0000-0001-6341-4639}{ORCID})
}

Other contributors:
\itemize{
  \item Linux Foundation (Uses some data from the "SPDX License List" <https://github.com/spdx/license-list-XML>) [data contributor]
  \item Delapouite <https://delapouite.com/> (Meeple shape extracted from "Meeple icon" <https://game-icons.net/1x1/delapouite/meeple.html> / "CC BY 3.0" <https://creativecommons.org/licenses/by/3.0/>) [illustrator]
  \item Creative Commons (`save_print_and_play()` uses "license badges" from Creative Commons to describe the generated print-and-play file's license) [illustrator]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piece-rayrender.R
\name{piece}
\alias{piece}
\title{Create rayrender board game piece objects}
\usage{
piece(
  piece_side = "tile_back",
  suit = NA,
  rank = NA,
  cfg = getOption("piecepackr.cfg", pp_cfg()),
  x = 0,
  y = 0,
  z = NA,
  angle = 0,
  axis_x = 0,
  axis_y = 0,
  width = NA,
  height = NA,
  depth = NA,
  envir = getOption("piecepackr.envir"),
  ...,
  scale = 1,
  res = 72
)
}
\arguments{
\item{piece_side}{A string with piece and side separated by a underscore e.g. "coin_face"}

\item{suit}{Number of suit (starting from 1).}

\item{rank}{Number of rank (starting from 1)}

\item{cfg}{Piecepack configuration list or \code{pp_cfg} object,
a list of \code{pp_cfg} objects,
or a character vector referring to names in \code{envir}
or a character vector referring to object names that
can be retrieved by \code{base::dynGet()}.}

\item{x}{Where to place piece on x axis of viewport}

\item{y}{Where to place piece on y axis of viewport}

\item{z}{z-coordinate of the piece.  Has no effect if \code{op_scale} is \code{0}.}

\item{angle}{Angle (on xy plane) to draw piece at}

\item{axis_x}{First coordinate of the axis unit vector.}

\item{axis_y}{Second coordinate of the axis unit vector.}

\item{width}{Width of piece}

\item{height}{Height of piece}

\item{depth}{Depth (thickness) of piece.  Has no effect if \code{op_scale} is \code{0}.}

\item{envir}{Environment (or named list) containing configuration list(s).}

\item{...}{Ignored.}

\item{scale}{Multiplicative scaling factor to apply to width, height, and depth.}

\item{res}{Resolution of the faces.}
}
\value{
A rayrender object.
}
\description{
\code{piece()} creates 3d board game piece objects for use with the rayrender package.
}
\examples{
opt <- options(cores = getOption("Ncpus"))
cfg <- game_systems("sans", border = FALSE)$piecepack
should_run_piece_example <- piecepackr:::donttest() &&
  requireNamespace("rayrender", quietly = TRUE) &&
  all(capabilities(c("cairo", "png")))
if (should_run_piece_example) {
  scene <- piece("tile_face", suit = 3, rank = 3, cfg = cfg)
  rayrender::render_scene(scene)
}
if (should_run_piece_example) {
  scene <- piece("coin_back", suit = 4, rank = 2, cfg = cfg)
  rayrender::render_scene(scene)
}
if (should_run_piece_example) {
  scene <- piece("pawn_face", suit = 2, cfg = cfg)
  rayrender::render_scene(scene)
}
options(opt)
}
\seealso{
See \url{https://www.rayrender.net} for more information about the \code{rayrender} package.
See \code{\link{geometry_utils}} for a discussion of the 3D rotation parameterization.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/piecepackr-defunct.R
\name{piecepackr-defunct}
\alias{piecepackr-defunct}
\alias{halmaGrob}
\alias{kiteGrob}
\alias{pyramidGrob}
\alias{convexGrobFn}
\alias{concaveGrobFn}
\alias{gridlinesGrob}
\alias{matGrob}
\alias{checkersGrob}
\alias{hexlinesGrob}
\alias{get_shape_grob_fn}
\title{Defunct functions}
\usage{
halmaGrob(...)

kiteGrob(...)

pyramidGrob(...)

convexGrobFn(...)

concaveGrobFn(...)

gridlinesGrob(...)

matGrob(...)

checkersGrob(...)

hexlinesGrob(...)

get_shape_grob_fn(...)
}
\arguments{
\item{...}{Ignored}
}
\description{
These functions are Defunct and have been removed from piecepackr.
}
\details{
\enumerate{
\item{For \code{get_shape_grob_fn} use \code{pp_shape()$shape} instead.}
\item{For \code{gridlinesGrob()} use \code{pp_shape()$gridlines()} instead.}
\item{For \code{matGrob()} use \code{pp_shape()$mat()} instead.}
\item{For \code{checkersGrob()}() use \code{pp_shape()$checkers()} instead.}
\item{For \code{hexlinesGrob()} use \code{pp_shape()$hexlines()} instead.}
\item{For \code{halmaGrob()} use \code{pp_shape("halma")$shape()} instead.}
\item{For \code{kiteGrob()} use \code{pp_shape("kite")$shape()} instead.}
\item{For \code{pyramidGrob()} use \code{pp_shape("pyramid")$shape()} instead.}
\item{For \code{convexGrobFn(n, t)} use \code{pp_shape(paste0("convex", n), t)$shape} instead.}
\item{For \code{concaveGrobFn(n, t, r)} use \code{pp_shape(paste0("concave", n), t, r)$shape} instead.}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aabb_piece.R
\name{aabb_piece}
\alias{aabb_piece}
\title{Calculate axis-aligned bounding box for set of game pieces}
\usage{
aabb_piece(
  df,
  cfg = getOption("piecepackr.cfg", pp_cfg()),
  envir = getOption("piecepackr.envir"),
  op_scale = getOption("piecepackr.op_scale", 0),
  op_angle = getOption("piecepackr.op_angle", 45),
  ...
)
}
\arguments{
\item{df}{A data frame of game piece information with (at least) the
named columns \dQuote{piece_side}, \dQuote{x}, and \dQuote{y}.}

\item{cfg}{Piecepack configuration list or \code{pp_cfg} object,
a list of \code{pp_cfg} objects,
or a character vector referring to names in \code{envir}
or a character vector referring to object names that
can be retrieved by \code{base::dynGet()}.}

\item{envir}{Environment (or named list) containing configuration list(s).}

\item{op_scale}{How much to scale the depth of the piece in the oblique projection
(viewed from the top of the board).
\code{0} (the default) leads to an \dQuote{orthographic} projection,
\code{0.5} is the most common scale used in the \dQuote{cabinet} projection,
and \code{1.0} is the scale used in the \dQuote{cavalier} projection.}

\item{op_angle}{What is the angle of the oblique projection?  Has no effect if \code{op_scale} is \code{0}.}

\item{...}{Ignored}
}
\value{
A named list of ranges with five named elements \code{x}, \code{y}, and \code{z} for
the axis-aligned bounding cube
in xyz-space plus \code{x_op} and \code{y_op} for the axis-aligned bounding box
of the \dQuote{oblique projection} onto the xy plane.
}
\description{
Calculate axis-aligned bounding box (AABB) for set of game pieces
with and without an \dQuote{oblique projection}.
}
\details{
The \dQuote{oblique projection} of a set of \eqn{(x,y,z)} points onto the xy-plane
is \eqn{(x + \lambda * z * cos(\alpha), y + \lambda * z * sin(\alpha))}
where \eqn{\lambda} is the scale factor and \eqn{\alpha} is the angle.
}
\examples{
 df_tiles <- data.frame(piece_side="tile_back", x=0.5+c(3,1,3,1), y=0.5+c(3,3,1,1),
                        suit=NA, angle=NA, z=NA, stringsAsFactors=FALSE)
 df_coins <- data.frame(piece_side="coin_back", x=rep(4:1, 4), y=rep(4:1, each=4),
                        suit=1:16\%\%2+rep(c(1,3), each=8),
                        angle=rep(c(180,0), each=8), z=1/4+1/16, stringsAsFactors=FALSE)
 df <- rbind(df_tiles, df_coins)

 aabb_piece(df, op_scale = 0)
 aabb_piece(df, op_scale = 1, op_angle = 45)
 aabb_piece(df, op_scale = 1, op_angle = -90)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pp_shape.R
\name{pp_shape}
\alias{pp_shape}
\title{Shape object for generating various grobs}
\usage{
pp_shape(
  label = "rect",
  theta = 90,
  radius = 0.2,
  back = FALSE,
  ...,
  width = 1,
  height = 1
)
}
\arguments{
\item{label}{Label of the shape.  One of \describe{
\item{\dQuote{circle}}{Circle.}
\item{\dQuote{convexN}}{An \code{N}-sided convex polygon.
\code{theta} controls which direction the first vertex is drawn.}
\item{\dQuote{concaveN}}{A \dQuote{star} (concave) polygon with \code{N} \dQuote{points}.
\code{theta} controls which direction the first point is drawn.
\code{radius} controls the distance of the \dQuote{inner} vertices from the center.}
\item{\dQuote{halma}}{A 2D outline of a \dQuote{Halma pawn}.}
\item{\dQuote{kite}}{\dQuote{Kite} quadrilateral shape.}
\item{\dQuote{meeple}}{A 2D outline of a \dQuote{meeple}.}
\item{\dQuote{oval}}{Oval.}
\item{\dQuote{pyramid}}{An \dQuote{Isosceles} triangle whose base is the bottom of the viewport.
Typically used to help draw the face of the \dQuote{pyramid} piece.}
\item{\dQuote{rect}}{Rectangle.}
\item{\dQuote{roundrect}}{\dQuote{Rounded} rectangle.  \code{radius} controls curvature of corners.}
}}

\item{theta}{\code{convex} and \code{concave} polygon shapes
use this to determine where the first point is drawn.}

\item{radius}{\code{concave} polygon and \code{roundrect} use this
to control appearance of the shape.}

\item{back}{Whether the shape should be reflected across a vertical line in the middle of the viewport.}

\item{...}{Should be empty}

\item{width, height}{Width and height of the piece.}
}
\description{
\code{pp_shape()} creates an R6 object with methods for creating various shape based grobs.
}
\details{
\code{pp_shape} objects serve the following purposes:\enumerate{
\item{Make it easier for developers to customize game piece appearances
either through a "grob_fn" or "op_grob_fn" styles in \code{pp_cfg()}
or manipulate a piece post drawing via functions like \code{grid::grid.edit()}.}
\item{Used internally to generate \code{piecepackr}'s built-in game piece grobs.}
}
}
\section{\code{pp_shape} R6 Class Method Arguments}{
\describe{
\item{\code{mat_width}}{Numeric vector of mat widths.}
\item{\code{clip}}{\dQuote{clip grob} to perform polyclip operation with.
See \code{\link[gridGeometry:grid.polyclip]{gridGeometry::grid.polyclip()}} for more info.}
\item{\code{op}}{Polyclip operation to perform.
See \code{\link[gridGeometry:grid.polyclip]{gridGeometry::grid.polyclip()}} for more info.}
\item{\code{pattern}}{Pattern to fill in shape with.
See \code{\link[gridpattern:patternGrob]{gridpattern::patternGrob()}} for more info.}
\item{\code{...}}{Passed to \code{gridpattern::patternGrob()}.}
\item{\code{name}}{Grid grob \code{name} value.}
\item{\code{gp}}{Grid \code{gpar} list.  See \code{\link[grid:gpar]{grid::gpar()}} for more info.}
\item{\code{vp}}{Grid viewport or \code{NULL}.}
}
}

\section{\code{pp_shape} R6 Class Methods}{
\describe{
\item{\code{checkers(name = NULL, gp = gpar(), vp = NULL)}}{Returns a grob of checkers for that shape.}
\item{\code{gridlines(name = NULL, gp = gpar(), vp = NULL)}}{Returns a grob of gridlines for that shape.}
\item{\code{hexlines(name = NULL, gp = gpar(), vp = NULL)}}{Returns a grob of hexlines for that shape.}
\item{\code{mat(mat_width = 0, name = NULL, gp = gpar(), vp = NULL)}}{
Returns a grob for a matting \dQuote{mat} for that shape.}
\item{\code{pattern(pattern = "stripe", ..., name = NULL, gp = gpar(), vp = NULL)}}{
Fills in the shape's \code{npc_coords} with a pattern.
See \code{\link[gridpattern:patternGrob]{gridpattern::patternGrob()}} for more information.
}
\item{\code{polyclip(clip, op = "intersection", name = NULL, gp = gpar(), vp = NULL)}}{
Returns a grob that is an \dQuote{intersection}, \dQuote{minus}, \dQuote{union}, or \dQuote{xor} of another grob.
Note unlike \code{gridGeometry::polyclipGrob} it can directly work with a \code{pieceGrob} "clip grob" argument.}
\item{\code{shape(name = NULL, gp = gpar(), vp = NULL)}}{Returns a grob of the shape.}
}
}

\section{\code{pp_shape} R6 Class Active Bindings}{
\describe{
\item{\code{label}}{The shape's label.}
\item{\code{theta}}{The shape's theta.}
\item{\code{radius}}{The shape's radius.}
\item{\code{back}}{A boolean of whether this is the shape's \dQuote{back} side.}
\item{\code{npc_coords}}{A named list of \dQuote{npc} coordinates along the perimeter of the shape.}
}
}

\examples{
 if (require("grid", quietly = TRUE)) {
     gp <- gpar(col="black", fill="yellow")
     rect <- pp_shape(label="rect")
     convex6 <- pp_shape(label="convex6")
     circle <- pp_shape(label="circle")

     pushViewport(viewport(x=0.25, y=0.75, width=1/2, height=1/2))
     grid.draw(rect$shape(gp=gp))
     grid.draw(rect$gridlines(gp=gpar(col="blue", lex=4)))
     grid.draw(rect$hexlines(gp=gpar(col="green")))
     popViewport()

     pushViewport(viewport(x=0.75, y=0.75, width=1/2, height=1/2))
     grid.draw(convex6$shape(gp=gp))
     grid.draw(convex6$checkers(gp=gpar(fill="blue")))
     popViewport()

     pushViewport(viewport(x=0.25, y=0.25, width=1/2, height=1/2))
     grid.draw(circle$shape(gp=gp))
     grid.draw(circle$mat(mat_width=0.2, gp=gpar(fill="blue")))
     popViewport()

     pushViewport(viewport(x=0.75, y=0.25, width=1/2, height=1/2))
     grid.draw(rect$shape(gp=gp))
     grid.draw(rect$mat(mat_width=c(0.2, 0.1, 0.3, 0.4), gp=gpar(fill="blue")))
     popViewport()
 }
 if (require("grid", quietly = TRUE)) {
     grid.newpage()
     gp <- gpar(col="black", fill="yellow")

     vp <- viewport(x=1/4, y=1/4, width=1/2, height=1/2)
     grid.draw(pp_shape("halma")$shape(gp=gp, vp=vp))
     vp <- viewport(x=3/4, y=1/4, width=1/2, height=1/2)
     grid.draw(pp_shape("pyramid")$shape(gp=gp, vp=vp))
     vp <- viewport(x=3/4, y=3/4, width=1/2, height=1/2)
     grid.draw(pp_shape("kite")$shape(gp=gp, vp=vp))
     vp <- viewport(x=1/4, y=3/4, width=1/2, height=1/2)
     grid.draw(pp_shape("meeple")$shape(gp=gp, vp=vp))
 }
 if (require("grid", quietly = TRUE)) {
     grid.newpage()
     vp <- viewport(x=1/4, y=1/4, width=1/2, height=1/2)
     grid.draw(pp_shape("convex3", 0)$shape(gp=gp, vp=vp))
     vp <- viewport(x=3/4, y=1/4, width=1/2, height=1/2)
     grid.draw(pp_shape("convex4", 90)$shape(gp=gp, vp=vp))
     vp <- viewport(x=3/4, y=3/4, width=1/2, height=1/2)
     grid.draw(pp_shape("convex5", 180)$shape(gp=gp, vp=vp))
     vp <- viewport(x=1/4, y=3/4, width=1/2, height=1/2)
     grid.draw(pp_shape("convex6", 270)$shape(gp=gp, vp=vp))
 }
 if (require("grid", quietly = TRUE)) {
     grid.newpage()
     vp <- viewport(x=1/4, y=1/4, width=1/2, height=1/2)
     grid.draw(pp_shape("concave3", 0, 0.1)$shape(gp=gp, vp=vp))
     vp <- viewport(x=3/4, y=1/4, width=1/2, height=1/2)
     grid.draw(pp_shape("concave4", 90, 0.2)$shape(gp=gp, vp=vp))
     vp <- viewport(x=3/4, y=3/4, width=1/2, height=1/2)
     grid.draw(pp_shape("concave5", 180, 0.3)$shape(gp=gp, vp=vp))
     vp <- viewport(x=1/4, y=3/4, width=1/2, height=1/2)
     grid.draw(pp_shape("concave6", 270)$shape(gp=gp, vp=vp))
 }
 if (require("grid", quietly = TRUE) &&
     requireNamespace("gridpattern", quietly = TRUE)) {
     grid.newpage()
     hex <- pp_shape("convex6")
     gp <- gpar(fill = c("blue", "yellow", "red"), col = "black")
     grid.draw(hex$pattern("polygon_tiling", gp = gp, spacing = 0.1,
                           type = "truncated_trihexagonal"))
     gp <- gpar(fill = "black", col = NA)
     grid.draw(hex$mat(mat_width = 0.025, gp = gp))
 }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/basicPieceGrob.R
\name{basicPieceGrobs}
\alias{basicPieceGrobs}
\alias{basicPieceGrob}
\alias{picturePieceGrobFn}
\alias{pyramidTopGrob}
\alias{previewLayoutGrob}
\title{Piece Grob Functions}
\usage{
basicPieceGrob(piece_side, suit, rank, cfg = pp_cfg())

picturePieceGrobFn(directory, filename_fn = find_pp_file)

pyramidTopGrob(piece_side, suit, rank, cfg = pp_cfg())

previewLayoutGrob(piece_side, suit, rank, cfg = pp_cfg())
}
\arguments{
\item{piece_side}{A string with piece and side separated by a underscore e.g. "coin_face"}

\item{suit}{Number of suit (starting from 1).}

\item{rank}{Number of rank (starting from 1)}

\item{cfg}{Piecepack configuration list or \code{\link[=pp_cfg]{pp_cfg()}} object.}

\item{directory}{Directory that \code{picturePieceGrobFn} will look in for piece graphics.}

\item{filename_fn}{Function that takes arguments \code{directory}, \code{piece_side}, \code{suit},
\code{rank}, and optionally \code{cfg} and returns the (full path) filename of the image that the
function returned by \code{picturePieceGrobFn} should import.}
}
\description{
\code{basicPieceGrob()} is the most common
\dQuote{grob} function that \code{\link[=grid.piece]{grid.piece()}} uses
to create \code{grid} graphical \code{grob} objects.
\code{picturePieceGrobFn()} is a function that returns a \dQuote{grob} function
that imports graphics from files found in its \code{directory} argument.
}
\examples{

 if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {
   cfg <- pp_cfg(list(grob_fn.tile=basicPieceGrob, invert_colors=TRUE))
   grid.piece("tile_face", suit=1, rank=3, cfg=cfg)
 }

 \donttest{# May take more than 5 seconds on CRAN servers
 try({
   if (requireNamespace("grid", quietly = TRUE) && capabilities(c("cairo"))) {
     cfg <- pp_cfg(list(grob_fn.tile=basicPieceGrob, invert_colors=TRUE))
     directory <- tempdir()
     save_piece_images(cfg, directory=directory, format="svg", angle=0)
     cfg2 <- pp_cfg(list(grob_fn=picturePieceGrobFn(directory)))

     grid::grid.newpage()
     grid.piece("coin_back", suit=3, rank=5, cfg=cfg2)
   }
 })
 }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cropmarkGrob.R
\name{grid.cropmark}
\alias{grid.cropmark}
\alias{cropmarkGrob}
\title{Draw crop marks with grid}
\usage{
cropmarkGrob(
  ...,
  piece_side = "tile_back",
  suit = NA,
  rank = NA,
  cfg = getOption("piecepackr.cfg", pp_cfg()),
  x = unit(0.5, "npc"),
  y = unit(0.5, "npc"),
  angle = 0,
  width = NA,
  height = NA,
  scale = 1,
  default.units = "npc",
  envir = getOption("piecepackr.envir"),
  name = NULL,
  gp = gpar(),
  vp = NULL,
  bleed = unit(0.125, "in"),
  cm_select = "12345678",
  cm_width = unit(0.25, "mm"),
  cm_length = unit(0.125, "in")
)

grid.cropmark(..., draw = TRUE)
}
\arguments{
\item{...}{\code{cropmarkGrob()} ignores; \code{grid.cropmark()} passes to \code{cropmarkGrob()}.}

\item{piece_side}{A string with piece and side separated by a underscore e.g. "coin_face"}

\item{suit}{Number of suit (starting from 1).}

\item{rank}{Number of rank (starting from 1)}

\item{cfg}{Piecepack configuration list or \code{pp_cfg} object,
a list of \code{pp_cfg} objects,
or a character vector referring to names in \code{envir}
or a character vector referring to object names that
can be retrieved by \code{base::dynGet()}.}

\item{x}{Where to place piece on x axis of viewport}

\item{y}{Where to place piece on y axis of viewport}

\item{angle}{Angle (on xy plane) to draw piece at}

\item{width}{Width of piece}

\item{height}{Height of piece}

\item{scale}{Multiplicative scaling factor to apply to width, height, and depth.}

\item{default.units}{A string indicating the default units to use if
'x', 'y', 'width', and/or 'height' are only given as numeric vectors.}

\item{envir}{Environment (or named list) containing configuration list(s).}

\item{name}{A character identifier (for grid)}

\item{gp}{An object of class \dQuote{gpar}.}

\item{vp}{A \code{grid} viewport object (or \code{NULL}).}

\item{bleed}{Bleed zone size to assume:
\itemize{
\item{If \code{bleed} is a \code{\link[grid:unit]{grid::unit()}} simply use it}
\item{If \code{bleed} is numeric then convert via \code{grid::unit(bleed, default.units)}}
\item{If \code{bleed} is \code{TRUE} assume 1/8 inch bleed zone size}
\item{If \code{bleed} is \code{FALSE} assume 0 inch bleed zone size}
}}

\item{cm_select}{A string of integers from "1" to "8" indicating which
crop marks to draw.  "1" represents the top right crop mark
then we proceeding clockwise to "8" which represents the
top left crop mark.
Default "12345678" draws all eight crop marks.}

\item{cm_width}{Width of crop mark.}

\item{cm_length}{Length of crop mark.}

\item{draw}{A logical value indicating whether graphics output should be produced.}
}
\value{
A grid grob.
}
\description{
\code{grid.cropmark()} draws \dQuote{crop marks} to the active graphics device.
\code{cropmarkGrob()} is its grid grob counterpart.
Intended for use in adding crop marks around
game pieces in print-and-play layouts.
}
\examples{
if (requireNamespace("grid", quietly = TRUE) &&
    piecepackr:::device_supports_unicode()) {
  cfg <- pp_cfg(list(mat_color = "pink", mat_width=0.05, border_color=NA))
  grid::grid.newpage()
  df <- data.frame(piece_side = "tile_face", suit = 2, rank = 2,
                   x = 2, y = 2, angle = 0,
                   stringsAsFactors = FALSE)
  pmap_piece(df, grid.cropmark, cfg = cfg, default.units = "in")
  pmap_piece(df, grid.piece, cfg = cfg, default.units = "in", bleed=TRUE)
}
if (requireNamespace("grid", quietly = TRUE) &&
    piecepackr:::device_supports_unicode()) {
  grid::grid.newpage()
  df <- data.frame(piece_side = "coin_back", suit = 2, rank = 2,
                   x = 2, y = 2, angle = 0,
                   stringsAsFactors = FALSE)
  pmap_piece(df, grid.cropmark, cfg = cfg, default.units = "in", bleed=TRUE)
  pmap_piece(df, grid.piece, cfg = cfg, default.units = "in", bleed=TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pmap_piece.R
\name{pmap_piece}
\alias{pmap_piece}
\title{Create graphics using data frame input}
\usage{
pmap_piece(
  .l,
  .f = pieceGrob,
  ...,
  cfg = getOption("piecepackr.cfg"),
  envir = getOption("piecepackr.envir"),
  trans = getOption("piecepackr.trans"),
  draw = TRUE,
  name = NULL,
  gp = NULL,
  vp = NULL
)
}
\arguments{
\item{.l}{A list of vectors, such as a data frame. The length of \code{.l}
determines the number of arguments that \code{.f}
will be called  with. List names will be used if present.}

\item{.f}{Function to be applied to \code{.l} after adjustments to
\code{cfg} and \code{envir} and the application of \code{trans}.
Usually \code{\link[=grid.piece]{grid.piece()}}, \code{\link[=pieceGrob]{pieceGrob()}}, \code{\link[=piece3d]{piece3d()}}, or \code{\link[=piece]{piece()}}.}

\item{...}{Extra arguments to pass to \code{.f}.}

\item{cfg}{Piecepack configuration list or \code{pp_cfg} object,
a list of \code{pp_cfg} objects,
or a character vector referring to names in \code{envir}
or a character vector referring to object names that
can be retrieved by \code{base::dynGet()}.}

\item{envir}{Environment (or named list) containing configuration list(s).}

\item{trans}{Function to modify \code{.l} before drawing.
Default (\code{NULL}) is to not modify \code{.l}.  \code{op_transform}
can help with using an oblique projection (i.e. \code{op_scale} over 0).}

\item{draw}{A logical value indicating whether graphics output should be produced.}

\item{name}{A character identifier (for grid)}

\item{gp}{An object of class \dQuote{gpar}.}

\item{vp}{A \code{grid} viewport object (or \code{NULL}).}
}
\description{
\code{pmap_piece()} operates on the rows of a data frame
applying \code{.f} to each row (usually \code{grid.piece}).
}
\details{
\code{pmap_piece()} differs from \code{purrr::pmap()} in a few ways:
\enumerate{
\item{If \code{cfg} and/or \code{envir} are missing attempts to set reasonable defaults.}
\item{If not \code{NULL} will first apply function \code{trans} to \code{.l}.}
\item{If the output of \code{.f} is a grid grob object then \code{pmap_piece}
will return a \code{gTree} object with
specified \code{name}, \code{gp}, and \code{vp} values and if \code{draw} is true draw it.}
\item{If \code{.l} lacks a \code{name} column or if \code{name} column is non-unique
attempts to generate a reasonable new default \code{name} column
and use that to name the return \code{gTree} children  or \code{list} values.}
}
}
\examples{
  if (requireNamespace("grid", quietly = TRUE) && piecepackr:::device_supports_unicode()) {
       dark_colorscheme <- list(suit_color="darkred,black,darkgreen,darkblue,black",
                            invert_colors.suited=TRUE, border_color="black", border_lex=2)
       traditional_ranks <- list(use_suit_as_ace=TRUE, rank_text=",a,2,3,4,5")
       cfg3d <- list(width.pawn=0.75, height.pawn=0.75, depth.pawn=1,
                          dm_text.pawn="", shape.pawn="convex6", invert_colors.pawn=TRUE,
                          edge_color.coin="tan", edge_color.tile="tan")
       cfg <- pp_cfg(c(dark_colorscheme, traditional_ranks, cfg3d))
       grid::grid.newpage()
       df_tiles <- data.frame(piece_side="tile_back", x=0.5+c(3,1,3,1), y=0.5+c(3,3,1,1),
                              suit=NA, angle=NA, z=NA, stringsAsFactors=FALSE)
       df_coins <- data.frame(piece_side="coin_back", x=rep(4:1, 4), y=rep(4:1, each=4),
                              suit=1:16\%\%2+rep(c(1,3), each=8),
                              angle=rep(c(180,0), each=8), z=1/4+1/16, stringsAsFactors=FALSE)
       df <- rbind(df_tiles, df_coins)
       pmap_piece(df, cfg=cfg, op_scale=0.5, default.units="in")
  }
}
\seealso{
\code{\link[=render_piece]{render_piece()}} is a higher-level function that wraps this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/op_transform.R
\name{op_transform}
\alias{op_transform}
\alias{marbles_transform}
\title{Oblique projection helper function}
\usage{
op_transform(
  df,
  ...,
  cfg = getOption("piecepackr.cfg", pp_cfg()),
  envir = getOption("piecepackr.envir"),
  op_angle = getOption("piecepackr.op_angle", 45),
  pt_thickness = 0.01,
  as_top = character(0),
  cfg_class = "list"
)

marbles_transform(df, ...)
}
\arguments{
\item{df}{A data frame with coordinates and dimensions in inches}

\item{...}{Ignored by \code{op_transform()}.  \code{marbles_transform()} passes to \code{op_transform()}.}

\item{cfg}{Piecepack configuration list or \code{\link[=pp_cfg]{pp_cfg()}} object,
a list of \code{\link[=pp_cfg]{pp_cfg()}} objects,
or a character vector of the names of \code{\link[=pp_cfg]{pp_cfg()}} objects contained in \code{envir}.}

\item{envir}{Environment (or named list) containing configuration list(s).}

\item{op_angle}{Intended oblique projection angle (used for re-sorting)}

\item{pt_thickness}{Thickness of pyramid tip i.e. value to add to the z-value of a pyramid top
if it is a (weakly) smaller ranked pyramid (top)
placed on top of a larger ranked pyramid (top).}

\item{as_top}{Character vector of components whose \dQuote{side}
should be converted to \dQuote{top} e.g. \code{"pawn_face"}.}

\item{cfg_class}{Either \code{"list"} (default) or \code{"character"}.
Desired class of the \code{cfg} column in the returned tibble.
\code{"list"} is more efficient for use with \code{pmap_piece()} but
\code{geom_piece()} needs \code{"character"}.}
}
\value{
A tibble with extra columns added
and re-sorted rows
}
\description{
\code{op_transform()} guesses \code{z} coordinates and
sorting order to more easily
make 3D graphics with \code{\link[=pmap_piece]{pmap_piece()}}.
\code{marbles_transform()} is a wrapper around \code{op_transform()} that handles the special case
of stacking 1" marbles in a pyramid with a square base on a holed board.
}
\details{
The heuristics used to generate guesses
for \code{z} coordinates and sorting order
aren't guaranteed to work in every case.
In some cases you may get better sorting results
by changing the \code{op_angle} or the dimensions of pieces.
}
\examples{
df <- tibble::tibble(piece_side="tile_back",
                     x=c(2,2,2,4,6,6,4,2,5),
                     y=c(4,4,4,4,4,2,2,2,3))
cfg <- game_systems()$piecepack
pmap_piece(df, op_angle=135, trans=op_transform,
           op_scale=0.5, default.units="in", cfg=cfg)
}
\seealso{
\url{https://trevorldavis.com/piecepackr/3d-projections.html} for more details
and examples of oblique projections in \code{piecepackr}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pp_cfg.R
\name{pp_cfg}
\alias{pp_cfg}
\alias{is_pp_cfg}
\alias{as_pp_cfg}
\title{Configuration list R6 object}
\usage{
pp_cfg(cfg = list())

is_pp_cfg(cfg)

as_pp_cfg(cfg = list())
}
\arguments{
\item{cfg}{List of configuration options}
}
\description{
\code{pp_cfg()} and \code{as_pp_cfg()} create piecepack configuration list R6 objects.
\code{is_pp_cfg()} returns \code{TRUE} if object is a piecepack configuration list R6 object.
\code{as.list()} will convert it into a list.
}
\details{
\code{pp_cfg} R6 class objects serve the following purposes:
\itemize{
\item Customize the appearance of pieces drawn by \code{grid.piece()}.
\item Speed up the drawing of graphics through use of caching.
\item Allow the setting and querying of information about the board game components
that maybe of use to developers:
\itemize{
\item Number of suits
\item Number of ranks
\item Suit colors
\item Which types of components are included and/or properly supported
\item What would be a good color to use when adding annotations on top of these components.
\item Title, Description, Copyright, License, and Credit metadata
}
}
}
\section{\code{pp_cfg} R6 Class Method Arguments}{
\describe{
\item{\code{piece_side}}{A string with piece and side separated by a underscore e.g. "coin_face".}
\item{\code{suit}}{Number of suit (starting from 1).}
\item{\code{rank}}{Number of rank (starting from 1).}
\item{\code{type}}{Which type of grob to return, either \code{"normal"}, \code{"picture"}, \code{"raster"}, or \code{"transformation"}.}
\item{\code{scale}}{"scale" factor}
\item{\code{alpha}}{"alpha" value}
}
}

\section{\code{pp_cfg} R6 Class Methods}{
\describe{
\item{\code{get_grob()}}{Returns a \code{grid} \dQuote{grob} for drawing the piece.}
\item{\code{get_piece_opt()}}{Returns a list with info useful for drawing the piece.}
\item{\code{get_suit_color()}}{Returns the suit colors.}
\item{\code{get_width()}, \code{get_height()}, \code{get_depth()}}{
Dimensions (of the bounding cube) of the piece in inches}
}
}

\section{\code{pp_cfg} R6 Class Fields and Active Bindings}{
\describe{
\item{\code{annotation_color}}{Suggestion of a good color to annotate with}
\item{\code{cache}}{Cache object which stores intermediate graphical calculations.
Default is a memory-cache that does not prune.
This can be replaced by another cache that
implements the cache API used by the \code{cachem} package}
\item{\code{cache_grob}}{Whether we should cache (2D) grobs}
\item{\code{cache_grob_with_bleed_fn}}{Whether we should cache the grob with bleed functions}
\item{\code{cache_piece_opt}}{Whether we should cache piece opt information}
\item{\code{cache_op_fn}}{Whether we should cache the oblique projection functions}
\item{\code{cache_obj_fn}}{Whether we should cache any 3D rendering functions}
\item{\code{copyright}}{Design copyright information}
\item{\code{credit}}{Design credits}
\item{\code{description}}{Design description}
\item{\code{fontfamily}}{Main font family}
\item{\code{has_bits}}{Whether we should assume this supports "bit" pieces}
\item{\code{has_boards}}{Whether we should assume this supports "board" pieces}
\item{\code{has_cards}}{Whether we should assume this supports "card" pieces}
\item{\code{has_coins}}{Whether we should assume this supports "coin" pieces}
\item{\code{has_dice}}{Whether we should assume this supports "die" pieces}
\item{\code{has_matchsticks}}{Whether we should assume this supports "matchstick" pieces}
\item{\code{has_pawns}}{Whether we should assume this supports "pawn" pieces}
\item{\code{has_piecepack}}{Binding which simultaneously checks/sets
\code{has_coins}, \code{has_tiles}, \code{has_pawns}, \code{has_dice}}
\item{\code{has_pyramids}}{Whether we should assume this supports "pyramid" pieces}
\item{\code{has_saucers}}{Whether we should assume this supports "saucer" pieces}
\item{\code{has_tiles}}{Whether we should assume this supports "tile" pieces}
\item{\code{spdx_id}}{SPDX Identifier for graphical design license.
See \url{https://spdx.org/licenses/} for full list.}
\item{\code{title}}{Design title}
}
}

\section{Defunct \code{pp_cfg} R6 Class attributes which have been removed}{
\describe{
\item{\code{cache_shadow}}{Use \code{cache_op_fn} instead}
\item{\code{i_unsuit}}{Instead add \code{1L} to \code{n_suits}}
\item{\code{get_pictureGrob()}}{Use \code{get_grob(..., type = "picture")} instead}
\item{\code{get_shadow_fn}}{\code{get_op_grob()} returns complete oblique projection grob}
}
}

\examples{
 cfg <- pp_cfg(list(invert_colors=TRUE))
 as.list(cfg)
 is_pp_cfg(cfg)
 as_pp_cfg(list(suit_color="darkred,black,darkgreen,darkblue,grey"))
 cfg$get_suit_color(suit=3)
 cfg$annotation_color
 cfg$has_matchsticks
 cfg$has_matchsticks <- TRUE
 cfg$has_matchsticks
 cfg$get_width("tile_back")
 cfg$get_height("die_face")
 cfg$get_depth("coin_face")
 \donttest{# May take more than 5 seconds on CRAN servers
 # `pp_cfg()` objects use a cache to speed up repeated drawing
 pdf(tempfile(fileext = ".pdf"))
 cfg <- list()
 system.time(replicate(100, grid.piece("tile_back", 4, 4, cfg)))
 cfg <- pp_cfg(list())
 system.time(replicate(100, grid.piece("tile_back", 4, 4, cfg)))
 invisible(dev.off())
 }

}
\seealso{
\code{\link[=game_systems]{game_systems()}} for functions that return configuration list
objects for several game systems.
\url{https://trevorldavis.com/piecepackr/configuration-lists.html} for more details
about \code{piecepackr} configuration lists.
}
